// Copyright 2022 xensik. All rights reserved.
//
// Use of this source code is governed by a GNU GPLv3 license
// that can be found in the LICENSE file.

#include "stdafx.hpp"
#include "iw6.hpp"

#ifdef _MSC_VER
#pragma warning(push)
#pragma warning(disable:4244)
#endif

namespace xsk::gsc::iw6
{

std::unordered_map<std::uint8_t, std::string_view> opcode_map;
std::unordered_map<std::uint16_t, std::string_view> function_map;
std::unordered_map<std::uint16_t, std::string_view> method_map;
std::unordered_map<std::uint16_t, std::string_view> file_map;
std::unordered_map<std::uint16_t, std::string_view> token_map;
std::unordered_map<std::string_view, std::uint8_t> opcode_map_rev;
std::unordered_map<std::string_view, std::uint16_t> function_map_rev;
std::unordered_map<std::string_view, std::uint16_t> method_map_rev;
std::unordered_map<std::string_view, std::uint16_t> file_map_rev;
std::unordered_map<std::string_view, std::uint16_t> token_map_rev;
std::unordered_map<std::string, std::vector<std::uint8_t>> files;
read_cb_type read_callback = nullptr;
std::set<std::string> string_map;

void resolver::init(read_cb_type callback)
{
    read_callback = callback;
}

void resolver::cleanup()
{
    files.clear();
}

auto resolver::opcode_id(const std::string& name) -> std::uint8_t
{
    const auto itr = opcode_map_rev.find(name);

    if (itr != opcode_map_rev.end())
    {
        return itr->second;
    }

    throw error(utils::string::va("Couldn't resolve opcode id for name '%s'!", name.data()));
}

auto resolver::opcode_name(std::uint8_t id) -> std::string
{
    const auto itr = opcode_map.find(id);

    if (itr != opcode_map.end())
    {
        return std::string(itr->second);
    }

    throw error(utils::string::va("Couldn't resolve opcode name for id '0x%hhX'!", id));
}

auto resolver::function_id(const std::string& name) -> std::uint16_t
{
    if (name.starts_with("_func_"))
    {
        return static_cast<std::uint16_t>(std::stoul(name.substr(6), nullptr, 16));
    }

    const auto itr = function_map_rev.find(name);

    if (itr != function_map_rev.end())
    {
        return itr->second;
    }

    throw error(utils::string::va("Couldn't resolve builtin function id for name '%s'!", name.data()));
}

auto resolver::function_name(std::uint16_t id) -> std::string
{
    const auto itr = function_map.find(id);

    if (itr != function_map.end())
    {
        return std::string(itr->second);
    }

    return utils::string::va("_func_%04X", id);
}

auto resolver::method_id(const std::string& name) -> std::uint16_t
{
    if (name.starts_with("_meth_"))
    {
        return static_cast<std::uint16_t>(std::stoul(name.substr(6), nullptr, 16));
    }

    const auto itr = method_map_rev.find(name);

    if (itr != method_map_rev.end())
    {
        return itr->second;
    }

    throw error(utils::string::va("Couldn't resolve builtin method id for name '%s'!", name.data()));
}

auto resolver::method_name(std::uint16_t id) -> std::string
{
    const auto itr = method_map.find(id);

    if (itr != method_map.end())
    {
        return std::string(itr->second);
    }

    return utils::string::va("_meth_%04X", id);
}

auto resolver::file_id(const std::string& name) -> std::uint16_t
{
    if (name.starts_with("_id_"))
    {
        return static_cast<std::uint16_t>(std::stoul(name.substr(4), nullptr, 16));
    }

    const auto itr = file_map_rev.find(name);

    if (itr != file_map_rev.end())
    {
        return itr->second;
    }

    return 0;
}

auto resolver::file_name(std::uint16_t id) -> std::string
{
    const auto itr = file_map.find(id);

    if (itr != file_map.end())
    {
        return std::string(itr->second);
    }

    return utils::string::va("_id_%04X", id);
}

auto resolver::token_id(const std::string& name) -> std::uint16_t
{
    if (name.starts_with("_id_"))
    {
        return static_cast<std::uint16_t>(std::stoul(name.substr(4), nullptr, 16));
    }

    const auto itr = token_map_rev.find(name);

    if (itr != token_map_rev.end())
    {
        return itr->second;
    }

    return 0;
}

auto resolver::token_name(std::uint16_t id) -> std::string
{
    const auto itr = token_map.find(id);

    if (itr != token_map.end())
    {
        return std::string(itr->second);
    }

    return utils::string::va("_id_%04X", id);
}

auto resolver::find_function(const std::string& name) -> bool
{
    if (name.starts_with("_func_")) return true;

    const auto itr = function_map_rev.find(name);

    if (itr != function_map_rev.end())
    {
        return true;
    }

    return false;
}

auto resolver::find_method(const std::string& name) -> bool
{
    if (name.starts_with("_meth_")) return true;

    const auto itr = method_map_rev.find(name);

    if (itr != method_map_rev.end())
    {
        return true;
    }

    return false;
}

auto resolver::make_token(std::string_view str) -> std::string
{
    if (str.starts_with("_id_") || str.starts_with("_func_") || str.starts_with("_meth_"))
    {
        return std::string(str);
    }

    auto data = std::string(str.begin(), str.end());

    for (std::size_t i = 0; i < data.size(); i++)
    {
        data[i] = static_cast<char>(std::tolower(static_cast<unsigned char>(str[i])));
        if (data[i] == '\\') data[i] = '/';
    }

    return data;
}

auto resolver::file_data(const std::string& name) -> std::tuple<const std::string*, char*, size_t>
{
    const auto& itr = files.find(name);

    if (itr != files.end())
    {
        return { &itr->first ,reinterpret_cast<char*>(itr->second.data()), itr->second.size() };
    }

    auto data = read_callback(name);

    const auto& res = files.insert({ name, std::move(data)});

    if (res.second)
    {
        return { &res.first->first, reinterpret_cast<char*>(res.first->second.data()), res.first->second.size() };
    }

    throw error("couldn't open gsc file '" + name + "'");
}

std::set<std::string_view> paths
{
    "character"sv,
    "codescripts"sv,
    "common_scripts"sv,
    "destructible_scripts"sv,
    "maps"sv,
    "vehicle_scripts"sv,
    "xmodelalias"sv,
    "maps/animated_models"sv,
    "maps/createart"sv,
    "maps/createfx"sv,
    "maps/interactive_models"sv,
    "maps/mp"sv,
    "maps/mp/agents"sv,
    "maps/mp/alien"sv,
    "maps/mp/bots"sv,
    "maps/mp/gametypes"sv,
    "maps/mp/killstreaks"sv,
    "maps/mp/perks"sv,
    "maps/mp/agents/alien"sv,
    "maps/mp/agents/dog"sv,
    "maps/mp/agents/alien/alien_ancestor"sv,
    "maps/mp/agents/alien/alien_kraken"sv,
    "maps/mp/agents/alien/alien_spider"sv,
};

auto resolver::fs_to_game_path(const std::filesystem::path& file) -> std::filesystem::path
{
    auto result = std::filesystem::path();
    auto root = false;

    for (auto& entry : file)
    {
        if (!root && paths.contains(entry.string()))
        {
            result = entry;
            root = true;
        }
        else if (paths.contains(result.string()))
        {
            result /= entry;
        }
    }

    return result.empty() ? file : result;
}

const std::array<std::pair<std::uint8_t, const char*>, 153> opcode_list
{{
    { 0x17, "SET_NEW_LOCAL_VARIABLE_FIELD_CACHED0" },
    { 0x18, "EVAL_SELF_FIELD_VARIABLE" },
    { 0x19, "RETN" },
    { 0x1A, "CALL_BUILTIN_FUNC_0" },
    { 0x1B, "CALL_BUILTIN_FUNC_1" },
    { 0x1C, "CALL_BUILTIN_FUNC_2" },
    { 0x1D, "CALL_BUILTIN_FUNC_3" },
    { 0x1E, "CALL_BUILTIN_FUNC_4" },
    { 0x1F, "CALL_BUILTIN_FUNC_5" },
    { 0x20, "CALL_BUILTIN_FUNC" },
    { 0x21, "BOOL_NOT" },
    { 0x22, "CALL_FAR_METHOD_THEAD" },
    { 0x23, "JMP_EXPR_TRUE" },
    { 0x24, "SET_LEVEL_FIELD_VARIABLE_FIELD" },
    { 0x25, "CAST_BOOL" },
    { 0x26, "EVAL_NEW_LOCAL_ARRAY_REF_CACHED0" },
    { 0x27, "CALL_BUILTIN_FUNC_POINTER" },
    { 0x28, "INEQUALITY" },
    { 0x29, "GET_THISTHREAD" },
    { 0x2A, "CLEAR_FIELD_VARIABLE" },
    { 0x2B, "GET_FLOAT" },
    { 0x2C, "SAFE_CREATE_VARIABLE_FIELD_CACHED" },
    { 0x2D, "CALL_FAR_FUNC2" },
    { 0x2E, "CALL_FAR_FUNC" },
    { 0x2F, "CALL_FAR_FUNC_CHILD_THREAD" },
    { 0x30, "CLEAR_LOCAL_VARIABLE_FIELD_CACHED0" },
    { 0x31, "CLEAR_LOCAL_VARIABLE_FIELD_CACHED" },
    { 0x32, "CHECK_CLEAR_PARAMS" },
    { 0x33, "CAST_FIELD_OBJ" },
    { 0x34, "END" },
    { 0x35, "SIZE" },
    { 0x36, "EMPTY_ARRAY" },
    { 0x37, "BIT_AND" },
    { 0x38, "LESSEQUAL" },
    { 0x39, "VOIDCODEPOS" },
    { 0x3A, "CALL_METHOD_THREAD_POINTER" },
    { 0x3B, "ENDSWITCH" },
    { 0x3C, "CLEAR_VARIABLE_FIELD" },
    { 0x3D, "DIV" },
    { 0x3E, "CALL_FAR_METHOD_CHILD_THEAD" },
    { 0x3F, "GET_USHORT" },
    { 0x40, "JMP_TRUE" },
    { 0x41, "GET_SELF" },
    { 0x42, "CALL_FAR_FUNC_THREAD" },
    { 0x43, "CALL_LOCAL_FUNC_THREAD" },
    { 0x44, "SET_LOCAL_VARIABLE_FIELD_CACHED0" },
    { 0x45, "SET_LOCAL_VARIABLE_FIELD_CACHED" },
    { 0x46, "PLUS" },
    { 0x47, "BOOL_COMPLEMENT" },
    { 0x48, "CALL_METHOD_POINTER" },
    { 0x49, "INC" },
    { 0x4A, "REMOVE_LOCAL_VARIABLES" },
    { 0x4B, "JMP_EXPR_FALSE" },
    { 0x4C, "SWITCH" },
    { 0x4D, "CLEAR_PARAMS" },
    { 0x4E, "EVAL_LOCAL_VARIABLE_REF_CACHED0" },
    { 0x4F, "EVAL_LOCAL_VARIABLE_REF_CACHED" },
    { 0x50, "CALL_LOCAL_METHOD" },
    { 0x51, "EVAL_FIELD_VARIABLE" },
    { 0x52, "EVAL_FIELD_VARIABLE_REF" },
    { 0x53, "GET_STRING" },
    { 0x54, "CALL_FUNC_POINTER" },
    { 0x55, "EVAL_LEVEL_FIELD_VARIABLE" },
    { 0x56, "GET_VECTOR" },
    { 0x57, "ENDON" },
    { 0x58, "GREATEREQUAL" },
    { 0x59, "GET_SELF_OBJ" },
    { 0x5A, "SET_ANIM_FIELD_VARIABLE_FIELD" },
    { 0x5B, "SET_VARIABLE_FIELD" },
    { 0x5C, "CALL_LOCAL_FUNC2" },
    { 0x5D, "CALL_LOCAL_FUNC" },
    { 0x5E, "EVAL_LOCAL_ARRAY_REF_CACHED0" },
    { 0x5F, "EVAL_LOCAL_ARRAY_REF_CACHED" },
    { 0x60, "GET_FAR_FUNC" },
    { 0x61, "LESS" },
    { 0x62, "GET_GAME_REF" },
    { 0x63, "WAITTILLFRAMEEND" },
    { 0x64, "SAFE_SET_VARIABLE_FIELD_CACHED0" },
    { 0x65, "SAFE_SET_VARIABLE_FIELD_CACHED" },
    { 0x66, "CALL_METHOD_CHILD_THREAD_POINTER" },
    { 0x67, "GET_LEVEL" },
    { 0x68, "NOTIFY" },
    { 0x69, "DEC_TOP" },
    { 0x6A, "SHIFT_LEFT" },
    { 0x6B, "CALL_LOCAL_METHOD_THREAD" },
    { 0x6C, "CALL_LOCAL_METHOD_CHILD_THREAD" },
    { 0x6D, "GREATER" },
    { 0x6E, "EVAL_LOCAL_VARIABLE_CACHED0" },
    { 0x6F, "EVAL_LOCAL_VARIABLE_CACHED1" },
    { 0x70, "EVAL_LOCAL_VARIABLE_CACHED2" },
    { 0x71, "EVAL_LOCAL_VARIABLE_CACHED3" },
    { 0x72, "EVAL_LOCAL_VARIABLE_CACHED4" },
    { 0x73, "EVAL_LOCAL_VARIABLE_CACHED5" },
    { 0x74, "EVAL_LOCAL_VARIABLE_CACHED" },
    { 0x75, "SAFE_SET_WAITTILL_VARIABLE_FIELD_CACHED" },
    { 0x76, "JMP" },
    { 0x77, "CALL_FUNC_THREAD_POINTER" },
    { 0x78, "GET_ZERO" },
    { 0x79, "WAIT" },
    { 0x7A, "MINUS" },
    { 0x7B, "SET_SELF_FIELD_VARIABLE_FIELD" },
    { 0x7C, "EVAL_NEW_LOCAL_VARIABLE_REF_CACHED0" },
    { 0x7D, "MULT" },
    { 0x7E, "CREATE_LOCAL_VARIABLE" },
    { 0x7F, "CALL_LOCAL_FUNC_CHILD_THREAD" },
    { 0x80, "GET_INT" },
    { 0x81, "MOD" },
    { 0x82, "EVAL_ANIM_FIELD_VARIABLE_REF" },
    { 0x83, "GET_BUILTIN_FUNC" },
    { 0x84, "GET_GAME" },
    { 0x85, "WAITTILL" },
    { 0x86, "DEC" },
    { 0x87, "EVAL_LOCAL_VARIABLE_OBJECT_CACHED" },
    { 0x88, "PRE_CALL" },
    { 0x89, "GET_ANIM" },
    { 0x8A, "GET_UNDEFINED" },
    { 0x8B, "EVAL_LEVEL_FIELD_VARIABLE_REF" },
    { 0x8C, "GET_ANIM_OBJ" },
    { 0x8D, "GET_LEVEL_OBJ" },
    { 0x8E, "BIT_EXOR" },
    { 0x8F, "EQUALITY" },
    { 0x90, "CLEAR_ARRAY" },
    { 0x91, "JMP_BACK" },
    { 0x92, "GET_ANIMATION" },
    { 0x93, "EVAL_ANIM_FIELD_VARIABLE" },
    { 0x94, "GET_ANIMTREE" },
    { 0x95, "GET_ISTRING" },
    { 0x96, "EVAL_ARRAY_REF" },
    { 0x97, "EVAL_SELF_FIELD_VARIABLE_REF" },
    { 0x98, "GET_NBYTE" },
    { 0x99, "GET_BUILTIN_METHOD" },
    { 0x9A, "CALL_BUILTIN_METHOD_POINTER" },
    { 0x9B, "EVAL_ARRAY" },
    { 0x9C, "VECTOR" },
    { 0x9D, "CALL_FAR_METHOD" },
    { 0x9E, "EVAL_LOCAL_ARRAY_CACHED" },
    { 0x9F, "GET_BYTE" },
    { 0xA0, "CALL_FUNC_CHILD_THREAD_POINTER" },
    { 0xA1, "BIT_OR" },
    { 0xA2, "ADD_ARRAY" },
    { 0xA3, "WAITTILLMATCH2" },
    { 0xA4, "WAITTILLMATCH" },
    { 0xA5, "GET_LOCAL_FUNC" },
    { 0xA6, "GET_NUSHORT" },
    { 0xA7, "SHIFT_RIGHT" },
    { 0xA8, "CALL_BUILTIN_METHOD_0" },
    { 0xA9, "CALL_BUILTIN_METHOD_1" },
    { 0xAA, "CALL_BUILTIN_METHOD_2" },
    { 0xAB, "CALL_BUILTIN_METHOD_3" },
    { 0xAC, "CALL_BUILTIN_METHOD_4" },
    { 0xAD, "CALL_BUILTIN_METHOD_5" },
    { 0xAE, "CALL_BUILTIN_METHOD" },
    { 0xAF, "JMP_FALSE" },
}};

const std::array<std::pair<std::uint16_t, const char*>, 605> function_list
{{
    { 0x001, "precacheturret" },
    { 0x002, "getweaponarray" },
    { 0x003, "createprintchannel" },
    { 0x004, "updategamerprofileall" },
    { 0x005, "clearlocalizedstrings" },
    { 0x006, "setphysicsgravitydir" },
    { 0x007, "gettimescale" },
    { 0x008, "settimescale" },
    { 0x009, "setslowmotionview" },
    { 0x00A, "teleportscene" },
    { 0x00B, "forcesharedammo" },
    { 0x00C, "refreshhudcompass" },
    { 0x00D, "refreshhudammocounter" },
    { 0x00E, "notifyoncommand" },
    { 0x00F, "setprintchannel" },
    { 0x010, "print" },
    { 0x011, "println" },
    { 0x012, "print3d" },
    { 0x013, "line" },
    { 0x014, "box" },
    { 0x015, "orientedbox" },
    { 0x016, "sphere" },
    { 0x017, "cylinder" },
    { 0x018, "spawnturret" },
    { 0x019, "canspawnturret" },
    { 0x01A, "assert" },
    { 0x01B, "pausecinematicingame" },
    { 0x01C, "drawcompassfriendlies" },
    { 0x01D, "bulletspread" },
    { 0x01E, "bullettracer" },
    { 0x01F, "badplace_delete" },
    { 0x020, "badplace_cylinder" },
    { 0x021, "badplace_arc" },
    { 0x022, "badplace_brush" },
    { 0x023, "clearallcorpses" },
    { 0x024, "setturretnode" },
    { 0x025, "unsetturretnode" },
    { 0x026, "setnodepriority" },
    { 0x027, "isnodeoccupied" },
    { 0x028, "setdebugorigin" },
    { 0x029, "setdebugangles" },
    { 0x02A, "updategamerprofile" },
    { 0x02B, "assertex" },
    { 0x02C, "assertmsg" },
    { 0x02D, "isdefined" },
    { 0x02E, "isvalidmissile" },
    { 0x02F, "isstring" },
    { 0x030, "setomnvar" },
    { 0x031, "getomnvar" },
    { 0x032, "setdvar" },
    { 0x033, "setdynamicdvar" },
    { 0x034, "setdvarifuninitialized" },
    { 0x035, "setdevdvar" },
    { 0x036, "setdevdvarifuninitialized" },
    { 0x037, "getdvar" },
    { 0x038, "getdvarint" },
    { 0x039, "getdvarfloat" },
    { 0x03A, "getdvarvector" },
    { 0x03B, "gettime" },
    { 0x03C, "getentbynum" },
    { 0x03D, "getweaponmodel" },
    { 0x03E, "getculldist" },
    { 0x03F, "sethalfresparticles" },
    { 0x040, "getmapsunlight" },
    { 0x041, "setsunlight" },
    { 0x042, "resetsunlight" },
    { 0x043, "getmapsundirection" },
    { 0x044, "getmapsunangles" },
    { 0x045, "setsundirection" },
    { 0x046, "lerpsundirection" },
    { 0x047, "lerpsunangles" },
    { 0x048, "resetsundirection" },
    { 0x049, "enableforcedsunshadows" },
    { 0x04A, "enableforcednosunshadows" },
    { 0x04B, "disableforcedsunshadows" },
    { 0x04C, "enableouterspacemodellighting" },
    { 0x04D, "disableouterspacemodellighting" },
    { 0x04E, "remapstage" },
    { 0x04F, "changelevel" },
    { 0x050, "missionsuccess" },
    { 0x051, "missionfailed" },
    { 0x052, "cinematic" },
    { 0x053, "cinematicingame" },
    { 0x054, "cinematicingamesync" },
    { 0x055, "cinematicingameloop" },
    { 0x056, "cinematicingameloopresident" },
    { 0x057, "iscinematicplaying" },
    { 0x058, "stopcinematicingame" },
    { 0x059, "getweapondisplayname" },
    { 0x05A, "getweaponbasename" },
    { 0x05B, "getweaponattachments" },
    { 0x05C, "getweaponattachmentdisplaynames" },
    { 0x05D, "getweaponcamoname" },
    { 0x05E, "getweaponreticlename" },
    { 0x05F, "getweaponhidetags" },
    { 0x060, "getanimlength" },
    { 0x061, "animhasnotetrack" },
    { 0x062, "getnotetracktimes" },
    { 0x063, "spawn" },
    { 0x064, "spawnloopsound" },
    { 0x065, "spawnloopingsound" },
    { 0x066, "bullettrace" },
    { 0x067, "target_setmaxsize" },
    { 0x068, "target_setcolor" },
    { 0x069, "target_setdelay" },
    { 0x06A, "getstartorigin" },
    { 0x06B, "getstartangles" },
    { 0x06C, "getcycleoriginoffset" },
    { 0x06D, "magicgrenade" },
    { 0x06E, "magicgrenademanual" },
    { 0x06F, "setblur" },
    { 0x070, "musicplay" },
    { 0x071, "musicstop" },
    { 0x072, "soundfade" },
    { 0x073, "soundsettimescalefactor" },
    { 0x074, "soundresettimescale" },
    { 0x075, "setocclusionpreset" },
    { 0x076, "levelsoundfade" },
    { 0x077, "precachenightvisioncodeassets" },
    { 0x078, "precachedigitaldistortcodeassets" },
    { 0x079, "precacheminimapsentrycodeassets" },
    { 0x07A, "savegame" },
    { 0x07B, "issavesuccessful" },
    { 0x07C, "issaverecentlyloaded" },
    { 0x07D, "savegamenocommit" },
    { 0x07E, "commitsave" },
    { 0x07F, "commitwouldbevalid" },
    { 0x080, "getfxvisibility" },
    { 0x081, "setculldist" },
    { 0x082, "bullettracepassed" },
    { 0x083, "sighttracepassed" },
    { 0x084, "physicstrace" },
    { 0x085, "playerphysicstrace" },
    { 0x086, "getgroundposition" },
    { 0x087, "getmovedelta" },
    { 0x088, "getangledelta" },
    { 0x089, "getnorthyaw" },
    { 0x08A, "getcommandfromkey" },
    { 0x08B, "getsticksconfig" },
    { 0x08C, "weaponfightdist" },
    { 0x08D, "weaponmaxdist" },
    { 0x08E, "isturretactive" },
    { 0x08F, "target_alloc" },
    { 0x090, "target_flush" },
    { 0x091, "target_set" },
    { 0x092, "target_remove" },
    { 0x093, "target_setshader" },
    { 0x094, "target_setoffscreenshader" },
    { 0x095, "target_isinrect" },
    { 0x096, "target_isincircle" },
    { 0x097, "target_startreticlelockon" },
    { 0x098, "target_clearreticlelockon" },
    { 0x099, "target_getarray" },
    { 0x09A, "target_istarget" },
    { 0x09B, "target_setattackmode" },
    { 0x09C, "target_setjavelinonly" },
    { 0x09D, "target_hidefromplayer" },
    { 0x09E, "target_showtoplayer" },
    { 0x09F, "target_setscaledrendermode" },
    { 0x0A0, "target_drawcornersonly" },
    { 0x0A1, "target_drawsquare" },
    { 0x0A2, "target_drawsingle" },
    { 0x0A3, "target_setminsize" },
    { 0x0A4, "setnorthyaw" },
    { 0x0A5, "setslowmotion" },
    { 0x0A6, "randomint" },
    { 0x0A7, "randomfloat" },
    { 0x0A8, "randomintrange" },
    { 0x0A9, "randomfloatrange" },
    { 0x0AA, "sin" },
    { 0x0AB, "cos" },
    { 0x0AC, "tan" },
    { 0x0AD, "asin" },
    { 0x0AE, "acos" },
    { 0x0AF, "atan" },
    { 0x0B0, "int" },
    { 0x0B1, "float" },
    { 0x0B2, "abs" },
    { 0x0B3, "min" },
    { 0x0B4, "objective_additionalcurrent" },
    { 0x0B5, "objective_ring" },
    { 0x0B6, "objective_setpointertextoverride" },
    { 0x0B7, "getnode" },
    { 0x0B8, "getnodearray" },
    { 0x0B9, "getallnodes" },
    { 0x0BA, "getnodesinradius" },
    { 0x0BB, "getnodesinradiussorted" },
    { 0x0BC, "getclosestnodeinsight" },
    { 0x0BD, "getreflectionlocs" },
    { 0x0BE, "getreflectionreferencelocs" },
    { 0x0BF, "getvehicletracksegment" },
    { 0x0C0, "getvehicletracksegmentarray" },
    { 0x0C1, "getallvehicletracksegments" },
    { 0x0C2, "isarray" },
    { 0x0C3, "isai" },
    { 0x0C4, "getindexforluincstring" },
    { 0x0C5, "issentient" },
    { 0x0C6, "isgodmode" },
    { 0x0C7, "getdebugdvar" },
    { 0x0C8, "getdebugdvarint" },
    { 0x0C9, "getdebugdvarfloat" },
    { 0x0CA, "setsaveddvar" },
    { 0x0CB, "getfreeaicount" },
    { 0x0CC, "getaicount" },
    { 0x0CD, "getaiarray" },
    { 0x0CE, "getaispeciesarray" },
    { 0x0CF, "getspawnerarray" },
    { 0x0D0, "getcorpsearray" },
    { 0x0D1, "getspawnerteamarray" },
    { 0x0D2, "getweaponclipmodel" },
    { 0x0D3, "getbrushmodelcenter" },
    { 0x0D4, "getkeybinding" },
    { 0x0D5, "max" },
    { 0x0D6, "floor" },
    { 0x0D7, "ceil" },
    { 0x0D8, "exp" },
    { 0x0D9, "log" },
    { 0x0DA, "sqrt" },
    { 0x0DB, "squared" },
    { 0x0DC, "clamp" },
    { 0x0DD, "angleclamp" },
    { 0x0DE, "angleclamp180" },
    { 0x0DF, "vectorfromlinetopoint" },
    { 0x0E0, "pointonsegmentnearesttopoint" },
    { 0x0E1, "distance" },
    { 0x0E2, "distance2d" },
    { 0x0E3, "distancesquared" },
    { 0x0E4, "length" },
    { 0x0E5, "length2d" },
    { 0x0E6, "lengthsquared" },
    { 0x0E7, "length2dsquared" },
    { 0x0E8, "closer" },
    { 0x0E9, "vectordot" },
    { 0x0EA, "vectorcross" },
    { 0x0EB, "axistoangles" },
    { 0x0EC, "visionsetthermal" },
    { 0x0ED, "visionsetpain" },
    { 0x0EE, "endlobby" },
    { 0x0EF, "setac130ambience" },
    { 0x0F0, "getmapcustom" },
    { 0x0F1, "spawnsighttrace" },
    { 0x0F2, "incrementcounter" },
    { 0x0F3, "getcountertotal" },
    { 0x0F4, "getlevelticks" },
    { 0x0F5, "perlinnoise2d" },
    { 0x0F6, "calcrockingangles" },
    { 0x0F7, "reconevent" },
    { 0x0F8, "reconspatialevent" },
    { 0x0F9, "setsunflareposition" },
    { 0x0FA, "createthreatbiasgroup" },
    { 0x0FB, "threatbiasgroupexists" },
    { 0x0FC, "getthreatbias" },
    { 0x0FD, "setthreatbias" },
    { 0x0FE, "setthreatbiasagainstall" },
    { 0x0FF, "setignoremegroup" },
    { 0x100, "isenemyteam" },
    { 0x101, "objective_additionalentity" },
    { 0x102, "objective_state_nomessage" },
    { 0x103, "objective_string" },
    { 0x104, "objective_string_nomessage" },
    { 0x105, "objective_additionalposition" },
    { 0x106, "objective_current_nomessage" },
    { 0x107, "vectornormalize" },
    { 0x108, "vectortoangles" },
    { 0x109, "vectortoyaw" },
    { 0x10A, "vectorlerp" },
    { 0x10B, "anglestoup" },
    { 0x10C, "anglestoright" },
    { 0x10D, "anglestoforward" },
    { 0x10E, "anglesdelta" },
    { 0x10F, "combineangles" },
    { 0x110, "transformmove" },
    { 0x111, "rotatevector" },
    { 0x112, "rotatepointaroundvector" },
    { 0x113, "issubstr" },
    { 0x114, "isendstr" },
    { 0x115, "getsubstr" },
    { 0x116, "tolower" },
    { 0x117, "strtok" },
    { 0x118, "stricmp" },
    { 0x119, "ambientplay" },
    { 0x11A, "getuavstrengthmax" },
    { 0x11B, "getuavstrengthlevelneutral" },
    { 0x11C, "getuavstrengthlevelshowenemyfastsweep" },
    { 0x11D, "getuavstrengthlevelshowenemydirectional" },
    { 0x11E, "blockteamradar" },
    { 0x11F, "unblockteamradar" },
    { 0x120, "isteamradarblocked" },
    { 0x121, "getassignedteam" },
    { 0x122, "setmatchdata" },
    { 0x123, "getmatchdata" },
    { 0x124, "sendmatchdata" },
    { 0x125, "clearmatchdata" },
    { 0x126, "setmatchdatadef" },
    { 0x127, "setmatchclientip" },
    { 0x128, "setmatchdataid" },
    { 0x129, "setclientmatchdata" },
    { 0x12A, "getclientmatchdata" },
    { 0x12B, "setclientmatchdatadef" },
    { 0x12C, "sendclientmatchdata" },
    { 0x12D, "getbuildversion" },
    { 0x12E, "getbuildnumber" },
    { 0x12F, "getsystemtime" },
    { 0x130, "getmatchrulesdata" },
    { 0x131, "isusingmatchrulesdata" },
    { 0x132, "kick" },
    { 0x133, "issplitscreen" },
    { 0x134, "setmapcenter" },
    { 0x135, "setgameendtime" },
    { 0x136, "visionsetnaked" },
    { 0x137, "visionsetnight" },
    { 0x138, "visionsetmissilecam" },
    { 0x139, "ambientstop" },
    { 0x13A, "precachemodel" },
    { 0x13B, "precacheshellshock" },
    { 0x13C, "precacheitem" },
    { 0x13D, "precacheshader" },
    { 0x13E, "precachestring" },
    { 0x13F, "precachemenu" },
    { 0x140, "precacherumble" },
    { 0x141, "precachelocationselector" },
    { 0x142, "precacheleaderboards" },
    { 0x143, "loadfx" },
    { 0x144, "playfx" },
    { 0x145, "playfxontag" },
    { 0x146, "stopfxontag" },
    { 0x147, "killfxontag" },
    { 0x148, "playloopedfx" },
    { 0x149, "spawnfx" },
    { 0x14A, "triggerfx" },
    { 0x14B, "playfxontagforclients" },
    { 0x14C, "setwinningteam" },
    { 0x14D, "announcement" },
    { 0x14E, "clientannouncement" },
    { 0x14F, "setteammode" },
    { 0x150, "getteamscore" },
    { 0x151, "setteamscore" },
    { 0x152, "setclientnamemode" },
    { 0x153, "updateclientnames" },
    { 0x154, "getteamplayersalive" },
    { 0x155, "logprint" },
    { 0x156, "worldentnumber" },
    { 0x157, "obituary" },
    { 0x158, "positionwouldtelefrag" },
    { 0x159, "canspawn" },
    { 0x15A, "getstarttime" },
    { 0x15B, "precachestatusicon" },
    { 0x15C, "precacheheadicon" },
    { 0x15D, "precacheminimapicon" },
    { 0x15E, "precachempanim" },
    { 0x15F, "map_restart" },
    { 0x160, "exitlevel" },
    { 0x161, "addtestclient" },
    { 0x162, "addagent" },
    { 0x163, "setarchive" },
    { 0x164, "allclientsprint" },
    { 0x165, "clientprint" },
    { 0x166, "mapexists" },
    { 0x167, "isvalidgametype" },
    { 0x168, "matchend" },
    { 0x169, "setplayerteamrank" },
    { 0x16A, "endparty" },
    { 0x16B, "setteamradar" },
    { 0x16C, "getteamradar" },
    { 0x16D, "setteamradarstrength" },
    { 0x16E, "getteamradarstrength" },
    { 0x16F, "getuavstrengthmin" },
    { 0x170, "physicsexplosionsphere" },
    { 0x171, "physicsexplosioncylinder" },
    { 0x172, "physicsjolt" },
    { 0x173, "physicsjitter" },
    { 0x174, "setexpfog" },
    { 0x175, "isexplosivedamagemod" },
    { 0x176, "radiusdamage" },
    { 0x177, "setplayerignoreradiusdamage" },
    { 0x178, "glassradiusdamage" },
    { 0x179, "earthquake" },
    { 0x17A, "getnumparts" },
    { 0x17B, "objective_onentity" },
    { 0x17C, "objective_onentitywithrotation" },
    { 0x17D, "objective_team" },
    { 0x17E, "objective_player" },
    { 0x17F, "objective_playerteam" },
    { 0x180, "objective_playerenemyteam" },
    { 0x181, "objective_playermask_hidefromall" },
    { 0x182, "objective_playermask_hidefrom" },
    { 0x183, "objective_playermask_showtoall" },
    { 0x184, "objective_playermask_showto" },
    { 0x185, "iprintln" },
    { 0x186, "iprintlnbold" },
    { 0x187, "logstring" },
    { 0x188, "getent" },
    { 0x189, "getentarray" },
    { 0x18A, "getspawnarray" },
    { 0x18B, "spawnplane" },
    { 0x18C, "spawnstruct" },
    { 0x18D, "spawnhelicopter" },
    { 0x18E, "isalive" },
    { 0x18F, "isspawner" },
    { 0x190, "missile_createattractorent" },
    { 0x191, "missile_createattractororigin" },
    { 0x192, "missile_createrepulsorent" },
    { 0x193, "missile_createrepulsororigin" },
    { 0x194, "missile_deleteattractor" },
    { 0x195, "playsoundatpos" },
    { 0x196, "newhudelem" },
    { 0x197, "newclienthudelem" },
    { 0x198, "newteamhudelem" },
    { 0x199, "resettimeout" },
    { 0x19A, "isplayer" },
    { 0x19B, "isplayernumber" },
    { 0x19C, "getpartname" },
    { 0x19D, "weaponfiretime" },
    { 0x19E, "weaponclipsize" },
    { 0x19F, "weaponisauto" },
    { 0x1A0, "weaponissemiauto" },
    { 0x1A1, "weaponisboltaction" },
    { 0x1A2, "weaponinheritsperks" },
    { 0x1A3, "weaponburstcount" },
    { 0x1A4, "weapontype" },
    { 0x1A5, "weaponclass" },
    { 0x1A6, "getnextarraykey" },
    { 0x1A7, "sortbydistance" },
    { 0x1A8, "tablelookup" },
    { 0x1A9, "tablelookupbyrow" },
    { 0x1AA, "tablelookupistring" },
    { 0x1AB, "tablelookupistringbyrow" },
    { 0x1AC, "tablelookuprownum" },
    { 0x1AD, "tableexists" },
    { 0x1AE, "getmissileowner" },
    { 0x1AF, "magicbullet" },
    { 0x1B0, "getweaponflashtagname" },
    { 0x1B1, "averagepoint" },
    { 0x1B2, "averagenormal" },
    { 0x1B3, "vehicle_getspawnerarray" },
    { 0x1B4, "playrumbleonposition" },
    { 0x1B5, "playrumblelooponposition" },
    { 0x1B6, "stopallrumbles" },
    { 0x1B7, "soundexists" },
    { 0x1B8, "openfile" },
    { 0x1B9, "closefile" },
    { 0x1BA, "fprintln" },
    { 0x1BB, "fprintfields" },
    { 0x1BC, "freadln" },
    { 0x1BD, "fgetarg" },
    { 0x1BE, "setminimap" },
    { 0x1BF, "setthermalbodymaterial" },
    { 0x1C0, "getarraykeys" },
    { 0x1C1, "getfirstarraykey" },
    { 0x1C2, "getglass" },
    { 0x1C3, "getglassarray" },
    { 0x1C4, "getglassorigin" },
    { 0x1C5, "isglassdestroyed" },
    { 0x1C6, "destroyglass" },
    { 0x1C7, "deleteglass" },
    { 0x1C8, "getentchannelscount" },
    { 0x1C9, "getentchannelname" },
    { 0x1CA, "objective_add" },
    { 0x1CB, "objective_delete" },
    { 0x1CC, "objective_state" },
    { 0x1CD, "objective_icon" },
    { 0x1CE, "objective_position" },
    { 0x1CF, "objective_current" },
    { 0x1D0, "weaponinventorytype" },
    { 0x1D1, "weaponstartammo" },
    { 0x1D2, "weaponmaxammo" },
    { 0x1D3, "weaponaltweaponname" },
    { 0x1D4, "isweaponcliponly" },
    { 0x1D5, "isweapondetonationtimed" },
    { 0x1D6, "weaponhasthermalscope" },
    { 0x1D7, "getvehiclenode" },
    { 0x1D8, "getvehiclenodearray" },
    { 0x1D9, "getallvehiclenodes" },
    { 0x1DA, "getnumvehicles" },
    { 0x1DB, "precachevehicle" },
    { 0x1DC, "spawnvehicle" },
    { 0x1DD, "vehicle_getarray" },
    { 0x1DE, "pow" },
    { 0x1DF, "botgetmemoryevents" },
    { 0x1E0, "botautoconnectenabled" },
    { 0x1E1, "botzonegetcount" },
    { 0x1E2, "botzonesetteam" },
    { 0x1E3, "botzonenearestcount" },
    { 0x1E4, "botmemoryflags" },
    { 0x1E5, "botflagmemoryevents" },
    { 0x1E6, "botzonegetindoorpercent" },
    { 0x1E7, "botsentientswap" },
    { 0x1E8, "isbot" },
    { 0x1E9, "isagent" },
    { 0x1EA, "getmaxagents" },
    { 0x1EB, "botdebugdrawtrigger" },
    { 0x1EC, "botgetclosestnavigablepoint" },
    { 0x1ED, "gettnodesintrigger" },
    { 0x1EE, "nodesvisible" },
    { 0x1EF, "getnodesonpath" },
    { 0x1F0, "getzonecount" },
    { 0x1F1, "getzonenearest" },
    { 0x1F2, "getzonenodes" },
    { 0x1F3, "getzonepath" },
    { 0x1F4, "getzoneorigin" },
    { 0x1F5, "getnodezone" },
    { 0x1F6, "getzonenodesbydist" },
    { 0x1F7, "getzonenodeforindex" },
    { 0x1F8, "getweaponexplosionradius" },
    { 0x1F9, "markdangerousnodes" },
    { 0x1FA, "markdangerousnodesintrigger" },
    { 0x1FB, "nodeexposedtosky" },
    { 0x1FC, "findentrances" },
    { 0x1FD, "badplace_global" },
    { 0x1FE, "getpathdist" },
    { 0x1FF, "getlinkednodes" },
    { 0x200, "disconnectnodepair" },
    { 0x201, "connectnodepair" },
    { 0x202, "drawsoundshape" },
    { 0x203, "gettimesincelastpaused" },
    { 0x204, "setlasermaterial" },
    { 0x205, "precachefxontag" },
    { 0x206, "precachetag" },
    { 0x207, "precachesound" },
    { 0x208, "devsetminimapdvarsettings" },
    { 0x209, "loadtransient" },
    { 0x20A, "unloadtransient" },
    { 0x20B, "unloadalltransients" },
    { 0x20C, "synctransients" },
    { 0x20D, "istransientqueued" },
    { 0x20E, "istransientloaded" },
    { 0x20F, "loadstartpointtransient" },
    { 0x210, "distance2dsquared" },
    { 0x211, "getangledelta3d" },
    { 0x212, "activateclientexploder" },
    { 0x213, "trajectorycalculateinitialvelocity" },
    { 0x214, "trajectorycalculateminimumvelocity" },
    { 0x215, "trajectorycalculateexitangle" },
    { 0x216, "trajectoryestimatedesiredinairtime" },
    { 0x217, "trajectorycomputedeltaheightattime" },
    { 0x218, "trajectorycanattemptaccuratejump" },
    { 0x219, "adddebugcommand" },
    { 0x21A, "ispointinvolume" },
    { 0x21B, "cinematicgettimeinmsec" },
    { 0x21C, "cinematicgetframe" },
    { 0x21D, "iscinematicloaded" },
    { 0x21E, "bbprint" },
    { 0x21F, "getenemysquaddata" },
    { 0x220, "lookupsoundlength" },
    { 0x221, "getscriptablearray" },
    { 0x222, "clearfog" },
    { 0x223, "setleveldopplerpreset" },
    { 0x224, "screenshake" },
    { 0x225, "anglestoaxis" },
    { 0x226, "visionsetwater" },
    { 0x227, "sendscriptusageanalysisdata" },
    { 0x228, "resetscriptusageanalysisdata" },
    { 0x229, "instantlylogusageanalysisdata" },
    { 0x22A, "invertangles" },
    { 0x22B, "rotatevectorinverted" },
    { 0x22C, "calculatestartorientation" },
    { 0x22D, "getcsplinecount" },
    { 0x22E, "getcsplinepointcount" },
    { 0x22F, "getcsplinelength" },
    { 0x230, "getcsplinepointid" },
    { 0x231, "getcsplinepointlabel" },
    { 0x232, "getcsplinepointtension" },
    { 0x233, "getcsplinepointposition" },
    { 0x234, "getcsplinepointcorridordims" },
    { 0x235, "getcsplinepointtangent" },
    { 0x236, "getcsplinepointdisttonextpoint" },
    { 0x237, "calccsplineposition" },
    { 0x238, "calccsplinetangent" },
    { 0x239, "calccsplinecorridor" },
    { 0x23A, "setnojipscore" },
    { 0x23B, "setnojiptime" },
    { 0x23C, "getpredictedentityposition" },
    { 0x23D, "gamedvrprohibitrecording" },
    { 0x23E, "gamedvrstartrecording" },
    { 0x23F, "gamedvrstoprecording" },
    { 0x240, "gamedvrsetvideometadata" },
    { 0x241, "gamedvrprohibitscreenshots" },
    { 0x242, "gamedvrsetscreenshotmetadata" },
    { 0x243, "queuedialog" },
    { 0x244, "speechenablegrammar" },
    { 0x245, "speechenable" },
    { 0x246, "livestreamingenable" },
    { 0x247, "livestreamingsetbitrate" },
    { 0x248, "triggerportableradarping" },
    { 0x249, "setglaregrimematerial" },
    { 0x24A, "botgetteamlimit" },
    { 0x24B, "spawnfxforclient" },
    { 0x24C, "botgetteamdifficulty" },
    { 0x24D, "getsquadassaultelo" },
    { 0x24E, "loadluifile" },
    { 0x24F, "isdedicatedserver" },
    { 0x250, "getplaylistversion" },
    { 0x251, "getplaylistid" },
    { 0x252, "getactiveclientcount" },
    { 0x253, "issquadsmode" },
    { 0x254, "getsquadassaultsquadindex" },
    { 0x255, "visionsetpostapply" },
    { 0x256, "addbot" },
    { 0x257, "ishairrunning" },
    { 0x258, "getsquadassaultenemyprestige" },
    { 0x259, "playcinematicforall" },
    { 0x25A, "preloadcinematicforall" },
    { 0x25B, "stopcinematicforall" },
    { 0x25C, "getenemysquaddogtype" },
    { 0x25D, "capsuletracepassed" },
}};

const std::array<std::pair<std::uint16_t, const char*>, 1066> method_list
{{
    { 0x8000, "thermaldrawdisable" },
    { 0x8001, "setturretdismountorg" },
    { 0x8002, "setdamagestage" },
    { 0x8003, "playsoundtoteam" },
    { 0x8004, "playsoundtoplayer" },
    { 0x8005, "playerhide" },
    { 0x8006, "showtoplayer" },
    { 0x8007, "enableplayeruse" },
    { 0x8008, "disableplayeruse" },
    { 0x8009, "makescrambler" },
    { 0x800A, "makeportableradar" },
    { 0x800B, "clearscrambler" },
    { 0x800C, "clearportableradar" },
    { 0x800D, "placespawnpoint" },
    { 0x800E, "setteamfortrigger" },
    { 0x800F, "clientclaimtrigger" },
    { 0x8010, "clientreleasetrigger" },
    { 0x8011, "releaseclaimedtrigger" },
    { 0x8012, "isusingonlinedataoffline" },
    { 0x8013, "getrestedtime" },
    { 0x8014, "sendleaderboards" },
    { 0x8015, "isonladder" },
    { 0x8016, "getcorpseanim" },
    { 0x8017, "playerforcedeathanim" },
    { 0x8018, "attach" },
    { 0x8019, "attachshieldmodel" },
    { 0x801A, "getlightfovinner" },
    { 0x801B, "getlightfovouter" },
    { 0x801C, "setlightfovrange" },
    { 0x801D, "getlightexponent" },
    { 0x801E, "setlightexponent" },
    { 0x801F, "startragdoll" },
    { 0x8020, "startragdollfromimpact" },
    { 0x8021, "queryshouldearlyragdoll" },
    { 0x8022, "logstring" },
    { 0x8023, "laserhidefromclient" },
    { 0x8024, "stopsoundchannel" },
    { 0x8025, "thermaldrawenable" },
    { 0x8026, "detach" },
    { 0x8027, "detachshieldmodel" },
    { 0x8028, "moveshieldmodel" },
    { 0x8029, "detachall" },
    { 0x802A, "getattachsize" },
    { 0x802B, "getattachmodelname" },
    { 0x802C, "getattachtagname" },
    { 0x802D, "setturretcanaidetach" },
    { 0x802E, "setturretfov" },
    { 0x802F, "lerpfov" },
    { 0x8030, "getvalidcoverpeekouts" },
    { 0x8031, "gethighestnodestance" },
    { 0x8032, "doesnodeallowstance" },
    { 0x8033, "getgunangles" },
    { 0x8034, "magicgrenade" },
    { 0x8035, "magicgrenademanual" },
    { 0x8036, "getentnum" },
    { 0x8037, "launch" },
    { 0x8038, "setsoundblend" },
    { 0x8039, "makefakeai" },
    { 0x803A, "spawndrone" },
    { 0x803B, "setcorpseremovetimer" },
    { 0x803C, "setlookattext" },
    { 0x803D, "setspawnerteam" },
    { 0x803E, "addaieventlistener" },
    { 0x803F, "removeaieventlistener" },
    { 0x8040, "getlightcolor" },
    { 0x8041, "setlightcolor" },
    { 0x8042, "getlightradius" },
    { 0x8043, "setlightradius" },
    { 0x8044, "getattachignorecollision" },
    { 0x8045, "hidepart" },
    { 0x8046, "hidepart_allinstances" },
    { 0x8047, "hideallparts" },
    { 0x8048, "showpart" },
    { 0x8049, "showallparts" },
    { 0x804A, "linkto" },
    { 0x804B, "linktoblendtotag" },
    { 0x804C, "unlink" },
    { 0x804D, "setnormalhealth" },
    { 0x804E, "dodamage" },
    { 0x804F, "kill" },
    { 0x8050, "show" },
    { 0x8051, "hide" },
    { 0x8052, "showonclient" },
    { 0x8053, "hideonclient" },
    { 0x8054, "laserforceon" },
    { 0x8055, "laserforceoff" },
    { 0x8056, "disconnectpaths" },
    { 0x8057, "connectpaths" },
    { 0x8058, "disconnectnode" },
    { 0x8059, "connectnode" },
    { 0x805A, "startusingheroonlylighting" },
    { 0x805B, "stopusingheroonlylighting" },
    { 0x805C, "startusinglessfrequentlighting" },
    { 0x805D, "stopusinglessfrequentlighting" },
    { 0x805E, "setmovingplatformplayerturnrate" },
    { 0x805F, "setthermalfog" },
    { 0x8060, "setnightvisionfog" },
    { 0x8061, "clearthermalfog" },
    { 0x8062, "clearnightvisionfog" },
    { 0x8063, "digitaldistortsetparams" },
    { 0x8064, "setmode" },
    { 0x8065, "getmode" },
    { 0x8066, "setturretignoregoals" },
    { 0x8067, "islinked" },
    { 0x8068, "enablelinkto" },
    { 0x8069, "playsoundatviewheight" },
    { 0x806A, "prefetchsound" },
    { 0x806B, "setpitch" },
    { 0x806C, "scalepitch" },
    { 0x806D, "setvolume" },
    { 0x806E, "scalevolume" },
    { 0x806F, "setspeakermapmonotostereo" },
    { 0x8070, "setspeakermapmonoto51" },
    { 0x8071, "setdistributed2dsound" },
    { 0x8072, "playsoundasmaster" },
    { 0x8073, "playloopsound" },
    { 0x8074, "eqon" },
    { 0x8075, "eqoff" },
    { 0x8076, "haseq" },
    { 0x8077, "iswaitingonsound" },
    { 0x8078, "getnormalhealth" },
    { 0x8079, "playerlinkto" },
    { 0x807A, "playerlinktodelta" },
    { 0x807B, "playerlinkweaponviewtodelta" },
    { 0x807C, "playerlinktoabsolute" },
    { 0x807D, "playerlinktoblend" },
    { 0x807E, "playerlinkedoffsetenable" },
    { 0x807F, "setwaypointedgestyle_secondaryarrow" },
    { 0x8080, "setwaypointiconoffscreenonly" },
    { 0x8081, "fadeovertime" },
    { 0x8082, "scaleovertime" },
    { 0x8083, "moveovertime" },
    { 0x8084, "reset" },
    { 0x8085, "destroy" },
    { 0x8086, "setpulsefx" },
    { 0x8087, "setplayernamestring" },
    { 0x8088, "changefontscaleovertime" },
    { 0x8089, "startignoringspotlight" },
    { 0x808A, "stopignoringspotlight" },
    { 0x808B, "dontcastshadows" },
    { 0x808C, "castshadows" },
    { 0x808D, "setstablemissile" },
    { 0x808E, "playersetgroundreferenceent" },
    { 0x808F, "dontinterpolate" },
    { 0x8090, "dospawn" },
    { 0x8091, "stalingradspawn" },
    { 0x8092, "getorigin" },
    { 0x8093, "getcentroid" },
    { 0x8094, "getshootatpos" },
    { 0x8095, "getdebugeye" },
    { 0x8096, "useby" },
    { 0x8097, "playsound" },
    { 0x8098, "playerlinkedoffsetdisable" },
    { 0x8099, "playerlinkedsetviewznear" },
    { 0x809A, "playerlinkedsetusebaseangleforviewclamp" },
    { 0x809B, "lerpviewangleclamp" },
    { 0x809C, "setviewangleresistance" },
    { 0x809D, "springcamenabled" },
    { 0x809E, "springcamdisabled" },
    { 0x809F, "linktoplayerview" },
    { 0x80A0, "unlinkfromplayerview" },
    { 0x80A1, "geteye" },
    { 0x80A2, "istouching" },
    { 0x80A3, "getistouchingentities" },
    { 0x80A4, "stoploopsound" },
    { 0x80A5, "stopsounds" },
    { 0x80A6, "playrumbleonentity" },
    { 0x80A7, "playrumblelooponentity" },
    { 0x80A8, "stoprumble" },
    { 0x80A9, "delete" },
    { 0x80AA, "setmodel" },
    { 0x80AB, "laseron" },
    { 0x80AC, "laseroff" },
    { 0x80AD, "laseraltviewon" },
    { 0x80AE, "laseraltviewoff" },
    { 0x80AF, "thermalvisionon" },
    { 0x80B0, "thermalvisiononshadowoff" },
    { 0x80B1, "thermalvisionoff" },
    { 0x80B2, "thermalvisionfofoverlayon" },
    { 0x80B3, "thermalvisionfofoverlayoff" },
    { 0x80B4, "autospotoverlayon" },
    { 0x80B5, "autospotoverlayoff" },
    { 0x80B6, "seteyesonuplinkenabled" },
    { 0x80B7, "setcontents" },
    { 0x80B8, "makeusable" },
    { 0x80B9, "makeunusable" },
    { 0x80BA, "setwhizbyprobabilities" },
    { 0x80BB, "visionsetnakedforplayer_lerp" },
    { 0x80BC, "setwaitnode" },
    { 0x80BD, "returnplayercontrol" },
    { 0x80BE, "vehphys_starttrack" },
    { 0x80BF, "vehphys_clearautodisable" },
    { 0x80C0, "vehicleusealtblendedaudio" },
    { 0x80C1, "settext" },
    { 0x80C2, "clearalltextafterhudelem" },
    { 0x80C3, "setshader" },
    { 0x80C4, "settargetent" },
    { 0x80C5, "cleartargetent" },
    { 0x80C6, "settimer" },
    { 0x80C7, "settimerup" },
    { 0x80C8, "settimerstatic" },
    { 0x80C9, "settenthstimer" },
    { 0x80CA, "settenthstimerup" },
    { 0x80CB, "settenthstimerstatic" },
    { 0x80CC, "setclock" },
    { 0x80CD, "setclockup" },
    { 0x80CE, "setvalue" },
    { 0x80CF, "setwaypoint" },
    { 0x80D0, "setwaypointedgestyle_rotatingicon" },
    { 0x80D1, "setcursorhint" },
    { 0x80D2, "sethintstring" },
    { 0x80D3, "forceusehinton" },
    { 0x80D4, "forceusehintoff" },
    { 0x80D5, "makesoft" },
    { 0x80D6, "makehard" },
    { 0x80D7, "willneverchange" },
    { 0x80D8, "startfiring" },
    { 0x80D9, "stopfiring" },
    { 0x80DA, "isfiringturret" },
    { 0x80DB, "startbarrelspin" },
    { 0x80DC, "stopbarrelspin" },
    { 0x80DD, "getbarrelspinrate" },
    { 0x80DE, "remotecontrolturret" },
    { 0x80DF, "remotecontrolturretoff" },
    { 0x80E0, "shootturret" },
    { 0x80E1, "getturretowner" },
    { 0x80E2, "enabledeathshield" },
    { 0x80E3, "nightvisiongogglesforceon" },
    { 0x80E4, "nightvisiongogglesforceoff" },
    { 0x80E5, "enableinvulnerability" },
    { 0x80E6, "disableinvulnerability" },
    { 0x80E7, "enablebreaching" },
    { 0x80E8, "disablebreaching" },
    { 0x80E9, "forceviewmodelanimation" },
    { 0x80EA, "disableturretdismount" },
    { 0x80EB, "enableturretdismount" },
    { 0x80EC, "uploadscore" },
    { 0x80ED, "uploadtime" },
    { 0x80EE, "uploadleaderboards" },
    { 0x80EF, "giveachievement" },
    { 0x80F0, "hidehud" },
    { 0x80F1, "showhud" },
    { 0x80F2, "mountvehicle" },
    { 0x80F3, "dismountvehicle" },
    { 0x80F4, "enableslowaim" },
    { 0x80F5, "disableslowaim" },
    { 0x80F6, "usehintsinvehicle" },
    { 0x80F7, "vehicleattackbuttonpressed" },
    { 0x80F8, "setwhizbyoffset" },
    { 0x80F9, "setsentryowner" },
    { 0x80FA, "setsentrycarrier" },
    { 0x80FB, "setturretminimapvisible" },
    { 0x80FC, "settargetentity" },
    { 0x80FD, "snaptotargetentity" },
    { 0x80FE, "cleartargetentity" },
    { 0x80FF, "getturrettarget" },
    { 0x8100, "setplayerspread" },
    { 0x8101, "setaispread" },
    { 0x8102, "setsuppressiontime" },
    { 0x8103, "setflaggedanimknobrestart" },
    { 0x8104, "setflaggedanimknoblimitedrestart" },
    { 0x8105, "setflaggedanimknoball" },
    { 0x8106, "setflaggedanimknoballrestart" },
    { 0x8107, "setflaggedanim" },
    { 0x8108, "setflaggedanimlimited" },
    { 0x8109, "setflaggedanimrestart" },
    { 0x810A, "setflaggedanimlimitedrestart" },
    { 0x810B, "useanimtree" },
    { 0x810C, "stopuseanimtree" },
    { 0x810D, "setanimtime" },
    { 0x810E, "showviewmodel" },
    { 0x810F, "hideviewmodel" },
    { 0x8110, "allowstand" },
    { 0x8111, "allowcrouch" },
    { 0x8112, "allowprone" },
    { 0x8113, "allowlean" },
    { 0x8114, "allowswim" },
    { 0x8115, "setocclusion" },
    { 0x8116, "setocclusionfromtable" },
    { 0x8117, "deactivateocclusion" },
    { 0x8118, "deactivateallocclusion" },
    { 0x8119, "isocclusionenabled" },
    { 0x811A, "setreverbfromtable" },
    { 0x811B, "setvolmodfromtable" },
    { 0x811C, "settimescalefactorfromtable" },
    { 0x811D, "setwhizbyfromtable" },
    { 0x811E, "seteqfromtable" },
    { 0x811F, "iseqenabled" },
    { 0x8120, "seteq" },
    { 0x8121, "seteqbands" },
    { 0x8122, "deactivateeq" },
    { 0x8123, "seteqlerp" },
    { 0x8124, "islookingat" },
    { 0x8125, "isthrowinggrenade" },
    { 0x8126, "isfiring" },
    { 0x8127, "ismeleeing" },
    { 0x8128, "setautopickup" },
    { 0x8129, "allowmelee" },
    { 0x812A, "allowfire" },
    { 0x812B, "enablehealthshield" },
    { 0x812C, "setconvergencetime" },
    { 0x812D, "setconvergenceheightpercent" },
    { 0x812E, "setturretteam" },
    { 0x812F, "maketurretsolid" },
    { 0x8130, "maketurretoperable" },
    { 0x8131, "maketurretinoperable" },
    { 0x8132, "makeentitysentient" },
    { 0x8133, "freeentitysentient" },
    { 0x8134, "isindoor" },
    { 0x8135, "getdroptofloorposition" },
    { 0x8136, "isbadguy" },
    { 0x8137, "animscripted" },
    { 0x8138, "animscriptedthirdperson" },
    { 0x8139, "animrelative" },
    { 0x813A, "stopanimscripted" },
    { 0x813B, "clearanim" },
    { 0x813C, "setanimknob" },
    { 0x813D, "setanimknoblimited" },
    { 0x813E, "setanimknobrestart" },
    { 0x813F, "setanimknoblimitedrestart" },
    { 0x8140, "setanimknoball" },
    { 0x8141, "setanimknoballlimited" },
    { 0x8142, "setanimknoballrestart" },
    { 0x8143, "setanimknoballlimitedrestart" },
    { 0x8144, "setanim" },
    { 0x8145, "setanimlimited" },
    { 0x8146, "setanimrestart" },
    { 0x8147, "setanimlimitedrestart" },
    { 0x8148, "getanimtime" },
    { 0x8149, "getanimweight" },
    { 0x814A, "getanimassettype" },
    { 0x814B, "setflaggedanimknob" },
    { 0x814C, "setflaggedanimknoblimited" },
    { 0x814D, "setturretaccuracy" },
    { 0x814E, "setrightarc" },
    { 0x814F, "setleftarc" },
    { 0x8150, "settoparc" },
    { 0x8151, "setbottomarc" },
    { 0x8152, "setautorotationdelay" },
    { 0x8153, "setdefaultdroppitch" },
    { 0x8154, "restoredefaultdroppitch" },
    { 0x8155, "turretfiredisable" },
    { 0x8156, "getfixednodesafevolume" },
    { 0x8157, "clearfixednodesafevolume" },
    { 0x8158, "isingoal" },
    { 0x8159, "setruntopos" },
    { 0x815A, "nearnode" },
    { 0x815B, "nearclaimnode" },
    { 0x815C, "nearclaimnodeandangle" },
    { 0x815D, "atdangerousnode" },
    { 0x815E, "getenemyinfo" },
    { 0x815F, "clearenemy" },
    { 0x8160, "setentitytarget" },
    { 0x8161, "clearentitytarget" },
    { 0x8162, "setpotentialthreat" },
    { 0x8163, "clearpotentialthreat" },
    { 0x8164, "setflashbanged" },
    { 0x8165, "setengagementmindist" },
    { 0x8166, "setengagementmaxdist" },
    { 0x8167, "isknownenemyinradius" },
    { 0x8168, "isknownenemyinvolume" },
    { 0x8169, "settalktospecies" },
    { 0x816A, "laseralton" },
    { 0x816B, "laseraltoff" },
    { 0x816C, "invisiblenotsolid" },
    { 0x816D, "visiblesolid" },
    { 0x816E, "setdefaultaimlimits" },
    { 0x816F, "initriotshieldhealth" },
    { 0x8170, "getenemysqdist" },
    { 0x8171, "getclosestenemysqdist" },
    { 0x8172, "setthreatbiasgroup" },
    { 0x8173, "getthreatbiasgroup" },
    { 0x8174, "turretfireenable" },
    { 0x8175, "setturretmodechangewait" },
    { 0x8176, "usetriggerrequirelookat" },
    { 0x8177, "getstance" },
    { 0x8178, "setstance" },
    { 0x8179, "itemweaponsetammo" },
    { 0x817A, "getammocount" },
    { 0x817B, "gettagorigin" },
    { 0x817C, "gettagangles" },
    { 0x817D, "shellshock" },
    { 0x817E, "stunplayer" },
    { 0x817F, "stopshellshock" },
    { 0x8180, "fadeoutshellshock" },
    { 0x8181, "setdepthoffield" },
    { 0x8182, "setviewmodeldepthoffield" },
    { 0x8183, "setmotionblurmovescale" },
    { 0x8184, "pickupgrenade" },
    { 0x8185, "useturret" },
    { 0x8186, "stopuseturret" },
    { 0x8187, "canuseturret" },
    { 0x8188, "traversemode" },
    { 0x8189, "animmode" },
    { 0x818A, "orientmode" },
    { 0x818B, "getmotionangle" },
    { 0x818C, "shouldfacemotion" },
    { 0x818D, "getanglestolikelyenemypath" },
    { 0x818E, "setturretanim" },
    { 0x818F, "getturret" },
    { 0x8190, "getgroundenttype" },
    { 0x8191, "forcestartnegotiation" },
    { 0x8192, "setalienjumpcostscale" },
    { 0x8193, "setalienruncostscale" },
    { 0x8194, "setalientraversecostscale" },
    { 0x8195, "animcustom" },
    { 0x8196, "isinscriptedstate" },
    { 0x8197, "canattackenemynode" },
    { 0x8198, "getnegotiationstartnode" },
    { 0x8199, "getnegotiationendnode" },
    { 0x819A, "getnegotiationnextnode" },
    { 0x819B, "getdoorpathnode" },
    { 0x819C, "comparenodedirtopathdir" },
    { 0x819D, "checkprone" },
    { 0x819E, "pushplayer" },
    { 0x819F, "checkgrenadethrowpos" },
    { 0x81A0, "setgoalnode" },
    { 0x81A1, "setgoalpos" },
    { 0x81A2, "setgoalentity" },
    { 0x81A3, "setgoalvolume" },
    { 0x81A4, "setgoalvolumeauto" },
    { 0x81A5, "getgoalvolume" },
    { 0x81A6, "cleargoalvolume" },
    { 0x81A7, "setfixednodesafevolume" },
    { 0x81A8, "setmotionblurturnscale" },
    { 0x81A9, "setmotionblurzoomscale" },
    { 0x81AA, "viewkick" },
    { 0x81AB, "localtoworldcoords" },
    { 0x81AC, "getentitynumber" },
    { 0x81AD, "getentityvelocity" },
    { 0x81AE, "enablegrenadetouchdamage" },
    { 0x81AF, "disablegrenadetouchdamage" },
    { 0x81B0, "enableaimassist" },
    { 0x81B1, "setlookatyawlimits" },
    { 0x81B2, "stoplookat" },
    { 0x81B3, "getmuzzlepos" },
    { 0x81B4, "getmuzzleangle" },
    { 0x81B5, "getmuzzlesideoffsetpos" },
    { 0x81B6, "getaimangle" },
    { 0x81B7, "canshoot" },
    { 0x81B8, "canshootenemy" },
    { 0x81B9, "cansee" },
    { 0x81BA, "seerecently" },
    { 0x81BB, "lastknowntime" },
    { 0x81BC, "lastknownpos" },
    { 0x81BD, "dropweapon" },
    { 0x81BE, "maymovetopoint" },
    { 0x81BF, "maymovefrompointtopoint" },
    { 0x81C0, "teleport" },
    { 0x81C1, "forceteleport" },
    { 0x81C2, "safeteleport" },
    { 0x81C3, "withinapproxpathdist" },
    { 0x81C4, "ispathdirect" },
    { 0x81C5, "allowedstances" },
    { 0x81C6, "isstanceallowed" },
    { 0x81C7, "issuppressionwaiting" },
    { 0x81C8, "issuppressed" },
    { 0x81C9, "ismovesuppressed" },
    { 0x81CA, "isgrenadepossafe" },
    { 0x81CB, "checkgrenadethrow" },
    { 0x81CC, "checkgrenadelaunch" },
    { 0x81CD, "checkgrenadelaunchpos" },
    { 0x81CE, "throwgrenade" },
    { 0x81CF, "disableaimassist" },
    { 0x81D0, "radiusdamage" },
    { 0x81D1, "detonate" },
    { 0x81D2, "damageconetrace" },
    { 0x81D3, "sightconetrace" },
    { 0x81D4, "missile_settargetent" },
    { 0x81D5, "missile_settargetpos" },
    { 0x81D6, "missile_cleartarget" },
    { 0x81D7, "missile_setflightmodedirect" },
    { 0x81D8, "missile_setflightmodetop" },
    { 0x81D9, "getlightintensity" },
    { 0x81DA, "setlightintensity" },
    { 0x81DB, "isragdoll" },
    { 0x81DC, "setmovespeedscale" },
    { 0x81DD, "cameralinkto" },
    { 0x81DE, "cameraunlink" },
    { 0x81DF, "startcoverarrival" },
    { 0x81E0, "starttraversearrival" },
    { 0x81E1, "checkcoverexitposwithpath" },
    { 0x81E2, "shoot" },
    { 0x81E3, "shootblank" },
    { 0x81E4, "melee" },
    { 0x81E5, "updateplayersightaccuracy" },
    { 0x81E6, "findshufflecovernode" },
    { 0x81E7, "findnearbycovernode" },
    { 0x81E8, "findcovernode" },
    { 0x81E9, "findbestcovernode" },
    { 0x81EA, "getcovernode" },
    { 0x81EB, "usecovernode" },
    { 0x81EC, "iscovervalidagainstenemy" },
    { 0x81ED, "reacquirestep" },
    { 0x81EE, "findreacquiredirectpath" },
    { 0x81EF, "trimpathtoattack" },
    { 0x81F0, "reacquiremove" },
    { 0x81F1, "findreacquireproximatepath" },
    { 0x81F2, "flagenemyunattackable" },
    { 0x81F3, "enterprone" },
    { 0x81F4, "exitprone" },
    { 0x81F5, "setproneanimnodes" },
    { 0x81F6, "updateprone" },
    { 0x81F7, "clearpitchorient" },
    { 0x81F8, "setlookatanimnodes" },
    { 0x81F9, "setlookat" },
    { 0x81FA, "setlookatentity" },
    { 0x81FB, "controlslinkto" },
    { 0x81FC, "controlsunlink" },
    { 0x81FD, "makevehiclesolidcapsule" },
    { 0x81FE, "teleportentityrelative" },
    { 0x81FF, "makevehiclesolidsphere" },
    { 0x8200, "makevehiclesolid" },
    { 0x8201, "remotecontrolvehicle" },
    { 0x8202, "remotecontrolvehicleoff" },
    { 0x8203, "isfiringvehicleturret" },
    { 0x8204, "remotecontrolvehicletarget" },
    { 0x8205, "remotecontrolvehicletargetoff" },
    { 0x8206, "drivevehicleandcontrolturret" },
    { 0x8207, "drivevehicleandcontrolturretoff" },
    { 0x8208, "getplayersetting" },
    { 0x8209, "getlocalplayerprofiledata" },
    { 0x820A, "setlocalplayerprofiledata" },
    { 0x820B, "remotecamerasoundscapeon" },
    { 0x820C, "remotecamerasoundscapeoff" },
    { 0x820D, "setmotiontrackervisible" },
    { 0x820E, "getmotiontrackervisible" },
    { 0x820F, "worldpointinreticle_circle" },
    { 0x8210, "worldpointinreticle_rect" },
    { 0x8211, "getpointinbounds" },
    { 0x8212, "transfermarkstonewscriptmodel" },
    { 0x8213, "setwatersheeting" },
    { 0x8214, "addontoviewmodel" },
    { 0x8215, "clearviewmodeladdons" },
    { 0x8216, "setweaponhudiconoverride" },
    { 0x8217, "getweaponhudiconoverride" },
    { 0x8218, "setempjammed" },
    { 0x8219, "playersetexpfog" },
    { 0x821A, "isitemunlocked" },
    { 0x821B, "getplayerdata" },
    { 0x821C, "getrankedplayerdata" },
    { 0x821D, "getprivateplayerdata" },
    { 0x821E, "getcoopplayerdata" },
    { 0x821F, "getcommonplayerdata" },
    { 0x8220, "vehicleturretcontroloff" },
    { 0x8221, "isturretready" },
    { 0x8222, "vehicledriveto" },
    { 0x8223, "vehicle_dospawn" },
    { 0x8224, "vehicle_isphysveh" },
    { 0x8225, "vehphys_crash" },
    { 0x8226, "vehphys_launch" },
    { 0x8227, "vehphys_disablecrashing" },
    { 0x8228, "vehphys_enablecrashing" },
    { 0x8229, "vehphys_setspeed" },
    { 0x822A, "vehphys_setconveyorbelt" },
    { 0x822B, "freehelicopter" },
    { 0x822C, "playerlinkedturretanglesenable" },
    { 0x822D, "playerlinkedturretanglesdisable" },
    { 0x822E, "playersetstreamorigin" },
    { 0x822F, "playerclearstreamorigin" },
    { 0x8230, "nightvisionviewon" },
    { 0x8231, "nightvisionviewoff" },
    { 0x8232, "painvisionon" },
    { 0x8233, "painvisionoff" },
    { 0x8234, "getplayerintelisfound" },
    { 0x8235, "setplayerintelfound" },
    { 0x8236, "newpip" },
    { 0x8237, "sethuddynlight" },
    { 0x8238, "startscriptedanim" },
    { 0x8239, "startcoverbehavior" },
    { 0x823A, "setplayerdata" },
    { 0x823B, "setrankedplayerdata" },
    { 0x823C, "setprivateplayerdata" },
    { 0x823D, "setcoopplayerdata" },
    { 0x823E, "setcommonplayerdata" },
    { 0x823F, "getcacplayerdata" },
    { 0x8240, "trackerupdate" },
    { 0x8241, "pingplayer" },
    { 0x8242, "buttonpressed" },
    { 0x8243, "sayall" },
    { 0x8244, "sayteam" },
    { 0x8245, "setspawnweapon" },
    { 0x8246, "dropitem" },
    { 0x8247, "dropscavengerbag" },
    { 0x8248, "setjitterparams" },
    { 0x8249, "sethoverparams" },
    { 0x824A, "joltbody" },
    { 0x824B, "freevehicle" },
    { 0x824C, "getwheelsurface" },
    { 0x824D, "getvehicleowner" },
    { 0x824E, "setvehiclelookattext" },
    { 0x824F, "setvehicleteam" },
    { 0x8250, "setneargoalnotifydist" },
    { 0x8251, "setvehgoalpos" },
    { 0x8252, "setgoalyaw" },
    { 0x8253, "cleargoalyaw" },
    { 0x8254, "settargetyaw" },
    { 0x8255, "cleartargetyaw" },
    { 0x8256, "vehicle_helisetai" },
    { 0x8257, "setturrettargetvec" },
    { 0x8258, "setturrettargetent" },
    { 0x8259, "clearturrettarget" },
    { 0x825A, "vehicle_canturrettargetpoint" },
    { 0x825B, "setlookatent" },
    { 0x825C, "clearlookatent" },
    { 0x825D, "setvehweapon" },
    { 0x825E, "fireweapon" },
    { 0x825F, "vehicleturretcontrolon" },
    { 0x8260, "finishplayerdamage" },
    { 0x8261, "suicide" },
    { 0x8262, "closeingamemenu" },
    { 0x8263, "iprintln" },
    { 0x8264, "iprintlnbold" },
    { 0x8265, "spawn" },
    { 0x8266, "setentertime" },
    { 0x8267, "cloneplayer" },
    { 0x8268, "istalking" },
    { 0x8269, "allowspectateteam" },
    { 0x826A, "getguid" },
    { 0x826B, "physicslaunchserver" },
    { 0x826C, "physicslaunchserveritem" },
    { 0x826D, "clonebrushmodeltoscriptmodel" },
    { 0x826E, "scriptmodelplayanim" },
    { 0x826F, "scriptmodelclearanim" },
    { 0x8270, "scriptmodelplayanimdeltamotion" },
    { 0x8271, "vehicle_teleport" },
    { 0x8272, "attachpath" },
    { 0x8273, "getattachpos" },
    { 0x8274, "startpath" },
    { 0x8275, "setswitchnode" },
    { 0x8276, "setwaitspeed" },
    { 0x8277, "vehicle_finishdamage" },
    { 0x8278, "vehicle_setspeed" },
    { 0x8279, "vehicle_setspeedimmediate" },
    { 0x827A, "vehicle_rotateyaw" },
    { 0x827B, "vehicle_getspeed" },
    { 0x827C, "vehicle_getvelocity" },
    { 0x827D, "vehicle_getbodyvelocity" },
    { 0x827E, "vehicle_getsteering" },
    { 0x827F, "vehicle_getthrottle" },
    { 0x8280, "vehicle_turnengineoff" },
    { 0x8281, "vehicle_turnengineon" },
    { 0x8282, "vehicle_orientto" },
    { 0x8283, "getgoalspeedmph" },
    { 0x8284, "setacceleration" },
    { 0x8285, "setdeceleration" },
    { 0x8286, "resumespeed" },
    { 0x8287, "setyawspeed" },
    { 0x8288, "setyawspeedbyname" },
    { 0x8289, "setmaxpitchroll" },
    { 0x828A, "setairresistance" },
    { 0x828B, "setturningability" },
    { 0x828C, "getxuid" },
    { 0x828D, "getucdidhigh" },
    { 0x828E, "getucdidlow" },
    { 0x828F, "getclanidhigh" },
    { 0x8290, "getclanidlow" },
    { 0x8291, "ishost" },
    { 0x8292, "getspectatingplayer" },
    { 0x8293, "predictstreampos" },
    { 0x8294, "updatescores" },
    { 0x8295, "updatedmscores" },
    { 0x8296, "setrank" },
    { 0x8297, "setcardtitle" },
    { 0x8298, "weaponlocknoclearance" },
    { 0x8299, "visionsyncwithplayer" },
    { 0x829A, "showhudsplash" },
    { 0x829B, "setperk" },
    { 0x829C, "hasperk" },
    { 0x829D, "clearperks" },
    { 0x829E, "unsetperk" },
    { 0x829F, "registerparty" },
    { 0x82A0, "getfireteammembers" },
    { 0x82A1, "noclip" },
    { 0x82A2, "ufo" },
    { 0x82A3, "moveto" },
    { 0x82A4, "movex" },
    { 0x82A5, "movey" },
    { 0x82A6, "movez" },
    { 0x82A7, "movegravity" },
    { 0x82A8, "moveslide" },
    { 0x82A9, "stopmoveslide" },
    { 0x82AA, "rotateto" },
    { 0x82AB, "rotatepitch" },
    { 0x82AC, "rotateyaw" },
    { 0x82AD, "rotateroll" },
    { 0x82AE, "addpitch" },
    { 0x82AF, "addyaw" },
    { 0x82B0, "addroll" },
    { 0x82B1, "vibrate" },
    { 0x82B2, "rotatevelocity" },
    { 0x82B3, "solid" },
    { 0x82B4, "notsolid" },
    { 0x82B5, "setcandamage" },
    { 0x82B6, "setcanradiusdamage" },
    { 0x82B7, "physicslaunchclient" },
    { 0x82B8, "setcardicon" },
    { 0x82B9, "setcardnameplate" },
    { 0x82BA, "setcarddisplayslot" },
    { 0x82BB, "kc_regweaponforfxremoval" },
    { 0x82BC, "laststandrevive" },
    { 0x82BD, "setspectatedefaults" },
    { 0x82BE, "getthirdpersoncrosshairoffset" },
    { 0x82BF, "disableweaponpickup" },
    { 0x82C0, "enableweaponpickup" },
    { 0x82C1, "issplitscreenplayer" },
    { 0x82C2, "getweaponslistoffhands" },
    { 0x82C3, "getweaponslistitems" },
    { 0x82C4, "getweaponslistexclusives" },
    { 0x82C5, "getweaponslist" },
    { 0x82C6, "canplayerplacesentry" },
    { 0x82C7, "canplayerplacetank" },
    { 0x82C8, "visionsetnakedforplayer" },
    { 0x82C9, "visionsetnightforplayer" },
    { 0x82CA, "visionsetmissilecamforplayer" },
    { 0x82CB, "visionsetthermalforplayer" },
    { 0x82CC, "visionsetpainforplayer" },
    { 0x82CD, "setblurforplayer" },
    { 0x82CE, "getplayerweaponmodel" },
    { 0x82CF, "getplayerknifemodel" },
    { 0x82D0, "updateentitywithweapons" },
    { 0x82D1, "notifyonplayercommand" },
    { 0x82D2, "canmantle" },
    { 0x82D3, "forcemantle" },
    { 0x82D4, "ismantling" },
    { 0x82D5, "playfx" },
    { 0x82D6, "player_recoilscaleon" },
    { 0x82D7, "player_recoilscaleoff" },
    { 0x82D8, "weaponlockstart" },
    { 0x82D9, "weaponlockfinalize" },
    { 0x82DA, "weaponlockfree" },
    { 0x82DB, "weaponlocktargettooclose" },
    { 0x82DC, "issplitscreenplayerprimary" },
    { 0x82DD, "markforeyeson" },
    { 0x82DE, "issighted" },
    { 0x82DF, "getsightedplayers" },
    { 0x82E0, "getplayerssightingme" },
    { 0x82E1, "getviewmodel" },
    { 0x82E2, "fragbuttonpressed" },
    { 0x82E3, "secondaryoffhandbuttonpressed" },
    { 0x82E4, "getcurrentweaponclipammo" },
    { 0x82E5, "setvelocity" },
    { 0x82E6, "getplayerviewheight" },
    { 0x82E7, "getnormalizedmovement" },
    { 0x82E8, "setchannelvolumes" },
    { 0x82E9, "deactivatechannelvolumes" },
    { 0x82EA, "playlocalsound" },
    { 0x82EB, "stoplocalsound" },
    { 0x82EC, "setweaponammoclip" },
    { 0x82ED, "setweaponammostock" },
    { 0x82EE, "getweaponammoclip" },
    { 0x82EF, "getweaponammostock" },
    { 0x82F0, "anyammoforweaponmodes" },
    { 0x82F1, "setclientomnvar" },
    { 0x82F2, "setclientdvar" },
    { 0x82F3, "setclientdvars" },
    { 0x82F4, "setclientspawnsighttraces" },
    { 0x82F5, "clientspawnsighttracepassed" },
    { 0x82F6, "allowads" },
    { 0x82F7, "allowjump" },
    { 0x82F8, "allowsprint" },
    { 0x82F9, "setspreadoverride" },
    { 0x82FA, "resetspreadoverride" },
    { 0x82FB, "setaimspreadmovementscale" },
    { 0x82FC, "setactionslot" },
    { 0x82FD, "setviewkickscale" },
    { 0x82FE, "getviewkickscale" },
    { 0x82FF, "getweaponslistall" },
    { 0x8300, "getweaponslistprimaries" },
    { 0x8301, "getnormalizedcameramovement" },
    { 0x8302, "giveweapon" },
    { 0x8303, "takeweapon" },
    { 0x8304, "takeallweapons" },
    { 0x8305, "getcurrentweapon" },
    { 0x8306, "getcurrentprimaryweapon" },
    { 0x8307, "getcurrentoffhand" },
    { 0x8308, "hasweapon" },
    { 0x8309, "switchtoweapon" },
    { 0x830A, "switchtoweaponimmediate" },
    { 0x830B, "switchtooffhand" },
    { 0x830C, "setoffhandsecondaryclass" },
    { 0x830D, "getoffhandsecondaryclass" },
    { 0x830E, "beginlocationselection" },
    { 0x830F, "endlocationselection" },
    { 0x8310, "disableweapons" },
    { 0x8311, "enableweapons" },
    { 0x8312, "disableoffhandweapons" },
    { 0x8313, "enableoffhandweapons" },
    { 0x8314, "disableweaponswitch" },
    { 0x8315, "enableweaponswitch" },
    { 0x8316, "openpopupmenu" },
    { 0x8317, "openpopupmenunomouse" },
    { 0x8318, "closepopupmenu" },
    { 0x8319, "openmenu" },
    { 0x831A, "closemenu" },
    { 0x831B, "savematchrulestohistory" },
    { 0x831C, "freezecontrols" },
    { 0x831D, "disableusability" },
    { 0x831E, "enableusability" },
    { 0x831F, "setwhizbyspreads" },
    { 0x8320, "setwhizbyradii" },
    { 0x8321, "setreverb" },
    { 0x8322, "deactivatereverb" },
    { 0x8323, "setvolmod" },
    { 0x8324, "setchannelvolume" },
    { 0x8325, "givestartammo" },
    { 0x8326, "givemaxammo" },
    { 0x8327, "getfractionstartammo" },
    { 0x8328, "getfractionmaxammo" },
    { 0x8329, "isdualwielding" },
    { 0x832A, "isreloading" },
    { 0x832B, "isswitchingweapon" },
    { 0x832C, "setorigin" },
    { 0x832D, "getvelocity" },
    { 0x832E, "setplayerangles" },
    { 0x832F, "getplayerangles" },
    { 0x8330, "usebuttonpressed" },
    { 0x8331, "attackbuttonpressed" },
    { 0x8332, "adsbuttonpressed" },
    { 0x8333, "meleebuttonpressed" },
    { 0x8334, "playerads" },
    { 0x8335, "isonground" },
    { 0x8336, "isusingturret" },
    { 0x8337, "setviewmodel" },
    { 0x8338, "setoffhandprimaryclass" },
    { 0x8339, "getoffhandprimaryclass" },
    { 0x833A, "startac130" },
    { 0x833B, "stopac130" },
    { 0x833C, "enablemousesteer" },
    { 0x833D, "setscriptmoverkillcam" },
    { 0x833E, "usinggamepad" },
    { 0x833F, "forcethirdpersonwhenfollowing" },
    { 0x8340, "disableforcethirdpersonwhenfollowing" },
    { 0x8341, "botsetflag" },
    { 0x8342, "botsetstance" },
    { 0x8343, "botsetscriptmove" },
    { 0x8344, "botsetscriptgoal" },
    { 0x8345, "botsetscriptgoalnode" },
    { 0x8346, "botclearscriptgoal" },
    { 0x8347, "botsetscriptenemy" },
    { 0x8348, "botclearscriptenemy" },
    { 0x8349, "botsetattacker" },
    { 0x834A, "botgetscriptgoal" },
    { 0x834B, "botgetscriptgoalradius" },
    { 0x834C, "botgetscriptgoalyaw" },
    { 0x834D, "botgetscriptgoaltype" },
    { 0x834E, "botgetentrancepoint" },
    { 0x834F, "botgetworldsize" },
    { 0x8350, "botnodeavailable" },
    { 0x8351, "botfindnoderandom" },
    { 0x8352, "botmemoryevent" },
    { 0x8353, "botmemoryselectpos" },
    { 0x8354, "botnodepick" },
    { 0x8355, "bothasscriptgoal" },
    { 0x8356, "botgetpersonality" },
    { 0x8357, "botthrowgrenade" },
    { 0x8358, "botgetmemoryevents" },
    { 0x8359, "botsetpersonality" },
    { 0x835A, "botsetdifficulty" },
    { 0x835B, "botgetdifficulty" },
    { 0x835C, "botgetworldclosestedge" },
    { 0x835D, "botlookatpoint" },
    { 0x835E, "botpredictseepoint" },
    { 0x835F, "botcanseeentity" },
    { 0x8360, "botgetnodesonpath" },
    { 0x8361, "botnodepickmultiple" },
    { 0x8362, "botgetnearestnode" },
    { 0x8363, "botgetfovdot" },
    { 0x8364, "botsetawareness" },
    { 0x8365, "botpursuingscriptgoal" },
    { 0x8366, "botgetscriptgoalnode" },
    { 0x8367, "botgetimperfectenemyinfo" },
    { 0x8368, "botflagmemoryevents" },
    { 0x8369, "botsetpathingstyle" },
    { 0x836A, "botsetdifficultysetting" },
    { 0x836B, "botgetdifficultysetting" },
    { 0x836C, "botgetpathdist" },
    { 0x836D, "botisrandomized" },
    { 0x836E, "botpressbutton" },
    { 0x836F, "botclearbutton" },
    { 0x8370, "botnodescoremultiple" },
    { 0x8371, "getnodenumber" },
    { 0x8372, "setclientowner" },
    { 0x8373, "setotherent" },
    { 0x8374, "playercommandbot" },
    { 0x8375, "setaisightlinevisible" },
    { 0x8376, "setentityowner" },
    { 0x8377, "nodeisdisconnected" },
    { 0x8378, "getnearestnode" },
    { 0x8379, "makeentitynomeleetarget" },
    { 0x837A, "isagent" },
    { 0x837B, "spawnagent" },
    { 0x837C, "finishagentdamage" },
    { 0x837D, "setagentattacker" },
    { 0x837E, "cloneagent" },
    { 0x837F, "agentcanseesentient" },
    { 0x8380, "scragentsetwaypoint" },
    { 0x8381, "scragentsetgoalpos" },
    { 0x8382, "scragentgetgoalpos" },
    { 0x8383, "scragentsetgoalnode" },
    { 0x8384, "scragentsetgoalentity" },
    { 0x8385, "scragentsetgoalradius" },
    { 0x8386, "scragentsetanimscale" },
    { 0x8387, "scragentsetorientmode" },
    { 0x8388, "scragentsetanimmode" },
    { 0x8389, "scragentsetphysicsmode" },
    { 0x838A, "scragentsetclipmode" },
    { 0x838B, "scragentsetmaxturnspeed" },
    { 0x838C, "scragentgetmaxturnspeed" },
    { 0x838D, "scragentbeginmelee" },
    { 0x838E, "scragentsetscripted" },
    { 0x838F, "scragentdotrajectory" },
    { 0x8390, "scragentdoanimlerp" },
    { 0x8391, "scragentsetviewheight" },
    { 0x8392, "scragentclaimnode" },
    { 0x8393, "scragentrelinquishclaimednode" },
    { 0x8394, "setdoghandler" },
    { 0x8395, "setdogcommand" },
    { 0x8396, "setdogattackradius" },
    { 0x8397, "isdogfollowinghandler" },
    { 0x8398, "setdogmaxdrivespeed" },
    { 0x8399, "isdogbeingdriven" },
    { 0x839A, "setdogautoattackwhendriven" },
    { 0x839B, "getdogattackbeginnode" },
    { 0x839C, "setanimclass" },
    { 0x839D, "enableanimstate" },
    { 0x839E, "setanimstate" },
    { 0x839F, "getanimentry" },
    { 0x83A0, "getanimentryname" },
    { 0x83A1, "getanimentryalias" },
    { 0x83A2, "getanimentrycount" },
    { 0x83A3, "pushplayervector" },
    { 0x83A4, "issprinting" },
    { 0x83A5, "playerlinkeduselinkedvelocity" },
    { 0x83A6, "shootstopsound" },
    { 0x83A7, "setclothtype" },
    { 0x83A8, "getclothmovesound" },
    { 0x83A9, "getequipmovesound" },
    { 0x83AA, "jumpbuttonpressed" },
    { 0x83AB, "rotateby" },
    { 0x83AC, "getlookaheaddir" },
    { 0x83AD, "getpathgoalpos" },
    { 0x83AE, "setrocketcorpse" },
    { 0x83AF, "setcorpsefalling" },
    { 0x83B0, "setsurfacetype" },
    { 0x83B1, "aiphysicstrace" },
    { 0x83B2, "aiphysicstracepassed" },
    { 0x83B3, "setdevtext" },
    { 0x83B4, "forcemovingplatformentity" },
    { 0x83B5, "setmovingplatformtrigger" },
    { 0x83B6, "visionsetstage" },
    { 0x83B7, "linkwaypointtotargetwithoffset" },
    { 0x83B8, "getlinkedparent" },
    { 0x83B9, "getmovingplatformparent" },
    { 0x83BA, "setnameplatematerial" },
    { 0x83BB, "retargetscriptmodellighting" },
    { 0x83BC, "setmantlesoundflavor" },
    { 0x83BD, "clearclienttriggeraudiozone" },
    { 0x83BE, "setclienttriggeraudiozone" },
    { 0x83BF, "makevehiclenotcollidewithplayers" },
    { 0x83C0, "getbobrate" },
    { 0x83C1, "setbobrate" },
    { 0x83C2, "setscriptablepartstate" },
    { 0x83C3, "stopsliding" },
    { 0x83C4, "cancelrocketcorpse" },
    { 0x83C5, "setdronegoalpos" },
    { 0x83C6, "hudoutlineenable" },
    { 0x83C7, "hudoutlinedisable" },
    { 0x83C8, "motionblurhqenable" },
    { 0x83C9, "motionblurhqdisable" },
    { 0x83CA, "screenshakeonentity" },
    { 0x83CB, "setpitchorient" },
    { 0x83CC, "worldpointtoscreenpos" },
    { 0x83CD, "linktoplayerviewignoreparentrot" },
    { 0x83CE, "shouldplaymeleedeathanim" },
    { 0x83CF, "botfirstavailablegrenade" },
    { 0x83D0, "visionsetwaterforplayer" },
    { 0x83D1, "setwatersurfacetransitionfx" },
    { 0x83D2, "linktoplayerviewfollowwatersurface" },
    { 0x83D3, "linktoplayerviewattachwatersurfacetransitioneffects" },
    { 0x83D4, "playersetwaterfog" },
    { 0x83D5, "emissiveblend" },
    { 0x83D6, "enableforceviewmodeldof" },
    { 0x83D7, "disableforceviewmodeldof" },
    { 0x83D8, "getcustomizationbody" },
    { 0x83D9, "getcustomizationhead" },
    { 0x83DA, "getcustomizationviewmodel" },
    { 0x83DB, "setviewmodeldepth" },
    { 0x83DC, "isenemyaware" },
    { 0x83DD, "hasenemybeenseen" },
    { 0x83DE, "physicssetmaxlinvel" },
    { 0x83DF, "physicssetmaxangvel" },
    { 0x83E0, "physicsgetlinvel" },
    { 0x83E1, "physicsgetlinspeed" },
    { 0x83E2, "physicsgetangvel" },
    { 0x83E3, "physicsgetangspeed" },
    { 0x83E4, "disablemissileboosting" },
    { 0x83E5, "enablemissileboosting" },
    { 0x83E6, "canspawntestclient" },
    { 0x83E7, "spawntestclient" },
    { 0x83E8, "loadcustomizationplayerview" },
    { 0x83E9, "setgrenadethrowscale" },
    { 0x83EA, "setgrenadecookscale" },
    { 0x83EB, "setplanesplineid" },
    { 0x83EC, "hudoutlineenableforclient" },
    { 0x83ED, "hudoutlinedisableforclient" },
    { 0x83EE, "hudoutlineenableforclients" },
    { 0x83EF, "hudoutlinedisableforclients" },
    { 0x83F0, "turretsetbarrelspinenabled" },
    { 0x83F1, "hasloadedcustomizationplayerview" },
    { 0x83F2, "setclienttriggeraudiozonelerp" },
    { 0x83F3, "setclienttriggeraudiozonepartial" },
    { 0x83F4, "scragentdoanimrelative" },
    { 0x83F5, "rotatetolinked" },
    { 0x83F6, "rotatebylinked" },
    { 0x83F7, "setlinkedangles" },
    { 0x83F8, "getcorpseentity" },
    { 0x83F9, "removefrommovingplatformsystem" },
    { 0x83FA, "logmatchdatalife" },
    { 0x83FB, "logmatchdatadeath" },
    { 0x83FC, "queuedialogforplayer" },
    { 0x83FD, "setmlgcameradefaults" },
    { 0x83FE, "ismlgspectator" },
    { 0x83FF, "disableautoreload" },
    { 0x8400, "enableautoreload" },
    { 0x8401, "issprintsliding" },
    { 0x8402, "getlinkedchildren" },
    { 0x8403, "botpredictenemycampspots" },
    { 0x8404, "playsoundonmovingent" },
    { 0x8405, "cancelmantle" },
    { 0x8406, "hasfemalecustomizationmodel" },
    { 0x8407, "setscriptabledamageowner" },
    { 0x8408, "setfxkilldefondelete" },
    { 0x8409, "getdogavoident" },
    { 0x840A, "enabledogavoidance" },
    { 0x840B, "forcedeathfall" },
    { 0x840C, "gethybridscopestate" },
    { 0x840D, "sethybridscopestate" },
    { 0x840E, "getvieworigin" },
    { 0x840F, "setweaponmodelvariant" },
    { 0x8410, "ridevehicle" },
    { 0x8411, "stopridingvehicle" },
    { 0x8412, "getmantlesound" },
    { 0x8413, "autoboltmissileeffects" },
    { 0x8414, "disablemissilestick" },
    { 0x8415, "enablemissilestick" },
    { 0x8416, "setmissileminimapvisible" },
    { 0x8417, "isoffhandweaponreadytothrow" },
    { 0x8418, "isleaning" },
    { 0x8419, "makecollidewithitemclip" },
    { 0x841A, "ismovementfromgamepad" },
    { 0x841B, "visionsetpostapplyforplayer" },
    { 0x841C, "setcommonplayerdatareservedint" },
    { 0x841D, "getclanwarsbonus" },
    { 0x841E, "getrankedplayerdatareservedint" },
    { 0x841F, "setrankedplayerdatareservedint" },
    { 0x8420, "setignorefoliagesightingme" },
    { 0x8421, "scragentusemodelcollisionbounds" },
    { 0x8422, "setmlgspectator" },
    { 0x8423, "getmlgspectatorteam" },
    { 0x8424, "getcommonplayerdatareservedint" },
    { 0x8425, "getcoopplayerdatareservedint" },
    { 0x8426, "setcoopplayerdatareservedint" },
    { 0x8427, "getjointype" },
    { 0x8428, "scragentsetwallruncost" },
    { 0x8429, "alienscheckisitempurchased" },
}};

const std::array<std::pair<std::uint16_t, const char*>, 831> file_list
{{
    { 0x301, "character/character_ajax_flood_a" },
    { 0x302, "character/character_ajax_wood_shotgun_a" },
    { 0x303, "character/character_ajax_wood_shotgun_b" },
    { 0x304, "character/character_ajax_wood_shotgun_injured" },
    { 0x305, "character/character_almagro_assault_a" },
    { 0x306, "character/character_chemwar_russian_assault_b" },
    { 0x307, "character/character_chemwar_russian_assault_bb" },
    { 0x308, "character/character_chemwar_russian_assault_c" },
    { 0x309, "character/character_chemwar_russian_assault_cc" },
    { 0x30A, "character/character_chemwar_russian_assault_d" },
    { 0x30B, "character/character_chemwar_russian_assault_dd" },
    { 0x30C, "character/character_chemwar_russian_assault_e" },
    { 0x30D, "character/character_chemwar_russian_assault_ee" },
    { 0x30E, "character/character_deer_a" },
    { 0x30F, "character/character_deer_b" },
    { 0x310, "character/character_deer_c" },
    { 0x311, "character/character_elias_a" },
    { 0x312, "character/character_elias_assault_a" },
    { 0x313, "character/character_elias_b" },
    { 0x314, "character/character_elias_e" },
    { 0x315, "character/character_elias_hc" },
    { 0x316, "character/character_elias_wood_shotgun_b" },
    { 0x317, "character/character_elite_pmc_assault_a" },
    { 0x318, "character/character_elite_pmc_assault_a_black" },
    { 0x319, "character/character_elite_pmc_assault_a_white" },
    { 0x31A, "character/character_elite_pmc_assault_b" },
    { 0x31B, "character/character_elite_pmc_assault_b_desert" },
    { 0x31C, "character/character_elite_pmc_lmg_b_desert" },
    { 0x31D, "character/character_elite_pmc_shotgun_b_desert" },
    { 0x31E, "character/character_elite_pmc_skyway" },
    { 0x31F, "character/character_elite_pmc_smg_b_desert" },
    { 0x320, "character/character_enemy_shark" },
    { 0x321, "character/character_enemy_wolf" },
    { 0x322, "character/character_fed_army_assault_a" },
    { 0x323, "character/character_fed_army_assault_a_arctic" },
    { 0x324, "character/character_fed_army_assault_a_elite" },
    { 0x325, "character/character_fed_army_assault_a_nohelmet" },
    { 0x326, "character/character_fed_army_assault_a_urban" },
    { 0x327, "character/character_fed_army_assault_b" },
    { 0x328, "character/character_fed_army_assault_b_arctic" },
    { 0x329, "character/character_fed_army_assault_b_elite" },
    { 0x32A, "character/character_fed_army_assault_b_nohelmet" },
    { 0x32B, "character/character_fed_army_assault_b_urban" },
    { 0x32C, "character/character_fed_army_drones" },
    { 0x32D, "character/character_fed_army_lmg_a" },
    { 0x32E, "character/character_fed_army_lmg_a_arctic" },
    { 0x32F, "character/character_fed_army_lmg_a_urban" },
    { 0x330, "character/character_fed_army_shotgun_a" },
    { 0x331, "character/character_fed_army_shotgun_a_arctic" },
    { 0x332, "character/character_fed_army_shotgun_a_urban" },
    { 0x333, "character/character_fed_army_smg_a" },
    { 0x334, "character/character_fed_army_smg_a_arctic" },
    { 0x335, "character/character_fed_army_smg_a_nohelmet" },
    { 0x336, "character/character_fed_army_smg_a_urban" },
    { 0x337, "character/character_fed_basic_assault_a" },
    { 0x338, "character/character_fed_basic_smg_a" },
    { 0x339, "character/character_fed_space_assault_a" },
    { 0x33A, "character/character_fed_space_assault_b" },
    { 0x33B, "character/character_fed_udt_assault_a" },
    { 0x33C, "character/character_hazmat_a" },
    { 0x33D, "character/character_hero_flood_vargas_test" },
    { 0x33E, "character/character_hesh_assault_a" },
    { 0x33F, "character/character_hesh_cornered_b" },
    { 0x340, "character/character_hesh_end_a" },
    { 0x341, "character/character_hesh_end_a_water" },
    { 0x342, "character/character_hesh_end_b" },
    { 0x343, "character/character_hesh_fed_shotgun_a" },
    { 0x344, "character/character_hesh_hostage_ab" },
    { 0x345, "character/character_hesh_ranger_assault_a" },
    { 0x346, "character/character_hesh_ranger_assault_b" },
    { 0x347, "character/character_hesh_stealth_mask_c" },
    { 0x348, "character/character_hesh_stealth_mask_d" },
    { 0x349, "character/character_hesh_udt_assault_a" },
    { 0x34A, "character/character_hesh_young_a" },
    { 0x34B, "character/character_iw6_sp_enemy_dog" },
    { 0x34C, "character/character_iw6_sp_german_shepherd_dog_a" },
    { 0x34D, "character/character_iw6_sp_german_shepherd_dog_b" },
    { 0x34E, "character/character_iw6_sp_german_shepherd_dog_c" },
    { 0x34F, "character/character_keegan_assault_a" },
    { 0x350, "character/character_keegan_end_smg_a" },
    { 0x351, "character/character_keegan_fed_smg_a" },
    { 0x352, "character/character_keegan_udt_assault_a" },
    { 0x353, "character/character_keegan_udt_water_b" },
    { 0x354, "character/character_keegan_wood_sniper_b" },
    { 0x355, "character/character_kick_udt_assault_b" },
    { 0x356, "character/character_kick_udt_water_glowstick" },
    { 0x357, "character/character_kyra_us_space_a" },
    { 0x358, "character/character_merrick_assault_a" },
    { 0x359, "character/character_merrick_assault_b" },
    { 0x35A, "character/character_merrick_assault_c" },
    { 0x35B, "character/character_merrick_end_a" },
    { 0x35C, "character/character_merrick_fed_assault_a" },
    { 0x35D, "character/character_merrick_flood_a" },
    { 0x35E, "character/character_merrick_hostage_a" },
    { 0x35F, "character/character_merrick_udt_assault_b" },
    { 0x360, "character/character_merrick_udt_water_glowstick" },
    { 0x361, "character/character_merrick_wood_smg_c_skull" },
    { 0x362, "character/character_oil_worker" },
    { 0x363, "character/character_oil_worker_bi_foreman" },
    { 0x364, "character/character_oil_worker_mask" },
    { 0x365, "character/character_opforce_henchmen_smg_a" },
    { 0x366, "character/character_opforce_henchmen_smg_b" },
    { 0x367, "character/character_pilot_a" },
    { 0x368, "character/character_pilot_b" },
    { 0x369, "character/character_pilot_b_tan" },
    { 0x36A, "character/character_pilot_c" },
    { 0x36B, "character/character_pilot_c_blue" },
    { 0x36C, "character/character_pilot_c_green" },
    { 0x36D, "character/character_pilot_c_purple" },
    { 0x36E, "character/character_pilot_c_red" },
    { 0x36F, "character/character_ppilot_crew_d" },
    { 0x370, "character/character_ramos_a" },
    { 0x371, "character/character_rorke_assault" },
    { 0x372, "character/character_rorke_basic_a" },
    { 0x373, "character/character_rorke_flood_a" },
    { 0x374, "character/character_scientist_a" },
    { 0x375, "character/character_scientist_b" },
    { 0x376, "character/character_scientist_c" },
    { 0x377, "character/character_scientist_d" },
    { 0x378, "character/character_us_civ_female_yb_a" },
    { 0x379, "character/character_us_civ_female_yb_b" },
    { 0x37A, "character/character_us_civ_male_yb_a" },
    { 0x37B, "character/character_us_civ_male_yb_b" },
    { 0x37C, "character/character_us_rangers_assault_a" },
    { 0x37D, "character/character_us_rangers_assault_a_desert" },
    { 0x37E, "character/character_us_rangers_assault_a_urban" },
    { 0x37F, "character/character_us_rangers_drones" },
    { 0x380, "character/character_us_rangers_lmg_a" },
    { 0x381, "character/character_us_rangers_lmg_a_urban" },
    { 0x382, "character/character_us_rangers_shotgun_a" },
    { 0x383, "character/character_us_rangers_shotgun_a_urban" },
    { 0x384, "character/character_us_rangers_smg_a" },
    { 0x385, "character/character_us_rangers_smg_a_desert" },
    { 0x386, "character/character_us_space_assault_a" },
    { 0x387, "character/character_us_space_assault_b" },
    { 0x388, "character/character_us_space_collins" },
    { 0x389, "character/character_us_space_int_a" },
    { 0x38A, "character/character_us_space_int_b" },
    { 0x38B, "character/character_us_space_int_c" },
    { 0x38C, "character/character_venezuela_army_assault_a" },
    { 0x38D, "character/character_venezuela_army_smg_a" },
    { 0x38E, "character/character_venezuela_army_smg_a_head_d" },
    { 0x38F, "codescripts/character" },
    { 0x390, "codescripts/delete" },
    { 0x391, "codescripts/struct" },
    { 0x392, "common_scripts/_artcommon" },
    { 0x393, "common_scripts/_bcs_location_trigs" },
    { 0x394, "common_scripts/_createfx" },
    { 0x395, "common_scripts/_createfxmenu" },
    { 0x396, "common_scripts/_csplines" },
    { 0x397, "common_scripts/_destructible" },
    { 0x398, "common_scripts/_dynamic_world" },
    { 0x399, "common_scripts/_elevator" },
    { 0x39A, "common_scripts/_exploder" },
    { 0x39B, "common_scripts/_fx" },
    { 0x39C, "common_scripts/_pipes" },
    { 0x39D, "common_scripts/_sentry" },
    { 0x39E, "common_scripts/_wind" },
    { 0x39F, "common_scripts/utility" },
    { 0x3A0, "destructible_scripts/destructible_civilian_sedan_iw6" },
    { 0x3A1, "destructible_scripts/destructible_civilian_sedan_water_iw6" },
    { 0x3A2, "destructible_scripts/destructible_van_water_iw6" },
    { 0x3A3, "destructible_scripts/destructible_vehicle_city_car_water" },
    { 0x3A4, "destructible_scripts/electrical_transformer_large" },
    { 0x3A5, "destructible_scripts/explodable_barrel" },
    { 0x3A6, "destructible_scripts/toy_electricbox2" },
    { 0x3A7, "destructible_scripts/toy_filecabinet" },
    { 0x3A8, "destructible_scripts/toy_generator_on" },
    { 0x3A9, "destructible_scripts/toy_light_ceiling_fluorescent" },
    { 0x3AA, "destructible_scripts/toy_lv_trash_can_vegas" },
    { 0x3AB, "destructible_scripts/toy_sp_panel_box" },
    { 0x3AC, "destructible_scripts/toy_transformer_small01" },
    { 0x3AD, "destructible_scripts/toy_trashcan_metal_closed" },
    { 0x3AE, "destructible_scripts/toy_tv_flatscreen" },
    { 0x3AF, "destructible_scripts/toy_tv_flatscreen_wallmount_02" },
    { 0x3B0, "destructible_scripts/toy_tv_video_monitor" },
    { 0x3B1, "destructible_scripts/toy_usa_gas_station_trash_bin_02" },
    { 0x3B2, "maps/_ambient" },
    { 0x3B3, "maps/_anim" },
    { 0x3B4, "maps/_animatedmodels" },
    { 0x3B5, "maps/_art" },
    { 0x3B6, "maps/_ash_falling" },
    { 0x3B7, "maps/_audio" },
    { 0x3B8, "maps/_audio_ambient" },
    { 0x3B9, "maps/_audio_code" },
    { 0x3BA, "maps/_autosave" },
    { 0x3BB, "maps/_breach" },
    { 0x3BC, "maps/_c4" },
    { 0x3BD, "maps/_chaingun_player" },
    { 0x3BE, "maps/_chopperboss" },
    { 0x3BF, "maps/_chopperboss_utility" },
    { 0x3C0, "maps/_colors" },
    { 0x3C1, "maps/_compass" },
    { 0x3C2, "maps/_coop" },
    { 0x3C3, "maps/_createfx" },
    { 0x3C4, "maps/_credits_iw6" },
    { 0x3C5, "maps/_damagefeedback" },
    { 0x3C6, "maps/_deadbody" },
    { 0x3C7, "maps/_debug" },
    { 0x3C8, "maps/_detonategrenades" },
    { 0x3C9, "maps/_dog_control" },
    { 0x3CA, "maps/_dog_drive" },
    { 0x3CB, "maps/_dog_kinect" },
    { 0x3CC, "maps/_dog_pip" },
    { 0x3CD, "maps/_drone" },
    { 0x3CE, "maps/_drone_ai" },
    { 0x3CF, "maps/_drone_base" },
    { 0x3D0, "maps/_drone_civilian" },
    { 0x3D1, "maps/_drone_deer" },
    { 0x3D2, "maps/_dynamic_run_speed" },
    { 0x3D3, "maps/_endmission" },
    { 0x3D4, "maps/_friendlyfire" },
    { 0x3D5, "maps/_gameskill" },
    { 0x3D6, "maps/_geo_mover" },
    { 0x3D7, "maps/_global_fx" },
    { 0x3D8, "maps/_global_fx_code" },
    { 0x3D9, "maps/_hand_signals" },
    { 0x3DA, "maps/_helicopter_ai" },
    { 0x3DB, "maps/_helicopter_globals" },
    { 0x3DC, "maps/_hud" },
    { 0x3DD, "maps/_hud_util" },
    { 0x3DE, "maps/_idle" },
    { 0x3DF, "maps/_idle_coffee" },
    { 0x3E0, "maps/_idle_lean_smoke" },
    { 0x3E1, "maps/_idle_phone" },
    { 0x3E2, "maps/_idle_sit_load_ak" },
    { 0x3E3, "maps/_idle_sleep" },
    { 0x3E4, "maps/_idle_smoke" },
    { 0x3E5, "maps/_idle_smoke_balcony" },
    { 0x3E6, "maps/_intelligence" },
    { 0x3E7, "maps/_introscreen" },
    { 0x3E8, "maps/_inventory" },
    { 0x3E9, "maps/_lights" },
    { 0x3EA, "maps/_load" },
    { 0x3EB, "maps/_loadout" },
    { 0x3EC, "maps/_loadout_code" },
    { 0x3ED, "maps/_mg_penetration" },
    { 0x3EE, "maps/_mgturret" },
    { 0x3EF, "maps/_minigun" },
    { 0x3F0, "maps/_minigun_viewmodel" },
    { 0x3F1, "maps/_mocap_ar" },
    { 0x3F2, "maps/_mortar" },
    { 0x3F3, "maps/_names" },
    { 0x3F4, "maps/_nightvision" },
    { 0x3F5, "maps/_ocean" },
    { 0x3F6, "maps/_patrol" },
    { 0x3F7, "maps/_patrol_anims" },
    { 0x3F8, "maps/_patrol_anims_creepwalk" },
    { 0x3F9, "maps/_patrol_anims_gundown" },
    { 0x3FA, "maps/_patrol_anims_patroljog" },
    { 0x3FB, "maps/_perlin_noise" },
    { 0x3FC, "maps/_player_death" },
    { 0x3FD, "maps/_player_limp" },
    { 0x3FE, "maps/_player_rig" },
    { 0x3FF, "maps/_player_stats" },
    { 0x400, "maps/_props" },
    { 0x401, "maps/_radiation" },
    { 0x402, "maps/_rambo" },
    { 0x403, "maps/_readystand_anims" },
    { 0x404, "maps/_remoteturret" },
    { 0x405, "maps/_rv_vfx" },
    { 0x406, "maps/_shellshock" },
    { 0x407, "maps/_shg_common" },
    { 0x408, "maps/_slowmo_breach" },
    { 0x409, "maps/_space" },
    { 0x40A, "maps/_space_ai" },
    { 0x40B, "maps/_space_player" },
    { 0x40C, "maps/_spawner" },
    { 0x40D, "maps/_stealth" },
    { 0x40E, "maps/_stealth_accuracy_friendly" },
    { 0x40F, "maps/_stealth_animation_funcs" },
    { 0x410, "maps/_stealth_anims" },
    { 0x411, "maps/_stealth_behavior_enemy" },
    { 0x412, "maps/_stealth_behavior_friendly" },
    { 0x413, "maps/_stealth_behavior_system" },
    { 0x414, "maps/_stealth_color_friendly" },
    { 0x415, "maps/_stealth_corpse_enemy" },
    { 0x416, "maps/_stealth_corpse_system" },
    { 0x417, "maps/_stealth_event_enemy" },
    { 0x418, "maps/_stealth_shared_utilities" },
    { 0x419, "maps/_stealth_smartstance_friendly" },
    { 0x41A, "maps/_stealth_threat_enemy" },
    { 0x41B, "maps/_stealth_utility" },
    { 0x41C, "maps/_stealth_visibility_enemy" },
    { 0x41D, "maps/_stealth_visibility_friendly" },
    { 0x41E, "maps/_stealth_visibility_system" },
    { 0x41F, "maps/_swim_ai" },
    { 0x420, "maps/_swim_ai_common" },
    { 0x421, "maps/_swim_player" },
    { 0x422, "maps/_tank_battlechatter" },
    { 0x423, "maps/_teargas" },
    { 0x424, "maps/_treadfx" },
    { 0x425, "maps/_trigger" },
    { 0x426, "maps/_underwater" },
    { 0x427, "maps/_utility" },
    { 0x428, "maps/_utility_code" },
    { 0x429, "maps/_utility_dogs" },
    { 0x42A, "maps/_vehicle" },
    { 0x42B, "maps/_vehicle_aianim" },
    { 0x42C, "maps/_vehicle_code" },
    { 0x42D, "maps/_vehicle_spline_zodiac" },
    { 0x42E, "maps/_vignette_util" },
    { 0x42F, "maps/_weather" },
    { 0x430, "maps/black_ice_fx" },
    { 0x431, "maps/black_ice_precache" },
    { 0x432, "maps/carrier_fx" },
    { 0x433, "maps/carrier_precache" },
    { 0x434, "maps/clockwork_fx" },
    { 0x435, "maps/clockwork_precache" },
    { 0x436, "maps/cornered_fx" },
    { 0x437, "maps/cornered_precache" },
    { 0x438, "maps/deer_hunt_fx" },
    { 0x439, "maps/deer_hunt_precache" },
    { 0x43A, "maps/enemyhq_fx" },
    { 0x43B, "maps/enemyhq_precache" },
    { 0x43C, "maps/flood_fx" },
    { 0x43D, "maps/flood_precache" },
    { 0x43E, "maps/homecoming_fx" },
    { 0x43F, "maps/homecoming_precache" },
    { 0x440, "maps/iplane_fx" },
    { 0x441, "maps/jungle_ghosts_fx" },
    { 0x442, "maps/jungle_ghosts_precache" },
    { 0x443, "maps/loki_fx" },
    { 0x444, "maps/loki_precache" },
    { 0x445, "maps/nml_fx" },
    { 0x446, "maps/nml_precache" },
    { 0x447, "maps/odin_fx" },
    { 0x448, "maps/odin_precache" },
    { 0x449, "maps/oilrocks_fx" },
    { 0x44A, "maps/oilrocks_precache" },
    { 0x44B, "maps/prologue_fx" },
    { 0x44C, "maps/prologue_precache" },
    { 0x44D, "maps/satfarm_b_fx" },
    { 0x44E, "maps/satfarm_b_precache" },
    { 0x44F, "maps/satfarm_fx" },
    { 0x450, "maps/satfarm_precache" },
    { 0x451, "maps/ship_graveyard_fx" },
    { 0x452, "maps/ship_graveyard_precache" },
    { 0x453, "maps/skyway_fx" },
    { 0x454, "maps/skyway_precache" },
    { 0x455, "maps/youngblood_fx" },
    { 0x456, "maps/youngblood_precache" },
    { 0x457, "vehicle_scripts/_a10_warthog" },
    { 0x458, "vehicle_scripts/_apache" },
    { 0x459, "vehicle_scripts/_apache_player" },
    { 0x45A, "vehicle_scripts/_apache_player_audio" },
    { 0x45B, "vehicle_scripts/_apache_player_difficulty" },
    { 0x45C, "vehicle_scripts/_apache_player_missile_hydra_and_lockon" },
    { 0x45D, "vehicle_scripts/_apache_player_pilot" },
    { 0x45E, "vehicle_scripts/_apache_player_raining_missile" },
    { 0x45F, "vehicle_scripts/_apache_player_targeting" },
    { 0x460, "vehicle_scripts/_attack_heli" },
    { 0x461, "vehicle_scripts/_btr80" },
    { 0x462, "vehicle_scripts/_c17" },
    { 0x463, "vehicle_scripts/_chinese_brave_warrior" },
    { 0x464, "vehicle_scripts/_chopper_ai_missile_defense" },
    { 0x465, "vehicle_scripts/_chopper_missile_defense_utility" },
    { 0x466, "vehicle_scripts/_chopper_player_missile_defense" },
    { 0x467, "vehicle_scripts/_empty" },
    { 0x468, "vehicle_scripts/_f15" },
    { 0x469, "vehicle_scripts/_f18" },
    { 0x46A, "vehicle_scripts/_foliage_tumbleweed_vehicle" },
    { 0x46B, "vehicle_scripts/_gaz" },
    { 0x46C, "vehicle_scripts/_gaz_dshk_oilrocks" },
    { 0x46D, "vehicle_scripts/_gunboat" },
    { 0x46E, "vehicle_scripts/_hind" },
    { 0x46F, "vehicle_scripts/_lcs" },
    { 0x470, "vehicle_scripts/_m1a1_minigun" },
    { 0x471, "vehicle_scripts/_m880_launcher" },
    { 0x472, "vehicle_scripts/_man_7t" },
    { 0x473, "vehicle_scripts/_mig29" },
    { 0x474, "vehicle_scripts/_missile_boat" },
    { 0x475, "vehicle_scripts/_osprey" },
    { 0x476, "vehicle_scripts/_snowmobile" },
    { 0x477, "vehicle_scripts/_snowmobile_friendly" },
    { 0x478, "vehicle_scripts/_soc_r" },
    { 0x479, "vehicle_scripts/_submarine_sdv" },
    { 0x47A, "vehicle_scripts/_tank_crush" },
    { 0x47B, "vehicle_scripts/_zodiac" },
    { 0x47C, "vehicle_scripts/_zpu_antiair" },
    { 0x47D, "vehicle_scripts/_zpu_antiair_oilrocks" },
    { 0x47E, "vehicle_scripts/aas_72x" },
    { 0x47F, "vehicle_scripts/apache" },
    { 0x480, "vehicle_scripts/artemis" },
    { 0x481, "vehicle_scripts/hind" },
    { 0x482, "vehicle_scripts/hind_battle_carrier" },
    { 0x483, "vehicle_scripts/hind_battle_oilrocks" },
    { 0x484, "vehicle_scripts/hovercraft" },
    { 0x485, "vehicle_scripts/iveco_lynx" },
    { 0x486, "vehicle_scripts/iveco_lynx_turret" },
    { 0x487, "vehicle_scripts/m1a2" },
    { 0x488, "vehicle_scripts/m1a2_player" },
    { 0x489, "vehicle_scripts/m800" },
    { 0x48A, "vehicle_scripts/matv" },
    { 0x48B, "vehicle_scripts/mk23" },
    { 0x48C, "vehicle_scripts/nh90" },
    { 0x48D, "vehicle_scripts/pickup_truck_civ" },
    { 0x48E, "vehicle_scripts/sedan_4door" },
    { 0x48F, "vehicle_scripts/silenthawk" },
    { 0x490, "vehicle_scripts/silenthawk_landing" },
    { 0x491, "vehicle_scripts/t90ms" },
    { 0x492, "vehicle_scripts/tatra_t815" },
    { 0x493, "vehicle_scripts/y_8_gunship" },
    { 0x494, "xmodelalias/alias_chemwar_russian_heads" },
    { 0x495, "xmodelalias/alias_elite_pmc_heads" },
    { 0x496, "xmodelalias/alias_fed_army_heads_a" },
    { 0x497, "xmodelalias/alias_fed_army_heads_a_arctic_noshield" },
    { 0x498, "xmodelalias/alias_fed_army_heads_a_nofaceshield" },
    { 0x499, "xmodelalias/alias_fed_army_heads_a_urban" },
    { 0x49A, "xmodelalias/alias_fed_basic_heads" },
    { 0x49B, "xmodelalias/alias_fed_space_assault_heads" },
    { 0x49C, "xmodelalias/alias_fed_udt_heads" },
    { 0x49D, "xmodelalias/alias_henchmen_heads" },
    { 0x49E, "xmodelalias/alias_oil_worker_bodies" },
    { 0x49F, "xmodelalias/alias_oil_worker_heads" },
    { 0x4A0, "xmodelalias/alias_pilot_heads" },
    { 0x4A1, "xmodelalias/alias_pilot_heads_blue" },
    { 0x4A2, "xmodelalias/alias_pilot_heads_green" },
    { 0x4A3, "xmodelalias/alias_pilot_heads_purple" },
    { 0x4A4, "xmodelalias/alias_pilot_heads_red" },
    { 0x4A5, "xmodelalias/alias_pilot_heads_yellow" },
    { 0x4A6, "xmodelalias/alias_russian_military_heads_skyway" },
    { 0x4A7, "xmodelalias/alias_scientist_heads" },
    { 0x4A8, "xmodelalias/alias_us_rangers_drone_bodies" },
    { 0x4A9, "xmodelalias/alias_us_rangers_drone_heads" },
    { 0x4AA, "xmodelalias/alias_us_rangers_heads_a" },
    { 0x4AB, "xmodelalias/alias_us_rangers_heads_a_desert" },
    { 0x4AC, "xmodelalias/alias_us_rangers_heads_a_urban" },
    { 0x4AD, "xmodelalias/alias_us_space_assault_heads" },
    { 0x4AE, "xmodelalias/alias_us_space_int_heads" },
    { 0x4AF, "xmodelalias/alias_venezuela_army_heads" },
    { 0x4B0, "maps/animated_models/accessories_windsock_wind_medium" },
    { 0x4B1, "maps/animated_models/com_roofvent2" },
    { 0x4B2, "maps/animated_models/com_roofvent3" },
    { 0x4B3, "maps/animated_models/flood_palm_tree_tall" },
    { 0x4B4, "maps/animated_models/flood_palm_tree_tall_no_shadow" },
    { 0x4B5, "maps/animated_models/hanging_sheet_wind_medium" },
    { 0x4B6, "maps/animated_models/mp_flooded_water_debris_bob" },
    { 0x4B7, "maps/animated_models/mp_flooded_water_debris_spiral" },
    { 0x4B8, "maps/animated_models/mp_flooded_water_street" },
    { 0x4B9, "maps/animated_models/mp_frag_crane" },
    { 0x4BA, "maps/animated_models/ow_crane_hook" },
    { 0x4BB, "maps/animated_models/tarp_tattered_thin_02" },
    { 0x4BC, "maps/animated_models/tattered_cloth_medium_01" },
    { 0x4BD, "maps/animated_models/tattered_cloth_small_02" },
    { 0x4BE, "maps/createart/black_ice_art" },
    { 0x4BF, "maps/createart/black_ice_fog" },
    { 0x4C0, "maps/createart/carrier_art" },
    { 0x4C1, "maps/createart/carrier_fog" },
    { 0x4C2, "maps/createart/clockwork_art" },
    { 0x4C3, "maps/createart/clockwork_fog" },
    { 0x4C4, "maps/createart/cornered_art" },
    { 0x4C5, "maps/createart/cornered_fog" },
    { 0x4C6, "maps/createart/deer_hunt_art" },
    { 0x4C7, "maps/createart/deer_hunt_fog" },
    { 0x4C8, "maps/createart/enemyhq_art" },
    { 0x4C9, "maps/createart/enemyhq_fog" },
    { 0x4CA, "maps/createart/flood_art" },
    { 0x4CB, "maps/createart/flood_fog" },
    { 0x4CC, "maps/createart/homecoming_art" },
    { 0x4CD, "maps/createart/homecoming_fog" },
    { 0x4CE, "maps/createart/iplane_art" },
    { 0x4CF, "maps/createart/iplane_fog" },
    { 0x4D0, "maps/createart/jungle_ghosts_art" },
    { 0x4D1, "maps/createart/jungle_ghosts_fog" },
    { 0x4D2, "maps/createart/loki_art" },
    { 0x4D3, "maps/createart/loki_fog" },
    { 0x4D4, "maps/createart/mp_alien_town_art" },
    { 0x4D5, "maps/createart/mp_alien_town_fog" },
    { 0x4D6, "maps/createart/mp_chasm_art" },
    { 0x4D7, "maps/createart/mp_chasm_fog" },
    { 0x4D8, "maps/createart/mp_dart_art" },
    { 0x4D9, "maps/createart/mp_dart_fog" },
    { 0x4DA, "maps/createart/mp_fahrenheit_art" },
    { 0x4DB, "maps/createart/mp_fahrenheit_fog" },
    { 0x4DC, "maps/createart/mp_flooded_art" },
    { 0x4DD, "maps/createart/mp_flooded_fog" },
    { 0x4DE, "maps/createart/mp_frag_art" },
    { 0x4DF, "maps/createart/mp_frag_fog" },
    { 0x4E0, "maps/createart/mp_hashima_art" },
    { 0x4E1, "maps/createart/mp_hashima_fog" },
    { 0x4E2, "maps/createart/mp_lonestar_art" },
    { 0x4E3, "maps/createart/mp_lonestar_fog" },
    { 0x4E4, "maps/createart/mp_prisonbreak_art" },
    { 0x4E5, "maps/createart/mp_prisonbreak_fog" },
    { 0x4E6, "maps/createart/mp_siege_buildings_fog" },
    { 0x4E7, "maps/createart/mp_skeleton_art" },
    { 0x4E8, "maps/createart/mp_skeleton_fog" },
    { 0x4E9, "maps/createart/mp_snow_art" },
    { 0x4EA, "maps/createart/mp_snow_fog" },
    { 0x4EB, "maps/createart/mp_sovereign_art" },
    { 0x4EC, "maps/createart/mp_sovereign_fog" },
    { 0x4ED, "maps/createart/mp_strikezone_art" },
    { 0x4EE, "maps/createart/mp_strikezone_fog" },
    { 0x4EF, "maps/createart/mp_warhawk_art" },
    { 0x4F0, "maps/createart/mp_warhawk_fog" },
    { 0x4F1, "maps/createart/mp_zebra_art" },
    { 0x4F2, "maps/createart/mp_zebra_fog" },
    { 0x4F3, "maps/createart/nml_art" },
    { 0x4F4, "maps/createart/nml_fog" },
    { 0x4F5, "maps/createart/odin_art" },
    { 0x4F6, "maps/createart/odin_fog" },
    { 0x4F7, "maps/createart/oilrocks_art" },
    { 0x4F8, "maps/createart/oilrocks_fog" },
    { 0x4F9, "maps/createart/prologue_art" },
    { 0x4FA, "maps/createart/prologue_fog" },
    { 0x4FB, "maps/createart/satfarm_art" },
    { 0x4FC, "maps/createart/satfarm_b_art" },
    { 0x4FD, "maps/createart/satfarm_b_fog" },
    { 0x4FE, "maps/createart/satfarm_fog" },
    { 0x4FF, "maps/createart/ship_graveyard_art" },
    { 0x500, "maps/createart/ship_graveyard_fog" },
    { 0x501, "maps/createart/skyway_art" },
    { 0x502, "maps/createart/skyway_fog" },
    { 0x503, "maps/createart/youngblood_art" },
    { 0x504, "maps/createart/youngblood_fog" },
    { 0x505, "maps/createfx/black_ice_fx" },
    { 0x506, "maps/createfx/black_ice_sound" },
    { 0x507, "maps/createfx/carrier_fx" },
    { 0x508, "maps/createfx/carrier_sound" },
    { 0x509, "maps/createfx/clockwork_fx" },
    { 0x50A, "maps/createfx/clockwork_sound" },
    { 0x50B, "maps/createfx/cornered_fx" },
    { 0x50C, "maps/createfx/cornered_sound" },
    { 0x50D, "maps/createfx/deer_hunt_fx" },
    { 0x50E, "maps/createfx/deer_hunt_sound" },
    { 0x50F, "maps/createfx/enemyhq_fx" },
    { 0x510, "maps/createfx/enemyhq_sound" },
    { 0x511, "maps/createfx/flood_fx" },
    { 0x512, "maps/createfx/flood_sound" },
    { 0x513, "maps/createfx/homecoming_fx" },
    { 0x514, "maps/createfx/homecoming_sound" },
    { 0x515, "maps/createfx/jungle_ghosts_fx" },
    { 0x516, "maps/createfx/jungle_ghosts_sound" },
    { 0x517, "maps/createfx/loki_fx" },
    { 0x518, "maps/createfx/loki_sound" },
    { 0x519, "maps/createfx/mp_alien_town_fx" },
    { 0x51A, "maps/createfx/mp_chasm_fx" },
    { 0x51B, "maps/createfx/mp_dart_fx" },
    { 0x51C, "maps/createfx/mp_fahrenheit_fx" },
    { 0x51D, "maps/createfx/mp_flooded_fx" },
    { 0x51E, "maps/createfx/mp_frag_fx" },
    { 0x51F, "maps/createfx/mp_hashima_fx" },
    { 0x520, "maps/createfx/mp_lonestar_fx" },
    { 0x521, "maps/createfx/mp_prisonbreak_fx" },
    { 0x522, "maps/createfx/mp_skeleton_fx" },
    { 0x523, "maps/createfx/mp_snow_fx" },
    { 0x524, "maps/createfx/mp_sovereign_fx" },
    { 0x525, "maps/createfx/mp_strikezone_fx" },
    { 0x526, "maps/createfx/mp_warhawk_fx" },
    { 0x527, "maps/createfx/mp_zebra_fx" },
    { 0x528, "maps/createfx/nml_fx" },
    { 0x529, "maps/createfx/nml_sound" },
    { 0x52A, "maps/createfx/odin_fx" },
    { 0x52B, "maps/createfx/odin_sound" },
    { 0x52C, "maps/createfx/oilrocks_fx" },
    { 0x52D, "maps/createfx/oilrocks_sound" },
    { 0x52E, "maps/createfx/prologue_fx" },
    { 0x52F, "maps/createfx/prologue_sound" },
    { 0x530, "maps/createfx/satfarm_b_fx" },
    { 0x531, "maps/createfx/satfarm_b_sound" },
    { 0x532, "maps/createfx/satfarm_fx" },
    { 0x533, "maps/createfx/satfarm_sound" },
    { 0x534, "maps/createfx/ship_graveyard_fx" },
    { 0x535, "maps/createfx/ship_graveyard_sound" },
    { 0x536, "maps/createfx/skyway_fx" },
    { 0x537, "maps/createfx/skyway_sound" },
    { 0x538, "maps/createfx/youngblood_fx" },
    { 0x539, "maps/createfx/youngblood_sound" },
    { 0x53A, "maps/interactive_models/_birds" },
    { 0x53B, "maps/interactive_models/_fish" },
    { 0x53C, "maps/interactive_models/_interactive_utility" },
    { 0x53D, "maps/interactive_models/_interactive_utility_sp" },
    { 0x53E, "maps/interactive_models/bldg_01_dest" },
    { 0x53F, "maps/interactive_models/egrets" },
    { 0x540, "maps/interactive_models/fish_bannerfish" },
    { 0x541, "maps/interactive_models/fish_school_sardines" },
    { 0x542, "maps/interactive_models/fish_school_snapper" },
    { 0x543, "maps/interactive_models/oilrig_hanging_jumpsuit" },
    { 0x544, "maps/interactive_models/parakeets" },
    { 0x545, "maps/interactive_models/pigeons" },
    { 0x546, "maps/interactive_models/vulture" },
    { 0x547, "maps/mp/_animatedmodels" },
    { 0x548, "maps/mp/_areas" },
    { 0x549, "maps/mp/_art" },
    { 0x54A, "maps/mp/_audio" },
    { 0x54B, "maps/mp/_awards" },
    { 0x54C, "maps/mp/_breach" },
    { 0x54D, "maps/mp/_compass" },
    { 0x54E, "maps/mp/_createfx" },
    { 0x54F, "maps/mp/_crib" },
    { 0x550, "maps/mp/_defcon" },
    { 0x551, "maps/mp/_destructables" },
    { 0x552, "maps/mp/_elevator" },
    { 0x553, "maps/mp/_empgrenade" },
    { 0x554, "maps/mp/_entityheadicons" },
    { 0x555, "maps/mp/_events" },
    { 0x556, "maps/mp/_flashgrenades" },
    { 0x557, "maps/mp/_fx" },
    { 0x558, "maps/mp/_global_fx" },
    { 0x559, "maps/mp/_global_fx_code" },
    { 0x55A, "maps/mp/_highlights" },
    { 0x55B, "maps/mp/_javelin" },
    { 0x55C, "maps/mp/_laserguidedlauncher" },
    { 0x55D, "maps/mp/_load" },
    { 0x55E, "maps/mp/_matchdata" },
    { 0x55F, "maps/mp/_matchevents" },
    { 0x560, "maps/mp/_menus" },
    { 0x561, "maps/mp/_minefields" },
    { 0x562, "maps/mp/_movable_cover" },
    { 0x563, "maps/mp/_movers" },
    { 0x564, "maps/mp/_radiation" },
    { 0x565, "maps/mp/_scoreboard" },
    { 0x566, "maps/mp/_shutter" },
    { 0x567, "maps/mp/_stinger" },
    { 0x568, "maps/mp/_teleport" },
    { 0x569, "maps/mp/_utility" },
    { 0x56A, "maps/mp/_water" },
    { 0x56B, "maps/mp/_zipline" },
    { 0x56C, "maps/mp/mp_alien_town_fx" },
    { 0x56D, "maps/mp/mp_alien_town_precache" },
    { 0x56E, "maps/mp/mp_chasm_fx" },
    { 0x56F, "maps/mp/mp_chasm_precache" },
    { 0x570, "maps/mp/mp_dart_fx" },
    { 0x571, "maps/mp/mp_dart_precache" },
    { 0x572, "maps/mp/mp_fahrenheit_fx" },
    { 0x573, "maps/mp/mp_fahrenheit_precache" },
    { 0x574, "maps/mp/mp_flooded_fx" },
    { 0x575, "maps/mp/mp_flooded_precache" },
    { 0x576, "maps/mp/mp_frag_fx" },
    { 0x577, "maps/mp/mp_frag_precache" },
    { 0x578, "maps/mp/mp_hashima_fx" },
    { 0x579, "maps/mp/mp_hashima_precache" },
    { 0x57A, "maps/mp/mp_lonestar_fx" },
    { 0x57B, "maps/mp/mp_lonestar_precache" },
    { 0x57C, "maps/mp/mp_prisonbreak_fx" },
    { 0x57D, "maps/mp/mp_prisonbreak_precache" },
    { 0x57E, "maps/mp/mp_skeleton_fx" },
    { 0x57F, "maps/mp/mp_skeleton_precache" },
    { 0x580, "maps/mp/mp_snow_fx" },
    { 0x581, "maps/mp/mp_snow_precache" },
    { 0x582, "maps/mp/mp_sovereign_fx" },
    { 0x583, "maps/mp/mp_sovereign_precache" },
    { 0x584, "maps/mp/mp_strikezone_fx" },
    { 0x585, "maps/mp/mp_strikezone_precache" },
    { 0x586, "maps/mp/mp_warhawk_fx" },
    { 0x587, "maps/mp/mp_warhawk_precache" },
    { 0x588, "maps/mp/mp_zebra_fx" },
    { 0x589, "maps/mp/mp_zebra_precache" },
    { 0x58A, "maps/mp/alien/_achievement" },
    { 0x58B, "maps/mp/alien/_airdrop" },
    { 0x58C, "maps/mp/alien/_alien_fx" },
    { 0x58D, "maps/mp/alien/_autosentry_alien" },
    { 0x58E, "maps/mp/alien/_unk1422" },
    { 0x58F, "maps/mp/alien/_challenge_function" },
    { 0x590, "maps/mp/alien/_collectibles" },
    { 0x591, "maps/mp/alien/_combat_resources" },
    { 0x592, "maps/mp/alien/_damage" },
    { 0x593, "maps/mp/alien/_death" },
    { 0x594, "maps/mp/alien/_debug" },
    { 0x595, "maps/mp/alien/_deployablebox" },
    { 0x596, "maps/mp/alien/_deployablebox_adrenalinebox" },
    { 0x597, "maps/mp/alien/_deployablebox_ammo" },
    { 0x598, "maps/mp/alien/_deployablebox_currency" },
    { 0x599, "maps/mp/alien/_deployablebox_explosives" },
    { 0x59A, "maps/mp/alien/_deployablebox_juicebox" },
    { 0x59B, "maps/mp/alien/_deployablebox_randombox" },
    { 0x59C, "maps/mp/alien/_deployablebox_specialammo" },
    { 0x59D, "maps/mp/alien/_deployablebox_specialammo_ap" },
    { 0x59E, "maps/mp/alien/_deployablebox_specialammo_explo" },
    { 0x59F, "maps/mp/alien/_unk1439" },
    { 0x5A0, "maps/mp/alien/_deployablebox_vest" },
    { 0x5A1, "maps/mp/alien/_director" },
    { 0x5A2, "maps/mp/alien/_drill" },
    { 0x5A3, "maps/mp/alien/_unk1443" },
    { 0x5A4, "maps/mp/alien/_globallogic" },
    { 0x5A5, "maps/mp/alien/_hud" },
    { 0x5A6, "maps/mp/alien/_intro_sequence" },
    { 0x5A7, "maps/mp/alien/_lasedstrike_alien" },
    { 0x5A8, "maps/mp/alien/_laststand" },
    { 0x5A9, "maps/mp/alien/_music_and_dialog" },
    { 0x5AA, "maps/mp/alien/_nuke" },
    { 0x5AB, "maps/mp/alien/_outline_proto" },
    { 0x5AC, "maps/mp/alien/_perk_utility" },
    { 0x5AD, "maps/mp/alien/_perkfunctions" },
    { 0x5AE, "maps/mp/alien/_perks" },
    { 0x5AF, "maps/mp/alien/_persistence" },
    { 0x5B0, "maps/mp/alien/_pillage" },
    { 0x5B1, "maps/mp/alien/_prestige" },
    { 0x5B2, "maps/mp/alien/_progression" },
    { 0x5B3, "maps/mp/alien/_spawn_director" },
    { 0x5B4, "maps/mp/alien/_spawnlogic" },
    { 0x5B5, "maps/mp/alien/_switchblade_alien" },
    { 0x5B6, "maps/mp/alien/_trap" },
    { 0x5B7, "maps/mp/alien/_unlock" },
    { 0x5B8, "maps/mp/alien/_unk1464" },
    { 0x5B9, "maps/mp/gametypes/_battlebuddy" },
    { 0x5BA, "maps/mp/gametypes/_battlechatter_mp" },
    { 0x5BB, "maps/mp/gametypes/_callbacksetup" },
    { 0x5BC, "maps/mp/gametypes/_class" },
    { 0x5BD, "maps/mp/gametypes/_clientids" },
    { 0x5BE, "maps/mp/gametypes/_damage" },
    { 0x5BF, "maps/mp/gametypes/_damagefeedback" },
    { 0x5C0, "maps/mp/gametypes/_deathicons" },
    { 0x5C1, "maps/mp/gametypes/_dev" },
    { 0x5C2, "maps/mp/gametypes/_door" },
    { 0x5C3, "maps/mp/gametypes/_friendicons" },
    { 0x5C4, "maps/mp/gametypes/_gamelogic" },
    { 0x5C5, "maps/mp/gametypes/_gameobjects" },
    { 0x5C6, "maps/mp/gametypes/_gamescore" },
    { 0x5C7, "maps/mp/gametypes/_globalentities" },
    { 0x5C8, "maps/mp/gametypes/_globallogic" },
    { 0x5C9, "maps/mp/gametypes/_hardpoints" },
    { 0x5CA, "maps/mp/gametypes/_healthoverlay" },
    { 0x5CB, "maps/mp/gametypes/_horde_crates" },
    { 0x5CC, "maps/mp/gametypes/_horde_laststand" },
    { 0x5CD, "maps/mp/gametypes/_horde_util" },
    { 0x5CE, "maps/mp/gametypes/_hostmigration" },
    { 0x5CF, "maps/mp/gametypes/_hud" },
    { 0x5D0, "maps/mp/gametypes/_hud_message" },
    { 0x5D1, "maps/mp/gametypes/_hud_util" },
    { 0x5D2, "maps/mp/gametypes/_intel" },
    { 0x5D3, "maps/mp/gametypes/_intelchallenges" },
    { 0x5D4, "maps/mp/gametypes/_killcam" },
    { 0x5D5, "maps/mp/gametypes/_menus" },
    { 0x5D6, "maps/mp/gametypes/_missions" },
    { 0x5D7, "maps/mp/gametypes/_music_and_dialog" },
    { 0x5D8, "maps/mp/gametypes/_objpoints" },
    { 0x5D9, "maps/mp/gametypes/_outline" },
    { 0x5DA, "maps/mp/gametypes/_persistence" },
    { 0x5DB, "maps/mp/gametypes/_playercards" },
    { 0x5DC, "maps/mp/gametypes/_playerlogic" },
    { 0x5DD, "maps/mp/gametypes/_rank" },
    { 0x5DE, "maps/mp/gametypes/_serversettings" },
    { 0x5DF, "maps/mp/gametypes/_shellshock" },
    { 0x5E0, "maps/mp/gametypes/_spawnfactor" },
    { 0x5E1, "maps/mp/gametypes/_spawnlogic" },
    { 0x5E2, "maps/mp/gametypes/_spawnscoring" },
    { 0x5E3, "maps/mp/gametypes/_spectating" },
    { 0x5E4, "maps/mp/gametypes/_teams" },
    { 0x5E5, "maps/mp/gametypes/_trophy_system" },
    { 0x5E6, "maps/mp/gametypes/_tweakables" },
    { 0x5E7, "maps/mp/gametypes/_weapons" },
    { 0x5E8, "maps/mp/killstreaks/_a10" },
    { 0x5E9, "maps/mp/killstreaks/_unk1513" },
    { 0x5EA, "maps/mp/killstreaks/_aamissile" },
    { 0x5EB, "maps/mp/killstreaks/_unk1515" },
    { 0x5EC, "maps/mp/killstreaks/_ac130" },
    { 0x5ED, "maps/mp/killstreaks/_agent_killstreak" },
    { 0x5EE, "maps/mp/killstreaks/_air_superiority" },
    { 0x5EF, "maps/mp/killstreaks/_airdrop" },
    { 0x5F0, "maps/mp/killstreaks/_airstrike" },
    { 0x5F1, "maps/mp/killstreaks/_autosentry" },
    { 0x5F2, "maps/mp/killstreaks/_autoshotgun" },
    { 0x5F3, "maps/mp/killstreaks/_unk1523" },
    { 0x5F4, "maps/mp/killstreaks/_deployablebox" },
    { 0x5F5, "maps/mp/killstreaks/_deployablebox_ammo" },
    { 0x5F6, "maps/mp/killstreaks/_deployablebox_grenades" },
    { 0x5F7, "maps/mp/killstreaks/_deployablebox_gun" },
    { 0x5F8, "maps/mp/killstreaks/_deployablebox_juicebox" },
    { 0x5F9, "maps/mp/killstreaks/_deployablebox_vest" },
    { 0x5FA, "maps/mp/killstreaks/_designator_grenade" },
    { 0x5FB, "maps/mp/killstreaks/_dog_killstreak" },
    { 0x5FC, "maps/mp/killstreaks/_dronehive" },
    { 0x5FD, "maps/mp/killstreaks/_emp" },
    { 0x5FE, "maps/mp/killstreaks/_unk1534" },
    { 0x5FF, "maps/mp/killstreaks/_escortairdrop" },
    { 0x600, "maps/mp/killstreaks/_flares" },
    { 0x601, "maps/mp/killstreaks/_gas_airstrike" },
    { 0x602, "maps/mp/killstreaks/_harrier" },
    { 0x603, "maps/mp/killstreaks/_helicopter" },
    { 0x604, "maps/mp/killstreaks/_helicopter_flock" },
    { 0x605, "maps/mp/killstreaks/_helicopter_guard" },
    { 0x606, "maps/mp/killstreaks/_helicopter_pilot" },
    { 0x607, "maps/mp/killstreaks/_helisniper" },
    { 0x608, "maps/mp/killstreaks/_unk1544" },
    { 0x609, "maps/mp/killstreaks/_ims" },
    { 0x60A, "maps/mp/killstreaks/_jammer" },
    { 0x60B, "maps/mp/killstreaks/_juggernaut" },
    { 0x60C, "maps/mp/killstreaks/_killstreaks" },
    { 0x60D, "maps/mp/killstreaks/_killstreaks_init" },
    { 0x60E, "maps/mp/killstreaks/_lasedstrike" },
    { 0x60F, "maps/mp/killstreaks/_mobilemortar" },
    { 0x610, "maps/mp/killstreaks/_mrsiartillery" },
    { 0x611, "maps/mp/killstreaks/_nuke" },
    { 0x612, "maps/mp/killstreaks/_odin" },
    { 0x613, "maps/mp/killstreaks/_perkstreaks" },
    { 0x614, "maps/mp/killstreaks/_unk1556" },
    { 0x615, "maps/mp/killstreaks/_placeable_barrier" },
    { 0x616, "maps/mp/killstreaks/_plane" },
    { 0x617, "maps/mp/killstreaks/_portableaoegenerator" },
    { 0x618, "maps/mp/killstreaks/_remotemissile" },
    { 0x619, "maps/mp/killstreaks/_unk1561" },
    { 0x61A, "maps/mp/killstreaks/_remotetank" },
    { 0x61B, "maps/mp/killstreaks/_remoteturret" },
    { 0x61C, "maps/mp/killstreaks/_remoteuav" },
    { 0x61D, "maps/mp/killstreaks/_tank" },
    { 0x61E, "maps/mp/killstreaks/_teamammorefill" },
    { 0x61F, "maps/mp/killstreaks/_uav" },
    { 0x620, "maps/mp/killstreaks/_uplink" },
    { 0x621, "maps/mp/killstreaks/_vanguard" },
    { 0x622, "maps/mp/perks/_abilities" },
    { 0x623, "maps/mp/perks/_perkfunctions" },
    { 0x624, "maps/mp/perks/_perks" },
// onslaugh dlc
    { 0x8EFB, "destructible_scripts/toy_wall_fan" },
    { 36604, "destructible_scripts/vehicle_pickup" },
    { 36605, "maps/createart/mp_alien_armory_art" },
    { 36606, "maps/createart/mp_alien_armory_fog" },
    { 36607, "maps/createart/mp_boneyard_ns_art" },
    { 36608, "maps/createart/mp_boneyard_ns_fog" },
    { 36609, "maps/createart/mp_ca_red_river_art" },
    { 36610, "maps/createart/mp_ca_red_river_fog" },
    { 36611, "maps/createart/mp_ca_rumble_art" },
    { 36612, "maps/createart/mp_ca_rumble_fog" },
    { 36613, "maps/createart/mp_swamp_art" },
    { 36614, "maps/createart/mp_swamp_fog" },
    { 36615, "maps/cretaefx/mp_alien_armory_fx" },
    { 36616, "maps/createfx/mp_boneyard_ns_fx" },
    { 36617, "maps/createfx/mp_ca_red_river_fx" },
    { 36618, "maps/createfx/mp_ca_rumble_fx" },
    { 36619, "maps/createfx/mp_swamp_fx" },
    // { 36620, "maps/mp/dlc_unk" },
    { 36621, "maps/mp/mp_alien_armory_fx" },
    { 36622, "maps/mp/mp_alien_armory_precache" },
    { 36623, "maps/mp/mp_boneyard_ns_fx" },
    { 36624, "maps/mp/mp_boneyard_ns_precache" },
    { 36625, "maps/mp/mp_ca_red_river_fx" },
    { 36626, "maps/mp/mp_ca_red_river_precache" },
    { 36627, "maps/mp/mp_ca_rumble_fx" },
    { 36628, "maps/mp/mp_ca_rumble_precache" },
    { 36629, "maps/mp/mp_swamp_fx" },
    { 36630, "maps/mp/mp_swamp_precache" },
    // 36631-44, effects?
}};

const std::array<std::pair<std::uint16_t, const char*>, 13823> token_list
{{
    { 0x0000, "" },
    { 0x0001, "pl#" },
    { 0x0002, "-" },
    { 0x0003, "radius`" },
    { 0x0004, "note:" },
    { 0x0005, "_" },
    { 0x0006, "_custom" },
    { 0x0007, "accuracy" },
    { 0x0008, "actionslot1" },
    { 0x0009, "actionslot2" },
    { 0x000A, "actionslot3" },
    { 0x000B, "actionslot4" },
    { 0x000C, "actionslot5" },
    { 0x000D, "actionslot6" },
    { 0x000E, "actionslot7" },
    { 0x000F, "activator" },
    { 0x10, "active" },
    { 0x11, "activevisionset" },
    { 0x12, "activevisionsetduration" },
    { 0x13, "agent" },
    { 0x14, "agenthealth" },
    { 0x15, "agentname" },
    { 0x16, "agentteam" },
    { 0x17, "ai_event" },
    { 0x18, "ai_sight_line_cycle_group" },
    { 0x19, "ai_sight_line_group" },
    { 0x1A, "aim_highest_bone" },
    { 0x1B, "aim_vis_bone" },
    { 0x1C, "alert" },
    { 0x1D, "alertlevel" },
    { 0x1E, "alertlevelint" },
    { 0x1F, "alien" },
    { 0x20, "alignx" },
    { 0x21, "aligny" },
    { 0x22, "all" },
    { 0x23, "allies" },
    { 0x24, "allowdeath" },
    { 0x25, "allowjump" },
    { 0x26, "allowladders" },
    { 0x27, "allowpain" },
    { 0x28, "alpha" },
    { 0x29, "altmode" },
    { 0x2A, "always" },
    { 0x2B, "ambush" },
    { 0x2C, "ambush_nodes_only" },
    { 0x2D, "angle_deltas" },
    { 0x2E, "anglelerprate" },
    { 0x2F, "angles" },
    { 0x30, "anim_angle_delta" },
    { 0x31, "anim_deltas" },
    { 0x32, "anim_pose" },
    { 0x33, "anim_will_finish" },
    { 0x34, "animscript" },
    { 0x35, "archived" },
    { 0x36, "archivetime" },
    { 0x37, "asleep" },
    { 0x38, "aspectratio" },
    { 0x39, "assists" },
    { 0x3A, "attackeraccuracy" },
    { 0x3B, "attackercount" },
    { 0x3C, "auto_ai" },
    { 0x3D, "auto_change" },
    { 0x3E, "auto_nonai" },
    { 0x3F, "axis" },
    { 0x40, "back" },
    { 0x41, "back_left" },
    { 0x42, "back_low" },
    { 0x43, "back_mid" },
    { 0x44, "back_right" },
    { 0x45, "back_up" },
    { 0x46, "bad_guys" },
    { 0x47, "bad_path" },
    { 0x48, "badplaceawareness" },
    { 0x49, "begin" },
    { 0x4A, "begin_custom_anim" },
    { 0x4B, "begin_firing" },
    { 0x4C, "begin_firing_left" },
    { 0x4D, "bipods" },
    { 0x4E, "birthtime" },
    { 0x4F, "blade_hide" },
    { 0x50, "blade_show" },
    { 0x51, "blockfriendlies" },
    { 0x52, "blurradius" },
    { 0x53, "body_animate_jnt" },
    { 0x54, "bullet_hitshield" },
    { 0x55, "bullethit" },
    { 0x56, "bulletwhizby" },
    { 0x57, "call_vote" },
    { 0x58, "cancel_location" },
    { 0x59, "canclimbladders" },
    { 0x5A, "chainfallback" },
    { 0x5B, "chainnode" },
    { 0x5C, "chest" },
    { 0x5D, "chyron_message1" },
    { 0x5E, "chyron_message2" },
    { 0x5F, "chyron_message3" },
    { 0x60, "civilian" },
    { 0x61, "classname" },
    { 0x62, "clipdistance" },
    { 0x63, "code_classname" },
    { 0x64, "code_damageradius" },
    { 0x65, "code_move" },
    { 0x66, "code_move_slide" },
    { 0x67, "color" },
    { 0x68, "color_blind_toggled" },
    { 0x69, "combat" },
    { 0x6A, "combatmode" },
    { 0x6B, "confirm_location" },
    { 0x6C, "constrained" },
    { 0x6D, "contact" },
    { 0x6E, "contextleanenabled" },
    { 0x6F, "count" },
    { 0x70, "cover" },
    { 0x71, "cover_approach" },
    { 0x72, "coversearchinterval" },
    { 0x73, "criticalbulletdamagedist" },
    { 0x74, "crouch" },
    { 0x75, "current" },
    { 0x76, "custom_attach_00" },
    { 0x77, "custom_attach_01" },
    { 0x78, "custom_attach_02" },
    { 0x79, "custom_attach_03" },
    { 0x7A, "custom_attach_04" },
    { 0x7B, "custom_attach_05" },
    { 0x7C, "custom_attach_06" },
    { 0x7D, "custom_attach_07" },
    { 0x7E, "custom_attach_08" },
    { 0x7F, "custom_attach_09" },
    { 0x80, "custom_attach_10" },
    { 0x81, "custom_attach_11" },
    { 0x82, "custom_attach_12" },
    { 0x83, "custom_attach_13" },
    { 0x84, "custom_attach_14" },
    { 0x85, "custom_attach_15" },
    { 0x86, "damage" },
    { 0x87, "damage_notdone" },
    { 0x88, "damagedir" },
    { 0x89, "damagelocation" },
    { 0x8A, "damagemod" },
    { 0x8B, "damagemultiplier" },
    { 0x8C, "damageshield" },
    { 0x8D, "damagetaken" },
    { 0x8E, "damageweapon" },
    { 0x8F, "damageyaw" },
    { 0x90, "dangerreactduration" },
    { 0x91, "dead" },
    { 0x92, "death" },
    { 0x93, "deathinvulnerabletime" },
    { 0x94, "deathplant" },
    { 0x95, "deaths" },
    { 0x96, "deathshield" },
    { 0x97, "delayeddeath" },
    { 0x98, "desiredangle" },
    { 0x99, "detonate" },
    { 0x9A, "diequietly" },
    { 0x9B, "direct" },
    { 0x9C, "disableplayeradsloscheck" },
    { 0x9D, "dlight" },
    { 0x9E, "dmg" },
    { 0x9F, "dodamagetoall" },
    { 0xA0, "dodangerreact" },
    { 0xA1, "doffar" },
    { 0xA2, "dofnear" },
    { 0xA3, "dog" },
    { 0xA4, "doghandler" },
    { 0xA5, "doingambush" },
    { 0xA6, "done" },
    { 0xA7, "dontavoidplayer" },
    { 0xA8, "down" },
    { 0xA9, "downaimlimit" },
    { 0xAA, "drawoncompass" },
    { 0xAB, "dropweapon" },
    { 0xAC, "empty" },
    { 0xAD, "empty_offhand" },
    { 0xAE, "enable" },
    { 0xAF, "enableshadows" },
    { 0xB0, "end_firing" },
    { 0xB1, "end_firing_left" },
    { 0xB2, "enemy" },
    { 0xB3, "enemy_sighted" },
    { 0xB4, "enemy_sighted_lost" },
    { 0xB5, "enemy_visible" },
    { 0xB6, "engagemaxdist" },
    { 0xB7, "engagemaxfalloffdist" },
    { 0xB8, "engagemindist" },
    { 0xB9, "engageminfalloffdist" },
    { 0xBA, "entity" },
    { 0xBB, "exclusive" },
    { 0xBC, "explode" },
    { 0xBD, "exposedduration" },
    { 0xBE, "extrascore0" },
    { 0xBF, "extrascore1" },
    { 0xC0, "playercardpatch" },
    { 0xC1, "playercardpatchbacking" },
    { 0xC2, "playercardbackground" },
    { 0xC3, "face_angle" },
    { 0xC4, "face_angle_3d" },
    { 0xC5, "face_angle_abs" },
    { 0xC6, "face_angle_rel" },
    { 0xC7, "face_current" },
    { 0xC8, "face_default" },
    { 0xC9, "face_direction" },
    { 0xCA, "face_enemy" },
    { 0xCB, "face_enemy_or_motion" },
    { 0xCC, "face_goal" },
    { 0xCD, "face_motion" },
    { 0xCE, "face_point" },
    { 0xCF, "facemotion" },
    { 0xD0, "failed" },
    { 0xD1, "falling" },
    { 0xD2, "fast_radar" },
    { 0xD3, "favoriteenemy" },
    { 0xD4, "finalaccuracy" },
    { 0xD5, "first_person" },
    { 0xD6, "fixednode" },
    { 0xD7, "fixednodesaferadius" },
    { 0xD8, "flash" },
    { 0xD9, "flashbang" },
    { 0xDA, "follow" },
    { 0xDB, "followmax" },
    { 0xDC, "followmin" },
    { 0xDD, "font" },
    { 0xDE, "fontscale" },
    { 0xDF, "footstepdetectdist" },
    { 0xE0, "footstepdetectdistsprint" },
    { 0xE1, "footstepdetectdistwalk" },
    { 0xE2, "forceragdollimmediate" },
    { 0xE3, "forcespectatorclient" },
    { 0xE4, "foregrip_off" },
    { 0xE5, "foreground" },
    { 0xE6, "forward" },
    { 0xE7, "fov" },
    { 0xE8, "fovcosine" },
    { 0xE9, "fovcosinebusy" },
    { 0xEA, "fraction" },
    { 0xEB, "frag" },
    { 0xEC, "free" },
    { 0xED, "freecamera" },
    { 0xEE, "freelook" },
    { 0xEF, "front_left" },
    { 0xF0, "front_right" },
    { 0xF1, "frontshieldanglecos" },
    { 0xF2, "game_extrainfo" },
    { 0xF3, "glass_destroyed" },
    { 0xF4, "glowalpha" },
    { 0xF5, "glowcolor" },
    { 0xF6, "goal" },
    { 0xF7, "goal_changed" },
    { 0xF8, "goal_reached" },
    { 0xF9, "goal_yaw" },
    { 0xFA, "goalheight" },
    { 0xFB, "goalpos" },
    { 0xFC, "goalradius" },
    { 0xFD, "goingtoruntopos" },
    { 0xFE, "gravity" },
    { 0xFF, "grenade" },
    { 0x100, "grenade_fire" },
    { 0x101, "grenade_pullback" },
    { 0x102, "grenade_return_hand_tag" },
    { 0x103, "grenadeammo" },
    { 0x104, "grenadeawareness" },
    { 0x105, "grenadedanger" },
    { 0x106, "grenadeweapon" },
    { 0x107, "groundEntChanged" },
    { 0x108, "groundtype" },
    { 0x109, "gunblockedbywall" },
    { 0x10A, "gunshot" },
    { 0x10B, "gunshot_teammate" },
    { 0x10C, "hasradar" },
    { 0x10D, "headicon" },
    { 0x10E, "headiconteam" },
    { 0x10F, "health" },
    { 0x110, "height" },
    { 0x111, "showinkillcam" },
    { 0x112, "hidein3rdperson" },
    { 0x113, "hidewhendead" },
    { 0x114, "hidewhenindemo" },
    { 0x115, "enablehudlighting" },
    { 0x116, "hidewheninmenu" },
    { 0x117, "high_priority" },
    { 0x118, "highlyawareradius" },
    { 0x119, "hindlegstraceoffset" },
    { 0x11A, "hit_by_missile" },
    { 0x11B, "horzalign" },
    { 0x11C, "host_sucks_end_game" },
    { 0x11D, "human" },
    { 0x11E, "ignoreall" },
    { 0x11F, "ignoreclosefoliage" },
    { 0x120, "ignoreexplosionevents" },
    { 0x121, "ignoreforfixednodesafecheck" },
    { 0x122, "ignoreme" },
    { 0x123, "ignorerandombulletdamage" },
    { 0x124, "ignoresuppression" },
    { 0x125, "ignoretriggers" },
    { 0x126, "infinite_energy" },
    { 0x127, "info_notnull" },
    { 0x128, "info_player_start" },
    { 0x129, "insolid" },
    { 0x12A, "intermission" },
    { 0x12B, "interval" },
    { 0x12C, "invisible" },
    { 0x12D, "ironsight_off" },
    { 0x12E, "ironsight_on" },
    { 0x12F, "isradarblocked" },
    { 0x130, "item" },
    { 0x131, "j_eyeball_le" },
    { 0x132, "j_eyeball_ri" },
    { 0x133, "j_head" },
    { 0x134, "j_left_elbow" },
    { 0x135, "j_left_hand" },
    { 0x136, "j_left_shoulder" },
    { 0x137, "j_mainroot" },
    { 0x138, "j_neck" },
    { 0x139, "j_spine4" },
    { 0x13A, "j_spinelower" },
    { 0x13B, "j_spineupper" },
    { 0x13C, "jumpcost" },
    { 0x13D, "jumping" },
    { 0x13E, "keepclaimednode" },
    { 0x13F, "keepclaimednodeifvalid" },
    { 0x140, "keepnodeduringscriptedanim" },
    { 0x141, "key1" },
    { 0x142, "key2" },
    { 0x143, "killanimscript" },
    { 0x144, "killcamentity" },
    { 0x145, "killcamentitylookat" },
    { 0x146, "kills" },
    { 0x147, "known_event" },
    { 0x148, "label" },
    { 0x149, "ladder_down" },
    { 0x14A, "ladder_up" },
    { 0x14B, "land" },
    { 0x14C, "lastattacker" },
    { 0x14D, "lastenemysightpos" },
    { 0x14E, "laststand" },
    { 0x14F, "leanamount" },
    { 0x150, "ledge" },
    { 0x151, "left" },
    { 0x152, "leftaimlimit" },
    { 0x153, "light" },
    { 0x154, "lockorientation" },
    { 0x155, "lod" },
    { 0x156, "look" },
    { 0x157, "lookahead" },
    { 0x158, "lookaheaddir" },
    { 0x159, "lookaheaddist" },
    { 0x15A, "lookaheadhitsstairs" },
    { 0x15B, "lookforward" },
    { 0x15C, "lookright" },
    { 0x15D, "looktarget" },
    { 0x15E, "lookup" },
    { 0x15F, "low_priority" },
    { 0x160, "lowresbackground" },
    { 0x161, "luinotifyserver" },
    { 0x162, "mag_eject" },
    { 0x163, "mag_eject_left" },
    { 0x164, "manual" },
    { 0x165, "manual_ai" },
    { 0x166, "manual_change" },
    { 0x167, "max_time" },
    { 0x168, "maxfaceenemydist" },
    { 0x169, "maxhealth" },
    { 0x16A, "maxsightdistsqrd" },
    { 0x16B, "maxvisibledist" },
    { 0x16C, "meleeattackdist" },
    { 0x16D, "menuresponse" },
    { 0x16E, "middle_left" },
    { 0x16F, "middle_right" },
    { 0x170, "min_energy" },
    { 0x171, "min_time" },
    { 0x172, "minpaindamage" },
    { 0x173, "minusedistsq" },
    { 0x174, "missile_fire" },
    { 0x175, "missile_stuck" },
    { 0x176, "mod_crush" },
    { 0x177, "mod_explosive" },
    { 0x178, "mod_explosive_bullet" },
    { 0x179, "mod_falling" },
    { 0x17A, "mod_grenade" },
    { 0x17B, "mod_grenade_splash" },
    { 0x17C, "mod_head_shot" },
    { 0x17D, "mod_impact" },
    { 0x17E, "mod_melee" },
    { 0x17F, "mod_melee_alien" },
    { 0x180, "mod_melee_dog" },
    { 0x181, "mod_pistol_bullet" },
    { 0x182, "mod_projectile" },
    { 0x183, "mod_projectile_splash" },
    { 0x184, "mod_rifle_bullet" },
    { 0x185, "mod_suicide" },
    { 0x186, "mod_trigger_hurt" },
    { 0x187, "mod_unknown" },
    { 0x188, "model" },
    { 0x189, "motiontrackerenabled" },
    { 0x18A, "movedone" },
    { 0x18B, "movemode" },
    { 0x18C, "name" },
    { 0x18D, "near_goal" },
    { 0x18E, "nearz" },
    { 0x18F, "neutral" },
    { 0x190, "never" },
    { 0x191, "newenemyreactiondistsq" },
    { 0x192, "night_vision_off" },
    { 0x193, "night_vision_on" },
    { 0x194, "no_cover" },
    { 0x195, "no_gravity" },
    { 0x196, "noattackeraccuracymod" },
    { 0x197, "noclip" },
    { 0x198, "node" },
    { 0x199, "node_not_safe" },
    { 0x19A, "node_out_of_range" },
    { 0x19B, "node_relinquished" },
    { 0x19C, "node_taken" },
    { 0x19D, "nodeoffsetpos" },
    { 0x19E, "nododgemove" },
    { 0x19F, "nogravity" },
    { 0x1A0, "nogrenadereturnthrow" },
    { 0x1A1, "noncombat" },
    { 0x1A2, "none" },
    { 0x1A3, "nophysics" },
    { 0x1A4, "normal" },
    { 0x1A5, "normal_radar" },
    { 0x1A6, "notinsolid" },
    { 0x1A7, "obstacle" },
    { 0x1A8, "offhand" },
    { 0x1A9, "offhand_end" },
    { 0x1AA, "only_sky" },
    { 0x1AB, "onlygoodnearestnodes" },
    { 0x1AC, "oriented" },
    { 0x1AD, "orientto_complete" },
    { 0x1AE, "origin" },
    { 0x1AF, "other" },
    { 0x1B0, "over" },
    { 0x1B1, "owner" },
    { 0x1B2, "pacifist" },
    { 0x1B3, "pacifistwait" },
    { 0x1B4, "pain" },
    { 0x1B5, "parentindex" },
    { 0x1B6, "parentname" },
    { 0x1B7, "path_blocked" },
    { 0x1B8, "path_changed" },
    { 0x1B9, "path_dir_change" },
    { 0x1BA, "path_enemy" },
    { 0x1BB, "path_need_dodge" },
    { 0x1BC, "path_set" },
    { 0x1BD, "pathenemyfightdist" },
    { 0x1BE, "pathenemylookahead" },
    { 0x1BF, "pathgoalpos" },
    { 0x1C0, "pathrandompercent" },
    { 0x1C1, "pelvis" },
    { 0x1C2, "pers" },
    { 0x1C3, "physics_finished" },
    { 0x1C4, "pickup" },
    { 0x1C5, "pistol" },
    { 0x1C6, "pitchamount" },
    { 0x1C7, "plane_waypoint" },
    { 0x1C8, "player" },
    { 0x1C9, "player_pushed" },
    { 0x1CA, "playername" },
    { 0x1CB, "playing" },
    { 0x1CC, "position" },
    { 0x1CD, "predicted_projectile_impact" },
    { 0x1CE, "prevanimdelta" },
    { 0x1CF, "prevnode" },
    { 0x1D0, "prevscript" },
    { 0x1D1, "primary" },
    { 0x1D2, "primaryoffhand" },
    { 0x1D3, "projectile_impact" },
    { 0x1D4, "prone" },
    { 0x1D5, "proneok" },
    { 0x1D6, "providecoveringfire" },
    { 0x1D7, "psoffsettime" },
    { 0x1D8, "pushable" },
    { 0x1D9, "radarmode" },
    { 0x1DA, "radarshowenemydirection" },
    { 0x1DB, "radarstrength" },
    { 0x1DC, "radius" },
    { 0x1DD, "ragdoll_early_result" },
    { 0x1DE, "reached_end_node" },
    { 0x1DF, "reached_wait_node" },
    { 0x1E0, "reached_wait_speed" },
    { 0x1E1, "reactiontargetpos" },
    { 0x1E2, "receiver" },
    { 0x1E3, "relativedir" },
    { 0x1E4, "reload" },
    { 0x1E5, "reload_start" },
    { 0x1E6, "rendertotexture" },
    { 0x1E7, "requestarrivalnotify" },
    { 0x1E8, "result" },
    { 0x1E9, "return_pitch" },
    { 0x1EA, "reverse" },
    { 0x1EB, "right" },
    { 0x1EC, "rightaimlimit" },
    { 0x1ED, "riotshield_damaged" },
    { 0x1EE, "rocket" },
    { 0x1EF, "rotatedone" },
    { 0x1F0, "run" },
    { 0x1F1, "runcost" },
    { 0x1F2, "runto_arrived" },
    { 0x1F3, "safetochangescript" },
    { 0x1F4, "scavenger" },
    { 0x1F5, "score" },
    { 0x1F6, "script" },
    { 0x1F7, "script_brushmodel" },
    { 0x1F8, "script_linkname" },
    { 0x1F9, "script_model" },
    { 0x1FA, "script_noteworthy" },
    { 0x1FB, "script_origin" },
    { 0x1FC, "script_parent" },
    { 0x1FD, "script_parentname" },
    { 0x1FE, "script_pushable" },
    { 0x1FF, "script_vehicle" },
    { 0x200, "script_vehicle_collision" },
    { 0x201, "script_vehicle_collmap" },
    { 0x202, "script_vehicle_corpse" },
    { 0x203, "scriptable" },
    { 0x204, "scriptedarrivalent" },
    { 0x205, "scope_center" },
    { 0x206, "scope_top" },
    { 0x207, "scope_cap" },
    { 0x208, "search_end" },
    { 0x209, "secondaryoffhand" },
    { 0x20A, "sentry" },
    { 0x20B, "sentry_offline" },
    { 0x20C, "sessionstate" },
    { 0x20D, "sessionteam" },
    { 0x20E, "sharpturnnotifydist" },
    { 0x20F, "sightlatency" },
    { 0x210, "silenced_shot" },
    { 0x211, "slidevelocity" },
    { 0x212, "slowmo_active" },
    { 0x213, "slowmo_passive" },
    { 0x214, "smoke" },
    { 0x215, "snd_channelvolprio_holdbreath" },
    { 0x216, "snd_channelvolprio_pain" },
    { 0x217, "snd_channelvolprio_shellshock" },
    { 0x218, "snd_enveffectsprio_level" },
    { 0x219, "snd_enveffectsprio_shellshock" },
    { 0x21A, "sort" },
    { 0x21B, "sound_blend" },
    { 0x21C, "space" },
    { 0x21D, "spawned" },
    { 0x21E, "spawnflags" },
    { 0x21F, "spectatekillcam" },
    { 0x220, "spectating_cycle" },
    { 0x221, "spectator" },
    { 0x222, "speed" },
    { 0x223, "splatter" },
    { 0x224, "sprint_begin" },
    { 0x225, "sprint_end" },
    { 0x226, "stairsstate" },
    { 0x227, "stand" },
    { 0x228, "start_blend" },
    { 0x229, "start_move" },
    { 0x22A, "start_ragdoll" },
    { 0x22B, "statelocked" },
    { 0x22C, "statusicon" },
    { 0x22D, "stop" },
    { 0x22E, "stop_soon" },
    { 0x22F, "stopanimdistsq" },
    { 0x230, "stopsoonnotifydist" },
    { 0x231, "suppression" },
    { 0x232, "suppression_end" },
    { 0x233, "suppressionduration" },
    { 0x234, "suppressionmeter" },
    { 0x235, "suppressionstarttime" },
    { 0x236, "suppressionwait" },
    { 0x237, "surfacetype" },
    { 0x238, "surprisedbymedistsq" },
    { 0x239, "swimmer" },
    { 0x23A, "syncedmeleetarget" },
    { 0x23B, "tag" },
    { 0x23C, "tag_aim" },
    { 0x23D, "tag_aim_animated" },
    { 0x23E, "tag_aim_pivot" },
    { 0x23F, "tag_barrel" },
    { 0x240, "tag_blade_off" },
    { 0x241, "tag_body" },
    { 0x242, "tag_brass" },
    { 0x243, "tag_butt" },
    { 0x244, "tag_camera" },
    { 0x245, "tag_clip" },
    { 0x246, "tag_detach" },
    { 0x247, "tag_engine_left" },
    { 0x248, "tag_engine_right" },
    { 0x249, "tag_eotech_reticle" },
    { 0x24A, "tag_eye" },
    { 0x24B, "tag_flash" },
    { 0x24C, "tag_flash_11" },
    { 0x24D, "tag_flash_2" },
    { 0x24E, "tag_flash_22" },
    { 0x24F, "tag_flash_3" },
    { 0x250, "tag_flash_silenced" },
    { 0x251, "tag_fx" },
    { 0x252, "tag_gasmask" },
    { 0x253, "tag_gasmask2" },
    { 0x254, "tag_ik_loc_le" },
    { 0x255, "tag_ik_loc_le_foregrip" },
    { 0x256, "tag_ik_loc_le_launcher" },
    { 0x257, "tag_ik_loc_le_shotgun" },
    { 0x258, "tag_ik_target" },
    { 0x259, "tag_inhand" },
    { 0x25A, "tag_knife_fx" },
    { 0x25B, "tag_laser" },
    { 0x25C, "tag_launcher" },
    { 0x25D, "tag_magnifier_eotech_reticle" },
    { 0x25E, "tag_motion_tracker_bl" },
    { 0x25F, "tag_motion_tracker_br" },
    { 0x260, "tag_motion_tracker_fx" },
    { 0x261, "tag_motion_tracker_tl" },
    { 0x262, "tag_origin" },
    { 0x263, "tag_player" },
    { 0x264, "tag_popout" },
    { 0x265, "tag_reticle_acog" },
    { 0x266, "tag_reticle_hamr" },
    { 0x267, "tag_reticle_on" },
    { 0x268, "tag_reticle_red_dot" },
    { 0x269, "tag_reticle_reflex" },
    { 0x26A, "tag_reticle_tavor_scope" },
    { 0x26B, "tag_reticle_thermal_scope" },
    { 0x26C, "tag_shield_back" },
    { 0x26D, "tag_shotgun" },
    { 0x26E, "tag_show_alt" },
    { 0x26F, "tag_stowed_back" },
    { 0x270, "tag_stowed_hip_rear" },
    { 0x271, "tag_sync" },
    { 0x272, "tag_tip" },
    { 0x273, "tag_turret" },
    { 0x274, "tag_turret_base" },
    { 0x275, "tag_weapon" },
    { 0x276, "tag_weapon_chest" },
    { 0x277, "tag_weapon_left" },
    { 0x278, "tag_weapon_right" },
    { 0x279, "tag_wheel_back_left" },
    { 0x27A, "tag_wheel_back_right" },
    { 0x27B, "tag_wheel_front_left" },
    { 0x27C, "tag_wheel_front_right" },
    { 0x27D, "tag_wheel_middle_left" },
    { 0x27E, "tag_wheel_middle_right" },
    { 0x27F, "takedamage" },
    { 0x280, "target" },
    { 0x281, "target_script_trigger" },
    { 0x282, "targetname" },
    { 0x283, "team" },
    { 0x284, "team3" },
    { 0x285, "teammode_axisallies" },
    { 0x286, "teammode_ffa" },
    { 0x287, "teammovewaittime" },
    { 0x288, "thermal" },
    { 0x289, "thermalbodymaterial" },
    { 0x28A, "third_person" },
    { 0x28B, "threatbias" },
    { 0x28C, "threatbiasgroup" },
    { 0x28D, "throwingknife" },
    { 0x28E, "top" },
    { 0x28F, "touch" },
    { 0x290, "touching_platform" },
    { 0x291, "transients_synced" },
    { 0x292, "traverse_complete" },
    { 0x293, "traverse_soon" },
    { 0x294, "traversecost" },
    { 0x295, "traversesoonnotifydist" },
    { 0x296, "trigger" },
    { 0x297, "trigger_damage" },
    { 0x298, "trigger_use" },
    { 0x299, "trigger_use_touch" },
    { 0x29A, "truck_cam" },
    { 0x29B, "turnrate" },
    { 0x29C, "turret_deactivate" },
    { 0x29D, "turret_fire" },
    { 0x29E, "turret_no_vis" },
    { 0x29F, "turret_not_on_target" },
    { 0x2A0, "turret_on_target" },
    { 0x2A1, "turret_on_vistarget" },
    { 0x2A2, "turret_pitch_clamped" },
    { 0x2A3, "turret_rotate_stopped" },
    { 0x2A4, "turret_yaw_clamped" },
    { 0x2A5, "turretinvulnerability" },
    { 0x2A6, "turretownerchange" },
    { 0x2A7, "turretstatechange" },
    { 0x2A8, "type" },
    { 0x2A9, "unresolved_collision" },
    { 0x2AA, "up" },
    { 0x2AB, "upaimlimit" },
    { 0x2AC, "useable" },
    { 0x2AD, "usechokepoints" },
    { 0x2AE, "usecombatscriptatcover" },
    { 0x2AF, "veh_boatbounce" },
    { 0x2B0, "veh_brake" },
    { 0x2B1, "veh_collision" },
    { 0x2B2, "veh_jolt" },
    { 0x2B3, "veh_landed" },
    { 0x2B4, "veh_leftground" },
    { 0x2B5, "veh_pathdir" },
    { 0x2B6, "veh_pathspeed" },
    { 0x2B7, "veh_pathtype" },
    { 0x2B8, "veh_predictedcollision" },
    { 0x2B9, "veh_speed" },
    { 0x2BA, "veh_throttle" },
    { 0x2BB, "veh_topspeed" },
    { 0x2BC, "veh_transmission" },
    { 0x2BD, "vehicle_dismount" },
    { 0x2BE, "vehicle_mount" },
    { 0x2BF, "velocity" },
    { 0x2C0, "vertalign" },
    { 0x2C1, "visionsetmissilecam" },
    { 0x2C2, "visionsetmissilecamduration" },
    { 0x2C3, "visionsetnaked" },
    { 0x2C4, "visionsetnakedduration" },
    { 0x2C5, "visionsetnight" },
    { 0x2C6, "visionsetnightduration" },
    { 0x2C7, "visionsetpain" },
    { 0x2C8, "visionsetpainduration" },
    { 0x2C9, "visionsetthermal" },
    { 0x2CA, "visionsetthermalduration" },
    { 0x2CB, "vote" },
    { 0x2CC, "walk" },
    { 0x2CD, "walkdist" },
    { 0x2CE, "walkdistfacingmotion" },
    { 0x2CF, "waypoint_reached" },
    { 0x2D0, "weapon" },
    { 0x2D1, "weapon_change" },
    { 0x2D2, "weapon_dropped" },
    { 0x2D3, "weapon_fired" },
    { 0x2D4, "weapon_switch_started" },
    { 0x2D5, "weapon_taken" },
    { 0x2D6, "weaponchange" },
    { 0x2D7, "weaponrail_on" },
    { 0x2D8, "width" },
    { 0x2D9, "world" },
    { 0x2DA, "worldspawn" },
    { 0x2DB, "x" },
    { 0x2DC, "y" },
    { 0x2DD, "z" },
    { 0x2DE, "zonly_physics" },
    { 0x2DF, "accumulate" },
    { 0x2E0, "allowprone" },
    { 0x2E1, "aiSpread" },
    { 0x2E2, "ambienttrack" },
    { 0x2E3, "ambienttrack_ac130" },
    { 0x2E4, "bottomarc" },
    { 0x2E5, "convergencetime" },
    { 0x2E6, "cursorhint" },
    { 0x2E7, "destructible_type" },
    { 0x2E8, "diffusefraction" },
    { 0x2E9, "eftarc" },
    { 0x2EA, "leftarc" },
    { 0x2EB, "maxrange" },
    { 0x2EC, "northyaw" },
    { 0x2ED, "pitchconvergencetime" },
    { 0x2EE, "playerSpread" },
    { 0x2EF, "postsharpturnlookaheaddist" },
    { 0x2F0, "reflection_clear_color" },
    { 0x2F1, "rightarc" },
    { 0x2F2, "sharpturntooclosetodestdist" },
    { 0x2F3, "script_delay" },
    { 0x2F4, "script_visionset" },
    { 0x2F5, "script_zone" },
    { 0x2F6, "spawner" },
    { 0x2F7, "sunlight" },
    { 0x2F8, "suncolor" },
    { 0x2F9, "sundirection" },
    { 0x2FA, "suppressionTime" },
    { 0x2FB, "threshold" },
    { 0x2FC, "toparc" },
    { 0x2FD, "vehicletype" },
    { 0x2FE, "wait" },
    { 0x2FF, "weaponinfo" },
    { 0x300, "yawconvergencetime" },
// symbols
    { 0x625, "__smangles" },
    { 0x626, "__smid" },
    { 0x627, "__smname" },
    { 0x628, "__smorigin" },
    { 0x629, "_add_as_apache_target_on_spawn_iternal" },
    { 0x62A, "_ai_delete" },
    { 0x62B, "_ai_group" },
    { 0x62C, "_ai_health" },
    { 0x62D, "_aliveplayers" },
    { 0x62E, "_allies" },
    { 0x630, "_ally" },
    { 0x632, "_ally_dist" },
    { 0x633, "_ally_get_pitch_down_aim_weight" },
    { 0x634, "_ally_get_pitch_up_aim_weight" },
    { 0x635, "_ally_get_yaw_left_aim_weight" },
    { 0x636, "_ally_get_yaw_right_aim_weight" },
    { 0x637, "_ally_is_current_volume" },
    { 0x638, "_ally_set_last_volume" },
    { 0x639, "_ally_transition_to_weight" },
    { 0x63A, "_ally_trigs" },
    { 0x63B, "_anim" },
    { 0x63C, "_anim_node" },
    { 0x63D, "_animactive" },
    { 0x63F, "_animmode" },
    { 0x640, "_animname" },
    { 0x641, "_array_wait" },
    { 0x642, "_aud_zip_wind_1" },
    { 0x643, "_audio" },
    { 0x644, "_audio_trigger" },
    { 0x645, "_autosave_game_now" },
    { 0x646, "_autosave_game_now_nochecks" },
    { 0x647, "_autosave_game_now_notrestart" },
    { 0x648, "_autosave_stealthcheck" },
    { 0x649, "_autosave_stealthcheck_nml" },
    { 0x64A, "_battlechatter_off" },
    { 0x64B, "_battlechatter_on" },
    { 0x64C, "_beginlocationselection" },
    { 0x64D, "_boss" },
    { 0x64E, "_box_setactivehelper" },
    { 0x64F, "_bravo" },
    { 0x650, "_clearalltextafterhudelem" },
    { 0x651, "_clearperks" },
    { 0x652, "_col" },
    { 0x653, "_color" },
    { 0x654, "_color_friendly_spawners" },
    { 0x655, "_color_ng" },
    { 0x656, "_colors_go_line" },
    { 0x657, "_command" },
    { 0x659, "_current_goal_volume" },
    { 0x65A, "_current_index" },
    { 0x65B, "_custom_anim" },
    { 0x65C, "_custom_anim_loop" },
    { 0x65D, "_custom_anim_thread" },
    { 0x65E, "_death_anims" },
    { 0x65F, "_delay" },
    { 0x660, "_destroy" },
    { 0x661, "_destructible_preanims" },
    { 0x662, "_destructible_preanimtree" },
    { 0x663, "_detachall" },
    { 0x664, "_disableoffhandweapons" },
    { 0x665, "_disableusability" },
    { 0x666, "_disableweapon" },
    { 0x667, "_disableweaponswitch" },
    { 0x668, "_dmg" },
    { 0x669, "_dog_guard" },
    { 0x66A, "_dog_too_close_to_owner" },
    { 0x66B, "_domflageffect" },
    { 0x66D, "_effect_keys" },
    { 0x66E, "_effecttype" },
    { 0x670, "_enableusability" },
    { 0x671, "_enableweapon" },
    { 0x672, "_enableweaponswitch" },
    { 0x673, "_end" },
    { 0x674, "_end_swim" },
    { 0x675, "_end_wreck" },
    { 0x676, "_endbeach" },
    { 0x677, "_enemies" },
    { 0x678, "_enemy" },
    { 0x679, "_enemy_num" },
    { 0x67A, "_engine_room" },
    { 0x67B, "_essential_part" },
    { 0x67C, "_exfil" },
    { 0x67D, "_exfil_heli" },
    { 0x67E, "_exit_menu" },
    { 0x67F, "_explosion_fquakepower" },
    { 0x680, "_explosion_iblastradius" },
    { 0x681, "_explosion_idamagemax" },
    { 0x682, "_explosion_idamagemin" },
    { 0x683, "_explosion_imaxrange" },
    { 0x684, "_explosion_iminrange" },
    { 0x685, "_explosion_iquakeradius" },
    { 0x686, "_explosion_iquaketime" },
    { 0x687, "_explosion_last_incoming" },
    { 0x688, "_explosion_last_sound" },
    { 0x689, "_extra_autosave_checks" },
    { 0x68A, "_findunobstructedfiringpointhelper" },
    { 0x68B, "_fire" },
    { 0x68C, "_fire_damage_ent" },
    { 0x68D, "_fire_suppression" },
    { 0x68E, "_fires" },
    { 0x68F, "_firework_large" },
    { 0x690, "_firework_sunlight" },
    { 0x691, "_firework_wait" },
    { 0x692, "_fireworks_cleanup" },
    { 0x693, "_fireworks_internal" },
    { 0x694, "_fireworks_meteor_internal" },
    { 0x695, "_first_frame_anim" },
    { 0x696, "_flag_wait_trigger" },
    { 0x697, "_flags" },
    { 0x698, "_flarestack" },
    { 0x699, "_force_kill" },
    { 0x69A, "_freevehicle" },
    { 0x69B, "_fx" },
    { 0x69C, "_get_dummy" },
    { 0x69D, "_get_guard_node_behind_player" },
    { 0x69E, "_get_location_sunlight" },
    { 0x69F, "_get_player_tank_target" },
    { 0x6A0, "_getplayerdata" },
    { 0x6A1, "_getplayerscore" },
    { 0x6A2, "_getradarstrength" },
    { 0x6A5, "_getvehiclespawnerarray_by_spawngroup" },
    { 0x6A6, "_giveweapon" },
    { 0x6A7, "_global_fx_ents" },
    { 0x6A8, "_globals" },
    { 0x6A9, "_gopath" },
    { 0x6AA, "_hangar" },
    { 0x6AB, "_hasperk" },
    { 0x6AC, "_health_death" },
    { 0x6AD, "_heli_ai_pre_move_func_internal" },
    { 0x6AE, "_hint" },
    { 0x6AF, "_hint_stick_get_config_suffix" },
    { 0x6B0, "_hint_stick_update_breakfunc" },
    { 0x6B1, "_hint_stick_update_string" },
    { 0x6B2, "_ignore_settings_old" },
    { 0x6B3, "_index" },
    { 0x6B4, "_init" },
    { 0x6B5, "_interactive" },
    { 0x6B8, "_ishelicopter" },
    { 0x6B9, "_kill_fx" },
    { 0x6BA, "_lastanimtime" },
    { 0x6BC, "_lc" },
    { 0x6BD, "_lc_persists" },
    { 0x6BE, "_lever_col" },
    { 0x6BF, "_linked_triggers" },
    { 0x6C0, "_loadstarted" },
    { 0x6C1, "_max_ai" },
    { 0x6C3, "_mgoff" },
    { 0x6C5, "_missile_cleanup_fake_target" },
    { 0x6C6, "_missile_earthquake" },
    { 0x6C7, "_missile_start_lockon_notify" },
    { 0x6C9, "_mount_snowmobile" },
    { 0x6CA, "_nextcoverprint" },
    { 0x6CB, "_nextmission" },
    { 0x6CC, "_notetrackfx" },
    { 0x6CE, "_objective_delete" },
    { 0x6CF, "_old_goalradius" },
    { 0x6D0, "_patrol_endon_spotted_flag" },
    { 0x6D1, "_pick_best_node_behind_owner" },
    { 0x6D2, "_pick_best_node_heeled_by_owner" },
    { 0x6D3, "_pipe_deck" },
    { 0x6D4, "_pipe_fx_time" },
    { 0x6D5, "_pipe_methods" },
    { 0x6D6, "_pipes" },
    { 0x6D7, "_playlocalsound" },
    { 0x6D8, "_precache" },
    { 0x6D9, "_radio_queue" },
    { 0x6DA, "_refinery" },
    { 0x6DB, "_remote_turrets" },
    { 0x6DC, "_remoteturret_loc_table" },
    { 0x6DD, "_remove_nodes_too_close" },
    { 0x6DE, "_restorepreviousnameplatematerial" },
    { 0x6DF, "_retreat_current_volumes" },
    { 0x6E0, "_retreat_final" },
    { 0x6E1, "_retreat_standby" },
    { 0x6E4, "_rpl_legs_is_diagonal" },
    { 0x6E5, "_rpl_legs_is_horizontal" },
    { 0x6E6, "_sat" },
    { 0x6E7, "_script_exploders" },
    { 0x6E8, "_scripted_spawn" },
    { 0x6E9, "_set_anim_time" },
    { 0x6EA, "_setactionslot" },
    { 0x6EB, "_setextraperks" },
    { 0x6EC, "_sethighestmissionifnotcheating" },
    { 0x6ED, "_setmissiondiffstringifnotcheating" },
    { 0x6EE, "_setnameplatematerial" },
    { 0x6EF, "_setperk" },
    { 0x6F0, "_setplayerdata" },
    { 0x6F1, "_setplayerscore" },
    { 0x6F2, "_setswitchnode" },
    { 0x6F3, "_setteamscore" },
    { 0x6F4, "_setup_chair" },
    { 0x6F5, "_setvehgoalpos" },
    { 0x6F6, "_setvehgoalpos_wrap" },
    { 0x6F7, "_setvehgoalposadheretomesh" },
    { 0x6F8, "_sleeves_flap_internal" },
    { 0x6F9, "_sleeves_idle" },
    { 0x6FA, "_slomo_breach_blowback_guy" },
    { 0x6FC, "_slomo_breach_chair_guy_animated" },
    { 0x6FD, "_slomo_breach_chair_guy_normal" },
    { 0x6FE, "_slomo_breach_desk_guy" },
    { 0x6FF, "_slomo_breach_executed_guy" },
    { 0x700, "_slomo_breach_executed_guy_pushed_to_floor" },
    { 0x701, "_slomo_breach_executioner_knife" },
    { 0x702, "_slomo_breach_executioner_pistol" },
    { 0x703, "_slomo_breach_fightback_guy" },
    { 0x704, "_slomo_breach_hostage_react" },
    { 0x705, "_slomo_breach_knife_charger" },
    { 0x706, "_slomo_breach_knife_hostage_death" },
    { 0x707, "_slomo_breach_pistol_guy" },
    { 0x708, "_slowmo_breach_funcs" },
    { 0x709, "_slowmo_functions" },
    { 0x70A, "_sound" },
    { 0x70B, "_source" },
    { 0x70C, "_source_base" },
    { 0x70D, "_spawner_mg42_think" },
    { 0x70E, "_spawner_stealth_default" },
    { 0x70F, "_spawner_stealth_dog" },
    { 0x710, "_start" },
    { 0x711, "_stealth" },
    { 0x712, "_stealth_move_detection_cap" },
    { 0x713, "_suicide" },
    { 0x714, "_tag_entity" },
    { 0x715, "_takeweaponsexcept" },
    { 0x717, "_tanks" },
    { 0x718, "_target" },
    { 0x719, "_target_vols" },
    { 0x71A, "_thruster_ents" },
    { 0x71B, "_thruster_rig" },
    { 0x71C, "_timeout" },
    { 0x71D, "_timeout_pause_on_death_and_prematch" },
    { 0x71F, "_traverses" },
    { 0x720, "_trigger_handle_triggering" },
    { 0x721, "_turn_off_spec_sun_lerp" },
    { 0x722, "_turn_on_spec_sun_lerp" },
    { 0x723, "_unsetextraperks" },
    { 0x724, "_unsetperk" },
    { 0x725, "_up" },
    { 0x726, "_updateenemyusable" },
    { 0x727, "_updateteamusable" },
    { 0x728, "_useperkenabled" },
    { 0x729, "_validateattacker" },
    { 0x72A, "_vehicle_badplace" },
    { 0x72B, "_vehicle_effect" },
    { 0x72C, "_vehicle_is_crashing" },
    { 0x72D, "_vehicle_landvehicle" },
    { 0x72E, "_vehicle_paths" },
    { 0x72F, "_vehicle_resume_named" },
    { 0x730, "_vehicle_spawn" },
    { 0x731, "_vehicle_stop_named" },
    { 0x732, "_vehicle_unload" },
    { 0x733, "_vehicles" },
    { 0x734, "_vignette_active" },
    { 0x735, "_vision_sets_active" },
    { 0x736, "_vols" },
    { 0x738, "_walkway_brush" },
    { 0x739, "_walkway_brush_node" },
    { 0x73A, "_wavedelay" },
    { 0x73B, "_waveplayerspawnindex" },
    { 0x73C, "_window_imp" },
    { 0x73D, "a" },
    { 0x73E, "a10_30mm_fire" },
    { 0x73F, "a10_allies_target_logic" },
    { 0x740, "a10_ambient_clouds" },
    { 0x742, "a10_balcony_strafe_physics" },
    { 0x743, "a10_bridge_strike" },
    { 0x744, "a10_cockpit_breathing" },
    { 0x745, "a10_crash_approach" },
    { 0x746, "a10_crash_impact" },
    { 0x747, "a10_create_fake_ai" },
    { 0x748, "a10_create_fake_player" },
    { 0x749, "a10_delayed_hint" },
    { 0x74A, "a10_do_shots" },
    { 0x74B, "a10_enable_target" },
    { 0x74C, "a10_endrun_shooting" },
    { 0x74D, "a10_enemies_target_logic" },
    { 0x74E, "a10_explode" },
    { 0x74F, "a10_fake_ai_death" },
    { 0x750, "a10_fire_hint_func" },
    { 0x751, "a10_fire_missiles" },
    { 0x754, "a10_freezebuffer" },
    { 0x755, "a10_get_player_end_position" },
    { 0x756, "a10_gun_dives" },
    { 0x757, "a10_handledamage" },
    { 0x758, "a10_hint_func" },
    { 0x75A, "a10_hud_connected_pulse" },
    { 0x75B, "a10_hud_grain" },
    { 0x75C, "a10_hud_set_alpha" },
    { 0x75D, "a10_hud_set_altitude" },
    { 0x75E, "a10_hud_set_connection" },
    { 0x75F, "a10_hud_set_coords" },
    { 0x760, "a10_hud_set_id" },
    { 0x761, "a10_hud_set_speed" },
    { 0x762, "a10_inital_fire_check" },
    { 0x763, "a10_kill_notification" },
    { 0x764, "a10_lastweapon" },
    { 0x765, "a10_mechanic_off" },
    { 0x766, "a10_mechanic_skip_end" },
    { 0x767, "a10_missile_cleanup" },
    { 0x768, "a10_missile_dives" },
    { 0x769, "a10_missile_lockon" },
    { 0x76A, "a10_missile_set_target" },
    { 0x76B, "a10_originalaudiozone" },
    { 0x76C, "a10_player_30mm" },
    { 0x76D, "a10_player_30mm_fire" },
    { 0x76E, "a10_player_aftermission_report" },
    { 0x76F, "a10_player_hit_hudelem" },
    { 0x770, "a10_player_hit_strafe_vehicles" },
    { 0x771, "a10_player_hud_cleanup" },
    { 0x772, "a10_player_init" },
    { 0x773, "a10_player_init_hud" },
    { 0x774, "a10_player_kills" },
    { 0x775, "a10_player_lockon_warning" },
    { 0x776, "a10_player_shot_think" },
    { 0x777, "a10_precache" },
    { 0x778, "a10_remove_target" },
    { 0x779, "a10_spawn_funcs" },
    { 0x77A, "a10_squadron_friendlyfire_watcher" },
    { 0x77B, "a10_squadron_logic" },
    { 0x77C, "a10_squadron_shoot" },
    { 0x77D, "a10_squadron_tower_crash" },
    { 0x77E, "a10_squadron_tower_crash_flares_warthog" },
    { 0x77F, "a10_squadron_tower_crash_missile" },
    { 0x780, "a10_strafe_cleanup" },
    { 0x781, "a10_strafe_get_location_spawner" },
    { 0x782, "a10_strafe_groups" },
    { 0x783, "a10_strafe_impact" },
    { 0x784, "a10_strafe_impact_earthquake" },
    { 0x785, "a10_strafe_impacts" },
    { 0x786, "a10_strafe_mechanic" },
    { 0x787, "a10_strafe_respotioning" },
    { 0x788, "a10_strafe_respotioning_dialogue" },
    { 0x789, "a10_strafe_run" },
    { 0x78A, "a10_strafe_run_cheap" },
    { 0x78B, "a10_strafe_use_nag" },
    { 0x78C, "a10_target_logic" },
    { 0x78D, "a10_targeting_think" },
    { 0x78E, "a10_targeting_watcher" },
    { 0x78F, "a10_uses" },
    { 0x791, "a10_vista_strafe_group" },
    { 0x792, "a10_vista_strafe_group_delete" },
    { 0x793, "a10_vista_strafe_mig" },
    { 0x794, "a10_wait_fire_missile" },
    { 0x795, "a10_wait_start_firing" },
    { 0x796, "a10_wait_stop_firing" },
    { 0x797, "a10_warthog_strafe" },
    { 0x798, "a10endposition" },
    { 0x799, "a10fakeplayer" },
    { 0x79B, "a10splinesout" },
    { 0x79C, "a10strafeactive" },
    { 0x79E, "a_bad_guys" },
    { 0x7A1, "a_globals" },
    { 0x7A2, "a_guys" },
    { 0x7A3, "a_rel" },
    { 0x7A4, "aa_add_event" },
    { 0x7A5, "aa_add_event_float" },
    { 0x7A6, "aa_init_stats" },
    { 0x7A7, "aa_missile_fire" },
    { 0x7A8, "aa_player_ads_tracking" },
    { 0x7A9, "aa_player_attacks_enemy_with_ads" },
    { 0x7AA, "aa_player_health_tracking" },
    { 0x7AB, "aa_print_vals" },
    { 0x7AC, "aa_should_start_fresh" },
    { 0x7AD, "aa_time_tracking" },
    { 0x7AE, "aa_update_flags" },
    { 0x7AF, "aalauncherammo" },
    { 0x7B0, "aamissilelaunchhorz" },
    { 0x7B1, "aamissilelaunchtargetdist" },
    { 0x7B2, "aamissilelaunchvert" },
    { 0x7B3, "aammissilelaunchtargetdist" },
    { 0x7B4, "aas_guys_spawn_logic" },
    { 0x7B6, "aasoundmanager" },
    { 0x7B7, "abanglecutoff" },
    { 0x7B8, "abilitychosen" },
    { 0x7B9, "abilitymaxval" },
    { 0x7BA, "abort" },
    { 0x7BB, "abort_attack_requested" },
    { 0x7BC, "abort_count" },
    { 0x7BE, "abortapproachifthreatened" },
    { 0x7BF, "abortlevel" },
    { 0x7C0, "abortreloadwhencanshoot" },
    { 0x7C1, "abouttobebreached" },
    { 0x7C2, "above_water_start_setup" },
    { 0x7C3, "absangleclamp180" },
    { 0x7C4, "absolute" },
    { 0x7C5, "absyawtoangles" },
    { 0x7C6, "absyawtoenemy" },
    { 0x7C7, "absyawtoenemy2d" },
    { 0x7C9, "ac130" },
    { 0x7CA, "ac130_altscene" },
    { 0x7CB, "ac130_attack_random" },
    { 0x7CC, "ac130_attacked_player_count" },
    { 0x7CD, "ac130_constant_target" },
    { 0x7CE, "ac130_direct_attack_path" },
    { 0x7CF, "ac130_final_life" },
    { 0x7D0, "ac130_flood_respawn" },
    { 0x7D1, "ac130_kill_player" },
    { 0x7D2, "ac130_last_105_fire_time" },
    { 0x7D3, "ac130_magic_105" },
    { 0x7D4, "ac130_magic_105_fake" },
    { 0x7D5, "ac130_magic_105_impact" },
    { 0x7D7, "ac130_magic_bullet_fake" },
    { 0x7DA, "ac130_missile_take_hit" },
    { 0x7DC, "ac130_spawn" },
    { 0x7DD, "ac130_speed" },
    { 0x7DE, "ac130_use_duration" },
    { 0x7DF, "ac130gunner" },
    { 0x7E0, "ac130inuse" },
    { 0x7E1, "ac130player" },
    { 0x7E2, "ac130queue" },
    { 0x7E3, "ac130shellshock" },
    { 0x7E4, "ac_130" },
    { 0x7E5, "accaracy_mod" },
    { 0x7E6, "accel" },
    { 0x7E7, "accel_factor" },
    { 0x7E8, "accel_time" },
    { 0x7E9, "acceleration" },
    { 0x7EA, "acceleration_fov" },
    { 0x7EB, "accessory" },
    { 0x7EC, "accn" },
    { 0x7ED, "accumulated_damage" },
    { 0x7EE, "accuracy_ally" },
    { 0x7EF, "accuracy_enemy" },
    { 0x7F0, "accuracygrowthmultiplier" },
    { 0x7F1, "accuracystationarymod" },
    { 0x7F2, "achieve_birdie" },
    { 0x7F4, "achieve_jack_the_ripper" },
    { 0x7F5, "achieve_serrated_edge" },
    { 0x7F6, "achieve_slowmo_breach_kills" },
    { 0x7F7, "achieve_strike" },
    { 0x7F8, "achievement" },
    { 0x7F9, "achievement_attacker" },
    { 0x7FA, "achievement_completed" },
    { 0x7FB, "achievement_list" },
    { 0x7FC, "achievement_registration_func" },
    { 0x7FD, "acquiregroundtarget" },
    { 0x7FE, "acquireminitarget" },
    { 0x7FF, "acquiretarget" },
    { 0x800, "acquirevehicletarget" },
    { 0x801, "action_back" },
    { 0x802, "action_func" },
    { 0x803, "action_gears" },
    { 0x804, "action_killstreak" },
    { 0x806, "action_thread" },
    { 0x807, "action_weapons_primary" },
    { 0x808, "action_weapons_secondary" },
    { 0x809, "actionbinds" },
    { 0x80A, "actionnotify" },
    { 0x80B, "actionnotifymessage" },
    { 0x80C, "actionslotenabled" },
    { 0x80D, "actionslots" },
    { 0x80E, "activate" },
    { 0x80F, "activate_angered_state" },
    { 0x810, "activate_avoid_minion_exp" },
    { 0x811, "activate_clientside_exploder" },
    { 0x812, "activate_color_code_internal" },
    { 0x813, "activate_color_trigger" },
    { 0x815, "activate_destructibles_in_volume" },
    { 0x816, "activate_exploder" },
    { 0x817, "activate_exploders_in_volume" },
    { 0x819, "activate_fireworks_exploder" },
    { 0x81A, "activate_health_regen" },
    { 0x81C, "activate_individual_exploder" },
    { 0x81D, "activate_individual_exploder_proc" },
    { 0x81E, "activate_individual_fireworks_exploder" },
    { 0x81F, "activate_interactives_in_volume" },
    { 0x820, "activate_kill_10_in_30" },
    { 0x822, "activate_kill_10_with_traps" },
    { 0x823, "activate_kill_10_with_turrets" },
    { 0x824, "activate_kill_airborne_aliens" },
    { 0x826, "activate_melee_goons" },
    { 0x827, "activate_melee_only" },
    { 0x829, "activate_mortar" },
    { 0x82A, "activate_nerf" },
    { 0x82B, "activate_new_challenge" },
    { 0x82C, "activate_no_abilities" },
    { 0x82D, "activate_no_reloads" },
    { 0x82E, "activate_percent_accuracy" },
    { 0x82F, "activate_protect_a_player" },
    { 0x830, "activate_riders" },
    { 0x831, "activate_rotunda_fight" },
    { 0x832, "activate_spawn_event" },
    { 0x833, "activate_spend_currency" },
    { 0x834, "activate_spend_money_progress" },
    { 0x835, "activate_spend_no_money" },
    { 0x836, "activate_stay_prone" },
    { 0x837, "activate_stay_within_area" },
    { 0x838, "activate_trig_if_not_flag" },
    { 0x839, "activate_trigger" },
    { 0x83A, "activate_trigger_process" },
    { 0x83B, "activate_trigger_with_noteworthy" },
    { 0x83C, "activate_trigger_with_targetname" },
    { 0x83D, "activate_use_weapon_challenge" },
    { 0x83E, "activateagent" },
    { 0x83F, "activated" },
    { 0x840, "activated_color_trigger" },
    { 0x841, "activated_nerfs" },
    { 0x842, "activatefunc" },
    { 0x843, "activateratio" },
    { 0x844, "activatetime" },
    { 0x845, "activateweapon" },
    { 0x846, "activation_notify" },
    { 0x847, "activation_time" },
    { 0x848, "activation_trig" },
    { 0x84A, "active_deployables" },
    { 0x84B, "active_force_dog_talk" },
    { 0x84C, "active_mode" },
    { 0x84D, "active_objective" },
    { 0x84E, "active_odin" },
    { 0x84F, "active_rope" },
    { 0x850, "active_teargas" },
    { 0x851, "active_turrets" },
    { 0x852, "active_wait_spread" },
    { 0x853, "activebreaks" },
    { 0x854, "activecount" },
    { 0x855, "activecounteruavs" },
    { 0x857, "activegrenadetimer" },
    { 0x858, "activenodes" },
    { 0x85A, "activeplayers" },
    { 0x85B, "activesfx" },
    { 0x85C, "activeuavs" },
    { 0x85D, "actor_teleport" },
    { 0x85E, "actor_use_water_when_moving" },
    { 0x860, "actual_health" },
    { 0x861, "actuator_click_wait" },
    { 0x862, "adam" },
    { 0x863, "add_abort" },
    { 0x864, "add_actor_danger_listeners" },
    { 0x866, "add_and_select_entity" },
    { 0x867, "add_animation" },
    { 0x868, "add_animsound" },
    { 0x869, "add_array_to_destructible" },
    { 0x86A, "add_as_apache_target_on_spawn" },
    { 0x86B, "add_as_apaches_target" },
    { 0x86C, "add_attachment_to_weapon" },
    { 0x86E, "add_bcs_location_mapping" },
    { 0x86F, "add_beacon_effect" },
    { 0x870, "add_breach_func" },
    { 0x871, "add_breach_target" },
    { 0x872, "add_button" },
    { 0x873, "add_c4_glow" },
    { 0x874, "add_c4_to_spot" },
    { 0x875, "add_call" },
    { 0x876, "add_care_package" },
    { 0x877, "add_cellphone_notetracks" },
    { 0x878, "add_cleanup_ent" },
    { 0x87B, "add_collision_to_path_ent" },
    { 0x87C, "add_context_sensative_dialog" },
    { 0x87D, "add_context_sensative_timeout" },
    { 0x87E, "add_contrail" },
    { 0x880, "add_cover_node" },
    { 0x881, "add_cycle_scalar" },
    { 0x883, "add_damage_function" },
    { 0x884, "add_damage_owner_recorder" },
    { 0x885, "add_damagefeedback" },
    { 0x886, "add_dead_enemy_clip" },
    { 0x887, "add_debug_dialogue" },
    { 0x888, "add_destructible_fx" },
    { 0x889, "add_destructible_to_frame_queue" },
    { 0x88A, "add_destructible_type_function" },
    { 0x88C, "add_dialogue_line" },
    { 0x88D, "add_dialogue_line_timed" },
    { 0x88E, "add_earthquake" },
    { 0x88F, "add_endon" },
    { 0x890, "add_ent_objective_to_compass" },
    { 0x891, "add_extra_autosave_check" },
    { 0x892, "add_fire" },
    { 0x893, "add_fire_fx" },
    { 0x894, "add_fractional_data_point" },
    { 0x896, "add_fx" },
    { 0x898, "add_headlamp" },
    { 0x899, "add_hint_background" },
    { 0x89A, "add_hint_string" },
    { 0x89B, "add_hive_dependencies" },
    { 0x89C, "add_hud_line" },
    { 0x89D, "add_hudelm_position_internal" },
    { 0x89E, "add_humanoid_agent" },
    { 0x89F, "add_ice_radius" },
    { 0x8A0, "add_in_more_allies" },
    { 0x8A1, "add_jav_glow" },
    { 0x8A2, "add_kb_button" },
    { 0x8A3, "add_key" },
    { 0x8A4, "add_key_to_destructible" },
    { 0x8A5, "add_keypairs_to_destructible" },
    { 0x8A6, "add_large_firework" },
    { 0x8A7, "add_lcs_target" },
    { 0x8A8, "add_light_to_actor" },
    { 0x8A9, "add_magic_bullet_shield_if_off" },
    { 0x8AA, "add_meteor_firework" },
    { 0x8AB, "add_name" },
    { 0x8AC, "add_no_game_starts" },
    { 0x8AE, "add_noself_call" },
    { 0x8AF, "add_notetrack_and_get_index" },
    { 0x8B0, "add_notetrack_array" },
    { 0x8B1, "add_option_to_selected_entities" },
    { 0x8B3, "add_path_start_and_end_refs" },
    { 0x8B4, "add_pet_bombs_to_pillage_system" },
    { 0x8B5, "add_proccess_trigger" },
    { 0x8B7, "add_random_killspawner_to_spawngroup" },
    { 0x8B8, "add_reactive_fx" },
    { 0x8B9, "add_reverb" },
    { 0x8BA, "add_scene_model" },
    { 0x8BB, "add_sit_load_ak_notetracks" },
    { 0x8BC, "add_slowmo_breach_custom_function" },
    { 0x8BD, "add_slowmo_breacher" },
    { 0x8BE, "add_small_firework" },
    { 0x8BF, "add_smoking_notetracks" },
    { 0x8C0, "add_spawn_function" },
    { 0x8C1, "add_start" },
    { 0x8C2, "add_start_assert" },
    { 0x8C3, "add_start_construct" },
    { 0x8C4, "add_target" },
    { 0x8C5, "add_target_pivot" },
    { 0x8C6, "add_teargas_cloud_radius" },
    { 0x8C9, "add_to_array" },
    { 0x8CA, "add_to_bot_damage_targets" },
    { 0x8CB, "add_to_bot_use_targets" },
    { 0x8CC, "add_to_destroyed_count" },
    { 0x8CD, "add_to_dialogue" },
    { 0x8CE, "add_to_dialogue_generic" },
    { 0x8CF, "add_to_enemygazs_until_dead" },
    { 0x8D0, "add_to_enemytanks_until_dead" },
    { 0x8D1, "add_to_group" },
    { 0x8D2, "add_to_interrupt_vo" },
    { 0x8D4, "add_to_outline_hive_watch_list" },
    { 0x8D5, "add_to_outline_pillage_watch_list" },
    { 0x8D6, "add_to_outline_watch_list" },
    { 0x8D7, "add_to_outline_weapon_watch_list" },
    { 0x8D8, "add_to_queue_at_priority" },
    { 0x8D9, "add_to_radio" },
    { 0x8DA, "add_to_spawngroup" },
    { 0x8DB, "add_to_standby" },
    { 0x8DC, "add_to_thrown_entity_list" },
    { 0x8DD, "add_to_vo_system" },
    { 0x8DE, "add_to_vo_system_internal" },
    { 0x8DF, "add_to_zone" },
    { 0x8E0, "add_tokens_to_trigger_flags" },
    { 0x8E1, "add_trace_fx" },
    { 0x8E2, "add_trace_fx_proc" },
    { 0x8E3, "add_trigger_func_thread" },
    { 0x8E4, "add_trigger_function" },
    { 0x8E5, "add_turret_to_heli" },
    { 0x8E6, "add_valid_evade" },
    { 0x8E7, "add_vol_to_node" },
    { 0x8E8, "add_volume_to_global_arrays" },
    { 0x8E9, "add_wait" },
    { 0x8EA, "add_wait_asserter" },
    { 0x8EB, "add_weapon" },
    { 0x8EC, "add_z" },
    { 0x8ED, "addactioncovermealiasex" },
    { 0x8EE, "addactivecounteruav" },
    { 0x8EF, "addactiveuav" },
    { 0x8F0, "addaieventlistener_func" },
    { 0x8F1, "addairexplosion" },
    { 0x8F2, "addalienagent" },
    { 0x8F3, "addalienweaponammo" },
    { 0x8F4, "addallowedthreatcallout" },
    { 0x8F5, "addallweaponammo" },
    { 0x8F6, "addalternatespawnpoint" },
    { 0x8F7, "addammo" },
    { 0x8F8, "addammoovertime" },
    { 0x8F9, "addasapachehudtarget" },
    { 0x8FA, "addattacker" },
    { 0x8FB, "addawardwinner" },
    { 0x8FC, "addblankiw6" },
    { 0x8FD, "addboxtolevelarray" },
    { 0x8FE, "addcalloutresponseevent" },
    { 0x8FF, "addcastiw" },
    { 0x900, "addcastname" },
    { 0x901, "addcenterdual" },
    { 0x902, "addcenterediw6" },
    { 0x903, "addcenterheading" },
    { 0x904, "addcenterimage" },
    { 0x905, "addcentername" },
    { 0x906, "addcenternamedouble" },
    { 0x908, "addcheckfirealias" },
    { 0x909, "addchild" },
    { 0x90A, "addconcatdirectionalias" },
    { 0x90B, "addconcattargetalias" },
    { 0x90C, "addcratetype" },
    { 0x90F, "addcreditiw6_3f" },
    { 0x910, "addcreditiw6_4" },
    { 0x911, "addcreditliw6" },
    { 0x912, "addcreditlsubheaderriw6" },
    { 0x913, "addcreditriw6" },
    { 0x914, "adddeathicon" },
    { 0x915, "addedtowave" },
    { 0x916, "addenemytominimap" },
    { 0x917, "addentryiw6" },
    { 0x918, "addfullcliptoallweapons" },
    { 0x91A, "addgap" },
    { 0x91B, "addgrenadethrowanimoffset" },
    { 0x91C, "addheaderiw6" },
    { 0x91D, "addhostileburstalias" },
    { 0x91E, "addimageiw" },
    { 0x91F, "addinformalias" },
    { 0x920, "addinformevent" },
    { 0x921, "addinformreloadingaliasex" },
    { 0x922, "additionalassets" },
    { 0x923, "additionalexit_vo" },
    { 0x924, "additionalsighttraceentities" },
    { 0x925, "additive_pain" },
    { 0x927, "additiveturretdriveidle" },
    { 0x928, "additiveturretfire" },
    { 0x929, "additiveturretidle" },
    { 0x92A, "additiveturretrotateleft" },
    { 0x92B, "additiveturretrotateright" },
    { 0x92C, "additiveusegunroot" },
    { 0x92D, "addleftimage" },
    { 0x92E, "addleftname" },
    { 0x92F, "addleftnamename" },
    { 0x930, "addlefttitle" },
    { 0x931, "addlefttitlename" },
    { 0x932, "addlefttitlenamespace" },
    { 0x933, "addlevel" },
    { 0x934, "addlevelstoexperience" },
    { 0x936, "addlinknode" },
    { 0x937, "addlockedontarget" },
    { 0x939, "addlowermessage" },
    { 0x93A, "addmovecombataliasex" },
    { 0x93B, "addmovenoncombataliasex" },
    { 0x93C, "addnamealias" },
    { 0x93D, "addnamealiasex" },
    { 0x93E, "addnameiw" },
    { 0x93F, "addnotetrack_animsound" },
    { 0x940, "addnotetrack_attach" },
    { 0x941, "addnotetrack_clockwork" },
    { 0x942, "addnotetrack_cornered" },
    { 0x943, "addnotetrack_customfunction" },
    { 0x944, "addnotetrack_detach" },
    { 0x945, "addnotetrack_detach_gun" },
    { 0x946, "addnotetrack_dialogue" },
    { 0x947, "addnotetrack_flag" },
    { 0x948, "addnotetrack_flag_clear" },
    { 0x949, "addnotetrack_notify" },
    { 0x94A, "addnotetrack_playersound" },
    { 0x94B, "addnotetrack_sound" },
    { 0x94C, "addnotetrack_startfxontag" },
    { 0x94D, "addnotetrack_stopfxontag" },
    { 0x94E, "addnotetrack_swapparttoefx" },
    { 0x94F, "addnotetrack_tracepartforefx" },
    { 0x950, "addofficertosquad" },
    { 0x953, "addoption" },
    { 0x954, "addorderalias" },
    { 0x955, "addorderevent" },
    { 0x956, "addpaneltimesarray" },
    { 0x957, "addplanetolist" },
    { 0x958, "addplayernamealias" },
    { 0x959, "addplayertosquad" },
    { 0x95A, "addpossiblethreatcallout" },
    { 0x95B, "addprereq" },
    { 0x95C, "addrankalias" },
    { 0x95D, "addratiomaxstocktoallweapons" },
    { 0x95E, "addreactionalias" },
    { 0x95F, "addreactionevent" },
    { 0x961, "addresponsealias" },
    { 0x962, "addresponseevent" },
    { 0x963, "addresponseevent_internal" },
    { 0x964, "addrightname" },
    { 0x965, "addrighttitle" },
    { 0x966, "addsafetyhealth" },
    { 0x967, "addsituationalcombatorder" },
    { 0x968, "addsituationalorder" },
    { 0x969, "addspace" },
    { 0x96A, "addspacesmall" },
    { 0x96C, "addspawnpoints" },
    { 0x96D, "addspeaker" },
    { 0x96E, "addstartspawnpoints" },
    { 0x96F, "addsubheaderiw6" },
    { 0x970, "addsubheaderliw6" },
    { 0x971, "addsubheaderriw6" },
    { 0x972, "addsubleftname" },
    { 0x973, "addsubleftnamename" },
    { 0x974, "addsubleftnamenamename" },
    { 0x976, "addsublefttitlename" },
    { 0x977, "addsublefttitlenamespace" },
    { 0x978, "addsubnameiw" },
    { 0x979, "addsubtitleiw" },
    { 0x97A, "addsubtitleiw6" },
    { 0x97B, "addsubtitlenameiw" },
    { 0x97C, "addtakingfirealias" },
    { 0x97D, "addtauntalias" },
    { 0x97E, "addthreatalias" },
    { 0x97F, "addthreatcalloutalias" },
    { 0x980, "addthreatcalloutecho" },
    { 0x981, "addthreatcalloutlandmarkalias" },
    { 0x982, "addthreatcalloutlocationalias" },
    { 0x983, "addthreatcalloutqa_nextline" },
    { 0x984, "addthreatcalloutresponsealias" },
    { 0x985, "addthreatdistancealias" },
    { 0x986, "addthreatelevationalias" },
    { 0x987, "addthreatevent" },
    { 0x988, "addthreatexposedalias" },
    { 0x989, "addthreatobviousalias" },
    { 0x98A, "addtime" },
    { 0x98B, "addtitleiw" },
    { 0x98C, "addtitleiw6" },
    { 0x98D, "addtitlenameiw" },
    { 0x98F, "addtoballdronelist" },
    { 0x990, "addtobattlebuddywaitlist" },
    { 0x991, "addtocharactersarray" },
    { 0x992, "addtoclosedlist" },
    { 0x993, "addtohelilist" },
    { 0x994, "addtoimslist" },
    { 0x995, "addtolittlebirdlist" },
    { 0x996, "addtolivescount" },
    { 0x997, "addtoopenlist" },
    { 0x998, "addtoparticipantsarray" },
    { 0x999, "addtosquad" },
    { 0x99B, "addtotanklist" },
    { 0x99C, "addtoteam" },
    { 0x99D, "addtoteamcount" },
    { 0x99E, "addtoturretlist" },
    { 0x99F, "addtougvlist" },
    { 0x9A2, "addtrackingtarget_ondeath" },
    { 0x9A3, "addtrackingtarget_update" },
    { 0x9A5, "adduavmodel" },
    { 0x9A6, "adduplinktolevellist" },
    { 0x9A7, "addzone" },
    { 0x9A8, "adjust_ally_movement" },
    { 0x9A9, "adjust_angles_to_player" },
    { 0x9AA, "adjust_forward_push" },
    { 0x9AB, "adjust_helo_sound_high" },
    { 0x9AC, "adjust_helo_sound_low" },
    { 0x9AD, "adjust_helo_sound_roll" },
    { 0x9AE, "adjust_movement_step_up" },
    { 0x9AF, "adjust_moving_grass" },
    { 0x9B0, "adjust_overall_apache_pitches" },
    { 0x9B1, "adjust_player_view" },
    { 0x9B2, "adjust_suppression_on_enemies" },
    { 0x9B3, "adrenaline" },
    { 0x9B4, "adrenalineinfo" },
    { 0x9B5, "adrenalinetime" },
    { 0x9B6, "ads_hint" },
    { 0x9B7, "adstime" },
    { 0x9B8, "adstoggled" },
    { 0x9B9, "adszoomed" },
    { 0x9BA, "advance_regardless_of_numbers" },
    { 0x9BB, "advancedtraverse" },
    { 0x9BC, "advancedtraverse2" },
    { 0x9BE, "advanceonhidingenemy" },
    { 0x9BF, "advancetoenemygroup" },
    { 0x9C0, "advancetoenemygroupmax" },
    { 0x9C1, "advancetoenemyinterval" },
    { 0x9C2, "aenemies" },
    { 0x9C3, "aent_flag_waitopen_either" },
    { 0x9C4, "aerial_vehicle_allowed" },
    { 0x9C5, "affected" },
    { 0x9C6, "afk" },
    { 0x9C7, "after_fall_bounce" },
    { 0x9C8, "after_hunt" },
    { 0x9C9, "after_hunt_dialogue" },
    { 0x9CA, "agent_damage_finished" },
    { 0x9CC, "agent_gameparticipant" },
    { 0x9CD, "agent_teamparticipant" },
    { 0x9CE, "agent_type" },
    { 0x9CF, "agentarray" },
    { 0x9D0, "agentdogthink" },
    { 0x9D1, "agentfunc" },
    { 0x9D2, "aggresivelookat" },
    { 0x9D3, "aggressivemode" },
    { 0x9D4, "ahdmode" },
    { 0x9D5, "ahdmode_ng" },
    { 0x9D7, "ai_3d_sighting_model" },
    { 0x9D8, "ai_advancing_logic" },
    { 0x9D9, "ai_alert" },
    { 0x9DA, "ai_alert_bullet" },
    { 0x9DB, "ai_alert_damage" },
    { 0x9DC, "ai_alert_friend_death" },
    { 0x9DD, "ai_alert_loop" },
    { 0x9DE, "ai_alert_player_break_stealth" },
    { 0x9DF, "ai_alert_range" },
    { 0x9E0, "ai_animate_props_on_death" },
    { 0x9E1, "ai_array" },
    { 0x9E2, "ai_array_killcount_flag_set" },
    { 0x9E3, "ai_attack_missile" },
    { 0x9E4, "ai_classname_in_level" },
    { 0x9E6, "ai_clean_up" },
    { 0x9E7, "ai_cleanup_fake_death" },
    { 0x9E8, "ai_clear_custom_animation_reaction" },
    { 0x9E9, "ai_clear_custom_animation_reaction_and_idle" },
    { 0x9EB, "ai_create_behavior_function" },
    { 0x9EC, "ai_damage_think" },
    { 0x9ED, "ai_deathflag" },
    { 0x9EE, "ai_debug" },
    { 0x9EF, "ai_delete_when_out_of_sight" },
    { 0x9F0, "ai_dont_glow_in_thermal" },
    { 0x9F1, "ai_enemy_target_underwater" },
    { 0x9F2, "ai_enemy_tracking" },
    { 0x9F3, "ai_event_settings" },
    { 0x9F4, "ai_event_settings_reset" },
    { 0x9F5, "ai_flee_from_teargas" },
    { 0x9F6, "ai_flooding_hip_anims" },
    { 0x9F7, "ai_flooding_under_anims" },
    { 0x9F8, "ai_follow_cover" },
    { 0x9FA, "ai_get_behavior_function" },
    { 0x9FB, "ai_go_to_player" },
    { 0x9FC, "ai_group_killcount_flag_set" },
    { 0x9FD, "ai_in_coverwater" },
    { 0x9FE, "ai_init" },
    { 0x9FF, "ai_lasers" },
    { 0xA00, "ai_message_handler_hidden" },
    { 0xA02, "ai_mode" },
    { 0xA03, "ai_number" },
    { 0xA05, "ai_out_time" },
    { 0xA06, "ai_picks_destination" },
    { 0xA08, "ai_react_to_teargas" },
    { 0xA09, "ai_record_spawn_pos" },
    { 0xA0A, "ai_remove_outline_waiter" },
    { 0xA0B, "ai_rider_invulnerable_until_vehicle_death" },
    { 0xA0C, "ai_rider_invulnerable_until_vehicle_death_or_jumping_out" },
    { 0xA0D, "ai_rpg_attack_delay_max" },
    { 0xA0E, "ai_rpg_attack_delay_min" },
    { 0xA0F, "ai_set_custom_animation_reaction" },
    { 0xA10, "ai_set_goback_override_function" },
    { 0xA11, "ai_sets_goal" },
    { 0xA13, "ai_should_be_added" },
    { 0xA14, "ai_sight_brushes" },
    { 0xA15, "ai_space_death" },
    { 0xA16, "ai_space_headshot_death" },
    { 0xA17, "ai_space_pain" },
    { 0xA19, "ai_stealth_init" },
    { 0xA1A, "ai_stealth_pause_handler" },
    { 0xA1B, "ai_swim_death" },
    { 0xA1C, "ai_swim_pain" },
    { 0xA1D, "ai_swim_sound" },
    { 0xA1E, "ai_swim_sound_idle" },
    { 0xA1F, "ai_test_count" },
    { 0xA20, "ai_to_kill" },
    { 0xA21, "ai_total_count" },
    { 0xA22, "ai_track" },
    { 0xA23, "ai_track_death_by_player" },
    { 0xA24, "ai_types" },
    { 0xA25, "ai_wait_go" },
    { 0xA26, "ai_waittill_entered_vehicle" },
    { 0xA27, "ai_water_rising_think" },
    { 0xA28, "aiamount" },
    { 0xA29, "aiareintheroom" },
    { 0xA2A, "aiarray" },
    { 0xA2B, "aibattlechatterloop" },
    { 0xA2C, "aicount" },
    { 0xA2D, "aideathenemy" },
    { 0xA2E, "aideatheventthread" },
    { 0xA2F, "aideathfriendly" },
    { 0xA30, "aidisplacewaiter" },
    { 0xA31, "aifolloworderwaiter" },
    { 0xA32, "aigrenadedangerwaiter" },
    { 0xA33, "aigroup_create" },
    { 0xA34, "aigroup_soldierthink" },
    { 0xA35, "aigroup_spawnerdeath" },
    { 0xA36, "aigroup_spawnerempty" },
    { 0xA37, "aigroup_spawnerthink" },
    { 0xA38, "aihasweapon" },
    { 0xA3A, "aikilleventthread" },
    { 0xA3B, "aim2_target" },
    { 0xA3C, "aim4_target" },
    { 0xA3D, "aim6_target" },
    { 0xA3E, "aim8_target" },
    { 0xA3F, "aim_arrow" },
    { 0xA40, "aim_arrow_on_target" },
    { 0xA42, "aim_idle_thread" },
    { 0xA43, "aim_missiles_2" },
    { 0xA44, "aim_turret_at_ambush_point_or_visible_enemy" },
    { 0xA45, "aim_while_moving_thread" },
    { 0xA46, "aimbutdontshoot" },
    { 0xA47, "aimed_at_shoot_ent_or_pos" },
    { 0xA48, "aimedatshootentorpos" },
    { 0xA49, "aimedsomewhatatenemy" },
    { 0xA4A, "aimidlethread" },
    { 0xA4B, "aiming_at_ally" },
    { 0xA4C, "aimpitchdifftolerance" },
    { 0xA4D, "aimweight" },
    { 0xA4E, "aimweight_end" },
    { 0xA4F, "aimweight_start" },
    { 0xA50, "aimweight_t" },
    { 0xA51, "aimweight_transframes" },
    { 0xA52, "aimyawdiffclosedistsq" },
    { 0xA55, "ainame" },
    { 0xA56, "ainameandrankwaiter" },
    { 0xA57, "aiofficerorders" },
    { 0xA58, "aiowner" },
    { 0xA59, "air_armada" },
    { 0xA5A, "air_dropped" },
    { 0xA5B, "air_node_mesh" },
    { 0xA5C, "air_raid" },
    { 0xA5E, "air_raid_fire" },
    { 0xA5F, "air_raid_siren" },
    { 0xA61, "air_raids" },
    { 0xA62, "air_start_nodes" },
    { 0xA63, "air_strip_ai_quick_cleanup_death_function" },
    { 0xA64, "air_strip_ai_quick_cleanup_spawn_function" },
    { 0xA65, "air_strip_ambient_a10_gun_dive_1" },
    { 0xA66, "air_strip_ambient_a10_gun_dive_2" },
    { 0xA67, "air_strip_ambient_a10_gun_dive_3" },
    { 0xA68, "air_strip_ambient_dogfight_1" },
    { 0xA69, "air_strip_ambient_dogfight_2" },
    { 0xA6B, "air_strip_begin" },
    { 0xA6C, "air_strip_choppers" },
    { 0xA6D, "air_strip_cleanup" },
    { 0xA6E, "air_strip_hints" },
    { 0xA6F, "air_strip_init" },
    { 0xA70, "air_strip_m880_corpses" },
    { 0xA71, "air_strip_m880_death_count" },
    { 0xA72, "air_strip_m880s" },
    { 0xA73, "air_strip_main" },
    { 0xA74, "air_strip_obj_markers" },
    { 0xA75, "air_strip_secured_ambient_enemies_setup" },
    { 0xA77, "air_strip_secured_init" },
    { 0xA78, "air_strip_secured_main" },
    { 0xA79, "air_strip_secured_vo" },
    { 0xA7A, "air_strip_take_off_mig_01" },
    { 0xA7B, "air_strip_take_off_mig_02" },
    { 0xA7C, "air_strip_temp_dialog" },
    { 0xA7D, "air_strip_to_chopper" },
    { 0xA7E, "air_strip_trucks_static_setup" },
    { 0xA7F, "air_strip_victory" },
    { 0xA81, "airank" },
    { 0xA82, "airburstbomb" },
    { 0xA83, "aircraft_wash" },
    { 0xA85, "airdeniedplayer" },
    { 0xA86, "airdrop_heli" },
    { 0xA87, "airdrop_icon" },
    { 0xA88, "airdrop_max_linear_velocity" },
    { 0xA89, "airdrop_override_death_moving_platform" },
    { 0xA8A, "airdrop_override_invalid_moving_platform" },
    { 0xA8B, "airdrop_reward" },
    { 0xA8C, "airdropcratecollision" },
    { 0xA8D, "airdropdetonateonstuck" },
    { 0xA8E, "airdropmarkeractivate" },
    { 0xA8F, "airdroptype" },
    { 0xA91, "airlock_glass_fog" },
    { 0xA92, "airlock_interior_hatch" },
    { 0xA93, "airlockexplode" },
    { 0xA94, "airplane_list" },
    { 0xA95, "airshipflydefense" },
    { 0xA97, "airshipfx" },
    { 0xA98, "airshipfxonclient" },
    { 0xA99, "airshipfxonconnect" },
    { 0xA9A, "airshippitchhatchdown" },
    { 0xA9B, "airshippitchhatchup" },
    { 0xA9C, "airshippitchpropsdown" },
    { 0xA9D, "airshippitchpropsup" },
    { 0xA9E, "airspread" },
    { 0xA9F, "airstrike_earthquake" },
    { 0xAA0, "airstrikedamagedents" },
    { 0xAA1, "airstrikedamagedentscount" },
    { 0xAA2, "airstrikedamagedentsindex" },
    { 0xAA3, "airstrikedamageentsthread" },
    { 0xAA4, "airstrikeexplosion" },
    { 0xAA6, "airstrikeheightscale" },
    { 0xAA7, "airstrikeinprogress" },
    { 0xAA8, "airstrikemadeselectionvo" },
    { 0xAA9, "airstrikessfx" },
    { 0xAAA, "airstriketype" },
    { 0xAAB, "aishootplayer" },
    { 0xAAC, "aispread" },
    { 0xAAD, "aistate" },
    { 0xAAE, "aisuppressai" },
    { 0xAAF, "aithreadthreader" },
    { 0xAB0, "aiturnnotifies" },
    { 0xAB1, "aitype_check" },
    { 0xAB2, "aiupdateanimstate" },
    { 0xAB3, "aiupdatecombat" },
    { 0xAB4, "aiupdatesuppressed" },
    { 0xAB5, "aiweapon" },
    { 0xAB6, "ajax_flare" },
    { 0xAB7, "alarm_annoyance" },
    { 0xAB8, "alarm_ent" },
    { 0xAB9, "alarm_interval" },
    { 0xABA, "alarm_playing" },
    { 0xABB, "alarm_validate_damage" },
    { 0xABC, "alarms" },
    { 0xABD, "alarms2" },
    { 0xABE, "alarms_1" },
    { 0xABF, "alarms_2" },
    { 0xAC0, "alarms_3" },
    { 0xAC1, "alert_all" },
    { 0xAC3, "alert_enemies" },
    { 0xAC4, "alert_enemies_early" },
    { 0xAC5, "alert_enemies_react" },
    { 0xAC6, "alert_level" },
    { 0xAC7, "alert_level_table" },
    { 0xAC8, "alert_on_chopper_damage" },
    { 0xAC9, "alert_team" },
    { 0xACA, "alertface" },
    { 0xACB, "alias" },
    { 0xACC, "alien_ai_debug_print" },
    { 0xACD, "alien_area_init" },
    { 0xACE, "alien_attack" },
    { 0xACF, "alien_attack_enemy" },
    { 0xAD0, "alien_attack_sequence" },
    { 0xAD1, "alien_attribute_table_init" },
    { 0xAD2, "alien_begindeployableviamarker" },
    { 0xAD3, "alien_challenge_table" },
    { 0xAD4, "alien_character_cac_table" },
    { 0xAD5, "alien_cloak" },
    { 0xAD6, "alien_collectibles_table" },
    { 0xAD7, "alien_combat_resource_callbacks" },
    { 0xAD8, "alien_combat_resources" },
    { 0xAD9, "alien_combat_resources_table" },
    { 0xADA, "alien_customprematchperiod" },
    { 0xADB, "alien_cycle_intermission" },
    { 0xADC, "alien_cycle_table" },
    { 0xADD, "alien_cycle_table_hardcore" },
    { 0xADE, "alien_eyes_off" },
    { 0xADF, "alien_eyes_off_on_death" },
    { 0xAE0, "alien_eyes_on" },
    { 0xAE1, "alien_eyes_on_threaded" },
    { 0xAE2, "alien_fire_off" },
    { 0xAE3, "alien_fire_on" },
    { 0xAE4, "alien_funcs" },
    { 0xAE5, "alien_health_per_player_init" },
    { 0xAE6, "alien_health_per_player_scalar" },
    { 0xAE8, "alien_jump_melee_speed" },
    { 0xAE9, "alien_loadout" },
    { 0xAEA, "alien_loot_initialized" },
    { 0xAEB, "alien_lurker_behavior" },
    { 0xAEC, "alien_lurker_init" },
    { 0xAED, "alien_lurkers" },
    { 0xAEE, "alien_main_loop" },
    { 0xAEF, "alien_make_entity_sentient" },
    { 0xAF0, "alien_max_rank" },
    { 0xAF1, "alien_melee" },
    { 0xAF2, "alien_mode" },
    { 0xAF3, "alien_mode_enable" },
    { 0xAF4, "alien_mode_enable_raw" },
    { 0xAF5, "alien_mode_feature" },
    { 0xAF6, "alien_mode_feature_strings" },
    { 0xAF7, "alien_mode_has" },
    { 0xAF8, "alien_noncombat" },
    { 0xAFA, "alien_perks" },
    { 0xAFB, "alien_perks_table" },
    { 0xAFC, "alien_pet" },
    { 0xAFD, "alien_pet_follow" },
    { 0xAFE, "alien_player_spawn_group" },
    { 0xAFF, "alien_pregame_delay" },
    { 0xB00, "alien_ranks" },
    { 0xB02, "alien_retreat" },
    { 0xB03, "alien_scene_behavior" },
    { 0xB04, "alien_scene_init" },
    { 0xB05, "alien_scripted" },
    { 0xB06, "alien_synch_attack_enemy" },
    { 0xB07, "alien_test_jump" },
    { 0xB08, "alien_test_loop" },
    { 0xB09, "alien_town_intro" },
    { 0xB0A, "alien_town_intro_precache_characters" },
    { 0xB0C, "alien_type" },
    { 0xB0E, "alien_unlock_data" },
    { 0xB0F, "alien_unlock_table" },
    { 0xB11, "alien_used_resource_rank" },
    { 0xB12, "alien_vo_priority_level" },
    { 0xB14, "alien_wave" },
    { 0xB15, "alien_wave_behavior" },
    { 0xB16, "alien_wave_init" },
    { 0xB18, "alien_wave_spawn_think" },
    { 0xB19, "alien_wave_status" },
    { 0xB1A, "alien_wave_table" },
    { 0xB1B, "alien_xp" },
    { 0xB1C, "alienagentspawn" },
    { 0xB1D, "alienagentthink" },
    { 0xB1E, "alienanimdata" },
    { 0xB1F, "alienbbdata" },
    { 0xB20, "alienclimbdown" },
    { 0xB21, "alienclimbup" },
    { 0xB22, "alienendgame" },
    { 0xB24, "alienmovebackanimchance" },
    { 0xB25, "alienoutcomenotify" },
    { 0xB26, "alienplayerarmor" },
    { 0xB29, "alienplayerpainbreathingsound" },
    { 0xB2A, "alienregulartraversal" },
    { 0xB2B, "aliens_give_currency_func" },
    { 0xB2C, "aliens_make_entity_sentient_func" },
    { 0xB2D, "aliensnarecount" },
    { 0xB2F, "alienspawnlogic" },
    { 0xB30, "alientraversenotetrackhandler" },
    { 0xB31, "alientypecandofriendlydamage" },
    { 0xB32, "align_chalk_marks" },
    { 0xB33, "aligntoverticaledge" },
    { 0xB35, "alivecount" },
    { 0xB36, "aliveplayers" },
    { 0xB37, "all_challenge_completed" },
    { 0xB38, "all_dom_flags" },
    { 0xB39, "all_players_istouching" },
    { 0xB3A, "all_rappel_pt3_downstairs_enemies" },
    { 0xB3C, "all_scenes" },
    { 0xB3D, "all_team_steak_col" },
    { 0xB3E, "all_team_streak_col" },
    { 0xB3F, "allcrashes" },
    { 0xB40, "alley_bokehdots" },
    { 0xB41, "alley_bokehdots_old" },
    { 0xB42, "alley_end_of_alley_fx" },
    { 0xB44, "alley_fill_shallow" },
    { 0xB45, "alley_flood" },
    { 0xB46, "alley_flood_collision_cheater" },
    { 0xB47, "alley_flood_far_vfx_attachments" },
    { 0xB48, "alley_flood_fx" },
    { 0xB49, "alley_flood_near_vfx_attachments" },
    { 0xB4A, "alley_flood_spawn" },
    { 0xB4C, "alley_flood_water" },
    { 0xB4D, "alley_froth_vfx" },
    { 0xB4E, "alley_giantsplashes_left" },
    { 0xB50, "alley_kill_triggers" },
    { 0xB51, "alley_near" },
    { 0xB52, "alley_stumble" },
    { 0xB53, "allie1_tussbubbs" },
    { 0xB55, "allies_baker_command_end_anim" },
    { 0xB56, "allies_baker_console_anims" },
    { 0xB57, "allies_baker_flarestack_exit" },
    { 0xB58, "allies_baker_hold" },
    { 0xB59, "allies_baker_hold_approach_and_idle" },
    { 0xB5A, "allies_breach_anim_node" },
    { 0xB5B, "allies_building_entry_movement" },
    { 0xB5C, "allies_building_entry_vo" },
    { 0xB5D, "allies_building_exit_hookup" },
    { 0xB60, "allies_convoy_dialogue" },
    { 0xB61, "allies_cqbwalk" },
    { 0xB62, "allies_dam_vign" },
    { 0xB63, "allies_dialog_col" },
    { 0xB64, "allies_dialog_playing" },
    { 0xB65, "allies_engine_room" },
    { 0xB66, "allies_first_advance" },
    { 0xB67, "allies_gunship_run" },
    { 0xB68, "allies_help_when_player_shoots_balcony_enemies" },
    { 0xB6A, "allies_help_when_player_shoots_second_floor_left" },
    { 0xB6B, "allies_help_when_player_shoots_second_floor_middle_or_right" },
    { 0xB6C, "allies_hesh_final_position" },
    { 0xB6D, "allies_inverted_rappel_movement" },
    { 0xB6E, "allies_inverted_rappel_vo" },
    { 0xB6F, "allies_jeep_sync_anim" },
    { 0xB70, "allies_move_down_to_first_floor_combat" },
    { 0xB71, "allies_move_down_to_second_floor_combat" },
    { 0xB72, "allies_move_down_to_third_floor_combat" },
    { 0xB73, "allies_move_to_checkpoint_start" },
    { 0xB76, "allies_movement_post_missile_launch" },
    { 0xB77, "allies_movement_tower" },
    { 0xB78, "allies_movement_warehouse" },
    { 0xB79, "allies_mudpumps" },
    { 0xB7A, "allies_rappel_anims" },
    { 0xB7B, "allies_rappel_stealth_anims" },
    { 0xB7F, "allies_ride_chopper" },
    { 0xB80, "allies_run_for_garage" },
    { 0xB81, "allies_sat1" },
    { 0xB82, "allies_sat2" },
    { 0xB83, "allies_shadow_kill_movement" },
    { 0xB84, "allies_shadow_kill_vo" },
    { 0xB85, "allies_start_cqb" },
    { 0xB86, "allies_stealth_behavior_end_count" },
    { 0xB87, "allies_tanks" },
    { 0xB88, "allies_throw_smoke" },
    { 0xB89, "allies_to_rappel" },
    { 0xB8A, "allies_vo_post_missile_launch" },
    { 0xB8B, "allies_vo_tower" },
    { 0xB8D, "alliescapturing" },
    { 0xB8E, "allieschopper" },
    { 0xB8F, "alliesinsertchopper" },
    { 0xB90, "alliesteletostartspot" },
    { 0xB91, "allow_dry_fire" },
    { 0xB92, "allow_fall" },
    { 0xB93, "allow_fire" },
    { 0xB94, "allow_glass_break_slide" },
    { 0xB96, "allow_level_killstreak" },
    { 0xB97, "allow_movement" },
    { 0xB98, "allow_pipe_damage" },
    { 0xB99, "allow_player_ascend_move" },
    { 0xB9A, "allow_sprint" },
    { 0xB9B, "allow_walk_up" },
    { 0xB9C, "allowable_double_attachments" },
    { 0xB9D, "allowboard" },
    { 0xB9F, "allowclasschoice" },
    { 0xBA0, "allowcrouch" },
    { 0xBA1, "allowed_cycles" },
    { 0xBA3, "allowedcallouts" },
    { 0xBA4, "allowedinsolo" },
    { 0xBA5, "allowedpartialreloadontheruntime" },
    { 0xBA6, "allowedtofire" },
    { 0xBA7, "allowempdamage" },
    { 0xBA8, "allowenemyspectate" },
    { 0xBA9, "allowfauxdeath" },
    { 0xBAA, "allowfreespectate" },
    { 0xBAB, "allowgrenadedamage" },
    { 0xBAC, "allowlaststandai" },
    { 0xBAD, "allowlatecomers" },
    { 0xBAF, "allowneutral" },
    { 0xBB0, "allowridekillstreakplayerexit" },
    { 0xBB1, "allowsafeeject" },
    { 0xBB2, "allowshoot" },
    { 0xBB3, "allowteamchoice" },
    { 0xBB4, "allowtelefrag" },
    { 0xBB5, "allowuse" },
    { 0xBB6, "allowvehicledamage" },
    { 0xBB7, "allowvote" },
    { 0xBB8, "allowweapons" },
    { 0xBB9, "ally" },
    { 0xBBA, "ally0_inhere" },
    { 0xBBB, "ally0_instruction_vo" },
    { 0xBBC, "ally0_instruction_vo_holdup" },
    { 0xBBD, "ally0_instruction_vo_table" },
    { 0xBBE, "ally0_main" },
    { 0xBBF, "ally0_main_int" },
    { 0xBC1, "ally0_move_to_end" },
    { 0xBC2, "ally0_start_path2" },
    { 0xBC3, "ally0_traversal_logic" },
    { 0xBC4, "ally1_ascend_ascender" },
    { 0xBC5, "ally1_ascend_launcher" },
    { 0xBC8, "ally1_main_int" },
    { 0xBC9, "ally1_mall" },
    { 0xBCA, "ally2_ascend_ascender" },
    { 0xBCB, "ally2_ascend_launcher" },
    { 0xBCC, "ally2_main" },
    { 0xBCD, "ally2_main_int" },
    { 0xBCE, "ally2_mall" },
    { 0xBCF, "ally_0_animation" },
    { 0xBD0, "ally_1_animation" },
    { 0xBD2, "ally_advance_watcher" },
    { 0xBD4, "ally_alley_flood_spawn" },
    { 0xBD5, "ally_animate_vault_scene" },
    { 0xBD6, "ally_breach_gas" },
    { 0xBD7, "ally_breach_goal" },
    { 0xBD8, "ally_breach_vo" },
    { 0xBD9, "ally_calm_idle_internal" },
    { 0xBDB, "ally_catchup" },
    { 0xBDC, "ally_catchup_solo" },
    { 0xBDF, "ally_cleanup" },
    { 0xBE0, "ally_clear_flee_behavior" },
    { 0xBE1, "ally_color_node_movement" },
    { 0xBE2, "ally_console_scene" },
    { 0xBE3, "ally_cqb_kill" },
    { 0xBE4, "ally_cqb_kill_solo" },
    { 0xBE5, "ally_crouch_walk_to_goal" },
    { 0xBE6, "ally_current_state" },
    { 0xBE7, "ally_dialogue" },
    { 0xBE8, "ally_do_reload_anim" },
    { 0xBE9, "ally_dog_attack_free" },
    { 0xBEB, "ally_dog_follow_owner" },
    { 0xBED, "ally_dog_guard_owner" },
    { 0xBEE, "ally_dog_guardpoint_radius" },
    { 0xBEF, "ally_dog_scripted" },
    { 0xBF0, "ally_dog_search_radius" },
    { 0xBF1, "ally_dog_sniff_mode" },
    { 0xBF2, "ally_dog_think" },
    { 0xBF3, "ally_dogs" },
    { 0xBF6, "ally_edge_nodes" },
    { 0xBF8, "ally_elevator_clip_back" },
    { 0xBF9, "ally_ent_del" },
    { 0xBFA, "ally_finale_logic" },
    { 0xBFB, "ally_finale_movement_setup" },
    { 0xBFC, "ally_fire_until_out_of_ammo" },
    { 0xBFD, "ally_fire_until_out_of_ammo_internal" },
    { 0xBFE, "ally_fired" },
    { 0xBFF, "ally_first_frame_check" },
    { 0xC00, "ally_flee_setup" },
    { 0xC01, "ally_garage" },
    { 0xC02, "ally_garage_sneak" },
    { 0xC03, "ally_get_fire_animation" },
    { 0xC04, "ally_get_horizontal_start_distance" },
    { 0xC05, "ally_get_horizontal_stop_distance" },
    { 0xC06, "ally_get_vertical_stop_anim_distance" },
    { 0xC0A, "ally_goggle_glow_on" },
    { 0xC0B, "ally_grenade" },
    { 0xC0C, "ally_groups" },
    { 0xC0E, "ally_gun_hide" },
    { 0xC0F, "ally_gun_show" },
    { 0xC11, "ally_impact_org" },
    { 0xC12, "ally_intro_anims_and_logic" },
    { 0xC13, "ally_invasion_scene_approach" },
    { 0xC14, "ally_is_aiming" },
    { 0xC17, "ally_kill" },
    { 0xC18, "ally_left_death" },
    { 0xC19, "ally_listen_dog_commands" },
    { 0xC1A, "ally_littlebird_1" },
    { 0xC1B, "ally_littlebird_2" },
    { 0xC1D, "ally_main" },
    { 0xC1E, "ally_main_walk" },
    { 0xC1F, "ally_make_fall" },
    { 0xC21, "ally_move_to_jeep" },
    { 0xC22, "ally_movement" },
    { 0xC24, "ally_nagging" },
    { 0xC25, "ally_new_state" },
    { 0xC26, "ally_node_logic" },
    { 0xC27, "ally_on_stairs" },
    { 0xC28, "ally_owner" },
    { 0xC29, "ally_physics_pulse" },
    { 0xC2A, "ally_push" },
    { 0xC2B, "ally_push_player_after_unload" },
    { 0xC2C, "ally_rappel_distance2dsquared_to_player" },
    { 0xC2F, "ally_rappel_get_aim_pitch" },
    { 0xC30, "ally_rappel_get_aim_yaw" },
    { 0xC31, "ally_rappel_get_enemy" },
    { 0xC32, "ally_rappel_get_rope_start" },
    { 0xC34, "ally_rappel_moving_change_direction" },
    { 0xC35, "ally_rappel_moving_stop_idle" },
    { 0xC36, "ally_rappel_moving_vertical_stop" },
    { 0xC37, "ally_rappel_pause_movement_horizontal" },
    { 0xC38, "ally_rappel_reload" },
    { 0xC39, "ally_rappel_rope" },
    { 0xC3B, "ally_rappel_set_perfect_accuracy" },
    { 0xC3C, "ally_rappel_setup_rope" },
    { 0xC3D, "ally_rappel_start_aiming" },
    { 0xC3E, "ally_rappel_start_movement_horizontal" },
    { 0xC3F, "ally_rappel_start_movement_horizontal_internal" },
    { 0xC40, "ally_rappel_start_rope" },
    { 0xC41, "ally_rappel_start_shooting" },
    { 0xC42, "ally_rappel_stealth_movement_vertical" },
    { 0xC43, "ally_rappel_stop_aiming" },
    { 0xC44, "ally_rappel_stop_rope" },
    { 0xC45, "ally_rappel_stop_shooting" },
    { 0xC47, "ally_reset_second_floor_right" },
    { 0xC48, "ally_reset_weights" },
    { 0xC4A, "ally_roof_collapsing_vo" },
    { 0xC4B, "ally_rooftop_water_to_debrisbridge" },
    { 0xC4C, "ally_rubber_banding_solo" },
    { 0xC4D, "ally_set_initial_weights" },
    { 0xC4E, "ally_setup" },
    { 0xC4F, "ally_setup_aim" },
    { 0xC51, "ally_shoot_at_enemy" },
    { 0xC52, "ally_shoot_convoy" },
    { 0xC53, "ally_shoot_nodes" },
    { 0xC54, "ally_shoot_think" },
    { 0xC55, "ally_shooting_loop" },
    { 0xC56, "ally_shoots_fire_extinguisher" },
    { 0xC57, "ally_sprint" },
    { 0xC58, "ally_sprint_end" },
    { 0xC59, "ally_sprint_setup" },
    { 0xC5A, "ally_start" },
    { 0xC5B, "ally_start_1" },
    { 0xC5C, "ally_start_2" },
    { 0xC5D, "ally_start_calm_idle" },
    { 0xC5E, "ally_start_cornerwaving" },
    { 0xC5F, "ally_stealth_kill" },
    { 0xC60, "ally_stealth_settings" },
    { 0xC61, "ally_stop_calm_idle" },
    { 0xC67, "ally_through_sat_panel" },
    { 0xC68, "ally_to_magicbullet" },
    { 0xC69, "ally_transition_to_target_weights" },
    { 0xC6A, "ally_turnanim_hack" },
    { 0xC6B, "ally_vault_props" },
    { 0xC6C, "ally_vo" },
    { 0xC6D, "ally_wait_to_get_impatient" },
    { 0xC6E, "ally_zipline_count" },
    { 0xC6F, "ally_zipline_nag" },
    { 0xC70, "allyagentthink" },
    { 0xC71, "allyc17_right_waits" },
    { 0xC72, "allyheli" },
    { 0xC76, "allytank1" },
    { 0xC7A, "allytime" },
    { 0xC7B, "alpha1" },
    { 0xC7C, "alpha2" },
    { 0xC7E, "alpha_ascend_rubberband_cleanup" },
    { 0xC7F, "alpha_curr_rate" },
    { 0xC80, "alphabetize" },
    { 0xC81, "already_checking_udwfx" },
    { 0xC82, "already_dumpped" },
    { 0xC83, "already_issued" },
    { 0xC84, "already_ran_function" },
    { 0xC85, "already_spawned" },
    { 0xC86, "already_stuck" },
    { 0xC87, "already_thrown" },
    { 0xC89, "alreadyaddedtoalivecount" },
    { 0xC8B, "alreadyrotating" },
    { 0xC8C, "alreadystumbling" },
    { 0xC8D, "alreadytarget" },
    { 0xC8E, "alsodamageparent" },
    { 0xC8F, "alt_override" },
    { 0xC90, "alternates" },
    { 0xC91, "altitude" },
    { 0xC92, "altitude_adjusted" },
    { 0xC93, "altitude_min_override" },
    { 0xC94, "altitude_min_override_remove" },
    { 0xC95, "altitude_override_over_time" },
    { 0xC96, "altitudepolls" },
    { 0xC97, "altwarning" },
    { 0xC98, "always_pain" },
    { 0xC99, "always_play_pain_sound" },
    { 0xC9A, "alwaysdrawfriendlynames" },
    { 0xC9B, "alwaysgamemodeclass" },
    { 0xC9C, "alwayslookatfirsttarget" },
    { 0xC9D, "alwaysquake" },
    { 0xC9E, "alwaysrocketdeath" },
    { 0xC9F, "alwaysrunforward" },
    { 0xCA0, "amb_quakes" },
    { 0xCA1, "ambience_inner" },
    { 0xCA3, "ambient" },
    { 0xCA4, "ambient_a10s" },
    { 0xCA5, "ambient_airburst_periph_cleanup" },
    { 0xCA6, "ambient_airbursts" },
    { 0xCA7, "ambient_airbursts_player_effcts" },
    { 0xCA8, "ambient_airbursts_startpoint" },
    { 0xCAA, "ambient_animate" },
    { 0xCAB, "ambient_breach_tank_a10_pass" },
    { 0xCAC, "ambient_building_elevators" },
    { 0xCAD, "ambient_building_explosions" },
    { 0xCAE, "ambient_building_lights" },
    { 0xCAF, "ambient_building_lights_internal" },
    { 0xCB0, "ambient_canyon_airburst_close_shake" },
    { 0xCB1, "ambient_canyon_airburst_fx_teleport_and_delete" },
    { 0xCB3, "ambient_canyon_airbursts_periph" },
    { 0xCB5, "ambient_choppers" },
    { 0xCB6, "ambient_combat_guys" },
    { 0xCB7, "ambient_derrick_animation" },
    { 0xCB8, "ambient_drones" },
    { 0xCB9, "ambient_event_thread" },
    { 0xCBA, "ambient_ext" },
    { 0xCBB, "ambient_fence_shocks" },
    { 0xCBC, "ambient_guys_anims" },
    { 0xCBD, "ambient_hinds" },
    { 0xCBE, "ambient_int" },
    { 0xCBF, "ambient_jet_by_sound" },
    { 0xCC2, "ambient_nh90_landers" },
    { 0xCC4, "ambient_road_runners" },
    { 0xCC5, "ambient_road_vehicles" },
    { 0xCC6, "ambient_rog_strike" },
    { 0xCC7, "ambient_rog_strike_intro" },
    { 0xCC9, "ambient_rog_strike_timer" },
    { 0xCCA, "ambient_runner_think" },
    { 0xCCC, "ambient_track" },
    { 0xCCD, "ambient_vehicle_spawning" },
    { 0xCCF, "ambush_duration" },
    { 0xCD1, "ambush_enemies" },
    { 0xCD5, "ambush_guy_outcome_logic" },
    { 0xCD6, "ambush_hidden_settings" },
    { 0xCD7, "ambush_init" },
    { 0xCD8, "ambush_jeep2_backr" },
    { 0xCD9, "ambush_jeep2_driver" },
    { 0xCDA, "ambush_jeep2_guy_wave" },
    { 0xCDB, "ambush_jeep2_passenger" },
    { 0xCDD, "ambush_jeep_latch_close_lf" },
    { 0xCDF, "ambush_jeep_latch_close_rr" },
    { 0xCE0, "ambush_jeep_latch_open_lf" },
    { 0xCE1, "ambush_jeep_latch_open_rf" },
    { 0xCE2, "ambush_jeep_latch_open_rr" },
    { 0xCE3, "ambush_jeep_passenger" },
    { 0xCE6, "ambush_kill_ally" },
    { 0xCE7, "ambush_kill_driver" },
    { 0xCE8, "ambush_kill_driver_cypher" },
    { 0xCE9, "ambush_kill_driver_player" },
    { 0xCEA, "ambush_main" },
    { 0xCEB, "ambush_notify_on_player_kill" },
    { 0xCEC, "ambush_notify_on_player_shot" },
    { 0xCEE, "ambush_patrol_guys" },
    { 0xCEF, "ambush_patrol_logic" },
    { 0xCF0, "ambush_patrollers" },
    { 0xCF1, "ambush_player_did_ambush" },
    { 0xCF2, "ambush_player_ran_ahead" },
    { 0xCF3, "ambush_recover_anim" },
    { 0xCF4, "ambush_script" },
    { 0xCF5, "ambush_stealth_settings" },
    { 0xCF6, "ambush_trap_ent" },
    { 0xCF8, "ambush_yaw" },
    { 0xCFB, "ammo" },
    { 0xCFC, "ammo_cache_think_global" },
    { 0xCFD, "ammo_crate_failsafe" },
    { 0xCFE, "ammo_hack" },
    { 0xCFF, "ammo_icon" },
    { 0xD00, "ammo_icon_fade_in" },
    { 0xD01, "ammo_icon_fade_out" },
    { 0xD04, "ammo_pickup" },
    { 0xD05, "ammo_thrower_scene" },
    { 0xD06, "ammocheatinterval" },
    { 0xD07, "ammocheattime" },
    { 0xD08, "ammocount" },
    { 0xD09, "ammocounterhide" },
    { 0xD0A, "ammoomnvar" },
    { 0xD0B, "ammopickup" },
    { 0xD0C, "ammopickupfunc" },
    { 0xD0D, "ammopickupmodel" },
    { 0xD0E, "ammorefillprimary" },
    { 0xD0F, "ammorefillsecondary" },
    { 0xD10, "ammorestockcheckfreq" },
    { 0xD13, "amount" },
    { 0xD14, "amplitude" },
    { 0xD16, "anchor_line_impact" },
    { 0xD17, "angel_flare" },
    { 0xD18, "angel_flare_burst" },
    { 0xD19, "angelflareprecache" },
    { 0xD1A, "angered" },
    { 0xD1B, "angle_180" },
    { 0xD1D, "angle_lerp" },
    { 0xD1E, "angle_offset" },
    { 0xD1F, "anglerangethread" },
    { 0xD20, "angles3d" },
    { 0xD23, "angles_within" },
    { 0xD24, "anglescheck" },
    { 0xD25, "anglesclamp180" },
    { 0xD28, "angry_flood_big_wave_water" },
    { 0xD2B, "angry_flood_collision_cheater" },
    { 0xD2C, "angry_flood_collision_cheater_spawn" },
    { 0xD2D, "angry_flood_collision_dodamage" },
    { 0xD2E, "angry_flood_collision_spawn" },
    { 0xD2F, "angry_flood_finishing_move" },
    { 0xD30, "angry_flood_rumble" },
    { 0xD31, "angry_flood_rumble_loop" },
    { 0xD32, "angry_flood_splash_sequencing_lf" },
    { 0xD34, "angry_flood_street_mist" },
    { 0xD35, "angry_flood_water" },
    { 0xD36, "anim_addmodel" },
    { 0xD37, "anim_animationendnotify" },
    { 0xD38, "anim_array" },
    { 0xD39, "anim_at_entity" },
    { 0xD3A, "anim_at_self" },
    { 0xD3B, "anim_base" },
    { 0xD3C, "anim_blend_time_override" },
    { 0xD3D, "anim_boost" },
    { 0xD3E, "anim_break_entrance" },
    { 0xD3F, "anim_changes_pushplayer" },
    { 0xD41, "anim_custom_animmode" },
    { 0xD42, "anim_custom_animmode_loop" },
    { 0xD43, "anim_custom_animmode_loop_solo" },
    { 0xD44, "anim_custom_animmode_on_guy" },
    { 0xD46, "anim_deathnotify" },
    { 0xD47, "anim_dialogueendnotify" },
    { 0xD48, "anim_disablepain" },
    { 0xD49, "anim_dontpushplayer" },
    { 0xD4A, "anim_down_move_strength" },
    { 0xD4B, "anim_drop_bone" },
    { 0xD4C, "anim_end_early" },
    { 0xD4D, "anim_end_early_animationendnotify" },
    { 0xD4E, "anim_end_early_deathnotify" },
    { 0xD4F, "anim_end_early_dialogueendnotify" },
    { 0xD50, "anim_end_early_facialendnotify" },
    { 0xD53, "anim_enemies_non_combat" },
    { 0xD54, "anim_enemies_non_combat_rate" },
    { 0xD57, "anim_ent_a" },
    { 0xD58, "anim_ent_b" },
    { 0xD59, "anim_ent_c" },
    { 0xD5C, "anim_facialanim" },
    { 0xD5F, "anim_fake_loop_endon" },
    { 0xD60, "anim_first_frame" },
    { 0xD61, "anim_first_frame_on_guy" },
    { 0xD62, "anim_first_frame_solo" },
    { 0xD63, "anim_first_roll_everyone" },
    { 0xD66, "anim_generic_custom_animmode_loop" },
    { 0xD69, "anim_generic_gravity_run" },
    { 0xD6A, "anim_generic_loop" },
    { 0xD6B, "anim_generic_queue" },
    { 0xD6C, "anim_generic_reach" },
    { 0xD6D, "anim_generic_reach_and_animate" },
    { 0xD6E, "anim_generic_reach_and_arrive" },
    { 0xD6F, "anim_generic_run" },
    { 0xD70, "anim_generic_teleport" },
    { 0xD71, "anim_gunhand" },
    { 0xD72, "anim_guninhand" },
    { 0xD73, "anim_handle_notetrack" },
    { 0xD74, "anim_head_faceplate" },
    { 0xD75, "anim_index" },
    { 0xD76, "anim_intro_dog" },
    { 0xD77, "anim_intro_sniper" },
    { 0xD79, "anim_last_frame_solo" },
    { 0xD7A, "anim_length" },
    { 0xD7B, "anim_length_frames_left" },
    { 0xD7C, "anim_link_tag_model" },
    { 0xD7D, "anim_loop" },
    { 0xD7E, "anim_loop_packet" },
    { 0xD80, "anim_loop_solo" },
    { 0xD81, "anim_minigun_hands" },
    { 0xD82, "anim_models" },
    { 0xD83, "anim_moveto" },
    { 0xD84, "anim_node" },
    { 0xD85, "anim_org" },
    { 0xD86, "anim_player_and_allies" },
    { 0xD87, "anim_player_target_enemy" },
    { 0xD89, "anim_precache" },
    { 0xD8A, "anim_prop_init_threads" },
    { 0xD8B, "anim_prop_models" },
    { 0xD8C, "anim_prop_models_animtree" },
    { 0xD8E, "anim_props_animated" },
    { 0xD8F, "anim_pushplayer" },
    { 0xD90, "anim_reach" },
    { 0xD91, "anim_reach_and_approach" },
    { 0xD92, "anim_reach_and_approach_node_solo" },
    { 0xD93, "anim_reach_and_approach_solo" },
    { 0xD94, "anim_reach_and_idle" },
    { 0xD95, "anim_reach_and_idle_solo" },
    { 0xD96, "anim_reach_and_plant" },
    { 0xD98, "anim_reach_failsafe" },
    { 0xD99, "anim_reach_idle" },
    { 0xD9A, "anim_reach_play" },
    { 0xD9B, "anim_reach_solo" },
    { 0xD9D, "anim_reach_together" },
    { 0xD9E, "anim_reach_with_funcs" },
    { 0xD9F, "anim_ref" },
    { 0xDA1, "anim_removemodel" },
    { 0xDA2, "anim_right_move_strength" },
    { 0xDA3, "anim_scene" },
    { 0xDA4, "anim_self_set_time" },
    { 0xDA5, "anim_sequence" },
    { 0xDA6, "anim_set_rate" },
    { 0xDA7, "anim_set_rate_internal" },
    { 0xDA8, "anim_set_rate_single" },
    { 0xDA9, "anim_set_time" },
    { 0xDAA, "anim_single" },
    { 0xDAB, "anim_single_end_early" },
    { 0xDAC, "anim_single_failsafe" },
    { 0xDAD, "anim_single_failsafeonguy" },
    { 0xDAE, "anim_single_internal" },
    { 0xDAF, "anim_single_queue" },
    { 0xDB0, "anim_single_run" },
    { 0xDB1, "anim_single_run_solo" },
    { 0xDB2, "anim_single_solo" },
    { 0xDB4, "anim_spawn_generic_model" },
    { 0xDB6, "anim_spawn_replace_with_model" },
    { 0xDB7, "anim_spawn_tag_model" },
    { 0xDB8, "anim_spawner_teleport" },
    { 0xDB9, "anim_start_at_groundpos" },
    { 0xDBA, "anim_start_pos" },
    { 0xDBB, "anim_start_pos_solo" },
    { 0xDBC, "anim_stopanimscripted" },
    { 0xDBD, "anim_teleport" },
    { 0xDBE, "anim_teleport_solo" },
    { 0xDBF, "anim_time" },
    { 0xDC0, "anim_timing" },
    { 0xDC1, "anim_up_down_boost" },
    { 0xDC2, "anim_wait_func" },
    { 0xDC3, "anim_weight" },
    { 0xDC5, "animarray" },
    { 0xDC6, "animarrayanyexist" },
    { 0xDC7, "animarrayfuncs" },
    { 0xDC8, "animarraypickrandom" },
    { 0xDC9, "animate_after_movement" },
    { 0xDCA, "animate_allies_to_train" },
    { 0xDCB, "animate_ally_breach" },
    { 0xDCC, "animate_ambush_scene_enemies" },
    { 0xDCD, "animate_and_get_shot" },
    { 0xDCE, "animate_arms" },
    { 0xDCF, "animate_chairs" },
    { 0xDD0, "animate_combat_two_intro_debris" },
    { 0xDD1, "animate_drive_idle" },
    { 0xDD3, "animate_ent" },
    { 0xDD4, "animate_front_station_01" },
    { 0xDD5, "animate_front_station_02" },
    { 0xDD6, "animate_front_station_03" },
    { 0xDD7, "animate_front_station_04" },
    { 0xDD8, "animate_front_station_05" },
    { 0xDD9, "animate_front_station_and_return_to_idle" },
    { 0xDDA, "animate_guys" },
    { 0xDDB, "animate_opposite_direction" },
    { 0xDDC, "animate_props_on_death" },
    { 0xDDD, "animate_til_volume" },
    { 0xDDE, "animate_turret_with_viewmodel" },
    { 0xDDF, "animate_vault_door" },
    { 0xDE0, "animate_vault_light" },
    { 0xDE1, "animate_vip_enemies" },
    { 0xDE2, "animated_doors" },
    { 0xDE3, "animated_logs" },
    { 0xDE4, "animated_model" },
    { 0xDE5, "animated_models" },
    { 0xDE6, "animated_prop" },
    { 0xDE7, "animated_prop_anims" },
    { 0xDE9, "animated_sat_part" },
    { 0xDEA, "animated_scene" },
    { 0xDEB, "animated_scene_org" },
    { 0xDEC, "animated_script_model" },
    { 0xDED, "animated_warehouse_guys" },
    { 0xDEF, "animatedduffle" },
    { 0xDF0, "animatemodel" },
    { 0xDF1, "animatemoveintoplace" },
    { 0xDF5, "animcbs" },
    { 0xDF9, "animflagnameindex" },
    { 0xDFA, "animgaz" },
    { 0xDFB, "animhasfacialoverride" },
    { 0xDFC, "animlengths" },
    { 0xDFD, "animname" },
    { 0xDFE, "animnode" },
    { 0xDFF, "animoffset" },
    { 0xE00, "animontag" },
    { 0xE03, "animplaybackrate" },
    { 0xE05, "anims_curr" },
    { 0xE06, "animsapplied" },
    { 0xE07, "animscriptdonotetracksthread" },
    { 0xE09, "animset" },
    { 0xE0A, "animsets" },
    { 0xE0B, "animsound_aliases" },
    { 0xE0C, "animsound_exists" },
    { 0xE0D, "animsound_hud_extralines" },
    { 0xE0E, "animsound_hudlimit" },
    { 0xE0F, "animsound_start_tracker" },
    { 0xE10, "animsound_start_tracker_loop" },
    { 0xE12, "animsound_tracker" },
    { 0xE13, "animsounds" },
    { 0xE14, "animsounds_thisframe" },
    { 0xE15, "animspeeds" },
    { 0xE16, "animspot" },
    { 0xE18, "animsubstate" },
    { 0xE19, "animtree" },
    { 0xE1A, "animtype_idle" },
    { 0xE1B, "animtype_idle_shift" },
    { 0xE1C, "animtype_loop" },
    { 0xE1D, "animtype_loop_run" },
    { 0xE1E, "animtype_parent" },
    { 0xE1F, "animtype_run_stop" },
    { 0xE20, "animtype_shift_back" },
    { 0xE22, "animtype_stop" },
    { 0xE23, "anmimname" },
    { 0xE24, "announce_past_those_doors" },
    { 0xE25, "annoyance_tracker" },
    { 0xE26, "ant_races" },
    { 0xE27, "antenna" },
    { 0xE28, "antenna_kill" },
    { 0xE29, "antenna_rumble" },
    { 0xE2A, "anti_air_objectiv_trigger" },
    { 0xE2B, "antithreat" },
    { 0xE2C, "any_player_near_sentry" },
    { 0xE2D, "any_player_nearby" },
    { 0xE2E, "any_players_istouching" },
    { 0xE2F, "anyone_touching_blocker" },
    { 0xE30, "anyplayersinkillcam" },
    { 0xE31, "anythingtouchingtrigger" },
    { 0xE32, "ao" },
    { 0xE33, "aoeradius" },
    { 0xE34, "aoexponent" },
    { 0xE35, "aogain" },
    { 0xE36, "aomaxdist" },
    { 0xE37, "aoscalemin" },
    { 0xE39, "apache" },
    { 0xE3A, "apache1" },
    { 0xE3B, "apache2" },
    { 0xE3C, "apache_ally_path_attack_filter_targets" },
    { 0xE3D, "apache_ally_path_attack_func" },
    { 0xE3E, "apache_ally_path_attack_internal" },
    { 0xE3F, "apache_autosave_check" },
    { 0xE40, "apache_chase_allies_apache" },
    { 0xE41, "apache_chase_ally_apache_think" },
    { 0xE42, "apache_chase_ally_blackhawk_think" },
    { 0xE43, "apache_chase_enemies" },
    { 0xE44, "apache_chase_enemies_turret_think_delay" },
    { 0xE45, "apache_chatter" },
    { 0xE46, "apache_chatter_func" },
    { 0xE47, "apache_chatter_last" },
    { 0xE48, "apache_chatter_queue" },
    { 0xE49, "apache_chopper_enemies" },
    { 0xE4A, "apache_chopper_hind_on_death" },
    { 0xE4B, "apache_chopper_hind_spawn" },
    { 0xE4D, "apache_difficulty" },
    { 0xE4E, "apache_dmg_recent" },
    { 0xE4F, "apache_dmg_time" },
    { 0xE50, "apache_escort_encounter_final_wave_on_spawn" },
    { 0xE51, "apache_factory_ai_roof_on_spawn" },
    { 0xE52, "apache_factory_allies_apache_think" },
    { 0xE53, "apache_factory_ally_blackhawk_think" },
    { 0xE54, "apache_factory_enemies" },
    { 0xE55, "apache_factory_enemy_counts" },
    { 0xE56, "apache_factory_enemy_on_death" },
    { 0xE57, "apache_factory_hind_parked_on_death" },
    { 0xE58, "apache_factory_hind_parked_on_spawn" },
    { 0xE59, "apache_factory_objective" },
    { 0xE5B, "apache_health" },
    { 0xE5D, "apache_health_at_max" },
    { 0xE5E, "apache_health_get" },
    { 0xE5F, "apache_health_init" },
    { 0xE60, "apache_health_max" },
    { 0xE61, "apache_health_max_get" },
    { 0xE62, "apache_health_max_set" },
    { 0xE63, "apache_health_min" },
    { 0xE64, "apache_health_pct_get" },
    { 0xE65, "apache_health_pct_min_clear" },
    { 0xE66, "apache_health_pct_min_set" },
    { 0xE67, "apache_health_pct_set" },
    { 0xE68, "apache_health_set" },
    { 0xE69, "apache_health_state_next" },
    { 0xE6A, "apache_health_state_think" },
    { 0xE6C, "apache_hints_break_flares" },
    { 0xE6D, "apache_hints_break_mg" },
    { 0xE6E, "apache_hints_break_missile_lockon" },
    { 0xE6F, "apache_hints_break_missile_straight" },
    { 0xE70, "apache_hints_chase" },
    { 0xE72, "apache_hints_display_hint_timeout" },
    { 0xE73, "apache_hints_factory" },
    { 0xE74, "apache_hints_island" },
    { 0xE75, "apache_hints_move" },
    { 0xE78, "apache_hints_tutorial" },
    { 0xE79, "apache_missile_water_z" },
    { 0xE7B, "apache_mission_difficulty" },
    { 0xE7C, "apache_mission_heli_ai_collision" },
    { 0xE7D, "apache_mission_impact_water_missile_think" },
    { 0xE7E, "apache_mission_impact_water_missiles" },
    { 0xE7F, "apache_mission_impact_water_think" },
    { 0xE80, "apache_mission_vo_antiair" },
    { 0xE81, "apache_mission_vo_chopper" },
    { 0xE82, "apache_mission_vo_factory" },
    { 0xE84, "apache_mission_vo_finale" },
    { 0xE85, "apache_mission_vo_player_crashing" },
    { 0xE89, "apache_owner_notify_on_input_camera" },
    { 0xE8A, "apache_owner_notify_on_input_move" },
    { 0xE8B, "apache_player_adjust" },
    { 0xE8C, "apache_player_dead" },
    { 0xE8D, "apache_player_difficulty" },
    { 0xE8E, "apache_precache" },
    { 0xE8F, "apache_savecheck" },
    { 0xE90, "apache_sun_settings" },
    { 0xE92, "apache_target_manager" },
    { 0xE93, "apache_tutorial_fly_allies" },
    { 0xE94, "apache_tutorial_fly_allies_govern_speed" },
    { 0xE95, "apache_tutorial_fly_allies_govern_speed_stop" },
    { 0xE97, "apache_tutorial_fly_player_pitch_think" },
    { 0xE98, "apache_tutorial_fly_reactive_foliage" },
    { 0xE99, "apextraversaldeathvector" },
    { 0xE9B, "apply_difficulty_frac_with_func" },
    { 0xE9D, "apply_end_fog" },
    { 0xE9E, "apply_extra_last_stand" },
    { 0xE9F, "apply_fog" },
    { 0xEA0, "apply_friendly_fire_damage_modifier" },
    { 0xEA3, "apply_synch_attack_damage" },
    { 0xEA4, "apply_truckjunk" },
    { 0xEA5, "apply_weapons_status" },
    { 0xEA6, "apply_whizby" },
    { 0xEA7, "applyaliensnare" },
    { 0xEA8, "applyaliensnareinternal" },
    { 0xEA9, "applybombcarrierclass" },
    { 0xEAB, "applyflash" },
    { 0xEAC, "applyformat" },
    { 0xEAD, "applyformattype" },
    { 0xEAE, "applygaseffect" },
    { 0xEAF, "applyglobalempeffects" },
    { 0xEB0, "applyoutline" },
    { 0xEB1, "applyperplayerempeffects" },
    { 0xEB2, "applyperplayerempeffects_ondetonate" },
    { 0xEB3, "applystunresistence" },
    { 0xEB4, "approach_anims" },
    { 0xEB5, "approach_enemy" },
    { 0xEB6, "approach_types" },
    { 0xEB8, "approachnumber" },
    { 0xEBB, "approachwaittillclose" },
    { 0xEBC, "approx_dist_from_edge" },
    { 0xEBD, "arc_cached" },
    { 0xEBF, "arcademode_hud_timer" },
    { 0xEC0, "arcademode_stop_timer" },
    { 0xEC1, "arcademode_stoptime" },
    { 0xEC2, "archetype_exists" },
    { 0xEC3, "archetypechanged" },
    { 0xEC4, "archetypeexists" },
    { 0xEC5, "archetypes" },
    { 0xEC6, "archive" },
    { 0xEC7, "are_all_players_using_nuke" },
    { 0xEC8, "are_opposite_sign" },
    { 0xEC9, "area1_ents" },
    { 0xECA, "area_damage_and_impulse" },
    { 0xECB, "areaparallelpipid" },
    { 0xECD, "areatriange" },
    { 0xECE, "aredifferent" },
    { 0xECF, "arm_mine" },
    { 0xED0, "arm_player" },
    { 0xED1, "armingdelay" },
    { 0xED3, "armormitigation" },
    { 0xED4, "armorpiercingmod" },
    { 0xED5, "armorvestmod" },
    { 0xED6, "arms_and_legs" },
    { 0xED7, "armtime" },
    { 0xED9, "array_2dadd" },
    { 0xEDA, "array_add" },
    { 0xEDB, "array_call" },
    { 0xEDC, "array_combine" },
    { 0xEDD, "array_combine_keys" },
    { 0xEDE, "array_combine_non_integer_indices" },
    { 0xEE0, "array_compare" },
    { 0xEE1, "array_contains" },
    { 0xEE3, "array_contains_script_linkto" },
    { 0xEE4, "array_delete" },
    { 0xEE5, "array_delete_evenly" },
    { 0xEE6, "array_exclude" },
    { 0xEE7, "array_find" },
    { 0xEE8, "array_index_by_classname" },
    { 0xEE9, "array_index_by_parameters" },
    { 0xEEB, "array_insert" },
    { 0xEEC, "array_is_defined" },
    { 0xEED, "array_is_greater_than" },
    { 0xEEE, "array_keep_key_values" },
    { 0xEEF, "array_keep_values" },
    { 0xEF0, "array_kill" },
    { 0xEF1, "array_levelcall" },
    { 0xEF2, "array_levelthread" },
    { 0xEF3, "array_levelthread_safe" },
    { 0xEF5, "array_merge_links" },
    { 0xEF6, "array_notify" },
    { 0xEF7, "array_of_triggers1" },
    { 0xEF8, "array_of_triggers2" },
    { 0xEF9, "array_of_triggers3" },
    { 0xEFA, "array_randomize" },
    { 0xEFB, "array_remove" },
    { 0xEFC, "array_remove_array" },
    { 0xEFD, "array_remove_dupes" },
    { 0xEFE, "array_remove_duplicates" },
    { 0xF00, "array_remove_nokeys" },
    { 0xF01, "array_remove_perk" },
    { 0xF03, "array_remove_when_dead" },
    { 0xF04, "array_removedead" },
    { 0xF05, "array_removedead_keepkeys" },
    { 0xF07, "array_removedead_zodiac" },
    { 0xF08, "array_removeinvalidmissiles" },
    { 0xF09, "array_removeundefined" },
    { 0xF0A, "array_reverse" },
    { 0xF0B, "array_setgoalvolume" },
    { 0xF0C, "array_shift" },
    { 0xF0D, "array_sort_by_handler" },
    { 0xF0E, "array_sort_by_handler_parameter" },
    { 0xF11, "array_sortbysorter" },
    { 0xF12, "array_spawn" },
    { 0xF13, "array_spawn_allow_fail" },
    { 0xF14, "array_spawn_function" },
    { 0xF15, "array_spawn_function_noteworthy" },
    { 0xF16, "array_spawn_function_targetname" },
    { 0xF18, "array_spawn_targetname" },
    { 0xF19, "array_spawn_targetname_allow_fail" },
    { 0xF1A, "array_spawn_targetname_allow_fail_setthreat_insideaware" },
    { 0xF1D, "array_thread5" },
    { 0xF1F, "array_thread_safe" },
    { 0xF20, "array_thread_targetname" },
    { 0xF21, "array_wait" },
    { 0xF23, "array_waitlogic2" },
    { 0xF24, "arrayinsertion" },
    { 0xF25, "arrays_of_colorcoded_ai" },
    { 0xF26, "arrays_of_colorcoded_nodes" },
    { 0xF27, "arrays_of_colorcoded_spawners" },
    { 0xF28, "arrays_of_colorcoded_volumes" },
    { 0xF29, "arrays_of_colorforced_ai" },
    { 0xF2A, "arrivalanim" },
    { 0xF2B, "arrivalendstance" },
    { 0xF2C, "arrivalnodetype" },
    { 0xF2D, "arrivalpathgoalpos" },
    { 0xF2E, "arrivalstance" },
    { 0xF2F, "arrivalstartdist" },
    { 0xF30, "arrivaltype" },
    { 0xF31, "arrive" },
    { 0xF32, "arrived" },
    { 0xF33, "artemis_fire" },
    { 0xF34, "artemis_fire_think" },
    { 0xF35, "artemis_think" },
    { 0xF37, "artifact_pulse" },
    { 0xF38, "artifacts" },
    { 0xF39, "artifacts_fade" },
    { 0xF3A, "artillery_balcony_stumble" },
    { 0xF3B, "artillery_balcony_stumble_skip" },
    { 0xF3C, "artillery_balcony_stumblers_setup" },
    { 0xF3D, "artillery_disable_player_mg" },
    { 0xF3E, "artillery_disableweapons" },
    { 0xF40, "artillery_hesco_tower_drones" },
    { 0xF41, "artillery_hit_drones" },
    { 0xF43, "artillery_player_slide" },
    { 0xF45, "artillery_smoke_grenade" },
    { 0xF47, "artilleryshellshock" },
    { 0xF48, "ascend_aim_lerp_anims" },
    { 0xF49, "ascend_aim_logic" },
    { 0xF4A, "ascend_aim_logic_cleanup" },
    { 0xF4B, "ascend_anim_node" },
    { 0xF4C, "ascend_anims_rate" },
    { 0xF4D, "ascend_ascend_state" },
    { 0xF4E, "ascend_ascend_state_transition" },
    { 0xF4F, "ascend_ascender" },
    { 0xF50, "ascend_current_rate" },
    { 0xF51, "ascend_dialog" },
    { 0xF53, "ascend_hook" },
    { 0xF55, "ascend_hook_ally2" },
    { 0xF56, "ascend_hook_ally3" },
    { 0xF57, "ascend_hook_ally4" },
    { 0xF58, "ascend_idle_state" },
    { 0xF59, "ascend_idle_state_transition" },
    { 0xF5A, "ascend_ignoreme_loop" },
    { 0xF5B, "ascend_launch_pos" },
    { 0xF5C, "ascend_launcher" },
    { 0xF5F, "ascend_mechanics" },
    { 0xF60, "ascend_pendulum" },
    { 0xF61, "ascend_rope1" },
    { 0xF62, "ascend_rope2" },
    { 0xF63, "ascend_rope3" },
    { 0xF64, "ascend_snow_fx" },
    { 0xF65, "ascend_state" },
    { 0xF66, "ascend_state_transition" },
    { 0xF67, "ascend_stop_state" },
    { 0xF68, "ascend_stop_state_transition" },
    { 0xF69, "ascend_target_rate" },
    { 0xF6A, "ascend_vision_sets" },
    { 0xF6B, "ascend_waiting" },
    { 0xF6C, "ascend_wind_01" },
    { 0xF6D, "ascend_wind_02" },
    { 0xF6E, "ascend_wind_03" },
    { 0xF6F, "ascend_wind_04" },
    { 0xF70, "ascender" },
    { 0xF72, "ash_fall_thread" },
    { 0xF73, "ash_init" },
    { 0xF74, "asign_blackhawk_riders" },
    { 0xF76, "assemble_plane_wing" },
    { 0xF77, "assembly_line" },
    { 0xF78, "assembly_line_animate" },
    { 0xF79, "assembly_line_notetrack_sound" },
    { 0xF7B, "assembly_line_piece" },
    { 0xF7C, "assembly_line_precache" },
    { 0xF7D, "assembly_line_tank_damage_watch" },
    { 0xF7E, "assembly_line_tank_notetracks" },
    { 0xF7F, "assembly_line_tank_part_damage_watch" },
    { 0xF80, "assembly_line_tank_part_damage_watch_end" },
    { 0xF81, "assembly_line_tank_part_explode" },
    { 0xF82, "assembly_line_tank_part_extinguish" },
    { 0xF83, "assembly_line_tank_part_visible" },
    { 0xF84, "assert_existance_of_anim" },
    { 0xF85, "assert_if_anim_not_defined" },
    { 0xF86, "assert_if_identical_origins" },
    { 0xF87, "assetname" },
    { 0xF88, "assign_alien_attributes" },
    { 0xF89, "assign_alpha" },
    { 0xF8A, "assign_animals_tree" },
    { 0xF8B, "assign_animtree" },
    { 0xF8C, "assign_archetypes" },
    { 0xF8D, "assign_bravo" },
    { 0xF8E, "assign_chatter" },
    { 0xF90, "assign_friendly_heros" },
    { 0xF91, "assign_fx_to_trigger" },
    { 0xF92, "assign_generic_human_tree" },
    { 0xF93, "assign_model" },
    { 0xF94, "assign_npcid" },
    { 0xF95, "assign_script_breachgroup_to_ents" },
    { 0xF96, "assign_unique_id" },
    { 0xF97, "assignaward" },
    { 0xF98, "assignawards" },
    { 0xF99, "assignteamspawns" },
    { 0xF9A, "assistedsuicide" },
    { 0xF9C, "assumed_match_length" },
    { 0xF9D, "at_edge" },
    { 0xF9E, "at_edge_sign" },
    { 0xF9F, "at_end_anim_freeze_frame" },
    { 0xFA0, "at_goal_node" },
    { 0xFA1, "at_least_goal" },
    { 0xFA2, "atbrinkofdeath" },
    { 0xFA3, "atconcealmentnode" },
    { 0xFA4, "atrium_checkpoint" },
    { 0xFA5, "atrium_done" },
    { 0xFA6, "atrium_final_stand" },
    { 0xFA8, "atrium_player_outside" },
    { 0xFA9, "atrium_wave2" },
    { 0xFAA, "atrium_wave3" },
    { 0xFAB, "atrium_wave4" },
    { 0xFAC, "attach_audio_points_to_player" },
    { 0xFAD, "attach_cig" },
    { 0xFAE, "attach_cig_self" },
    { 0xFB0, "attach_deck_gun" },
    { 0xFB1, "attach_dust_to_mover" },
    { 0xFB2, "attach_flashlight" },
    { 0xFB3, "attach_fx_anim_model" },
    { 0xFB5, "attach_fx_anim_model_mall_debris" },
    { 0xFB6, "attach_fx_anim_model_street_flood" },
    { 0xFB7, "attach_in_volume" },
    { 0xFB8, "attach_model_override" },
    { 0xFBA, "attach_phone" },
    { 0xFBC, "attach_pistol_right" },
    { 0xFBE, "attached_actor" },
    { 0xFBF, "attached_item" },
    { 0xFC0, "attachedguys" },
    { 0xFC1, "attachedmodels" },
    { 0xFC2, "attachedpath" },
    { 0xFC3, "attachedusemodel" },
    { 0xFC4, "attachgrenademodel" },
    { 0xFC5, "attachhat" },
    { 0xFC6, "attachhead" },
    { 0xFC7, "attachlid" },
    { 0xFC9, "attachmentcheck" },
    { 0xFCA, "attachmentgroup" },
    { 0xFCB, "attachmentmap_attachtoperk" },
    { 0xFCC, "attachmentmap_basetounique" },
    { 0xFCD, "attachmentmap_tobase" },
    { 0xFCE, "attachmentmap_tounique" },
    { 0xFCF, "attachmentmap_uniquetobase" },
    { 0xFD1, "attachments" },
    { 0xFD2, "attachmentscompatible" },
    { 0xFD3, "attachmissiles" },
    { 0xFD4, "attachplayer" },
    { 0xFD5, "attachplayertochopper" },
    { 0xFD6, "attachprops" },
    { 0xFD8, "attachturret" },
    { 0xFDA, "attack_chopper_monitoruse" },
    { 0xFDB, "attack_chopper_pot" },
    { 0xFDC, "attack_chopper_reward_pool" },
    { 0xFDD, "attack_damage_trigger" },
    { 0xFDE, "attack_dir" },
    { 0xFE0, "attack_heli_cleanup" },
    { 0xFE1, "attack_heli_fx" },
    { 0xFE2, "attack_heli_safe_volumes" },
    { 0xFE3, "attack_if_provoked" },
    { 0xFE4, "attack_indicator_off" },
    { 0xFE5, "attack_indicator_on" },
    { 0xFE6, "attack_missile_set_up_and_notify" },
    { 0xFE8, "attack_origin_condition_threadd" },
    { 0xFE9, "attack_player_after_death" },
    { 0xFEA, "attack_pos" },
    { 0xFEB, "attack_sequence_num" },
    { 0xFEC, "attack_sound_setup" },
    { 0xFED, "attack_targets" },
    { 0xFEE, "attack_times" },
    { 0xFEF, "attackable" },
    { 0xFF0, "attackable_ent" },
    { 0xFF2, "attackback" },
    { 0xFF3, "attacked" },
    { 0xFF5, "attackenemyaircraft" },
    { 0xFF6, "attacker" },
    { 0xFF7, "attacker_damage" },
    { 0xFF8, "attacker_isonmyteam" },
    { 0xFFA, "attackercandamageitem" },
    { 0xFFB, "attackerdata" },
    { 0xFFC, "attackerent" },
    { 0xFFD, "attackerinlaststand" },
    { 0xFFE, "attackerinremotekillstreak" },
    { 0xFFF, "attackerishittingteam" },
    { 0x1000, "attackernum" },
    { 0x1002, "attackerposition" },
    { 0x1003, "attackerride" },
    { 0x1004, "attackers" },
    { 0x1005, "attackerstance" },
    { 0x1006, "attackertable" },
    { 0x1008, "attackheightpos" },
    { 0x100A, "attackheliexcluders" },
    { 0x100B, "attackhelifov" },
    { 0x100C, "attackheligraceperiod" },
    { 0x100D, "attackhelikillsai" },
    { 0x100E, "attackhelimemory" },
    { 0x100F, "attackhelimovetime" },
    { 0x1010, "attackheliplayerbreak" },
    { 0x1011, "attackhelirange" },
    { 0x1012, "attackhelitargetreaquire" },
    { 0x1013, "attackhelitimeout" },
    { 0x1014, "attacking_player" },
    { 0x1017, "attackmisstracktargetthread" },
    { 0x1018, "attackmovetime" },
    { 0x1019, "attacknothingtodo" },
    { 0x101A, "attackoffset" },
    { 0x101B, "attackradiussq" },
    { 0x101C, "attacks" },
    { 0x101D, "attackstate" },
    { 0x101E, "attacksuppressableenemy" },
    { 0x101F, "attackteleportthread" },
    { 0x1022, "attackzheight" },
    { 0x1023, "attempt_bad_path_melee" },
    { 0x1024, "attempt_bad_path_move_nearby_node" },
    { 0x1025, "attempt_badpath_jump" },
    { 0x1026, "attempt_badpath_move_to_node" },
    { 0x1029, "attract_guys_to_dog" },
    { 0x102A, "attract_range" },
    { 0x102B, "attract_strength" },
    { 0x102C, "attractor" },
    { 0x102D, "attractor2" },
    { 0x102E, "attractor_flare" },
    { 0x102F, "attributes" },
    { 0x1031, "atv_death_launchslide" },
    { 0x1032, "atv_decide_shoot" },
    { 0x1033, "atv_decide_shoot_internal" },
    { 0x1034, "atv_do_event" },
    { 0x1035, "atv_get_death_anim" },
    { 0x1036, "atv_getoff" },
    { 0x1037, "atv_geton" },
    { 0x1038, "atv_handle_events" },
    { 0x1039, "atv_loop_driver" },
    { 0x103A, "atv_loop_driver_shooting" },
    { 0x103B, "atv_normal_death" },
    { 0x103C, "atv_reload" },
    { 0x103D, "atv_reload_internal" },
    { 0x103E, "atv_setanim_common" },
    { 0x1041, "atv_start_shooting" },
    { 0x1042, "atv_stop_shooting" },
    { 0x1044, "atv_waitfor_end" },
    { 0x1045, "atv_waitfor_start_aim" },
    { 0x1046, "atv_waitfor_start_lean" },
    { 0x1048, "aud_30mm_tail" },
    { 0x1049, "aud_add_progress_map" },
    { 0x104A, "aud_ajax_chair" },
    { 0x104B, "aud_ajax_flare" },
    { 0x104C, "aud_alarm" },
    { 0x104D, "aud_ally_gear_rustle" },
    { 0x104E, "aud_ally_gear_rustle_2" },
    { 0x104F, "aud_aux_explosions" },
    { 0x1050, "aud_bar" },
    { 0x1052, "aud_birds" },
    { 0x1053, "aud_blow_vehicle" },
    { 0x1054, "aud_blow_vehicle_low" },
    { 0x1056, "aud_bullet_count" },
    { 0x1057, "aud_bumpy_ride" },
    { 0x1059, "aud_bust_windshield" },
    { 0x105D, "aud_can_play_outside_wind_gusts" },
    { 0x105E, "aud_can_play_rappel_footsteps" },
    { 0x105F, "aud_can_play_rope_creak" },
    { 0x1060, "aud_can_play_tilt_screams" },
    { 0x1061, "aud_car_creak" },
    { 0x1063, "aud_carr_bg_rog_01" },
    { 0x1064, "aud_carr_bg_rog_02" },
    { 0x1065, "aud_carr_bg_rog_03" },
    { 0x1066, "aud_carr_dead_sparrow_ops" },
    { 0x1068, "aud_carr_elevator_exp" },
    { 0x1069, "aud_carr_elevator_front" },
    { 0x106A, "aud_carr_elevator_rear" },
    { 0x106B, "aud_carr_exfil_bg_heli" },
    { 0x106C, "aud_carr_exfil_heli" },
    { 0x106D, "aud_carr_exfil_rog" },
    { 0x106E, "aud_carr_exfil_rog_incoming" },
    { 0x106F, "aud_carr_exp_heli_blade" },
    { 0x1070, "aud_carr_exp_heli_bounce" },
    { 0x1072, "aud_carr_exp_heli_whoosh" },
    { 0x1073, "aud_carr_ghost_mask_on_plr" },
    { 0x1074, "aud_carr_gunship_attack_run" },
    { 0x1075, "aud_carr_gunship_killed" },
    { 0x1076, "aud_carr_hesh_talk_explode" },
    { 0x1079, "aud_carr_osprey_loader" },
    { 0x107A, "aud_carr_osprey_zone_off" },
    { 0x107B, "aud_carr_osprey_zone_on" },
    { 0x107C, "aud_carr_pharmacy_shut" },
    { 0x107D, "aud_carr_pickup_osprey_control" },
    { 0x107E, "aud_carr_player_cuts_rope" },
    { 0x107F, "aud_carr_promotion_dog" },
    { 0x1080, "aud_carr_promotion_hsh" },
    { 0x1081, "aud_carr_promotion_mrk" },
    { 0x1082, "aud_carr_promotion_plr" },
    { 0x1083, "aud_carr_slowmo_bg" },
    { 0x1084, "aud_carr_slowmo_in" },
    { 0x1085, "aud_carr_slowmo_out" },
    { 0x1086, "aud_carr_slowmo_roll" },
    { 0x1087, "aud_carr_slowmo_slide" },
    { 0x1088, "aud_carr_sparrow_105_hit" },
    { 0x1089, "aud_carr_sparrow_105_incoming" },
    { 0x108A, "aud_carr_sparrow_run_hit" },
    { 0x108B, "aud_carr_sparrow_zone_off" },
    { 0x108C, "aud_carr_sparrow_zone_on" },
    { 0x108D, "aud_carr_tilt_plr_death" },
    { 0x108E, "aud_carr_tilt_plr_vault" },
    { 0x108F, "aud_carr_victory_deck_checkpoint" },
    { 0x1090, "aud_carr_zodiac_deck_explode" },
    { 0x1091, "aud_carr_zodiac_deck_explode_vista" },
    { 0x1092, "aud_cart_crash" },
    { 0x1093, "aud_charge_set" },
    { 0x1094, "aud_check" },
    { 0x1096, "aud_chopper_second" },
    { 0x1097, "aud_clear_zone_medbay" },
    { 0x1098, "aud_collapse" },
    { 0x109A, "aud_convoy_done" },
    { 0x109B, "aud_convoy_start" },
    { 0x109C, "aud_coughing" },
    { 0x109D, "aud_deck_jet_catapult_01" },
    { 0x109E, "aud_deck_jet_catapult_02" },
    { 0x109F, "aud_deck_siren" },
    { 0x10A1, "aud_defend_zodiac_osprey_zone" },
    { 0x10A2, "aud_defend_zodiac_zone" },
    { 0x10A3, "aud_disable_deathsdoor_audio" },
    { 0x10A4, "aud_dish_crash" },
    { 0x10A5, "aud_distant_alarm" },
    { 0x10A7, "aud_dog_scratch" },
    { 0x10A8, "aud_dog_thread" },
    { 0x10AB, "aud_doppler_grenade" },
    { 0x10AC, "aud_drillholenumber" },
    { 0x10AD, "aud_dry_fire" },
    { 0x10AE, "aud_enable_deathsdoor_audio" },
    { 0x10AF, "aud_end_random_chatter_and_pa" },
    { 0x10B0, "aud_end_sniper" },
    { 0x10B3, "aud_enemy_muffled_vo" },
    { 0x10B4, "aud_engine_fail" },
    { 0x10B6, "aud_ext_bombs" },
    { 0x10B7, "aud_fast_jets" },
    { 0x10B8, "aud_filter_off" },
    { 0x10B9, "aud_filter_on" },
    { 0x10BA, "aud_finale_ally_locs" },
    { 0x10BB, "aud_finale_chopper1" },
    { 0x10BC, "aud_finale_chopper2" },
    { 0x10BE, "aud_finale_chopper4" },
    { 0x10BF, "aud_finale_pa_guys" },
    { 0x10C0, "aud_finale_sniper" },
    { 0x10C1, "aud_finale_sniper2" },
    { 0x10C2, "aud_flap_loop" },
    { 0x10C3, "aud_flare_grab" },
    { 0x10C4, "aud_flare_kill" },
    { 0x10C5, "aud_flight_deck_bell" },
    { 0x10C6, "aud_focus_zoom" },
    { 0x10C7, "aud_fx_planes" },
    { 0x10C8, "aud_gas_mask_on" },
    { 0x10C9, "aud_gear_sounds" },
    { 0x10CC, "aud_gunship_loc" },
    { 0x10CE, "aud_gunship_trans_4_105_01" },
    { 0x10CF, "aud_gunship_trans_4_105_02" },
    { 0x10D1, "aud_handle_remote_sniper_ai" },
    { 0x10D7, "aud_hvt_rescue_thread" },
    { 0x10D9, "aud_in_sparrow" },
    { 0x10DA, "aud_init_animation_sounds" },
    { 0x10DB, "aud_init_flags" },
    { 0x10DC, "aud_init_globals" },
    { 0x10DF, "aud_intro_alarms" },
    { 0x10E0, "aud_intro_cargo_doors" },
    { 0x10E1, "aud_intro_choppers" },
    { 0x10E2, "aud_intro_convoy_counter" },
    { 0x10E3, "aud_intro_keegan_tinkering" },
    { 0x10E4, "aud_intro_seq_lr" },
    { 0x10E6, "aud_invert" },
    { 0x10E7, "aud_inverted_kill_finish" },
    { 0x10E8, "aud_inverted_kill_firstguy" },
    { 0x10E9, "aud_jeep_flip" },
    { 0x10EA, "aud_jet_attack" },
    { 0x10EB, "aud_junction" },
    { 0x10EC, "aud_keegan_gunfire" },
    { 0x10ED, "aud_last_time" },
    { 0x10EF, "aud_lerp_eq_over_time" },
    { 0x10F1, "aud_listen_mk32_reload" },
    { 0x10F2, "aud_listen_sprint_or_switch" },
    { 0x10F5, "aud_lynx_rider_death_listener" },
    { 0x10F6, "aud_lynx_turrets" },
    { 0x10F8, "aud_map_range" },
    { 0x10F9, "aud_max_bullets" },
    { 0x10FA, "aud_medbay_alarms" },
    { 0x10FB, "aud_medbay_pa" },
    { 0x10FD, "aud_mk32_dud_beep_atrium" },
    { 0x1100, "aud_music" },
    { 0x1103, "aud_ocean02_line_emitter_create" },
    { 0x1104, "aud_old_height" },
    { 0x1105, "aud_old_rotation" },
    { 0x1106, "aud_osprey_controller_off" },
    { 0x1107, "aud_osprey_controller_on" },
    { 0x1108, "aud_osprey_fire" },
    { 0x110A, "aud_osprey_run" },
    { 0x110C, "aud_outside_crowd_rear" },
    { 0x110D, "aud_outside_music" },
    { 0x110E, "aud_outside_music_rear" },
    { 0x110F, "aud_pa_guys" },
    { 0x1110, "aud_party" },
    { 0x1111, "aud_pickup_mk32" },
    { 0x1113, "aud_play_and_move_sound" },
    { 0x1115, "aud_play_deck_music" },
    { 0x1118, "aud_play_intro_music" },
    { 0x1119, "aud_play_jet_flyby" },
    { 0x111A, "aud_play_jets_zoomby" },
    { 0x111B, "aud_play_loop_until_flag" },
    { 0x111C, "aud_play_medbay_music" },
    { 0x111E, "aud_play_random_wind_gust" },
    { 0x111F, "aud_play_tilt_music" },
    { 0x1121, "aud_play_whizby" },
    { 0x1123, "aud_player_gunfire" },
    { 0x1124, "aud_player_jumps_from_sparrow" },
    { 0x1125, "aud_player_reloading" },
    { 0x1127, "aud_player_tank_int_on" },
    { 0x1128, "aud_post_sparrow_music" },
    { 0x112A, "aud_pre_sniper_rpg_explode" },
    { 0x112B, "aud_pre_sniper_rpg_gunner_listener" },
    { 0x112C, "aud_pre_sniper_rpg_listener" },
    { 0x1130, "aud_random_enemy_chatter" },
    { 0x1131, "aud_random_enemy_pa" },
    { 0x1132, "aud_random_metal_crumbles" },
    { 0x1134, "aud_random_rumbles_intro" },
    { 0x1135, "aud_rappel" },
    { 0x1136, "aud_rappel_combat" },
    { 0x1137, "aud_rappel_jump_down" },
    { 0x1139, "aud_rpg_gunner_listener" },
    { 0x113D, "aud_screen_shake_jumps" },
    { 0x113E, "aud_set_spec_ops" },
    { 0x1143, "aud_sniper_fire" },
    { 0x1144, "aud_sniper_start_zoom" },
    { 0x1145, "aud_sniper_stop_zoom" },
    { 0x1146, "aud_sparrow_aiming" },
    { 0x1148, "aud_sparrow_run_spawn_fires" },
    { 0x1149, "aud_sparrow_tone" },
    { 0x114A, "aud_stadium_crumble1" },
    { 0x114C, "aud_start_garden_events" },
    { 0x114D, "aud_start_pseudo_occlusion" },
    { 0x114E, "aud_start_sniper" },
    { 0x114F, "aud_start_sniper_finale" },
    { 0x1151, "aud_stop_wind" },
    { 0x1152, "aud_switch_zone_medbay" },
    { 0x1153, "aud_temp_timer" },
    { 0x1154, "aud_tilt_barrels_01" },
    { 0x1155, "aud_tilt_debris_01" },
    { 0x1156, "aud_tilt_front_deck" },
    { 0x1158, "aud_tilt_sliding_cart_01" },
    { 0x1159, "aud_tilt_sliding_guya" },
    { 0x115B, "aud_tower_collapse" },
    { 0x115C, "aud_tower_to_deck" },
    { 0x115D, "aud_traverse_ally_locs" },
    { 0x115E, "aud_truck_drive" },
    { 0x115F, "aud_truck_enter" },
    { 0x1160, "aud_truck_ext_idle_loop" },
    { 0x1162, "aud_truck_start" },
    { 0x1163, "aud_turret_indices" },
    { 0x1165, "aud_updating_movement" },
    { 0x1166, "aud_vehicle_jolt" },
    { 0x1168, "aud_victory_deck_spawn_fires" },
    { 0x1169, "aud_vip_ally_locs" },
    { 0x116A, "aud_vip_breach" },
    { 0x116B, "aud_vip_combat" },
    { 0x116C, "aud_virus" },
    { 0x116D, "aud_wave2_ambient_jets" },
    { 0x116E, "aud_wave3_ambient_jets" },
    { 0x116F, "aud_weapon_strobe" },
    { 0x1170, "aud_wind_loop" },
    { 0x1171, "aud_wind_state_last" },
    { 0x1172, "aud_zipline" },
    { 0x1173, "aud_zipline_launcher_loop" },
    { 0x1175, "aud_zodiac_gunship_attack_105_fake" },
    { 0x1178, "aud_zodiac_jet_catapult_02" },
    { 0x117D, "audio" },
    { 0x117E, "audio_bink_percentage_beep_array" },
    { 0x117F, "audio_check_to_play_a_beep_or_not" },
    { 0x1180, "audio_derrick_explode_logic" },
    { 0x1181, "audio_entities" },
    { 0x1182, "audio_flag_init" },
    { 0x1185, "audio_jet_counter" },
    { 0x1186, "audio_odin_pressurized_variable" },
    { 0x1187, "audio_plane_engine_sounds_dying" },
    { 0x1189, "audio_player_falling_start" },
    { 0x118B, "audio_set_default_ambience" },
    { 0x118C, "audio_set_fadein_ambience" },
    { 0x118D, "audio_set_infil_ambience" },
    { 0x118E, "audio_set_initial_ambience" },
    { 0x118F, "audio_speed" },
    { 0x1190, "audio_start_destruction_loop" },
    { 0x1191, "audio_start_plane_engine_sounds" },
    { 0x1193, "audio_start_rushing_water_line_emitter_01" },
    { 0x1194, "audio_start_rushing_water_line_emitter_02" },
    { 0x1195, "audio_start_rushing_water_line_emitter_03" },
    { 0x1196, "audio_stereo_line_emitter" },
    { 0x1198, "audio_stringtable_mapname" },
    { 0x1199, "audio_underwater_breath_bubbles" },
    { 0x119A, "audio_underwater_breath_surfacing" },
    { 0x119B, "audio_underwater_choke" },
    { 0x119C, "audio_wait_to_delete_water_node" },
    { 0x119D, "audio_water_level_logic" },
    { 0x119E, "audio_zones" },
    { 0x11A0, "aurora_anims" },
    { 0x11A2, "auto_adjust_difficulty_player_movement_check" },
    { 0x11A3, "auto_adjust_difficulty_player_positioner" },
    { 0x11A4, "auto_adjust_difficulty_track_player_death" },
    { 0x11A7, "auto_adjust_enemy_died" },
    { 0x11A8, "auto_adjust_flags" },
    { 0x11AA, "auto_adjust_results" },
    { 0x11AB, "auto_adust_zone_complete" },
    { 0x11AC, "auto_breach" },
    { 0x11AD, "auto_breach_gametypes" },
    { 0x11AE, "auto_flash_text" },
    { 0x11AF, "auto_kill_enemies" },
    { 0x11B0, "auto_mg42_target" },
    { 0x11B2, "auto_pilot_rot_control_ref_ent" },
    { 0x11B4, "autoassign" },
    { 0x11B7, "automatic_sliding_door_detector" },
    { 0x11B9, "autoresettime" },
    { 0x11BA, "autosave_by_name" },
    { 0x11BB, "autosave_by_name_silent" },
    { 0x11BC, "autosave_by_name_thread" },
    { 0x11BD, "autosave_check_override" },
    { 0x11BE, "autosave_now" },
    { 0x11C0, "autosave_now_trigger" },
    { 0x11C1, "autosave_or_timeout" },
    { 0x11C2, "autosave_past_balcony" },
    { 0x11C4, "autosave_recon" },
    { 0x11C6, "autosave_stealth_silent" },
    { 0x11C7, "autosave_tactical" },
    { 0x11C8, "autosave_tactical_grenade_check" },
    { 0x11C9, "autosave_tactical_grenade_check_dieout" },
    { 0x11CA, "autosave_tactical_player_nades" },
    { 0x11CB, "autosave_tactical_proc" },
    { 0x11CD, "autosave_threat_check_enabled" },
    { 0x11CE, "autosave_timeout" },
    { 0x11CF, "autosaveammocheck" },
    { 0x11D0, "autosavecheck" },
    { 0x11D1, "autosavecheck_not_picky" },
    { 0x11D2, "autosavehealthcheck" },
    { 0x11D5, "autosaveprint" },
    { 0x11D6, "autosaves_think" },
    { 0x11D7, "autosavethreatcheck" },
    { 0x11D8, "autoshootanimrate" },
    { 0x11DA, "autospotadswatcher" },
    { 0x11DB, "autospotdeathwatcher" },
    { 0x11DC, "autotarget" },
    { 0x11DE, "available" },
    { 0x11DF, "availabledrones" },
    { 0x11E1, "avengeddog" },
    { 0x11E2, "avengedplayer" },
    { 0x11E3, "avoidairstrikelocations" },
    { 0x11E4, "avoidcarepackages" },
    { 0x11E5, "avoidcornervisibleenemies" },
    { 0x11E7, "avoidenemyspawn" },
    { 0x11E9, "avoidgrenades" },
    { 0x11EA, "avoidkillstreakonspawntimer" },
    { 0x11EB, "avoidlastattackerlocation" },
    { 0x11EC, "avoidlastdeathlocation" },
    { 0x11ED, "avoidmines" },
    { 0x11EE, "avoidrecentlyused" },
    { 0x11EF, "avoidsamespawn" },
    { 0x11F1, "awaittill_either" },
    { 0x11F2, "award1_ref" },
    { 0x11F3, "award2_ref" },
    { 0x11F4, "award_intel" },
    { 0x11F6, "awards" },
    { 0x11F7, "awardxp" },
    { 0x11F8, "aware_aievents" },
    { 0x11F9, "awareness" },
    { 0x11FA, "awareness_param" },
    { 0x11FB, "axes" },
    { 0x11FD, "axischopper" },
    { 0x11FF, "axismode" },
    { 0x1200, "b" },
    { 0x1201, "b_bad_guys" },
    { 0x1202, "b_falling" },
    { 0x1204, "back_alley_humvee" },
    { 0x1205, "back_jerk" },
    { 0x1206, "back_light_on" },
    { 0x1207, "back_velocity" },
    { 0x1208, "back_weight" },
    { 0x1209, "back_wheel_sfx" },
    { 0x120A, "backdoor_guy" },
    { 0x120B, "backend_friendly_stealth_logic" },
    { 0x120D, "background_block" },
    { 0x1210, "background_vo" },
    { 0x1211, "backplate_sound_on" },
    { 0x1212, "backstab" },
    { 0x1213, "backtank" },
    { 0x1214, "backtodefendlocation" },
    { 0x1215, "backtorunondamage" },
    { 0x1216, "backupbuddyplayfx" },
    { 0x1218, "baclonyguys" },
    { 0x121A, "bad_path_handled" },
    { 0x121C, "badger" },
    { 0x121E, "badpath_jump" },
    { 0x121F, "badpathcount" },
    { 0x1220, "badpathtime" },
    { 0x1221, "badplace" },
    { 0x1223, "badplace_brush_moving" },
    { 0x1224, "badplace_cylinder_func" },
    { 0x1226, "badplace_name" },
    { 0x1228, "badplace_think" },
    { 0x1229, "badplaceint" },
    { 0x122C, "badplacer" },
    { 0x122D, "badplaces" },
    { 0x122E, "badshot" },
    { 0x122F, "badshotcount" },
    { 0x1230, "badtarget" },
    { 0x1231, "badtargetreset" },
    { 0x1232, "bag_name" },
    { 0x1233, "bag_trigger_off_if_player_has_weapon" },
    { 0x1234, "bag_trigger_wait" },
    { 0x1235, "bag_vis_callback" },
    { 0x1236, "bags" },
    { 0x1237, "baker" },
    { 0x1238, "baker_anim" },
    { 0x1239, "baker_approach" },
    { 0x123A, "baker_building_entry_movement" },
    { 0x123B, "baker_crawl_dialogue" },
    { 0x123E, "baker_drop_bag" },
    { 0x123F, "baker_end" },
    { 0x1240, "baker_enemy" },
    { 0x1241, "baker_enter" },
    { 0x1242, "baker_enter_struct" },
    { 0x1243, "baker_enter_wreck" },
    { 0x1245, "baker_glint_off" },
    { 0x1246, "baker_glint_on" },
    { 0x1247, "baker_in_to_jeep_anim" },
    { 0x1248, "baker_junction_door_open" },
    { 0x1249, "baker_killfirms" },
    { 0x124A, "baker_move_to_stealth_2" },
    { 0x124B, "baker_noncombat" },
    { 0x124C, "baker_path_to_wreck" },
    { 0x124E, "baker_post_up_at_sharks" },
    { 0x1250, "baker_rappel_hookup" },
    { 0x1251, "baker_sonar_path" },
    { 0x1252, "baker_sonar_path_dialogue" },
    { 0x1254, "baker_struct" },
    { 0x1255, "baker_torpedo_position" },
    { 0x1256, "baker_wait_at_container" },
    { 0x1258, "baker_weld_door" },
    { 0x1259, "baker_wreck_cleanup" },
    { 0x125A, "baker_wreck_dialogue" },
    { 0x125B, "balanceteams" },
    { 0x125C, "balcony" },
    { 0x125D, "balcony_allies_playerhind_logic" },
    { 0x125E, "balcony_anims" },
    { 0x125F, "balcony_check" },
    { 0x1261, "balcony_death" },
    { 0x1262, "balcony_death_damage_watcher" },
    { 0x1264, "balcony_enemies_clip" },
    { 0x1265, "balcony_fall_deaths" },
    { 0x1267, "balcony_kill_trigger" },
    { 0x1268, "balcony_read" },
    { 0x1269, "balcony_turret" },
    { 0x126A, "balconyent" },
    { 0x126B, "balconystumblers" },
    { 0x126D, "ball_angle" },
    { 0x126E, "ball_i" },
    { 0x1270, "ball_offset" },
    { 0x1271, "ball_ring" },
    { 0x1272, "balldrone" },
    { 0x1277, "balldrone_burstfirestop" },
    { 0x1278, "balldrone_enemy_lightfx" },
    { 0x1279, "balldrone_followplayer" },
    { 0x127A, "balldrone_friendly_lightfx" },
    { 0x127B, "balldrone_handledamage" },
    { 0x127C, "balldrone_leave" },
    { 0x127E, "balldrone_moving_platform_death" },
    { 0x127F, "balldrone_stunned" },
    { 0x1280, "balldrone_watchdeath" },
    { 0x1281, "balldrone_watchforgoal" },
    { 0x1282, "balldrone_watchownerdeath" },
    { 0x1283, "balldrone_watchownerloss" },
    { 0x1284, "balldrone_watchroundend" },
    { 0x1285, "balldrone_watchtimeout" },
    { 0x1286, "balldronedestroyed" },
    { 0x1287, "balldroneexplode" },
    { 0x1288, "balldrones" },
    { 0x1289, "balldronesettings" },
    { 0x128A, "balldronetype" },
    { 0x128B, "balloon_count" },
    { 0x128C, "balloons" },
    { 0x128D, "ballytankmove" },
    { 0x128E, "bang_stick_rotate" },
    { 0x128F, "bar" },
    { 0x1290, "bar_animnode" },
    { 0x1291, "bar_death" },
    { 0x1292, "bar_drift_rate" },
    { 0x1293, "bar_enemies" },
    { 0x1296, "bar_enemy_lights_out_accuracy" },
    { 0x1297, "bar_enemy_panic_vo" },
    { 0x1298, "bar_enemy_reach" },
    { 0x1299, "bar_enemy_react" },
    { 0x129A, "bar_enemy_seek_player" },
    { 0x129B, "bar_enemy_setup" },
    { 0x129C, "bar_enemy_strobe_react" },
    { 0x129D, "bar_enemy_strobe_vo" },
    { 0x129E, "bar_enemy_vo" },
    { 0x129F, "bar_enemy_wave2_3_react" },
    { 0x12A0, "bar_enemy_wave2_behavior" },
    { 0x12A2, "bar_guy11" },
    { 0x12A3, "bar_guy9" },
    { 0x12A4, "bar_guy_watch_death" },
    { 0x12A5, "bar_light" },
    { 0x12A6, "bar_prep" },
    { 0x12A7, "bar_props" },
    { 0x12A8, "bar_react_variable_wait" },
    { 0x12A9, "bar_rorke" },
    { 0x12AA, "bar_rorke_move_on" },
    { 0x12AB, "bar_rorke_shoot_tv" },
    { 0x12AC, "bar_rorke_strobe_attack" },
    { 0x12AD, "bar_rorke_warning_vo" },
    { 0x12AE, "bar_spotted_func" },
    { 0x12AF, "bar_stool_anim" },
    { 0x12B1, "bar_strobe_player_force_off" },
    { 0x12B2, "bar_strobe_player_on" },
    { 0x12B3, "barbed_wire_waver" },
    { 0x12B4, "barbedwirerunners" },
    { 0x12B5, "bared_wire_scene" },
    { 0x12B6, "bark_trigger" },
    { 0x12B7, "barking_sound" },
    { 0x12B8, "barkingdog_handle_stealth_break" },
    { 0x12B9, "barrel" },
    { 0x12BA, "barrel_earthquake" },
    { 0x12BB, "barrel_model_1" },
    { 0x12BC, "barrel_model_2" },
    { 0x12BD, "barrel_model_3" },
    { 0x12BF, "barrel_model_5" },
    { 0x12C0, "barrel_roll" },
    { 0x12C4, "base" },
    { 0x12C5, "base_accuracy" },
    { 0x12C6, "base_alarm" },
    { 0x12C7, "base_alarm_loop" },
    { 0x12C9, "base_array_ai_cleanup_spawn_function" },
    { 0x12CC, "base_array_ambient_a10_gun_dive_1b" },
    { 0x12CD, "base_array_ambient_a10_gun_dive_2" },
    { 0x12CE, "base_array_ambient_a10_gun_dive_2b" },
    { 0x12D0, "base_array_ambient_a10_gun_dive_3b" },
    { 0x12D1, "base_array_ambient_dogfight_1" },
    { 0x12D2, "base_array_ambient_dogfight_1b" },
    { 0x12D3, "base_array_ambient_dogfight_2" },
    { 0x12D4, "base_array_ambient_dogfight_2b" },
    { 0x12D5, "base_array_ambient_dogfight_3" },
    { 0x12D6, "base_array_ambient_dogfight_3b" },
    { 0x12D7, "base_array_ambient_dogfight_4b" },
    { 0x12D8, "base_array_ambient_dogfight_5b" },
    { 0x12D9, "base_array_ambient_dogfight_6b" },
    { 0x12DA, "base_array_ambient_dogfight_6c" },
    { 0x12DB, "base_array_begin" },
    { 0x12DC, "base_array_choppers" },
    { 0x12DD, "base_array_cleanup" },
    { 0x12DE, "base_array_end_vo" },
    { 0x12DF, "base_array_enemies_setup" },
    { 0x12E0, "base_array_exit_rpg" },
    { 0x12E2, "base_array_init" },
    { 0x12E3, "base_array_main" },
    { 0x12E4, "base_array_mortar_strikes" },
    { 0x12E5, "base_array_pinned_down_allies" },
    { 0x12E7, "base_array_trucks_01_setup" },
    { 0x12E9, "base_array_vo" },
    { 0x12EA, "base_falling_hands_anim" },
    { 0x12EB, "base_falling_legs_anim" },
    { 0x12EC, "base_origin" },
    { 0x12ED, "baseaccuracy" },
    { 0x12EE, "basealpha" },
    { 0x12EF, "baseangles" },
    { 0x12F1, "baseeffectforward" },
    { 0x12F2, "baseeffectpos" },
    { 0x12F3, "baseeffectright" },
    { 0x12F4, "baseeffectswaitforjoined" },
    { 0x12F5, "basefontscale" },
    { 0x12F6, "baseheight" },
    { 0x12F7, "baseignorerandombulletdamage" },
    { 0x12F8, "baseline_speed" },
    { 0x12F9, "basement_ally_movement" },
    { 0x12FB, "basename" },
    { 0x12FD, "basetile" },
    { 0x12FE, "basetime" },
    { 0x12FF, "basewidth" },
    { 0x1300, "baseyaw" },
    { 0x1302, "batman_begins" },
    { 0x1303, "batman_rotate_plane" },
    { 0x1305, "battingcage_door_peek" },
    { 0x1306, "battle_chatter_controller_friendlies" },
    { 0x1307, "battlebuddy" },
    { 0x1308, "battlebuddyrespawntimestamp" },
    { 0x1309, "battlebuddywaitlist" },
    { 0x130A, "battlechatter" },
    { 0x130B, "battlechatter_canprint" },
    { 0x130D, "battlechatter_debugprint" },
    { 0x130F, "battlechatter_off" },
    { 0x1310, "battlechatter_on" },
    { 0x1311, "battlechatter_on_thread" },
    { 0x1312, "battlechatter_print" },
    { 0x1314, "battlechatter_printdumpline" },
    { 0x1315, "battlechatter_setup" },
    { 0x1317, "bay_door_lower_model" },
    { 0x1319, "bay_door_upper_model" },
    { 0x131A, "bbdata_init" },
    { 0x131C, "bc_ascend" },
    { 0x1320, "bc_enabled" },
    { 0x1322, "bc_end_fic" },
    { 0x1324, "bc_eventtypelastusedtimeplr" },
    { 0x1326, "bc_front_fic" },
    { 0x1327, "bc_helo_fic" },
    { 0x1328, "bc_helo_reinforce" },
    { 0x1329, "bc_helo_reinforce_kill" },
    { 0x132B, "bc_isspeaking" },
    { 0x132C, "bc_mid" },
    { 0x132E, "bc_rolling_door_open" },
    { 0x132F, "bc_snow_tweaks" },
    { 0x1330, "bc_street" },
    { 0x1332, "bc_street_fic" },
    { 0x1333, "bccountryid" },
    { 0x1335, "bcgetclaimednode" },
    { 0x1336, "bcinfo" },
    { 0x1338, "bclearstrafeturnrate" },
    { 0x1339, "bcname" },
    { 0x133B, "bcprintfailprefix" },
    { 0x133D, "bcqb_pa_playing" },
    { 0x133E, "bcrank" },
    { 0x133F, "bcrashmix" },
    { 0x1340, "bcs_location_mappings" },
    { 0x1341, "bcs_location_trigger_mapping" },
    { 0x1342, "bcs_location_trigs_init" },
    { 0x1343, "bcs_locations" },
    { 0x1344, "bcs_maxtalkingdistsqrdfromplayer" },
    { 0x1346, "bcs_minpriority" },
    { 0x1347, "bcs_on" },
    { 0x1348, "bcs_scripted_dialogue_start" },
    { 0x1349, "bcs_setup_chatter_toggle_array" },
    { 0x134B, "bcs_setup_flavorburst_toggle_array" },
    { 0x134C, "bcs_setup_teams_array" },
    { 0x134D, "bcs_setup_voice" },
    { 0x134E, "bcs_threatresettime" },
    { 0x1350, "bcsdebugwaiter" },
    { 0x1352, "bcsounds" },
    { 0x1353, "bdamagesoundplaying" },
    { 0x1354, "bdcheck" },
    { 0x1356, "bdefenddoorexplosionplaying" },
    { 0x1357, "bdisabledefaultfacialanims" },
    { 0x1358, "bdisablegearsounds" },
    { 0x1359, "bdisablemovetwitch" },
    { 0x135A, "bdoorbreakfoleyplayed" },
    { 0x135B, "bdoturnandmove" },
    { 0x135C, "bdrillon" },
    { 0x135D, "bdriverkilled" },
    { 0x135E, "beach_a10_return_flybys" },
    { 0x135F, "beach_ai" },
    { 0x1360, "beach_ally_default" },
    { 0x1362, "beach_artillery_balcony_logic" },
    { 0x1363, "beach_balcony_collapse_watcher" },
    { 0x1364, "beach_battlehind_default" },
    { 0x1365, "beach_battlehind_rpgers" },
    { 0x1367, "beach_battlehinds_start" },
    { 0x1368, "beach_bunker_backtrack_blocker" },
    { 0x1369, "beach_bunker_drones" },
    { 0x136A, "beach_bunker_stumble_event" },
    { 0x136B, "beach_bunker_stumbler_sound" },
    { 0x136C, "beach_dof_changes" },
    { 0x136D, "beach_enemy_attack_player_manager" },
    { 0x136E, "beach_enemy_default" },
    { 0x1370, "beach_flyover_helis" },
    { 0x1371, "beach_front_nodes_think" },
    { 0x1372, "beach_frontline_abrams" },
    { 0x1373, "beach_frontline_abrams_mg" },
    { 0x1374, "beach_hind_balcony_logic" },
    { 0x1375, "beach_hind_balcony_missile_logic" },
    { 0x1376, "beach_hovercraft_looper" },
    { 0x1377, "beach_hovercraft_tanks_default" },
    { 0x1379, "beach_m880_death_hack" },
    { 0x137B, "beach_nh90_damagestate" },
    { 0x137D, "beach_path_drones" },
    { 0x137E, "beach_playerhind_attack_logic" },
    { 0x137F, "beach_playerhind_attack_target" },
    { 0x1380, "beach_playerhind_attack_target_player" },
    { 0x1381, "beach_reveal_turn_off_mix_snapshot" },
    { 0x1382, "beach_runners" },
    { 0x1383, "beach_sequence_bunker_new" },
    { 0x1384, "beach_sequence_trenches" },
    { 0x1385, "beach_ship_ambient_artillery" },
    { 0x1386, "beach_ship_phalanx_start" },
    { 0x1387, "beach_ship_phalanx_system" },
    { 0x1389, "beach_ship_phalanx_think" },
    { 0x138B, "beach_tank_balcony_logic" },
    { 0x138C, "beach_tank_balcony_stumble" },
    { 0x138D, "beach_tank_balcony_stumbler_notetrack" },
    { 0x138E, "beach_tower_runners_off" },
    { 0x138F, "beach_tower_runners_on" },
    { 0x1390, "beach_trenches_combat" },
    { 0x1391, "beach_trenches_combat_part2" },
    { 0x1392, "beach_trenches_dialogue" },
    { 0x1393, "beach_trenches_part2_dialogue" },
    { 0x1394, "beach_vehicle_default" },
    { 0x1395, "beach_wave1_ai" },
    { 0x1397, "beach_wave1_artillery_drones" },
    { 0x1398, "beach_wave1_artillery_retreat" },
    { 0x1399, "beach_wave1_dialog" },
    { 0x139B, "beach_wave1_enemy_drones" },
    { 0x139C, "beach_wave1_hind_flybys" },
    { 0x139D, "beach_wave1_logic" },
    { 0x139E, "beach_wave2_dialogue" },
    { 0x139F, "beach_wave2_inithinds" },
    { 0x13A0, "beach_wave2_logic" },
    { 0x13A1, "beach_wave2_playerhind_deathfunc" },
    { 0x13A2, "beach_wave2_playerhind_evade" },
    { 0x13A5, "beach_wave2_playerhind_path_logic" },
    { 0x13A6, "beach_wave2_playerhind_pathing" },
    { 0x13A8, "beach_wave2_vehicle_watcher" },
    { 0x13A9, "beach_wave3_dialogue" },
    { 0x13AA, "beach_wave3_logic" },
    { 0x13AB, "beach_wave3_tank_dropoff" },
    { 0x13AC, "beach_wave3_tank_setup" },
    { 0x13AD, "beachfronelinedrones" },
    { 0x13AF, "beachhinds" },
    { 0x13B0, "beachlander" },
    { 0x13B1, "beachtanks" },
    { 0x13B4, "beforestairanim" },
    { 0x13B6, "begin_anim_reach" },
    { 0x13B7, "begin_atrium" },
    { 0x13BB, "begin_building_entry" },
    { 0x13BE, "begin_checkpoint" },
    { 0x13C0, "begin_combat" },
    { 0x13C2, "begin_courtyard" },
    { 0x13C3, "begin_deck_combat" },
    { 0x13C5, "begin_deck_transition" },
    { 0x13C6, "begin_deck_victory" },
    { 0x13C8, "begin_defend_blowdoors1" },
    { 0x13C9, "begin_defend_blowdoors2" },
    { 0x13CB, "begin_defend_plat" },
    { 0x13CD, "begin_defend_zodiac" },
    { 0x13CE, "begin_drive_in" },
    { 0x13D2, "begin_exfil_tank" },
    { 0x13D3, "begin_finale" },
    { 0x13D4, "begin_garden" },
    { 0x13D6, "begin_hvt_test" },
    { 0x13D9, "begin_interior_cqb" },
    { 0x13DA, "begin_interior_vault_scene" },
    { 0x13DB, "begin_intro" },
    { 0x13DC, "begin_inverted_rappel" },
    { 0x13E1, "begin_rail_vo" },
    { 0x13E2, "begin_rappel" },
    { 0x13E3, "begin_rappel_stealth" },
    { 0x13E4, "begin_rooftop_intro" },
    { 0x13E5, "begin_rooftop_shoot" },
    { 0x13E6, "begin_run_to_sparrow" },
    { 0x13E7, "begin_semtex_grenade_tracking" },
    { 0x13E8, "begin_shadow_kill" },
    { 0x13E9, "begin_slow_intro" },
    { 0x13EA, "begin_stairwell" },
    { 0x13ED, "begin_traverse" },
    { 0x13EF, "begin_zipline" },
    { 0x13F0, "beginairdropmarkertracking" },
    { 0x13F1, "beginairdropviamarker" },
    { 0x13F3, "beginclasschoice" },
    { 0x13F4, "begincustomevent" },
    { 0x13F6, "begingrenadetracking" },
    { 0x13F7, "beginharrier" },
    { 0x13F8, "beginningoflevelsave" },
    { 0x13FA, "beginsmokegrenadetracking" },
    { 0x13FB, "beginteamchoice" },
    { 0x13FC, "behavior" },
    { 0x13FD, "being_charged" },
    { 0x13FF, "beingartilleryshellshocked" },
    { 0x1400, "beingdestroyed" },
    { 0x1401, "beingrevived" },
    { 0x1402, "belowcumulativepainthreshold" },
    { 0x1404, "bestminitarget" },
    { 0x1405, "bestspawnflag" },
    { 0x1407, "better" },
    { 0x1408, "betty_tutorial_given" },
    { 0x140A, "bfirstmoveanim" },
    { 0x140B, "bg" },
    { 0x140C, "bg_loop_sound" },
    { 0x140D, "bg_rog_hit" },
    { 0x140E, "bg_rog_impact" },
    { 0x140F, "bg_viewbobmax" },
    { 0x1410, "bg_weaponbobamplitudebase" },
    { 0x1411, "bg_zodiac_respawn" },
    { 0x1414, "bhasgunwhileriding" },
    { 0x1415, "bhasnopath" },
    { 0x1416, "bidlehitreaction" },
    { 0x1417, "bidlelooking" },
    { 0x1418, "big_dish_fall" },
    { 0x1419, "big_message" },
    { 0x141A, "big_wave_2" },
    { 0x141B, "big_wave_addl_effects" },
    { 0x141C, "big_wreck_2_dialogue" },
    { 0x141D, "big_wreck_2_setup" },
    { 0x141E, "big_wreck_baker_stealth" },
    { 0x1420, "big_wreck_dialogue" },
    { 0x1421, "big_wreck_encounter" },
    { 0x1423, "big_wreck_kill_when_outside" },
    { 0x1424, "big_wreck_setup" },
    { 0x1425, "big_wreck_shark" },
    { 0x1426, "big_wreck_shark_baker_teleport" },
    { 0x1427, "big_wreck_tilt" },
    { 0x1428, "big_wreck_track_player_gunfire" },
    { 0x142A, "bigjump" },
    { 0x142B, "bigjump_player_blend_to_anim" },
    { 0x142C, "bigjump_timedelta" },
    { 0x142D, "bigmissile1" },
    { 0x142E, "bike" },
    { 0x1430, "bike_avoids_obstacles" },
    { 0x1431, "bike_death_score" },
    { 0x1432, "bike_drives_path" },
    { 0x1433, "bike_ent_wipe_out_check" },
    { 0x1434, "bike_randomly_changes_lanes" },
    { 0x1436, "bike_turns" },
    { 0x1437, "bin_failsafe" },
    { 0x1438, "binding" },
    { 0x143A, "bink_is_paused" },
    { 0x143B, "bink_percentage" },
    { 0x143C, "bink_start_time" },
    { 0x143D, "binoc_target" },
    { 0x143E, "binocular_body_features_left" },
    { 0x1440, "binocular_face_scanning" },
    { 0x1441, "binocular_face_scanning_data" },
    { 0x1443, "binocular_face_scanning_lines_complete" },
    { 0x1446, "binocular_profile_materials" },
    { 0x1448, "binocular_reticle_target_reaction" },
    { 0x1449, "binocular_status_update" },
    { 0x144A, "binocular_target" },
    { 0x144B, "binocular_zoom_levels" },
    { 0x144E, "binoculars_angles_display" },
    { 0x144F, "binoculars_calculate_range" },
    { 0x1450, "binoculars_clear_hud" },
    { 0x1452, "binoculars_default_zoom_level" },
    { 0x1454, "binoculars_hide_deactive_hint" },
    { 0x1455, "binoculars_hide_hint" },
    { 0x1456, "binoculars_hud" },
    { 0x1457, "binoculars_hud_item" },
    { 0x1458, "binoculars_init" },
    { 0x145A, "binoculars_linked_to_target" },
    { 0x145B, "binoculars_lock_to_target" },
    { 0x145C, "binoculars_monitor_scanning" },
    { 0x145D, "binoculars_monitor_scanning_button" },
    { 0x1461, "binoculars_pip_update_position" },
    { 0x1462, "binoculars_remove_target_on_death" },
    { 0x1463, "binoculars_reticle_lerp_to_tag" },
    { 0x1464, "binoculars_scan_for_targets" },
    { 0x1465, "binoculars_scan_target" },
    { 0x1466, "binoculars_scan_target_points" },
    { 0x1468, "binoculars_set_default_zoom_level" },
    { 0x1469, "binoculars_set_vision_set" },
    { 0x146B, "binoculars_trace" },
    { 0x146D, "binoculars_vision_set" },
    { 0x146E, "binoculars_zoom_display" },
    { 0x146F, "binoculars_zooming" },
    { 0x1471, "bird_fly" },
    { 0x1473, "bird_fx" },
    { 0x1475, "bird_model" },
    { 0x1476, "bird_sit" },
    { 0x1477, "bird_startle" },
    { 0x1478, "bird_waitfordamage" },
    { 0x1479, "birdexists" },
    { 0x147A, "birdmodel_anims" },
    { 0x147B, "birds" },
    { 0x147C, "birds_createents" },
    { 0x147F, "birds_finishbirdtypesetup" },
    { 0x1480, "birds_fly" },
    { 0x1482, "birds_get_last_takeoff" },
    { 0x1483, "birds_isperchsafe" },
    { 0x1484, "birds_loadfromstruct" },
    { 0x1485, "birds_on_baddy" },
    { 0x1486, "birds_path_move_first_point" },
    { 0x1487, "birds_perchdangertrigger" },
    { 0x1489, "birds_perchsetuppath" },
    { 0x148B, "birds_savetostruct" },
    { 0x148C, "birds_savetostructandwaitfortriggerstart" },
    { 0x148E, "birds_setup" },
    { 0x148F, "birds_setupconnectedperches" },
    { 0x1492, "birth_time" },
    { 0x1493, "bisgunner" },
    { 0x1494, "bishop" },
    { 0x1497, "bishop_in_crosshairs" },
    { 0x1498, "bishop_loop_carry_pose" },
    { 0x149A, "bishop_speaks" },
    { 0x149B, "bishop_stool" },
    { 0x149D, "bkillplayer" },
    { 0x149F, "black_fade" },
    { 0x14A0, "black_ice_geyser2_pulse" },
    { 0x14A1, "black_ice_geyser_pulse" },
    { 0x14A2, "black_ice_hide_hud" },
    { 0x14A3, "black_ice_hud" },
    { 0x14A4, "black_ice_hud_actionslotshide" },
    { 0x14A6, "black_ice_hud_compass" },
    { 0x14A7, "black_ice_hud_showstance" },
    { 0x14A8, "black_ice_show_previous_hud" },
    { 0x14AA, "black_screen_vo" },
    { 0x14AB, "blackbox_alienkilled" },
    { 0x14AC, "blackbox_endgame" },
    { 0x14AD, "blackbox_endgame_score" },
    { 0x14AE, "blackbox_laststand" },
    { 0x14AF, "blackhawk_ally" },
    { 0x14B0, "blackhawk_countermeasure" },
    { 0x14B2, "blackhawk_idle_next_to_factory" },
    { 0x14B3, "blackhawk_into_position" },
    { 0x14B4, "blackhawk_landing" },
    { 0x14B5, "blackhawk_path_to_end" },
    { 0x14B6, "blackhawk_unloads_and_takes_off" },
    { 0x14B7, "blackice_exfil_music" },
    { 0x14B8, "blackice_exfil_stinger_music" },
    { 0x14B9, "blackice_ice_chunks_truck" },
    { 0x14BA, "blackice_locations" },
    { 0x14BB, "blackice_pre_ascend_music" },
    { 0x14BC, "blackout" },
    { 0x14BD, "blackout_50" },
    { 0x14BE, "blackout_enemy1and2_react_anims" },
    { 0x14BF, "blackout_enemy3" },
    { 0x14C0, "blackout_enemy45" },
    { 0x14C1, "blackout_loop_anims" },
    { 0x14C2, "blackout_moment_anims" },
    { 0x14C4, "blackout_no_blur" },
    { 0x14C6, "blackscreen" },
    { 0x14C7, "blackscreen_intro" },
    { 0x14C8, "blank" },
    { 0x14C9, "blast_shield" },
    { 0x14CB, "blastshieldclamp" },
    { 0x14CC, "blastshieldmod" },
    { 0x14CD, "blastshieldusetracker" },
    { 0x14CE, "blend_droppitch" },
    { 0x14CF, "blend_link_over_time" },
    { 0x14D0, "blend_movespeedscale" },
    { 0x14D1, "blend_movespeedscale_custom" },
    { 0x14D2, "blend_movespeedscale_default" },
    { 0x14D3, "blend_movespeedscale_percent" },
    { 0x14D4, "blend_player_position" },
    { 0x14D5, "blend_player_to_arms" },
    { 0x14D6, "blend_struct" },
    { 0x14D9, "blend_to_exploder" },
    { 0x14DA, "blend_wind_setting_internal" },
    { 0x14DB, "blend_zones" },
    { 0x14DE, "blended_link" },
    { 0x14DF, "blendintocrouchrun" },
    { 0x14E0, "blendintocrouchwalk" },
    { 0x14E1, "blendintostandrun" },
    { 0x14E2, "blendintostandwalk" },
    { 0x14E3, "blendtreeanims" },
    { 0x14E4, "blindfire" },
    { 0x14E5, "blinkinglightfx" },
    { 0x14E8, "blitz_allies_trigger_origin" },
    { 0x14E9, "blitz_axis_trigger_origin" },
    { 0x14EA, "blitzgetteam" },
    { 0x14EB, "block_ally_cornerwaving" },
    { 0x14EC, "block_ally_sneak_to_node" },
    { 0x14EE, "block_garage_exit" },
    { 0x14EF, "block_off_road_during_convoy" },
    { 0x14F0, "block_until_at_struct" },
    { 0x14F1, "block_until_fully_stopped_and_idle" },
    { 0x14F3, "blockarea" },
    { 0x14F4, "blockentsinarea" },
    { 0x14F5, "blocker_hive_active" },
    { 0x14F6, "blocker_hive_burn" },
    { 0x14F7, "blocker_hive_chopper_hp_bar" },
    { 0x14F8, "blocker_hive_cycle" },
    { 0x14F9, "blocker_hive_hp_bar" },
    { 0x14FA, "blocker_hive_pain_monitor" },
    { 0x14FD, "blockgoalpos" },
    { 0x14FE, "blocking" },
    { 0x14FF, "blockingpain" },
    { 0x1500, "blockplayeruav" },
    { 0x1501, "blockweapondrops" },
    { 0x1502, "blood_splat_on_screen" },
    { 0x1505, "bloodrushregenhealthmod" },
    { 0x1506, "bloodrushregenspeedmod" },
    { 0x1507, "bloodsplateffect" },
    { 0x1508, "bloody_death" },
    { 0x150C, "blowdoors" },
    { 0x150D, "blowit_beep" },
    { 0x150E, "blowout_goalradius_on_pathend" },
    { 0x150F, "blue" },
    { 0x1510, "blue_flare" },
    { 0x1511, "blur" },
    { 0x1512, "blur_death" },
    { 0x1513, "blur_pulse" },
    { 0x1514, "blur_sine" },
    { 0x1515, "blurview" },
    { 0x1516, "bmantisexplosionplaying" },
    { 0x1517, "bmcd_debug_loop" },
    { 0x1518, "bmovingstraight" },
    { 0x1519, "bnoanimunload" },
    { 0x151A, "bnpcpostsoundplaying" },
    { 0x151B, "bo_enemy3" },
    { 0x151D, "bo_enemy5" },
    { 0x1520, "boat2_in_sounds" },
    { 0x1521, "boat_crashing_waves" },
    { 0x1522, "boat_fall_trigs" },
    { 0x1523, "boat_mginit" },
    { 0x1524, "boat_populate" },
    { 0x1527, "boat_teleport" },
    { 0x1529, "boat_vo" },
    { 0x152A, "boats" },
    { 0x152B, "boats_struct" },
    { 0x152C, "bob_axis" },
    { 0x152D, "bob_mask" },
    { 0x152F, "bob_value" },
    { 0x1530, "bobbing_actor" },
    { 0x1532, "bobbing_jitter_cleanup" },
    { 0x1533, "bobbing_object" },
    { 0x1535, "bobbing_ripple" },
    { 0x1536, "bobbing_updown" },
    { 0x1537, "bobcat" },
    { 0x1538, "bodies" },
    { 0x153A, "body_ext" },
    { 0x153C, "body_is_falling" },
    { 0x153D, "bodyarmorhp" },
    { 0x153E, "bodyindex" },
    { 0x1540, "bog_style_mortar" },
    { 0x1542, "bog_style_mortar_cleanup" },
    { 0x1543, "bog_style_mortar_cooldown" },
    { 0x1544, "bog_style_mortar_explode" },
    { 0x1545, "bog_style_mortar_off" },
    { 0x1546, "bog_style_mortar_on" },
    { 0x1547, "bog_style_mortar_think" },
    { 0x154A, "bokeh_ent" },
    { 0x154B, "bokehdots_audition_test" },
    { 0x154C, "bold_dog_jeep" },
    { 0x154D, "bolthit" },
    { 0x154E, "bomb" },
    { 0x154F, "bomb_squad" },
    { 0x1552, "bombdefusetrig" },
    { 0x1554, "bombexploded" },
    { 0x1556, "bombowner" },
    { 0x1557, "bombplanted" },
    { 0x1558, "bombplantedtime" },
    { 0x155A, "bombsquadicon" },
    { 0x155B, "bombsquadicons" },
    { 0x155D, "bombsquadmodel" },
    { 0x155E, "bombsquadvisibilityupdater" },
    { 0x155F, "bombsquadwaiter_missilefire" },
    { 0x1560, "bombstrike" },
    { 0x1561, "bombtimer" },
    { 0x1563, "bombwatcher" },
    { 0x1564, "bombzones" },
    { 0x1565, "bone" },
    { 0x1567, "boneyard_style_heli_missile_attack" },
    { 0x156A, "bool" },
    { 0x156B, "border" },
    { 0x156C, "border_thickness" },
    { 0x156E, "borescope" },
    { 0x1571, "bot" },
    { 0x1572, "bot_3d_sighting_model" },
    { 0x1573, "bot_3d_sighting_model_thread" },
    { 0x1574, "bot_abort_tactical_goal" },
    { 0x1575, "bot_add_ambush_time_delayed" },
    { 0x1577, "bot_add_to_bot_damage_targets" },
    { 0x1578, "bot_add_to_bot_level_targets" },
    { 0x1579, "bot_add_to_bot_use_targets" },
    { 0x157A, "bot_allowed_to_use_killstreaks" },
    { 0x157B, "bot_ambush_end" },
    { 0x157D, "bot_attachment_reticle" },
    { 0x157E, "bot_attachmenttable" },
    { 0x157F, "bot_balance_personality" },
    { 0x1580, "bot_body_is_dead" },
    { 0x1581, "bot_bots_enabled_or_added" },
    { 0x1582, "bot_cache_entrances" },
    { 0x1583, "bot_cache_entrances_to_bombzones" },
    { 0x1584, "bot_cache_entrances_to_flags_or_radios" },
    { 0x1585, "bot_can_revive" },
    { 0x1586, "bot_can_use_aa_launcher" },
    { 0x1587, "bot_can_use_air_superiority" },
    { 0x1588, "bot_can_use_ball_drone" },
    { 0x1589, "bot_can_use_box_by_type" },
    { 0x158B, "bot_can_use_point_in_defend" },
    { 0x158D, "bot_capture_zone" },
    { 0x158E, "bot_capture_zone_get_furthest_distance" },
    { 0x158F, "bot_check_team_is_using_position" },
    { 0x1590, "bot_choose_difficulty_for_default" },
    { 0x1591, "bot_chosen_difficulty" },
    { 0x1592, "bot_class" },
    { 0x1593, "bot_client_counts" },
    { 0x1594, "bot_connect_monitor" },
    { 0x1595, "bot_control_heli" },
    { 0x1596, "bot_control_heli_main_move_loop" },
    { 0x1597, "bot_control_heli_pilot" },
    { 0x1598, "bot_control_heli_sniper" },
    { 0x1599, "bot_control_odin" },
    { 0x159D, "bot_control_vanguard" },
    { 0x159E, "bot_crate_is_command_goal" },
    { 0x159F, "bot_crate_valid" },
    { 0x15A0, "bot_damage_callback" },
    { 0x15A3, "bot_defend_get_random_entrance_point_for_current_area" },
    { 0x15A5, "bot_defend_stop" },
    { 0x15A6, "bot_defend_think" },
    { 0x15A7, "bot_defending" },
    { 0x15A8, "bot_defending_center" },
    { 0x15AB, "bot_defending_radius" },
    { 0x15AC, "bot_defending_trigger" },
    { 0x15AD, "bot_defending_type" },
    { 0x15B0, "bot_disable_tactical_goals" },
    { 0x15B1, "bot_draw_circle" },
    { 0x15B2, "bot_draw_cylinder" },
    { 0x15B3, "bot_draw_cylinder_think" },
    { 0x15B4, "bot_enable_tactical_goals" },
    { 0x15B5, "bot_end_control_on_respawn" },
    { 0x15B6, "bot_end_control_on_vehicle_death" },
    { 0x15B7, "bot_end_control_watcher" },
    { 0x15B8, "bot_end_odin_watcher" },
    { 0x15B9, "bot_ent_is_anonymous_mine" },
    { 0x15BA, "bot_filter_ambush_inuse" },
    { 0x15BB, "bot_filter_ambush_vicinity" },
    { 0x15BC, "bot_find_ambush_entrances" },
    { 0x15BD, "bot_find_defend_node_func" },
    { 0x15BE, "bot_find_node_that_protects_point" },
    { 0x15BF, "bot_find_node_to_capture_point" },
    { 0x15C1, "bot_find_node_to_guard_player" },
    { 0x15C2, "bot_find_random_midpoint" },
    { 0x15C3, "bot_fireteam_buddy_search" },
    { 0x15C4, "bot_fireteam_buddy_think" },
    { 0x15C5, "bot_fireteam_buddy_up" },
    { 0x15C6, "bot_fireteam_cac_getperk" },
    { 0x15C7, "bot_fireteam_cac_getprimarygrenade" },
    { 0x15C8, "bot_fireteam_cac_getsecondarygrenade" },
    { 0x15C9, "bot_fireteam_cac_getstreak" },
    { 0x15CA, "bot_fireteam_cac_getweapon" },
    { 0x15CC, "bot_fireteam_cac_getweaponattachmenttwo" },
    { 0x15CD, "bot_fireteam_cac_getweaponbuff" },
    { 0x15CE, "bot_fireteam_cac_getweaponcamo" },
    { 0x15CF, "bot_fireteam_cac_getweaponreticle" },
    { 0x15D1, "bot_fireteam_follower" },
    { 0x15D4, "bot_fireteam_loadout_class_callback" },
    { 0x15D5, "bot_fireteam_monitor_killstreak_earned" },
    { 0x15D6, "bot_fireteam_setup_callback_class" },
    { 0x15D7, "bot_fireteam_setup_callbacks" },
    { 0x15D9, "bot_free_to_move" },
    { 0x15DA, "bot_funcs" },
    { 0x15DB, "bot_gametype_chooses_class" },
    { 0x15DC, "bot_gametype_chooses_team" },
    { 0x15DF, "bot_get_client_limit" },
    { 0x15E0, "bot_get_entrances_for_stance_and_index" },
    { 0x15E1, "bot_get_grenade_ammo" },
    { 0x15E2, "bot_get_grenade_for_purpose" },
    { 0x15E3, "bot_get_heli_goal_dist_sq" },
    { 0x15E4, "bot_get_heli_slowdown_dist_sq" },
    { 0x15E5, "bot_get_known_attacker" },
    { 0x15E7, "bot_get_nodes_in_cone" },
    { 0x15E8, "bot_get_player_team" },
    { 0x15E9, "bot_get_rank_xp" },
    { 0x15EA, "bot_get_string_index_for_integer" },
    { 0x15EB, "bot_get_team_limit" },
    { 0x15EC, "bot_get_teammates_currently_defending_point" },
    { 0x15EE, "bot_get_total_gun_ammo" },
    { 0x15EF, "bot_get_zones_within_dist" },
    { 0x15F0, "bot_get_zones_within_dist_recurs" },
    { 0x15F1, "bot_goal_can_override" },
    { 0x15F2, "bot_grenade_matches_purpose" },
    { 0x15F3, "bot_guard_player" },
    { 0x15F5, "bot_has_tactical_goal" },
    { 0x15F6, "bot_heli_find_unvisited_nodes" },
    { 0x15F7, "bot_heli_nodes" },
    { 0x15F8, "bot_heli_pilot_traceoffset" },
    { 0x15F9, "bot_in_combat" },
    { 0x15FA, "bot_initialized_remote_vehicles" },
    { 0x15FB, "bot_interaction_type" },
    { 0x15FC, "bot_invalid_attachment_combos" },
    { 0x15FD, "bot_is_bodyguarding" },
    { 0x15FE, "bot_is_capturing" },
    { 0x15FF, "bot_is_defending" },
    { 0x1600, "bot_is_defending_point" },
    { 0x1601, "bot_is_fireteam_mode" },
    { 0x1602, "bot_is_guarding_player" },
    { 0x1605, "bot_is_remote_or_linked" },
    { 0x1606, "bot_killstreak_choose_loc_enemies" },
    { 0x1607, "bot_killstreak_drop" },
    { 0x1609, "bot_killstreak_drop_hidden" },
    { 0x160A, "bot_killstreak_drop_outside" },
    { 0x160B, "bot_killstreak_get_all_outside_allies" },
    { 0x160C, "bot_killstreak_get_all_outside_enemies" },
    { 0x160E, "bot_killstreak_get_zone_allies_outside" },
    { 0x160F, "bot_killstreak_get_zone_enemies_outside" },
    { 0x1610, "bot_killstreak_is_valid_internal" },
    { 0x1612, "bot_killstreak_never_use" },
    { 0x1613, "bot_killstreak_remote_control" },
    { 0x1614, "bot_killstreak_sentry" },
    { 0x1615, "bot_killstreak_setup" },
    { 0x1616, "bot_killstreak_simple_use" },
    { 0x1617, "bot_killstreak_valid_for_specific_streaktype" },
    { 0x161B, "bot_ks_funcs" },
    { 0x161F, "bot_loadout_choose_from_attachmenttable" },
    { 0x1621, "bot_loadout_choose_from_perktable" },
    { 0x1622, "bot_loadout_choose_from_set" },
    { 0x1626, "bot_loadout_copy_from_client" },
    { 0x1627, "bot_loadout_fields" },
    { 0x1629, "bot_loadout_pick" },
    { 0x162A, "bot_loadout_set" },
    { 0x162B, "bot_loadout_setup_perks" },
    { 0x162C, "bot_loadout_setup_squad_match" },
    { 0x162D, "bot_loadout_valid_choice" },
    { 0x162F, "bot_lui_convert_team_to_int" },
    { 0x1631, "bot_map_center" },
    { 0x1632, "bot_map_max_x" },
    { 0x1633, "bot_map_max_y" },
    { 0x1635, "bot_map_min_x" },
    { 0x1636, "bot_map_min_y" },
    { 0x1637, "bot_map_min_z" },
    { 0x1638, "bot_melee_tactical_insertion_check" },
    { 0x1639, "bot_memory_goal" },
    { 0x163A, "bot_memory_goal_time" },
    { 0x163D, "bot_monitor_watch_entrances_bodyguard" },
    { 0x163F, "bot_new_tactical_goal" },
    { 0x1640, "bot_notify_streak_used" },
    { 0x1641, "bot_odin_find_target_for_airdrop" },
    { 0x1642, "bot_odin_find_target_for_rods" },
    { 0x1643, "bot_odin_get_closest_visible_outside_player" },
    { 0x1646, "bot_odin_get_player_target_point" },
    { 0x1647, "bot_odin_get_visible_outside_players" },
    { 0x1648, "bot_odin_should_airdrop_at_marker" },
    { 0x164A, "bot_odin_should_fire_flash_at_marker" },
    { 0x164B, "bot_odin_should_fire_rod_at_marker" },
    { 0x164C, "bot_odin_time_to_move" },
    { 0x164D, "bot_odin_try_airdrop" },
    { 0x164E, "bot_odin_try_flash" },
    { 0x164F, "bot_odin_try_rods" },
    { 0x1650, "bot_odin_try_smoke" },
    { 0x1651, "bot_odin_try_spawn_juggernaut" },
    { 0x1652, "bot_out_of_ammo" },
    { 0x1654, "bot_outside_gate_watch" },
    { 0x1656, "bot_perk_cost" },
    { 0x1657, "bot_perktable" },
    { 0x1659, "bot_pers_init" },
    { 0x165A, "bot_pers_update" },
    { 0x165E, "bot_pick_random_point_from_set" },
    { 0x1660, "bot_pickup_weapon" },
    { 0x1662, "bot_point_is_on_pathgrid" },
    { 0x1663, "bot_post_teleport" },
    { 0x1664, "bot_post_use_ammo_crate" },
    { 0x1665, "bot_post_use_box_of_type" },
    { 0x1667, "bot_pre_use_box_of_type" },
    { 0x1669, "bot_protect_point" },
    { 0x166A, "bot_queued_process" },
    { 0x166B, "bot_queued_process_level_thread" },
    { 0x166C, "bot_queued_process_level_thread_active" },
    { 0x166E, "bot_random_path" },
    { 0x166F, "bot_random_path_default" },
    { 0x1670, "bot_random_path_function" },
    { 0x1671, "bot_random_ranks_for_difficulty" },
    { 0x1673, "bot_register_killstreak_func" },
    { 0x1674, "bot_remove_from_bot_level_targets" },
    { 0x1675, "bot_respawn_launcher_name" },
    { 0x1676, "bot_restart_think_threads" },
    { 0x1677, "bot_rnd_prestige" },
    { 0x1678, "bot_rnd_rank" },
    { 0x167A, "bot_seek_dropped_weapon" },
    { 0x167B, "bot_send_cancel_notify" },
    { 0x167D, "bot_sentry_activate" },
    { 0x167E, "bot_sentry_add_goal" },
    { 0x1680, "bot_sentry_carried_obj" },
    { 0x1681, "bot_sentry_choose_placement" },
    { 0x1682, "bot_sentry_choose_target" },
    { 0x1683, "bot_sentry_ensure_exit" },
    { 0x1684, "bot_sentry_force_cancel" },
    { 0x1685, "bot_sentry_path_start" },
    { 0x1686, "bot_sentry_path_thread" },
    { 0x1688, "bot_set_ambush_trap" },
    { 0x1689, "bot_set_ambush_trap_wait_fire" },
    { 0x168A, "bot_set_bombzone_bottargets" },
    { 0x168B, "bot_set_difficulty" },
    { 0x168D, "bot_set_personality" },
    { 0x168F, "bot_setup_bot_targets" },
    { 0x1690, "bot_setup_callback_class" },
    { 0x1692, "bot_setup_radio_bottargets" },
    { 0x1693, "bot_shootable_target_watch" },
    { 0x1694, "bot_should_do_killcam" },
    { 0x1696, "bot_should_pickup_weapons" },
    { 0x1697, "bot_should_use_ammo_crate" },
    { 0x1699, "bot_should_use_grenade_crate" },
    { 0x169B, "bot_should_use_scavenger_bag" },
    { 0x169C, "bot_spawned_before" },
    { 0x169D, "bot_squad_lookup" },
    { 0x169E, "bot_squad_lookup_enemy" },
    { 0x16A0, "bot_squad_lookup_ranked" },
    { 0x16A1, "bot_start_aa_launcher_tracking" },
    { 0x16A4, "bot_supported_killstreaks" },
    { 0x16A5, "bot_switch_to_killstreak_weapon" },
    { 0x16A6, "bot_target" },
    { 0x16A7, "bot_targets" },
    { 0x16A8, "bot_tdm_apply_commander_tactics" },
    { 0x16AC, "bot_think_crate" },
    { 0x16AD, "bot_think_crate_blocking_path" },
    { 0x16AE, "bot_think_gametype" },
    { 0x16AF, "bot_think_killstreak" },
    { 0x16B1, "bot_think_revive" },
    { 0x16B2, "bot_think_seek_dropped_weapons" },
    { 0x16B3, "bot_think_tactical_goals" },
    { 0x16B5, "bot_think_watch_enemy" },
    { 0x16B6, "bot_underground_trapped_watch" },
    { 0x16B8, "bot_usebutton_wait" },
    { 0x16B9, "bot_valid_camp_assassin" },
    { 0x16BA, "bot_validate_perk" },
    { 0x16BB, "bot_validate_reticle" },
    { 0x16BC, "bot_validate_weapon" },
    { 0x16BD, "bot_vanguard_find_unvisited_nodes" },
    { 0x16BE, "bot_vanguard_height_trace_size" },
    { 0x16C5, "bot_waittill_goal_or_fail" },
    { 0x16C6, "bot_waittill_out_of_combat_or_time" },
    { 0x16C7, "bot_waittill_using_vehicle" },
    { 0x16C8, "bot_war_think" },
    { 0x16CB, "bot_watch_manual_detonate" },
    { 0x16CC, "bot_watch_nodes" },
    { 0x16CD, "bot_weap_personality" },
    { 0x16CE, "bot_weap_statstable" },
    { 0x16D0, "botlastloadout" },
    { 0x16D1, "botlastloadoutdifficulty" },
    { 0x16D2, "botlastloadoutpersonality" },
    { 0x16D3, "botloadoutfavoritecamo" },
    { 0x16D7, "bots_disable_team_switching" },
    { 0x16D8, "bots_exist" },
    { 0x16DA, "bots_fireteam_num_classes_loaded" },
    { 0x16DB, "bots_gametype_handles_class_choice" },
    { 0x16DC, "bots_gametype_handles_team_choice" },
    { 0x16DD, "bots_ignore_team_balance" },
    { 0x16DE, "bots_notify_on_disconnect" },
    { 0x16E1, "bots_update_difficulty" },
    { 0x16E2, "bottarget" },
    { 0x16E3, "bottargets" },
    { 0x16E4, "bottom" },
    { 0x16E5, "bottom_arc" },
    { 0x16E6, "bottom_missile_target" },
    { 0x16E7, "bottom_target" },
    { 0x16E8, "bottom_tower_enemies" },
    { 0x16E9, "bouncelight" },
    { 0x16EB, "boundry_radius_cache" },
    { 0x16EC, "boundryradius" },
    { 0x16EE, "box_agentconnected" },
    { 0x16EF, "box_disableplayeruse" },
    { 0x16F0, "box_enableplayeruse" },
    { 0x16F1, "box_handledamage" },
    { 0x16F2, "box_handledeath" },
    { 0x16F3, "box_handledeathdamage" },
    { 0x16F4, "box_handleownerdisconnect" },
    { 0x16F6, "box_modelteamupdater" },
    { 0x16F9, "box_playerjoinedteam" },
    { 0x16FA, "box_setactive" },
    { 0x16FB, "box_seticon" },
    { 0x16FC, "box_setinactive" },
    { 0x16FD, "box_should_leave_immediately" },
    { 0x1702, "boxes" },
    { 0x1704, "boxmodifydamage" },
    { 0x1706, "boxsettings" },
    { 0x1707, "boxthink" },
    { 0x1708, "boxtouchonly" },
    { 0x1709, "boxtype" },
    { 0x170A, "bp" },
    { 0x170C, "bplrpostsoundplaying" },
    { 0x170D, "branchnodes" },
    { 0x170F, "bravo1_ascend_ascender" },
    { 0x1710, "bravo1_ascend_launcher" },
    { 0x1712, "bravo2_ascend_launcher" },
    { 0x1713, "bravo_ascend" },
    { 0x1714, "bravo_ascend_anim_node" },
    { 0x1716, "bravo_ascend_rope2" },
    { 0x1717, "bravo_ascend_rubberband" },
    { 0x1719, "bravo_ascend_rubberband_cleanup" },
    { 0x171B, "bravo_friendly_logic" },
    { 0x171C, "bravo_post_snake_cam" },
    { 0x171E, "breach_abort" },
    { 0x1720, "breach_ai_space_death" },
    { 0x1721, "breach_allies" },
    { 0x1722, "breach_ally_drones_drone_think" },
    { 0x1725, "breach_anim_struct" },
    { 0x1726, "breach_animated_door_init" },
    { 0x1728, "breach_bad_weapon_hint" },
    { 0x172B, "breach_charge_fx_activate" },
    { 0x172E, "breach_cleanup" },
    { 0x1731, "breach_damage_radius" },
    { 0x1732, "breach_damage_watch" },
    { 0x1736, "breach_dont_fire" },
    { 0x1737, "breach_door" },
    { 0x1738, "breach_door_clips" },
    { 0x1739, "breach_door_close" },
    { 0x173A, "breach_door_init" },
    { 0x173B, "breach_door_open" },
    { 0x173D, "breach_door_rig" },
    { 0x173E, "breach_doors" },
    { 0x1740, "breach_enemies_stunned" },
    { 0x1741, "breach_enemy_array" },
    { 0x1743, "breach_enemy_catch_exceptions" },
    { 0x1745, "breach_enemy_death_dmg" },
    { 0x1747, "breach_enemy_player_stab" },
    { 0x1748, "breach_enemy_ragdoll_on_death" },
    { 0x1749, "breach_enemy_setup" },
    { 0x174A, "breach_enemy_spawner_think" },
    { 0x174E, "breach_enemy_waitfor_breach_ending" },
    { 0x1751, "breach_explosion" },
    { 0x1752, "breach_explosion_notify" },
    { 0x1754, "breach_fire_straight" },
    { 0x1756, "breach_friendlies_restore_grenades" },
    { 0x1759, "breach_functions" },
    { 0x175A, "breach_fx_setup" },
    { 0x175B, "breach_grenade" },
    { 0x175C, "breach_grenade_smoke" },
    { 0x175D, "breach_group_trigger_think" },
    { 0x175E, "breach_groups" },
    { 0x175F, "breach_gun_up" },
    { 0x1760, "breach_heli_door" },
    { 0x1761, "breach_hide" },
    { 0x1765, "breach_icon" },
    { 0x1766, "breach_icon_count" },
    { 0x1767, "breach_icon_create" },
    { 0x1768, "breach_icon_fade_in" },
    { 0x1769, "breach_icon_fade_out" },
    { 0x176A, "breach_icon_think" },
    { 0x176B, "breach_icon_update" },
    { 0x176C, "breach_icon_update_is_player_in_range" },
    { 0x176D, "breach_icons" },
    { 0x176E, "breach_index" },
    { 0x176F, "breach_init" },
    { 0x1771, "breach_mines" },
    { 0x1772, "breach_missionfailed" },
    { 0x1774, "breach_nag" },
    { 0x1776, "breach_no_auto_reload" },
    { 0x1777, "breach_not_ready_hint" },
    { 0x1778, "breach_on_activate" },
    { 0x1779, "breach_on_event" },
    { 0x177A, "breach_onenduse" },
    { 0x177B, "breach_onuse" },
    { 0x177C, "breach_open_door" },
    { 0x177D, "breach_open_watch" },
    { 0x177E, "breach_origin" },
    { 0x177F, "breach_other_watch" },
    { 0x1780, "breach_outside_ambience" },
    { 0x1781, "breach_participants_ready_to_proceed" },
    { 0x1783, "breach_passive_time" },
    { 0x1788, "breach_precache" },
    { 0x178A, "breach_pt" },
    { 0x178E, "breach_reset_goaladius" },
    { 0x1790, "breach_rumble" },
    { 0x1791, "breach_set_2dicon" },
    { 0x1793, "breach_set_can_use" },
    { 0x1794, "breach_set_goaladius" },
    { 0x1795, "breach_setup" },
    { 0x1796, "breach_sfx" },
    { 0x1797, "breach_shot_blood_fx" },
    { 0x1799, "breach_show" },
    { 0x179A, "breach_spawner_setup" },
    { 0x179B, "breach_spawners" },
    { 0x179C, "breach_targets" },
    { 0x179D, "breach_think" },
    { 0x179E, "breach_too_many_enemies_hint" },
    { 0x179F, "breach_trigger_cleanup" },
    { 0x17A1, "breach_use_triggers" },
    { 0x17A3, "breach_vehicles" },
    { 0x17A4, "breach_wait_move" },
    { 0x17A5, "breach_wait_nag_proc" },
    { 0x17A8, "breach_weapon" },
    { 0x17A9, "breachdonotfire" },
    { 0x17AA, "breached" },
    { 0x17AE, "breacher_think" },
    { 0x17AF, "breachers" },
    { 0x17B2, "breachfriendlies" },
    { 0x17B3, "breachfriendlies_can_teleport" },
    { 0x17B4, "breachfriendlies_grenades_empty" },
    { 0x17B6, "breaching" },
    { 0x17B7, "breaching_shots_fired" },
    { 0x17B8, "breachless_door_opens" },
    { 0x17BB, "break_cockpit_glass" },
    { 0x17BE, "break_light" },
    { 0x17BF, "break_nearest_light" },
    { 0x17C0, "break_stealth_mg" },
    { 0x17C1, "break_window" },
    { 0x17C2, "breakable_light" },
    { 0x17C5, "breaktarget" },
    { 0x17C7, "breathing_overlay" },
    { 0x17C8, "breathingmanager" },
    { 0x17CA, "bren_switchout_logic" },
    { 0x17CB, "brian" },
    { 0x17D0, "bridge_anim" },
    { 0x17D1, "bridge_begin" },
    { 0x17D2, "bridge_deploy_begin" },
    { 0x17D3, "bridge_deploy_enemy_a10_gun_dives" },
    { 0x17D4, "bridge_deploy_enemy_tanks_setup" },
    { 0x17D5, "bridge_deploy_init" },
    { 0x17D7, "bridge_destroyed" },
    { 0x17D8, "bridge_enemies" },
    { 0x17D9, "bridge_event" },
    { 0x17DD, "bridge_init" },
    { 0x17DE, "bridge_jeep_by" },
    { 0x17E0, "bridge_main" },
    { 0x17E1, "bridge_push" },
    { 0x17E3, "bridge_push_4" },
    { 0x17E4, "bridge_scene_anims" },
    { 0x17E5, "bridge_scene_tank_anims" },
    { 0x17E8, "bridges" },
    { 0x17E9, "briefcase" },
    { 0x17EA, "bright_light_flash_into_console" },
    { 0x17EC, "bring_up_osprey" },
    { 0x17ED, "brinkofdeathkillstreak" },
    { 0x17EE, "broadcast_player_input" },
    { 0x17F1, "broken_wall_delete" },
    { 0x17F2, "broken_wall_init" },
    { 0x17F3, "broken_wall_show" },
    { 0x17F4, "broken_walls" },
    { 0x17FA, "brushmodels" },
    { 0x17FB, "bsharpturnduringsharpturn" },
    { 0x17FC, "bshootwhilemoving" },
    { 0x17FD, "bsnowmobilesstarted" },
    { 0x17FE, "bsoundlooping" },
    { 0x17FF, "bsplashplaying" },
    { 0x1800, "bsprinton" },
    { 0x1801, "bstopbarrage" },
    { 0x1802, "btankmoving" },
    { 0x1803, "bthermalon" },
    { 0x1804, "btr_ambush" },
    { 0x1805, "btr_attack_player_on_flag" },
    { 0x1806, "btr_burst" },
    { 0x1807, "btr_by_mountainside" },
    { 0x1808, "btr_check_player_fire" },
    { 0x1809, "btr_fire_logic" },
    { 0x180A, "btr_look_logic" },
    { 0x180B, "btr_mg_off" },
    { 0x180C, "btr_sees_body" },
    { 0x180D, "btr_stop_when_not_normal" },
    { 0x180E, "btr_target_player" },
    { 0x180F, "btr_turret_follow" },
    { 0x1810, "buddyspawn" },
    { 0x1813, "bufferedstats" },
    { 0x1814, "bufferedstatsmax" },
    { 0x1817, "build_aianims" },
    { 0x1818, "build_all_treadfx" },
    { 0x181A, "build_bulletshield" },
    { 0x181C, "build_car" },
    { 0x181D, "build_chopperboss_defaults" },
    { 0x181E, "build_data_override" },
    { 0x181F, "build_death_badplace" },
    { 0x1820, "build_death_jolt_delay" },
    { 0x1821, "build_deathanim" },
    { 0x1822, "build_deathfx" },
    { 0x1823, "build_deathfx_override" },
    { 0x1824, "build_deathmodel" },
    { 0x1829, "build_elevators" },
    { 0x182A, "build_exhaust" },
    { 0x182B, "build_frontarmor" },
    { 0x182C, "build_fx" },
    { 0x182D, "build_gaz_death" },
    { 0x182E, "build_grenadeshield" },
    { 0x182F, "build_hideparts" },
    { 0x1830, "build_human_model" },
    { 0x1831, "build_humvee_anims" },
    { 0x1833, "build_is_airplane" },
    { 0x1834, "build_is_helicopter" },
    { 0x1835, "build_life" },
    { 0x1837, "build_light_override" },
    { 0x1839, "build_lynx_death" },
    { 0x183C, "build_pillageitem_array" },
    { 0x183D, "build_pillageitem_arrays" },
    { 0x183E, "build_quake" },
    { 0x183F, "build_radiusdamage" },
    { 0x1840, "build_rider_death_func" },
    { 0x1841, "build_rocket_deathfx" },
    { 0x1844, "build_rumble_unique" },
    { 0x1845, "build_shoot_shock" },
    { 0x1846, "build_single_tread" },
    { 0x1847, "build_spawn_zones" },
    { 0x1849, "build_template" },
    { 0x184A, "build_treadfx" },
    { 0x184C, "build_turret" },
    { 0x184D, "build_turrets" },
    { 0x1850, "buildalienweaponname" },
    { 0x1851, "buildalienweaponnamecamo" },
    { 0x1856, "buildbaseweaponlist" },
    { 0x1858, "buildchallengetableinfo" },
    { 0x1859, "builddot_damage" },
    { 0x185A, "builddot_ontick" },
    { 0x185C, "builddot_wait" },
    { 0x1861, "building_entry_exit_anim_struct" },
    { 0x1864, "building_entry_tv" },
    { 0x1865, "building_fall_anim_rig" },
    { 0x1867, "building_hit_fx" },
    { 0x1869, "building_slide_control_hint" },
    { 0x186A, "buildscoreboardtype" },
    { 0x186B, "buildshadowgeomopt" },
    { 0x186D, "buildweaponname" },
    { 0x186F, "buildweaponnamereticle" },
    { 0x1870, "bullet_armor" },
    { 0x1871, "bullet_attack" },
    { 0x1872, "bullet_blender" },
    { 0x1874, "bullet_caused_fuel_leaks" },
    { 0x1875, "bullet_damage_scalar" },
    { 0x1878, "bullet_holes2" },
    { 0x1879, "bullet_holes3" },
    { 0x187A, "bullet_pinhole_tag_fx_orgs" },
    { 0x187B, "bullet_resistance" },
    { 0x187C, "bullet_start" },
    { 0x187D, "bullet_strafe_start" },
    { 0x1880, "bullet_watcher_ambush" },
    { 0x1881, "bulletcount" },
    { 0x1884, "bulletsinclip" },
    { 0x1885, "bulletwhizbycheck_whilemoving" },
    { 0x1886, "bulletwhizbycheckloop" },
    { 0x1887, "bulletwhizbyreaction" },
    { 0x1888, "bully" },
    { 0x1889, "bully_shoots" },
    { 0x188A, "bump_player" },
    { 0x188C, "bunker_balcony_bullet_impacts" },
    { 0x188E, "bunker_balcony_bullet_impacts_think" },
    { 0x188F, "bunker_balcony_cleanup" },
    { 0x1890, "bunker_balcony_damage_state" },
    { 0x1893, "bunker_beach_ai" },
    { 0x1894, "bunker_beach_attackers" },
    { 0x1895, "bunker_beach_attackers_death" },
    { 0x1896, "bunker_beach_attackers_think" },
    { 0x1897, "bunker_beach_vehicles" },
    { 0x1898, "bunker_enemy_cover_drones" },
    { 0x1899, "bunker_final_strafe" },
    { 0x189A, "bunker_final_strafe_hind_missiles" },
    { 0x189C, "bunker_final_strafe_think" },
    { 0x189E, "bunker_mg_guy" },
    { 0x189F, "bunker_reinforcement_helis" },
    { 0x18A1, "bunker_style_mortar_activate" },
    { 0x18A5, "bunker_style_mortar_on" },
    { 0x18A6, "bunker_style_mortar_think" },
    { 0x18A7, "bunker_style_mortar_trigger" },
    { 0x18A8, "bunker_trench_drone_runners" },
    { 0x18AA, "bunker_vehicle_javelin_watcher" },
    { 0x18AB, "burn_trig" },
    { 0x18AC, "burning" },
    { 0x18AF, "burnville_paratrooper_hack_loop" },
    { 0x18B2, "burst_fire_settings" },
    { 0x18B5, "burst_pause_max" },
    { 0x18B6, "burst_pause_min" },
    { 0x18B7, "burst_shots_max" },
    { 0x18B9, "burstdelay" },
    { 0x18BA, "burstfirenumshots" },
    { 0x18BC, "burstmin" },
    { 0x18BD, "burstshootanimrate" },
    { 0x18BF, "bus_movement_internal" },
    { 0x18C0, "bus_movement_model_logic" },
    { 0x18C1, "bus_movement_sounds_rumble_etc" },
    { 0x18C2, "busdust" },
    { 0x18C3, "busereadyidle" },
    { 0x18C6, "bust_thru_prep" },
    { 0x18C7, "bust_thru_prep_dog" },
    { 0x18C8, "butchdance" },
    { 0x18C9, "butchdance_combat" },
    { 0x18CA, "button_is_clicked" },
    { 0x18CB, "button_is_held" },
    { 0x18CD, "button_parse_parameters" },
    { 0x18CE, "button_pressed_from_string" },
    { 0x18D0, "button_sound" },
    { 0x18D1, "button_switch" },
    { 0x18D2, "button_toggles" },
    { 0x18D3, "buttonclick" },
    { 0x18D5, "buttonisheld" },
    { 0x18D6, "buttonpressed_internal" },
    { 0x18D8, "buttonthink" },
    { 0x18DA, "buzzkill" },
    { 0x18DB, "bwalldestroyed" },
    { 0x18DC, "bwalldestroyedsoundplayed" },
    { 0x18DD, "bypass_max_attacker_counter" },
    { 0x18DE, "bypassclasschoice" },
    { 0x18DF, "bypassclasschoicefunc" },
    { 0x18E0, "c" },
    { 0x18E1, "c130" },
    { 0x18E2, "c130setup" },
    { 0x18E3, "c17_drops" },
    { 0x18E4, "c4_cancel_flag" },
    { 0x18E6, "c4_earthquake" },
    { 0x18E7, "c4_hint" },
    { 0x18E8, "c4_hintstring" },
    { 0x18E9, "c4_location" },
    { 0x18EA, "c4_sound_override" },
    { 0x18EB, "c4_spot_glow" },
    { 0x18EC, "c4_weaponname" },
    { 0x18ED, "c4activate" },
    { 0x18EF, "c4damage" },
    { 0x18F0, "c4death" },
    { 0x18F1, "c4deathdetonate" },
    { 0x18F2, "c4detectiontrigger" },
    { 0x18F4, "c4empdamage" },
    { 0x18F5, "c4empkillstreakwait" },
    { 0x18F7, "c4used" },
    { 0x18F9, "cac_getkillstreak" },
    { 0x18FA, "cac_getkillstreakwithtype" },
    { 0x18FB, "cac_getperk" },
    { 0x18FD, "cac_getsecondarygrenade" },
    { 0x18FE, "cac_getweapon" },
    { 0x18FF, "cac_getweaponattachment" },
    { 0x1901, "cac_getweaponattachmenttwo" },
    { 0x1902, "cac_getweaponbuff" },
    { 0x1903, "cac_getweaponcamo" },
    { 0x1904, "cac_getweaponreticle" },
    { 0x1906, "cac_selector" },
    { 0x1907, "cache_ambient" },
    { 0x1908, "cache_ambient_element" },
    { 0x1909, "cache_ambient_event" },
    { 0x190D, "cache_mix_default" },
    { 0x190E, "cache_occlusion" },
    { 0x190F, "cache_reverb" },
    { 0x1910, "cache_timescale" },
    { 0x1911, "cache_whizby" },
    { 0x1912, "cache_zone" },
    { 0x1913, "cached" },
    { 0x1914, "cached_ambients" },
    { 0x1915, "cached_arcs" },
    { 0x1916, "cached_elems" },
    { 0x1919, "cage_guys3" },
    { 0x191A, "cairo_church_whines" },
    { 0x191B, "cairo_scared_dialogue" },
    { 0x191C, "calc_max_last_stands" },
    { 0x191D, "calcanimstartpos" },
    { 0x191E, "calctrackingyaw" },
    { 0x191F, "calculate_and_show_hive_scores" },
    { 0x1920, "calculate_arc" },
    { 0x1921, "calculate_bezier_curve" },
    { 0x1922, "calculate_bravo_rubberband_base" },
    { 0x1923, "calculate_challenge_score" },
    { 0x1924, "calculate_current_intensity_level" },
    { 0x1925, "calculate_defend_stance" },
    { 0x1926, "calculate_discount" },
    { 0x1927, "calculate_drill_protection_score" },
    { 0x1929, "calculate_maxmovedeltainanimstate" },
    { 0x192A, "calculate_partial_hive_scores" },
    { 0x192B, "calculate_personal_skill_score" },
    { 0x192D, "calculate_players_total_end_game_score" },
    { 0x192E, "calculate_stopsoonnotifydist" },
    { 0x1930, "calculate_total_end_game_score" },
    { 0x1931, "calculateanimdata" },
    { 0x1932, "calculated_closest_point" },
    { 0x1933, "calculated_nearest_node" },
    { 0x1934, "calculatekillcamtime" },
    { 0x1936, "calculatenodeoffsetfromanimationdelta" },
    { 0x1938, "call_elevator" },
    { 0x1939, "call_fire_exploder_on_spawn" },
    { 0x193A, "call_flag_when_clacked" },
    { 0x193B, "call_in_airdrop_heli" },
    { 0x193C, "call_in_attack_heli" },
    { 0x193D, "call_in_hive_heli" },
    { 0x193E, "call_in_rescue_heli" },
    { 0x193F, "call_jolt_by_flag" },
    { 0x1942, "callairsupport" },
    { 0x1943, "callback_alienplayerdamage" },
    { 0x1947, "callback_playerconnect" },
    { 0x1948, "callback_playerdamage" },
    { 0x1949, "callback_playerdamage_internal" },
    { 0x194C, "callback_playerlaststand" },
    { 0x194D, "callback_playerlaststandalien" },
    { 0x194F, "callback_playermigrated" },
    { 0x1950, "callback_startgametype" },
    { 0x1951, "callback_vehicledamage" },
    { 0x1953, "callbackhostmigration" },
    { 0x1955, "callbackplayerdamage" },
    { 0x1956, "callbackplayerdisconnect" },
    { 0x1957, "callbackplayerkilled" },
    { 0x1958, "callbackplayerlaststand" },
    { 0x1959, "callbackplayermigrated" },
    { 0x195B, "callbackstartgametype" },
    { 0x195E, "caller" },
    { 0x195F, "calloptionalbehaviorcallback" },
    { 0x1960, "callout" },
    { 0x1961, "calloutdestroyed" },
    { 0x1962, "callouttypewillrepeat" },
    { 0x1963, "callstrike" },
    { 0x1965, "callstrike_bombeffect" },
    { 0x1966, "calm_idle_get_random" },
    { 0x1967, "cam" },
    { 0x1968, "cam_sound_sources" },
    { 0x1969, "camera" },
    { 0x196C, "camera_zoom" },
    { 0x196D, "camera_zoomout" },
    { 0x196E, "cameramodel" },
    { 0x1970, "camlanding_from_apache" },
    { 0x1971, "camp_mblur_changes" },
    { 0x1974, "camper_guy" },
    { 0x1975, "camper_time_started_hunting" },
    { 0x1976, "camper_trigger_think" },
    { 0x1978, "campfire_temp_dialog" },
    { 0x1979, "camping_needs_fallback_camp_location" },
    { 0x197A, "camtime" },
    { 0x197B, "can_activate_trap" },
    { 0x197D, "can_attempt_badpath_move" },
    { 0x1980, "can_cut_rope" },
    { 0x1981, "can_do_charge_attack" },
    { 0x1982, "can_early_melee" },
    { 0x1984, "can_leap_melee" },
    { 0x1986, "can_place_sentry" },
    { 0x1987, "can_purchase_chopper" },
    { 0x1989, "can_say_event_type" },
    { 0x198B, "can_say_soundalias" },
    { 0x198C, "can_see_attacker_for_a_bit" },
    { 0x198D, "can_see_enemy" },
    { 0x198E, "can_see_origin" },
    { 0x1990, "can_spawn_at_any_node" },
    { 0x1991, "can_spawn_meteoroid_alien" },
    { 0x1992, "can_spit_gas_cloud" },
    { 0x1994, "can_use_health" },
    { 0x1995, "canactivatefunc" },
    { 0x1996, "canal_event" },
    { 0x1997, "canbepickedup" },
    { 0x1998, "canbeplaced" },
    { 0x1999, "canbetargeted" },
    { 0x199C, "cancalloutlocation" },
    { 0x199E, "cancel_repair_on_hive_death" },
    { 0x19A1, "cancelmode" },
    { 0x19A2, "cancelnukeondeath" },
    { 0x19A4, "canceluse_default_deployable_box" },
    { 0x19A5, "canceluse_dpad_airstrike" },
    { 0x19A6, "canceluse_dpad_backup_buddy" },
    { 0x19A7, "canceluse_dpad_death_machine" },
    { 0x19A8, "canceluse_dpad_glsentry" },
    { 0x19A9, "canceluse_dpad_ims" },
    { 0x19AA, "canceluse_dpad_minigun_turret" },
    { 0x19AB, "canceluse_dpad_predator" },
    { 0x19AC, "canceluse_dpad_riotshield" },
    { 0x19AD, "canceluse_dpad_sentry" },
    { 0x19AE, "canceluse_dpad_war_machine" },
    { 0x19B2, "candodge" },
    { 0x19B3, "candoflavorburst" },
    { 0x19B4, "candojumpforend" },
    { 0x19B5, "candopain" },
    { 0x19B6, "candostartmove" },
    { 0x19B7, "candoturnanim" },
    { 0x19B8, "canenttriggerplatform" },
    { 0x19B9, "canfire" },
    { 0x19BC, "cangive_booster" },
    { 0x19BE, "cangive_health" },
    { 0x19BF, "cangive_loot_blood" },
    { 0x19C1, "cangive_maxammo" },
    { 0x19C2, "cangive_offhand" },
    { 0x19C3, "cangive_self_revive" },
    { 0x19C4, "cangive_shock_ammo" },
    { 0x19C5, "cangive_slotted_explosive" },
    { 0x19C7, "cangive_throwable_weapon" },
    { 0x19CB, "canlean" },
    { 0x19CC, "canlogclient" },
    { 0x19CD, "canlogevent" },
    { 0x19CE, "canlogkillstreak" },
    { 0x19CF, "canloglife" },
    { 0x19D0, "canmovepointtopoint" },
    { 0x19D1, "cannedtraverseanims" },
    { 0x19D3, "cannonfirevfx" },
    { 0x19D4, "cannonrumble" },
    { 0x19D5, "cannotplacestring" },
    { 0x19D6, "canperformclienttraces" },
    { 0x19D7, "canpickcrate" },
    { 0x19D8, "canpickupobject" },
    { 0x19DC, "canpurchase_dpad_backup_buddy" },
    { 0x19DD, "canpurchase_dpad_death_machine" },
    { 0x19E0, "canpurchase_dpad_minigun_turret" },
    { 0x19E1, "canpurchase_dpad_predator" },
    { 0x19E2, "canpurchase_dpad_riotshield" },
    { 0x19E6, "canpurchase_dpad_team_armor" },
    { 0x19E7, "canpurchase_dpad_team_boost" },
    { 0x19E8, "canpurchase_dpad_team_currency" },
    { 0x19EB, "canpurchase_dpad_team_specialammo" },
    { 0x19ED, "canreactagain" },
    { 0x19EE, "canreadtext" },
    { 0x19EF, "canregenhealth" },
    { 0x19F0, "canreturntocover" },
    { 0x19F1, "cansave" },
    { 0x19F5, "canseeandshootpoint" },
    { 0x19F7, "canseeenemyfromexposed" },
    { 0x19F9, "canseepointfromexposedatnode" },
    { 0x19FA, "canshootwhilerunning" },
    { 0x19FB, "canshootwhilerunningbackward" },
    { 0x19FC, "canshootwhilerunningforward" },
    { 0x19FE, "canstoppeeking" },
    { 0x19FF, "cansuppressenemy" },
    { 0x1A02, "cantfindanythingtodo" },
    { 0x1A03, "canthittarget" },
    { 0x1A04, "canthrowgrenade" },
    { 0x1A05, "cantraceto" },
    { 0x1A06, "cantseeenemybehavior" },
    { 0x1A07, "cantseeenemywait" },
    { 0x1A09, "canuse" },
    { 0x1A0B, "canuse_dpad_backup_buddy" },
    { 0x1A0C, "canuse_dpad_death_machine" },
    { 0x1A0D, "canuse_dpad_glsentry" },
    { 0x1A0E, "canuse_dpad_ims" },
    { 0x1A0F, "canuse_dpad_minigun_turret" },
    { 0x1A10, "canuse_dpad_predator" },
    { 0x1A14, "canuse_dpad_team_ammo" },
    { 0x1A15, "canuse_dpad_team_armor" },
    { 0x1A16, "canuse_dpad_team_boost" },
    { 0x1A17, "canuse_dpad_team_currency" },
    { 0x1A18, "canuse_dpad_team_explosives" },
    { 0x1A1C, "canusecallback" },
    { 0x1A1D, "canusedeployable" },
    { 0x1A1F, "canuseotherboxes" },
    { 0x1A21, "canyon_init" },
    { 0x1A22, "canyon_jumper_setup" },
    { 0x1A23, "canyon_main" },
    { 0x1A24, "cap_range" },
    { 0x1A25, "capacity" },
    { 0x1A26, "capsule_ninja" },
    { 0x1A27, "capturednuke" },
    { 0x1A2A, "capturingstring" },
    { 0x1A2C, "car_1" },
    { 0x1A2D, "car_alarm_org" },
    { 0x1A2E, "car_alarm_timeout" },
    { 0x1A2F, "car_damage_owner_recorder" },
    { 0x1A32, "care_package_watch" },
    { 0x1A34, "career_stat_increment" },
    { 0x1A35, "career_stats_init" },
    { 0x1A38, "cargo" },
    { 0x1A39, "cargo_amb" },
    { 0x1A3A, "cargo_choppers" },
    { 0x1A3B, "cargo_container" },
    { 0x1A3C, "cargo_item_spawners" },
    { 0x1A3E, "cargo_winds_front" },
    { 0x1A3F, "cargo_winds_rear" },
    { 0x1A40, "carried_grape" },
    { 0x1A41, "carriedby" },
    { 0x1A43, "carrieditem" },
    { 0x1A45, "carrier" },
    { 0x1A46, "carrier_grape" },
    { 0x1A48, "carrier_life_jet_takeoff_guys" },
    { 0x1A49, "carrier_life_jet_takeoff_jet" },
    { 0x1A4A, "carrier_liferaft" },
    { 0x1A4C, "carrier_light_post_sparrow" },
    { 0x1A4D, "carrier_locations" },
    { 0x1A4E, "carrier_planes_precache" },
    { 0x1A4F, "carrier_post_load" },
    { 0x1A50, "carrier_set_vision_rog_tilt" },
    { 0x1A51, "carrier_starts" },
    { 0x1A52, "carriervisible" },
    { 0x1A53, "carry_bishop" },
    { 0x1A54, "carry_in" },
    { 0x1A55, "carry_vanguard" },
    { 0x1A56, "carryflag" },
    { 0x1A57, "carryicon" },
    { 0x1A58, "carrying_pickedup_sentry" },
    { 0x1A59, "carryobject" },
    { 0x1A5A, "carryobject_overridemovingplatformdeath" },
    { 0x1A5B, "carryobjectproxthink" },
    { 0x1A5C, "carryobjectproxthinkdelayed" },
    { 0x1A5E, "carryobjectusethink" },
    { 0x1A5F, "carryremoteuav_delete" },
    { 0x1A61, "carryremoteuav_setcarried" },
    { 0x1A62, "cars" },
    { 0x1A63, "cart_player_kill_volume" },
    { 0x1A64, "cart_runner" },
    { 0x1A66, "casualtytracking" },
    { 0x1A67, "cat_array_add" },
    { 0x1A68, "cat_array_get" },
    { 0x1A69, "catch_alien_on_fire" },
    { 0x1A6B, "catch_death_notify" },
    { 0x1A6C, "catch_interrupt_notify" },
    { 0x1A6F, "catchup_deck_combat" },
    { 0x1A70, "catchup_deck_tilt" },
    { 0x1A73, "catchup_defend_sparrow" },
    { 0x1A75, "catchup_function" },
    { 0x1A78, "catchup_slow_intro" },
    { 0x1A7C, "catwalk_melee_glass_break" },
    { 0x1A7D, "catwalks_end" },
    { 0x1A82, "cave_cairo_whine" },
    { 0x1A83, "cave_dialogue" },
    { 0x1A84, "cave_dialogue_2" },
    { 0x1A85, "cave_dust" },
    { 0x1A86, "cave_earthquake" },
    { 0x1A8D, "ceiling_rubble" },
    { 0x1A8E, "ceiling_rubble_missile_explode_watch" },
    { 0x1A8F, "ceiling_rubble_onplayerconnect" },
    { 0x1A90, "ceiling_rubble_watchusage" },
    { 0x1A94, "centeraimpoint" },
    { 0x1A95, "centered" },
    { 0x1A97, "centerref" },
    { 0x1A98, "cfxprintln" },
    { 0x1A99, "cfxprintlnend" },
    { 0x1A9A, "cfxprintlnstart" },
    { 0x1A9B, "cg_drawbreathhint" },
    { 0x1A9C, "ch_assists" },
    { 0x1A9D, "ch_bulletdamagecommon" },
    { 0x1AA0, "ch_getstate" },
    { 0x1AA1, "ch_gettarget" },
    { 0x1AA3, "ch_isactivechallenge" },
    { 0x1AA5, "ch_roundplayed" },
    { 0x1AA6, "ch_roundwin" },
    { 0x1AA7, "ch_setprogress" },
    { 0x1AA9, "ch_vehicle_killed" },
    { 0x1AAA, "ch_vehicle_kills" },
    { 0x1AAB, "ch_vo" },
    { 0x1AAC, "chain_gate" },
    { 0x1AAF, "chaingun_shelleject_fx" },
    { 0x1AB4, "chair_anim_ent" },
    { 0x1AB5, "chair_animate" },
    { 0x1AB8, "chair_nags" },
    { 0x1AB9, "chair_vargas" },
    { 0x1ABC, "chalk_swipe1" },
    { 0x1ABD, "chalk_swipe2" },
    { 0x1ABE, "challenge_countdown" },
    { 0x1AC0, "challenge_registration_func" },
    { 0x1AC1, "challenge_rewardval" },
    { 0x1AC4, "challengedata" },
    { 0x1AC5, "challengeinfo" },
    { 0x1AC7, "challengereward" },
    { 0x1AC8, "challengescompleted" },
    { 0x1AC9, "challengeslot" },
    { 0x1ACB, "challengetarget" },
    { 0x1ACD, "chance" },
    { 0x1ACE, "chancefordoground" },
    { 0x1ACF, "chancetospawndog" },
    { 0x1AD0, "chancetospawnpickup" },
    { 0x1AD1, "change_character_model" },
    { 0x1AD3, "change_enemy_priority" },
    { 0x1AD4, "change_guard_to_hazmat" },
    { 0x1AD5, "change_light" },
    { 0x1AD6, "change_mix" },
    { 0x1AD8, "change_zone_stairwell" },
    { 0x1ADA, "changecrateweight" },
    { 0x1ADB, "changed_team" },
    { 0x1ADE, "changelightcolortoworkerthread" },
    { 0x1AE0, "changestanceforfuntime" },
    { 0x1AE1, "changestepoutpos" },
    { 0x1AE3, "changingcoverpos" },
    { 0x1AE5, "changingtoregularinfected" },
    { 0x1AE6, "changingtoregularinfectedbykill" },
    { 0x1AE7, "changingweapon" },
    { 0x1AE8, "chaos_a" },
    { 0x1AEA, "chaos_a_hesh_landing" },
    { 0x1AEB, "chaos_a_hesh_run" },
    { 0x1AEE, "chaos_airdrop_locs" },
    { 0x1AEF, "chaos_b" },
    { 0x1AF0, "chaos_b_hide_debris" },
    { 0x1AF1, "chaos_b_slow_zone" },
    { 0x1AF2, "chaos_checkpoint" },
    { 0x1AF3, "chaos_checkpoint_trigger" },
    { 0x1AF7, "chaos_hide_on_start" },
    { 0x1AF8, "chaos_kill_after_time" },
    { 0x1AFA, "chaos_kill_player" },
    { 0x1AFB, "chaos_moving_clip" },
    { 0x1AFC, "chaos_moving_head" },
    { 0x1AFD, "chaos_music" },
    { 0x1AFE, "chaos_player_kill" },
    { 0x1AFF, "chaos_quake_trigger_think" },
    { 0x1B00, "chaos_rog_think" },
    { 0x1B02, "chaos_walker_wait" },
    { 0x1B09, "character_type" },
    { 0x1B0A, "characters" },
    { 0x1B0B, "charactertypebonusxp" },
    { 0x1B0C, "charge" },
    { 0x1B0D, "charge1" },
    { 0x1B0E, "charge2" },
    { 0x1B0F, "charge_attack" },
    { 0x1B10, "charge_tracking_enemy" },
    { 0x1B12, "chase_amb_enter_chasm" },
    { 0x1B13, "chase_amb_enter_ravine" },
    { 0x1B16, "chase_amb_garage_exit" },
    { 0x1B19, "chase_amb_garage_punch_it" },
    { 0x1B1A, "chase_amb_garage_start" },
    { 0x1B1B, "chase_amb_sharp_turn" },
    { 0x1B1C, "chase_amb_sub_comes_up" },
    { 0x1B1D, "chase_amb_tight_spot" },
    { 0x1B1F, "chase_amb_under_bridge_2" },
    { 0x1B20, "chase_amb_under_bridge_3" },
    { 0x1B21, "chase_amb_under_bridge_4" },
    { 0x1B22, "chase_cam_ent" },
    { 0x1B23, "chase_cam_last_num" },
    { 0x1B26, "chase_concussion" },
    { 0x1B27, "chase_crack_icehole" },
    { 0x1B28, "chase_crashmix" },
    { 0x1B29, "chase_dog" },
    { 0x1B2A, "chase_dog_blocker_think" },
    { 0x1B2B, "chase_dog_dialogue" },
    { 0x1B2C, "chase_land_roof" },
    { 0x1B2D, "chase_land_tires_big" },
    { 0x1B2E, "chase_land_tires_small" },
    { 0x1B30, "chase_music" },
    { 0x1B31, "chase_pileup_counter" },
    { 0x1B32, "chase_player" },
    { 0x1B33, "chase_player_collision" },
    { 0x1B34, "chase_player_jolt" },
    { 0x1B35, "chase_roadblock_smash" },
    { 0x1B36, "chase_script" },
    { 0x1B37, "chase_sink" },
    { 0x1B3C, "chase_tunnel_jeep" },
    { 0x1B3D, "chasecam" },
    { 0x1B3E, "chasecam_onent" },
    { 0x1B41, "chatqueue" },
    { 0x1B42, "cheap_effect_id" },
    { 0x1B43, "cheap_vehicles_have_shields" },
    { 0x1B44, "cheatammoifnecessary" },
    { 0x1B46, "check_ai_array_for_death" },
    { 0x1B47, "check_already_touching" },
    { 0x1B4B, "check_atrium_done" },
    { 0x1B4C, "check_break_stealth" },
    { 0x1B4D, "check_break_stealth_end" },
    { 0x1B51, "check_debug_hud_dvar" },
    { 0x1B53, "check_dog_ready_to_attack" },
    { 0x1B54, "check_dog_sprinting" },
    { 0x1B55, "check_enemy_clump" },
    { 0x1B56, "check_enemy_index" },
    { 0x1B57, "check_failed_spawn_groups" },
    { 0x1B58, "check_feature_dependencies" },
    { 0x1B5A, "check_flag_for_stat_tracking" },
    { 0x1B5C, "check_for_block" },
    { 0x1B5D, "check_for_combat_two_done" },
    { 0x1B5E, "check_for_death_during_traversal" },
    { 0x1B5F, "check_for_enemies" },
    { 0x1B60, "check_for_existing_pet_bombs" },
    { 0x1B61, "check_for_melee_stab" },
    { 0x1B63, "check_for_player_impact" },
    { 0x1B64, "check_for_player_meleeing" },
    { 0x1B65, "check_for_player_near_hive_with_drill" },
    { 0x1B66, "check_for_rog_death" },
    { 0x1B67, "check_for_special_damage" },
    { 0x1B68, "check_for_weapon_pickup" },
    { 0x1B69, "check_force_color" },
    { 0x1B6B, "check_grenade" },
    { 0x1B6C, "check_guard_death_to_stop_pistol_fire" },
    { 0x1B6E, "check_hold_bar_in_green" },
    { 0x1B6F, "check_hold_bar_in_green_or_die" },
    { 0x1B70, "check_hostage_flag_a" },
    { 0x1B71, "check_hostage_flag_b" },
    { 0x1B72, "check_if_player_close_to_checkpoint" },
    { 0x1B73, "check_if_went_hot_late" },
    { 0x1B75, "check_kill_traversal" },
    { 0x1B77, "check_melee_interaction_active" },
    { 0x1B78, "check_missing_animation" },
    { 0x1B79, "check_other_haslevelveteranachievement" },
    { 0x1B7A, "check_overrides" },
    { 0x1B7C, "check_pain_lockout" },
    { 0x1B7F, "check_player_warehouse_mantle" },
    { 0x1B80, "check_player_zoom" },
    { 0x1B81, "check_script_char_group_ratio" },
    { 0x1B82, "check_sound_tag_dupe" },
    { 0x1B83, "check_sound_tag_dupe_reset" },
    { 0x1B86, "check_submix_hud_dvar" },
    { 0x1B89, "check_triggers_flagset" },
    { 0x1B8A, "check_unloadgroup" },
    { 0x1B8C, "check_weapon" },
    { 0x1B8F, "checkadvanceonenemyconditions" },
    { 0x1B90, "checkallgeneratorsofthistype" },
    { 0x1B93, "checkapproachpreconditions" },
    { 0x1B94, "checkarrivalenterpositions" },
    { 0x1B97, "checkchanged" },
    { 0x1B98, "checkcombatstate" },
    { 0x1B99, "checkcoverenterpos" },
    { 0x1B9B, "checkdynamicspawns" },
    { 0x1B9C, "checked" },
    { 0x1B9E, "checkexpiretime" },
    { 0x1B9F, "checkforbestweapon" },
    { 0x1BA0, "checkforcebleedout" },
    { 0x1BA1, "checkforceragdoll" },
    { 0x1BA3, "checkforperkupgrade" },
    { 0x1BA4, "checkforpersonalbests" },
    { 0x1BA5, "checkforthrowback" },
    { 0x1BA6, "checkgeneratorplacement" },
    { 0x1BA7, "checkgrenadethrowdist" },
    { 0x1BA8, "checkhit" },
    { 0x1BAA, "checkhitsthismag_regularmp" },
    { 0x1BAB, "checking_for_hits" },
    { 0x1BAC, "checkisfacing" },
    { 0x1BAD, "checkkillsteal" },
    { 0x1BAE, "checkmatchdataequipmentkills" },
    { 0x1BB2, "checkowner" },
    { 0x1BB3, "checkpitchvisibility" },
    { 0x1BB4, "checkplayer" },
    { 0x1BB5, "checkplayerscorelimitsoon" },
    { 0x1BB8, "checkpoint_base_array" },
    { 0x1BB9, "checkpoint_bridge" },
    { 0x1BBA, "checkpoint_canyon" },
    { 0x1BBB, "checkpoint_cave" },
    { 0x1BBC, "checkpoint_chaos" },
    { 0x1BBF, "checkpoint_crash_site" },
    { 0x1BC0, "checkpoint_defend" },
    { 0x1BC3, "checkpoint_exfil" },
    { 0x1BC5, "checkpoint_foley" },
    { 0x1BC6, "checkpoint_interior" },
    { 0x1BC7, "checkpoint_interior_combat" },
    { 0x1BC8, "checkpoint_interior_cqb" },
    { 0x1BCA, "checkpoint_osprey" },
    { 0x1BCC, "checkpoint_patrol_anim" },
    { 0x1BCE, "checkpoint_ride" },
    { 0x1BD2, "checkpoint_tank" },
    { 0x1BD3, "checkpoint_tower" },
    { 0x1BD4, "checkpredictedspawnpointcorrectness" },
    { 0x1BD8, "checkstreakreward" },
    { 0x1BD9, "checkteam" },
    { 0x1BDE, "checktransitionpreconditions" },
    { 0x1BE5, "choke_loc" },
    { 0x1BE6, "choke_trigs" },
    { 0x1BE9, "choose_fire" },
    { 0x1BEA, "choose_from_weighted_array" },
    { 0x1BEB, "choose_goal_vol" },
    { 0x1BEC, "choose_goal_vol_chain" },
    { 0x1BEE, "choose_random_explosive_type" },
    { 0x1BEF, "choose_spit_type" },
    { 0x1BF1, "chooseattackidle" },
    { 0x1BF2, "choosedirection" },
    { 0x1BF4, "choosepose" },
    { 0x1BF5, "chooseposefunc" },
    { 0x1BF6, "choper_fly_in_think" },
    { 0x1BF7, "chopper" },
    { 0x1BFA, "chopper_achievement_check" },
    { 0x1BFB, "chopper_active" },
    { 0x1BFC, "chopper_ai_init" },
    { 0x1BFD, "chopper_air_support_activate" },
    { 0x1BFE, "chopper_arrive_wind_gust" },
    { 0x1C00, "chopper_attacker_check" },
    { 0x1C01, "chopper_boss_agro" },
    { 0x1C03, "chopper_boss_attempt_firing" },
    { 0x1C05, "chopper_boss_can_hit_from_tag_turret" },
    { 0x1C0C, "chopper_boss_forced_target_set" },
    { 0x1C0D, "chopper_boss_gather_targets" },
    { 0x1C0E, "chopper_boss_get_best_location_and_target" },
    { 0x1C10, "chopper_boss_get_best_target_proc" },
    { 0x1C12, "chopper_boss_get_closest_neighbor_2d" },
    { 0x1C13, "chopper_boss_get_closest_target_2d" },
    { 0x1C14, "chopper_boss_get_target" },
    { 0x1C15, "chopper_boss_goto_hangout" },
    { 0x1C17, "chopper_boss_in_range" },
    { 0x1C1C, "chopper_boss_locs_disabled" },
    { 0x1C1D, "chopper_boss_locs_monitor_disable" },
    { 0x1C1E, "chopper_boss_locs_monitor_disable_clean_up" },
    { 0x1C1F, "chopper_boss_locs_monitor_disable_reset" },
    { 0x1C21, "chopper_boss_locs_populate" },
    { 0x1C22, "chopper_boss_locs_populate_thread" },
    { 0x1C23, "chopper_boss_locs_populate_thread_optimized" },
    { 0x1C24, "chopper_boss_locs_populated" },
    { 0x1C27, "chopper_boss_move" },
    { 0x1C28, "chopper_boss_path_override" },
    { 0x1C29, "chopper_boss_path_paused" },
    { 0x1C2B, "chopper_boss_post_move_func" },
    { 0x1C2C, "chopper_boss_pre_move_func" },
    { 0x1C2D, "chopper_boss_resume_path_finding" },
    { 0x1C31, "chopper_boss_think" },
    { 0x1C33, "chopper_boss_wait_populate" },
    { 0x1C35, "chopper_crash" },
    { 0x1C36, "chopper_crash_enemies" },
    { 0x1C38, "chopper_crash_guy_logic" },
    { 0x1C39, "chopper_custom_death_spin" },
    { 0x1C3D, "chopper_decoys_and_evade" },
    { 0x1C3E, "chopper_destroys_cover" },
    { 0x1C3F, "chopper_does_first_flyby" },
    { 0x1C42, "chopper_hear_vo" },
    { 0x1C43, "chopper_infantry_tweak" },
    { 0x1C44, "chopper_insta_kill" },
    { 0x1C46, "chopper_leaves_after_time" },
    { 0x1C49, "chopper_missile_burst" },
    { 0x1C4B, "chopper_outline_monitor" },
    { 0x1C4E, "chopper_reacts_to_lockon" },
    { 0x1C4F, "chopper_resume_path" },
    { 0x1C50, "chopper_reward" },
    { 0x1C51, "chopper_rumble_earthquake" },
    { 0x1C52, "chopper_scripted_death" },
    { 0x1C53, "chopper_side_dodge" },
    { 0x1C54, "chopper_sound" },
    { 0x1C58, "chopper_start_at_path" },
    { 0x1C5B, "chopper_turret_on" },
    { 0x1C5D, "chopperboss_const" },
    { 0x1C5F, "choppers_do_strafe_attacks" },
    { 0x1C64, "chopperturretofffunc" },
    { 0x1C65, "chopperturretonfunc" },
    { 0x1C66, "chosen" },
    { 0x1C67, "chosentemplates" },
    { 0x1C68, "chunk_spark_fx_tag" },
    { 0x1C6A, "church_destruction_init" },
    { 0x1C6B, "church_fall" },
    { 0x1C6C, "church_weapon_pullout" },
    { 0x1C6E, "chyron" },
    { 0x1C6F, "chyron_sound" },
    { 0x1C70, "cig_throwing" },
    { 0x1C72, "cinematic_off" },
    { 0x1C73, "cinematic_on" },
    { 0x1C74, "cinematic_over" },
    { 0x1C75, "cinematicmode_off" },
    { 0x1C77, "cipher_ambush_approach" },
    { 0x1C79, "cipher_keegan_tower_kill" },
    { 0x1C7A, "cipher_podium" },
    { 0x1C7B, "cipher_vo" },
    { 0x1C7E, "circle_radius" },
    { 0x1C7F, "circle_time" },
    { 0x1C80, "circling" },
    { 0x1C82, "city_collapse" },
    { 0x1C84, "civilian_combatmoveturn" },
    { 0x1C86, "civilian_jet_flyby" },
    { 0x1C87, "civilian_noncombatmoveturn" },
    { 0x1C88, "civilian_walk_animation" },
    { 0x1C8A, "civilianjetflyby" },
    { 0x1C8B, "civilianjetflyby_timer" },
    { 0x1C8C, "civilianprops" },
    { 0x1C8D, "civs" },
    { 0x1C8E, "civs_set_flag_on_damage" },
    { 0x1C8F, "clacker_vo_failsafe" },
    { 0x1C93, "claimed_node" },
    { 0x1C96, "claimtrigger" },
    { 0x1C97, "clampandnormalize" },
    { 0x1C98, "clampedhealth" },
    { 0x1C9A, "class" },
    { 0x1C9B, "class_num" },
    { 0x1C9D, "classcallback" },
    { 0x1C9F, "classhasperk" },
    { 0x1CA0, "classmap" },
    { 0x1CA1, "classname_ng" },
    { 0x1CA2, "classtablename" },
    { 0x1CA3, "classtablenum" },
    { 0x1CA4, "classtweaks" },
    { 0x1CA6, "claymore_tutorial_given" },
    { 0x1CAA, "claymoredetectiontrigger" },
    { 0x1CAB, "claymoredetonateradius" },
    { 0x1CAC, "claymoredetonation" },
    { 0x1CAE, "claymores" },
    { 0x1CAF, "claymoresentientfunc" },
    { 0x1CB0, "claymoretracking" },
    { 0x1CB1, "claymoreused" },
    { 0x1CB3, "clean_tow_hud" },
    { 0x1CB9, "clean_up_intro_exterior_props" },
    { 0x1CBA, "clean_up_lagged_aliens" },
    { 0x1CBB, "clean_up_on_death" },
    { 0x1CBC, "clean_up_on_owner_death" },
    { 0x1CBD, "clean_window_player" },
    { 0x1CC0, "cleanup" },
    { 0x1CC2, "cleanup_boats" },
    { 0x1CC4, "cleanup_complex" },
    { 0x1CC5, "cleanup_deck_transition" },
    { 0x1CC6, "cleanup_delayed" },
    { 0x1CC7, "cleanup_enemies" },
    { 0x1CC8, "cleanup_enemy_destroyer" },
    { 0x1CC9, "cleanup_ent_on_dismount" },
    { 0x1CCA, "cleanup_ents" },
    { 0x1CCB, "cleanup_ents_removing_bullet_shield" },
    { 0x1CCC, "cleanup_fixtures" },
    { 0x1CCE, "cleanup_intermission_enemies" },
    { 0x1CD0, "cleanup_land_vert_missile_fx" },
    { 0x1CD1, "cleanup_launcher" },
    { 0x1CD3, "cleanup_missile_fx" },
    { 0x1CD6, "cleanup_phys_obj" },
    { 0x1CD9, "cleanup_rope" },
    { 0x1CDA, "cleanup_rope_on_zodiac_death" },
    { 0x1CDE, "cleanup_sentry_hud_outline" },
    { 0x1CE0, "cleanup_teargas_on_exit" },
    { 0x1CE1, "cleanup_topdrive" },
    { 0x1CE2, "cleanup_triggers" },
    { 0x1CE3, "cleanup_vehicles" },
    { 0x1CE4, "cleanup_vert_missile_fx" },
    { 0x1CE6, "cleanup_zodiac_bodies" },
    { 0x1CE7, "cleanup_zodiacs" },
    { 0x1CE8, "cleanup_zone_arrays" },
    { 0x1CEC, "cleanupents" },
    { 0x1CED, "cleanupequipment" },
    { 0x1CEE, "cleanupflyby" },
    { 0x1CEF, "cleanuplasertargetingdevice" },
    { 0x1CF3, "clear_all_corpses" },
    { 0x1CF4, "clear_ally_drones" },
    { 0x1CF5, "clear_animation" },
    { 0x1CF9, "clear_attached_items" },
    { 0x1CFA, "clear_attacking" },
    { 0x1CFB, "clear_audio_filter" },
    { 0x1CFC, "clear_audio_mix" },
    { 0x1CFD, "clear_audio_occlusion" },
    { 0x1CFE, "clear_bark_cairo" },
    { 0x1D03, "clear_colors" },
    { 0x1D05, "clear_custom_gameskill_func" },
    { 0x1D06, "clear_deathanim" },
    { 0x1D07, "clear_deck_props" },
    { 0x1D08, "clear_dog_master" },
    { 0x1D0C, "clear_entity_selection" },
    { 0x1D0D, "clear_exception" },
    { 0x1D0E, "clear_filter" },
    { 0x1D12, "clear_generic_run_anim" },
    { 0x1D13, "clear_groundref" },
    { 0x1D16, "clear_ignore_everything" },
    { 0x1D17, "clear_ignore_states" },
    { 0x1D1B, "clear_mix" },
    { 0x1D1E, "clear_promotion_order" },
    { 0x1D1F, "clear_rappel_move_flag" },
    { 0x1D20, "clear_reverb" },
    { 0x1D21, "clear_run_anim" },
    { 0x1D22, "clear_sam_missiles" },
    { 0x1D23, "clear_script_goal_on" },
    { 0x1D24, "clear_settable_fx" },
    { 0x1D25, "clear_target_array" },
    { 0x1D26, "clear_target_on_vehicle_death" },
    { 0x1D29, "clear_this_node" },
    { 0x1D2A, "clear_to_go_flag_set_once" },
    { 0x1D2B, "clear_tool_hud" },
    { 0x1D2E, "clear_turret_ammo_counter_on_dismount" },
    { 0x1D30, "clear_vista_vehicles" },
    { 0x1D31, "clear_volume_flag" },
    { 0x1D32, "clearafterfade" },
    { 0x1D33, "clearall" },
    { 0x1D34, "clearawardwinners" },
    { 0x1D36, "clearbuddymessage" },
    { 0x1D38, "cleardamagehistory" },
    { 0x1D39, "cleardodgeanims" },
    { 0x1D3B, "clearempondeath" },
    { 0x1D3C, "clearfaceanimonanimdone" },
    { 0x1D3D, "clearfxondeath" },
    { 0x1D40, "clearidshortly" },
    { 0x1D41, "clearing_bob_anim" },
    { 0x1D42, "clearing_sways" },
    { 0x1D44, "clearkillcamstate" },
    { 0x1D45, "clearkillstreaks" },
    { 0x1D49, "clearondeath" },
    { 0x1D4A, "clearonvictimdisconnect" },
    { 0x1D4B, "clearplayerlockfromremoteuavlaunch" },
    { 0x1D4D, "clearportalfx" },
    { 0x1D4E, "clearprevioustispawnpoint" },
    { 0x1D4F, "clearprogress" },
    { 0x1D50, "clearrideintro" },
    { 0x1D52, "clearsightposnear" },
    { 0x1D53, "clearspawnpointdistancedata" },
    { 0x1D56, "clearteamspawnpointsightdata" },
    { 0x1D58, "clearthreatbias" },
    { 0x1D5A, "clearwatervarsonspawn" },
    { 0x1D5B, "clientid" },
    { 0x1D5C, "clientmatchdataid" },
    { 0x1D5D, "clienttracespawnclass" },
    { 0x1D5E, "clienttweakables" },
    { 0x1D61, "cliff_choppers_move_on" },
    { 0x1D62, "cliff_guy_logic" },
    { 0x1D63, "climb_over" },
    { 0x1D64, "climbover_enemy" },
    { 0x1D66, "clip_brush" },
    { 0x1D67, "clip_delete" },
    { 0x1D68, "clip_down" },
    { 0x1D69, "clip_move" },
    { 0x1D6B, "clip_up" },
    { 0x1D6D, "clipammo" },
    { 0x1D71, "clipmovementtolerance" },
    { 0x1D73, "clockwork_exfil_pre_load" },
    { 0x1D75, "clockwork_interior_nvg_pre_load" },
    { 0x1D77, "clockwork_intro_pre_load" },
    { 0x1D78, "clockwork_locations" },
    { 0x1D79, "clockwork_starts" },
    { 0x1D7A, "clockwork_thermobaric_mines" },
    { 0x1D7C, "clockwork_treadfx_override" },
    { 0x1D7D, "close_distance_sq" },
    { 0x1D7E, "close_door" },
    { 0x1D7F, "close_elevator_doors" },
    { 0x1D80, "close_enemy_volume" },
    { 0x1D82, "close_gate" },
    { 0x1D84, "close_inner_doors" },
    { 0x1D86, "close_outer_doors" },
    { 0x1D87, "close_range_retreat" },
    { 0x1D88, "close_stream_enemies" },
    { 0x1D89, "close_to_waterfall_enemy_logic" },
    { 0x1D8B, "close_warehouse_doors" },
    { 0x1D8C, "closed_angles" },
    { 0x1D8D, "closed_defend_fx_ent" },
    { 0x1D8F, "closed_height" },
    { 0x1D90, "closed_pos" },
    { 0x1D92, "closed_x" },
    { 0x1D94, "closedlist" },
    { 0x1D96, "closeelevatordoors" },
    { 0x1D97, "closeenoughaimdegrees" },
    { 0x1D98, "closeomamenuondeath" },
    { 0x1D99, "closepos" },
    { 0x1D9A, "closerfunc" },
    { 0x1D9C, "closest_enemy_snowmobile_to_player" },
    { 0x1D9E, "closest_start_struct" },
    { 0x1DA1, "clouds" },
    { 0x1DA6, "cnd_delay_rope_link" },
    { 0x1DA7, "cnd_ent_flag_clear" },
    { 0x1DAC, "cnd_generic_flicker" },
    { 0x1DAD, "cnd_generic_flicker_msg_watcher" },
    { 0x1DAE, "cnd_generic_flicker_pause" },
    { 0x1DB0, "cnd_get_rope_anim_origin" },
    { 0x1DB4, "cnd_plyr_rpl_death" },
    { 0x1DB6, "cnd_plyr_rpl_handle_view_lerp" },
    { 0x1DBA, "cnd_plyr_rpl_setup_dvars" },
    { 0x1DBB, "cnd_plyr_rpl_setup_globals" },
    { 0x1DBD, "cnd_rappel_player_rope" },
    { 0x1DBE, "cnd_rappel_railing_obj" },
    { 0x1DC3, "cnd_rpl_cleanup" },
    { 0x1DC6, "cnd_rpl_do_move_bob" },
    { 0x1DC7, "cnd_rpl_do_rope" },
    { 0x1DC8, "cnd_rpl_do_stop_sway" },
    { 0x1DCB, "cnd_rpl_do_wind" },
    { 0x1DCD, "cnd_rpl_legs_notetracks" },
    { 0x1DCE, "cnd_rpl_plyr_link_to_rope" },
    { 0x1DD2, "cnd_shaft_flickering" },
    { 0x1DD6, "cobra_missile_models" },
    { 0x1DD7, "cobra_weapon_tags" },
    { 0x1DD9, "cockpit_tubes" },
    { 0x1DDA, "codecallback_agentadded" },
    { 0x1DDB, "codecallback_agentdamaged" },
    { 0x1DDC, "codecallback_agentkilled" },
    { 0x1DDD, "codecallback_codeendgame" },
    { 0x1DDE, "codecallback_hostmigration" },
    { 0x1DDF, "codecallback_leaderdialog" },
    { 0x1DE0, "codecallback_playerconnect" },
    { 0x1DE1, "codecallback_playerdamage" },
    { 0x1DE3, "codecallback_playerkilled" },
    { 0x1DE4, "codecallback_playerlaststand" },
    { 0x1DE5, "codecallback_playermigrated" },
    { 0x1DE7, "codecallback_vehicledamage" },
    { 0x1DEB, "col_lines" },
    { 0x1DEC, "col_radiuses" },
    { 0x1DED, "col_volumes" },
    { 0x1DEE, "coldbreath_player" },
    { 0x1DEF, "coldbreathfx" },
    { 0x1DF0, "colheight" },
    { 0x1DF2, "collapsed_derrick_wire_anims" },
    { 0x1DF3, "collect_func" },
    { 0x1DF5, "collectibles_hint_precache" },
    { 0x1DF7, "collectibles_model_precache" },
    { 0x1DFA, "collectibles_worldcount" },
    { 0x1DFB, "collision" },
    { 0x1DFC, "collision_brush_post_explosion" },
    { 0x1DFD, "collision_brush_pre_explosion" },
    { 0x1DFE, "collision_origin" },
    { 0x1E01, "color_debug_println" },
    { 0x1E02, "color_doesnt_care_about_classname" },
    { 0x1E03, "color_doesnt_care_about_heroes" },
    { 0x1E04, "color_helper_trigger" },
    { 0x1E05, "color_node" },
    { 0x1E08, "color_node_finds_user_from_colorcodes" },
    { 0x1E09, "color_node_type_function" },
    { 0x1E0A, "color_ordered_node_assignment" },
    { 0x1E0B, "color_status" },
    { 0x1E0D, "color_teams" },
    { 0x1E0E, "color_trig_moves_enemies" },
    { 0x1E10, "color_user" },
    { 0x1E11, "color_volume_enemies" },
    { 0x1E12, "colorblind" },
    { 0x1E15, "colordebug" },
    { 0x1E16, "colorindex" },
    { 0x1E17, "colorislegit" },
    { 0x1E18, "colorlist" },
    { 0x1E19, "colornode_func" },
    { 0x1E1B, "colornode_setgoal_func" },
    { 0x1E1D, "colornodes_debug_array" },
    { 0x1E1E, "colors" },
    { 0x1E20, "combat_allies_mainhall" },
    { 0x1E22, "combat_ambient_guys2" },
    { 0x1E24, "combat_handle_allies" },
    { 0x1E25, "combat_one" },
    { 0x1E28, "combat_one_door_col" },
    { 0x1E2A, "combat_one_enemy" },
    { 0x1E2B, "combat_one_start" },
    { 0x1E2C, "combat_one_wave_node" },
    { 0x1E2D, "combat_playfacialanim" },
    { 0x1E30, "combat_rappel_garden_entry" },
    { 0x1E34, "combat_rappel_garden_entry_enemies" },
    { 0x1E36, "combat_rappel_garden_entry_finish_player_rope" },
    { 0x1E37, "combat_rappel_garden_entry_first_frame" },
    { 0x1E3C, "combat_rappel_garden_entry_rotate_legs" },
    { 0x1E3D, "combat_rappel_garden_entry_set_small_legs_jump" },
    { 0x1E3E, "combat_rappel_garden_entry_set_small_rotate_jump" },
    { 0x1E3F, "combat_rappel_garden_entry_setup_glass" },
    { 0x1E40, "combat_rappel_garden_entry_setup_jumpoints" },
    { 0x1E41, "combat_rappel_garden_entry_setup_weapon" },
    { 0x1E42, "combat_rappel_garden_entry_should_shift_allies" },
    { 0x1E44, "combat_rappel_garden_entry_tree" },
    { 0x1E45, "combat_rappel_rope_coil_baker" },
    { 0x1E46, "combat_rappel_rope_coil_player" },
    { 0x1E47, "combat_rappel_rope_coil_rorke" },
    { 0x1E48, "combat_rappel_rope_rorke" },
    { 0x1E49, "combat_rappel_spawn_garden_entry_enemies" },
    { 0x1E4B, "combat_resource" },
    { 0x1E4C, "combat_two" },
    { 0x1E4E, "combat_two_dialogue" },
    { 0x1E54, "combatbehavior" },
    { 0x1E57, "combathigh" },
    { 0x1E58, "combathighicon" },
    { 0x1E59, "combathighoverlay" },
    { 0x1E5B, "combatidle" },
    { 0x1E5C, "combatidlepreventoverlappingplayer" },
    { 0x1E5D, "combatmemorytimeconst" },
    { 0x1E60, "combatspeedscalar" },
    { 0x1E62, "combatstate" },
    { 0x1E63, "combattime" },
    { 0x1E64, "comeback" },
    { 0x1E65, "comexpfuncs" },
    { 0x1E67, "comm_center_2_pieces" },
    { 0x1E68, "command_fail_late_death" },
    { 0x1E69, "command_geyser_light" },
    { 0x1E6A, "command_goal" },
    { 0x1E6D, "command_platform_bag_baker" },
    { 0x1E6E, "command_platform_bag_cypher" },
    { 0x1E6F, "command_platform_bag_keegan" },
    { 0x1E70, "command_platform_bag_player" },
    { 0x1E75, "commander_can_takeover_bot" },
    { 0x1E77, "commander_create_dom_obj" },
    { 0x1E79, "commander_gametype_initialized" },
    { 0x1E7A, "commander_gave_hint" },
    { 0x1E7B, "commander_handle_notify_quick" },
    { 0x1E7D, "commander_hint_delete_on_commander_menu" },
    { 0x1E7F, "commander_initialize_gametype" },
    { 0x1E80, "commander_last_tactic_applied" },
    { 0x1E81, "commander_last_tactic_selected" },
    { 0x1E82, "commander_loadout_class_callback" },
    { 0x1E83, "commander_moment" },
    { 0x1E84, "commander_monitor_tactics" },
    { 0x1E85, "commander_or_bot_change_class" },
    { 0x1E86, "commander_order_ack" },
    { 0x1E87, "commander_speaking" },
    { 0x1E89, "commander_spectate_first_available_bot" },
    { 0x1E8A, "commander_spectate_next_bot" },
    { 0x1E8B, "commander_takeover_first_available_bot" },
    { 0x1E8C, "commander_vo" },
    { 0x1E8D, "commander_wait_connect" },
    { 0x1E8F, "commanding_bot" },
    { 0x1E90, "common_flash_check" },
    { 0x1E92, "common_tryuse_actions" },
    { 0x1E93, "commonstarttime" },
    { 0x1E94, "compare" },
    { 0x1E96, "comparesizesfx" },
    { 0x1E98, "compass_flash" },
    { 0x1E99, "compass_objectives" },
    { 0x1E9C, "compassiconenemy" },
    { 0x1E9D, "compassiconfriendly" },
    { 0x1E9F, "complete_me" },
    { 0x1EA2, "completednodes" },
    { 0x1EA3, "complex_expl" },
    { 0x1EA6, "complex_script" },
    { 0x1EA7, "complexobjectives" },
    { 0x1EA8, "compute_round_based_percentages" },
    { 0x1EA9, "computer_guys_runin" },
    { 0x1EAA, "computer_guys_runin_vo" },
    { 0x1EAB, "computer_idf" },
    { 0x1EAD, "conf_fx" },
    { 0x1EAE, "config" },
    { 0x1EAF, "confirmed" },
    { 0x1EB0, "connect_and_start_tarps" },
    { 0x1EB1, "connect_door_path" },
    { 0x1EB3, "connect_nodes_after_crash" },
    { 0x1EB4, "connect_time" },
    { 0x1EB5, "connect_traverse" },
    { 0x1EB6, "connect_watch" },
    { 0x1EB9, "connected_nodes" },
    { 0x1EBC, "connectedpostgame" },
    { 0x1EBD, "connection_sound" },
    { 0x1EC1, "connecttime" },
    { 0x1EC2, "connecttraverses" },
    { 0x1EC5, "considerthrowgrenade" },
    { 0x1EC6, "consolation" },
    { 0x1EC7, "console" },
    { 0x1EC8, "console_scene_dialogue" },
    { 0x1EC9, "console_scene_player_blocker" },
    { 0x1ECA, "const_baker" },
    { 0x1ED0, "constraingametype" },
    { 0x1ED2, "constrict_view" },
    { 0x1ED3, "constrict_view_count" },
    { 0x1ED4, "contact_points" },
    { 0x1ED8, "context_sensative_dialog_filler" },
    { 0x1EDA, "context_sensative_dialog_guy_in_sight" },
    { 0x1EDB, "context_sensative_dialog_guy_in_sight_check" },
    { 0x1EDD, "context_sensative_dialog_kill" },
    { 0x1EDE, "context_sensative_dialog_kill_thread" },
    { 0x1EDF, "context_sensative_dialog_locations" },
    { 0x1EE0, "context_sensative_dialog_locations_add_notify_event" },
    { 0x1EE1, "context_sensative_dialog_locations_thread" },
    { 0x1EE2, "context_sensative_dialog_play_random_group_sound" },
    { 0x1EE3, "context_sensative_dialog_secondary_explosion_vehicle" },
    { 0x1EE4, "context_sensative_dialog_timedout" },
    { 0x1EE6, "context_sensative_dialog_vehicledeath" },
    { 0x1EE7, "context_sensative_dialog_vehiclespawn" },
    { 0x1EE8, "continous_fire" },
    { 0x1EE9, "continuedrivenmovement" },
    { 0x1EEA, "continuemovement" },
    { 0x1EEC, "control_nvg_lightmodels" },
    { 0x1EED, "control_nvg_staticscreens_off" },
    { 0x1EEE, "control_nvg_staticscreens_on" },
    { 0x1EEF, "control_panel_setup_lights" },
    { 0x1EF0, "control_room_combat" },
    { 0x1EF1, "control_room_enemies" },
    { 0x1EF2, "control_room_enemies_upper_setup" },
    { 0x1EF3, "control_room_enemy_setup" },
    { 0x1EF6, "control_screen_bink_init" },
    { 0x1EF7, "controlled_dog" },
    { 0x1EF8, "controlling_dog" },
    { 0x1EFA, "controlroom_top_2" },
    { 0x1EFB, "controlsfrozen" },
    { 0x1EFD, "contuing_to_move_check" },
    { 0x1EFE, "converge_on_target" },
    { 0x1EFF, "conversation_start" },
    { 0x1F00, "conversation_stop" },
    { 0x1F01, "convert_color_to_short_string" },
    { 0x1F02, "convert_fov_string" },
    { 0x1F03, "convert_guy_to_drone" },
    { 0x1F04, "convert_moving_cover_enemies_to_wave_3" },
    { 0x1F06, "convert_to_fake_riders" },
    { 0x1F07, "convert_to_fireworks_effect" },
    { 0x1F08, "convert_to_time_string" },
    { 0x1F09, "convo_generator" },
    { 0x1F0A, "convoy" },
    { 0x1F0C, "convoy_checkpoint" },
    { 0x1F0D, "convoy_checkpoint_spawn" },
    { 0x1F0E, "convoy_death_func" },
    { 0x1F0F, "convoy_kill_player" },
    { 0x1F10, "convoy_riders_death_func" },
    { 0x1F11, "convoy_riders_react_func" },
    { 0x1F13, "convoy_spawn_func" },
    { 0x1F14, "convoy_spawn_logic" },
    { 0x1F16, "convoy_tall_barricade_01" },
    { 0x1F17, "convoy_tall_barricade_02" },
    { 0x1F18, "convoy_think" },
    { 0x1F19, "convoy_veh_01" },
    { 0x1F1A, "convoy_veh_01a" },
    { 0x1F1C, "convoy_veh_03" },
    { 0x1F1F, "convoy_veh_05" },
    { 0x1F22, "cool_spawn" },
    { 0x1F23, "cool_walk" },
    { 0x1F26, "cooldowntime" },
    { 0x1F27, "cooldownwaittime" },
    { 0x1F28, "coop_breached_from_same_door_in_a_muliti_door_room" },
    { 0x1F29, "coop_icon_blinkcrement" },
    { 0x1F2B, "coop_icon_color_blink" },
    { 0x1F2C, "coop_icon_color_damage" },
    { 0x1F2D, "coop_icon_color_downed" },
    { 0x1F2E, "coop_icon_color_dying" },
    { 0x1F2F, "coop_icon_color_normal" },
    { 0x1F30, "coop_icon_color_shoot" },
    { 0x1F31, "coop_icon_state" },
    { 0x1F32, "coop_player_in_special_ops_survival" },
    { 0x1F33, "coop_player_touching_valid_door_volume" },
    { 0x1F35, "coop_with_one_player_downed" },
    { 0x1F37, "copier" },
    { 0x1F38, "copy_adrenaline" },
    { 0x1F3B, "copy_bar" },
    { 0x1F3C, "copy_cat" },
    { 0x1F3D, "copy_ents" },
    { 0x1F3E, "copy_fullweaponlist" },
    { 0x1F3F, "copy_killstreak_status" },
    { 0x1F41, "copy_node" },
    { 0x1F42, "copy_weapon_ammo_clip" },
    { 0x1F43, "copy_weapon_ammo_stock" },
    { 0x1F44, "copy_weapon_current" },
    { 0x1F46, "copymachine_ai" },
    { 0x1F48, "copymachine_cleanup" },
    { 0x1F49, "copymachine_falling_vo" },
    { 0x1F4C, "copymachine_window_event" },
    { 0x1F4D, "core_lights_red" },
    { 0x1F4E, "corner_clearfacialanim" },
    { 0x1F4F, "corner_playaimfacialanim" },
    { 0x1F50, "corner_playcornerfacialanim" },
    { 0x1F55, "cornerdirection" },
    { 0x1F56, "cornered_building_entry_pre_load" },
    { 0x1F57, "cornered_destruct_pre_load" },
    { 0x1F58, "cornered_falling_death" },
    { 0x1F59, "cornered_garden_pre_load" },
    { 0x1F5B, "cornered_interior_pre_load" },
    { 0x1F5E, "cornered_player_arms" },
    { 0x1F5F, "cornered_player_legs" },
    { 0x1F60, "cornered_rappel_pre_load" },
    { 0x1F61, "cornered_start_random_wind" },
    { 0x1F63, "cornered_starts" },
    { 0x1F64, "cornered_stop_random_wind" },
    { 0x1F65, "cornered_stop_rappel" },
    { 0x1F66, "cornerline_height" },
    { 0x1F67, "cornermode" },
    { 0x1F68, "cornerreload" },
    { 0x1F69, "cornerrightgrenadedeath" },
    { 0x1F6A, "cornersights" },
    { 0x1F6B, "corpse" },
    { 0x1F6C, "corpse_behavior_doesnt_require_player_sight" },
    { 0x1F6D, "corpse_clear" },
    { 0x1F6E, "corpse_entity" },
    { 0x1F6F, "corpse_entnums" },
    { 0x1F70, "corpse_height" },
    { 0x1F71, "corpse_ragdoll_when_vertical" },
    { 0x1F73, "corpseflag" },
    { 0x1F76, "cos45" },
    { 0x1F77, "cosine" },
    { 0x1F78, "cosine11_25" },
    { 0x1F79, "cosine15" },
    { 0x1F7B, "cosine45" },
    { 0x1F7D, "cost" },
    { 0x1F7E, "cost_display" },
    { 0x1F7F, "coughing_active" },
    { 0x1F81, "countdown_sound" },
    { 0x1F83, "counted" },
    { 0x1F86, "countryids" },
    { 0x1F87, "courtyard_cleanup_enemies" },
    { 0x1F88, "courtyard_directory" },
    { 0x1F8A, "courtyard_guys_notify_on_death" },
    { 0x1F8B, "courtyard_guys_walla" },
    { 0x1F8C, "courtyard_hazmat_guy_think" },
    { 0x1F8D, "courtyard_intro_ally_vo" },
    { 0x1F8F, "courtyard_intro_elevator" },
    { 0x1F90, "courtyard_intro_elevator_guy" },
    { 0x1F91, "courtyard_intro_elevator_guy_fail" },
    { 0x1F92, "courtyard_intro_elevator_guys" },
    { 0x1F95, "courtyard_office_a_doors" },
    { 0x1F97, "courtyard_office_chair" },
    { 0x1F99, "courtyard_office_death" },
    { 0x1F9A, "courtyard_office_enemies" },
    { 0x1F9B, "courtyard_office_enemy_anim" },
    { 0x1F9C, "courtyard_office_glass" },
    { 0x1F9D, "courtyard_office_props" },
    { 0x1F9E, "courtyard_office_stealth_end" },
    { 0x1F9F, "courtyard_rig_kill" },
    { 0x1FA2, "courtyard_transient_unload" },
    { 0x1FAA, "cover_warnings_disabled" },
    { 0x1FAD, "covercrouchfail" },
    { 0x1FAE, "covercrouchgrenade" },
    { 0x1FAF, "covercrouchlean_aimmode" },
    { 0x1FB0, "covercrouchleanpitch" },
    { 0x1FB1, "coverenterpos" },
    { 0x1FB2, "coverexit" },
    { 0x1FB3, "coverexitangles" },
    { 0x1FB6, "coverexitpostdist" },
    { 0x1FBB, "covermode" },
    { 0x1FBC, "covermulti_choosehidestate" },
    { 0x1FBD, "covermulti_dotransition" },
    { 0x1FBE, "covermulti_enterstate" },
    { 0x1FBF, "covermulti_exitstate" },
    { 0x1FC0, "covermulti_getanimtransition" },
    { 0x1FC2, "covermulti_getnonrandomvaliddir" },
    { 0x1FC3, "covermulti_getrandomvaliddir" },
    { 0x1FC4, "covermulti_getstatefromdir" },
    { 0x1FC5, "covermulti_isvaliddir" },
    { 0x1FC7, "covermulti_setstate" },
    { 0x1FC9, "covermulti_think" },
    { 0x1FCC, "coverposestablishedtime" },
    { 0x1FCE, "coverreload" },
    { 0x1FD1, "coverrightgrenade" },
    { 0x1FD2, "coversetupanim" },
    { 0x1FD3, "coverstand" },
    { 0x1FD4, "coverstandfail" },
    { 0x1FD5, "coverstandgrenade" },
    { 0x1FD8, "covertransangles" },
    { 0x1FDB, "covertranspredist" },
    { 0x1FDC, "covertype" },
    { 0x1FDF, "cq_combat_movement" },
    { 0x1FE0, "cq_enemies" },
    { 0x1FE1, "cq_room_destruction" },
    { 0x1FE2, "cqb_aim" },
    { 0x1FE4, "cqb_anims" },
    { 0x1FE5, "cqb_clearfacialanim" },
    { 0x1FE6, "cqb_dog" },
    { 0x1FE8, "cqb_door_open_slow" },
    { 0x1FE9, "cqb_door_shove" },
    { 0x1FEA, "cqb_encounter" },
    { 0x1FEB, "cqb_encounter_allies_move_up" },
    { 0x1FEC, "cqb_fans" },
    { 0x1FED, "cqb_off_sprint_on" },
    { 0x1FEF, "cqb_playfacialanim" },
    { 0x1FF0, "cqb_point_of_interest" },
    { 0x1FF1, "cqb_reloadinternal" },
    { 0x1FF3, "cqb_time" },
    { 0x1FF5, "cqb_wide_poi_track" },
    { 0x1FF6, "cqb_wide_target_track" },
    { 0x1FF7, "cqbenabled" },
    { 0x1FF8, "cqbpointsofinterest" },
    { 0x1FF9, "cqbtracking" },
    { 0x1FFE, "crack_exploder_test" },
    { 0x2000, "cranked" },
    { 0x2001, "cranked_end_time" },
    { 0x2002, "crankedbombtimer" },
    { 0x2003, "crash_blackhawk_fire_think" },
    { 0x2005, "crash_blackhawk_think" },
    { 0x2006, "crash_derailed_check" },
    { 0x2009, "crash_event" },
    { 0x200A, "crash_hidden_settings" },
    { 0x200B, "crash_model_damage" },
    { 0x200E, "crash_path_check" },
    { 0x200F, "crash_pilot_logic" },
    { 0x2011, "crash_site_a10_missile_dive_1" },
    { 0x2012, "crash_site_a10_overhead" },
    { 0x2014, "crash_site_begin" },
    { 0x2016, "crash_site_init" },
    { 0x2017, "crash_site_main" },
    { 0x2019, "crash_site_vo" },
    { 0x201A, "crash_test_start" },
    { 0x201C, "crash_vehicle_on_death" },
    { 0x201E, "crashed_truck1" },
    { 0x201F, "crashed_truck2" },
    { 0x2020, "crashed_trucks" },
    { 0x2022, "crashedc17_waits" },
    { 0x2023, "crashedtank_waits" },
    { 0x2024, "crashing" },
    { 0x2025, "crashplane" },
    { 0x2026, "crashtimer" },
    { 0x2027, "crate_calculate_on_path_grid" },
    { 0x2028, "crate_can_use_always" },
    { 0x202A, "crate_clip" },
    { 0x202B, "crate_clip_of_doom" },
    { 0x202C, "crate_death_fling" },
    { 0x202D, "crate_get_bot_target" },
    { 0x2033, "crate_low_ammo_check" },
    { 0x2035, "crate_picked_up" },
    { 0x2038, "crateallcapturethink" },
    { 0x203B, "cratekill" },
    { 0x203D, "cratemodelenemyteamsupdater" },
    { 0x203E, "cratemodelplayerupdater" },
    { 0x2040, "cratenonownerusetime" },
    { 0x2041, "crateothercapturethink" },
    { 0x2042, "crateownercapturethink" },
    { 0x2043, "crateownerusetime" },
    { 0x2046, "crater_ledge_walk" },
    { 0x2047, "crater_reveal_dialogue" },
    { 0x2048, "cratesetupforuse" },
    { 0x2049, "crateteammodelupdater" },
    { 0x204A, "cratetype" },
    { 0x204B, "cratetypes" },
    { 0x204D, "crateusepostjuggernautupdater" },
    { 0x204E, "crateuseteamupdater" },
    { 0x2051, "crawl_breath_start" },
    { 0x2052, "crawl_fx" },
    { 0x2053, "crawl_fx_rate" },
    { 0x2054, "crawl_hurt_pulse" },
    { 0x2055, "crawl_out" },
    { 0x2056, "crawl_target_and_init_flags" },
    { 0x2057, "crawl_through_targets_to_init_flags" },
    { 0x2059, "crawling_guys_spawnfunc" },
    { 0x205B, "crawlingpainanimoverridefunc" },
    { 0x205C, "crawlingpaintransanim" },
    { 0x205F, "creaky_board" },
    { 0x2061, "create_anim_event" },
    { 0x2062, "create_anim_node" },
    { 0x2063, "create_anim_scene" },
    { 0x2065, "create_array_of_intel_items" },
    { 0x2067, "create_artillery_shell" },
    { 0x206A, "create_chyron_text" },
    { 0x206B, "create_client_overlay" },
    { 0x206C, "create_client_overlay_custom_size" },
    { 0x2074, "create_dog_start_point_ent" },
    { 0x2075, "create_door_sound_ents" },
    { 0x2077, "create_dvar" },
    { 0x2078, "create_escape_doors" },
    { 0x207A, "create_flare" },
    { 0x207C, "create_fresh_friendly_icon" },
    { 0x207D, "create_fx_ent_setup" },
    { 0x207E, "create_fx_menu" },
    { 0x207F, "create_fxlighting_object" },
    { 0x2080, "create_gain_credit_hud" },
    { 0x2081, "create_gamemessage_text" },
    { 0x2083, "create_hatchet" },
    { 0x2084, "create_hud" },
    { 0x208B, "create_hud_lower_right" },
    { 0x208D, "create_hud_section" },
    { 0x208E, "create_hud_static_overlay" },
    { 0x208F, "create_hud_upper_left" },
    { 0x2090, "create_hud_upper_right" },
    { 0x2091, "create_hud_vertical_meter" },
    { 0x2093, "create_hud_xm25_screen" },
    { 0x2095, "create_jump_point" },
    { 0x2096, "create_line" },
    { 0x2097, "create_lock" },
    { 0x2098, "create_looper" },
    { 0x2099, "create_loopsound" },
    { 0x209A, "create_mantle" },
    { 0x209E, "create_new_spawner_pool" },
    { 0x20A0, "create_outofrange_hud" },
    { 0x20A1, "create_overlay_element" },
    { 0x20A2, "create_passengers" },
    { 0x20A3, "create_path" },
    { 0x20A4, "create_persistent_ice_breach_props" },
    { 0x20A7, "create_player_rig" },
    { 0x20A8, "create_player_surfacing_effects" },
    { 0x20A9, "create_pushed_dropped_weapon" },
    { 0x20AA, "create_qte_prompt" },
    { 0x20AB, "create_random_animation" },
    { 0x20AE, "create_reflection_objects" },
    { 0x20AF, "create_rog_controls" },
    { 0x20B0, "create_rumble_ent" },
    { 0x20B1, "create_single_missile_truck" },
    { 0x20B2, "create_sliding_space_door" },
    { 0x20B4, "create_smoke_and_ambience" },
    { 0x20B5, "create_start" },
    { 0x20B8, "create_sunflare_setting" },
    { 0x20B9, "create_trace_cache" },
    { 0x20BA, "create_track_struct" },
    { 0x20BF, "create_view_particle_source_locked" },
    { 0x20C0, "create_vision_set_fog" },
    { 0x20C1, "create_vo_data" },
    { 0x20C2, "create_warning_elem" },
    { 0x20C3, "create_world_model_from_ent_weapon" },
    { 0x20C5, "createairdropcrate" },
    { 0x20C7, "createart_transient_thread" },
    { 0x20C8, "createballdrone" },
    { 0x20C9, "createbar" },
    { 0x20CA, "createbombsquadmodel" },
    { 0x20CD, "createcarryremoteuav" },
    { 0x20D2, "createclientfontstring_func" },
    { 0x20D3, "createclienticon" },
    { 0x20D4, "createclientprogressbar" },
    { 0x20D5, "createclienttimer" },
    { 0x20D7, "createdogenemy" },
    { 0x20D8, "createdot" },
    { 0x20D9, "createdot_radius" },
    { 0x20DA, "createdropzones" },
    { 0x20DB, "createechoalias" },
    { 0x20DC, "createeffect" },
    { 0x20DF, "createexploder" },
    { 0x20E0, "createexploderex" },
    { 0x20E1, "createfontstring" },
    { 0x20E5, "createfx_centerprint" },
    { 0x20E6, "createfx_centerprint_thread" },
    { 0x20E8, "createfx_draw_enabled" },
    { 0x20EB, "createfx_inputlocked" },
    { 0x20EC, "createfx_loopcounter" },
    { 0x20ED, "createfx_offset" },
    { 0x20EE, "createfx_only_triggers" },
    { 0x20EF, "createfx_print3d" },
    { 0x20F0, "createfx_selecting" },
    { 0x20F1, "createfxbyfxid" },
    { 0x20F2, "createfxcursor" },
    { 0x20F4, "createfxexploders" },
    { 0x20F5, "createfxlogic" },
    { 0x20F6, "createfxmasks" },
    { 0x20F7, "createheli" },
    { 0x20F9, "createhordecrates" },
    { 0x20FB, "createhumanoidenemy" },
    { 0x20FD, "createicon_hudelem" },
    { 0x20FE, "createims" },
    { 0x20FF, "createimsforplayer" },
    { 0x2100, "createintervalsound" },
    { 0x2101, "createlbguard" },
    { 0x2102, "createline" },
    { 0x2104, "createloopeffect" },
    { 0x2105, "createloopsound" },
    { 0x2108, "createmuggercrates" },
    { 0x2109, "createmultipliertext" },
    { 0x210A, "createnewexploder" },
    { 0x210B, "createobject" },
    { 0x210C, "createobjective" },
    { 0x210E, "createoneshoteffect" },
    { 0x2110, "createplaneasheli" },
    { 0x2112, "createportal" },
    { 0x2113, "createportals" },
    { 0x2114, "createprimaryprogressbar" },
    { 0x2115, "createprimaryprogressbartext" },
    { 0x2116, "createprisoner" },
    { 0x2118, "createremoteuav" },
    { 0x2119, "createrpgrepulsors" },
    { 0x211D, "createserverfontstring" },
    { 0x211F, "createservertimer" },
    { 0x2120, "createspendhinthud" },
    { 0x2121, "createsquad" },
    { 0x2122, "createstruct" },
    { 0x2123, "createtags" },
    { 0x2124, "createtank" },
    { 0x2126, "createteamobjpoint" },
    { 0x2128, "createteamprogressbartext" },
    { 0x2129, "createtimer" },
    { 0x212A, "createturret" },
    { 0x212F, "createzones" },
    { 0x2130, "credits" },
    { 0x2133, "credits_spacing" },
    { 0x2136, "creepwalk_traverse_over_small" },
    { 0x2138, "creepwalk_traverse_under" },
    { 0x213A, "crew_quarters_combat" },
    { 0x213C, "crew_quarters_crew_killed" },
    { 0x213E, "critchance" },
    { 0x213F, "critialhit" },
    { 0x2141, "critical_factor" },
    { 0x2142, "criticalfactors_awayfromenemies" },
    { 0x2143, "criticalfactors_domination" },
    { 0x2145, "criticalfactors_nearteam" },
    { 0x2146, "criticalfactors_safeguard" },
    { 0x2148, "crnd_bar_amb" },
    { 0x214E, "crossproduct" },
    { 0x2153, "crouchrun_combatanim" },
    { 0x2154, "crouchrun_runnormal" },
    { 0x2155, "crouchrun_runoverride" },
    { 0x2156, "crouchruntocrouch" },
    { 0x2158, "crouchruntopronerun" },
    { 0x2159, "crouchruntopronewalk" },
    { 0x215B, "crouchstop_begin" },
    { 0x215C, "crouchtocrouchrun" },
    { 0x215E, "crouchtoprone" },
    { 0x2160, "crouchtopronewalk" },
    { 0x2161, "crouchtostand" },
    { 0x2167, "crt_plane" },
    { 0x2168, "crush_front_gaz" },
    { 0x2169, "crush_mobile_gaz" },
    { 0x216A, "crush_player_with_floating_lynx" },
    { 0x216B, "crush_rear_gaz" },
    { 0x216C, "crush_stop_sign_when_near_tank" },
    { 0x216F, "cspline_adjusttime" },
    { 0x2170, "cspline_calctangent" },
    { 0x2173, "cspline_findpathnodes" },
    { 0x2174, "cspline_getnodes" },
    { 0x2177, "cspline_initnoise" },
    { 0x217A, "cspline_makenoisepathnodes" },
    { 0x217B, "cspline_makepath" },
    { 0x217C, "cspline_makepath1seg" },
    { 0x217D, "cspline_makepathtopoint" },
    { 0x217F, "cspline_noise" },
    { 0x2180, "cspline_speedfromdistance" },
    { 0x2181, "cspline_test" },
    { 0x2182, "cspline_testnodes" },
    { 0x2183, "cspline_time" },
    { 0x2184, "csplineseg_calccoeffs" },
    { 0x2185, "csplineseg_calccoeffscapspeed" },
    { 0x2186, "csplineseg_calclengthbystepping" },
    { 0x2187, "csplineseg_calctopspeed" },
    { 0x2189, "csplineseg_calctopspeedbystepping" },
    { 0x218B, "csplineseg_getpoint" },
    { 0x218C, "csv_include" },
    { 0x218D, "csv_lines" },
    { 0x218F, "cull_spawners_from_killspawner" },
    { 0x2191, "cull_trigger_logic" },
    { 0x2192, "cur_defend_angle_override" },
    { 0x2194, "cur_defend_point_override" },
    { 0x2195, "cur_defend_stance" },
    { 0x2197, "cur_node" },
    { 0x2198, "cur_node_check_delete" },
    { 0x2199, "curautosave" },
    { 0x219A, "curclass" },
    { 0x219B, "curdefvalue" },
    { 0x219E, "curmeleetarget" },
    { 0x219F, "curorigin" },
    { 0x21A0, "curprogress" },
    { 0x21A1, "curr_betty" },
    { 0x21A2, "curr_interruption" },
    { 0x21A3, "curr_movetime" },
    { 0x21A5, "currectlocindex" },
    { 0x21A7, "currency_scale_per_hive" },
    { 0x21A8, "current_accuracy" },
    { 0x21A9, "current_alien_count" },
    { 0x21AA, "current_anim" },
    { 0x21AB, "current_anim_data_scene" },
    { 0x21AD, "current_attackers" },
    { 0x21AE, "current_audio_zone" },
    { 0x21AF, "current_binocular_zoom_level" },
    { 0x21B0, "current_challenge" },
    { 0x21B1, "current_challenge_exist" },
    { 0x21B2, "current_challenge_index" },
    { 0x21B3, "current_challenge_is" },
    { 0x21B4, "current_challenge_percent" },
    { 0x21B5, "current_challenge_progress_current" },
    { 0x21B6, "current_challenge_progress_max" },
    { 0x21B7, "current_challenge_target_player" },
    { 0x21BA, "current_color_order" },
    { 0x21BC, "current_cycle_num" },
    { 0x21BD, "current_dist_to_top" },
    { 0x21BE, "current_drill_health" },
    { 0x21BF, "current_drill_time" },
    { 0x21C0, "current_end" },
    { 0x21C1, "current_enemy" },
    { 0x21C2, "current_enemy_vol" },
    { 0x21C3, "current_event" },
    { 0x21C4, "current_fallbackers" },
    { 0x21C5, "current_follow_path" },
    { 0x21C6, "current_foot" },
    { 0x21C7, "current_global_hint" },
    { 0x21C8, "current_high_volume" },
    { 0x21C9, "current_hint" },
    { 0x21CB, "current_hive_name" },
    { 0x21CC, "current_intensity" },
    { 0x21CE, "current_interrupt_status" },
    { 0x21D0, "current_lane_monitor" },
    { 0x21D1, "current_mode_hud" },
    { 0x21D3, "current_position" },
    { 0x21D7, "current_sniff_zone" },
    { 0x21D8, "current_speed_percent" },
    { 0x21D9, "current_spit_node" },
    { 0x21DD, "current_sun" },
    { 0x21DE, "current_sunflare_setting" },
    { 0x21DF, "current_swing_pos" },
    { 0x21E0, "current_target" },
    { 0x21E1, "current_target_hud_count" },
    { 0x21E2, "current_vol_acceptable" },
    { 0x21E3, "current_volume" },
    { 0x21E4, "current_volumes" },
    { 0x21E5, "current_yellow_enemy_replenisher" },
    { 0x21EB, "currentanimstate" },
    { 0x21EF, "currentcolorcode" },
    { 0x21F0, "currentcolorforced" },
    { 0x21F1, "currentdialogimportance" },
    { 0x21F4, "currentdodgeanim" },
    { 0x21F5, "currentenemycount" },
    { 0x21F6, "currentfireworkslocation" },
    { 0x21FA, "currentlylocking" },
    { 0x21FC, "currentowner" },
    { 0x21FF, "currentroundnumber" },
    { 0x2200, "currentstate" },
    { 0x2207, "currenttrig" },
    { 0x2208, "currentvehicletarget" },
    { 0x2209, "currentvels" },
    { 0x220A, "currentweapon" },
    { 0x220C, "cursor" },
    { 0x220D, "curtain_dust" },
    { 0x2211, "custom_aim_internal" },
    { 0x2212, "custom_anim_wait" },
    { 0x2215, "custom_balcony_death_animscript" },
    { 0x2216, "custom_bar_enemy_state_spotted" },
    { 0x2219, "custom_battlechatter_init_valid_phrases" },
    { 0x221C, "custom_cornered_stealth_settings" },
    { 0x221F, "custom_death_script" },
    { 0x2220, "custom_dismount_hint_return_when_dismounted" },
    { 0x2222, "custom_fade_out" },
    { 0x2223, "custom_flavor_burst_ent_delete" },
    { 0x2225, "custom_flavor_bursts" },
    { 0x2227, "custom_followpath_parameter_func" },
    { 0x2228, "custom_friendly_fire_message" },
    { 0x2229, "custom_friendly_fire_shader" },
    { 0x222B, "custom_giveloadout" },
    { 0x222C, "custom_hind_death" },
    { 0x2230, "custom_laser_function" },
    { 0x2232, "custom_mount_hint_return_when_mounted" },
    { 0x2233, "custom_no_game_setupfunc" },
    { 0x2234, "custom_player_attacker" },
    { 0x2235, "custom_radius_damage_for_exploders" },
    { 0x2238, "custom_waterfx" },
    { 0x2239, "customanim" },
    { 0x223A, "customarrivalfunc" },
    { 0x223B, "customautosavecheck" },
    { 0x223C, "custombadplacethread" },
    { 0x223D, "custombcs_validphrases" },
    { 0x223E, "custombreathingtime" },
    { 0x223F, "customchatevent" },
    { 0x2244, "customidleanimset" },
    { 0x2245, "customidleanimweights" },
    { 0x2247, "customjuiced" },
    { 0x2248, "custommoveanimset" },
    { 0x224A, "custommovetransitionfunc" },
    { 0x224C, "customprematchperiod" },
    { 0x224D, "customrunningreacttobullets" },
    { 0x224E, "cut_exploders" },
    { 0x224F, "cut_hint" },
    { 0x2257, "cut_window_rorke" },
    { 0x2259, "cw_barracks" },
    { 0x225E, "cw_barracks_setup" },
    { 0x225F, "cw_barracks_slow" },
    { 0x2260, "cw_bloom_above" },
    { 0x2261, "cw_bravo_breach" },
    { 0x2266, "cw_common" },
    { 0x2268, "cw_common_breach_allies" },
    { 0x226A, "cw_common_breach_nag" },
    { 0x226B, "cw_common_breach_player" },
    { 0x226C, "cw_common_breach_trig_proc" },
    { 0x226D, "cw_common_fic" },
    { 0x226F, "cw_fizzle_flashbangs_thinking" },
    { 0x2270, "cw_fog_above" },
    { 0x2271, "cw_fog_under" },
    { 0x2272, "cw_in_rising_water" },
    { 0x2273, "cw_lights_out_ng" },
    { 0x2274, "cw_lights_out_script_1" },
    { 0x2276, "cw_lights_out_script_3" },
    { 0x2278, "cw_lights_out_script_5" },
    { 0x227B, "cw_mid" },
    { 0x227F, "cw_player_damage_underwater_time_exceeded" },
    { 0x2280, "cw_player_drowning_animate" },
    { 0x2281, "cw_player_drowning_damage_count" },
    { 0x2282, "cw_player_in_rising_water" },
    { 0x2283, "cw_player_is_drowning" },
    { 0x2284, "cw_player_make_stand" },
    { 0x2285, "cw_player_underwater_max_blur" },
    { 0x2287, "cw_player_view_fx_source" },
    { 0x2288, "cw_player_view_water_level_fx_source" },
    { 0x2289, "cw_playing_wet_fx" },
    { 0x228B, "cw_previous_longdeath" },
    { 0x228C, "cw_snowmobile_headlight" },
    { 0x228D, "cw_snowmobile_headlight_cg" },
    { 0x228E, "cw_snowmobile_light" },
    { 0x228F, "cw_stairs_fic" },
    { 0x2291, "cw_tape_breach" },
    { 0x2294, "cw_thermite_charge_light" },
    { 0x2295, "cw_trigger_volumes" },
    { 0x2296, "cw_vision_above" },
    { 0x2297, "cw_waterwipe_above" },
    { 0x2299, "cw_waterwipe_under" },
    { 0x229A, "cw_waterwipe_under_playing" },
    { 0x229C, "cycle_begin_intensity_monitor" },
    { 0x229D, "cycle_count" },
    { 0x229F, "cycle_data" },
    { 0x22A4, "cycle_reward_scalar" },
    { 0x22A7, "cycletargets" },
    { 0x22A8, "cypher_defend_self" },
    { 0x22AC, "d_dialogue_queue" },
    { 0x22AD, "dam" },
    { 0x22AE, "dam_break" },
    { 0x22B1, "dam_break_church_spire" },
    { 0x22B5, "dam_break_m880_launch_prep" },
    { 0x22B6, "dam_break_m880_launch_prep_spawn" },
    { 0x22BA, "dam_break_missile_02" },
    { 0x22BB, "dam_break_missile_03" },
    { 0x22BF, "dam_break_street_water" },
    { 0x22C0, "dam_break_street_water_01" },
    { 0x22C1, "dam_break_street_water_02" },
    { 0x22C5, "dam_break_weapon" },
    { 0x22C7, "dam_heli_flyover_hover" },
    { 0x22C8, "dam_start" },
    { 0x22C9, "dam_street_flood_big_splashes_fx" },
    { 0x22CA, "dam_street_flood_church_hits" },
    { 0x22CD, "damage_alien_over_time" },
    { 0x22CF, "damage_done" },
    { 0x22D1, "damage_feedback_setup" },
    { 0x22D2, "damage_functions" },
    { 0x22D4, "damage_hint_bullet_only" },
    { 0x22D6, "damage_hints_cleanup" },
    { 0x22D7, "damage_is_valid_for_friendlyfire_warning" },
    { 0x22D8, "damage_level" },
    { 0x22DA, "damage_not" },
    { 0x22DB, "damage_player" },
    { 0x22DF, "damage_slide_time" },
    { 0x22E1, "damage_smoke_time" },
    { 0x22E2, "damage_state_apply" },
    { 0x22E3, "damage_state_build" },
    { 0x22E4, "damage_state_choose" },
    { 0x22E6, "damage_state_fx_add" },
    { 0x22EB, "damage_state_play_fx_on_tag" },
    { 0x22EF, "damage_state_stop_fx_on_tag" },
    { 0x22F0, "damage_state_tag_ent_clear_all" },
    { 0x22F4, "damage_time" },
    { 0x22F5, "damage_time_player" },
    { 0x22F7, "damage_type" },
    { 0x22F8, "damage_vehicle_think" },
    { 0x22F9, "damage_vehicles_path1" },
    { 0x22FA, "damage_vehicles_path2" },
    { 0x22FB, "damage_watcher_ambush" },
    { 0x22FE, "damageattacker" },
    { 0x2300, "damagecallback" },
    { 0x2303, "damaged_crate" },
    { 0x2304, "damagedplayer" },
    { 0x2308, "damagefeedback" },
    { 0x2309, "damagefeedback_took_damage" },
    { 0x230B, "damageinfocolorindex" },
    { 0x230F, "damagelistsize" },
    { 0x2312, "damageorigin" },
    { 0x2314, "damagepoint" },
    { 0x2316, "damageshieldcounter" },
    { 0x2318, "damageshieldpain" },
    { 0x231B, "damping_factor" },
    { 0x231D, "dangermaxradius" },
    { 0x231E, "dangerminradius" },
    { 0x231F, "dangerovalscale" },
    { 0x2320, "dangersprinttime" },
    { 0x2321, "dart" },
    { 0x2323, "dart_breach_precache" },
    { 0x2324, "dash_wipe" },
    { 0x2325, "data" },
    { 0x2327, "data_value_suffix" },
    { 0x2329, "dc_old_moveplaybackrate" },
    { 0x232A, "dc_shotgun_seek" },
    { 0x232C, "ddtimetoadd" },
    { 0x232D, "deactivate" },
    { 0x232E, "deactivate_avoid_minion_exp" },
    { 0x232F, "deactivate_current_challenge" },
    { 0x2333, "deactivate_kill_10_with_turrets" },
    { 0x2334, "deactivate_kill_airborne_aliens" },
    { 0x2336, "deactivate_melee_goons" },
    { 0x2337, "deactivate_melee_only" },
    { 0x2339, "deactivate_occlusion" },
    { 0x233A, "deactivate_percent_accuracy" },
    { 0x233B, "deactivate_protect_a_player" },
    { 0x2347, "dead_guy" },
    { 0x234B, "dead_hive_model" },
    { 0x234C, "dead_pilot_hang" },
    { 0x234E, "deadly_sharks" },
    { 0x2350, "death_a" },
    { 0x2351, "death_achievements" },
    { 0x2352, "death_achievements_rappel" },
    { 0x2353, "death_anim" },
    { 0x2354, "death_anim_anytime" },
    { 0x2355, "death_anim_played" },
    { 0x2358, "death_d" },
    { 0x2359, "death_delayed_ragdoll" },
    { 0x235A, "death_effect" },
    { 0x235B, "death_firesound" },
    { 0x235C, "death_func" },
    { 0x235F, "death_getincomingdirection" },
    { 0x2360, "death_model_col" },
    { 0x2364, "death_only_ragdoll" },
    { 0x2365, "death_override" },
    { 0x2366, "death_waiter" },
    { 0x2367, "death_warning_vo" },
    { 0x2368, "death_watcher" },
    { 0x2369, "deathanim" },
    { 0x236B, "deathanimscript" },
    { 0x236C, "deathchain_goalvolume" },
    { 0x236D, "deathchainainotify" },
    { 0x236E, "deathchainspawnerlogic" },
    { 0x2372, "deathdamagemax" },
    { 0x2374, "deathdamageradius" },
    { 0x2375, "deathflags" },
    { 0x2376, "deathfunc" },
    { 0x2379, "deathfunc_vol_num_decrement" },
    { 0x237A, "deathfuncs" },
    { 0x237D, "deathfx_ent" },
    { 0x237E, "deathoverridecallback" },
    { 0x2380, "deathquote_on_death" },
    { 0x2381, "deathrolloff" },
    { 0x2383, "deathscript" },
    { 0x2385, "deathsdoor_enabled" },
    { 0x2388, "deathspawnerents" },
    { 0x2389, "deathspawnerpreview" },
    { 0x238A, "deathstring" },
    { 0x238E, "deathvfx" },
    { 0x238F, "deathwatcher" },
    { 0x2391, "debgugprintdestructiblelist" },
    { 0x2392, "debris" },
    { 0x2393, "debris_bridge_actor_vign_and_transition_to_combat" },
    { 0x2394, "debris_bridge_allies_loop" },
    { 0x2396, "debris_bridge_bus_sparks" },
    { 0x2397, "debris_bridge_cleanup" },
    { 0x2399, "debris_bridge_fx" },
    { 0x239A, "debris_bridge_loop1" },
    { 0x239C, "debris_bridge_sfx" },
    { 0x23A0, "debris_bridge_vign2_and_loop3" },
    { 0x23A1, "debris_bridge_vign2_and_loop3_ally" },
    { 0x23A2, "debris_chunk" },
    { 0x23A4, "debris_remove_after_time" },
    { 0x23A5, "debris_spawner" },
    { 0x23A8, "debrisbridge_cleanup" },
    { 0x23A9, "debrisbridge_clear_enemies_bottom" },
    { 0x23AA, "debrisbridge_clear_enemies_top" },
    { 0x23AC, "debrisbridge_crossing" },
    { 0x23AE, "debrisbridge_enemy_aggrisive_logic" },
    { 0x23AF, "debrisbridge_enemy_logic" },
    { 0x23B1, "debrisbridge_enemy_spawn_logic" },
    { 0x23B2, "debrisbridge_fodder" },
    { 0x23B3, "debrisbridge_fodder_extra" },
    { 0x23B4, "debrisbridge_hide_glass_parts" },
    { 0x23B5, "debrisbridge_kill_enemies_top" },
    { 0x23B6, "debrisbridge_loc" },
    { 0x23B7, "debrisbridge_move_trigger" },
    { 0x23B9, "debrisbridge_origins" },
    { 0x23BA, "debrisbridge_path_logic" },
    { 0x23BB, "debrisbridge_prevent_frogger" },
    { 0x23BC, "debrisbridge_setup_ally_for_kill_shot" },
    { 0x23BD, "debrisbridge_setup_enemies_for_clearance" },
    { 0x23BE, "debrisbridge_shot_count" },
    { 0x23C0, "debrisbridge_start" },
    { 0x23C1, "debrisbridge_wall_break_logic" },
    { 0x23C2, "debrisbridge_water_enter_combat_space" },
    { 0x23C6, "debug" },
    { 0x23C7, "debug_ai_drone_amounts" },
    { 0x23C8, "debug_ai_drone_amounts_logic" },
    { 0x23C9, "debug_anim" },
    { 0x23CB, "debug_animsoundtagselected" },
    { 0x23CD, "debug_arrivals_on_actor" },
    { 0x23CE, "debug_audio_hud" },
    { 0x23CF, "debug_bike_line" },
    { 0x23D2, "debug_circle" },
    { 0x23D5, "debug_color_huds" },
    { 0x23D6, "debug_color_system_hud" },
    { 0x23D8, "debug_corner" },
    { 0x23DA, "debug_cross" },
    { 0x23DB, "debug_death" },
    { 0x23DE, "debug_draw_arrow" },
    { 0x23E6, "debug_enemyposproc" },
    { 0x23E7, "debug_enemyposreplay" },
    { 0x23E8, "debug_error" },
    { 0x23E9, "debug_friendlyfire" },
    { 0x23EA, "debug_function" },
    { 0x23EC, "debug_fxlighting_buttons" },
    { 0x23ED, "debug_heli_gun" },
    { 0x23EF, "debug_hud_disabled" },
    { 0x23F1, "debug_j_prop" },
    { 0x23F3, "debug_jolt_vehs" },
    { 0x23F4, "debug_kill_enemies_in_order" },
    { 0x23F5, "debug_line" },
    { 0x23F6, "debug_lookflag" },
    { 0x23F7, "debug_message" },
    { 0x23FA, "debug_no_move" },
    { 0x23FC, "debug_org" },
    { 0x23FE, "debug_player_in_post_clip" },
    { 0x2400, "debug_print3d" },
    { 0x2401, "debug_print3d_simple" },
    { 0x2403, "debug_radiusdamage_circle" },
    { 0x2404, "debug_reflection" },
    { 0x2405, "debug_reflection_buttons" },
    { 0x2408, "debug_show_ai_counts" },
    { 0x240A, "debug_spawn_director_active" },
    { 0x240C, "debug_stopenemypos" },
    { 0x240E, "debug_test_thruster_audio" },
    { 0x2410, "debug_thruster_text" },
    { 0x2411, "debug_timer" },
    { 0x2415, "debug_wait" },
    { 0x2416, "debug_warning" },
    { 0x2417, "debug_wibble" },
    { 0x2419, "debugchains" },
    { 0x241A, "debugcircle" },
    { 0x241B, "debugcolorfriendlies" },
    { 0x241E, "debuggoalpos" },
    { 0x2421, "debugleft" },
    { 0x2423, "debugmisstime" },
    { 0x2426, "debugpos" },
    { 0x2427, "debugposinternal" },
    { 0x2428, "debugpossize" },
    { 0x242A, "debugprint3d" },
    { 0x242B, "debugprintln2" },
    { 0x242C, "debugsightline" },
    { 0x242D, "debugspawners" },
    { 0x242E, "debugthreat" },
    { 0x242F, "debugthreatcalc" },
    { 0x2430, "debugtimeout" },
    { 0x2431, "debugvar" },
    { 0x2433, "decal" },
    { 0x2435, "deceleration" },
    { 0x2436, "decidenumcrawls" },
    { 0x243B, "deck_ac130_dmg_badplace_size" },
    { 0x243C, "deck_ac130_dmg_clip" },
    { 0x2441, "deck_combat_vo" },
    { 0x2443, "deck_combat_wave1_flank" },
    { 0x2444, "deck_combat_wave1_helis" },
    { 0x2446, "deck_combat_wave2_helis" },
    { 0x2448, "deck_combat_wave3_ambient_jets" },
    { 0x2449, "deck_combat_wave3_helis" },
    { 0x244A, "deck_damage" },
    { 0x244B, "deck_destroyed_odin" },
    { 0x244C, "deck_explode" },
    { 0x244D, "deck_explode_guys" },
    { 0x244E, "deck_explode_vista" },
    { 0x244F, "deck_tilt_pre_load" },
    { 0x2450, "deck_tilt_water_kill_trigger" },
    { 0x2452, "deck_transition_cleanup_vol" },
    { 0x2453, "deck_transition_pre_load" },
    { 0x2454, "deck_victory_pre_load" },
    { 0x2456, "decomp_anim_line_1" },
    { 0x2457, "decomp_anim_line_2" },
    { 0x2458, "decomp_anim_notify_done" },
    { 0x2459, "decomp_anim_notify_end_slomo" },
    { 0x245A, "decomp_anim_notify_player_link" },
    { 0x245B, "decomp_anim_notify_slomo" },
    { 0x245D, "decomp_explosion_anim_ally" },
    { 0x245E, "decomp_explosion_anim_enemies" },
    { 0x245F, "decomp_explosion_anim_player" },
    { 0x2460, "decomp_explosion_anim_player_legs" },
    { 0x2462, "decomp_heavy_rumble" },
    { 0x2464, "decreasexpboost" },
    { 0x2465, "decrement_list_offset" },
    { 0x2466, "decrementalivecount" },
    { 0x2468, "decrementcolorusers" },
    { 0x246A, "decrementfauxvehiclecount" },
    { 0x246B, "decrementing_slide" },
    { 0x246C, "decrementlevelbreaks" },
    { 0x246D, "deep_array_call" },
    { 0x246E, "deep_array_thread" },
    { 0x246F, "deer" },
    { 0x2471, "deer_anims" },
    { 0x2474, "deer_detects_when_to_run" },
    { 0x2475, "deer_drone_custom_idle" },
    { 0x2477, "deer_drone_spawn_func" },
    { 0x2478, "deer_dronespawn" },
    { 0x2479, "deer_dronespawn_internal" },
    { 0x247B, "deer_fade_in" },
    { 0x247C, "deer_init" },
    { 0x247D, "deer_player_aim_detection" },
    { 0x247E, "deer_player_leaning_detect" },
    { 0x247F, "deer_reveal_chairs" },
    { 0x2481, "deer_ruckus_trig_logic" },
    { 0x2482, "deer_rumble_movewait" },
    { 0x2483, "deer_stampede_logic" },
    { 0x2484, "deerhunttransitionanim" },
    { 0x2485, "def" },
    { 0x2486, "def_fire" },
    { 0x248B, "default_canactivatefunc" },
    { 0x248D, "default_dof" },
    { 0x248E, "default_dog_limits" },
    { 0x248F, "default_door_node_flashbang_frequency" },
    { 0x2490, "default_drop_pitch" },
    { 0x2491, "default_failactivatefunc" },
    { 0x2492, "default_failfunc" },
    { 0x2494, "default_fov" },
    { 0x2496, "default_goalheight" },
    { 0x2498, "default_heartbeat_rate" },
    { 0x2499, "default_init" },
    { 0x249B, "default_loadout_if_notset" },
    { 0x249D, "default_mg_drone" },
    { 0x249F, "default_node" },
    { 0x24A0, "default_ondeadevent" },
    { 0x24A2, "default_ononeleftevent" },
    { 0x24A4, "default_resetsuccess" },
    { 0x24A6, "default_run_speed" },
    { 0x24A7, "default_should_update" },
    { 0x24A8, "default_sprint" },
    { 0x24A9, "default_start" },
    { 0x24AC, "default_stealth_override" },
    { 0x24AD, "default_stealth_settings" },
    { 0x24B0, "default_unresolved_collision_handler" },
    { 0x24B1, "default_visionset" },
    { 0x24B2, "default_wander_radius" },
    { 0x24B4, "defaultbodymodels" },
    { 0x24B5, "defaultclass" },
    { 0x24B6, "defaultdroneragdoll" },
    { 0x24B7, "defaultdroppitch" },
    { 0x24B8, "defaultdropyaw" },
    { 0x24B9, "defaultemissive" },
    { 0x24BA, "defaultexception" },
    { 0x24BB, "defaultheadmodels" },
    { 0x24BE, "defaultmoveplaybackrate" },
    { 0x24BF, "defaultonmode" },
    { 0x24C0, "defaults" },
    { 0x24C1, "defaulttarget" },
    { 0x24C2, "defaultturnthreshold" },
    { 0x24C6, "defaultweapon" },
    { 0x24C7, "defconkillstreakthread" },
    { 0x24C8, "defconkillstreakwait" },
    { 0x24C9, "defconmode" },
    { 0x24CA, "defconpointmod" },
    { 0x24CB, "defconsplashnotify" },
    { 0x24CE, "defend_combat" },
    { 0x24D2, "defend_door_explosion" },
    { 0x24D3, "defend_door_open" },
    { 0x24D7, "defend_fire" },
    { 0x24DA, "defend_intro" },
    { 0x24DB, "defend_platform" },
    { 0x24DC, "defend_quick" },
    { 0x24DD, "defend_save_safe" },
    { 0x24DF, "defend_sparrow_control" },
    { 0x24E0, "defend_sparrow_fx" },
    { 0x24E1, "defend_sparrow_pre_load" },
    { 0x24E2, "defend_sparrow_ships" },
    { 0x24E3, "defend_start" },
    { 0x24E4, "defend_start_player_shield" },
    { 0x24E5, "defend_valid_center" },
    { 0x24E9, "defend_zodiac_arrived_catwalk" },
    { 0x24EA, "defend_zodiac_arrived_right" },
    { 0x24EB, "defend_zodiac_autosave" },
    { 0x24ED, "defend_zodiac_osprey" },
    { 0x24F0, "defend_zodiac_spawn_hidden_reinforcement" },
    { 0x24F1, "defend_zodiac_spawn_reinforcement" },
    { 0x24F2, "defend_zodiac_waittill_enemies_remaining" },
    { 0x24F3, "defendedplayer" },
    { 0x24F4, "defendloc" },
    { 0x24F5, "defendlocation" },
    { 0x24F6, "defends" },
    { 0x24F7, "defense_cautious_approach" },
    { 0x24F8, "defense_death_monitor" },
    { 0x24F9, "defense_force_next_node_goal" },
    { 0x24FA, "defense_get_initial_entrances" },
    { 0x24FC, "defense_override_entrances" },
    { 0x24FF, "define_loadout" },
    { 0x2500, "definechestweapons" },
    { 0x2502, "defuseendtime" },
    { 0x2503, "defuses" },
    { 0x2504, "defusetime" },
    { 0x2505, "delay" },
    { 0x2506, "delay_drop_boat" },
    { 0x2508, "delay_keegan_color" },
    { 0x2509, "delay_multi_fx" },
    { 0x250A, "delay_multi_fx_proc" },
    { 0x250C, "delay_remove_invul" },
    { 0x250D, "delay_reset_swim_shock" },
    { 0x2511, "delay_show_legs_entry" },
    { 0x2512, "delay_with_bubbles" },
    { 0x2513, "delaybetweenlockon" },
    { 0x2514, "delaybetweenrockets" },
    { 0x2515, "delaycall" },
    { 0x2516, "delaycall_proc" },
    { 0x2517, "delaychildthread" },
    { 0x2518, "delaychildthread_proc" },
    { 0x2519, "delaydestroy" },
    { 0x251A, "delayed_delete" },
    { 0x251C, "delayed_intro" },
    { 0x251D, "delayed_kill" },
    { 0x251E, "delayed_player_seek_think" },
    { 0x251F, "delayed_setgoalvolumeauto" },
    { 0x2520, "delayed_setup" },
    { 0x2522, "delayedbadplace" },
    { 0x2524, "delayedexception" },
    { 0x2525, "delayedfofoverlay" },
    { 0x2528, "delayer" },
    { 0x2529, "delayhudoutline" },
    { 0x252E, "delaytankunload" },
    { 0x2530, "delaythread_nuke" },
    { 0x2531, "delaythread_proc" },
    { 0x2532, "deletable_magic_bullet_shield" },
    { 0x2534, "delete_accessories_on_death" },
    { 0x2538, "delete_ai_at_path_end" },
    { 0x2539, "delete_all_by_type" },
    { 0x253A, "delete_all_ents" },
    { 0x253B, "delete_all_grenades" },
    { 0x253C, "delete_all_triggers" },
    { 0x253E, "delete_at_anim_end" },
    { 0x2540, "delete_beginning_ents" },
    { 0x2543, "delete_building_glow" },
    { 0x2544, "delete_corpse_in_volume" },
    { 0x2546, "delete_corpses_around_vehicle" },
    { 0x2547, "delete_cracked_glass" },
    { 0x2548, "delete_destructibles_in_volumes" },
    { 0x2549, "delete_ent" },
    { 0x254A, "delete_ent_by_script_noteworthy" },
    { 0x254C, "delete_ents" },
    { 0x2552, "delete_fish_in_volume" },
    { 0x2553, "delete_glass_cutter" },
    { 0x2554, "delete_hazmat_guns" },
    { 0x2556, "delete_if_player_cant_see_me" },
    { 0x2557, "delete_intel" },
    { 0x2558, "delete_interactives_in_volumes" },
    { 0x2559, "delete_light_on_clearup" },
    { 0x255A, "delete_links_then_self" },
    { 0x255B, "delete_me" },
    { 0x255C, "delete_me_after_time" },
    { 0x255E, "delete_me_on_parent_ai_death" },
    { 0x2560, "delete_moving_cover_objects" },
    { 0x2564, "delete_on_crack1_done" },
    { 0x2565, "delete_on_crack2_done" },
    { 0x2566, "delete_on_death" },
    { 0x2567, "delete_on_death_wait_sound" },
    { 0x2568, "delete_on_end" },
    { 0x2569, "delete_on_flag" },
    { 0x256B, "delete_on_house_done" },
    { 0x256D, "delete_on_notify" },
    { 0x256E, "delete_on_path_end" },
    { 0x256F, "delete_on_removed" },
    { 0x2570, "delete_on_sounddone" },
    { 0x2571, "delete_path_clip" },
    { 0x2574, "delete_removables" },
    { 0x2577, "delete_sentry_turret" },
    { 0x2578, "delete_stage_one_fx" },
    { 0x2579, "delete_start" },
    { 0x257A, "delete_temp_sat_view_targets" },
    { 0x257D, "delete_trolley_ally" },
    { 0x257E, "delete_volmod_hud" },
    { 0x257F, "delete_wand_at_lights_out" },
    { 0x2580, "delete_when_behind_player" },
    { 0x2584, "deletecrate" },
    { 0x2585, "deletedebugtexthud" },
    { 0x2587, "deletedestructiblekillcament" },
    { 0x2589, "deleteexplosive" },
    { 0x258B, "deleteifnotused" },
    { 0x258C, "deletelethalequipmentondisconnect" },
    { 0x258D, "deleteobjpoint" },
    { 0x258E, "deleteonac130playerremoved" },
    { 0x258F, "deleteondeath" },
    { 0x2590, "deleteonentnotify" },
    { 0x2592, "deleteonreviveordeathordisconnect" },
    { 0x2594, "deletepickupafterawhile" },
    { 0x2595, "deleteplacedentity" },
    { 0x2596, "deletereviveicon" },
    { 0x2599, "deletestruct_ref" },
    { 0x259A, "deletestructarray" },
    { 0x259B, "deletestructarray_ref" },
    { 0x259C, "deleteti" },
    { 0x259D, "deleteuseent" },
    { 0x259E, "deleteuseobject" },
    { 0x25A1, "demo_setup_allies" },
    { 0x25A2, "demo_setup_sound" },
    { 0x25A3, "demo_switch_checkpoints" },
    { 0x25A5, "denied" },
    { 0x25A6, "dependent_hives_removed" },
    { 0x25A7, "deploy_currency" },
    { 0x25A8, "deploy_health_regen_shield" },
    { 0x25A9, "deployable" },
    { 0x25AB, "deployable_ammo_placed_listener" },
    { 0x25AC, "deployable_armor_placed_listener" },
    { 0x25AF, "deployable_box_onuse_message" },
    { 0x25B0, "deployable_currency_placed_listener" },
    { 0x25B3, "deployable_randombox_placed_listener" },
    { 0x25B4, "deployable_specialammo_placed_listener" },
    { 0x25B5, "deployable_tutorial_given" },
    { 0x25B6, "deployablebox_adrenalinebox_max" },
    { 0x25B7, "deployablebox_adrenalinebox_rank" },
    { 0x25B9, "deployablebox_juicebox_rank" },
    { 0x25BD, "deployablegunbox_bonusinxuses" },
    { 0x25C3, "depth" },
    { 0x25C6, "depth_allow_stand" },
    { 0x25C8, "depth_charge_check_failure" },
    { 0x25C9, "depth_charge_clear_hud" },
    { 0x25CA, "depth_charge_control_down" },
    { 0x25CB, "depth_charge_control_up" },
    { 0x25CD, "depth_charge_current_slow_pitch" },
    { 0x25D0, "depth_charge_death_trig" },
    { 0x25D2, "depth_charge_default_pitch" },
    { 0x25D3, "depth_charge_dialogue" },
    { 0x25D4, "depth_charge_drop" },
    { 0x25D7, "depth_charge_exit" },
    { 0x25D9, "depth_charge_firing_slow_aim_modifier" },
    { 0x25DB, "depth_charge_friendly_fire_kills_to_fail" },
    { 0x25DC, "depth_charge_give_control" },
    { 0x25DF, "depth_charge_hint" },
    { 0x25E2, "depth_charge_init" },
    { 0x25E8, "depth_charge_max_slow_aim_pitch" },
    { 0x25EA, "depth_charge_min_fov" },
    { 0x25EC, "depth_charge_min_slow_aim_yaw" },
    { 0x25EE, "depth_charge_monitor_drop" },
    { 0x25F0, "depth_charge_org" },
    { 0x25F1, "depth_charge_pitch_down_allowance" },
    { 0x25F3, "depth_charge_remove_control" },
    { 0x25F4, "depth_charge_replay" },
    { 0x25F6, "depth_charge_set_slow_aim" },
    { 0x25F7, "depth_charge_start_time" },
    { 0x25F8, "depth_charge_target" },
    { 0x25F9, "depth_charge_turn" },
    { 0x25FB, "depth_charge_use" },
    { 0x25FD, "depth_charges" },
    { 0x25FF, "depth_charges_setup" },
    { 0x2600, "depth_marker" },
    { 0x2601, "depthenable" },
    { 0x2603, "derrick_metal_debris_sfx" },
    { 0x2604, "derrick_model" },
    { 0x2605, "derrick_pop_and_explode" },
    { 0x2606, "derrick_scene_struct" },
    { 0x2607, "derrick_struct" },
    { 0x2608, "desc" },
    { 0x260C, "deselect_entity" },
    { 0x260D, "designator_disableusabilityduringgrenadepullback" },
    { 0x260E, "designator_enableusabilitywhendesignatorfinishes" },
    { 0x260F, "designator_ontargetacquired" },
    { 0x2611, "designator_start" },
    { 0x2612, "designator_waitforgrenadefire" },
    { 0x2615, "desired_anim_pose" },
    { 0x2616, "desired_bob_pitch" },
    { 0x2617, "desired_bob_roll" },
    { 0x2618, "desired_speed" },
    { 0x2619, "desiredtimeofdeath" },
    { 0x261A, "desk_animate" },
    { 0x261D, "dest_tree_test" },
    { 0x261E, "destination" },
    { 0x261F, "destinations" },
    { 0x2621, "destroy_all_hangar_walls_wait" },
    { 0x2622, "destroy_breach_mines_and_fx" },
    { 0x2624, "destroy_corner" },
    { 0x2627, "destroy_hint" },
    { 0x2628, "destroy_hint_on_friendlyfire" },
    { 0x2629, "destroy_hive_icon" },
    { 0x262A, "destroy_lip_debris_fx" },
    { 0x262B, "destroy_persistent_ice_breach_props" },
    { 0x262F, "destroy_player_ambush_vo" },
    { 0x2633, "destroy_warning_elem" },
    { 0x263A, "destroyactiveimss" },
    { 0x263B, "destroyactivelittlebirds" },
    { 0x263C, "destroyactiverockets" },
    { 0x263D, "destroyactiveturrets" },
    { 0x263E, "destroyactiveuavs" },
    { 0x263F, "destroyactiveugvs" },
    { 0x2641, "destroyed" },
    { 0x2642, "destroyed_derrick_models" },
    { 0x2643, "destroyed_fire_fx" },
    { 0x2644, "destroyed_hesco_count" },
    { 0x2645, "destroyed_tank" },
    { 0x2647, "destroyedcounttimeout" },
    { 0x2648, "destroyedmodel" },
    { 0x264A, "destroyefx" },
    { 0x264B, "destroyelem" },
    { 0x264D, "destroyer_ac130_exchange" },
    { 0x2650, "destroyer_mg_monitor" },
    { 0x2651, "destroyer_target" },
    { 0x2652, "destroyer_targets_big" },
    { 0x2655, "destroyheadiconsondeath" },
    { 0x2659, "destroyslowly" },
    { 0x265E, "destructable_think" },
    { 0x265F, "destructible" },
    { 0x2663, "destructible_attachmodel" },
    { 0x2667, "destructible_builddot_startloop" },
    { 0x2668, "destructible_builddot_wait" },
    { 0x2669, "destructible_car_alarm" },
    { 0x266A, "destructible_civilian_sedan_iw6" },
    { 0x266B, "destructible_cleans_up_more" },
    { 0x266C, "destructible_create" },
    { 0x266E, "destructible_createdot_radius" },
    { 0x266F, "destructible_damage_threshold" },
    { 0x2670, "destructible_death" },
    { 0x2672, "destructible_dots" },
    { 0x2673, "destructible_dragons_teeth" },
    { 0x2679, "destructible_function" },
    { 0x267C, "destructible_fx_think" },
    { 0x2680, "destructible_handles_collision_brushes" },
    { 0x2681, "destructible_health_drain_amount_multiplier" },
    { 0x2683, "destructible_info" },
    { 0x2684, "destructible_lights_out" },
    { 0x2685, "destructible_loopfx" },
    { 0x268C, "destructible_parts" },
    { 0x268D, "destructible_physics" },
    { 0x268E, "destructible_protection_func" },
    { 0x268F, "destructible_quakes_off" },
    { 0x2690, "destructible_quakes_on" },
    { 0x2693, "destructible_setdot_ontickfunc" },
    { 0x2694, "destructible_setup" },
    { 0x2696, "destructible_sound_think" },
    { 0x269A, "destructible_spotlight" },
    { 0x269B, "destructible_spotlight_think" },
    { 0x269C, "destructible_state" },
    { 0x269E, "destructible_transient" },
    { 0x269F, "destructible_update_part" },
    { 0x26A0, "destructible_van_iw6" },
    { 0x26A4, "destructibles" },
    { 0x26A5, "destructiblespawnedents" },
    { 0x26A9, "destructions" },
    { 0x26AA, "destructisolarpanelsinit" },
    { 0x26AC, "detach_cig" },
    { 0x26AD, "detach_getoutrigs" },
    { 0x26AE, "detach_hatchet" },
    { 0x26AF, "detach_idle_clip" },
    { 0x26B0, "detach_models_with_substr" },
    { 0x26B1, "detach_phone" },
    { 0x26B2, "detach_pistol_left" },
    { 0x26B3, "detach_pistol_right" },
    { 0x26B4, "detach_rope_ally" },
    { 0x26B7, "detach_rope_rorke" },
    { 0x26B8, "detachallweaponmodels" },
    { 0x26BA, "detachifattached" },
    { 0x26BB, "detachprops" },
    { 0x26BC, "detachusemodels" },
    { 0x26BF, "detect_distsqrd" },
    { 0x26C1, "detect_people" },
    { 0x26C2, "detect_people_trigger" },
    { 0x26C3, "detect_player_death" },
    { 0x26C4, "detect_player_event" },
    { 0x26C6, "detect_range" },
    { 0x26CA, "detectexplosives" },
    { 0x26CB, "detectfriendlyfireonentity" },
    { 0x26CD, "detectid" },
    { 0x26CE, "detection_level" },
    { 0x26CF, "detection_timeout" },
    { 0x26D0, "detectiongraceperiod" },
    { 0x26D2, "detectionradius" },
    { 0x26D3, "detectkill" },
    { 0x26D4, "determine_closest_cars" },
    { 0x26D6, "determinecqbanim" },
    { 0x26D7, "determineexposedapproachtype" },
    { 0x26D8, "determineheatcoverexittype" },
    { 0x26D9, "determinenodeapproachtype" },
    { 0x26DA, "determinenodeexittype" },
    { 0x26DB, "determinenonnodeexittype" },
    { 0x26DD, "detonate_dominator" },
    { 0x26E1, "detonateonuse" },
    { 0x26E4, "detoured" },
    { 0x26E5, "detouringpath" },
    { 0x26E6, "detpackstunradius" },
    { 0x26E7, "devaliengiveplayersmoney" },
    { 0x26E8, "device" },
    { 0x26EA, "dialog" },
    { 0x26EB, "dialog_baker_init_control" },
    { 0x26EC, "dialog_baker_player_control" },
    { 0x26ED, "dialog_baker_start" },
    { 0x26EF, "dialog_explode" },
    { 0x26F0, "dialog_found_a_body" },
    { 0x26F1, "dialog_jump" },
    { 0x26F2, "dialog_line_with_flag" },
    { 0x26F5, "dialog_nags_heli" },
    { 0x26F8, "dialog_radiobravo" },
    { 0x26F9, "dialog_seehelo" },
    { 0x26FB, "dialogue" },
    { 0x26FD, "dialogue_baker_waitforit" },
    { 0x2700, "dialogue_end" },
    { 0x2704, "dialogue_intro" },
    { 0x2705, "dialogue_mantle" },
    { 0x2706, "dialogue_mgs_1" },
    { 0x2708, "dialogue_mudpumps" },
    { 0x270A, "dialogue_outro" },
    { 0x270C, "dialogue_queue_single" },
    { 0x270D, "dialogue_random_last_line" },
    { 0x2710, "dialogue_sat1" },
    { 0x2711, "dialogue_sat2" },
    { 0x2712, "dialogue_streets_to_dam" },
    { 0x2713, "dialogue_streets_to_dam_2" },
    { 0x2714, "dialogue_tanks" },
    { 0x2715, "dialoguenotetrack" },
    { 0x2716, "diaz_door_kick_sfx" },
    { 0x2717, "did_flares" },
    { 0x2718, "did_handsignal" },
    { 0x2719, "did_inactive_vo" },
    { 0x271B, "diddamagestate" },
    { 0x271C, "didpastmeleefail" },
    { 0x271D, "didpastpursuitfail" },
    { 0x271F, "didsomethingotherthanshooting" },
    { 0x2722, "die_and_ragdoll" },
    { 0x2723, "die_from_explosion" },
    { 0x2724, "die_hard_explosion_fx" },
    { 0x2725, "die_quietly" },
    { 0x2727, "died_of_headshot" },
    { 0x2728, "diehardmode" },
    { 0x272A, "dieingrevival_vo" },
    { 0x272F, "difficultystring" },
    { 0x2730, "difficultytype" },
    { 0x2731, "diffusefraction_ng" },
    { 0x2734, "digitalflash" },
    { 0x2737, "diraimlimit" },
    { 0x2738, "direct_chopper_crate_anim" },
    { 0x2739, "direction" },
    { 0x273A, "direction_change_smoothing" },
    { 0x273E, "dirt_on_screen_from_position" },
    { 0x2742, "disable_achievement_harder_they_fall_guy" },
    { 0x2746, "disable_all_triggers" },
    { 0x2749, "disable_arrivals" },
    { 0x274A, "disable_arrivals_and_exits" },
    { 0x274D, "disable_blindfire" },
    { 0x274F, "disable_breathing_sound" },
    { 0x2750, "disable_bulletwhizbyreaction" },
    { 0x2751, "disable_careful" },
    { 0x2752, "disable_checkpoint_patrol" },
    { 0x2754, "disable_control" },
    { 0x2755, "disable_cqb_points_of_interest" },
    { 0x2756, "disable_cqbwalk" },
    { 0x2757, "disable_damagefeedback" },
    { 0x2758, "disable_danger_react" },
    { 0x275A, "disable_demigod" },
    { 0x275D, "disable_dog_control" },
    { 0x275F, "disable_dog_sneak" },
    { 0x2760, "disable_dog_sniff" },
    { 0x2761, "disable_dog_walk" },
    { 0x2763, "disable_dontevershoot" },
    { 0x2764, "disable_dynamic_run_speed" },
    { 0x2765, "disable_exits" },
    { 0x2767, "disable_flinch" },
    { 0x2768, "disable_fx_on_death" },
    { 0x2769, "disable_gun_recall" },
    { 0x276A, "disable_health_regen_shield" },
    { 0x276B, "disable_heat_behavior" },
    { 0x276C, "disable_hudoutline_on_death" },
    { 0x276D, "disable_ignorerandombulletdamage_drone" },
    { 0x276F, "disable_killcam" },
    { 0x2772, "disable_militia_behavior" },
    { 0x2773, "disable_nodes_before_allies_move" },
    { 0x2775, "disable_oneshotfx_with_noteworthy" },
    { 0x2777, "disable_pain" },
    { 0x2779, "disable_player_swim" },
    { 0x277A, "disable_readystand" },
    { 0x277D, "disable_sniper_outline" },
    { 0x2782, "disable_sprint" },
    { 0x2783, "disable_stealth" },
    { 0x2787, "disable_strikezone_rog" },
    { 0x2789, "disable_swim" },
    { 0x278A, "disable_target_on_death" },
    { 0x278C, "disable_teamflashbangimmunity" },
    { 0x278E, "disable_tired" },
    { 0x278F, "disable_torpedo_ui" },
    { 0x2790, "disable_trigger" },
    { 0x2791, "disable_trigger_helper" },
    { 0x2792, "disable_trigger_with_noteworthy" },
    { 0x2794, "disable_turnanims" },
    { 0x2795, "disable_turns_arrivals_exits" },
    { 0x2799, "disableallstreaks" },
    { 0x279C, "disablebattlechatter" },
    { 0x279E, "disablebutton" },
    { 0x279F, "disableclientspawntraces" },
    { 0x27A0, "disablecollision" },
    { 0x27A1, "disablecoverarrivalsonly" },
    { 0x27A4, "disabled_use_for" },
    { 0x27A5, "disabledamageshieldpain" },
    { 0x27A6, "disabledefaultfacialanims" },
    { 0x27A7, "disabledoffhandweapons" },
    { 0x27AA, "disabledweapon" },
    { 0x27AF, "disablegeardrop" },
    { 0x27B0, "disablegrenadetracking" },
    { 0x27B3, "disablelongdeath" },
    { 0x27B6, "disablemovementtracker" },
    { 0x27B8, "disablepain" },
    { 0x27C0, "disablevehiclescripts" },
    { 0x27C2, "disablewhenjuggernaut" },
    { 0x27C3, "discardtime" },
    { 0x27C4, "disconnect_and_bad_place" },
    { 0x27C5, "disconnect_path_periodic" },
    { 0x27C6, "disconnect_paths" },
    { 0x27C8, "disconnect_paths_whenstopped" },
    { 0x27CD, "disconnecttraverses" },
    { 0x27D0, "discrete_waittill" },
    { 0x27D1, "disguise_head" },
    { 0x27D3, "disomount_hint" },
    { 0x27D5, "dispatch_activated_zone" },
    { 0x27D6, "dispatch_activated_zone_thread" },
    { 0x27D7, "dispatch_dog_think" },
    { 0x27D8, "dispatchnotify" },
    { 0x27D9, "displacement_uvscale" },
    { 0x27DC, "display_fx_add_options" },
    { 0x27DE, "display_hint" },
    { 0x27DF, "display_hint_stick" },
    { 0x27E0, "display_hint_stick_timeout" },
    { 0x27E1, "display_hint_stick_timeout_mintime" },
    { 0x27E2, "display_hint_timeout" },
    { 0x27E3, "display_hint_timeout_mintime" },
    { 0x27E4, "display_no_target" },
    { 0x27E8, "display_starts_pressed" },
    { 0x27ED, "displaybuddystatusmessage" },
    { 0x27EE, "displayed_hints" },
    { 0x27EF, "displaygameend" },
    { 0x27F0, "displayincomingairdropmessage" },
    { 0x27F2, "displayroundend" },
    { 0x27F4, "displaythreat" },
    { 0x27F5, "displaythrowmessage" },
    { 0x27FA, "dist_to_next_targ" },
    { 0x2800, "distance_from_next_volume_sq" },
    { 0x2804, "distances" },
    { 0x2805, "distfromplayer" },
    { 0x2808, "distorg" },
    { 0x2809, "distort_pct" },
    { 0x280C, "disttoline" },
    { 0x280D, "dlight_on_me" },
    { 0x2810, "dmg_bullet_chance_player_static" },
    { 0x2813, "dmg_player_health_adjust_chance" },
    { 0x2814, "dmg_player_speed_evade_max_pct" },
    { 0x2815, "dmg_player_speed_evade_min_pct" },
    { 0x2817, "dmg_rear_elevator" },
    { 0x2819, "do_bird_single" },
    { 0x281A, "do_bird_single_enemy" },
    { 0x281C, "do_bokeh" },
    { 0x281D, "do_bokehdot_volume" },
    { 0x281F, "do_charge_attack" },
    { 0x2820, "do_creak" },
    { 0x2821, "do_damage_over_time" },
    { 0x2823, "do_decal_square" },
    { 0x2824, "do_dud_damage" },
    { 0x2825, "do_earthquake" },
    { 0x2826, "do_elias_head_look" },
    { 0x2829, "do_exit_wait_on_movement" },
    { 0x282C, "do_funcs" },
    { 0x282D, "do_fx_plane_break" },
    { 0x282E, "do_goto_trace" },
    { 0x282F, "do_ground_slam" },
    { 0x2831, "do_hand_wheeling_anim" },
    { 0x2832, "do_heavy_delete_wait" },
    { 0x2835, "do_knife_throw_blood" },
    { 0x2837, "do_multiple_treads" },
    { 0x2838, "do_nag_dialog" },
    { 0x283B, "do_no_game_start_teleport" },
    { 0x283E, "do_nothing" },
    { 0x283F, "do_physics_pulse" },
    { 0x2841, "do_rag_death" },
    { 0x2842, "do_ramp_light" },
    { 0x2844, "do_random_pilot_chatter" },
    { 0x2845, "do_rope_animation" },
    { 0x2846, "do_rubble" },
    { 0x2849, "do_single_tread" },
    { 0x284A, "do_spacejump_debris" },
    { 0x284B, "do_specular_sun_lerp" },
    { 0x284E, "do_tarps" },
    { 0x2850, "do_unique_debris" },
    { 0x2853, "do_wait_endons_array" },
    { 0x2856, "doaim" },
    { 0x2859, "doairstrikeflyby" },
    { 0x285B, "dobj_manager" },
    { 0x285C, "doblinkinglight" },
    { 0x285D, "dobomberstrike" },
    { 0x285F, "dobuilddot_wait" },
    { 0x2862, "docustomidle" },
    { 0x2863, "dodeathfromarray" },
    { 0x2864, "dodge_dir" },
    { 0x2865, "dodge_distance" },
    { 0x2869, "dodgemoveloopoverride" },
    { 0x286B, "dodgerightanimoffset" },
    { 0x286E, "dodot_delayfunc" },
    { 0x286F, "dodot_fadeinblackout" },
    { 0x2874, "dodrivenidle" },
    { 0x2875, "dodropff" },
    { 0x2876, "doempartifactloop" },
    { 0x2878, "doexposedcalloutresponse" },
    { 0x287C, "dof_blend_interior_ads" },
    { 0x287D, "dof_blend_interior_ads_element" },
    { 0x287E, "dof_blend_interior_ads_scalar" },
    { 0x287F, "dof_blend_interior_generic" },
    { 0x2882, "dof_disable_ads" },
    { 0x2884, "dof_enable_ads" },
    { 0x2885, "dof_enable_script" },
    { 0x2887, "dof_outro_pt2" },
    { 0x2888, "dof_process_ads" },
    { 0x288B, "dof_ref_ent" },
    { 0x288D, "dof_set_generic" },
    { 0x2890, "dof_underwater_general" },
    { 0x2892, "dofacialdialogue" },
    { 0x2894, "dofinalkillcamfx" },
    { 0x2895, "dofiring" },
    { 0x2897, "dofriendlyfirereaction" },
    { 0x2899, "dog_active_touch_zones" },
    { 0x289B, "dog_additional_drivein_anims" },
    { 0x28A1, "dog_animation_generic" },
    { 0x28A3, "dog_animation_sawcorpse" },
    { 0x28A5, "dog_animation_wakeup_slow" },
    { 0x28A9, "dog_attack_alt_func" },
    { 0x28AA, "dog_attack_back_enemies_logic" },
    { 0x28AB, "dog_attack_command_internal" },
    { 0x28AC, "dog_attack_damage_tracking" },
    { 0x28B2, "dog_attack_guy" },
    { 0x28B4, "dog_attack_range" },
    { 0x28B5, "dog_attack_slowmo" },
    { 0x28B7, "dog_attack_tunnel_sniper" },
    { 0x28B8, "dog_attack_victim_logic" },
    { 0x28B9, "dog_attacking_me" },
    { 0x28BB, "dog_bark" },
    { 0x28BD, "dog_bark_think" },
    { 0x28BF, "dog_cam_fov_default" },
    { 0x28C1, "dog_command_attack" },
    { 0x28C7, "dog_death_quote" },
    { 0x28CB, "dog_delayed_allow_damage" },
    { 0x28CC, "dog_delayed_unlink" },
    { 0x28CD, "dog_disable_ai_color" },
    { 0x28CE, "dog_downed_player" },
    { 0x28CF, "dog_drag_to_cover" },
    { 0x28D0, "dog_drive_animscript" },
    { 0x28D4, "dog_enemy_laststand_check" },
    { 0x28D5, "dog_exits_jeep" },
    { 0x28D9, "dog_force_talk" },
    { 0x28DB, "dog_friendly_spotted" },
    { 0x28DC, "dog_gasstation_logic" },
    { 0x28DF, "dog_goto_veh_and_enter" },
    { 0x28E3, "dog_health" },
    { 0x28E4, "dog_hint" },
    { 0x28E5, "dog_hint_check" },
    { 0x28E7, "dog_hits_before_kill" },
    { 0x28E8, "dog_hud_active" },
    { 0x28EA, "dog_hunt2_delete" },
    { 0x28EB, "dog_hunt_spotted" },
    { 0x28EC, "dog_in_exfil_jeep" },
    { 0x28EE, "dog_intro" },
    { 0x28EF, "dog_is_in_combat" },
    { 0x28F0, "dog_jump_down" },
    { 0x28F1, "dog_jump_guy" },
    { 0x28F5, "dog_jumpup_wait" },
    { 0x28F6, "dog_kills_roof_guy" },
    { 0x28F8, "dog_link" },
    { 0x28F9, "dog_lock_check" },
    { 0x28FA, "dog_lock_flag" },
    { 0x28FB, "dog_logic" },
    { 0x28FC, "dog_long_jump" },
    { 0x28FE, "dog_manage_damage" },
    { 0x28FF, "dog_marker" },
    { 0x2901, "dog_melee_index" },
    { 0x2902, "dog_melee_timing_array" },
    { 0x2904, "dog_navigation_logic" },
    { 0x2905, "dog_neck_snapped" },
    { 0x2906, "dog_node_wait" },
    { 0x2909, "dog_pant" },
    { 0x290A, "dog_pant_think" },
    { 0x2911, "dog_return_to_sender" },
    { 0x2912, "dog_reunite" },
    { 0x2916, "dog_run_earthquakes" },
    { 0x2917, "dog_scratch_and_path" },
    { 0x291B, "dog_setup" },
    { 0x291C, "dog_sniff_spots" },
    { 0x291F, "dog_sprint_disable" },
    { 0x2921, "dog_stealth" },
    { 0x2922, "dog_stealth_visibility" },
    { 0x2923, "dog_swap_enemy" },
    { 0x2924, "dog_tag_origin" },
    { 0x2927, "dog_traverse_kill" },
    { 0x292A, "dog_victim_death_internal" },
    { 0x292B, "dog_victim_enemy_early_damage_detection" },
    { 0x292E, "dog_wait_anim_finished" },
    { 0x2930, "dog_walk" },
    { 0x2931, "dog_walk_for_time" },
    { 0x2936, "dogattackaidist" },
    { 0x2939, "dogattackradius" },
    { 0x293B, "dogdamagedradiussq" },
    { 0x293C, "doghintelem" },
    { 0x293D, "doginited" },
    { 0x293E, "doglookpose" },
    { 0x2942, "dognextidletwitchtime" },
    { 0x2945, "dogs" },
    { 0x2948, "dogsinitialized" },
    { 0x2949, "dogstartmovedist" },
    { 0x294B, "dogstoppingdistsq" },
    { 0x294C, "dogtag" },
    { 0x294D, "dogtags" },
    { 0x2951, "dogturnadjust" },
    { 0x2955, "doimmediateragdolldeath" },
    { 0x2957, "doing_hand_signal" },
    { 0x2958, "doing_in_space_rotation" },
    { 0x295A, "doing_start_anim" },
    { 0x295B, "doingadditivepain" },
    { 0x295C, "doingdamagestate" },
    { 0x295E, "doinglongdeath" },
    { 0x295F, "doingreacquirestep" },
    { 0x2960, "doingsplash" },
    { 0x2961, "dointrovignetteanim" },
    { 0x2963, "dolastminuteexposedapproach" },
    { 0x2964, "dolastminuteexposedapproachwrapper" },
    { 0x2965, "dolbstrike" },
    { 0x2967, "dolerp" },
    { 0x296B, "dom" },
    { 0x296D, "domarkingflash" },
    { 0x296E, "domegac130flyby" },
    { 0x2971, "domeleevsai" },
    { 0x2972, "domeleevsai_simple" },
    { 0x2974, "domeleevsai_simple_animcustom_cleanup" },
    { 0x2975, "domeleevsdog" },
    { 0x2979, "dominator_earthquake" },
    { 0x297B, "domissioncallback" },
    { 0x297C, "domortar" },
    { 0x297D, "dompointa" },
    { 0x297F, "dompointc" },
    { 0x2980, "dompointnumber" },
    { 0x2984, "donodeexitanimation" },
    { 0x2985, "dononattackcoverbehavior" },
    { 0x2987, "donotetracksforever" },
    { 0x298A, "donotetracksforpopup" },
    { 0x298B, "donotetracksfortime" },
    { 0x2991, "donotetrackspostcallback" },
    { 0x2993, "donotetrackswithendon" },
    { 0x2994, "donotetrackswithtimeout" },
    { 0x2995, "donothingfunc" },
    { 0x2996, "dont_allow_ammo_cache" },
    { 0x2998, "dont_auto_ride" },
    { 0x2999, "dont_crush_player" },
    { 0x299B, "dont_delete_mines_on_next_spawn" },
    { 0x299E, "dont_unlink_after_breach" },
    { 0x29A0, "dontallowexplode" },
    { 0x29A2, "dontchangemoveplaybackrate" },
    { 0x29A3, "dontchangepushplayer" },
    { 0x29A4, "dontcolormove" },
    { 0x29A5, "dontcrouchtime" },
    { 0x29A7, "dontdonotetracks" },
    { 0x29A8, "dontdrawoncompass" },
    { 0x29A9, "dontdropweapon" },
    { 0x29AD, "dontgetonpath" },
    { 0x29B0, "dontshoot" },
    { 0x29B2, "dontshootwhilemoving" },
    { 0x29B3, "dontunloadondeath" },
    { 0x29B4, "dontunloadonend" },
    { 0x29B7, "donuke_fx" },
    { 0x29B8, "donuke_fx_strikezone" },
    { 0x29B9, "donukesimple" },
    { 0x29BA, "dooneflyby" },
    { 0x29BB, "door" },
    { 0x29BC, "door_break_foley" },
    { 0x29BD, "door_burst_splashes" },
    { 0x29BF, "door_close" },
    { 0x29C0, "door_close_behind" },
    { 0x29C3, "door_current_x" },
    { 0x29C4, "door_debris_l" },
    { 0x29C7, "door_fall_over" },
    { 0x29C8, "door_in" },
    { 0x29CB, "door_knockback" },
    { 0x29CC, "door_knockdown" },
    { 0x29CD, "door_l_modelswap" },
    { 0x29CE, "door_name" },
    { 0x29D2, "door_out" },
    { 0x29D3, "door_parse_parameters" },
    { 0x29D4, "door_play_sounds" },
    { 0x29D9, "door_state_exit" },
    { 0x29DD, "door_state_update_sound" },
    { 0x29DE, "door_switch_icon" },
    { 0x29E2, "door_switch_icon_fade_out" },
    { 0x29E7, "door_traversal" },
    { 0x29E8, "door_traversal_move" },
    { 0x29E9, "door_volume" },
    { 0x29EA, "doorangle" },
    { 0x29EB, "doorclose" },
    { 0x29EE, "doorenter_enable_cqbwalk" },
    { 0x29F1, "doorexit" },
    { 0x29F4, "doorfragchance" },
    { 0x29F6, "doormovetime" },
    { 0x29FB, "doors_open_flicker" },
    { 0x29FC, "doors_swingopen" },
    { 0x29FD, "doorspeed" },
    { 0x29FF, "doortype" },
    { 0x2A00, "doospreyinsertion" },
    { 0x2A01, "dopain" },
    { 0x2A02, "dopainfromarray" },
    { 0x2A03, "dopavelow" },
    { 0x2A08, "doreload" },
    { 0x2A0C, "doshoot" },
    { 0x2A10, "dosounddistant" },
    { 0x2A11, "dospacetraverse" },
    { 0x2A15, "dostrike" },
    { 0x2A16, "dostumble" },
    { 0x2A1A, "dot_player" },
    { 0x2A1C, "dot_to_apache_player_facing_2d_inverse" },
    { 0x2A20, "dotraverse_custom" },
    { 0x2A23, "doturnnotetracks" },
    { 0x2A24, "doturretearthquake" },
    { 0x2A25, "dotypelimit" },
    { 0x2A2A, "doubletapcount" },
    { 0x2A2B, "doubletapthink" },
    { 0x2A2D, "dowalkanimoverride" },
    { 0x2A2F, "down_leg_move_percent" },
    { 0x2A31, "down_part2_proc_ran" },
    { 0x2A32, "down_velocity" },
    { 0x2A35, "downed_enemy_monitor" },
    { 0x2A37, "download_time" },
    { 0x2A38, "download_timer" },
    { 0x2A39, "dpad_icon" },
    { 0x2A3A, "dpad_icon_col" },
    { 0x2A3E, "drag_hints" },
    { 0x2A3F, "drag_interrogate_scene" },
    { 0x2A41, "drag_metal_detector" },
    { 0x2A44, "drag_player_arms" },
    { 0x2A45, "drag_sounds" },
    { 0x2A48, "draw_arrow" },
    { 0x2A49, "draw_arrow_time" },
    { 0x2A4B, "draw_bike_debug" },
    { 0x2A4C, "draw_boundry_sphere" },
    { 0x2A50, "draw_col_vol_offset" },
    { 0x2A51, "draw_color_friendlies" },
    { 0x2A52, "draw_colored_nodes" },
    { 0x2A53, "draw_colornodes" },
    { 0x2A55, "draw_cut_hint" },
    { 0x2A59, "draw_effects_list" },
    { 0x2A5A, "draw_entity_bounds" },
    { 0x2A5D, "draw_line_from_ent_for_time" },
    { 0x2A5E, "draw_line_from_ent_to_ent_for_time" },
    { 0x2A5F, "draw_line_from_ent_to_ent_until_notify" },
    { 0x2A61, "draw_line_to_ent_for_time" },
    { 0x2A62, "draw_line_until_notify" },
    { 0x2A63, "draw_model_path" },
    { 0x2A64, "draw_move_path" },
    { 0x2A66, "draw_path" },
    { 0x2A69, "draw_radial_buttons" },
    { 0x2A6B, "draw_trigger" },
    { 0x2A6D, "draw_turret_target_line" },
    { 0x2A6E, "draw_volume" },
    { 0x2A71, "drawarrow" },
    { 0x2A72, "drawarrowforever" },
    { 0x2A75, "drawchopperattackarrow" },
    { 0x2A77, "drawcircleuntilnotify" },
    { 0x2A7A, "drawenttag" },
    { 0x2A7B, "drawforwardforever" },
    { 0x2A7C, "drawfriend" },
    { 0x2A7D, "drawgraph" },
    { 0x2A80, "drawminimapbounds" },
    { 0x2A81, "drawmyoff" },
    { 0x2A83, "drawoffset" },
    { 0x2A84, "draworgforever" },
    { 0x2A85, "draworiginforever" },
    { 0x2A86, "drawpath" },
    { 0x2A87, "drawplayerviewforever" },
    { 0x2A88, "drawsphere" },
    { 0x2A89, "drawstring" },
    { 0x2A8B, "drawtag" },
    { 0x2A8C, "drawtagforever" },
    { 0x2A91, "drift_hit" },
    { 0x2A93, "driftspeed" },
    { 0x2A94, "drill" },
    { 0x2A95, "drill_bink" },
    { 0x2A96, "drill_carrier" },
    { 0x2A98, "drill_detonate_bbprint" },
    { 0x2A99, "drill_fail_animation" },
    { 0x2A9A, "drill_generic_bbprint" },
    { 0x2A9B, "drill_health_hardcore" },
    { 0x2A9C, "drill_health_scalar" },
    { 0x2A9D, "drill_icon" },
    { 0x2AA0, "drill_locs" },
    { 0x2AA2, "drill_monitor" },
    { 0x2AA3, "drill_nag" },
    { 0x2AAA, "drill_plant" },
    { 0x2AAC, "drill_progress_fx" },
    { 0x2AAF, "drill_reset_bbprint" },
    { 0x2AB0, "drill_reset_pos" },
    { 0x2AB1, "drill_sfx_dist_lp" },
    { 0x2AB2, "drill_sfx_lp" },
    { 0x2AB4, "drill_sfx_offline_lp" },
    { 0x2AB5, "drill_sfx_overheat_lp" },
    { 0x2AB7, "drill_synch_attack_play_anim" },
    { 0x2AB9, "drill_think" },
    { 0x2ABA, "drill_threat_think" },
    { 0x2ABB, "drill_time_scalar" },
    { 0x2ABC, "drill_tutorial_given" },
    { 0x2AC2, "drillthrough_plate_sound" },
    { 0x2AC3, "drip_already_played" },
    { 0x2AC7, "drip_on_player_choose_location" },
    { 0x2AC8, "drive_in_vo" },
    { 0x2ACA, "drivenanimupdate" },
    { 0x2ACD, "driver_dies" },
    { 0x2ACF, "driver_shooting" },
    { 0x2AD0, "driverdead" },
    { 0x2AD1, "drivingvehicle" },
    { 0x2AD2, "drone_animate_on_path" },
    { 0x2AD3, "drone_anims" },
    { 0x2AD4, "drone_array_handling" },
    { 0x2AD7, "drone_civs_init" },
    { 0x2ADA, "drone_death_handler" },
    { 0x2ADD, "drone_die" },
    { 0x2AE0, "drone_fight" },
    { 0x2AE2, "drone_fire_fake_javelin_loop" },
    { 0x2AE5, "drone_frontline_respawner" },
    { 0x2AE7, "drone_give_soul" },
    { 0x2AE8, "drone_gun_remove" },
    { 0x2AE9, "drone_idle" },
    { 0x2AEA, "drone_idle_custom" },
    { 0x2AEE, "drone_init_path" },
    { 0x2AF1, "drone_look_ahead_point" },
    { 0x2AF2, "drone_lookahead_value" },
    { 0x2AF3, "drone_loop_custom" },
    { 0x2AF6, "drone_move_callback" },
    { 0x2AF9, "drone_move_z" },
    { 0x2AFB, "drone_play_anim" },
    { 0x2AFC, "drone_play_looping_anim" },
    { 0x2AFD, "drone_play_scripted_anim" },
    { 0x2AFE, "drone_play_weapon_sound" },
    { 0x2B00, "drone_respawner" },
    { 0x2B02, "drone_runanim" },
    { 0x2B04, "drone_set_runanim" },
    { 0x2B05, "drone_setname" },
    { 0x2B06, "drone_shoot" },
    { 0x2B08, "drone_shoot_rpg" },
    { 0x2B0A, "drone_spawn" },
    { 0x2B0E, "drone_targets" },
    { 0x2B0F, "drone_thermal_draw_disable" },
    { 0x2B10, "drone_traverse_check" },
    { 0x2B1A, "dronemissilespawnarray" },
    { 0x2B1B, "dronepathstarts" },
    { 0x2B1D, "dronerowend" },
    { 0x2B1E, "dronerows" },
    { 0x2B1F, "dronerowsamount" },
    { 0x2B21, "dronerunoffset" },
    { 0x2B23, "drones_crew" },
    { 0x2B24, "drones_death_watcher" },
    { 0x2B27, "drones_request_queue" },
    { 0x2B28, "drones_request_think" },
    { 0x2B29, "drones_targets_sets_to_default" },
    { 0x2B2D, "dronespawner_init" },
    { 0x2B2E, "dronesthermalteamselect" },
    { 0x2B2F, "dronestruct" },
    { 0x2B31, "drop_and_angle_to_ground" },
    { 0x2B32, "drop_ascender" },
    { 0x2B35, "drop_drill" },
    { 0x2B37, "drop_drill_on_disconnect" },
    { 0x2B3B, "drop_geiger_counter" },
    { 0x2B3C, "drop_grenade_bag" },
    { 0x2B3E, "drop_health_timeout_thread" },
    { 0x2B41, "drop_loc" },
    { 0x2B43, "drop_path_to_ground" },
    { 0x2B44, "drop_pillage_item_on_ground" },
    { 0x2B45, "drop_selection_to_ground" },
    { 0x2B46, "drop_to_ground" },
    { 0x2B47, "drop_turret" },
    { 0x2B4A, "dropanim" },
    { 0x2B4B, "dropbombs" },
    { 0x2B4D, "dropcrateexistence" },
    { 0x2B4F, "dropimpulse" },
    { 0x2B51, "dropnuke" },
    { 0x2B52, "dropoff_height" },
    { 0x2B53, "droponebomb" },
    { 0x2B54, "dropparachutebomb" },
    { 0x2B55, "droppeddeathweapon" },
    { 0x2B57, "droppedlinez" },
    { 0x2B58, "droppedoff" },
    { 0x2B5A, "droppingoff" },
    { 0x2B5B, "droppingtoground" },
    { 0x2B5C, "droppostoground" },
    { 0x2B5D, "dropscavengerfordeath" },
    { 0x2B60, "droptimeout" },
    { 0x2B61, "dropturret" },
    { 0x2B63, "droptype" },
    { 0x2B64, "dropweaponfordeath" },
    { 0x2B65, "dropweaponfordeathhorde" },
    { 0x2B68, "drown_choke" },
    { 0x2B6A, "drown_die" },
    { 0x2B6B, "drown_heartbeat" },
    { 0x2B6C, "drown_max_alpha" },
    { 0x2B6D, "drowning_dialogue" },
    { 0x2B6F, "drowning_hudfx" },
    { 0x2B71, "drs_ahead_test" },
    { 0x2B72, "ds_vo_timer" },
    { 0x2B73, "ds_vo_timer_left" },
    { 0x2B74, "ds_vo_timer_right" },
    { 0x2B75, "dshk_death_anim" },
    { 0x2B78, "dsq_2d_ents_lt" },
    { 0x2B7C, "dual_firing" },
    { 0x2B7D, "dual_waittill" },
    { 0x2B7E, "duck_once" },
    { 0x2B7F, "duckidle" },
    { 0x2B80, "duckidleoccurrence" },
    { 0x2B81, "duckin" },
    { 0x2B83, "duckout" },
    { 0x2B84, "dude_kicks_in_the_door" },
    { 0x2B85, "dude_kicks_in_the_door_setup" },
    { 0x2B86, "dudes_in_place_breach" },
    { 0x2B87, "dudes_in_place_cover" },
    { 0x2B88, "duffle_bag_anims" },
    { 0x2B89, "dumb_tank_shoot" },
    { 0x2B8A, "dummy_ent" },
    { 0x2B8D, "dummy_to_vehicle" },
    { 0x2B8F, "dummyspeed" },
    { 0x2B91, "dump_missing_anims" },
    { 0x2B92, "dumpit" },
    { 0x2B93, "dupe_hud" },
    { 0x2B96, "dusteffect" },
    { 0x2B97, "dvar" },
    { 0x2B9B, "dyingcrawlaiming" },
    { 0x2B9C, "dyingcrawlbackaim" },
    { 0x2B9D, "dyingcrawlbloodsmear" },
    { 0x2B9E, "dyn_balloon_bob" },
    { 0x2B9F, "dyn_balloon_delete" },
    { 0x2BA4, "dyn_sniff_disable" },
    { 0x2BA6, "dyn_speed" },
    { 0x2BA8, "dyn_swimspeed_enable" },
    { 0x2BAC, "dynamic_player_crash" },
    { 0x2BAD, "dynamic_repulsor" },
    { 0x2BB3, "dynamic_sun_sample_size" },
    { 0x2BB4, "dynamicspawns" },
    { 0x2BB9, "dz_osprey_missiles" },
    { 0x2BBA, "e" },
    { 0x2BBB, "e3_demo" },
    { 0x2BBC, "e3_fadein" },
    { 0x2BBD, "e3_fadeout" },
    { 0x2BBE, "e3_section" },
    { 0x2BBF, "e3_section_start" },
    { 0x2BC0, "e3_start" },
    { 0x2BC6, "early_level" },
    { 0x2BC7, "early_prompt" },
    { 0x2BC8, "early_weapon_enabled" },
    { 0x2BC9, "earlyabortwatcher" },
    { 0x2BCB, "earned_dialog_col" },
    { 0x2BCC, "earned_hint_col" },
    { 0x2BCD, "earnedstreaklevel" },
    { 0x2BCE, "earnkillstreak" },
    { 0x2BD0, "earth_model" },
    { 0x2BD1, "earth_origin_start" },
    { 0x2BD3, "earthqauke_wait" },
    { 0x2BD4, "earthquake" },
    { 0x2BD5, "earthquake_destructibles_monitor" },
    { 0x2BD7, "earthquake_max" },
    { 0x2BD9, "earthquake_on_death" },
    { 0x2BDA, "earthquake_on_death_missile" },
    { 0x2BDB, "earthquake_player" },
    { 0x2BDD, "earthquake_strength" },
    { 0x2BDE, "earthquake_trigger" },
    { 0x2BE3, "ebreachmodel" },
    { 0x2BE4, "edge_lean_natural" },
    { 0x2BE7, "edgetiles" },
    { 0x2BEB, "effect_id" },
    { 0x2BEF, "effect_monitors" },
    { 0x2BF0, "effect_soundalias" },
    { 0x2BF1, "effectcheckfrequency" },
    { 0x2BF3, "effectheight" },
    { 0x2BF4, "effectlifespan" },
    { 0x2BF5, "effectmaxdelay" },
    { 0x2BF7, "effectradius" },
    { 0x2BFB, "ehq_intro_flag_wait_all" },
    { 0x2BFC, "ehq_znear_default" },
    { 0x2BFD, "einflictor" },
    { 0x2BFE, "either_player_looking_at" },
    { 0x2BFF, "ejected" },
    { 0x2C00, "ele_enemy1" },
    { 0x2C03, "electric_fences" },
    { 0x2C06, "elem" },
    { 0x2C0A, "elevated_circling_retreat" },
    { 0x2C0B, "elevated_delay_retreat" },
    { 0x2C0D, "elevator_105" },
    { 0x2C0F, "elevator_ac130_dmg" },
    { 0x2C11, "elevator_accel" },
    { 0x2C13, "elevator_anims" },
    { 0x2C14, "elevator_call" },
    { 0x2C15, "elevator_callbutton_link_h" },
    { 0x2C18, "elevator_debug" },
    { 0x2C1C, "elevator_door_open" },
    { 0x2C1E, "elevator_floor_update" },
    { 0x2C25, "elevator_interrupt" },
    { 0x2C27, "elevator_lean_over" },
    { 0x2C29, "elevator_move" },
    { 0x2C2A, "elevator_movement" },
    { 0x2C2C, "elevator_outterdoorspeed" },
    { 0x2C2D, "elevator_return" },
    { 0x2C2E, "elevator_room_combat" },
    { 0x2C31, "elevator_room_left_flank" },
    { 0x2C33, "elevator_sound_think" },
    { 0x2C35, "elevator_think" },
    { 0x2C36, "elevator_update_global_dvars" },
    { 0x2C39, "elevator_zoom" },
    { 0x2C3A, "elevatorblockpath" },
    { 0x2C3B, "elevatorclearpath" },
    { 0x2C3C, "elevatordoorsautoclose" },
    { 0x2C3D, "elevatormovetofloor" },
    { 0x2C3E, "elevators" },
    { 0x2C40, "elias" },
    { 0x2C41, "elias_house_attack" },
    { 0x2C42, "elias_house_attack_dialogue" },
    { 0x2C45, "elias_house_attack_grab_sound" },
    { 0x2C46, "elias_house_attack_hesh_shot" },
    { 0x2C47, "elias_house_attack_hit_ground" },
    { 0x2C48, "elias_house_attack_knife_anim_logic" },
    { 0x2C4A, "elias_house_attack_knife_fail_killplayer" },
    { 0x2C4C, "elias_house_attack_knife_hint" },
    { 0x2C4D, "elias_house_attack_knife_hint_logic" },
    { 0x2C4F, "elias_house_attack_knife_think" },
    { 0x2C50, "elias_house_attack_knife_vision" },
    { 0x2C53, "elias_house_attack_wall" },
    { 0x2C55, "elias_house_attacker_breath_knife" },
    { 0x2C56, "elias_house_attacker_breath_sequence" },
    { 0x2C5C, "elias_house_lift_rubble_scene" },
    { 0x2C5D, "elias_house_sequence" },
    { 0x2C5E, "elias_house_window_explosion" },
    { 0x2C5F, "elias_line" },
    { 0x2C60, "elias_logic" },
    { 0x2C62, "elias_start" },
    { 0x2C63, "elias_street_advancing_drones" },
    { 0x2C65, "elias_street_artemis" },
    { 0x2C69, "elias_street_flee_guys" },
    { 0x2C6A, "elias_street_flee_guys_enemies" },
    { 0x2C6C, "elias_street_helicopter_flyover" },
    { 0x2C6D, "elias_street_helicopter_spawn" },
    { 0x2C71, "elite_angered" },
    { 0x2C72, "elite_approach" },
    { 0x2C75, "elite_monitor" },
    { 0x2C78, "emerge_node" },
    { 0x2C79, "emitfalldamage" },
    { 0x2C7A, "emp_jamplayers" },
    { 0x2C7E, "emp_teamtracker" },
    { 0x2C7F, "emp_use" },
    { 0x2C80, "empcandamage" },
    { 0x2C82, "empeffects" },
    { 0x2C84, "empgrenaded" },
    { 0x2C89, "emprumbleloop" },
    { 0x2C8A, "emptimeout" },
    { 0x2C8B, "emptimeremaining" },
    { 0x2C8C, "empty_func" },
    { 0x2C8D, "empty_init_func" },
    { 0x2C8E, "empty_kill_func" },
    { 0x2C8F, "empty_spawner" },
    { 0x2C91, "empty_suit_animation" },
    { 0x2C94, "enable_achievement_harder_they_fall" },
    { 0x2CA0, "enable_blood_pool" },
    { 0x2CA1, "enable_bulletwhizbyreaction" },
    { 0x2CA2, "enable_careful" },
    { 0x2CA3, "enable_cqbwalk" },
    { 0x2CA6, "enable_danger_react" },
    { 0x2CAB, "enable_dog_sniff" },
    { 0x2CAC, "enable_dog_walk" },
    { 0x2CAF, "enable_elias_walk" },
    { 0x2CB4, "enable_global_vehicle_spawn_functions" },
    { 0x2CB7, "enable_hesh_walk" },
    { 0x2CB8, "enable_ignorerandombulletdamage_drone" },
    { 0x2CBB, "enable_limp" },
    { 0x2CC2, "enable_player_space" },
    { 0x2CC3, "enable_player_swim" },
    { 0x2CC6, "enable_space" },
    { 0x2CC8, "enable_spline_path_think" },
    { 0x2CC9, "enable_sprint" },
    { 0x2CCB, "enable_stealth_smart_stance" },
    { 0x2CCC, "enable_stealth_system" },
    { 0x2CCE, "enable_surprise" },
    { 0x2CD0, "enable_team_color" },
    { 0x2CD1, "enable_teamflashbangimmunity" },
    { 0x2CD2, "enable_teamflashbangimmunity_proc" },
    { 0x2CD3, "enable_tired" },
    { 0x2CD7, "enable_trigger_with_noteworthy" },
    { 0x2CD8, "enable_trigger_with_targetname" },
    { 0x2CD9, "enable_turnanims" },
    { 0x2CDA, "enable_turns_arrivals_exits" },
    { 0x2CDE, "enable_weapons" },
    { 0x2CDF, "enableairdropoutlineai" },
    { 0x2CE4, "enableextendedkill" },
    { 0x2CE6, "enablejammedeffect" },
    { 0x2CE7, "enablejuggernaut" },
    { 0x2CE9, "enablelaststandweapons" },
    { 0x2CEB, "enableplayerweapons" },
    { 0x2CED, "enablesomecontrol" },
    { 0x2CEF, "enablestop" },
    { 0x2CF0, "enableteamintel" },
    { 0x2CF1, "enableweaponlaser" },
    { 0x2CF2, "encoderatio" },
    { 0x2CF6, "end" },
    { 0x2CF7, "end_aim_idle_thread" },
    { 0x2CF9, "end_anim_then_loop" },
    { 0x2CFB, "end_birds" },
    { 0x2CFD, "end_breach_enemies_killed" },
    { 0x2CFE, "end_breach_enemy_spawn_single" },
    { 0x2CFF, "end_breach_engines_sieze" },
    { 0x2D02, "end_breach_impulse_player_logic" },
    { 0x2D03, "end_breach_impulse_player_single" },
    { 0x2D05, "end_breach_player_death_rpg" },
    { 0x2D06, "end_breach_player_unlink" },
    { 0x2D07, "end_breach_rpg_explosion" },
    { 0x2D0A, "end_breach_rpg_guy_fire" },
    { 0x2D0E, "end_color" },
    { 0x2D14, "end_dialogue" },
    { 0x2D15, "end_enemies" },
    { 0x2D16, "end_fade" },
    { 0x2D17, "end_fardist" },
    { 0x2D18, "end_fire_and_anim_idle_thread" },
    { 0x2D19, "end_game_score" },
    { 0x2D1A, "end_game_string_index" },
    { 0x2D1B, "end_game_string_override" },
    { 0x2D1D, "end_hdrsuncolorintensity" },
    { 0x2D1E, "end_heli_2_think" },
    { 0x2D1F, "end_heli_treadfx" },
    { 0x2D23, "end_legs_jump_anim" },
    { 0x2D24, "end_level" },
    { 0x2D27, "end_mantle_angle" },
    { 0x2D2A, "end_neardist" },
    { 0x2D2B, "end_nodes" },
    { 0x2D2E, "end_notifies" },
    { 0x2D30, "end_of_scripting" },
    { 0x2D31, "end_opacity" },
    { 0x2D32, "end_org" },
    { 0x2D34, "end_script" },
    { 0x2D35, "end_script_corner" },
    { 0x2D36, "end_skyfogintensity" },
    { 0x2D3A, "end_slide_effects" },
    { 0x2D3B, "end_sunbeginfadeangle" },
    { 0x2D3C, "end_suncolor" },
    { 0x2D3D, "end_sundir" },
    { 0x2D3E, "end_sunendfadeangle" },
    { 0x2D3F, "end_sunfogscale" },
    { 0x2D40, "end_surface" },
    { 0x2D41, "end_swim_anim_node" },
    { 0x2D43, "end_swim_logic" },
    { 0x2D44, "end_thread" },
    { 0x2D46, "end_to_start_length" },
    { 0x2D47, "end_tunnel_swim" },
    { 0x2D48, "end_turret_reservation" },
    { 0x2D4E, "endcustomevent" },
    { 0x2D4F, "endedkillcamcleanup" },
    { 0x2D54, "endgame" },
    { 0x2D5A, "endgameontimelimit" },
    { 0x2D5B, "endgameovertime" },
    { 0x2D5E, "endgameupdate" },
    { 0x2D61, "ending2_debri_fall" },
    { 0x2D65, "ending2_window_break" },
    { 0x2D67, "ending_animatic_setup" },
    { 0x2D68, "ending_arms" },
    { 0x2D69, "ending_autosave" },
    { 0x2D6A, "ending_bink_display" },
    { 0x2D6B, "ending_blink_qte_prompt" },
    { 0x2D6C, "ending_blood_wall" },
    { 0x2D6D, "ending_blur_logic" },
    { 0x2D6E, "ending_breach_ally" },
    { 0x2D77, "ending_dof_05" },
    { 0x2D7A, "ending_dof_08" },
    { 0x2D7C, "ending_door_r" },
    { 0x2D83, "ending_fx_opfor03_fire_pilot" },
    { 0x2D84, "ending_gate_l" },
    { 0x2D85, "ending_gate_node_left" },
    { 0x2D86, "ending_gate_node_right" },
    { 0x2D87, "ending_gate_r" },
    { 0x2D8A, "ending_harmless_shots_logic" },
    { 0x2D8B, "ending_heli" },
    { 0x2D8C, "ending_heli_callout_vo" },
    { 0x2D8D, "ending_heli_path" },
    { 0x2D8F, "ending_hvt_handle_damage" },
    { 0x2D90, "ending_hvt_shot_blood_fx" },
    { 0x2D91, "ending_init" },
    { 0x2D94, "ending_lower_raise_weapon_logic" },
    { 0x2D95, "ending_open_doors" },
    { 0x2D97, "ending_opfor_1" },
    { 0x2D9A, "ending_opfor_kill_pilot" },
    { 0x2D9B, "ending_player_anims" },
    { 0x2D9C, "ending_player_broken_nose" },
    { 0x2D9E, "ending_player_enemy_broken_nose" },
    { 0x2DA0, "ending_player_failed_qte_0" },
    { 0x2DA3, "ending_player_hvt_aim" },
    { 0x2DA8, "ending_player_qte_0_logic" },
    { 0x2DAB, "ending_player_qte_reach_logic" },
    { 0x2DAC, "ending_player_qte_shoot_logic" },
    { 0x2DB6, "ending_price_gets_capped" },
    { 0x2DB7, "ending_pt1_allies_sequence_start" },
    { 0x2DB8, "ending_pt1_ally_0_sequence" },
    { 0x2DB9, "ending_pt1_ally_1_sequence" },
    { 0x2DC0, "ending_pt2_player_sequence_save" },
    { 0x2DC1, "ending_qte_0_prompt_logic" },
    { 0x2DC3, "ending_qte_catch" },
    { 0x2DC8, "ending_qte_reach" },
    { 0x2DCA, "ending_scene" },
    { 0x2DCD, "ending_sequence" },
    { 0x2DCF, "ending_setup" },
    { 0x2DD4, "ending_transition" },
    { 0x2DD5, "ending_vo_main" },
    { 0x2DD6, "ending_white_fade" },
    { 0x2DD8, "endkillcamifnothingtoshow" },
    { 0x2DDA, "endmgstreakwhenleavemg" },
    { 0x2DDD, "endnotetrack" },
    { 0x2DDF, "endofspeedwatcher" },
    { 0x2DE2, "endonstring" },
    { 0x2DE5, "endpoint" },
    { 0x2DE6, "endpos" },
    { 0x2DE8, "endrespawnnotify" },
    { 0x2DF0, "endselectiononemp" },
    { 0x2DF1, "endselectiononendgame" },
    { 0x2DF2, "endsliding" },
    { 0x2DF4, "endtime" },
    { 0x2DF5, "enduavonlatejoiner" },
    { 0x2DF6, "endupvector" },
    { 0x2DF7, "endvel" },
    { 0x2DFA, "enemies_above_killed" },
    { 0x2DFC, "enemies_around_nitro" },
    { 0x2DFE, "enemies_engine_room" },
    { 0x2DFF, "enemies_engine_room_reinforcements" },
    { 0x2E00, "enemies_engineroom_entry" },
    { 0x2E02, "enemies_engineroom_extinguisher" },
    { 0x2E05, "enemies_final_move" },
    { 0x2E08, "enemies_magic_bullet_until_player_at_corner" },
    { 0x2E0A, "enemies_mg_watcher_shield_damage" },
    { 0x2E0F, "enemies_right_door" },
    { 0x2E10, "enemies_sat1" },
    { 0x2E11, "enemies_sat2" },
    { 0x2E12, "enemies_scene" },
    { 0x2E13, "enemies_setup_explosion_scene_guys" },
    { 0x2E17, "enemieskilledintimewindow" },
    { 0x2E1A, "enemy2" },
    { 0x2E1B, "enemy_13" },
    { 0x2E1E, "enemy_1_sitting" },
    { 0x2E1F, "enemy_1_standing" },
    { 0x2E21, "enemy_alert_level_attack" },
    { 0x2E22, "enemy_alert_level_attack_wrapper" },
    { 0x2E25, "enemy_alert_level_default_pre_spotted_func" },
    { 0x2E26, "enemy_alert_level_forget" },
    { 0x2E27, "enemy_alert_level_logic" },
    { 0x2E29, "enemy_alert_level_normal" },
    { 0x2E2A, "enemy_alert_level_normal_wrapper" },
    { 0x2E2B, "enemy_alert_level_reset_wrapper" },
    { 0x2E2E, "enemy_alert_level_warning1" },
    { 0x2E32, "enemy_animation_custom" },
    { 0x2E33, "enemy_animation_do_anim" },
    { 0x2E35, "enemy_animation_generic" },
    { 0x2E36, "enemy_animation_loop" },
    { 0x2E38, "enemy_animation_post_anim" },
    { 0x2E39, "enemy_animation_pre_anim" },
    { 0x2E3A, "enemy_animation_pre_anim_dog_special_first_condition" },
    { 0x2E40, "enemy_announce_corpse" },
    { 0x2E41, "enemy_announce_hmph" },
    { 0x2E46, "enemy_announce_spotted_acknowledge" },
    { 0x2E48, "enemy_announce_wtf" },
    { 0x2E4A, "enemy_attempt_melee" },
    { 0x2E4B, "enemy_breach_anims" },
    { 0x2E53, "enemy_checkpoint_dialogue" },
    { 0x2E57, "enemy_corpse_alert_level" },
    { 0x2E59, "enemy_corpse_found" },
    { 0x2E5B, "enemy_corpse_found_loop" },
    { 0x2E5D, "enemy_corpse_logic" },
    { 0x2E61, "enemy_corpse_saw_behavior" },
    { 0x2E64, "enemy_custom_corpse_behavior" },
    { 0x2E65, "enemy_custom_state_behavior" },
    { 0x2E69, "enemy_death_tracking" },
    { 0x2E6A, "enemy_debris_vo" },
    { 0x2E6D, "enemy_default_corpse_behavior" },
    { 0x2E6F, "enemy_default_threat_anim_behavior" },
    { 0x2E71, "enemy_destroyer_guy_logic" },
    { 0x2E72, "enemy_dialog_col" },
    { 0x2E74, "enemy_dog_init" },
    { 0x2E76, "enemy_drone_anim" },
    { 0x2E78, "enemy_drones_pt1_lower" },
    { 0x2E79, "enemy_drones_pt1_lower_runners" },
    { 0x2E7D, "enemy_event_awareness_notify" },
    { 0x2E7E, "enemy_event_category_awareness" },
    { 0x2E7F, "enemy_event_debug_print" },
    { 0x2E80, "enemy_event_declare_to_team" },
    { 0x2E83, "enemy_event_listeners_logic" },
    { 0x2E84, "enemy_event_listeners_proc" },
    { 0x2E85, "enemy_event_loop" },
    { 0x2E86, "enemy_event_reaction_dog_attack" },
    { 0x2E87, "enemy_event_reaction_dog_bark" },
    { 0x2E89, "enemy_event_reaction_flashbang" },
    { 0x2E8A, "enemy_event_reaction_heard_scream" },
    { 0x2E8D, "enemy_final_wave_run" },
    { 0x2E94, "enemy_gaz_health" },
    { 0x2E95, "enemy_get_nearby_pathnodes" },
    { 0x2E97, "enemy_go_back_clear_lastspot" },
    { 0x2E9E, "enemy_heli_attacking" },
    { 0x2E9F, "enemy_heli_killed" },
    { 0x2EA1, "enemy_heli_respawner" },
    { 0x2EA4, "enemy_infantry_set_up_on_spawn" },
    { 0x2EA7, "enemy_investigate_position" },
    { 0x2EAA, "enemy_jeep_b" },
    { 0x2EAF, "enemy_jeep_turret" },
    { 0x2EB4, "enemy_lookaround_for_time" },
    { 0x2EB5, "enemy_lower_level" },
    { 0x2EB9, "enemy_melee_readjust" },
    { 0x2EBA, "enemy_mg_adjust_if_target_dies" },
    { 0x2EBB, "enemy_mg_burst_fire" },
    { 0x2EBC, "enemy_mg_get_untargeted_random_target" },
    { 0x2EBD, "enemy_mg_pin_down_player" },
    { 0x2EBF, "enemy_near_position" },
    { 0x2EC0, "enemy_odin_assault_exists" },
    { 0x2EC1, "enemy_patrol_phone" },
    { 0x2EC3, "enemy_plane_behind_skipto" },
    { 0x2EC4, "enemy_plane_engine_loop_01" },
    { 0x2EC7, "enemy_plane_engine_loop_04" },
    { 0x2EC9, "enemy_plane_looping_sounds_start" },
    { 0x2ECA, "enemy_poker_cards" },
    { 0x2ECC, "enemy_prespotted_func_default" },
    { 0x2ECD, "enemy_process_synch_attack" },
    { 0x2ED1, "enemy_reaction_state_alert" },
    { 0x2ED3, "enemy_rnd_runner" },
    { 0x2ED5, "enemy_runto_and_lookaround" },
    { 0x2ED6, "enemy_saw_corpse_logic" },
    { 0x2ED7, "enemy_sentry_difficulty_settings" },
    { 0x2EDB, "enemy_set_threat_behavior" },
    { 0x2EDC, "enemy_setup" },
    { 0x2EDD, "enemy_setup_vign" },
    { 0x2EDE, "enemy_setup_vign_floor" },
    { 0x2EDF, "enemy_shoot_at_player" },
    { 0x2EE1, "enemy_snowmobiles" },
    { 0x2EE3, "enemy_spanish_vo" },
    { 0x2EE5, "enemy_spawn_func" },
    { 0x2EE7, "enemy_squad_spawn" },
    { 0x2EEC, "enemy_stop_current_behavior" },
    { 0x2EF0, "enemy_struggle_anim" },
    { 0x2EF1, "enemy_struggle_anim_additives" },
    { 0x2EF4, "enemy_switchblade_exists" },
    { 0x2EF8, "enemy_tank_soldiers_2_init" },
    { 0x2EFB, "enemy_team_name" },
    { 0x2EFC, "enemy_threat_anim_defaults" },
    { 0x2EFD, "enemy_threat_logic" },
    { 0x2EFE, "enemy_threat_logic_dog" },
    { 0x2EFF, "enemy_threat_logic_dog_wait" },
    { 0x2F01, "enemy_threat_loop" },
    { 0x2F04, "enemy_twitch" },
    { 0x2F05, "enemy_twitch_runstumble" },
    { 0x2F09, "enemy_wait_for_synch_attack_end" },
    { 0x2F0A, "enemy_wait_for_synch_end_notify" },
    { 0x2F0B, "enemy_wait_for_synch_invalid_enemy" },
    { 0x2F0C, "enemy_waittill_count" },
    { 0x2F0E, "enemy_weapons_force_use_silencer" },
    { 0x2F0F, "enemy_zodiacs_spawn_and_attack" },
    { 0x2F10, "enemy_zpu_health" },
    { 0x2F14, "enemygazs" },
    { 0x2F16, "enemyguys2" },
    { 0x2F19, "enemyhq_basement_pre_load" },
    { 0x2F1B, "enemyhq_footstep_override" },
    { 0x2F1E, "enemyhq_pre_load" },
    { 0x2F1F, "enemyhq_rooftop_intro_pre_load" },
    { 0x2F20, "enemyhq_starts" },
    { 0x2F26, "enemyisingeneraldirection" },
    { 0x2F2B, "enemytank_cleanup" },
    { 0x2F2D, "enemytanksbri1" },
    { 0x2F2E, "enemyteam" },
    { 0x2F2F, "enemyteamid" },
    { 0x2F30, "enemytrigger" },
    { 0x2F31, "enemyvelocity" },
    { 0x2F32, "engageground" },
    { 0x2F33, "engagementcomplimentdialognext" },
    { 0x2F34, "engagevehicle" },
    { 0x2F35, "engine_fxs" },
    { 0x2F37, "engineroom_headsmoke_fx_end" },
    { 0x2F38, "engineroom_headsmoke_fx_start" },
    { 0x2F39, "engineroom_heat_fx_shake" },
    { 0x2F3C, "enginevfxtag" },
    { 0x2F3D, "enhanced_vision" },
    { 0x2F42, "ent_flag" },
    { 0x2F43, "ent_flag_assert" },
    { 0x2F46, "ent_flag_exist" },
    { 0x2F4A, "ent_flag_wait" },
    { 0x2F4B, "ent_flag_wait_either" },
    { 0x2F4C, "ent_flag_wait_or_timeout" },
    { 0x2F4D, "ent_flag_wait_vehicle_node" },
    { 0x2F4E, "ent_flag_waitopen" },
    { 0x2F50, "ent_flags_lock" },
    { 0x2F51, "ent_is_highlighted" },
    { 0x2F54, "ent_parachute_from_plane_two" },
    { 0x2F55, "ent_wait_for_flag_or_time_elapses" },
    { 0x2F56, "ent_waits_for_trigger" },
    { 0x2F57, "enter" },
    { 0x2F58, "enter_blackbird_vo" },
    { 0x2F59, "enter_camera_zoomout" },
    { 0x2F5D, "enter_jeep" },
    { 0x2F60, "enter_revive_use_hold_think" },
    { 0x2F61, "enter_spectate" },
    { 0x2F64, "entermove" },
    { 0x2F65, "enterpronewrapper" },
    { 0x2F6A, "entities_are_selected" },
    { 0x2F6C, "entity_counter" },
    { 0x2F6D, "entity_fx_and_anims_think" },
    { 0x2F6E, "entity_highlight_disable" },
    { 0x2F6F, "entity_highlight_enable" },
    { 0x2F70, "entity_number" },
    { 0x2F71, "entity_path_disconnect_thread" },
    { 0x2F73, "entityheadiconoffset" },
    { 0x2F76, "entityheadiconteam" },
    { 0x2F77, "entitynumber" },
    { 0x2F78, "entnum" },
    { 0x2F7C, "entrance_points" },
    { 0x2F7F, "entrance_visible_from" },
    { 0x2F81, "entrance_watched_by_player" },
    { 0x2F82, "entry" },
    { 0x2F83, "entry_door_beeper" },
    { 0x2F85, "entry_door_close_layer2" },
    { 0x2F87, "ents" },
    { 0x2F89, "entsound" },
    { 0x2F8A, "eog_altimeter_conv_ratio" },
    { 0x2F8B, "eog_player_tracking_init" },
    { 0x2F8C, "eog_player_update_stat" },
    { 0x2F8E, "eq_table" },
    { 0x2F90, "eq_track" },
    { 0x2F91, "eq_trigger_num" },
    { 0x2F92, "equal_to_goal" },
    { 0x2F93, "equipment_enabled" },
    { 0x2F94, "equipmentdeathvfx" },
    { 0x2F95, "equipmentdeletevfx" },
    { 0x2F96, "equipmentempstunvfx" },
    { 0x2F98, "equipped_hide_weapon" },
    { 0x2F99, "equipped_weapon" },
    { 0x2F9A, "erasefinalkillcam" },
    { 0x2F9B, "error" },
    { 0x2FA3, "escape_choke_init" },
    { 0x2FA6, "escape_combat_door" },
    { 0x2FA7, "escape_cycle" },
    { 0x2FAB, "escape_dialogue" },
    { 0x2FAD, "escape_door_anim_player" },
    { 0x2FB0, "escape_enemies_and_vehicles" },
    { 0x2FB2, "escape_enemy_01_think" },
    { 0x2FB3, "escape_enemy_02_think" },
    { 0x2FB4, "escape_enemy_03_think" },
    { 0x2FB5, "escape_enemy_setup" },
    { 0x2FB7, "escape_fake_underwater_bullets" },
    { 0x2FB9, "escape_friendly_follow_spline" },
    { 0x2FBB, "escape_friendly_movement" },
    { 0x2FBD, "escape_jungle_start" },
    { 0x2FBE, "escape_loc" },
    { 0x2FBF, "escape_main" },
    { 0x2FC2, "escape_objective" },
    { 0x2FC3, "escape_play_slide_fx_on_npc" },
    { 0x2FC6, "escape_player_jump" },
    { 0x2FC7, "escape_player_speed" },
    { 0x2FC9, "escape_river_start" },
    { 0x2FCB, "escape_scripted_destruction" },
    { 0x2FCC, "escape_setup" },
    { 0x2FCE, "escape_setup_trees" },
    { 0x2FD1, "escape_socr_turret_own_target" },
    { 0x2FD2, "escape_spawn_special_minion_wave" },
    { 0x2FD8, "escape_vo" },
    { 0x2FD9, "escape_waterfall_enemies_logic" },
    { 0x2FDD, "estate" },
    { 0x2FDE, "estimatedtimetillscorelimit" },
    { 0x2FE0, "evad_anims" },
    { 0x2FE2, "eval" },
    { 0x2FE4, "evaluatefiringevent" },
    { 0x2FE5, "evaluatemeleeevent" },
    { 0x2FE8, "evaluatesuppressionevent" },
    { 0x2FE9, "evasive_addpoint" },
    { 0x2FED, "evasive_endmaneuvers" },
    { 0x2FF5, "event_awareness_dialogue_wrapper" },
    { 0x2FF6, "event_awareness_enders" },
    { 0x2FFE, "event_derrick_explode_catwalk_break" },
    { 0x3001, "event_derrick_explode_debris_bomb_throw_enemy" },
    { 0x3003, "event_derrick_explode_debris_main" },
    { 0x3005, "event_derrick_explode_debris_oiltank" },
    { 0x3007, "event_derrick_explode_debris_setup_collision" },
    { 0x300A, "event_derrick_explode_large" },
    { 0x300B, "event_derrick_explode_setup" },
    { 0x300D, "event_derrick_explode_stack_setup" },
    { 0x300F, "event_flarestack_enter" },
    { 0x3011, "event_flarestack_enter_fail_watcher" },
    { 0x3014, "event_intro_player_anims" },
    { 0x3015, "event_intro_tv_pip" },
    { 0x3016, "event_jet_flyby" },
    { 0x3017, "event_noticket" },
    { 0x3018, "event_pipe_explosions" },
    { 0x3019, "event_player_arm_break" },
    { 0x301B, "event_player_gets_kicked" },
    { 0x301D, "event_player_grabs_knife" },
    { 0x301F, "event_pressure_buildup" },
    { 0x3021, "event_qte_bullet_catch" },
    { 0x3024, "event_quaker_indoor" },
    { 0x3025, "event_quaker_outdoor" },
    { 0x3027, "event_rogs_finale" },
    { 0x302B, "event_scan_manager" },
    { 0x302C, "event_sinking_boats" },
    { 0x302E, "event_tanks_bridge_fall_scene" },
    { 0x302F, "event_thread" },
    { 0x3035, "eventactionminwait" },
    { 0x3036, "eventchance" },
    { 0x3037, "eventduration" },
    { 0x3038, "eventpriority" },
    { 0x303A, "eventstring" },
    { 0x303B, "eventtype" },
    { 0x303C, "eventtypeminwait" },
    { 0x303D, "everusessecondaryweapon" },
    { 0x303E, "exceededmaxballdrones" },
    { 0x3044, "exception" },
    { 0x3046, "excludedir" },
    { 0x304C, "execute_save" },
    { 0x304D, "execution" },
    { 0x304E, "execution_guy_dead" },
    { 0x304F, "execution_slowmo" },
    { 0x3051, "executor_kill_hostage_notetrack" },
    { 0x3054, "exfil_animnode" },
    { 0x3055, "exfil_anims_cornering" },
    { 0x3057, "exfil_blackice_exfil_heli_lights_fx" },
    { 0x3059, "exfil_chopper" },
    { 0x305A, "exfil_cypher_enter_jeep" },
    { 0x305D, "exfil_engine_fires" },
    { 0x3060, "exfil_enter_jeep" },
    { 0x3061, "exfil_get_on_turret" },
    { 0x3062, "exfil_gun_cock" },
    { 0x3063, "exfil_hall_explosions" },
    { 0x3065, "exfil_heli" },
    { 0x3066, "exfil_heli_smoke_fx_01" },
    { 0x3067, "exfil_heli_spotlight" },
    { 0x3068, "exfil_heli_tag" },
    { 0x3069, "exfil_hoodsmack" },
    { 0x306A, "exfil_jeep_latch_close_lf" },
    { 0x306C, "exfil_jeep_latch_close_rf" },
    { 0x306D, "exfil_jeep_latch_open_lf" },
    { 0x3070, "exfil_jet" },
    { 0x3072, "exfil_keegan_and_cypher_enter_jeep" },
    { 0x3075, "exfil_mblur_changes" },
    { 0x307A, "exfil_player_view_smoke_particles" },
    { 0x307B, "exfil_random_quaker" },
    { 0x307D, "exfil_slomo" },
    { 0x307E, "exfil_sniper" },
    { 0x3080, "exfil_teleport" },
    { 0x3081, "exfil_vision_bump" },
    { 0x3082, "exfil_yellow_alarms" },
    { 0x3086, "exit" },
    { 0x3088, "exit_gamemodespecificaction" },
    { 0x308B, "exit_laststand" },
    { 0x308C, "exit_nags" },
    { 0x3094, "exit_tunnel_jeep_by" },
    { 0x3095, "exit_water_tired" },
    { 0x3097, "exit_wreck_check" },
    { 0x3098, "exitaistate" },
    { 0x3099, "exiting_combat_player_fail" },
    { 0x309A, "exiting_combat_zone" },
    { 0x309C, "exitpronewrapper" },
    { 0x30A0, "expand_goalradius" },
    { 0x30A4, "expire_time" },
    { 0x30A6, "explicitabandonminitarget" },
    { 0x30A7, "explicitabandontarget" },
    { 0x30AA, "explode_gunboats" },
    { 0x30AE, "explode_zodiacs" },
    { 0x30AF, "exploded" },
    { 0x30B0, "explodeondamage" },
    { 0x30B1, "exploder" },
    { 0x30B2, "exploder_after_load" },
    { 0x30B4, "exploder_cache" },
    { 0x30B6, "exploder_damage" },
    { 0x30B7, "exploder_damage_loop" },
    { 0x30B9, "exploder_earthquake" },
    { 0x30BA, "exploder_flag_wait" },
    { 0x30BB, "exploder_fx_stuffs" },
    { 0x30BD, "exploder_model_is_chunk" },
    { 0x30C1, "exploder_radius_custom" },
    { 0x30C6, "exploder_with_connect_watch" },
    { 0x30C7, "exploderfunction" },
    { 0x30C8, "exploderfx" },
    { 0x30CB, "exploders_cached" },
    { 0x30CC, "exploders_explode_for_late_player" },
    { 0x30CE, "explodevfx" },
    { 0x30D1, "exploding_heli" },
    { 0x30D5, "explosion_anim" },
    { 0x30D7, "explosion_boom" },
    { 0x30D8, "explosion_death" },
    { 0x30D9, "explosion_death_offset" },
    { 0x30DB, "explosion_delay" },
    { 0x30DD, "explosion_incoming" },
    { 0x30E2, "explosion_throw_sandbags" },
    { 0x30E4, "explosion_vehicle_killer2" },
    { 0x30E5, "explosion_vehicle_killer3" },
    { 0x30E6, "explosion_vehicle_killer4" },
    { 0x30EB, "explosive_damage_watch" },
    { 0x30EC, "explosive_dist" },
    { 0x30ED, "explosive_type" },
    { 0x30EF, "explosivehandlemovers" },
    { 0x30F0, "explosiveinfo" },
    { 0x30F3, "explosiveplanttime" },
    { 0x30F4, "explosivetrigger" },
    { 0x30F5, "expltagroot" },
    { 0x30F8, "exposed_guy" },
    { 0x30FC, "exposedcombatcheckputawaypistol" },
    { 0x30FD, "exposedcombatcheckreloadorusepistol" },
    { 0x30FF, "exposedcombatmainloop" },
    { 0x3101, "exposedcombatpositionadjust" },
    { 0x3102, "exposedcombatstopusingrpgcheck" },
    { 0x310A, "extra_vehicle_cleanup" },
    { 0x310B, "extrafiretime_max" },
    { 0x310C, "extrafiretime_min" },
    { 0x310D, "extraspintime_max" },
    { 0x3112, "f_max" },
    { 0x3115, "face_players" },
    { 0x3116, "faceenemyaimtracking" },
    { 0x3117, "faceenemyarrival" },
    { 0x3118, "faceenemyatendofapproach" },
    { 0x311C, "faceresult" },
    { 0x311E, "facewaitforresult" },
    { 0x3120, "faceyaw" },
    { 0x3121, "facial" },
    { 0x3125, "facialidx" },
    { 0x3129, "facing_dot" },
    { 0x312C, "fade" },
    { 0x312D, "fade_ambient_elem_internal" },
    { 0x312F, "fade_ambient_elems" },
    { 0x3130, "fade_from_black" },
    { 0x3135, "fade_in_hint" },
    { 0x3136, "fade_in_logo" },
    { 0x3138, "fade_in_to_alpha" },
    { 0x313B, "fade_logo_run_credits" },
    { 0x313C, "fade_out" },
    { 0x313D, "fade_out_cut_hint" },
    { 0x3140, "fade_out_hint" },
    { 0x3141, "fade_out_in" },
    { 0x3145, "fade_out_time" },
    { 0x3148, "fade_qte_prompt" },
    { 0x3149, "fade_sun_in_out" },
    { 0x314B, "fade_tired" },
    { 0x314D, "fade_up_black" },
    { 0x314E, "fadeaway" },
    { 0x3150, "fadeinblackout" },
    { 0x3154, "fadeoverlay_no_blur" },
    { 0x3155, "fadetoalpha" },
    { 0x3156, "fadetoalphatime" },
    { 0x3157, "fading" },
    { 0x315A, "fail_mission_ambush" },
    { 0x315B, "fail_mission_leave_area" },
    { 0x315E, "fail_on_player_kill" },
    { 0x315F, "failactivatefunc" },
    { 0x3160, "failcase_blackout_early" },
    { 0x3162, "failed_spawnvehicles" },
    { 0x3165, "failonfriendlyfire" },
    { 0x3166, "fails" },
    { 0x3167, "failsafe_c_k_bags" },
    { 0x3168, "fake" },
    { 0x3169, "fake_ally_kill_me" },
    { 0x316B, "fake_dead" },
    { 0x316C, "fake_death_bullet" },
    { 0x316F, "fake_function" },
    { 0x3171, "fake_mantle_text" },
    { 0x3172, "fake_missile_launch" },
    { 0x3175, "fake_shellshock_sound" },
    { 0x3176, "fake_shooter_death" },
    { 0x317A, "fake_spot" },
    { 0x317C, "fake_tank_rumble" },
    { 0x317D, "fake_targets" },
    { 0x317F, "fake_teleport_notify" },
    { 0x3181, "fake_vehicle_moveto" },
    { 0x3185, "fakea10ai" },
    { 0x3186, "fakedrones" },
    { 0x3190, "fall_break_glass" },
    { 0x3194, "fall_enemies_rail_hit" },
    { 0x3196, "fall_enemies_start" },
    { 0x3197, "fall_enemy_anim" },
    { 0x319A, "fall_environment" },
    { 0x319B, "fall_fx_billboard_fx" },
    { 0x319C, "fall_fx_billboard_setup" },
    { 0x319E, "fall_fx_crowd_fx" },
    { 0x31A0, "fall_fx_end_bldg" },
    { 0x31A5, "fall_object_init" },
    { 0x31A6, "fall_object_run" },
    { 0x31AA, "fall_physics_debris_furniture" },
    { 0x31AB, "fall_physics_debris_lobby" },
    { 0x31AC, "fall_physics_debris_slide" },
    { 0x31B0, "fall_prop_merrick_window" },
    { 0x31B1, "fall_prop_picture" },
    { 0x31B5, "fall_props_debris_b" },
    { 0x31BB, "fall_props_player_parachute" },
    { 0x31BC, "fall_to" },
    { 0x31BF, "fallback_ai" },
    { 0x31C0, "fallback_ai_think" },
    { 0x31C2, "fallback_coverprint" },
    { 0x31C5, "fallback_initiated" },
    { 0x31C8, "fallback_text" },
    { 0x31CB, "falling_debris" },
    { 0x31D0, "falling_rock_init" },
    { 0x31D1, "falling_rock_kill_trigger" },
    { 0x31D2, "falling_rock_run" },
    { 0x31D3, "falling_rocks" },
    { 0x31D4, "falling_sat_dish" },
    { 0x31D5, "falling_tree_player_detection" },
    { 0x31D6, "fallingtileeffect" },
    { 0x31D7, "falloff" },
    { 0x31D9, "fandb_dodamage" },
    { 0x31DD, "fandb_handles_collision_brushes" },
    { 0x31DE, "fandb_hideshowtag" },
    { 0x31DF, "fandb_playsound" },
    { 0x31E5, "farthest_distance_down" },
    { 0x31E6, "fast_convoy" },
    { 0x31E7, "fast_destructible_explode" },
    { 0x31E8, "fast_jog" },
    { 0x31E9, "fast_walk" },
    { 0x31EB, "fastburstfirenumshots" },
    { 0x31EC, "fastcrouch" },
    { 0x31F0, "fastropeoffset" },
    { 0x31F4, "faux_spawn_stance" },
    { 0x31F5, "fauxdead" },
    { 0x31FB, "faze_out_finish" },
    { 0x31FD, "fbt_firstburst" },
    { 0x3203, "featherents" },
    { 0x3205, "fed_destroyer_fx_guns" },
    { 0x3207, "fed_destroyer_guys" },
    { 0x3208, "fed_destroyer_osprey" },
    { 0x320B, "fence_bottom_left" },
    { 0x320F, "fence_smash_setup" },
    { 0x3210, "fence_smash_wait" },
    { 0x321A, "ffpoints" },
    { 0x321C, "field_birds" },
    { 0x321D, "fifteen_minutes_earlier_feed_lines" },
    { 0x3220, "fight_objective" },
    { 0x3221, "fightdist" },
    { 0x3225, "fileprint_launcher_start_file" },
    { 0x3226, "fileprint_map_entity_end" },
    { 0x3227, "fileprint_map_entity_start" },
    { 0x322D, "fileprintlauncher_linecount" },
    { 0x3230, "filter" },
    { 0x3232, "filter_spawn_point_by_distance_from_player" },
    { 0x3236, "final_jungle_ai_cleanup" },
    { 0x3237, "final_main" },
    { 0x3239, "final_sequence_fail_condition" },
    { 0x3240, "finale_fx_handling" },
    { 0x3241, "finale_nags" },
    { 0x3242, "finale_sniping" },
    { 0x3243, "finale_timer" },
    { 0x3246, "finalhinds" },
    { 0x324D, "finalkillcam_inflictor" },
    { 0x324E, "finalkillcam_killcamentityindex" },
    { 0x3250, "finalkillcam_psoffsettime" },
    { 0x3252, "finalkillcam_sweapon" },
    { 0x3255, "finalkillcam_victim" },
    { 0x3256, "finalkillcam_winner" },
    { 0x3257, "finalkillcammusic" },
    { 0x325D, "find_available_collision_model" },
    { 0x3260, "find_camp_node_worker" },
    { 0x3264, "find_connected_turrets" },
    { 0x3266, "find_defend_node_capture" },
    { 0x3269, "find_defend_node_protect" },
    { 0x326A, "find_destructibles" },
    { 0x326B, "find_different_way_to_attack_last_seen_position" },
    { 0x326E, "find_lane_spawn_node" },
    { 0x3270, "find_new_chase_target" },
    { 0x3272, "find_random_spawn_node" },
    { 0x3274, "find_safe_spawn_spot_with_volumes" },
    { 0x3278, "find_spitter_attack_node" },
    { 0x327E, "findaveragepointvec" },
    { 0x327F, "findbattlebuddy" },
    { 0x3280, "findboxcenter" },
    { 0x3281, "findbuddypathnode" },
    { 0x3288, "findlocation" },
    { 0x3289, "findnewowner" },
    { 0x328A, "findpointnearowner" },
    { 0x328B, "findrandomtarget" },
    { 0x328D, "findspawnlocationnearplayer" },
    { 0x328E, "findtarget" },
    { 0x3290, "findunobstructedfiringpoint" },
    { 0x3294, "finish_zoom" },
    { 0x3297, "finishcoverexitnotetracks" },
    { 0x329A, "finished_spawning" },
    { 0x32A1, "finishsupportescortusage" },
    { 0x32A4, "fire_anim_start" },
    { 0x32A6, "fire_artillery" },
    { 0x32A8, "fire_at_chopper" },
    { 0x32B1, "fire_cloud_burn_alien" },
    { 0x32B2, "fire_cloud_burn_player" },
    { 0x32B3, "fire_damage" },
    { 0x32B4, "fire_death_watcher" },
    { 0x32B7, "fire_ext_grab" },
    { 0x32BA, "fire_extinguisher" },
    { 0x32BD, "fire_extinguisher_death" },
    { 0x32C6, "fire_fx" },
    { 0x32C8, "fire_grenade" },
    { 0x32C9, "fire_guns" },
    { 0x32CA, "fire_heli_missile" },
    { 0x32CD, "fire_link_structs" },
    { 0x32CF, "fire_magic_missile" },
    { 0x32D0, "fire_magic_missile_mig" },
    { 0x32D2, "fire_missile_at_vehicle" },
    { 0x32D3, "fire_missiles" },
    { 0x32D7, "fire_on" },
    { 0x32D8, "fire_on_non_vehicle" },
    { 0x32DA, "fire_radius" },
    { 0x32DD, "fire_rocket" },
    { 0x32E0, "fire_sat_rcs_thrusters" },
    { 0x32E1, "fire_seeking_missile" },
    { 0x32E2, "fire_sensor" },
    { 0x32E3, "fire_sound_spindown" },
    { 0x32E6, "fire_space_microtar" },
    { 0x32E7, "fire_spit_projectile" },
    { 0x32E9, "fire_supression_status" },
    { 0x32EB, "fire_tracers" },
    { 0x32ED, "fire_tracking_missile_at_vehicle" },
    { 0x32EE, "fire_tracking_missile_mig" },
    { 0x32EF, "fire_trap_burn" },
    { 0x32F6, "fire_turret" },
    { 0x32FA, "firecloudheight" },
    { 0x32FC, "firecloudlingertime" },
    { 0x3300, "firecloudsfx" },
    { 0x3301, "firecloudtickdamage" },
    { 0x3306, "fired_weapon" },
    { 0x3308, "firedirector" },
    { 0x3309, "firedonme" },
    { 0x330B, "fireinterval" },
    { 0x3310, "fireontarget" },
    { 0x3311, "fireonvehicletarget" },
    { 0x3312, "fires" },
    { 0x3313, "fireshield" },
    { 0x331C, "fireteam_hunt_next_zone_search_time" },
    { 0x331D, "fireteam_hunt_target_zone" },
    { 0x331E, "fireteam_menu" },
    { 0x331F, "fireteam_personality_original" },
    { 0x3323, "fireteam_tdm_set_hunt_leader" },
    { 0x3324, "fireteammembers" },
    { 0x3325, "firetime" },
    { 0x3328, "fireuntiloutofammointernal" },
    { 0x332C, "fireworklights" },
    { 0x332D, "fireworks" },
    { 0x332E, "fireworks_courtyard" },
    { 0x332F, "fireworks_courtyard_post" },
    { 0x3330, "fireworks_courtyard_stairs" },
    { 0x3331, "fireworks_finale" },
    { 0x3335, "fireworks_init" },
    { 0x3337, "fireworks_intro_post" },
    { 0x3338, "fireworks_junction" },
    { 0x3339, "fireworks_junction_post" },
    { 0x333B, "fireworks_start" },
    { 0x3341, "firing_into_bunks" },
    { 0x3342, "firing_position" },
    { 0x3343, "firing_sound_ent" },
    { 0x3347, "firingreaper" },
    { 0x3348, "first_distant_sat_launch" },
    { 0x3349, "first_explosion_fx" },
    { 0x334C, "first_floor_patroller_1_setup" },
    { 0x334D, "first_floor_patroller_2_setup" },
    { 0x334E, "first_floor_shift_right_trigger" },
    { 0x334F, "first_floor_sitting_laptop_guy_react" },
    { 0x3350, "first_floor_sitting_laptop_guy_react_laptop" },
    { 0x3352, "first_frame" },
    { 0x3356, "first_guy_leap_frog_logic" },
    { 0x335E, "first_time_depth_charge_minigun_hint" },
    { 0x3361, "first_zoom_level_fov" },
    { 0x3362, "firstblood" },
    { 0x3364, "firstcapture" },
    { 0x3365, "firstcontact" },
    { 0x3366, "firstcratedrop" },
    { 0x3367, "firstframe_combat_one_door" },
    { 0x336A, "firstinit" },
    { 0x336C, "firstload" },
    { 0x3370, "fish" },
    { 0x3377, "fizzle_flashbangs_think" },
    { 0x3378, "fizzle_tracked" },
    { 0x337E, "flag_count" },
    { 0x3383, "flag_exist" },
    { 0x3384, "flag_if_player_aims_knife_at_enemy" },
    { 0x3386, "flag_inits" },
    { 0x3387, "flag_on_death" },
    { 0x3388, "flag_set" },
    { 0x3389, "flag_set_delayed" },
    { 0x338A, "flag_set_near_ent" },
    { 0x338C, "flag_state" },
    { 0x3390, "flag_trigger_init" },
    { 0x3394, "flag_wait" },
    { 0x3395, "flag_wait_all" },
    { 0x3396, "flag_wait_any" },
    { 0x3399, "flag_wait_badplace_brush" },
    { 0x339C, "flag_wait_either_return" },
    { 0x339F, "flag_wait_or_timeout" },
    { 0x33A1, "flag_waitopen_any" },
    { 0x33A3, "flag_watcher" },
    { 0x33A4, "flagbasefxid" },
    { 0x33A9, "flagtowait" },
    { 0x33AA, "flagwaitthread" },
    { 0x33AB, "flagwaitthread_proc" },
    { 0x33AD, "flamestack_godrays" },
    { 0x33AF, "flare_doburnout" },
    { 0x33B2, "flare_fx" },
    { 0x33B5, "flare_knife" },
    { 0x33B9, "flare_trackvelocity" },
    { 0x33BB, "flareactiveradius" },
    { 0x33BD, "flareauto" },
    { 0x33C1, "flarefx" },
    { 0x33C2, "flarefxexplode" },
    { 0x33C3, "flareindex" },
    { 0x33C5, "flareprotection" },
    { 0x33C7, "flarerig_animrate" },
    { 0x33CA, "flarerig_tagangles" },
    { 0x33CC, "flares" },
    { 0x33D2, "flares_fire" },
    { 0x33DB, "flares_playfx" },
    { 0x33DC, "flares_redirect_missiles" },
    { 0x33DE, "flares_watchsamproximity" },
    { 0x33E1, "flaresound" },
    { 0x33EA, "flarestack_door" },
    { 0x33EC, "flarestack_pressurelp_01" },
    { 0x33ED, "flarestack_pressurelp_02" },
    { 0x33F4, "flash_credit_gain" },
    { 0x33F5, "flash_fxs" },
    { 0x33F7, "flashanimindex" },
    { 0x33FA, "flashbangedloop" },
    { 0x33FC, "flashbangimmunity" },
    { 0x33FD, "flashbangisactive" },
    { 0x33FF, "flashbangstart" },
    { 0x3400, "flashbangstop" },
    { 0x3403, "flashendtime" },
    { 0x3408, "flashlight_off" },
    { 0x340A, "flashlight_prop" },
    { 0x340C, "flashlight_toggle" },
    { 0x340D, "flashmonitor" },
    { 0x340E, "flashmonitorenablehealthshield" },
    { 0x3412, "flashrumbleloop" },
    { 0x3418, "flat_trigger" },
    { 0x3428, "flinching" },
    { 0x342B, "fling_debug" },
    { 0x3434, "flock_handledamage" },
    { 0x3436, "flood_amb_fx" },
    { 0x3438, "flood_and_secure_spawn" },
    { 0x3439, "flood_and_secure_spawn_goal" },
    { 0x343A, "flood_and_secure_spawner" },
    { 0x343B, "flood_and_secure_spawner_think" },
    { 0x343C, "flood_battlechatter_on" },
    { 0x343E, "flood_convoy_attackheli01_sfx" },
    { 0x3441, "flood_convoy_chopper2_sfx" },
    { 0x3443, "flood_convoy_exp_sfx" },
    { 0x3444, "flood_convoy_sfx" },
    { 0x3445, "flood_createfx_cleanup" },
    { 0x3446, "flood_current_goalnode" },
    { 0x3447, "flood_default_water_transion_fx" },
    { 0x3449, "flood_ending_fadeout" },
    { 0x344D, "flood_interactives_cleanup" },
    { 0x344E, "flood_intro_screen" },
    { 0x3450, "flood_launcher_crash_sfx" },
    { 0x3451, "flood_locations" },
    { 0x3452, "flood_mall_roof_door" },
    { 0x3453, "flood_mall_roof_door_model" },
    { 0x3459, "flood_og_movetransitionrate" },
    { 0x345B, "flood_player_default_jump_height" },
    { 0x345C, "flood_shake_tree" },
    { 0x345D, "flood_shake_tree_internal" },
    { 0x345F, "flood_spawn" },
    { 0x3462, "flood_spawner_scripted" },
    { 0x3463, "flood_spawner_stop" },
    { 0x3464, "flood_spawner_think" },
    { 0x3465, "flood_sweptaway" },
    { 0x346A, "flooding_ext_start" },
    { 0x346C, "flooding_int_start" },
    { 0x346D, "flooding_last_water_state" },
    { 0x346E, "flooding_stairs_vo" },
    { 0x3470, "flush_church_proc" },
    { 0x3471, "flush_sat_camp_1" },
    { 0x3473, "fly_around_start" },
    { 0x3476, "fly_distance" },
    { 0x347B, "flyout_lights" },
    { 0x347C, "fndogmeleevictim" },
    { 0x347E, "fnoverlord" },
    { 0x3487, "folded_sentry_use_wait" },
    { 0x3488, "foley" },
    { 0x348B, "follow_icon_manager" },
    { 0x348C, "follow_on_went_hot" },
    { 0x348D, "follow_on_went_hot_logic" },
    { 0x348E, "follow_origin" },
    { 0x348F, "follow_path" },
    { 0x3491, "follow_path_animate_set_ent" },
    { 0x3492, "follow_path_animate_set_node" },
    { 0x3493, "follow_path_animate_set_struct" },
    { 0x349A, "font_color" },
    { 0x349B, "font_size" },
    { 0x349D, "fontpulse" },
    { 0x349F, "fontscaler" },
    { 0x34A0, "foo" },
    { 0x34A3, "force_actor_space_rotation_update" },
    { 0x34A5, "force_all_complete" },
    { 0x34A7, "force_civilian_hunched_run" },
    { 0x34AB, "force_crawling_death_proc" },
    { 0x34AC, "force_crouch" },
    { 0x34AD, "force_deathquote" },
    { 0x34B2, "force_hint" },
    { 0x34B6, "force_mantle_trigs" },
    { 0x34B7, "force_next_earthquake" },
    { 0x34B8, "force_num_crawls" },
    { 0x34BB, "force_player_to_end_upload" },
    { 0x34BF, "force_stop_sound" },
    { 0x34C7, "forced_hint" },
    { 0x34C8, "forced_slowmo_breach_lerpout" },
    { 0x34CA, "forced_start_catchup" },
    { 0x34CB, "forced_startingposition" },
    { 0x34CC, "forcedend" },
    { 0x34CD, "forcedgameskill" },
    { 0x34D0, "forcedtarget" },
    { 0x34D6, "forcegoal" },
    { 0x34D7, "forcegoal_radius" },
    { 0x34DE, "forcespawnangles" },
    { 0x34E2, "forcetrigger" },
    { 0x34E5, "foreman" },
    { 0x34E7, "forfeitinprogress" },
    { 0x34E8, "forfeitwaitforabort" },
    { 0x34E9, "forget_time" },
    { 0x34EA, "format1" },
    { 0x34EB, "format2" },
    { 0x34ED, "forty_five_sec_vo" },
    { 0x34EE, "forward_direction_worldspace" },
    { 0x34EF, "forward_for_vision_change" },
    { 0x34F0, "forwardgraphid" },
    { 0x34F1, "forwardpush" },
    { 0x34F3, "forwardyaw" },
    { 0x34FB, "fov_get_default" },
    { 0x34FD, "fov_orig" },
    { 0x34FE, "fov_outer" },
    { 0x34FF, "fov_range" },
    { 0x3501, "frac" },
    { 0x3504, "frames_visible" },
    { 0x3505, "free_node" },
    { 0x3506, "free_on_death" },
    { 0x350A, "freeplayer" },
    { 0x350D, "freezeallplayers" },
    { 0x3510, "freon_leak_fx_turn_off_damage" },
    { 0x3512, "fridge_anims" },
    { 0x3516, "friendlies_breach" },
    { 0x3517, "friendlies_execute_enemies" },
    { 0x3518, "friendlies_shoot_while_breaching" },
    { 0x351F, "friendly_breach_thread" },
    { 0x3520, "friendly_bubbles" },
    { 0x3525, "friendly_color_hidden" },
    { 0x3526, "friendly_color_spotted" },
    { 0x352A, "friendly_custom_acc_behavior" },
    { 0x352D, "friendly_default_acc_behavior" },
    { 0x352E, "friendly_default_color_behavior" },
    { 0x3533, "friendly_fire_enable" },
    { 0x3534, "friendly_fire_fail_check" },
    { 0x3536, "friendly_getangles_ai" },
    { 0x3537, "friendly_getangles_player" },
    { 0x353A, "friendly_getvelocity" },
    { 0x353C, "friendly_hud_icon_blink_on_damage" },
    { 0x3540, "friendly_init" },
    { 0x3545, "friendly_mg42_doneusingturret" },
    { 0x3546, "friendly_mg42_endtrigger" },
    { 0x354B, "friendly_mgturret" },
    { 0x354C, "friendly_navigation" },
    { 0x354D, "friendly_nearby" },
    { 0x354E, "friendly_promotion_thread" },
    { 0x3550, "friendly_respawn_vision_checker_thread" },
    { 0x3553, "friendly_setup" },
    { 0x3554, "friendly_setup_apache_section" },
    { 0x3556, "friendly_spawner" },
    { 0x3558, "friendly_spotted_getup_from_prone" },
    { 0x3559, "friendly_stance_handler" },
    { 0x355C, "friendly_stance_handler_check_mightbeseen" },
    { 0x355E, "friendly_stance_handler_return_ai_sight" },
    { 0x3568, "friendly_visibility_logic" },
    { 0x356B, "friendly_wave_masterthread" },
    { 0x356C, "friendly_wave_trigger" },
    { 0x3570, "friendlychains" },
    { 0x3574, "friendlyfire_damage_modifier" },
    { 0x3575, "friendlyfire_destructible_attacker" },
    { 0x3577, "friendlyfire_max_participation" },
    { 0x3578, "friendlyfire_shield" },
    { 0x3579, "friendlyfire_warning" },
    { 0x357B, "friendlyfire_warnings_disable" },
    { 0x357C, "friendlyfire_warnings_off" },
    { 0x357E, "friendlyfire_whizby_distances_valid" },
    { 0x3581, "friendlyfiredisabledfordestructible" },
    { 0x3587, "friendlyhudicon_normal" },
    { 0x3588, "friendlyhudicon_rotating" },
    { 0x3589, "friendlyhudicon_showall" },
    { 0x358A, "friendlyhudicon_update" },
    { 0x358B, "friendlyicon" },
    { 0x358C, "friendlymodel" },
    { 0x358D, "friendlyspawnorg" },
    { 0x358F, "friendlyspawnwave" },
    { 0x3590, "friendlyspawnwave_triggerthink" },
    { 0x3591, "friendlytrigger" },
    { 0x3592, "friendlywave_thread" },
    { 0x3593, "friendlywaypoint" },
    { 0x359D, "front_wheel_sfx" },
    { 0x35A0, "fudge_ally_accuracy" },
    { 0x35A3, "fullmodel" },
    { 0x35A5, "func" },
    { 0x35AA, "func_get_level_fx" },
    { 0x35AB, "func_get_nodes_on_path" },
    { 0x35AD, "func_give" },
    { 0x35B2, "func_position_player" },
    { 0x35B3, "func_position_player_get" },
    { 0x35B6, "func_waittill_msg" },
    { 0x35B7, "function_stack" },
    { 0x35B8, "function_stack_caller_waits_for_turn" },
    { 0x35B9, "function_stack_clear" },
    { 0x35BE, "function_stack_wait" },
    { 0x35BF, "function_stack_wait_finish" },
    { 0x35C2, "funnel_player_internal" },
    { 0x35C3, "furniturepushsound" },
    { 0x35C7, "fx_airlock_ambient" },
    { 0x35C8, "fx_airstrike_afterburner" },
    { 0x35C9, "fx_airstrike_contrail" },
    { 0x35CB, "fx_airstrike_wingtip_light_red" },
    { 0x35CE, "fx_ambient_setup" },
    { 0x35CF, "fx_angry_flood_nearmiss" },
    { 0x35D0, "fx_array" },
    { 0x35D1, "fx_aurora_anim" },
    { 0x35D2, "fx_blood_splatter" },
    { 0x35D3, "fx_bokehdots_and_waterdrops_heavy" },
    { 0x35D5, "fx_bokehdots_far" },
    { 0x35D8, "fx_bridgefall_med_splash" },
    { 0x35D9, "fx_bridgefall_small_splash" },
    { 0x35DB, "fx_burnup_kyra" },
    { 0x35DC, "fx_burnup_player" },
    { 0x35DD, "fx_burnup_sat" },
    { 0x35DE, "fx_c2_rog_satelittes_close_01" },
    { 0x35E1, "fx_camp_truck_submerge_body" },
    { 0x35E5, "fx_charge2_pos" },
    { 0x35E9, "fx_command_snow" },
    { 0x35EB, "fx_command_window_light_on" },
    { 0x35EF, "fx_crater_plume" },
    { 0x35F0, "fx_create_bokehdots_source" },
    { 0x35F3, "fx_dam_explosion" },
    { 0x35F5, "fx_dam_missile_dust" },
    { 0x35F6, "fx_dam_missile_launch_01" },
    { 0x35FC, "fx_debri_combat_one" },
    { 0x35FD, "fx_door_open" },
    { 0x3605, "fx_ents_playfx" },
    { 0x3606, "fx_escape_ambient" },
    { 0x3608, "fx_escape_play_rog_impact" },
    { 0x360B, "fx_flarestack_motion" },
    { 0x3610, "fx_glass_front" },
    { 0x3612, "fx_glass_mid_1" },
    { 0x3620, "fx_init" },
    { 0x3621, "fx_intro_amb" },
    { 0x3622, "fx_intro_ambient" },
    { 0x3623, "fx_intro_friendly_glowsticks" },
    { 0x3624, "fx_jump_ambient" },
    { 0x3627, "fx_lens_drops_01" },
    { 0x362A, "fx_lens_drops_04" },
    { 0x362D, "fx_like_dust_in_the_wind_02" },
    { 0x362E, "fx_like_dust_in_the_wind_03" },
    { 0x3632, "fx_mall_roof_water_hide" },
    { 0x3633, "fx_mall_roof_water_show" },
    { 0x3634, "fx_mall_rooftop_debris" },
    { 0x3635, "fx_mall_rooftop_hide_shadow_geo" },
    { 0x3636, "fx_missile_launch" },
    { 0x3639, "fx_opfor3_pushdown_bubbles" },
    { 0x363B, "fx_org" },
    { 0x363D, "fx_org_belly2" },
    { 0x363E, "fx_org_body" },
    { 0x3645, "fx_origin_with_train_angles" },
    { 0x3646, "fx_parking_garage_ambient" },
    { 0x3650, "fx_retarget_warehouse_waters_lighting" },
    { 0x3651, "fx_rog_amb_exp" },
    { 0x3653, "fx_rog_satelittes_close_02" },
    { 0x3654, "fx_rog_satelittes_close_03" },
    { 0x3655, "fx_rog_satelittes_fire_fx_close" },
    { 0x3657, "fx_rog_satelittes_firing_c2" },
    { 0x3659, "fx_rooftop2_ambient" },
    { 0x365E, "fx_rotate_lights" },
    { 0x3660, "fx_rotating" },
    { 0x3662, "fx_sat_doors_close" },
    { 0x3663, "fx_sat_doors_open" },
    { 0x3665, "fx_sat_rcs_damage" },
    { 0x3666, "fx_sat_rcs_damage_kill" },
    { 0x3667, "fx_sat_thrusters_damage" },
    { 0x3669, "fx_screen_bokehdots_rain" },
    { 0x366D, "fx_screen_raindrops_fast_kill" },
    { 0x366E, "fx_screen_water_sheeting" },
    { 0x366F, "fx_sea_spray" },
    { 0x3675, "fx_skybridge_event" },
    { 0x3676, "fx_skybridge_room_bokeh_01" },
    { 0x3677, "fx_skybridge_room_bokeh_02" },
    { 0x3678, "fx_snakecam_lens_shmutz" },
    { 0x367B, "fx_snowmobile_light" },
    { 0x367D, "fx_solar_panel_collision_player" },
    { 0x367F, "fx_space_glass" },
    { 0x3682, "fx_spin_play_rog_impact" },
    { 0x3683, "fx_spin_player_debris" },
    { 0x3686, "fx_stealth_ambient" },
    { 0x3688, "fx_stealthkill_02_blood_02" },
    { 0x3689, "fx_stealthkill_02_opfor2_blood_01" },
    { 0x368C, "fx_swept_amb_fx" },
    { 0x368E, "fx_swept_water_hide" },
    { 0x3691, "fx_tag" },
    { 0x3694, "fx_tag_prop_wash" },
    { 0x3695, "fx_tag_prop_wash1" },
    { 0x3697, "fx_tank_window_break" },
    { 0x369A, "fx_turn_on_bokehdots_16_player" },
    { 0x369C, "fx_turn_on_bokehdots_32_player" },
    { 0x369E, "fx_turn_on_introfx" },
    { 0x369F, "fx_turn_on_tunnel_lights" },
    { 0x36A0, "fx_turnon_loco_exterior_lights" },
    { 0x36A2, "fx_turnon_rooftop_lights" },
    { 0x36B0, "fx_warehouse_door_burst" },
    { 0x36B2, "fx_warehouse_double_doors" },
    { 0x36B3, "fx_warehouse_floating_debris" },
    { 0x36B5, "fx_warehouse_splashes" },
    { 0x36B8, "fx_warehouse_underwater_fx_on" },
    { 0x36BD, "fx_waterwipe_above" },
    { 0x36BE, "fx_waterwipe_under" },
    { 0x36BF, "fx_wh_splashes" },
    { 0x36C0, "fxcost" },
    { 0x36C4, "fxfireloopmod" },
    { 0x36C5, "fxhudelements" },
    { 0x36C8, "fxid_explode" },
    { 0x36CB, "fxid_light2" },
    { 0x36D3, "fxstop" },
    { 0x36D5, "fxtags" },
    { 0x36D6, "fxtime" },
    { 0x36D8, "g_bob_scale_get_func" },
    { 0x36D9, "g_bob_scale_set_func" },
    { 0x36DB, "g_friendlynamedist_old" },
    { 0x36DC, "g_speed" },
    { 0x36DD, "g_speed_get_func" },
    { 0x36DE, "g_speed_set_func" },
    { 0x36DF, "gain" },
    { 0x36E0, "gain_credit_message" },
    { 0x36E1, "gamblerability" },
    { 0x36E2, "gambleranimwatcher" },
    { 0x36E3, "game_is_current_gen" },
    { 0x36E4, "game_is_ng" },
    { 0x36EE, "gameended" },
    { 0x36EF, "gameendlistener" },
    { 0x36F5, "gameflagwait" },
    { 0x36FA, "gamemodemaydropweapon" },
    { 0x36FB, "gamemodemodifyplayerdamage" },
    { 0x36FE, "gameskill" },
    { 0x3700, "gameskill_is_difficult" },
    { 0x3709, "gapfixlevel" },
    { 0x370A, "garage" },
    { 0x370D, "garage_ally_move476_jumpers" },
    { 0x3711, "garage_ally_move480" },
    { 0x3713, "garage_door_logic" },
    { 0x3714, "garage_door_nag" },
    { 0x3715, "garage_jeep_start_skid" },
    { 0x3716, "garage_jump_01" },
    { 0x3719, "garage_jump_02_spawn" },
    { 0x371A, "garage_malllight_off" },
    { 0x371B, "garage_misc" },
    { 0x371C, "garage_opening_collapse" },
    { 0x371F, "garage_start" },
    { 0x3722, "garage_velocity_loops" },
    { 0x3725, "garage_wave3_trig" },
    { 0x3726, "garage_wave4_trig" },
    { 0x3727, "garage_wave_get_furthest" },
    { 0x3729, "garden_enemies_cyan_needed" },
    { 0x372A, "garden_enemies_delete" },
    { 0x372F, "garden_entity_cleanup" },
    { 0x3731, "garden_gameplay" },
    { 0x3732, "garden_glass_room_guy_spawnfunc" },
    { 0x3733, "garden_idf_guys_orange" },
    { 0x3737, "garden_last_stand" },
    { 0x3738, "garden_move_allies_to_end" },
    { 0x3739, "garden_transient_sync" },
    { 0x373A, "garden_vo" },
    { 0x373B, "garden_wave1" },
    { 0x373F, "gas_cloud_available" },
    { 0x3742, "gas_flee_guys" },
    { 0x3743, "gas_mask_off_player_anim" },
    { 0x3746, "gas_station" },
    { 0x3748, "gas_station_connect_node" },
    { 0x3749, "gas_station_disconnect_node" },
    { 0x374A, "gas_station_earthquake" },
    { 0x374B, "gas_station_events" },
    { 0x374E, "gas_station_play_fx" },
    { 0x3750, "gas_station_precache" },
    { 0x3752, "gas_station_run_func_on_notify" },
    { 0x3753, "gas_station_update_clip" },
    { 0x3755, "gasmask" },
    { 0x3756, "gasmask_breathing" },
    { 0x3757, "gasmask_hud_elem" },
    { 0x375C, "gasstation_ambient_aa72" },
    { 0x375F, "gasstation_civs_logic" },
    { 0x3760, "gasstation_did_player_rush" },
    { 0x3763, "gasstation_execution_timing" },
    { 0x3766, "gasstation_takedown_knife_notetrack_func" },
    { 0x3768, "gate_being_used" },
    { 0x3769, "gateopen" },
    { 0x376A, "gates" },
    { 0x376C, "gather_delay_proc" },
    { 0x376F, "gaz_additive_anims_release" },
    { 0x3770, "gaz_crush" },
    { 0x3777, "gaz_relative_speed_stop" },
    { 0x3778, "gaz_spawn_setup" },
    { 0x3779, "gaz_turret_guy_gettin_func" },
    { 0x377A, "gazvehicles" },
    { 0x377E, "generatemaxweightedcratevalue" },
    { 0x3780, "generatesquadname" },
    { 0x3784, "generatortype" },
    { 0x3788, "generic_double_strobe" },
    { 0x3789, "generic_flicker" },
    { 0x378E, "generic_gaz_spawner_setup" },
    { 0x3792, "generic_human_beach" },
    { 0x3793, "generic_human_intro" },
    { 0x3794, "generic_human_recruits" },
    { 0x3797, "generic_index" },
    { 0x379C, "generic_shootable_double_doors" },
    { 0x379D, "generic_spawn_node_based_enemy_gaz" },
    { 0x37A5, "generic_swing_ents" },
    { 0x37AA, "geo_off" },
    { 0x37AB, "geo_on" },
    { 0x37AC, "get_a10_player_start" },
    { 0x37AD, "get_activation_result" },
    { 0x37AE, "get_active_turret" },
    { 0x37AF, "get_adjusted_armor" },
    { 0x37B0, "get_ai_array" },
    { 0x37B2, "get_ai_group_count" },
    { 0x37B4, "get_ai_number" },
    { 0x37B5, "get_ai_touching_volume" },
    { 0x37B6, "get_alias_from_stored" },
    { 0x37BB, "get_all_allies" },
    { 0x37C2, "get_all_wave_guys" },
    { 0x37C3, "get_altitude_min" },
    { 0x37C4, "get_angered_damage_scalar" },
    { 0x37C6, "get_anim_class" },
    { 0x37C9, "get_anim_start_time" },
    { 0x37CD, "get_apache_ally" },
    { 0x37CF, "get_apache_player" },
    { 0x37D1, "get_apache_spawn_struct" },
    { 0x37D3, "get_approach_node" },
    { 0x37D4, "get_area_for_cycle" },
    { 0x37D6, "get_array_of_farthest" },
    { 0x37DA, "get_attack_num" },
    { 0x37E0, "get_average_distance_to_players" },
    { 0x37E2, "get_badcountryidstr" },
    { 0x37EA, "get_best_weapons" },
    { 0x37F0, "get_blended_difficulty" },
    { 0x37F1, "get_blendtime_from_notetrack" },
    { 0x37F6, "get_breach_groups" },
    { 0x37F9, "get_breach_target" },
    { 0x3800, "get_challenge_scalar" },
    { 0x3801, "get_channel_array" },
    { 0x3802, "get_chaos_airdrop_loc" },
    { 0x3804, "get_choke_trig_id" },
    { 0x3805, "get_chopperboss_data" },
    { 0x3806, "get_clone_agent" },
    { 0x3809, "get_closest_ally" },
    { 0x380A, "get_closest_colored_friendly" },
    { 0x380B, "get_closest_colored_friendly_with_classname" },
    { 0x3812, "get_closest_player_healthy" },
    { 0x3815, "get_closest_to_player_view" },
    { 0x3816, "get_closest_vec_index" },
    { 0x3817, "get_closest_with_targetname" },
    { 0x3819, "get_color_index_item" },
    { 0x381F, "get_color_volume_from_trigger" },
    { 0x3820, "get_colorcoded_volume" },
    { 0x3822, "get_colorcodes_and_activate_trigger" },
    { 0x382D, "get_cumulative_weights" },
    { 0x382E, "get_curfloor" },
    { 0x382F, "get_current_agent_count" },
    { 0x3834, "get_current_hive" },
    { 0x3835, "get_current_hive_lane_index" },
    { 0x3838, "get_current_sam" },
    { 0x383B, "get_current_vo_alias" },
    { 0x3845, "get_damageable_mine" },
    { 0x3846, "get_damageable_player" },
    { 0x3849, "get_datascene" },
    { 0x384B, "get_death_anim_physics_mode" },
    { 0x384D, "get_delay" },
    { 0x384E, "get_desc_by_ref" },
    { 0x3852, "get_direction_away_from_players" },
    { 0x3855, "get_dot" },
    { 0x3856, "get_downed_posture_node" },
    { 0x3857, "get_download_state_hud" },
    { 0x3858, "get_dpad_down_level" },
    { 0x385B, "get_dpad_right_level" },
    { 0x385E, "get_drill_widget_color" },
    { 0x385F, "get_drones_touching_volume" },
    { 0x3860, "get_drones_with_targetname" },
    { 0x3861, "get_drop_loop_center" },
    { 0x3866, "get_edge_tiles" },
    { 0x3867, "get_elem_position" },
    { 0x3868, "get_elevated_jump_node" },
    { 0x386B, "get_enemy_count_touching_volume" },
    { 0x3871, "get_ent" },
    { 0x3873, "get_entry" },
    { 0x3877, "get_explosion_radius" },
    { 0x387A, "get_eye_position" },
    { 0x387C, "get_flare_node" },
    { 0x387D, "get_flashed_anim" },
    { 0x3884, "get_force_color" },
    { 0x3885, "get_force_color_guys" },
    { 0x3886, "get_fov" },
    { 0x3888, "get_free_edge_node" },
    { 0x3889, "get_free_shoot_node" },
    { 0x388A, "get_friendly_crate_model" },
    { 0x388B, "get_friendly_juggernaut_crate_model" },
    { 0x388D, "get_from_entity_target" },
    { 0x388E, "get_from_spawnstruct" },
    { 0x388F, "get_from_spawnstruct_target" },
    { 0x3893, "get_func_give" },
    { 0x3894, "get_furthest_zodiac" },
    { 0x3895, "get_generic_anime" },
    { 0x389A, "get_gun_tag" },
    { 0x389D, "get_helicopter_crash_location" },
    { 0x389E, "get_heroes" },
    { 0x38A2, "get_hintstring_for_item_pickup" },
    { 0x38A3, "get_hintstring_for_pillaged_item" },
    { 0x38A5, "get_hives_destroyed_stat" },
    { 0x38A9, "get_housing_door_trigger" },
    { 0x38AF, "get_housing_musak_model" },
    { 0x38B0, "get_housing_primarylight" },
    { 0x38B3, "get_human_player" },
    { 0x38B6, "get_idle_anim" },
    { 0x38B9, "get_in_moving_vehicle" },
    { 0x38BA, "get_in_truck_nag_vo" },
    { 0x38BC, "get_in_world_area" },
    { 0x38C1, "get_is_upgrade_by_ref" },
    { 0x38C3, "get_item_desc" },
    { 0x38C4, "get_item_name" },
    { 0x38C5, "get_item_outline_color" },
    { 0x38C6, "get_jump_info" },
    { 0x38C7, "get_key" },
    { 0x38C9, "get_knife_reticle" },
    { 0x38CD, "get_last_selected_ent" },
    { 0x38CF, "get_latest_struct" },
    { 0x38D3, "get_leper_retreat_node" },
    { 0x38D6, "get_leveltime" },
    { 0x38D8, "get_linked_ents" },
    { 0x38DA, "get_linked_nodes_and_delete" },
    { 0x38DB, "get_linked_points" },
    { 0x38DC, "get_linked_struct" },
    { 0x38DD, "get_linked_structs" },
    { 0x38DE, "get_linked_vehicle_node" },
    { 0x38DF, "get_linked_vehicle_nodes" },
    { 0x38E0, "get_links" },
    { 0x38E1, "get_living_ai" },
    { 0x38E2, "get_living_ai_array" },
    { 0x38E5, "get_load_trigger_classes" },
    { 0x38E6, "get_load_trigger_funcs" },
    { 0x38E7, "get_loadout" },
    { 0x38EB, "get_locked_difficulty_val_global" },
    { 0x38EC, "get_locked_difficulty_val_player" },
    { 0x38EE, "get_lookahead_target_location" },
    { 0x38F1, "get_manhandled" },
    { 0x38F2, "get_map_soundtable" },
    { 0x38F3, "get_max_alien_count" },
    { 0x38F6, "get_max_sentry_count" },
    { 0x38FA, "get_maxxp_by_id" },
    { 0x38FB, "get_melee_painstate_info" },
    { 0x38FE, "get_meteoroid_impact_node_escape" },
    { 0x3901, "get_minutes_and_seconds" },
    { 0x3904, "get_modified_alien_amount" },
    { 0x3909, "get_name" },
    { 0x390A, "get_name_by_ref" },
    { 0x390C, "get_nerf_scalar" },
    { 0x390D, "get_next_allow_melee_time" },
    { 0x3910, "get_node_funcs_based_on_target" },
    { 0x3912, "get_noteworthy_ent" },
    { 0x3916, "get_obj_ent_hvt" },
    { 0x3917, "get_obj_event" },
    { 0x3918, "get_obj_origin" },
    { 0x3919, "get_obstacle_dodge_amount" },
    { 0x391B, "get_offset_percent" },
    { 0x391D, "get_optimal_next_path_los_check" },
    { 0x3923, "get_out_override" },
    { 0x3927, "get_outer_leftdoor" },
    { 0x392B, "get_outside_range" },
    { 0x392E, "get_pacifist" },
    { 0x3930, "get_part_fx_cost_for_action_state" },
    { 0x3931, "get_parts" },
    { 0x3932, "get_path_end_node" },
    { 0x3933, "get_path_getfunc" },
    { 0x3936, "get_perk_1_level" },
    { 0x3937, "get_perk_ref_at_upgrade_level" },
    { 0x3939, "get_pet_follow_node" },
    { 0x393A, "get_photo_copier" },
    { 0x393B, "get_phraseinvalidstr" },
    { 0x393E, "get_pistol_plane_attack" },
    { 0x393F, "get_player_currency" },
    { 0x3940, "get_player_escaped" },
    { 0x3941, "get_player_feet_from_view" },
    { 0x3943, "get_player_gameskill" },
    { 0x3944, "get_player_hesco" },
    { 0x3946, "get_player_loop_center" },
    { 0x3947, "get_player_max_currency" },
    { 0x394D, "get_player_rig" },
    { 0x394E, "get_player_score" },
    { 0x394F, "get_player_session_rankup" },
    { 0x3951, "get_player_targ" },
    { 0x3955, "get_player_xp" },
    { 0x3956, "get_players" },
    { 0x395C, "get_portable_mg_spot" },
    { 0x395D, "get_position_from_progress" },
    { 0x3961, "get_possible_attachments_by_weaponclass" },
    { 0x3967, "get_primary_death_anim_state" },
    { 0x396B, "get_prioritized_colorcoded_nodes" },
    { 0x3971, "get_random_death_sound" },
    { 0x3974, "get_random_firework_num" },
    { 0x3977, "get_random_streamer_num" },
    { 0x397D, "get_ref_by_id" },
    { 0x397E, "get_reflected_point" },
    { 0x3980, "get_resource_ref_by_index" },
    { 0x3982, "get_respawn_delay_by_index" },
    { 0x3983, "get_respawn_threshold_by_index" },
    { 0x3984, "get_retreat_node_rated" },
    { 0x3987, "get_rider_by_position" },
    { 0x398A, "get_rl_toward" },
    { 0x398D, "get_rumble_ent" },
    { 0x398E, "get_rumble_ent_linked" },
    { 0x398F, "get_scaled_alien_amount" },
    { 0x3991, "get_script_linkto_targets" },
    { 0x3993, "get_selected_dpad_down" },
    { 0x3997, "get_selected_move_vector" },
    { 0x3998, "get_selected_nerf" },
    { 0x399F, "get_shuffle_to_corner_start_anim" },
    { 0x39A0, "get_side_next_missile" },
    { 0x39A1, "get_sign" },
    { 0x39A2, "get_skill_from_index" },
    { 0x39A3, "get_sound_ent" },
    { 0x39A4, "get_spawn_event_types_array" },
    { 0x39A5, "get_spawn_event_wave_blocking_by_index" },
    { 0x39A6, "get_spawn_position" },
    { 0x39AD, "get_start_dvars" },
    { 0x39B1, "get_strip_color" },
    { 0x39B2, "get_strip_settings" },
    { 0x39B4, "get_suppress_point" },
    { 0x39B5, "get_surface_types" },
    { 0x39BE, "get_target_ent_origin" },
    { 0x39C0, "get_target_ent_target_ent" },
    { 0x39C2, "get_target_ents" },
    { 0x39C6, "get_team_prefix" },
    { 0x39C7, "get_team_substr" },
    { 0x39C9, "get_teleport_optimized_breachfriendly" },
    { 0x39CD, "get_to_train_wait_node" },
    { 0x39CE, "get_tool_hudelem" },
    { 0x39CF, "get_touching_tiles" },
    { 0x39D0, "get_trace_loc_for_target" },
    { 0x39D1, "get_trace_types" },
    { 0x39D7, "get_trigger_targs" },
    { 0x39D8, "get_turret_setup_anim" },
    { 0x39DA, "get_type_data" },
    { 0x39DB, "get_types_array" },
    { 0x39E2, "get_use_trigger" },
    { 0x39E3, "get_valid_challenge" },
    { 0x39E9, "get_vehicle_ai_spawners" },
    { 0x39EA, "get_vehicle_array" },
    { 0x39F4, "get_wash_effect" },
    { 0x39F6, "get_wave_data" },
    { 0x39F7, "get_waves" },
    { 0x39F9, "get_weapon_outline_color" },
    { 0x39FE, "get_white_overlay" },
    { 0x39FF, "get_within_range" },
    { 0x3A00, "get_world_item_spawn_pos" },
    { 0x3A05, "get_zone_to" },
    { 0x3A0A, "getactionbind" },
    { 0x3A0B, "getactiveagentsoftype" },
    { 0x3A0C, "getactiveplayerlist" },
    { 0x3A0D, "getagentdamagescalar" },
    { 0x3A0F, "getaicurrentweaponslot" },
    { 0x3A10, "getaimpitchtoshootentorpos" },
    { 0x3A14, "getairstrikedanger" },
    { 0x3A16, "getaisidearmweapon" },
    { 0x3A17, "getaliastypefromsoundalias" },
    { 0x3A18, "getallactiveparts" },
    { 0x3A19, "getallweapons" },
    { 0x3A1B, "getangleindex" },
    { 0x3A1E, "getangles_func" },
    { 0x3A24, "getanimscalefactors" },
    { 0x3A27, "getarrivalnode" },
    { 0x3A28, "getarrivalprestartpos" },
    { 0x3A2B, "getattachmentlistbasenames" },
    { 0x3A2C, "getattachmentlistuniqenames" },
    { 0x3A2D, "getattachmenttype" },
    { 0x3A2E, "getattackpoint" },
    { 0x3A31, "getaverageplayerorigin" },
    { 0x3A32, "getawardrecord" },
    { 0x3A34, "getawardtype" },
    { 0x3A35, "getawardwinners" },
    { 0x3A3B, "getbestminitarget" },
    { 0x3A3F, "getbestspawnpoint" },
    { 0x3A40, "getbeststepoutpos" },
    { 0x3A41, "getbesttarget" },
    { 0x3A42, "getbetterplayer" },
    { 0x3A45, "getbuddyspawnangles" },
    { 0x3A4A, "getcapxpscale" },
    { 0x3A4C, "getchain" },
    { 0x3A4F, "getchallengestatus" },
    { 0x3A57, "getclosest2d" },
    { 0x3A59, "getclosestfriendlyspeaker" },
    { 0x3A5A, "getclosestfx" },
    { 0x3A5C, "getclosestnode" },
    { 0x3A5D, "getclosests_flickering_model" },
    { 0x3A5E, "getcloseststartnode" },
    { 0x3A60, "getcombatspeedscalar" },
    { 0x3A62, "getcorrectcoverangles" },
    { 0x3A64, "getcovermultipretendtype" },
    { 0x3A65, "getcratetypefordroptype" },
    { 0x3A67, "getcrouchdeathanim" },
    { 0x3A6B, "getcurrentdifficultysetting" },
    { 0x3A6C, "getcurrentfraction" },
    { 0x3A6E, "getcurrentweaponxp" },
    { 0x3A73, "getdamagetype" },
    { 0x3A74, "getdeathanim" },
    { 0x3A75, "getdeathanimindex" },
    { 0x3A7A, "getdegreeselevation" },
    { 0x3A7B, "getdescription" },
    { 0x3A7C, "getdesireddrivenmovemode" },
    { 0x3A7E, "getdesiredidlepose" },
    { 0x3A80, "getdirectioncompass" },
    { 0x3A82, "getdirectionfacingclock" },
    { 0x3A87, "getdogdeathanim" },
    { 0x3A89, "getdogmoveanim" },
    { 0x3A8A, "getdognexttwitchtime" },
    { 0x3A8E, "getdogstopanimlook" },
    { 0x3A91, "getdoorside" },
    { 0x3A93, "getdvarvec" },
    { 0x3A94, "getearliestclaimplayer" },
    { 0x3A95, "getempdamageents" },
    { 0x3A96, "getenemyeyepos" },
    { 0x3A9B, "getent_or_struct" },
    { 0x3A9E, "getentarraywithflag" },
    { 0x3A9F, "getentwithflag" },
    { 0x3AA0, "geteventstate" },
    { 0x3AA1, "getexitnode" },
    { 0x3AA2, "getexitsplittime" },
    { 0x3AA3, "getexplodedistance" },
    { 0x3AAC, "getflavorburstaliases" },
    { 0x3AAD, "getflavorburstid" },
    { 0x3AB1, "getfollowmovemode" },
    { 0x3AB4, "getfraggrenadecount" },
    { 0x3AB5, "getfreeagent" },
    { 0x3AB6, "getfriendlyspawnstart" },
    { 0x3ABC, "getgenericanim" },
    { 0x3AC2, "getgunpitchtoshootentorpos" },
    { 0x3AC4, "getgvalue" },
    { 0x3AC5, "gethalftime" },
    { 0x3AC7, "gethealthcap" },
    { 0x3AC8, "gethelipilotmeshoffset" },
    { 0x3ACB, "gethighestscoringplayer" },
    { 0x3ACE, "gethostplayer" },
    { 0x3ACF, "gethvalue" },
    { 0x3AD0, "geticonshader" },
    { 0x3AD2, "getimpactpainanimindex" },
    { 0x3AD3, "getimpactpainanimstate" },
    { 0x3ADB, "getintervalsounddelaymaxdefault" },
    { 0x3ADF, "getitemweaponname" },
    { 0x3AE5, "getjumpendangles" },
    { 0x3AE6, "getjumpgravity" },
    { 0x3AE7, "getjumpinfo" },
    { 0x3AE9, "getjumpplaybackrate" },
    { 0x3AEA, "getjumpspeedmultiplier" },
    { 0x3AEC, "getjumpvelocity" },
    { 0x3AF0, "getkillstreakallteamstreak" },
    { 0x3AF1, "getkillstreakawardref" },
    { 0x3AF2, "getkillstreakcount" },
    { 0x3AF6, "getkillstreakearnedhint" },
    { 0x3AF9, "getkillstreakfromchallenge" },
    { 0x3AFA, "getkillstreakicon" },
    { 0x3AFE, "getkillstreakoverheadicon" },
    { 0x3AFF, "getkillstreakreference" },
    { 0x3B00, "getkillstreakrownum" },
    { 0x3B01, "getkillstreakscore" },
    { 0x3B04, "getkillstreakweapon" },
    { 0x3B0B, "getlaunchanimentry" },
    { 0x3B0D, "getlerptime" },
    { 0x3B0E, "getlevelcompleted" },
    { 0x3B13, "getlinkedvehiclenodes" },
    { 0x3B14, "getlinknamenodes" },
    { 0x3B18, "getlocation" },
    { 0x3B19, "getloccalloutalias" },
    { 0x3B1A, "getlookattarget" },
    { 0x3B1C, "getlosingplayers" },
    { 0x3B1D, "getlowermessage" },
    { 0x3B21, "getmapcenter" },
    { 0x3B25, "getmaxaliveenemycount" },
    { 0x3B29, "getmaxrounds" },
    { 0x3B2A, "getmaxstreakcost" },
    { 0x3B2C, "getmeleepainanimindex" },
    { 0x3B33, "getmoveanim" },
    { 0x3B34, "getmovebackentry" },
    { 0x3B35, "getmovebackstate" },
    { 0x3B3C, "getnewenemyreactionanim" },
    { 0x3B3F, "getnextflashanim" },
    { 0x3B40, "getnextlevelindex" },
    { 0x3B41, "getnextmissilespawnindex" },
    { 0x3B44, "getnextobjid" },
    { 0x3B4C, "getnodeforwardangles" },
    { 0x3B4D, "getnodeforwardyaw" },
    { 0x3B4E, "getnodefunction" },
    { 0x3B51, "getnodeorigin" },
    { 0x3B52, "getnodetype" },
    { 0x3B54, "getnodeyawtoorigin" },
    { 0x3B56, "getnormalanimtime" },
    { 0x3B57, "getnumabilitycategories" },
    { 0x3B58, "getnumactiveagents" },
    { 0x3B59, "getnumownedactiveagents" },
    { 0x3B5D, "getnumsubability" },
    { 0x3B60, "getobjectivetext" },
    { 0x3B62, "getobjpointbyname" },
    { 0x3B67, "getotherteam" },
    { 0x3B73, "getoutrig_model" },
    { 0x3B79, "getpainanim" },
    { 0x3B7D, "getparent" },
    { 0x3B7E, "getpartandstateindex" },
    { 0x3B82, "getpathnodenearplayer" },
    { 0x3B83, "getpathstart" },
    { 0x3B84, "getpercentchancetodrop" },
    { 0x3B86, "getperkforclass" },
    { 0x3B89, "getperkupgrade" },
    { 0x3B8C, "getpermutation" },
    { 0x3B8E, "getpitchtoenemy" },
    { 0x3B90, "getpitchtoshootentorpos" },
    { 0x3B94, "getplayerclaymores" },
    { 0x3B95, "getplayerfoleytype" },
    { 0x3B98, "getplayerhelispeed" },
    { 0x3B99, "getplayermodelindex" },
    { 0x3B9A, "getplayermodelname" },
    { 0x3B9B, "getplayerstat" },
    { 0x3B9D, "getplayertraceheight" },
    { 0x3B9E, "getplayerweaponhorde" },
    { 0x3BA1, "getpredictedaimyawtoshootentorpos" },
    { 0x3BA2, "getpredictedpathmidpoint" },
    { 0x3BA4, "getpreferredweapon" },
    { 0x3BA7, "getprojectiondata" },
    { 0x3BAC, "getquadrant" },
    { 0x3BAE, "getradarstrengthforcomexp" },
    { 0x3BAF, "getradarstrengthforplayer" },
    { 0x3BB1, "getradialanglefroment" },
    { 0x3BB2, "getrandomability" },
    { 0x3BB3, "getrandomanimentry" },
    { 0x3BB4, "getrandomattachments" },
    { 0x3BB7, "getrandomcratetype" },
    { 0x3BB8, "getrandomcratetypeforgamemode" },
    { 0x3BBA, "getrandomindex" },
    { 0x3BC0, "getrank" },
    { 0x3BC1, "getrankforxp" },
    { 0x3BC2, "getrankfromname" },
    { 0x3BCB, "getratiohival" },
    { 0x3BCC, "getratioloval" },
    { 0x3BCD, "getrawbaseweaponname" },
    { 0x3BCE, "getregendata" },
    { 0x3BCF, "getregenspeed" },
    { 0x3BD2, "getremainingburstdelaytime" },
    { 0x3BD7, "getreversegraphnode" },
    { 0x3BD8, "getroundintermissiontimer" },
    { 0x3BD9, "getroundswon" },
    { 0x3BE1, "getscorelimit" },
    { 0x3BE2, "getscoreperminute" },
    { 0x3BED, "getspawnpoint" },
    { 0x3BEF, "getspawnpoint_domination" },
    { 0x3BF7, "getspawnteam" },
    { 0x3BF8, "getspeakers" },
    { 0x3BFA, "getspecialroundtimer" },
    { 0x3BFB, "getsplittimes" },
    { 0x3BFE, "getsprintanim" },
    { 0x3C01, "getstance_func" },
    { 0x3C07, "getstartmoveangleindex" },
    { 0x3C09, "getstartnode" },
    { 0x3C0C, "getstat_easy" },
    { 0x3C11, "getstat_veteran" },
    { 0x3C12, "getstopanimindex" },
    { 0x3C14, "getstopdata" },
    { 0x3C23, "gettargetentpos" },
    { 0x3C25, "gettargetpredictedposition" },
    { 0x3C2A, "getteamarray" },
    { 0x3C2B, "getteamassignment" },
    { 0x3C2C, "getteambalance" },
    { 0x3C2D, "getteamcolor" },
    { 0x3C2E, "getteamcratemodel" },
    { 0x3C2F, "getteamdeploymodel" },
    { 0x3C36, "getteamflagmodel" },
    { 0x3C37, "getteamforfeitedstring" },
    { 0x3C39, "getteamhudicon" },
    { 0x3C40, "getteamspawnmusic" },
    { 0x3C42, "getteamvoiceprefix" },
    { 0x3C43, "getteamwinmusic" },
    { 0x3C47, "getthreatsovertime" },
    { 0x3C48, "gettierfromtable" },
    { 0x3C49, "gettimeinterval" },
    { 0x3C4C, "gettimepassedpercentage" },
    { 0x3C4D, "gettimeremaining" },
    { 0x3C4F, "gettimesincedompointcapture" },
    { 0x3C53, "gettranssplittime" },
    { 0x3C54, "gettruenodeangles" },
    { 0x3C55, "getturnanim" },
    { 0x3C56, "getturnanimstate" },
    { 0x3C57, "getturninplaceindex" },
    { 0x3C58, "gettweakabledvar" },
    { 0x3C59, "gettweakabledvarvalue" },
    { 0x3C5C, "getuniqueflagnameindex" },
    { 0x3C64, "getvalueinrange" },
    { 0x3C65, "getvectorarrayaverage" },
    { 0x3C6A, "getvelocity_func" },
    { 0x3C6D, "getwalkanim" },
    { 0x3C6E, "getwatcheddvar" },
    { 0x3C70, "getweaponattachmentfromchallenge" },
    { 0x3C71, "getweaponattachmentfromstats" },
    { 0x3C73, "getweaponbarsize" },
    { 0x3C76, "getweaponchoice" },
    { 0x3C77, "getweaponclass" },
    { 0x3C78, "getweaponforpos" },
    { 0x3C79, "getweaponfromchallenge" },
    { 0x3C7C, "getweaponmaxrankxp" },
    { 0x3C7E, "getweaponrankforxp" },
    { 0x3C7F, "getweaponrankinfomaxxp" },
    { 0x3C80, "getweaponrankinfominxp" },
    { 0x3C81, "getweaponrankinfoxpamt" },
    { 0x3C82, "getweaponrankxp" },
    { 0x3C84, "getweeklyref" },
    { 0x3C88, "getyaw2d" },
    { 0x3C89, "getyawangles" },
    { 0x3C8C, "getyawtoorigin" },
    { 0x3C8D, "getyawtospot" },
    { 0x3C8E, "getyawtotag" },
    { 0x3C90, "ghetto_tag_create" },
    { 0x3C92, "ghille_top" },
    { 0x3C97, "ghost_town_end" },
    { 0x3C9C, "ghost_town_sneak_end" },
    { 0x3CA1, "give_ammo_clip" },
    { 0x3CA5, "give_back_aim_over_time" },
    { 0x3CA7, "give_binoculars" },
    { 0x3CA8, "give_boost_item" },
    { 0x3CA9, "give_default" },
    { 0x3CAF, "give_hatchet" },
    { 0x3CB0, "give_health" },
    { 0x3CB3, "give_initial_perks" },
    { 0x3CB4, "give_laser" },
    { 0x3CB7, "give_loadout" },
    { 0x3CBC, "give_maxammo" },
    { 0x3CC1, "give_player_new_ref_ent_look" },
    { 0x3CC3, "give_player_points" },
    { 0x3CC5, "give_player_skill_point" },
    { 0x3CC8, "give_player_xp" },
    { 0x3CC9, "give_players_rewards" },
    { 0x3CCA, "give_point" },
    { 0x3CCB, "give_power_back" },
    { 0x3CCF, "give_soflam" },
    { 0x3CD4, "give_trophy_item" },
    { 0x3CD5, "give_underwater_weapon" },
    { 0x3CDC, "giveassistbonus" },
    { 0x3CE0, "givebonusperks" },
    { 0x3CE1, "givecarryremoteuav" },
    { 0x3CE4, "givecombathigh" },
    { 0x3CEA, "giveflagcapturexp" },
    { 0x3CEE, "givehighlight" },
    { 0x3CEF, "giveims" },
    { 0x3CF0, "givejuggernaut" },
    { 0x3CF1, "givekillreward" },
    { 0x3CF3, "givekillstreakweapon" },
    { 0x3CF5, "giveloadout" },
    { 0x3CF6, "givemarker" },
    { 0x3CF7, "givematchbonus" },
    { 0x3CF8, "giveobject" },
    { 0x3CFC, "giveperkequipment" },
    { 0x3D03, "givepointsforevent" },
    { 0x3D04, "giverandomdeployable" },
    { 0x3D05, "giverandomgun" },
    { 0x3D06, "giverankxp" },
    { 0x3D07, "giverankxp_regularmp" },
    { 0x3D09, "giverecentshieldxp" },
    { 0x3D0A, "giveresistanceperks" },
    { 0x3D0C, "giveselectedkillstreakitem" },
    { 0x3D10, "givetank" },
    { 0x3D14, "giveteamscoreforobjective" },
    { 0x3D15, "giveultimatekillstreak" },
    { 0x3D17, "giveuponsuppressiontime" },
    { 0x3D18, "giveuptime" },
    { 0x3D1E, "glass_cutter_on" },
    { 0x3D1F, "glass_cutting_fx_notetrack_handler" },
    { 0x3D28, "global_breakables_think" },
    { 0x3D29, "global_callbacks" },
    { 0x3D2A, "global_damage_func" },
    { 0x3D2E, "global_fx_array_to_string" },
    { 0x3D2F, "global_kill_func" },
    { 0x3D30, "global_spawn_functions" },
    { 0x3D31, "global_tables" },
    { 0x3D32, "globalthink" },
    { 0x3D34, "glow" },
    { 0x3D35, "glow_console" },
    { 0x3D37, "glow_model" },
    { 0x3D3E, "glowstick2_prop" },
    { 0x3D3F, "glowstick_hacking" },
    { 0x3D45, "glowstickenemyuselistener" },
    { 0x3D46, "glowstickhandledeathdamage" },
    { 0x3D48, "glowsticks" },
    { 0x3D4B, "glowstickuselistener" },
    { 0x3D4C, "glsentry_placed_listener" },
    { 0x3D4E, "gndlt" },
    { 0x3D51, "go_hit_geo" },
    { 0x3D52, "go_path_by_targetname" },
    { 0x3D54, "go_street" },
    { 0x3D57, "go_to_leaping_melee_node" },
    { 0x3D59, "go_to_node" },
    { 0x3D5C, "go_to_node_set_goal_node" },
    { 0x3D61, "go_to_struct" },
    { 0x3D62, "go_to_waiting" },
    { 0x3D63, "goal_dir" },
    { 0x3D64, "goal_fx_ent" },
    { 0x3D65, "goal_node" },
    { 0x3D6C, "goalraidus" },
    { 0x3D6F, "goback_startfunc" },
    { 0x3D71, "god_rays_from_moving_source" },
    { 0x3D72, "god_rays_from_rog" },
    { 0x3D74, "godmode" },
    { 0x3D75, "godoff" },
    { 0x3D76, "godon" },
    { 0x3D77, "godrays" },
    { 0x3D7A, "goingtoproneaim" },
    { 0x3D7D, "gold_jeep_player_door_exfil" },
    { 0x3D7E, "gold_player_door" },
    { 0x3D80, "goodaccuracy" },
    { 0x3D82, "goodfriendlydistancefromplayersquared" },
    { 0x3D83, "goodshootpos" },
    { 0x3D84, "gopath" },
    { 0x3D89, "gotocover" },
    { 0x3D8F, "grabvignetteinfo" },
    { 0x3D91, "graceperiodgrenademod" },
    { 0x3D92, "graph_position" },
    { 0x3D95, "grass_aas_approach" },
    { 0x3D96, "gravity_shift" },
    { 0x3D98, "greatestuniqueplayerkills" },
    { 0x3DA3, "grenade_cache_index" },
    { 0x3DAA, "grenade_roll_grenade" },
    { 0x3DAB, "grenade_tossed" },
    { 0x3DAD, "grenadecooldownelapsed" },
    { 0x3DB4, "grenadeorigin" },
    { 0x3DB6, "grenades" },
    { 0x3DB7, "grenades_by_difficulty" },
    { 0x3DB8, "grenades_merrick" },
    { 0x3DB9, "grenadeshielded" },
    { 0x3DBA, "grenadesplashing" },
    { 0x3DBC, "grenadethrowoffsets" },
    { 0x3DBD, "grenadethrowpose" },
    { 0x3DBE, "grenadetimers" },
    { 0x3DBF, "grenadetracking" },
    { 0x3DC6, "ground_movement" },
    { 0x3DC8, "ground_ref" },
    { 0x3DC9, "ground_ref_ent" },
    { 0x3DCB, "ground_support_locs" },
    { 0x3DD0, "group_anim" },
    { 0x3DD3, "group_flag_init" },
    { 0x3DD6, "group_get_flagname" },
    { 0x3DD7, "group_get_flagname_from_group" },
    { 0x3DDA, "group_return_groups_with_flag_set" },
    { 0x3DDB, "group_wait_group_spawned" },
    { 0x3DDC, "group_walla" },
    { 0x3DDD, "groupedanim_pos" },
    { 0x3DE0, "growl_on_path_end" },
    { 0x3DE3, "gt_get_to_cover_after_landing" },
    { 0x3DE8, "guard_b_1" },
    { 0x3DE9, "guard_b_2" },
    { 0x3DEF, "gun_down_trigger" },
    { 0x3DF7, "gun_on_ground" },
    { 0x3DF8, "gun_pickup_left" },
    { 0x3DFA, "gun_r_off" },
    { 0x3DFC, "gun_recall" },
    { 0x3DFE, "gun_removed" },
    { 0x3E00, "gun_up_for_reload" },
    { 0x3E09, "gundown_twitch" },
    { 0x3E0A, "gundown_walk" },
    { 0x3E0F, "gunfireloopfxvecthread" },
    { 0x3E12, "gunner_pain_init" },
    { 0x3E13, "gunner_pain_reset" },
    { 0x3E16, "gunnerweapon" },
    { 0x3E18, "gunship_attack_autosave" },
    { 0x3E1A, "gunship_damage" },
    { 0x3E1D, "gunship_line_attack" },
    { 0x3E1F, "gunship_line_attack_fake" },
    { 0x3E20, "gunship_sparrow_platform_loop" },
    { 0x3E21, "gunship_trans_1" },
    { 0x3E22, "gunship_trans_2" },
    { 0x3E23, "gunship_trans_3" },
    { 0x3E24, "gunship_trans_4" },
    { 0x3E28, "guy1" },
    { 0x3E2A, "guy3" },
    { 0x3E2E, "guy_cleanup_vehiclevars" },
    { 0x3E2F, "guy_death" },
    { 0x3E30, "guy_deathhandle" },
    { 0x3E31, "guy_deathimate_me" },
    { 0x3E32, "guy_do_animation" },
    { 0x3E35, "guy_duck_once" },
    { 0x3E37, "guy_duck_out" },
    { 0x3E38, "guy_enter" },
    { 0x3E3A, "guy_fridge_beers" },
    { 0x3E3C, "guy_goes_directly_to_turret" },
    { 0x3E3D, "guy_goes_to_struct_animates" },
    { 0x3E43, "guy_idle_alert_to_casual" },
    { 0x3E4E, "guy_runtovehicle_load" },
    { 0x3E4F, "guy_runtovehicle_loaded" },
    { 0x3E52, "guy_should_man_turret" },
    { 0x3E57, "guy_turn_hardleft" },
    { 0x3E59, "guy_turn_left" },
    { 0x3E5B, "guy_turn_right" },
    { 0x3E5D, "guy_turret_fire" },
    { 0x3E5E, "guy_turret_turnleft" },
    { 0x3E63, "guy_unset_allowdeath" },
    { 0x3E66, "guy_weave" },
    { 0x3E67, "guy_weave_check" },
    { 0x3E6C, "hacking_eyes_and_ears" },
    { 0x3E6D, "hacking_music" },
    { 0x3E71, "halftimeroundenddelay" },
    { 0x3E72, "halftimesubcaption" },
    { 0x3E75, "hall_clear_vo" },
    { 0x3E77, "hall_redshirt_talk" },
    { 0x3E7A, "hallway_crossing_middle" },
    { 0x3E7C, "hallway_encounter" },
    { 0x3E85, "halon_system_fog_on" },
    { 0x3E86, "halon_system_killstreak" },
    { 0x3E8A, "hand_icon" },
    { 0x3E8B, "hand_icon_count" },
    { 0x3E90, "hand_icons" },
    { 0x3E95, "handle_ally_bag_vis" },
    { 0x3E98, "handle_ambush_return_spot" },
    { 0x3E9B, "handle_approach_anims_end" },
    { 0x3E9E, "handle_autoturret_chatter" },
    { 0x3EA0, "handle_backpack_objective" },
    { 0x3EA4, "handle_baker_intro_anim" },
    { 0x3EA6, "handle_basement_guys2" },
    { 0x3EAA, "handle_binocular_zoom_magnet" },
    { 0x3EAE, "handle_breach_door" },
    { 0x3EB4, "handle_cipher_intro_anim" },
    { 0x3EB9, "handle_combat_guys2" },
    { 0x3EBD, "handle_cqb_pip_cams" },
    { 0x3EC2, "handle_death" },
    { 0x3EC6, "handle_delete" },
    { 0x3ECB, "handle_dismount" },
    { 0x3ECC, "handle_dog_ambush" },
    { 0x3ECD, "handle_dog_chaos" },
    { 0x3ECF, "handle_dog_combat_defend" },
    { 0x3ED1, "handle_dog_defend" },
    { 0x3ED3, "handle_dog_hud" },
    { 0x3ED5, "handle_dog_interior_attack" },
    { 0x3ED8, "handle_dog_modes" },
    { 0x3ED9, "handle_dog_targeting_chaos" },
    { 0x3EDB, "handle_dogbite_notetrack" },
    { 0x3EDC, "handle_drill_spot" },
    { 0x3EDD, "handle_drilling" },
    { 0x3EDE, "handle_drone" },
    { 0x3EDF, "handle_drone_nags" },
    { 0x3EE1, "handle_enemies_behind_player" },
    { 0x3EE6, "handle_exploders" },
    { 0x3EEB, "handle_finale_drones" },
    { 0x3EEE, "handle_first_doorway_allies" },
    { 0x3EEF, "handle_flare_slow" },
    { 0x3EF1, "handle_fog_changes" },
    { 0x3EF2, "handle_friendly_fail" },
    { 0x3EF3, "handle_front_elevator" },
    { 0x3EF9, "handle_ghost_chopper_removal" },
    { 0x3EFA, "handle_glowstick" },
    { 0x3EFD, "handle_grenade_thrown_failcase" },
    { 0x3EFF, "handle_gunner_pain" },
    { 0x3F00, "handle_hesh_teleport" },
    { 0x3F04, "handle_intro_fx" },
    { 0x3F05, "handle_intro_sniper_outline" },
    { 0x3F0C, "handle_jet3_fx" },
    { 0x3F0F, "handle_jet_taxi" },
    { 0x3F10, "handle_keegan_intro_anim" },
    { 0x3F12, "handle_landing" },
    { 0x3F15, "handle_lights" },
    { 0x3F18, "handle_looper" },
    { 0x3F1B, "handle_m32_launcher" },
    { 0x3F1C, "handle_mg_firing" },
    { 0x3F23, "handle_moving_platform_invalid" },
    { 0x3F2B, "handle_phys_debris" },
    { 0x3F2E, "handle_platform_blockers" },
    { 0x3F34, "handle_radiotower_guy" },
    { 0x3F38, "handle_rappel_inverted" },
    { 0x3F3E, "handle_rider_death" },
    { 0x3F45, "handle_sentry_placement_failed" },
    { 0x3F46, "handle_shadow_kill" },
    { 0x3F47, "handle_skip_rpg_sniping" },
    { 0x3F48, "handle_sliding_osprey_fx" },
    { 0x3F49, "handle_slow_movement" },
    { 0x3F4C, "handle_spotlight_enemy_b" },
    { 0x3F4D, "handle_stage_two_fx" },
    { 0x3F4E, "handle_stairs" },
    { 0x3F4F, "handle_starts" },
    { 0x3F53, "handle_strobe_on_hint" },
    { 0x3F54, "handle_tank_death" },
    { 0x3F57, "handle_teargas_grenades" },
    { 0x3F5D, "handle_truck_shooting" },
    { 0x3F60, "handle_unresolved_collision" },
    { 0x3F62, "handle_vehicles_near_iceholes" },
    { 0x3F67, "handle_weapon_melee_toggle" },
    { 0x3F6D, "handleapdamage" },
    { 0x3F71, "handlebranchnode" },
    { 0x3F73, "handlecleanupputaway" },
    { 0x3F76, "handledeathdamage" },
    { 0x3F7B, "handleemp" },
    { 0x3F7C, "handleempdamage" },
    { 0x3F7D, "handlefirstelitearrival" },
    { 0x3F7F, "handlefriendlyfiredeath" },
    { 0x3F80, "handlegrenadedamage" },
    { 0x3F81, "handlehostmigration" },
    { 0x3F84, "handleinsolid" },
    { 0x3F88, "handlemeleedamage" },
    { 0x3F8C, "handlenotetrack" },
    { 0x3F8F, "handleownerleft" },
    { 0x3F96, "handlescavengerbagpickup" },
    { 0x3F97, "handlesidetosidenotetracks" },
    { 0x3FA0, "handletraversedeathnotetrack" },
    { 0x3FA4, "handleunreslovedcollision" },
    { 0x3FA5, "handleuse" },
    { 0x3FA7, "handleworlddeath" },
    { 0x3FB0, "hangar_wall_debris" },
    { 0x3FB2, "hangar_wall_shot" },
    { 0x3FB9, "hanging_cargo_motion" },
    { 0x3FBA, "harassers_ignore_player" },
    { 0x3FBB, "hardcoremode" },
    { 0x3FBD, "hardpointtweaks" },
    { 0x3FBE, "hardpointtype" },
    { 0x3FC7, "harrierexplode" },
    { 0x3FCA, "harriers" },
    { 0x3FCF, "has_attack_abort_been_requested" },
    { 0x3FD4, "has_char_group" },
    { 0x3FD5, "has_cold_breath" },
    { 0x3FDB, "has_health_pack" },
    { 0x3FDD, "has_homing_missiles" },
    { 0x3FDE, "has_ims" },
    { 0x3FE1, "has_loadout" },
    { 0x3FE3, "has_model" },
    { 0x3FE5, "has_passive_breacher" },
    { 0x3FEB, "has_script_parameters" },
    { 0x3FF0, "has_special_weapon" },
    { 0x3FF1, "has_straight_missiles" },
    { 0x3FF2, "has_stun_ammo" },
    { 0x3FF3, "has_target_shader" },
    { 0x3FF4, "has_terrain" },
    { 0x3FF7, "has_weapon_variation" },
    { 0x3FFE, "hasattachedweapons" },
    { 0x4004, "hasdied" },
    { 0x4005, "hasdisplayvalue" },
    { 0x4008, "hasdonecombat" },
    { 0x400A, "hasenemysightpos" },
    { 0x400B, "hasexplosivefired" },
    { 0x400E, "hashelicopterdustkickup" },
    { 0x400F, "hashelicopterturret" },
    { 0x4010, "hashima" },
    { 0x4012, "hashimacustomcratefunc" },
    { 0x4019, "haslightarmor" },
    { 0x401A, "hasmissionhardenedaward" },
    { 0x401E, "hasprogressbar" },
    { 0x401F, "hasregenfaster" },
    { 0x4020, "hasriotshield" },
    { 0x4023, "hasspawned" },
    { 0x4025, "hasstarted" },
    { 0x4026, "hassuppressableenemy" },
    { 0x402F, "have_flashlight" },
    { 0x4032, "havenothingtoshoot" },
    { 0x4035, "hazmat_find_gun" },
    { 0x4039, "hazmat_if_hazmat" },
    { 0x403B, "hazmat_react" },
    { 0x403C, "hazmat_run_away" },
    { 0x4042, "hdrsuncolorintensity" },
    { 0x4043, "head_shield_activated" },
    { 0x4047, "heading" },
    { 0x4049, "headlamp_death_blink" },
    { 0x4050, "health_drain" },
    { 0x4058, "healthregen" },
    { 0x405A, "healthregendisabled" },
    { 0x405C, "healthregenerated_regularmp" },
    { 0x405F, "healthregeninit" },
    { 0x4063, "heartbeat_rumble" },
    { 0x4064, "heartbeatrate" },
    { 0x4065, "heat" },
    { 0x4066, "heat_column_fx" },
    { 0x4069, "heat_warn_toggle" },
    { 0x406A, "heatlevel" },
    { 0x406B, "heavy_quake" },
    { 0x406C, "heavyarmorhp" },
    { 0x406D, "heightforhighcallout" },
    { 0x406E, "heightrange" },
    { 0x406F, "heli" },
    { 0x4070, "heli_ai_can_hit_target" },
    { 0x4073, "heli_ai_collision_cylinder_setup" },
    { 0x4074, "heli_ai_gather_targets" },
    { 0x4078, "heli_ai_shoot_target" },
    { 0x407A, "heli_angle_offset" },
    { 0x407C, "heli_anims_length" },
    { 0x407E, "heli_array_setup" },
    { 0x407F, "heli_attack1" },
    { 0x4083, "heli_attract_range" },
    { 0x4089, "heli_beach_lander_leave" },
    { 0x408C, "heli_can_see_target" },
    { 0x408D, "heli_chopper_impact" },
    { 0x4093, "heli_combat_path" },
    { 0x4094, "heli_combat_respawn" },
    { 0x4098, "heli_crash_blade_hit_2" },
    { 0x409B, "heli_crash_box_hit_1" },
    { 0x409C, "heli_crash_box_hit_2" },
    { 0x409D, "heli_crash_engine_fail" },
    { 0x40A0, "heli_crash_kill" },
    { 0x40A6, "heli_damage_update" },
    { 0x40A9, "heli_decides_to_shoot_missile_at_ai" },
    { 0x40B1, "heli_engine_lp_02" },
    { 0x40B6, "heli_explode" },
    { 0x40BA, "heli_fire_turret" },
    { 0x40BF, "heli_firelink" },
    { 0x40C0, "heli_fireminigun_cheap" },
    { 0x40C1, "heli_firemissile_noexplode" },
    { 0x40C2, "heli_flock" },
    { 0x40C3, "heli_flock_victims" },
    { 0x40CB, "heli_flyin_mudpumps" },
    { 0x40D0, "heli_get_node_origin" },
    { 0x40D3, "heli_get_target_player_only" },
    { 0x40D5, "heli_goal_think" },
    { 0x40DE, "heli_jump_fire_fail" },
    { 0x40E4, "heli_leave_on_changeteams" },
    { 0x40E5, "heli_leave_on_disconnect" },
    { 0x40E7, "heli_leave_on_nuke" },
    { 0x40EC, "heli_loop_nodes" },
    { 0x40EF, "heli_loop_speed_control" },
    { 0x40F0, "heli_lp" },
    { 0x40F1, "heli_maxhealth" },
    { 0x40FB, "heli_pilot" },
    { 0x4100, "heli_pilot_monitor_flares" },
    { 0x4101, "heli_pilot_pick_node" },
    { 0x4102, "heli_pilot_waittill_initial_goal" },
    { 0x4108, "heli_setup" },
    { 0x4109, "heli_shells" },
    { 0x410A, "heli_shoot_think" },
    { 0x410F, "heli_sniper_waittill_initial_goal" },
    { 0x4111, "heli_spawn" },
    { 0x4119, "heli_spotlight_cleanup" },
    { 0x4121, "heli_spotlight_think" },
    { 0x4125, "heli_strafing_run" },
    { 0x4126, "heli_strafing_think" },
    { 0x4130, "heli_targeting_delay" },
    { 0x4132, "heli_think" },
    { 0x4136, "heli_turret_think" },
    { 0x4139, "heli_type" },
    { 0x413E, "heli_vs_heli_mg_range_2d_squared" },
    { 0x4140, "heli_wait_node" },
    { 0x4145, "heliconfigs" },
    { 0x4146, "helicopter_crash" },
    { 0x4148, "helicopter_crash_flavor" },
    { 0x414B, "helicopter_crash_path" },
    { 0x414E, "helicopter_fail" },
    { 0x4151, "helicopter_jump" },
    { 0x4153, "helicopter_predator_target_shader" },
    { 0x4157, "helidialog" },
    { 0x415B, "heliguardsettings" },
    { 0x415F, "helileave" },
    { 0x4161, "helipickup" },
    { 0x4166, "helipilot_leave" },
    { 0x4168, "helipilot_setairstartnodes" },
    { 0x416A, "helipilot_watchdeath" },
    { 0x416C, "helipilot_watchownerloss" },
    { 0x416E, "helipilot_watchtimeout" },
    { 0x416F, "helipilotsettings" },
    { 0x4170, "helipilottype" },
    { 0x4172, "heliridelifeid" },
    { 0x4174, "helis_can_respawn" },
    { 0x4176, "helisniper" },
    { 0x4177, "helispawning" },
    { 0x417A, "helmet_damage_test" },
    { 0x417F, "helmet_swap_keegan" },
    { 0x4181, "helmetpop" },
    { 0x4182, "helmetsidemodel" },
    { 0x418B, "help_baker_control_panel_vo" },
    { 0x418C, "help_near_comps" },
    { 0x4194, "hero_train_impact" },
    { 0x4199, "heroonly" },
    { 0x419C, "hesco_logic" },
    { 0x419F, "hesh" },
    { 0x41A5, "hesh_close_to_knockdown" },
    { 0x41A9, "hesh_dialogue_pacing" },
    { 0x41AE, "hesh_gate_logic" },
    { 0x41B0, "hesh_getin_truck" },
    { 0x41B1, "hesh_gunship_run" },
    { 0x41B3, "hesh_hide_dsm" },
    { 0x41B9, "hesh_logic" },
    { 0x41BA, "hesh_looks_at_cam" },
    { 0x41BE, "hesh_nag_til_flag" },
    { 0x41C1, "hesh_regroup_nag_vo" },
    { 0x41C2, "hesh_ride_chopper" },
    { 0x41CA, "hesh_waits_for_player" },
    { 0x41CD, "hidden_parts" },
    { 0x41D0, "hidden_terrain" },
    { 0x41D1, "hide_ai_sight_brushes" },
    { 0x41D2, "hide_and_drop_entity" },
    { 0x41D3, "hide_and_show" },
    { 0x41D5, "hide_bink_brush" },
    { 0x41D6, "hide_combat_two_intro_debris" },
    { 0x41DA, "hide_dufflebags" },
    { 0x41DB, "hide_end_bridge_geo" },
    { 0x41DD, "hide_exploder_models" },
    { 0x41E6, "hide_hint" },
    { 0x41E8, "hide_if_defined" },
    { 0x41EA, "hide_loco_exterior" },
    { 0x41EC, "hide_mask" },
    { 0x41EE, "hide_missile_launcher_collision" },
    { 0x41EF, "hide_models_by_targetname" },
    { 0x41F0, "hide_normal_hud_elements" },
    { 0x41F1, "hide_notsolid" },
    { 0x41F3, "hide_player_arms_sleeve_flaps" },
    { 0x41FA, "hide_turret" },
    { 0x41FB, "hide_turret_icon" },
    { 0x4201, "hidecarryiconongameend" },
    { 0x4207, "hidehudelementongameend" },
    { 0x420C, "hidespendhinticon" },
    { 0x420E, "hideworldiconongameend" },
    { 0x4210, "hiding" },
    { 0x4213, "high_priority_for" },
    { 0x4214, "high_priority_hint" },
    { 0x4216, "highestwins" },
    { 0x4217, "highlightairdrop" },
    { 0x4218, "highlightallenemies" },
    { 0x421E, "highy" },
    { 0x4220, "hill_chopper" },
    { 0x4221, "hill_enemy_on_spotted" },
    { 0x4224, "hill_holders" },
    { 0x4227, "hill_squad" },
    { 0x422C, "hind_projectile_damagestate" },
    { 0x4230, "hint_active" },
    { 0x4231, "hint_ascend_func" },
    { 0x4235, "hint_breakfunc" },
    { 0x4237, "hint_bridge_2" },
    { 0x423B, "hint_create" },
    { 0x423D, "hint_delete" },
    { 0x4240, "hint_fontscale" },
    { 0x4241, "hint_for_first_attack" },
    { 0x4243, "hint_guided_round_guiding_off" },
    { 0x4244, "hint_jump_off" },
    { 0x4246, "hint_machine_gun_off" },
    { 0x4247, "hint_missile_lock" },
    { 0x4248, "hint_neverbreak" },
    { 0x424C, "hint_smoke_off" },
    { 0x4252, "hint_timeout" },
    { 0x4253, "hint_toggle_thermal_off" },
    { 0x4256, "hint_zoom_off" },
    { 0x4257, "hintbackground" },
    { 0x4258, "hintbackground1" },
    { 0x4259, "hintbackground2" },
    { 0x425E, "hintexit" },
    { 0x4260, "hintmessage" },
    { 0x426A, "hit_bullet_armor" },
    { 0x426B, "hit_chain_dialogue" },
    { 0x426F, "hit_panel_rumble" },
    { 0x4270, "hit_player" },
    { 0x4272, "hit_surface" },
    { 0x4273, "hit_vo" },
    { 0x4277, "hitloc" },
    { 0x427B, "hitroundlimit" },
    { 0x4284, "hive_dependencies" },
    { 0x4285, "hive_drill_listener" },
    { 0x4289, "hive_outline_monitor" },
    { 0x428D, "hive_play_death_animations" },
    { 0x428F, "hive_play_first_pain_animations" },
    { 0x4290, "hive_play_second_pain_animations" },
    { 0x4291, "hive_score_bbprint" },
    { 0x4298, "hold_count_check" },
    { 0x4299, "hold_fire_unless_ads" },
    { 0x429D, "holdstancechange" },
    { 0x42A0, "homecoming_locations" },
    { 0x42A5, "hopper_wheel_think" },
    { 0x42AB, "hordecrateperkthink" },
    { 0x42AC, "hordecrateweaponthink" },
    { 0x42AE, "hordedroplocations" },
    { 0x42B0, "hordeendgame" },
    { 0x42B4, "hordemaydropweapon" },
    { 0x42B5, "hordesetupdogstate" },
    { 0x42B9, "horn" },
    { 0x42BB, "hose_fx" },
    { 0x42BD, "hostage_b" },
    { 0x42BF, "hostage_fire_single_right" },
    { 0x42C0, "hostage_fire_spray" },
    { 0x42C1, "hostage_health_regen" },
    { 0x42C7, "hostilesspawned" },
    { 0x42CA, "hostmigrationtimer" },
    { 0x42CC, "hostmigrationtimerthink_internal" },
    { 0x42D0, "hotel_parking_car_hide" },
    { 0x42D2, "house_attack_slowmo" },
    { 0x42D3, "house_deer" },
    { 0x42D4, "house_enter" },
    { 0x42D9, "house_secondfloor_exploder" },
    { 0x42DA, "house_secondfloor_exploder_off" },
    { 0x42DD, "hovercraft_ai_pathstarts" },
    { 0x42E0, "hovercraft_artillery_incoming_missile" },
    { 0x42E1, "hovercraft_artillery_player_weapon" },
    { 0x42E8, "hovercraft_drone_fightspots" },
    { 0x42EE, "hovercraft_missile_barrage" },
    { 0x42EF, "hovercraft_missile_barrage_player" },
    { 0x42F2, "hovercraft_tanks_setup" },
    { 0x42FA, "hovercraftdroneunloader" },
    { 0x42FB, "hovercraftlanders" },
    { 0x42FF, "hovercraftsmokeents" },
    { 0x4300, "hovercrafttanks" },
    { 0x4302, "hp" },
    { 0x4303, "hud" },
    { 0x4306, "hud2" },
    { 0x430C, "hud_color_ally" },
    { 0x430D, "hud_compass_elem" },
    { 0x430E, "hud_countfree_flares" },
    { 0x4311, "hud_debug_add" },
    { 0x4313, "hud_debug_add_frac" },
    { 0x4316, "hud_debug_add_second_string" },
    { 0x4317, "hud_debug_add_string" },
    { 0x431C, "hud_enemy_missile_lockon" },
    { 0x431D, "hud_enemy_tracker" },
    { 0x4325, "hud_highlight_homing_missile" },
    { 0x432D, "hud_makenotavailable_availablemissileicon" },
    { 0x432F, "hud_mantle" },
    { 0x4330, "hud_markavailable_firstusedmissileicon" },
    { 0x4332, "hud_marklocked_firstavailablemissileicon" },
    { 0x4333, "hud_markunlocked_firstlockedmissileicon" },
    { 0x4335, "hud_markused_freemissileicon" },
    { 0x4336, "hud_mask_model" },
    { 0x4338, "hud_mg_inactive" },
    { 0x4339, "hud_missile_active" },
    { 0x433B, "hud_missile_warning" },
    { 0x433C, "hud_monitorplayerownership" },
    { 0x433E, "hud_offset" },
    { 0x4342, "hud_outlineenable_static" },
    { 0x4344, "hud_player_target_hide_at_min" },
    { 0x4345, "hud_scubamask" },
    { 0x434D, "hud_space_helmet_rim" },
    { 0x434E, "hud_start" },
    { 0x434F, "hud_static_overlay" },
    { 0x4357, "hud_update_speed" },
    { 0x435B, "huddelete" },
    { 0x435D, "hudelem_count" },
    { 0x4360, "hudelems" },
    { 0x4362, "huditemshide" },
    { 0x4363, "huditemsshow" },
    { 0x436A, "hudoutline_wait_death" },
    { 0x436B, "huds" },
    { 0x436C, "hudsetpoint_func" },
    { 0x4373, "hummer_spawn_func" },
    { 0x4377, "humvee_turret_guy_gettin_func" },
    { 0x4379, "hunted_style_door_open" },
    { 0x437F, "hurtgen_style" },
    { 0x4380, "hvt_dog_bark" },
    { 0x4382, "hvt_guys_in_position" },
    { 0x4385, "hvt_office_chair" },
    { 0x4386, "hvt_office_doors" },
    { 0x4389, "hvt_office_hesh" },
    { 0x438B, "hvt_office_hvt_death" },
    { 0x438D, "hvt_office_light" },
    { 0x438E, "hvt_office_merrick" },
    { 0x438F, "hvt_office_player" },
    { 0x4390, "hvt_office_props" },
    { 0x4394, "hvt_reach_and_animate" },
    { 0x43A5, "icehole_achievement" },
    { 0x43A7, "icehole_shards" },
    { 0x43AA, "icon_always_show" },
    { 0x43AC, "icon_fade_in" },
    { 0x43AD, "icon_fade_out" },
    { 0x43AE, "icon_name" },
    { 0x43B0, "iconelem" },
    { 0x43B5, "iconvisall" },
    { 0x43B9, "idflags" },
    { 0x43BB, "idflags_no_knockback" },
    { 0x43BF, "idflags_penetration" },
    { 0x43C2, "idflags_shield_explosive_impact_huge" },
    { 0x43D0, "idle_break_anim_active" },
    { 0x43D3, "idle_hardleft" },
    { 0x43DA, "idle_reach_node" },
    { 0x43DB, "idle_right" },
    { 0x43DD, "idle_struct_animating_node" },
    { 0x43DF, "idleanim" },
    { 0x43E0, "idleanimtype" },
    { 0x43E2, "idlelookattargets" },
    { 0x43E3, "idleoccurrence" },
    { 0x43E6, "idlesound_waitfordoneordeath" },
    { 0x43EC, "idlewait" },
    { 0x43EE, "idlingatcover" },
    { 0x43F2, "iflashfuse" },
    { 0x43F7, "ignore_me_ignore_all" },
    { 0x43FB, "ignore_move_suppression" },
    { 0x4400, "ignore_run" },
    { 0x4408, "ignoreall_false_end_anim" },
    { 0x4409, "ignoreall_on_end" },
    { 0x440B, "ignored_by_attack_heli" },
    { 0x440D, "ignoreeachother" },
    { 0x440F, "ignoreorigin" },
    { 0x4412, "ignoresightpos" },
    { 0x4417, "impact_loc" },
    { 0x4421, "impulse_push" },
    { 0x4428, "ims_handleownerdisconnect" },
    { 0x442A, "ims_hideallparts" },
    { 0x442B, "ims_modifydamage" },
    { 0x442D, "ims_oncarrierdeath" },
    { 0x442E, "ims_oncarrierdisconnect" },
    { 0x4433, "ims_setactive" },
    { 0x4434, "ims_setcancelled" },
    { 0x4435, "ims_setcarried" },
    { 0x443B, "imscreateexplosive" },
    { 0x4441, "imsopendoor" },
    { 0x4445, "in_alien_mode" },
    { 0x4446, "in_attack" },
    { 0x4448, "in_deathsdoor" },
    { 0x4450, "in_spawnspectator" },
    { 0x4451, "in_to_jeep" },
    { 0x4454, "in_world_area" },
    { 0x4458, "inairanimentry" },
    { 0x445A, "inboundsfx" },
    { 0x445C, "inc" },
    { 0x445D, "inc4death" },
    { 0x4465, "inc_escaped_stat" },
    { 0x4466, "inc_hives_destroyed_stat" },
    { 0x4468, "inc_prestige_nerfs_stat" },
    { 0x446A, "inc_score_stat" },
    { 0x446B, "inc_session_stat" },
    { 0x446C, "inc_stat" },
    { 0x446D, "inc_xp_stat" },
    { 0x4474, "incoming_sound" },
    { 0x447C, "incranimaimweight" },
    { 0x447F, "increase_difficulty" },
    { 0x4481, "increase_threatbias" },
    { 0x4485, "incrementalivecount" },
    { 0x4489, "index" },
    { 0x448A, "index_col" },
    { 0x448B, "index_is_selected" },
    { 0x448C, "index_map" },
    { 0x448D, "indexarray" },
    { 0x448E, "indexnumber" },
    { 0x4490, "indoor_think" },
    { 0x4491, "indoorcqbtogglecheck" },
    { 0x4492, "infantry_guys" },
    { 0x4493, "infantry_teleport_start" },
    { 0x4497, "infect_chosefirstinfected" },
    { 0x44A1, "infil_allow_movement" },
    { 0x44A2, "infil_allowed_slide" },
    { 0x44A3, "infil_ally" },
    { 0x44A5, "infil_cleanup" },
    { 0x44A7, "infil_entry_lock" },
    { 0x44A8, "infil_explosions" },
    { 0x44A9, "infil_flyin_allies" },
    { 0x44AB, "infil_flyin_battle_init" },
    { 0x44AC, "infil_flyin_old" },
    { 0x44B0, "infil_grenade" },
    { 0x44B1, "infil_grenade_delete" },
    { 0x44B3, "infil_heli_anim_skip" },
    { 0x44B7, "infil_lights_and_vision" },
    { 0x44B8, "infil_movement_step" },
    { 0x44B9, "infil_object_hide" },
    { 0x44BC, "infil_path_offset" },
    { 0x44C4, "infil_rpg_guy" },
    { 0x44C5, "infil_sidestreet" },
    { 0x44C8, "infil_vignette" },
    { 0x44CD, "infinite_event_index" },
    { 0x44D3, "informattacking" },
    { 0x44D6, "informreloading" },
    { 0x44DA, "infront" },
    { 0x44E0, "init_a10" },
    { 0x44E2, "init_ai_space_animsets" },
    { 0x44E7, "init_alien_idle" },
    { 0x44E9, "init_ambient" },
    { 0x44ED, "init_animated_dufflebags_baker" },
    { 0x44EE, "init_animated_dufflebags_candk" },
    { 0x44EF, "init_animatedmodels" },
    { 0x44F1, "init_anims_human" },
    { 0x44F2, "init_anims_player" },
    { 0x44F3, "init_animset_ambush" },
    { 0x44F4, "init_animset_combat" },
    { 0x44F8, "init_animset_cover_multi" },
    { 0x44FA, "init_animset_cover_right" },
    { 0x44FB, "init_animset_cover_wall" },
    { 0x44FC, "init_animset_cqb_move" },
    { 0x44FD, "init_animset_cqb_stand" },
    { 0x44FF, "init_animset_custom_stand" },
    { 0x4500, "init_animset_death" },
    { 0x4501, "init_animset_default_crouch" },
    { 0x4504, "init_animset_flashed" },
    { 0x4507, "init_animset_heat_stand" },
    { 0x4508, "init_animset_idle" },
    { 0x450B, "init_animset_pain" },
    { 0x450D, "init_animset_reactions" },
    { 0x450E, "init_animset_rpg_crouch" },
    { 0x4513, "init_animset_shotgun_stand" },
    { 0x4516, "init_animsounds_for_animname" },
    { 0x4518, "init_audio" },
    { 0x451B, "init_biasgroups" },
    { 0x451C, "init_blocker_hive_animation_state" },
    { 0x451D, "init_bot_attachmenttable" },
    { 0x4522, "init_challenge_type" },
    { 0x4524, "init_chaos_animset" },
    { 0x4527, "init_color_grouping" },
    { 0x4528, "init_color_helper_triggers" },
    { 0x452A, "init_colors" },
    { 0x452C, "init_combat_resource_from_table" },
    { 0x452E, "init_combat_resource_overrides" },
    { 0x452F, "init_combat_resources" },
    { 0x4531, "init_creepwalk_archetype" },
    { 0x4533, "init_dam_destruction_anim" },
    { 0x4534, "init_damage_feedback" },
    { 0x453B, "init_dog_anims" },
    { 0x453E, "init_drill_drop_loc" },
    { 0x4547, "init_filter" },
    { 0x454B, "init_fog_transition" },
    { 0x454D, "init_fx" },
    { 0x454F, "init_gamescore" },
    { 0x4552, "init_helicopters" },
    { 0x4553, "init_helo" },
    { 0x4558, "init_hive_locs" },
    { 0x4559, "init_hover" },
    { 0x4563, "init_levelvariables" },
    { 0x4564, "init_light_def" },
    { 0x4565, "init_lights" },
    { 0x4567, "init_lit_model" },
    { 0x456A, "init_localoilrocks" },
    { 0x456E, "init_mgturretsettings" },
    { 0x4570, "init_mix" },
    { 0x4571, "init_move_transition_arrays" },
    { 0x4577, "init_occlusion" },
    { 0x4579, "init_patrol_paths" },
    { 0x457A, "init_perk_bullet_damage" },
    { 0x4580, "init_perks" },
    { 0x4581, "init_perks_callback" },
    { 0x4584, "init_personality_camper" },
    { 0x458A, "init_player_hud_onconnect" },
    { 0x458B, "init_player_limp" },
    { 0x458E, "init_player_rig" },
    { 0x458F, "init_player_score" },
    { 0x4590, "init_player_space" },
    { 0x4594, "init_player_unlock" },
    { 0x4595, "init_post_main" },
    { 0x4597, "init_radio_dialogue" },
    { 0x4599, "init_reverb" },
    { 0x459D, "init_screeneffect_vars" },
    { 0x45A0, "init_scripted_light" },
    { 0x45A4, "init_shooter" },
    { 0x45A5, "init_shooter_anims" },
    { 0x45AC, "init_spawn_node_info" },
    { 0x45AE, "init_squadbattlechatter" },
    { 0x45B7, "init_take_cover_warnings" },
    { 0x45BC, "init_threatbiasgroups" },
    { 0x45C1, "init_tracks" },
    { 0x45C3, "init_tunnel" },
    { 0x45C6, "init_uphill_jog_animset" },
    { 0x45C9, "init_vehicles" },
    { 0x45CC, "init_vo_system" },
    { 0x45CF, "init_whizby" },
    { 0x45D0, "init_wind_if_uninitialized" },
    { 0x45D3, "initactivisioncredits" },
    { 0x45D8, "initalienanims" },
    { 0x45D9, "initaliencannedtraverses" },
    { 0x45DA, "initaliendeath" },
    { 0x45DD, "initalienvosystem" },
    { 0x45E0, "initanimtree" },
    { 0x45E2, "initatvicredits_blade" },
    { 0x45E4, "initatvicredits_central_tech" },
    { 0x45E8, "initatvicredits_global" },
    { 0x45EA, "initatvicredits_qa1" },
    { 0x45EB, "initatvicredits_qa2" },
    { 0x45F1, "initbaseaward" },
    { 0x45F2, "initbattlechatter" },
    { 0x45F3, "initbombsquaddata" },
    { 0x45F7, "initcharacterface" },
    { 0x45FB, "initcontact" },
    { 0x4601, "initdoganimations" },
    { 0x4605, "initdogarchetype_reaction" },
    { 0x4608, "initdogvestanimations" },
    { 0x4609, "initdot" },
    { 0x4610, "initgameflags" },
    { 0x4612, "initglobals" },
    { 0x4615, "inithandsignals" },
    { 0x4616, "inithordesettings" },
    { 0x4619, "initial_dist_z_from_top" },
    { 0x461B, "initial_edge_ally_left" },
    { 0x461E, "initial_spawn" },
    { 0x4620, "initialdelay" },
    { 0x4621, "initialdmscoreupdate" },
    { 0x4628, "initializematchrules" },
    { 0x462A, "initiated" },
    { 0x462B, "initinsertionvehicles" },
    { 0x462D, "initiw6atvicredits" },
    { 0x462E, "initiw6credits" },
    { 0x4634, "initlevelflags" },
    { 0x4637, "initmovestartstoptransitions" },
    { 0x463C, "initpainfx" },
    { 0x4643, "initplayerstat" },
    { 0x4644, "initplayerstats" },
    { 0x4646, "initprisonerloadout" },
    { 0x464A, "initridekillstreak_internal" },
    { 0x464C, "initscoreboard" },
    { 0x4650, "initspawns" },
    { 0x4651, "initstat" },
    { 0x4653, "initstate" },
    { 0x4659, "initweapon" },
    { 0x465A, "initwindowtraverse" },
    { 0x465C, "inovertime" },
    { 0x465D, "inplayerportableradar" },
    { 0x465E, "inplayerscrambler" },
    { 0x4674, "insyncmeleewithtarget" },
    { 0x4677, "intel_items" },
    { 0x4678, "intel_think" },
    { 0x467E, "intelcrouchkillschallenge" },
    { 0x4684, "intelfoundshotkillschallenge" },
    { 0x4685, "intelheadshotchallenge" },
    { 0x4689, "intelknifekillchallenge" },
    { 0x468A, "intelminigun" },
    { 0x4693, "intelteammelee" },
    { 0x4694, "intelteamreward" },
    { 0x4696, "intensity" },
    { 0x4697, "intensity_monitor_update_loop" },
    { 0x4699, "intensity_spawning_paused_count" },
    { 0x469A, "intensitylevels" },
    { 0x469E, "interactive_number" },
    { 0x46A0, "interactive_tv" },
    { 0x46AD, "intermission_vo" },
    { 0x46AE, "internalmain" },
    { 0x46B5, "interrupt_level" },
    { 0x46C0, "intro_agent" },
    { 0x46C1, "intro_allies_killed_by_mig" },
    { 0x46C3, "intro_ally_idle" },
    { 0x46C4, "intro_ambush_vo" },
    { 0x46C5, "intro_and_crash_site_ally_setup" },
    { 0x46C8, "intro_anim_from_keegan" },
    { 0x46D7, "intro_binocs_not_target_vo" },
    { 0x46D8, "intro_binocs_oracle_scanning_vo" },
    { 0x46D9, "intro_binocular_monitor" },
    { 0x46DD, "intro_binoculars_use_hint" },
    { 0x46DF, "intro_birds" },
    { 0x46E4, "intro_bunker_house_runners" },
    { 0x46E5, "intro_bunker_turrets" },
    { 0x46EB, "intro_check_binocular_activate" },
    { 0x46EC, "intro_check_binocular_deactivate" },
    { 0x46ED, "intro_check_binocular_range" },
    { 0x46EF, "intro_check_binocular_zoom" },
    { 0x46F1, "intro_chopper_fx" },
    { 0x46F3, "intro_cleanup" },
    { 0x46F4, "intro_cliff_tanks" },
    { 0x46FC, "intro_dialogue_2" },
    { 0x46FF, "intro_dof" },
    { 0x4703, "intro_drive" },
    { 0x4707, "intro_enemy11" },
    { 0x4709, "intro_enemy2" },
    { 0x470B, "intro_enemy5" },
    { 0x470F, "intro_enemy_scene" },
    { 0x4711, "intro_ent_del" },
    { 0x4713, "intro_explosion2" },
    { 0x4714, "intro_fake_mortars" },
    { 0x4715, "intro_flavorburst" },
    { 0x4716, "intro_flyby_a10" },
    { 0x4721, "intro_helis_overhead" },
    { 0x4726, "intro_house_blocker_truck" },
    { 0x472B, "intro_logic" },
    { 0x472C, "intro_main" },
    { 0x472D, "intro_mask" },
    { 0x472E, "intro_masks" },
    { 0x472F, "intro_medic_osprey" },
    { 0x4733, "intro_offse" },
    { 0x4736, "intro_overlook_static" },
    { 0x4739, "intro_player_blur" },
    { 0x473A, "intro_player_goggles_watersheeting_fx" },
    { 0x473D, "intro_prep_vip_heli" },
    { 0x4748, "intro_rorke_gun" },
    { 0x474A, "intro_runners" },
    { 0x474B, "intro_save_check" },
    { 0x4750, "intro_screen_custom_timing" },
    { 0x4752, "intro_scripted_audio_fade_in_mixing" },
    { 0x4753, "intro_scripted_sequence" },
    { 0x4757, "intro_sequence_precache" },
    { 0x4758, "intro_sequence_street" },
    { 0x4759, "intro_setup" },
    { 0x4762, "intro_street_abrams" },
    { 0x4764, "intro_street_drones" },
    { 0x4767, "intro_tank_foley" },
    { 0x4768, "intro_target_monitor" },
    { 0x476A, "intro_transport_mover" },
    { 0x476B, "intro_turn_off_prop_bm21_1_lights_fx" },
    { 0x4779, "intro_watch" },
    { 0x477A, "introcp_guy_radio" },
    { 0x477B, "introcp_guys_remaining" },
    { 0x477C, "introcp_guys_tower" },
    { 0x477D, "introscreen" },
    { 0x4786, "introscreen_generic_fade_in_on_flag" },
    { 0x4788, "introscreen_line_1" },
    { 0x478B, "introscreen_line_4" },
    { 0x4792, "invasion_door_anim" },
    { 0x4794, "inventory" },
    { 0x47A0, "inverted_rappel" },
    { 0x47A1, "inverted_rappel_ally_idles" },
    { 0x47A2, "inverted_rappel_ally_movement" },
    { 0x47A5, "inverted_rappel_movement_rorke" },
    { 0x47A8, "invincible" },
    { 0x47B2, "iplane_start" },
    { 0x47B3, "iplane_start_dialogue" },
    { 0x47B5, "is3d" },
    { 0x47BA, "is_alien_agent" },
    { 0x47BC, "is_ambience_blend_valid" },
    { 0x47BF, "is_apache_player" },
    { 0x47C0, "is_array_close" },
    { 0x47C4, "is_blocker_hive" },
    { 0x47C7, "is_burning" },
    { 0x47D4, "is_currently_mg_target" },
    { 0x47D9, "is_dead_sentient" },
    { 0x47E0, "is_dog_really_attacking" },
    { 0x47E2, "is_drop_button_pressed" },
    { 0x47E6, "is_dynamic_path" },
    { 0x47E9, "is_empty_string" },
    { 0x47EB, "is_escape_sequence_active" },
    { 0x47EC, "is_exploding" },
    { 0x47EF, "is_fake_zodiac" },
    { 0x47F0, "is_firing" },
    { 0x47F1, "is_first" },
    { 0x47F3, "is_frag_grenade" },
    { 0x47F9, "is_godmode" },
    { 0x47FC, "is_hero_ai" },
    { 0x4807, "is_in_upload" },
    { 0x4808, "is_indoor_map" },
    { 0x480B, "is_lastblast" },
    { 0x480C, "is_later_in_alphabet" },
    { 0x480F, "is_locked" },
    { 0x4811, "is_moving" },
    { 0x4814, "is_no_nerf" },
    { 0x4817, "is_normal_upright" },
    { 0x481A, "is_on_heli" },
    { 0x481D, "is_overrode" },
    { 0x4822, "is_path_start_node" },
    { 0x4823, "is_perk_set" },
    { 0x482B, "is_player_right" },
    { 0x482D, "is_playing_initial_blocker_anim" },
    { 0x4830, "is_precalculated_entrance" },
    { 0x4833, "is_resource_set" },
    { 0x4834, "is_rider" },
    { 0x483E, "is_spawn_debug_info_requested" },
    { 0x4842, "is_spitter_spit" },
    { 0x4844, "is_survival" },
    { 0x4846, "is_targeting" },
    { 0x484B, "is_turret_enabled" },
    { 0x4858, "is_valid_spit_target" },
    { 0x485D, "is_visable" },
    { 0x4860, "isactive" },
    { 0x4865, "isagent" },
    { 0x4866, "isaifunc" },
    { 0x4867, "isaigameparticipant" },
    { 0x4868, "isairdenied" },
    { 0x486A, "isairdropmarker" },
    { 0x486B, "isairplane" },
    { 0x4876, "isanymissilefiredonme" },
    { 0x4878, "isassaultkillstreak" },
    { 0x487B, "isatmost" },
    { 0x487D, "isattachment" },
    { 0x487F, "isattachmentsniperscopedefaulttokenized" },
    { 0x4880, "isattachmentunlocked" },
    { 0x4885, "isawardexclusive" },
    { 0x4886, "isawardflag" },
    { 0x4887, "isbombcarrier" },
    { 0x488A, "isbreaching" },
    { 0x488E, "isbulletweapon" },
    { 0x488F, "iscacprimaryweapon" },
    { 0x4893, "iscallouttypereport" },
    { 0x4895, "iscapturingcrate" },
    { 0x4898, "ischallengeunlocked" },
    { 0x489A, "ischeap" },
    { 0x48A1, "iscqbwalkingorfacingenemy" },
    { 0x48A4, "iscurrentlyholdingkillstreakweapon" },
    { 0x48A8, "isdefusing" },
    { 0x48AD, "isdestructibleweapon" },
    { 0x48B0, "isdrone" },
    { 0x48BB, "isexposed" },
    { 0x48BC, "isfacing" },
    { 0x48BD, "isferal" },
    { 0x48C4, "isflashed" },
    { 0x48C5, "isflashing" },
    { 0x48C8, "isfmjdamage" },
    { 0x48CA, "isfriendlyteam" },
    { 0x48CC, "isfriendlytoims" },
    { 0x48D1, "ishackweapon" },
    { 0x48D5, "ishelicopter" },
    { 0x48DF, "isincontact" },
    { 0x48E5, "isinset" },
    { 0x48E8, "isinzone" },
    { 0x48E9, "isitem" },
    { 0x48F4, "iskillstreakaffectedbyjammer" },
    { 0x48F5, "iskillstreakchallenge" },
    { 0x48F8, "isknifeonly" },
    { 0x48FE, "islaserguidedmissile" },
    { 0x4901, "islastround" },
    { 0x4906, "islockedontarget" },
    { 0x4907, "islongrangeai" },
    { 0x490A, "isloot" },
    { 0x490D, "ismembersaying" },
    { 0x490E, "ismidstarted" },
    { 0x490F, "isminitarget" },
    { 0x4913, "ismoving" },
    { 0x491E, "isoffhandweapon" },
    { 0x491F, "isoffhandweaponenabled" },
    { 0x4921, "isonemanarmymenu" },
    { 0x4922, "isonhumanteam" },
    { 0x4925, "ispainted" },
    { 0x4926, "ispartiallysuppressedwrapper" },
    { 0x4927, "ispathclear" },
    { 0x492A, "ispeekoutposclear" },
    { 0x492C, "isperk" },
    { 0x492D, "isperkstreakon" },
    { 0x4930, "isplanting" },
    { 0x4936, "isplayeroutsideofanybombsite" },
    { 0x4937, "isplayertimer" },
    { 0x493E, "isprimaryweapon" },
    { 0x4941, "israppelshooting" },
    { 0x4942, "isreactivated" },
    { 0x4943, "isreactivating" },
    { 0x4947, "isrelativeteam" },
    { 0x4948, "isreloading" },
    { 0x4949, "isrepairing" },
    { 0x4954, "isscoring" },
    { 0x4955, "isscriptmodel" },
    { 0x4957, "issentient" },
    { 0x4959, "issentrygun" },
    { 0x495A, "issetup" },
    { 0x495B, "isshooting" },
    { 0x495C, "isshotgun" },
    { 0x495D, "isshotgunai" },
    { 0x495F, "issidearm" },
    { 0x4960, "issliding" },
    { 0x4962, "issniperrifle" },
    { 0x4964, "issp_towerdefense" },
    { 0x4967, "isspeakerinrange" },
    { 0x4969, "isspeakingfailsafe" },
    { 0x496A, "isspecialist" },
    { 0x4975, "issue_color_orders" },
    { 0x4976, "issue_leave_node_order_to_ai_and_get_ai" },
    { 0x4977, "issuffix" },
    { 0x4978, "issupportkillstreak" },
    { 0x4981, "isteam" },
    { 0x4982, "isteaminlaststand" },
    { 0x4986, "isteamspeaking" },
    { 0x4989, "isthresholdaward" },
    { 0x498A, "isthrowableitem" },
    { 0x498B, "istopbarrage" },
    { 0x498D, "istouchingrangetrigger" },
    { 0x498E, "istouchingtrigger" },
    { 0x4999, "isvalidattachment" },
    { 0x499F, "isvalidevent" },
    { 0x49A0, "isvalidffatarget" },
    { 0x49A3, "isvalidperk1" },
    { 0x49A4, "isvalidperk2" },
    { 0x49A5, "isvalidperk3" },
    { 0x49A8, "isvalidsecondary" },
    { 0x49A9, "isvalidsoundcause" },
    { 0x49AF, "isvehicle" },
    { 0x49B2, "isweapon" },
    { 0x49B4, "isweaponbuffunlocked" },
    { 0x49B5, "isweaponchallenge" },
    { 0x49B9, "isweaponswitchenabled" },
    { 0x49BA, "item_bob" },
    { 0x49BB, "item_ent" },
    { 0x49BD, "item_exist" },
    { 0x49BF, "item_min_distance_from_players" },
    { 0x49C0, "item_outline" },
    { 0x49C4, "item_pickup_listener" },
    { 0x49C5, "item_ref" },
    { 0x49C7, "itemexplodethisframe" },
    { 0x49CC, "jackpot_tag" },
    { 0x49CD, "jackpot_targetfx" },
    { 0x49CE, "jackpot_zone" },
    { 0x49CF, "jackpotpileicon" },
    { 0x49D1, "javelin" },
    { 0x49D2, "javelin_check_decent" },
    { 0x49D3, "javelin_dof" },
    { 0x49D6, "javelin_target_set" },
    { 0x49DB, "javelinlostsightlinetime" },
    { 0x49DE, "javelinstage" },
    { 0x49E2, "javelintargets" },
    { 0x49E4, "javelinuseentered" },
    { 0x49E5, "javtargets" },
    { 0x49E8, "jeep2" },
    { 0x49F1, "jeep_player_arms_sub" },
    { 0x49F4, "jeep_start_music" },
    { 0x49F5, "jeeps_by" },
    { 0x49FD, "jet_blast_shields" },
    { 0x4A03, "jet_flight_time" },
    { 0x4A0B, "jet_phalanx_tracking" },
    { 0x4A0C, "jet_planesound" },
    { 0x4A0D, "jet_reset" },
    { 0x4A0E, "jet_takeoff1" },
    { 0x4A0F, "jet_takeoff2" },
    { 0x4A13, "jets_check_deleted" },
    { 0x4A21, "joint1" },
    { 0x4A2B, "juggmovespeedscaler" },
    { 0x4A2E, "juggsettings" },
    { 0x4A2F, "juggtype" },
    { 0x4A32, "juicetime" },
    { 0x4A35, "jump_distance_allowed" },
    { 0x4A3A, "jumparms" },
    { 0x4A47, "jumporient" },
    { 0x4A50, "junction_baker_open_elevator_control_room" },
    { 0x4A51, "junction_banners" },
    { 0x4A53, "junction_elevator_control_doors_open" },
    { 0x4A56, "junction_enemies" },
    { 0x4A57, "junction_enemy_setup" },
    { 0x4A58, "junction_fireworks" },
    { 0x4A5E, "junction_pip_waver_drone" },
    { 0x4A5F, "junction_rorke_window" },
    { 0x4A6A, "jungle_start" },
    { 0x4A73, "keegan_additional_drivein_anims" },
    { 0x4A76, "keegan_breach_anim" },
    { 0x4A78, "keegan_breach_guys" },
    { 0x4A79, "keegan_drag_body" },
    { 0x4A7A, "keegan_drag_body2" },
    { 0x4A7B, "keegan_enter" },
    { 0x4A81, "keegan_gun_l" },
    { 0x4A83, "keegan_idle_with_bishop" },
    { 0x4A84, "keegan_kill_dialog" },
    { 0x4A8B, "keegan_snipes" },
    { 0x4A8C, "keegan_stands_behind_desk" },
    { 0x4A8D, "keegan_start_at_button" },
    { 0x4A8E, "keegan_swap_head_to_mask" },
    { 0x4A93, "keegan_turn_right_anims" },
    { 0x4A94, "keegan_turn_right_anims_rush" },
    { 0x4A99, "keep_player_below_surface" },
    { 0x4A9C, "keep_up_with_player_reset" },
    { 0x4AA0, "keepnukeemptimeremaining" },
    { 0x4AA3, "keeptryingtomelee" },
    { 0x4AA4, "keepweapons" },
    { 0x4AA9, "keyhint" },
    { 0x4AAD, "kick_bully_logic" },
    { 0x4AB9, "kill_allies" },
    { 0x4ABA, "kill_allies_on_next_shot" },
    { 0x4ABB, "kill_ally1_submerge_bubbles" },
    { 0x4ABC, "kill_ally_in_volume" },
    { 0x4AC0, "kill_barriers_when_close" },
    { 0x4AC6, "kill_death_anim_thread" },
    { 0x4AC8, "kill_deathflag" },
    { 0x4ACC, "kill_drone_respawner" },
    { 0x4ACD, "kill_dude" },
    { 0x4ACE, "kill_during_vignette" },
    { 0x4ACF, "kill_ending_heli_fx" },
    { 0x4AD1, "kill_enemies_touching_trigger" },
    { 0x4AD2, "kill_enemy_at_arm_node" },
    { 0x4AD3, "kill_estimates_vo" },
    { 0x4AD4, "kill_fire" },
    { 0x4AD6, "kill_fx" },
    { 0x4ADB, "kill_guys_right" },
    { 0x4ADE, "kill_heli_logic" },
    { 0x4AE1, "kill_hive_burning_on_death" },
    { 0x4AE2, "kill_infil_enemies" },
    { 0x4AE3, "kill_intro_chopper" },
    { 0x4AED, "kill_on_river_crossing_back_track" },
    { 0x4AEF, "kill_player" },
    { 0x4AF0, "kill_player_if_go_back_or_not_moving" },
    { 0x4AF3, "kill_players_touching_ent" },
    { 0x4AF5, "kill_remaining_gunboats" },
    { 0x4AFD, "kill_spawnernum" },
    { 0x4AFE, "kill_spawners_per_checkpoint" },
    { 0x4AFF, "kill_streak" },
    { 0x4B01, "kill_timer" },
    { 0x4B03, "kill_trigger" },
    { 0x4B04, "kill_trigger_linked" },
    { 0x4B0D, "killcamcleanup" },
    { 0x4B12, "killchains" },
    { 0x4B16, "killedaxis" },
    { 0x4B17, "killedbestenemyplayer" },
    { 0x4B19, "killedinuse" },
    { 0x4B1B, "killedplayernotifysys" },
    { 0x4B1D, "killedplayerscurrent" },
    { 0x4B22, "killing_will_down" },
    { 0x4B23, "killplayer" },
    { 0x4B24, "killplayerfromcrate_dodamage" },
    { 0x4B25, "killplayerfromcrate_fastvelocitypush" },
    { 0x4B28, "killsassurvivor" },
    { 0x4B31, "killstreak_botparm" },
    { 0x4B33, "killstreak_info" },
    { 0x4B35, "killstreakcratethink" },
    { 0x4B40, "killstreaks" },
    { 0x4B41, "killstreaks_array" },
    { 0x4B45, "killstreakspawnshield" },
    { 0x4B47, "killstreakthink" },
    { 0x4B49, "killstreakusepressed" },
    { 0x4B4B, "killstreakweapons" },
    { 0x4B4D, "killteaminlaststand" },
    { 0x4B51, "killwrapper" },
    { 0x4B57, "knife" },
    { 0x4B58, "knife_guy_cleanup" },
    { 0x4B5A, "knife_out_rorke_anims" },
    { 0x4B5B, "knife_reticle" },
    { 0x4B62, "knockback_player" },
    { 0x4B68, "ks_manualflares_watchuse" },
    { 0x4B69, "ks_setup_manual_flares" },
    { 0x4B6D, "la_river_defend_weapons_spawn" },
    { 0x4B6F, "ladder" },
    { 0x4B72, "ladder_spot_glow" },
    { 0x4B73, "ladder_vo" },
    { 0x4B78, "land_shoot_vert_missile" },
    { 0x4B79, "land_shoot_vert_missile_loop" },
    { 0x4B7B, "land_vert_missile_wait_max" },
    { 0x4B81, "landed" },
    { 0x4B83, "landing_gear_up" },
    { 0x4B87, "lanterns" },
    { 0x4B89, "laptop_close" },
    { 0x4B8C, "laptop_hdr" },
    { 0x4B8D, "laptop_off" },
    { 0x4B92, "largescale" },
    { 0x4B96, "lariver_balcony_friendly_logic" },
    { 0x4B97, "lariver_bridge_drones" },
    { 0x4B99, "lariver_bridge_rappel_enemies" },
    { 0x4B9B, "lariver_chopper_passanger_logic" },
    { 0x4BA0, "lariver_defend_enemy_population" },
    { 0x4BA2, "lariver_defend_guided_missile_setup" },
    { 0x4BA6, "lariver_defend_populate_close_area" },
    { 0x4BA7, "lariver_defend_slide_down_river_wall" },
    { 0x4BA9, "lariver_defend_start" },
    { 0x4BAA, "lariver_doors_siren" },
    { 0x4BAF, "lariver_enemy_chopper_logic" },
    { 0x4BB0, "lariver_exit" },
    { 0x4BB6, "lariver_matv_load_and_go" },
    { 0x4BBE, "lariver_spawn_wall_battle_guys_early" },
    { 0x4BC1, "lariver_team2_logic" },
    { 0x4BC7, "lasedstrikecratethink" },
    { 0x4BC8, "lasedstrikedrone" },
    { 0x4BCD, "laser_artillery" },
    { 0x4BD1, "laser_designator_disable_list" },
    { 0x4BD2, "laser_ent" },
    { 0x4BD6, "laser_targets" },
    { 0x4BDA, "laserforceon" },
    { 0x4BDC, "laserguidedmissileents_inuse" },
    { 0x4BE0, "laseron_func" },
    { 0x4BE6, "last_audio_bink_beep_array_num" },
    { 0x4BE7, "last_audio_bink_percentage" },
    { 0x4BE8, "last_balcony_death" },
    { 0x4BEA, "last_balcony_death_time" },
    { 0x4BEB, "last_bob_anim" },
    { 0x4BEE, "last_charge_time" },
    { 0x4BEF, "last_command" },
    { 0x4BF0, "last_commanded_bot" },
    { 0x4BF2, "last_death_pos" },
    { 0x4BF8, "last_dodge_time" },
    { 0x4BF9, "last_dog_attack" },
    { 0x4BFB, "last_drill_pickup_angles" },
    { 0x4BFD, "last_enemy_sight_time" },
    { 0x4C01, "last_goal_struct" },
    { 0x4C02, "last_heli_decides_to_shoot_missile_at_ai_time" },
    { 0x4C05, "last_hive_xp" },
    { 0x4C07, "last_investigation_time" },
    { 0x4C08, "last_killfirm_time" },
    { 0x4C09, "last_killfirm_timeout" },
    { 0x4C0B, "last_large_rod_time" },
    { 0x4C0C, "last_legs_offset" },
    { 0x4C13, "last_motion_time" },
    { 0x4C14, "last_move_state" },
    { 0x4C19, "last_player_damage" },
    { 0x4C1E, "last_runanim" },
    { 0x4C21, "last_set_goalent" },
    { 0x4C23, "last_set_goalpos" },
    { 0x4C25, "last_spawned_time" },
    { 0x4C27, "last_spawned_vehicle" },
    { 0x4C2D, "last_strips" },
    { 0x4C2E, "last_teleport_time" },
    { 0x4C30, "last_time" },
    { 0x4C36, "lastadvancetoenemyattacker" },
    { 0x4C37, "lastadvancetoenemydest" },
    { 0x4C38, "lastadvancetoenemysrc" },
    { 0x4C3A, "lastaieventtrigger" },
    { 0x4C3D, "lastattackedshieldplayer" },
    { 0x4C3F, "lastautosavetime" },
    { 0x4C46, "lastcampkilltime" },
    { 0x4C47, "lastcaralarmtime" },
    { 0x4C4A, "lastcarexplosionrange" },
    { 0x4C4E, "lastclass" },
    { 0x4C54, "lastdeathicon" },
    { 0x4C5C, "lastenemypos" },
    { 0x4C5D, "lastenemysightposold" },
    { 0x4C5E, "lastenemysightposselforigin" },
    { 0x4C61, "lastexplodingbarrel" },
    { 0x4C62, "lastflashedtime" },
    { 0x4C63, "lastfraggrenadetoplayerstart" },
    { 0x4C66, "lastgrenadelandednearplayertime" },
    { 0x4C6F, "lastkilldogtime" },
    { 0x4C70, "lastkilledby" },
    { 0x4C73, "lastkilllocation" },
    { 0x4C7B, "lastmansd" },
    { 0x4C7C, "lastmeleefailedmypos" },
    { 0x4C7D, "lastmeleefailedpos" },
    { 0x4C7F, "lastmortar" },
    { 0x4C85, "lastpaintime" },
    { 0x4C88, "lastplayercallouttime" },
    { 0x4C8B, "lastprimaryweaponswaptime" },
    { 0x4C97, "lastshoottime" },
    { 0x4C9A, "lastslowprocessframe" },
    { 0x4C9E, "lastspawnteam" },
    { 0x4C9F, "lastspawntime" },
    { 0x4CA4, "laststand_enabled" },
    { 0x4CAD, "laststandkeepoverlay" },
    { 0x4CB0, "laststandrespawnplayer" },
    { 0x4CB6, "laststandupdatereviveiconcoloralien" },
    { 0x4CB8, "laststandusetime" },
    { 0x4CB9, "laststandwaittilldeath" },
    { 0x4CBA, "laststandwaittilldeathhorde" },
    { 0x4CC0, "laststatustime" },
    { 0x4CC1, "laststoppedtime" },
    { 0x4CC4, "lastsurfacetype" },
    { 0x4CC5, "lasttarget" },
    { 0x4CC9, "lastteamthreatcallouttime" },
    { 0x4CCC, "lastturretindex" },
    { 0x4CCD, "lastupdatetime" },
    { 0x4CD3, "lastvectotarget" },
    { 0x4CD4, "lastvisionsetthermal" },
    { 0x4CD9, "lateral_change_this_update" },
    { 0x4CDE, "launch_button" },
    { 0x4CE1, "launch_ends" },
    { 0x4CE4, "launch_prep_static" },
    { 0x4CE6, "launch_ragdolls_zodiac" },
    { 0x4CE9, "launch_rope_ally" },
    { 0x4CEB, "launch_smoke" },
    { 0x4CEC, "launch_smoke_from_tag" },
    { 0x4CED, "launch_speed" },
    { 0x4CEE, "launchanimentry" },
    { 0x4CEF, "launchanimstate" },
    { 0x4CF0, "launcher" },
    { 0x4CF7, "launcher_destroy_slomo_sfx" },
    { 0x4CF8, "launcher_destroy_stop_slomo_sfx" },
    { 0x4CF9, "launcher_lynx" },
    { 0x4CFA, "launcher_lynx_spawn_func" },
    { 0x4CFD, "launchers_attached" },
    { 0x4D01, "launchtime" },
    { 0x4D02, "launchuav" },
    { 0x4D03, "launchvelocity" },
    { 0x4D04, "launchvelocity2d" },
    { 0x4D06, "layer_completed" },
    { 0x4D09, "lbexplode" },
    { 0x4D0E, "lbsupport_attacktargets" },
    { 0x4D10, "lbsupport_burstfirestop" },
    { 0x4D14, "lbsupport_getclosestnode" },
    { 0x4D1E, "lbsupport_watchownerdamage" },
    { 0x4D29, "lcs_lights_back" },
    { 0x4D2A, "lcs_lights_front" },
    { 0x4D2B, "lcs_setup" },
    { 0x4D2E, "lead_jeep" },
    { 0x4D32, "leaderdialoggroup" },
    { 0x4D33, "leaderdialoggroups" },
    { 0x4D39, "leadersound" },
    { 0x4D44, "leave_node_on_attacked" },
    { 0x4D45, "leave_node_on_distance_breach" },
    { 0x4D46, "leave_path_for_spline_path" },
    { 0x4D47, "leavebattlebuddysystem" },
    { 0x4D48, "leavebattlebuddysystemdisconnect" },
    { 0x4D49, "leavecoverandshoot" },
    { 0x4D4B, "leaveonownerdisconnect" },
    { 0x4D4E, "ledge_fx" },
    { 0x4D4F, "ledge_old_player_pos" },
    { 0x4D51, "ledge_sway" },
    { 0x4D52, "ledge_track_player_pos" },
    { 0x4D54, "left_arc" },
    { 0x4D56, "left_defend" },
    { 0x4D57, "left_ent" },
    { 0x4D58, "left_flank_spawn_proc" },
    { 0x4D5D, "left_post" },
    { 0x4D5F, "left_stick" },
    { 0x4D60, "left_velocity" },
    { 0x4D63, "leg_anim_blend_time" },
    { 0x4D64, "leg_anim_blend_time_fast" },
    { 0x4D65, "leg_clear_anim_blend_time" },
    { 0x4D6B, "legs_idle_anim" },
    { 0x4D6C, "legs_move_parent_node" },
    { 0x4D6D, "length" },
    { 0x4D6E, "leper" },
    { 0x4D72, "leper_combat" },
    { 0x4D73, "leper_despawn" },
    { 0x4D74, "leper_init" },
    { 0x4D75, "leper_retreat" },
    { 0x4D78, "leperdespawntime" },
    { 0x4D79, "lerp" },
    { 0x4D7E, "lerp_fovscale_overtime" },
    { 0x4D80, "lerp_intensity" },
    { 0x4D81, "lerp_maxalpha_overtime" },
    { 0x4D83, "lerp_out_drop_pitch" },
    { 0x4D88, "lerp_player_view_to_tag" },
    { 0x4D8C, "lerp_player_view_to_tag_oldstyle_internal" },
    { 0x4D8D, "lerp_rate" },
    { 0x4D90, "lerp_saveddvar_cg_ng" },
    { 0x4D98, "lerpfov_saved_thread" },
    { 0x4D99, "lerpfraction" },
    { 0x4D9B, "let_player_through" },
    { 0x4D9E, "level_2_weight" },
    { 0x4DA5, "level_fadein" },
    { 0x4DAD, "level_specific_dof" },
    { 0x4DAE, "level_specific_score_parameter" },
    { 0x4DB3, "levelflag" },
    { 0x4DB7, "levelflagset" },
    { 0x4DB9, "levelflagwaitopen" },
    { 0x4DBB, "levelhasvehicles" },
    { 0x4DBE, "lg_missileslocked" },
    { 0x4DC3, "lgm_firing_monitormissilefire" },
    { 0x4DC7, "lgm_locked_missileondeath" },
    { 0x4DC8, "lgm_locked_spawnmissiles" },
    { 0x4DCA, "lgm_locking_loopsound" },
    { 0x4DCC, "lgm_locking_think" },
    { 0x4DCD, "lgm_missilesnotifyandrelease" },
    { 0x4DD0, "lgm_onmissilenotifies" },
    { 0x4DD5, "lgm_update_launcherusage" },
    { 0x4DD6, "lgt_ally_sequence" },
    { 0x4DE0, "life_span" },
    { 0x4DE3, "liferaft_splash_on_hit_water" },
    { 0x4DE8, "light_brighten" },
    { 0x4DED, "light_fade" },
    { 0x4DF0, "light_flicker" },
    { 0x4DF2, "light_hall_light" },
    { 0x4DF3, "light_model" },
    { 0x4DF4, "light_pulse" },
    { 0x4DF5, "light_rog_threads" },
    { 0x4DF6, "light_tag" },
    { 0x4DFB, "lightgridsupersamplecount" },
    { 0x4DFC, "lighthouse" },
    { 0x4DFE, "lightmapcount_cg" },
    { 0x4E00, "lightning_flash" },
    { 0x4E02, "lightningexploder" },
    { 0x4E03, "lightningexploderindex" },
    { 0x4E05, "lightningthink" },
    { 0x4E07, "lights" },
    { 0x4E09, "lights_off" },
    { 0x4E16, "limping_guys_spawnfunc" },
    { 0x4E19, "line_debug" },
    { 0x4E1D, "linelist" },
    { 0x4E1E, "lines" },
    { 0x4E23, "link1" },
    { 0x4E25, "link_bag_to_jeep_after_anim" },
    { 0x4E26, "link_bags_to_spin_crates" },
    { 0x4E27, "link_dlight_to_dummy" },
    { 0x4E28, "link_dog_to_jeep" },
    { 0x4E2A, "link_final_exposed_nodes" },
    { 0x4E2C, "link_player_for_catch" },
    { 0x4E2F, "link_to_moving_target" },
    { 0x4E30, "link_to_sittag" },
    { 0x4E32, "linked" },
    { 0x4E35, "linked_ents" },
    { 0x4E39, "linked_prefab_ents" },
    { 0x4E3B, "linked_to_ent" },
    { 0x4E3D, "linked_world_space_forward" },
    { 0x4E3E, "linkedplayers" },
    { 0x4E42, "linkparent" },
    { 0x4E43, "linkpet" },
    { 0x4E44, "links" },
    { 0x4E45, "linkstartnodes" },
    { 0x4E46, "linktoblend" },
    { 0x4E48, "linktotrain" },
    { 0x4E4B, "listen_dog_attack" },
    { 0x4E4E, "listen_for_dog_commands" },
    { 0x4E50, "listen_for_landing_gear_messages" },
    { 0x4E51, "listen_for_mine_layed" },
    { 0x4E52, "listen_for_mine_trigger" },
    { 0x4E53, "listen_for_use_multi_turret" },
    { 0x4E62, "listen_player_leftground" },
    { 0x4E67, "lit_models" },
    { 0x4E6E, "littlebird_spotlight_death" },
    { 0x4E72, "littlebird_turrets_think" },
    { 0x4E75, "littlebirdmadeselectionvo" },
    { 0x4E79, "living_ai" },
    { 0x4E7A, "living_ai_prethink" },
    { 0x4E7E, "lnchr_slomo_sfx" },
    { 0x4E7F, "load" },
    { 0x4E8A, "load_player_anims" },
    { 0x4E90, "load_spawn_events_from_table" },
    { 0x4E92, "load_transient" },
    { 0x4E95, "load_variables_from_table" },
    { 0x4E9C, "loadingdocks_no_jump" },
    { 0x4EA3, "loadoutperkoffhand" },
    { 0x4EA7, "loadoutprimarycamo" },
    { 0x4EA8, "loadoutprimaryreticle" },
    { 0x4EAD, "loadoutsecondarycamo" },
    { 0x4EAF, "loadoutvalues" },
    { 0x4EB2, "lobby_ruckus" },
    { 0x4EB4, "loc1" },
    { 0x4EBA, "location" },
    { 0x4EBB, "location_add_last_callout_time" },
    { 0x4ECA, "lock_dummy_add" },
    { 0x4ECB, "lock_dummy_add_ondeath" },
    { 0x4ECC, "lock_dummy_remove" },
    { 0x4ED2, "lock_player_controls" },
    { 0x4ED3, "lock_spawner_for_awhile" },
    { 0x4EDA, "lockedtarget" },
    { 0x4EDE, "lockmissespassedthreshold" },
    { 0x4EE5, "lockontargets_stop" },
    { 0x4EEE, "loco_breach_opfor_start" },
    { 0x4EF7, "loco_falling_glass" },
    { 0x4EF8, "loco_gunhold_end_slowmo" },
    { 0x4F00, "loco_slide_player_view_lerp_up" },
    { 0x4F06, "log_anim_lengths" },
    { 0x4F0D, "log_pile_support_damage_watch" },
    { 0x4F12, "log_visual_link_joints" },
    { 0x4F14, "logaward" },
    { 0x4F17, "loggameevent" },
    { 0x4F1B, "logkillsconfirmed" },
    { 0x4F25, "logplayerdata" },
    { 0x4F2A, "logweaponstat" },
    { 0x4F2B, "logxpgains" },
    { 0x4F2C, "loki_breach_lighting" },
    { 0x4F30, "loki_drop_weapon" },
    { 0x4F3A, "loki_space_particulates_while_moving" },
    { 0x4F42, "look_at_roof_nag" },
    { 0x4F47, "look_at_tower_vo_timeout" },
    { 0x4F4D, "lookat_triggers" },
    { 0x4F4E, "lookatent" },
    { 0x4F53, "lookdown_zodiacs" },
    { 0x4F56, "lookforenemy" },
    { 0x4F5B, "lookout_guys_logic" },
    { 0x4F5C, "lookupanim" },
    { 0x4F61, "loop" },
    { 0x4F63, "loop_armada" },
    { 0x4F74, "loopers" },
    { 0x4F7E, "loopidlesound" },
    { 0x4F82, "loopmovesound" },
    { 0x4F83, "loopreflectoreffect" },
    { 0x4F8A, "loot_collection_timeout" },
    { 0x4F8D, "loot_pickup_listener" },
    { 0x4F8F, "lootbag" },
    { 0x4F92, "lootbox_world_init" },
    { 0x4F93, "lootboxes" },
    { 0x4F97, "losthope_vo" },
    { 0x4F98, "low_speed_mult" },
    { 0x4F9B, "lower_bottom_bay_door" },
    { 0x4F9D, "lower_shield1" },
    { 0x4F9E, "lower_shield2" },
    { 0x4FA5, "lowertextfontsize" },
    { 0x4FAC, "lowspeed_start_crossfade" },
    { 0x4FAE, "lowy" },
    { 0x4FAF, "lrtest" },
    { 0x4FB2, "lurker_listen_trigger" },
    { 0x4FBC, "m880_crash" },
    { 0x4FC5, "m880_crash_spawn" },
    { 0x4FC8, "m880_kill_collision_change" },
    { 0x4FCC, "m880_open_path" },
    { 0x4FCF, "m880_waits" },
    { 0x4FD6, "magic_bullet_shield" },
    { 0x4FD7, "magic_distance" },
    { 0x4FD8, "magic_missile_fire_at_ent" },
    { 0x4FDA, "magicbullet_spray" },
    { 0x4FE4, "main_chopper_test" },
    { 0x4FE6, "main_common" },
    { 0x4FE8, "main_end_beach" },
    { 0x4FF6, "main_roof_combat" },
    { 0x4FFA, "main_sp" },
    { 0x4FFB, "main_spot" },
    { 0x5001, "maingun_fx_override" },
    { 0x5002, "mainloopstart" },
    { 0x5004, "make_array" },
    { 0x5008, "make_enemy_squad_burst" },
    { 0x5019, "makereviveentity" },
    { 0x501F, "makesureturnworks" },
    { 0x5023, "mall" },
    { 0x5029, "mall_breach_enemy_2" },
    { 0x502A, "mall_breach_enemy_ragdoll_on_death" },
    { 0x502E, "mall_delete_rooftop_ents" },
    { 0x5032, "mall_dialogue_grp1" },
    { 0x503A, "mall_heli" },
    { 0x503E, "mall_lone_patrol_think" },
    { 0x503F, "mall_lookout" },
    { 0x5041, "mall_radioburst_volume" },
    { 0x5043, "mall_rooftop_floor_splash" },
    { 0x504B, "mall_rootop_event" },
    { 0x5051, "mall_teleport_dog" },
    { 0x5053, "mall_walla_volume" },
    { 0x5054, "mallroof_acboxes" },
    { 0x5055, "mallroof_array" },
    { 0x5063, "manage_all_rpg_ai_get_target" },
    { 0x506C, "manage_player_linked_view" },
    { 0x506F, "manage_sentry_count" },
    { 0x5072, "manage_use_region" },
    { 0x5076, "manhandler_hold" },
    { 0x5081, "mansion_hesh" },
    { 0x5086, "mantis_explosion" },
    { 0x5087, "mantis_explosion_fragment" },
    { 0x508B, "manual_clear_hint_on_trigger" },
    { 0x5091, "manualdropthink" },
    { 0x5097, "map_is_early_in_the_game" },
    { 0x5099, "map_without_loadout" },
    { 0x509D, "mapcustomkillstreakfunc" },
    { 0x509F, "mark_completed" },
    { 0x50A1, "mark_friendly_vehicles" },
    { 0x50A3, "mark_suits_for_delete" },
    { 0x50A4, "markedduration" },
    { 0x50AA, "markplayer" },
    { 0x50B0, "mask_interactives_in_volumes" },
    { 0x50B1, "mask_prop" },
    { 0x50B3, "massnodeinitfunctions" },
    { 0x50BB, "matchestargetteam" },
    { 0x50C2, "matchoutcomenotify" },
    { 0x50C5, "matchrules_numinitialinfected" },
    { 0x50CE, "matv" },
    { 0x50D4, "max_capacity" },
    { 0x50D5, "max_count" },
    { 0x50D7, "max_distance" },
    { 0x50D8, "max_drones" },
    { 0x50DC, "max_lurker_population" },
    { 0x50DF, "max_rotation_speed" },
    { 0x50E6, "max_velocity" },
    { 0x50E8, "max_zodiacs" },
    { 0x50E9, "max_zoom_dof" },
    { 0x50EA, "maxalienattackerdifficultyvalue" },
    { 0x50EB, "maxaliveenemycount" },
    { 0x50EE, "maxammopickupsperround" },
    { 0x50F6, "maxdestructions" },
    { 0x50F7, "maxdetpackdamage" },
    { 0x50F8, "maxdirections" },
    { 0x50FA, "maxdistance" },
    { 0x50FD, "maxemissive" },
    { 0x5101, "maxflashedseconds" },
    { 0x5107, "maxlaststands" },
    { 0x510C, "maxnamelength" },
    { 0x5111, "maxpickupsperround" },
    { 0x5114, "maxradius" },
    { 0x511A, "maxtracecount" },
    { 0x511D, "maxvehiclesallowed" },
    { 0x511E, "maxvisibledist_old" },
    { 0x5120, "maydolaststand" },
    { 0x5125, "mayonlydie" },
    { 0x512C, "medbay_cleanup" },
    { 0x512D, "medbay_player_anim" },
    { 0x5133, "meeting_guys" },
    { 0x5136, "meetup_vo" },
    { 0x5140, "melee_aivsai_exposed_chooseanimationandposition_behind" },
    { 0x5142, "melee_aivsai_exposed_chooseanimationandposition_flip" },
    { 0x514C, "melee_aivsai_targetlink" },
    { 0x514F, "melee_clean_up" },
    { 0x5156, "melee_deathhandler_regular" },
    { 0x515B, "melee_dogcleanup" },
    { 0x515D, "melee_droppedweaponrestore" },
    { 0x515E, "melee_endscript" },
    { 0x5160, "melee_endscript_checkpositionandmovement" },
    { 0x516A, "melee_in_posture" },
    { 0x516E, "melee_jumping_to_wall" },
    { 0x516F, "melee_kill_stab" },
    { 0x517B, "melee_player_lerp_back" },
    { 0x517C, "melee_playfacialanim" },
    { 0x517E, "melee_resetaction" },
    { 0x517F, "melee_scalar" },
    { 0x5180, "melee_setjumpanimstates" },
    { 0x5181, "melee_standard_checktimeconstraints" },
    { 0x5187, "melee_standard_resetgiveuptime" },
    { 0x518A, "melee_stealthcheck" },
    { 0x518C, "melee_success" },
    { 0x518F, "melee_synch" },
    { 0x5190, "melee_synch_attack" },
    { 0x5191, "melee_trigger" },
    { 0x519B, "meleebiteattackplayer" },
    { 0x51A0, "meleefailed" },
    { 0x51A1, "meleeforcedexposedflip" },
    { 0x51A2, "meleeforcedexposedwrestle" },
    { 0x51A6, "meleekilltarget" },
    { 0x51A7, "meleeplayerwhilemoving" },
    { 0x51AB, "meleestrength" },
    { 0x51AC, "meleestruggle_istraverse" },
    { 0x51AE, "meleestrugglevsai_first_attack" },
    { 0x51AF, "meleestrugglevsai_interrupted_animcustom" },
    { 0x51B0, "meleestrugglevsai_interrupted_animcustom_cleanup" },
    { 0x51B3, "meleestrugglevsai_supershort" },
    { 0x51B5, "meleestrugglevsdog" },
    { 0x51B9, "meleestrugglevsdog_justdie" },
    { 0x51BB, "meleestrugglevsdog_traverse" },
    { 0x51BC, "meleestun" },
    { 0x51C4, "memberremovestrings" },
    { 0x51C9, "menu_create" },
    { 0x51CB, "menu_fx_creation" },
    { 0x51CC, "menu_fx_option_set" },
    { 0x51CD, "menu_none" },
    { 0x51D2, "menuspectator" },
    { 0x51D4, "merrick_goes_green" },
    { 0x51D5, "merrick_handle_name_when_scanning" },
    { 0x51D7, "merrick_medbay" },
    { 0x51D8, "merrick_move_to_edge" },
    { 0x51DA, "merrick_scene_dialogue" },
    { 0x51DF, "merrick_shoots_first_guy" },
    { 0x51E1, "metal_detector" },
    { 0x51E2, "metal_detector_dmg_monitor" },
    { 0x51E3, "metal_detector_touch_monitor" },
    { 0x51E4, "metal_detector_weapons" },
    { 0x51E7, "meteorfireworkstructs" },
    { 0x51E8, "meteoroid" },
    { 0x51EF, "meteoroid_start_angles" },
    { 0x51F0, "meteoroid_start_pos" },
    { 0x51F1, "methodsinit" },
    { 0x51F2, "mg" },
    { 0x51F6, "mg42_gunner_manual_think" },
    { 0x51F7, "mg42_gunner_think" },
    { 0x51F9, "mg42_setdifficulty" },
    { 0x51FA, "mg42_suppressionfire" },
    { 0x51FC, "mg42_think" },
    { 0x51FF, "mg42badplace_mintime" },
    { 0x5200, "mg42pain" },
    { 0x5203, "mg_gunner_team" },
    { 0x5204, "mg_target" },
    { 0x5205, "mg_turret_do_something_while_waiting_for_player" },
    { 0x5206, "mginit" },
    { 0x5209, "mgon" },
    { 0x520D, "mgturret_auto" },
    { 0x5220, "mig29_missile_dives" },
    { 0x5234, "mindetpackdamage" },
    { 0x5235, "mindistancecallout" },
    { 0x5237, "mine_beacon" },
    { 0x523A, "mine_explode" },
    { 0x523D, "mine_notify_on_level" },
    { 0x5240, "mine_spin" },
    { 0x5246, "minedamagemin" },
    { 0x5247, "minedamagemonitor" },
    { 0x5248, "minedamageradius" },
    { 0x524B, "minedetectionheight" },
    { 0x524C, "minedetectionradius" },
    { 0x5252, "mineproximitytrigger" },
    { 0x5257, "mineselfdestructtime" },
    { 0x525A, "minethrown" },
    { 0x5262, "mini_sub_prop_wash_stop" },
    { 0x5266, "minigun_fire" },
    { 0x5269, "minigun_hints_off" },
    { 0x526A, "minigun_hints_on" },
    { 0x526B, "minigun_ignoreme" },
    { 0x526C, "minigun_rumble" },
    { 0x5274, "minigun_turret_placed_listener" },
    { 0x5275, "minigun_turret_watch_ammo" },
    { 0x5276, "minigun_used" },
    { 0x5277, "minigunchance" },
    { 0x527A, "minigunsspinning" },
    { 0x527F, "minimapheight" },
    { 0x5280, "minimaporigin" },
    { 0x5283, "minion_approach" },
    { 0x5288, "minradius" },
    { 0x528B, "missed_enemy" },
    { 0x5294, "missile_dist_internal" },
    { 0x5299, "missile_find_ground_target" },
    { 0x52A9, "missile_launcher_destruction_vignette" },
    { 0x52AF, "missile_move_firemissile" },
    { 0x52B5, "missile_starts" },
    { 0x52BA, "missile_towerbuzz" },
    { 0x52BB, "missile_trackrealtarget" },
    { 0x52BD, "missile_truck_fire_missile" },
    { 0x52C0, "missileattractor" },
    { 0x52C9, "missileisgoodtarget" },
    { 0x52CB, "missilelosetarget" },
    { 0x52D0, "missiles_chasing" },
    { 0x52D4, "missiletags" },
    { 0x52D5, "missiletargetangle" },
    { 0x52D6, "missiletargetflareradius" },
    { 0x52D8, "missing_animation_parameters" },
    { 0x52DA, "mission_fail_func" },
    { 0x52DC, "mission_failed_garage" },
    { 0x52E0, "mission_finished" },
    { 0x52E1, "mission_flag_inits" },
    { 0x52EA, "mission_post_inits" },
    { 0x52EC, "mission_recon" },
    { 0x52EF, "missionfail" },
    { 0x52F0, "missionfail_rorke" },
    { 0x52FA, "mix" },
    { 0x52FE, "mk32_badassery" },
    { 0x5308, "mlrs_start_qte" },
    { 0x530A, "mobilemortar" },
    { 0x530B, "mod" },
    { 0x530E, "model_dummy_death" },
    { 0x530F, "model_init" },
    { 0x5310, "model_name_dummy" },
    { 0x5311, "model_name_enemy" },
    { 0x5314, "modelbase" },
    { 0x5316, "modelbombsquad" },
    { 0x5319, "modeldummyon" },
    { 0x531A, "modelmgturret" },
    { 0x5322, "modify_player_speed" },
    { 0x5324, "modify_sentry_setting" },
    { 0x5325, "modifydamage" },
    { 0x5328, "mods_override" },
    { 0x5329, "modulate_speed_based_on_progress" },
    { 0x532B, "module_int" },
    { 0x5332, "monitor_attackable_ent_damage" },
    { 0x5334, "monitor_bar_drift" },
    { 0x5338, "monitor_can_cut_rope" },
    { 0x5339, "monitor_cautious_approach_dangerous_locations" },
    { 0x533E, "monitor_controls_and_fx" },
    { 0x533F, "monitor_deaths_on_dynamic_array" },
    { 0x5342, "monitor_drill_complete" },
    { 0x5343, "monitor_enemies_in_pods" },
    { 0x5345, "monitor_enemy_for_downed" },
    { 0x534A, "monitor_fx" },
    { 0x534C, "monitor_guy_moveup" },
    { 0x5350, "monitor_missile_distance" },
    { 0x5353, "monitor_pain" },
    { 0x535F, "monitor_spawners" },
    { 0x5362, "monitor_touching" },
    { 0x5363, "monitor_trig_activation" },
    { 0x5366, "monitorads_blend_dof" },
    { 0x5367, "monitorads_zoom_elem_offset" },
    { 0x5368, "monitorads_zoom_elem_reset" },
    { 0x5369, "monitorads_zoom_hud_delay" },
    { 0x536B, "monitorads_zoom_out" },
    { 0x5379, "monitorconcussion" },
    { 0x5384, "monitorenemymissilefire" },
    { 0x538B, "monitorflares" },
    { 0x5397, "monitorhealth_ondeath" },
    { 0x5398, "monitorhealth_ondeath_apache_crash" },
    { 0x539E, "monitorkills" },
    { 0x53A2, "monitorlivetime" },
    { 0x53A6, "monitormagcycle" },
    { 0x53A8, "monitormarkervisibility" },
    { 0x53AA, "monitormisc" },
    { 0x53AD, "monitormovedown" },
    { 0x53AE, "monitormovementdistance" },
    { 0x53B3, "monitorpositioncamping" },
    { 0x53B4, "monitorprocesschallenge" },
    { 0x53B5, "monitorreload" },
    { 0x53B9, "monitorrocketfire2" },
    { 0x53BD, "monitorscopechange" },
    { 0x53C1, "monitorsinglesprintdistance" },
    { 0x53C2, "monitorspecialroundend" },
    { 0x53C8, "monitorstateshud" },
    { 0x53CD, "monitorsupportdropprogress" },
    { 0x53D2, "monitorthermalvision" },
    { 0x53D5, "monitorturretfire" },
    { 0x53DE, "mortar_ends" },
    { 0x53DF, "mortar_fire_on_struct" },
    { 0x53EE, "mortardamageradius" },
    { 0x53EF, "mortardamagetriggerdist" },
    { 0x53F0, "mortarearthquakeradius" },
    { 0x53F3, "mortarfx" },
    { 0x53F4, "mortargroup" },
    { 0x53F7, "mortarnoincomingsound" },
    { 0x53F9, "mortarrecoil" },
    { 0x53FA, "mortars" },
    { 0x53FD, "mortartrigger" },
    { 0x5400, "mosley_airlock_ln_1" },
    { 0x5401, "mosley_airlock_ln_2" },
    { 0x5402, "motion_light" },
    { 0x5407, "motion_trigger" },
    { 0x5409, "motionsensormarkedby" },
    { 0x540D, "mount_tank" },
    { 0x5417, "movable_cover_init" },
    { 0x541C, "movable_cover_move_delay" },
    { 0x541E, "movable_cover_parse_parameters" },
    { 0x5421, "movable_cover_trigger" },
    { 0x5422, "movable_cover_update_use_icon" },
    { 0x5423, "movable_cover_use_icon" },
    { 0x5425, "movable_cover_wait_for_user_or_timeout" },
    { 0x5426, "movable_type" },
    { 0x5427, "movables" },
    { 0x5429, "move_all_fx" },
    { 0x542A, "move_allies_from_fire_blocker" },
    { 0x542C, "move_ally_to_mesh" },
    { 0x542E, "move_angles" },
    { 0x542F, "move_apache_to_main_island" },
    { 0x5430, "move_arc" },
    { 0x5431, "move_arc_dist" },
    { 0x5432, "move_arc_zodiac" },
    { 0x5437, "move_check" },
    { 0x5439, "move_controlroom_to_new_location" },
    { 0x5440, "move_explosion_buildup_rumble" },
    { 0x5448, "move_player_to_start_point" },
    { 0x544A, "move_previous" },
    { 0x544B, "move_primary_light" },
    { 0x5451, "move_side" },
    { 0x5452, "move_speed_scalar" },
    { 0x5458, "move_state_loop" },
    { 0x5459, "move_state_loop_run" },
    { 0x545A, "move_state_shift_back" },
    { 0x545B, "move_state_start" },
    { 0x545F, "move_to_breach" },
    { 0x5468, "move_transition_arrays" },
    { 0x5471, "move_with_plane" },
    { 0x5472, "move_with_rate" },
    { 0x5477, "movecovertocover_checkstartpose" },
    { 0x547A, "movedlow" },
    { 0x547C, "movedrecently" },
    { 0x547E, "moveloop" },
    { 0x5487, "movement_back" },
    { 0x548C, "movemsg" },
    { 0x548F, "moveplaybackrate" },
    { 0x5492, "mover_candidates" },
    { 0x5493, "mover_delete" },
    { 0x5496, "movercreate" },
    { 0x549A, "moverun" },
    { 0x549B, "moverwaitforuse" },
    { 0x549F, "movespeed_multiplier" },
    { 0x54A2, "movespeedscale" },
    { 0x54A5, "movestartbattlechatter" },
    { 0x54A8, "moveswim" },
    { 0x54AC, "moveswim_combat_forward_enter" },
    { 0x54AD, "moveswim_combat_forward_exit" },
    { 0x54B0, "moveswim_combat_strafe_exit" },
    { 0x54B5, "moveswim_set" },
    { 0x54BB, "moveto_speed" },
    { 0x54C6, "moving" },
    { 0x54C9, "moving_cover_death" },
    { 0x54CA, "moving_cover_guys" },
    { 0x54CB, "moving_cover_jumped" },
    { 0x54CC, "moving_cover_lightsoff" },
    { 0x54CE, "moving_cover_obj1" },
    { 0x54D4, "moving_crates_plane" },
    { 0x54D7, "moving_jeeps_and_crates" },
    { 0x54DC, "moving_platform_empty_func" },
    { 0x54DE, "moving_water" },
    { 0x54E2, "mp_alien_town_intro_drill_setup" },
    { 0x54EC, "mp_dart_is_light_entity" },
    { 0x54EE, "mp_dart_restarteffect" },
    { 0x54EF, "mp_dart_tv_flicker" },
    { 0x54F3, "mssl_launch_destory_sfx" },
    { 0x54F9, "mugger_add_extra_tag" },
    { 0x54FA, "mugger_bank_limit" },
    { 0x54FD, "mugger_delayed_banking" },
    { 0x5500, "mugger_first_unused_or_oldest_extra_tag" },
    { 0x5502, "mugger_fx_playing" },
    { 0x550E, "mugger_jackpot_tags_spawned" },
    { 0x5511, "mugger_jackpot_timer" },
    { 0x5514, "mugger_last_mega_drop" },
    { 0x5515, "mugger_max_extra_tags" },
    { 0x551B, "mugger_pile_icon_remove" },
    { 0x551C, "mugger_scorelimit" },
    { 0x551D, "mugger_tag_pickup_wait" },
    { 0x5526, "multikill" },
    { 0x5529, "multiple_c4" },
    { 0x552E, "multiteambased" },
    { 0x5532, "music_escape_hot" },
    { 0x5537, "music_loop_stealth" },
    { 0x5539, "music_on_flag" },
    { 0x553A, "music_play" },
    { 0x553C, "music_play_jg" },
    { 0x5540, "music_stealth_tension_loop" },
    { 0x5544, "musiclength" },
    { 0x5546, "muzzleflashoverride" },
    { 0x5547, "my_animnode" },
    { 0x554B, "my_speed" },
    { 0x5554, "n_door_knock" },
    { 0x5555, "n_hesh_stumble_and_path" },
    { 0x5556, "n_vehicle_1" },
    { 0x555A, "n_vehicle_3" },
    { 0x555F, "nag_if_shot" },
    { 0x5565, "nag_player_to_jump" },
    { 0x556C, "name1" },
    { 0x5573, "names" },
    { 0x5580, "near_dist_sq" },
    { 0x5584, "neardistance" },
    { 0x5586, "nearest_nodes" },
    { 0x5587, "nearest_point_on_pathgrid" },
    { 0x558A, "necksnapped" },
    { 0x558D, "needflexibleheightsupport" },
    { 0x558E, "needrecalculategoodshootpos" },
    { 0x55A3, "neutralflagfx" },
    { 0x55A6, "neverenablecqb" },
    { 0x55B6, "new_glowstick_scene" },
    { 0x55B9, "new_l_pip_corner" },
    { 0x55BD, "new_rope_roll" },
    { 0x55C0, "new_submix_hud" },
    { 0x55C5, "new_volmod_hud" },
    { 0x55C8, "newenemyreactiondistsq_old" },
    { 0x55C9, "newenemysurprisedreaction" },
    { 0x55CA, "newfallback_overmind" },
    { 0x55CC, "newradialbutton" },
    { 0x55CE, "newrandomcrate" },
    { 0x55D0, "next" },
    { 0x55D1, "next_bomb_damage_vo_time" },
    { 0x55D4, "next_earthquake" },
    { 0x55D8, "next_minion_vo_time" },
    { 0x55DF, "next_reactive_time" },
    { 0x55E3, "nextallowedlooktime" },
    { 0x55E8, "nextdoorgrenadetime" },
    { 0x55EB, "nextgrenadedrop" },
    { 0x55F1, "nextmeleechecktarget" },
    { 0x55F2, "nextmeleechecktime" },
    { 0x55F5, "nextmissiletag" },
    { 0x55F8, "nextorigin" },
    { 0x55F9, "nextpeekoutattempttime" },
    { 0x55FB, "nextsaytimes" },
    { 0x55FD, "nextstandinghitdying" },
    { 0x5600, "nh90_doors_open" },
    { 0x5603, "nightvision_dlight" },
    { 0x5605, "nightvision_effectsoff" },
    { 0x560A, "nightvision_started" },
    { 0x560D, "ninebangexplodewaiter" },
    { 0x5611, "nml_locations" },
    { 0x5612, "no_abilities_timer" },
    { 0x5614, "no_attack_hint" },
    { 0x5618, "no_crouch_or_prone_think_for_player" },
    { 0x5619, "no_delete" },
    { 0x561C, "no_edge_death" },
    { 0x5621, "no_go_back_into_field" },
    { 0x5627, "no_moving_unresolved_collisions" },
    { 0x562B, "no_prone" },
    { 0x5632, "no_swept_hint" },
    { 0x5634, "no_tremor" },
    { 0x5635, "no_vehicle_getoutanim" },
    { 0x5636, "no_vehicle_ragdoll" },
    { 0x563E, "node_array" },
    { 0x563F, "node_change_logic" },
    { 0x5640, "node_closest" },
    { 0x5646, "node_is_valid_outside_for_vanguard" },
    { 0x5649, "node_search" },
    { 0x564A, "node_test" },
    { 0x564E, "node_within_use_radius_of_crate" },
    { 0x5654, "nodes" },
    { 0x5655, "nodes_array" },
    { 0x5658, "nodes_start" },
    { 0x5659, "nodescore" },
    { 0x565B, "nodisconnect" },
    { 0x565C, "nodroneweaponsound" },
    { 0x565F, "nofirstframemelee" },
    { 0x5660, "nofour" },
    { 0x5662, "nogun" },
    { 0x5666, "noise_amplitude" },
    { 0x5669, "nomeleechargedelay" },
    { 0x5675, "noragdollents" },
    { 0x5678, "normal_speed" },
    { 0x5681, "norunreload" },
    { 0x5688, "not_closing" },
    { 0x568D, "notargethudelem" },
    { 0x5690, "note_track_start_fx_on_tag" },
    { 0x5695, "noteleport" },
    { 0x569B, "notetrack_blast_shake_early" },
    { 0x569D, "notetrack_command_dialog_end" },
    { 0x56A0, "notetrack_control_room_allow_free_look" },
    { 0x56A3, "notetrack_derrick_chunk_hit_barrels" },
    { 0x56A6, "notetrack_derrick_debris_hitground" },
    { 0x56A8, "notetrack_derrick_large_explosion" },
    { 0x56A9, "notetrack_derrick_small_explosion" },
    { 0x56AB, "notetrack_end_slomo" },
    { 0x56B8, "notetrack_oiltank_catwalk_swap" },
    { 0x56BC, "notetrack_player_breach_water" },
    { 0x56C3, "notetrack_release_allies" },
    { 0x56C5, "notetrack_shake_start" },
    { 0x56CD, "notetrack_start_slomo" },
    { 0x56CE, "notetrack_swim_begin_player_control" },
    { 0x56D8, "notetrackalertnesscasual" },
    { 0x56D9, "notetrackbodyfall" },
    { 0x56DD, "notetrackfire" },
    { 0x56E5, "notetrackguntochest" },
    { 0x56EA, "notetrackmovementstop" },
    { 0x56EC, "notetrackpistolpickup" },
    { 0x56ED, "notetrackpistolputaway" },
    { 0x56EE, "notetrackpistolrechamber" },
    { 0x56F0, "notetrackposecrawl" },
    { 0x56FC, "noteworthy_check" },
    { 0x56FD, "notfirsttime" },
    { 0x56FE, "notfirsttimedogs" },
    { 0x56FF, "notfirsttimedogvests" },
    { 0x5703, "notify_baddies_to_retreat" },
    { 0x5706, "notify_enable" },
    { 0x570B, "notify_moving_platform_invalid" },
    { 0x570F, "notify_on_damage" },
    { 0x571A, "notify_spotted_on_damage" },
    { 0x571C, "notify_trigger" },
    { 0x571E, "notifyaftertime" },
    { 0x5721, "notifydamage" },
    { 0x5722, "notifydamageafterframe" },
    { 0x5723, "notifydamagenotdone" },
    { 0x572A, "notifyonanimend" },
    { 0x572E, "notifyoverlay" },
    { 0x572F, "notifyroundover" },
    { 0x5733, "notifytext" },
    { 0x5737, "notsolid_ents" },
    { 0x573C, "npc_physics_pulse" },
    { 0x573E, "npc_tank_combat_init" },
    { 0x573F, "npcid" },
    { 0x5743, "nuke_blur" },
    { 0x5747, "nuke_empjam" },
    { 0x574E, "nuke_soundobject" },
    { 0x5753, "nukecankill" },
    { 0x5756, "nukecratethink" },
    { 0x5759, "nukedeathsimple" },
    { 0x575A, "nukedetonated" },
    { 0x5765, "nukeslowmo" },
    { 0x576A, "nukevisioninprogress" },
    { 0x5771, "num_players_left" },
    { 0x5773, "numagents" },
    { 0x5774, "numareas" },
    { 0x5776, "numberoffishinexistence" },
    { 0x577B, "numbombs" },
    { 0x5781, "numdropcrates" },
    { 0x5782, "numenemiesclose" },
    { 0x5785, "numexplosivesexceedmodelcapacity" },
    { 0x5787, "numfriendlyvoices" },
    { 0x5789, "numgasstrikeactive" },
    { 0x578A, "numgrenadesinprogresstowardsplayer" },
    { 0x5794, "numspeakers" },
    { 0x579D, "nvg_animted_scene" },
    { 0x57A0, "nvg_goggles_off" },
    { 0x57A5, "nvg_on_check" },
    { 0x57A9, "nvgs_on_blackout" },
    { 0x57AB, "obj" },
    { 0x57AF, "obj_capture_hvt" },
    { 0x57B0, "obj_clear_deck" },
    { 0x57B5, "obj_enterbase" },
    { 0x57B8, "obj_exists" },
    { 0x57BA, "obj_flags" },
    { 0x57BB, "obj_flight_deck" },
    { 0x57BC, "obj_getingetajaxgetout" },
    { 0x57BD, "obj_getinjeep" },
    { 0x57C0, "obj_origin" },
    { 0x57C3, "obj_sparrow" },
    { 0x57C9, "objective" },
    { 0x57CA, "objective_breach" },
    { 0x57CF, "objective_is_inactive" },
    { 0x57D0, "objective_outline_add" },
    { 0x57D1, "objective_outline_remove" },
    { 0x57DB, "objectivepointsmod" },
    { 0x57DC, "objectives" },
    { 0x57DD, "objectivescaler" },
    { 0x57E1, "objid_axis" },
    { 0x57E9, "objpoints" },
    { 0x57EF, "occlusion" },
    { 0x57F1, "occupied" },
    { 0x57FD, "odin_assault_perform_action" },
    { 0x57FE, "odin_clear_using" },
    { 0x57FF, "odin_control_player_speed" },
    { 0x5801, "odin_drop_weapon" },
    { 0x5806, "odin_fac_cart" },
    { 0x580A, "odin_fireweapon" },
    { 0x580C, "odin_flag_inits" },
    { 0x580E, "odin_fov" },
    { 0x5815, "odin_hall_escape_turn02_ally_spawn" },
    { 0x5817, "odin_hall_escape_turn02_player_spawn" },
    { 0x5818, "odin_hint_string_init" },
    { 0x5820, "odin_leave" },
    { 0x5829, "odin_overlay_ent" },
    { 0x5830, "odin_satellite" },
    { 0x5831, "odin_script_setup" },
    { 0x583B, "odin_waitfordonefiring" },
    { 0x583E, "odin_watchoutlines" },
    { 0x5844, "odin_zoom_up" },
    { 0x5845, "odinsettings" },
    { 0x5847, "off_flashlight" },
    { 0x5853, "office_enemy_vo" },
    { 0x585E, "officercount" },
    { 0x585F, "officerid" },
    { 0x5860, "officers" },
    { 0x5865, "offscreen_shader_blink" },
    { 0x5866, "offset" },
    { 0x5869, "offset_percent" },
    { 0x586D, "offsetdist" },
    { 0x586E, "offsetone" },
    { 0x5871, "offsetzero" },
    { 0x5873, "og_animname" },
    { 0x5876, "og_health" },
    { 0x5879, "og_script_stealthgroup" },
    { 0x587D, "oilrig_jumpsuit_ondamage" },
    { 0x587F, "oilrig_jumpsuit_precache" },
    { 0x5880, "oilrig_jumpsuits" },
    { 0x5881, "oilrocks_apache_hint_timers" },
    { 0x5882, "oilrocks_locations" },
    { 0x5883, "old" },
    { 0x5889, "old_ai_target" },
    { 0x588A, "old_alpha" },
    { 0x588B, "old_ang" },
    { 0x588E, "old_baseaccuracy" },
    { 0x588F, "old_color" },
    { 0x5892, "old_disablearrivals" },
    { 0x5893, "old_dist" },
    { 0x5895, "old_fixednode" },
    { 0x589A, "old_ignoreme" },
    { 0x58A1, "old_movementtype" },
    { 0x58AF, "old_weap" },
    { 0x58B0, "old_weapon" },
    { 0x58B4, "oldbaseaccuracy" },
    { 0x58B6, "oldcombatmode" },
    { 0x58BB, "oldfightdist" },
    { 0x58BD, "oldgoalradius" },
    { 0x58BF, "oldgrenadeawareness" },
    { 0x58C5, "oldmaxdist" },
    { 0x58CC, "oldorigin" },
    { 0x58CF, "oldprimarygun" },
    { 0x58D0, "oldradius" },
    { 0x58D9, "on_agent_squadmate_killed" },
    { 0x58DA, "on_alien_type_killed" },
    { 0x58DC, "on_bridge" },
    { 0x58DF, "on_damaged_finished" },
    { 0x58E3, "on_fire_sabot" },
    { 0x58E4, "on_hintstring" },
    { 0x58EA, "on_path_grid" },
    { 0x58EB, "on_pop_smoke" },
    { 0x58F0, "onagentkilled" },
    { 0x58F1, "onaiconnect" },
    { 0x58F3, "onalienagentkilled" },
    { 0x58F7, "onatv" },
    { 0x58F8, "onback" },
    { 0x5907, "ondamaged" },
    { 0x5908, "ondamageddelegate" },
    { 0x5909, "ondamagefinish" },
    { 0x590D, "ondeath" },
    { 0x5911, "ondeathfunc" },
    { 0x5914, "ondestroycallback" },
    { 0x5917, "ondetonateexplosive" },
    { 0x591D, "one_missile_kill" },
    { 0x591F, "onemanarmyweaponchangetracker" },
    { 0x5921, "onenter" },
    { 0x5923, "onentercallback" },
    { 0x5925, "onenterdot_player" },
    { 0x5927, "onenterdot_poisondamageoverlay" },
    { 0x592A, "onenterstate" },
    { 0x5937, "onfinalsurvivor" },
    { 0x5939, "onfirerocket" },
    { 0x593A, "onfirstspawnedplayer" },
    { 0x5940, "ongroundpos" },
    { 0x5941, "onhalftime" },
    { 0x5943, "onhitpitchclamp" },
    { 0x5944, "onjoinedspectators" },
    { 0x5947, "onkill" },
    { 0x5948, "onkillstreakdisowned" },
    { 0x5949, "onkillstreakkilled" },
    { 0x594F, "only_allowable_tactical_goals" },
    { 0x5954, "onmovingplatformcollision" },
    { 0x5957, "onpickup" },
    { 0x5961, "onplayerconnecting" },
    { 0x5969, "onplayerspawned" },
    { 0x596E, "onrespawndelay" },
    { 0x596F, "onrotatingvehicleturret" },
    { 0x5973, "onscorelimit" },
    { 0x5983, "ontriggeredsfx" },
    { 0x5986, "onusedefuseobject" },
    { 0x5988, "onuseplantobject" },
    { 0x598B, "onweapondamage" },
    { 0x598D, "op_barracks" },
    { 0x598E, "op_helo" },
    { 0x5990, "open_and_connect" },
    { 0x5991, "open_angles" },
    { 0x5992, "open_church_doors" },
    { 0x599E, "open_loading_dock_doors" },
    { 0x59A4, "open_up_player_view_during_weld" },
    { 0x59AC, "opened_y" },
    { 0x59B6, "opfor_bc_lmg" },
    { 0x59B7, "opfor_catwalk_falling_death" },
    { 0x59B9, "opfor_common_ambush" },
    { 0x59BA, "opfor_crawl" },
    { 0x59BB, "opfor_death_runners" },
    { 0x59BD, "opfor_infil" },
    { 0x59C1, "opfor_m880_escape" },
    { 0x59C6, "opfor_retreat" },
    { 0x59CA, "opfor_starting_runners" },
    { 0x59CB, "opsrey_control_nag_vo" },
    { 0x59CC, "optimal_height" },
    { 0x59D4, "options" },
    { 0x59D8, "orderonqueueddialog" },
    { 0x59E1, "orient_to_face" },
    { 0x59E2, "orientmeleevictim" },
    { 0x59E6, "orig_org" },
    { 0x59EC, "orig_wakeupradius" },
    { 0x59F1, "origin_offset" },
    { 0x59F3, "original_angles" },
    { 0x59F4, "original_height" },
    { 0x59F6, "original_pos" },
    { 0x59F9, "original_yaw" },
    { 0x5A03, "osprey2_gunship_attack" },
    { 0x5A07, "osprey_debug_ai" },
    { 0x5A09, "osprey_hit_fake_targets" },
    { 0x5A0B, "osprey_hit_zodiacs" },
    { 0x5A0E, "osprey_missile_side_left" },
    { 0x5A0F, "osprey_props" },
    { 0x5A17, "other_col_point" },
    { 0x5A1C, "otherdir" },
    { 0x5A31, "outline_init" },
    { 0x5A37, "outline_set_global_width" },
    { 0x5A38, "outline_switch" },
    { 0x5A49, "outlinegethighestinfoforplayer" },
    { 0x5A4A, "outlinegethighestpriorityid" },
    { 0x5A4C, "outlineonplayerdisconnect" },
    { 0x5A51, "outlineremoveplayerfromvisibletoarrays" },
    { 0x5A55, "outro_gate" },
    { 0x5A56, "outro_heli_front" },
    { 0x5A57, "outro_heli_mid" },
    { 0x5A58, "outro_heli_rear" },
    { 0x5A62, "outside_reads" },
    { 0x5A63, "outside_start" },
    { 0x5A6C, "overheat" },
    { 0x5A6E, "overheattime" },
    { 0x5A7C, "override_array_delete" },
    { 0x5A7F, "override_check" },
    { 0x5A80, "override_class_function" },
    { 0x5A81, "override_crawl_death_anims" },
    { 0x5A85, "override_ride_anims" },
    { 0x5A8E, "ownerboarded" },
    { 0x5A90, "ownerfx" },
    { 0x5A94, "ownerradiussq" },
    { 0x5A95, "ownersattacker" },
    { 0x5A97, "ownershipstring" },
    { 0x5A99, "ownerteamid" },
    { 0x5A9D, "pa_announcements_chaos_thread" },
    { 0x5AA7, "pain_interval_monitor" },
    { 0x5AAB, "pain_resistance" },
    { 0x5AAE, "pain_setflaggedanimknobrestart" },
    { 0x5AAF, "pain_test" },
    { 0x5AB0, "painai" },
    { 0x5AB4, "painonstairs" },
    { 0x5ABB, "pairbattlebuddy" },
    { 0x5AC8, "panel_array" },
    { 0x5ACC, "panicdistance" },
    { 0x5AD5, "parachute_function" },
    { 0x5ADA, "parachute_intro_sound" },
    { 0x5ADE, "parachute_player_land_sounds" },
    { 0x5AE1, "parachute_unload" },
    { 0x5AEA, "parm2" },
    { 0x5AFA, "part_2_draw_player" },
    { 0x5AFF, "part_2_gun_reload" },
    { 0x5B00, "part_2_help_ally" },
    { 0x5B05, "part_2_speedloader" },
    { 0x5B06, "parthealths" },
    { 0x5B09, "participation_point_cap" },
    { 0x5B12, "passenger2turret_anime" },
    { 0x5B16, "passes_spawn_node_filter" },
    { 0x5B17, "passive_missile_settargetandflightmode" },
    { 0x5B1A, "pastswitch" },
    { 0x5B1D, "path_clip" },
    { 0x5B24, "path_gate_open" },
    { 0x5B28, "path_solids" },
    { 0x5B31, "pathdataavailable" },
    { 0x5B33, "pathgoal" },
    { 0x5B35, "pathrandompercent_set" },
    { 0x5B37, "pathstart" },
    { 0x5B48, "patrol_react_anim_count" },
    { 0x5B4C, "patrol_scriptedanims" },
    { 0x5B5F, "pavelowmadeselectionvo" },
    { 0x5B67, "pc_dog_drive_end" },
    { 0x5B69, "pc_dog_drive_killed_by_sniper" },
    { 0x5B6B, "pc_house_baker_get_ready" },
    { 0x5B84, "perk_getdrillhealthscalar" },
    { 0x5B85, "perk_getdrilltimescalar" },
    { 0x5B8B, "perk_getpistoloverkill" },
    { 0x5B8F, "perk_getrevivetimescalar" },
    { 0x5B9E, "perksperkname" },
    { 0x5BAB, "persistence_weaponstats" },
    { 0x5BAF, "perslog_weaponstats" },
    { 0x5BB2, "personalcoldbreathstop" },
    { 0x5BB6, "personalitymanuallyset" },
    { 0x5BB9, "pet" },
    { 0x5BBC, "pet_patrol" },
    { 0x5BBE, "pet_patrol_get_available_origin" },
    { 0x5BC8, "phalanx_gun_fire_at_missiles" },
    { 0x5BCA, "phalanx_gun_fire_target" },
    { 0x5BCB, "phalanx_gun_offline" },
    { 0x5BD0, "pharm_roller_shut" },
    { 0x5BD1, "phone" },
    { 0x5BD2, "photo_copier" },
    { 0x5BD4, "photo_copier_init" },
    { 0x5BD8, "photo_light_flicker" },
    { 0x5BD9, "phys_blockers_brush" },
    { 0x5BDA, "phys_blockers_model" },
    { 0x5BE2, "physicsdefaultdamagepush" },
    { 0x5BE3, "physicsjolt_proximity" },
    { 0x5BE4, "physicsmodel" },
    { 0x5BEC, "pianodamagethink" },
    { 0x5BED, "pianothink" },
    { 0x5BF0, "pick_drone_gundown_dummy" },
    { 0x5BF4, "pickai" },
    { 0x5BF8, "pickup_guns" },
    { 0x5BFC, "pickupfunc" },
    { 0x5BFE, "pickupmodelobj" },
    { 0x5BFF, "pickupobjectdelay" },
    { 0x5C03, "pickuptimer" },
    { 0x5C0B, "pillage_area_has_petbomb" },
    { 0x5C0D, "pillage_init" },
    { 0x5C0E, "pillage_spot" },
    { 0x5C10, "pillage_trigger" },
    { 0x5C12, "pillageable_attachments" },
    { 0x5C15, "pillageable_attachments_dmr" },
    { 0x5C16, "pillageable_attachments_lmg" },
    { 0x5C1B, "pillageable_explosives" },
    { 0x5C1C, "pillageinfo" },
    { 0x5C1E, "pilot" },
    { 0x5C22, "pilot_ai_spawner" },
    { 0x5C29, "pip_ai_cam" },
    { 0x5C2A, "pip_border" },
    { 0x5C2B, "pip_camera" },
    { 0x5C30, "pip_enable" },
    { 0x5C33, "pip_open" },
    { 0x5C39, "pip_static_lines" },
    { 0x5C3B, "pip_toggle_ai_cam" },
    { 0x5C3F, "pipe_calc_ballistic" },
    { 0x5C41, "pipe_calc_splash" },
    { 0x5C43, "pipe_deck_fx" },
    { 0x5C45, "pipe_fx_array" },
    { 0x5C46, "pipe_logic" },
    { 0x5C52, "pistol_ammo_remaining" },
    { 0x5C54, "pistol_overkill" },
    { 0x5C57, "pistolcombatspeedscalar" },
    { 0x5C58, "pistolshoot" },
    { 0x5C5F, "pitch_offset_ground" },
    { 0x5C66, "place_defenses" },
    { 0x5C67, "place_sentry" },
    { 0x5C69, "place_weapon_on" },
    { 0x5C6B, "placed_sentry" },
    { 0x5C6C, "placedims" },
    { 0x5C6D, "placedmodel" },
    { 0x5C6E, "placedsfx" },
    { 0x5C78, "placeweaponon" },
    { 0x5C7C, "plan_crash_run" },
    { 0x5C8F, "plane_tail" },
    { 0x5C94, "planemodel" },
    { 0x5C95, "planemove" },
    { 0x5C96, "planes" },
    { 0x5CA1, "play2dspawnsound" },
    { 0x5CA3, "play_ally_anim" },
    { 0x5CA4, "play_ally_launcher_vignette" },
    { 0x5CA9, "play_ambient_sfx_int" },
    { 0x5CAA, "play_anim_and_end_slowmo_logic" },
    { 0x5CAF, "play_baker_anim" },
    { 0x5CB0, "play_basic_pain_overlay" },
    { 0x5CB4, "play_chatter" },
    { 0x5CBA, "play_console_scene" },
    { 0x5CC6, "play_distant_cloud_design" },
    { 0x5CCE, "play_fire" },
    { 0x5CD0, "play_foilage_sound_custom" },
    { 0x5CD4, "play_fullscreen_shader" },
    { 0x5CD8, "play_fx_for_sub_back" },
    { 0x5CD9, "play_fx_for_sub_blow" },
    { 0x5CDA, "play_fx_for_sub_front" },
    { 0x5CDC, "play_fx_on_tag" },
    { 0x5CE0, "play_health_regen_anim" },
    { 0x5CE3, "play_hive_anim" },
    { 0x5CED, "play_linked_sound" },
    { 0x5CF8, "play_mlrs_m880_end" },
    { 0x5CFA, "play_nag" },
    { 0x5CFD, "play_new_idle" },
    { 0x5D00, "play_nuke_rumble" },
    { 0x5D02, "play_pain_sound" },
    { 0x5D03, "play_pilot_vo" },
    { 0x5D04, "play_player_break_light" },
    { 0x5D11, "play_rumble_heavy" },
    { 0x5D12, "play_rumble_light" },
    { 0x5D15, "play_rumble_seconds" },
    { 0x5D16, "play_rumbles" },
    { 0x5D18, "play_satellite_static_on_connect" },
    { 0x5D27, "play_spark_fx_when_falling" },
    { 0x5D30, "play_sub_fx_icerise" },
    { 0x5D31, "play_sub_fx_settle" },
    { 0x5D32, "play_synch_attack" },
    { 0x5D36, "play_vo_for_grab_drill" },
    { 0x5D37, "play_vo_for_next_hive" },
    { 0x5D45, "playanimnatratefortime" },
    { 0x5D46, "playanimnatrateuntilnotetrack" },
    { 0x5D48, "playanimnuntilnotetrack" },
    { 0x5D4A, "playanimonscriptable" },
    { 0x5D4C, "playbackrate" },
    { 0x5D64, "player_acceleration_low" },
    { 0x5D67, "player_action_slot_internal" },
    { 0x5D6A, "player_aim_debug" },
    { 0x5D74, "player_animated_sequence_restrictions" },
    { 0x5D77, "player_animtree" },
    { 0x5D84, "player_beach" },
    { 0x5D86, "player_blend_swim_speed" },
    { 0x5D8A, "player_boundaries_off" },
    { 0x5D8C, "player_boundary_ai_focus" },
    { 0x5D90, "player_boundary_vo_player" },
    { 0x5D91, "player_breach" },
    { 0x5D98, "player_can_see_corpse" },
    { 0x5D9A, "player_cancel_internal" },
    { 0x5DA3, "player_chopper_anim_struct" },
    { 0x5DA4, "player_chopper_flyin_setup" },
    { 0x5DA9, "player_combat_rappel" },
    { 0x5DAE, "player_console_anims" },
    { 0x5DAF, "player_const_quake" },
    { 0x5DB1, "player_control" },
    { 0x5DB5, "player_controller" },
    { 0x5DC2, "player_died_recently" },
    { 0x5DC6, "player_dismount" },
    { 0x5DC8, "player_distsqrd" },
    { 0x5DD0, "player_downed_death_buffer_time" },
    { 0x5DD1, "player_drag_body" },
    { 0x5DD8, "player_embers" },
    { 0x5DDB, "player_enter_building" },
    { 0x5DE7, "player_fade_out" },
    { 0x5DE8, "player_fail_finale" },
    { 0x5DEE, "player_failcase_road_overrun" },
    { 0x5DF1, "player_failcase_tunnel_overrun" },
    { 0x5DF5, "player_fall_off_balcony" },
    { 0x5DF7, "player_falling_2" },
    { 0x5DFA, "player_flap_sleeves" },
    { 0x5DFC, "player_flashlight_toggle" },
    { 0x5DFF, "player_force_prone_recover" },
    { 0x5E07, "player_get_mk32" },
    { 0x5E12, "player_getvelocity_pc" },
    { 0x5E18, "player_gravity_slide" },
    { 0x5E1A, "player_grenade_check" },
    { 0x5E1B, "player_grenade_check_dieout" },
    { 0x5E1C, "player_ground_collapse" },
    { 0x5E1D, "player_ground_ref_mover" },
    { 0x5E22, "player_has_enough_currency" },
    { 0x5E24, "player_has_silenced_weapon" },
    { 0x5E2B, "player_heartbeat_rate" },
    { 0x5E2E, "player_heat_fx_end" },
    { 0x5E31, "player_helmet" },
    { 0x5E35, "player_hideviewmodelsleeveflaps" },
    { 0x5E38, "player_hit_ratio_override" },
    { 0x5E51, "player_initiates_inverted_kill" },
    { 0x5E52, "player_input_console_animate" },
    { 0x5E53, "player_inside_nvg_area" },
    { 0x5E56, "player_intro_anim" },
    { 0x5E5B, "player_inverted_kill" },
    { 0x5E5C, "player_inverted_kill_enemy" },
    { 0x5E5F, "player_inverted_kill_fail" },
    { 0x5E63, "player_is_good_missile_target" },
    { 0x5E74, "player_killzone" },
    { 0x5E75, "player_knife_throw_enemy" },
    { 0x5E7B, "player_lerp_speed" },
    { 0x5E7C, "player_lerp_swim_vars" },
    { 0x5E84, "player_littlebird" },
    { 0x5E86, "player_location_check" },
    { 0x5E8D, "player_looking_at_stabguy" },
    { 0x5E8E, "player_looking_at_vargas" },
    { 0x5E93, "player_mantle_wait" },
    { 0x5E95, "player_max" },
    { 0x5E98, "player_min_speed" },
    { 0x5E99, "player_missile_control" },
    { 0x5E9B, "player_missile_roll_control" },
    { 0x5EA9, "player_name_called_recently" },
    { 0x5EB1, "player_on_hill" },
    { 0x5EB3, "player_one_already_breached" },
    { 0x5EB4, "player_outline" },
    { 0x5EB6, "player_pain_vo" },
    { 0x5EBA, "player_persistence_init" },
    { 0x5EBC, "player_pipe_explosion_reaction" },
    { 0x5EBE, "player_plane_engine_right" },
    { 0x5EC4, "player_pounce_anim" },
    { 0x5EC5, "player_pulls_mask_down" },
    { 0x5ECD, "player_pushes_too_far" },
    { 0x5ECF, "player_quake_distance" },
    { 0x5ED3, "player_rambo_rpg" },
    { 0x5ED4, "player_ramp_up_wind" },
    { 0x5ED6, "player_random_blur_cleanup" },
    { 0x5ED7, "player_random_sway" },
    { 0x5ED9, "player_recoil" },
    { 0x5EDC, "player_ref_ent" },
    { 0x5EE3, "player_rides_shotgun_in_humvee" },
    { 0x5EE6, "player_rope_unwind_anim" },
    { 0x5EE8, "player_rotate_plane01" },
    { 0x5EEC, "player_rumble" },
    { 0x5EED, "player_rumble_amb_ent" },
    { 0x5EF3, "player_safe_from_sonar" },
    { 0x5EF4, "player_saw_kill" },
    { 0x5EF5, "player_scuba" },
    { 0x5EFC, "player_seek_disable" },
    { 0x5F00, "player_set_spotted" },
    { 0x5F07, "player_should_see_drill_hint" },
    { 0x5F09, "player_showviewmodelsleeveflaps" },
    { 0x5F15, "player_space_breathe_sound" },
    { 0x5F22, "player_spotted_logic" },
    { 0x5F23, "player_sprint" },
    { 0x5F24, "player_sprintcamerabob" },
    { 0x5F2B, "player_stealth" },
    { 0x5F2F, "player_struct" },
    { 0x5F30, "player_struggle_anim" },
    { 0x5F33, "player_stun_return_weapons_sprint" },
    { 0x5F35, "player_surface_blur_think" },
    { 0x5F37, "player_sway" },
    { 0x5F39, "player_sway_bump" },
    { 0x5F3C, "player_swim_rubberband" },
    { 0x5F3E, "player_swim_think" },
    { 0x5F3F, "player_swim_water_current_logic" },
    { 0x5F41, "player_tank_impeded_sound" },
    { 0x5F43, "player_tank_turret_sounds" },
    { 0x5F49, "player_tired" },
    { 0x5F4B, "player_took_too_long_to_open" },
    { 0x5F4D, "player_touched_arr" },
    { 0x5F4E, "player_touching_post_clip" },
    { 0x5F55, "player_underwater_set" },
    { 0x5F58, "player_unlink_slide_on_death" },
    { 0x5F5D, "player_using_missile" },
    { 0x5F65, "player_view_pitch_down" },
    { 0x5F68, "player_view_shake_blender" },
    { 0x5F6C, "player_viewhands_minigun_hand" },
    { 0x5F70, "player_viewkick_distance" },
    { 0x5F7D, "player_watch_upgrade_internal" },
    { 0x5F7E, "player_watch_use" },
    { 0x5F81, "player_watcher" },
    { 0x5F82, "player_water_breach_moment" },
    { 0x5F84, "player_water_height_think" },
    { 0x5F88, "player_water_wade_speed" },
    { 0x5F92, "playeractivatedairsupport" },
    { 0x5F93, "playeraffectedarray" },
    { 0x5F9F, "playerdamagerumble" },
    { 0x5FA8, "playerdrone_create" },
    { 0x5FAF, "playergrenaderangetime" },
    { 0x5FB4, "playerhealthregeninit" },
    { 0x5FB5, "playerhind_attackspots" },
    { 0x5FB8, "playerhurtcheck" },
    { 0x5FBA, "playerinvul" },
    { 0x5FC2, "playerkilled_internal" },
    { 0x5FC3, "playerkilled_regularmp" },
    { 0x5FC8, "playerlinktodeltablend" },
    { 0x5FCE, "playerpainbreathingsound" },
    { 0x5FD0, "players" },
    { 0x5FD5, "playerseesme" },
    { 0x5FE0, "playertank" },
    { 0x5FE3, "playertank_waits" },
    { 0x5FEF, "playerview_endsequence" },
    { 0x5FF4, "playerview_playmissanim" },
    { 0x6001, "playface_waitfornotify" },
    { 0x600D, "playfx_for_sub_slide" },
    { 0x600F, "playfx_targetname_endon" },
    { 0x6012, "playgrowl" },
    { 0x6014, "playheatfx" },
    { 0x6018, "playidleface" },
    { 0x601A, "playinairjumppainanims" },
    { 0x601C, "playing_effect" },
    { 0x6025, "playlocksound" },
    { 0x6026, "playlookanimation" },
    { 0x602E, "playmusicbeforereachlayer" },
    { 0x6031, "playpainoverlay" },
    { 0x6032, "playpanting" },
    { 0x6035, "playplayerandnpcsounds" },
    { 0x6039, "playreactionevent" },
    { 0x603E, "playsound_and_light" },
    { 0x6042, "playsound_wrapper" },
    { 0x604A, "playtankexhaust" },
    { 0x604C, "playtickingsound" },
    { 0x604F, "playtransitionanimationthread_withoutwaitsetstates" },
    { 0x6051, "playuplinkanimations" },
    { 0x605A, "playvoforbombplant" },
    { 0x605D, "playvofordeathmachine" },
    { 0x605E, "playvofordowned" },
    { 0x6067, "playvoformeteor" },
    { 0x606B, "playvofornukecountdown" },
    { 0x6070, "playvoforrandombox" },
    { 0x607C, "playvoforwavestart" },
    { 0x607E, "plot_points" },
    { 0x6087, "point_cost" },
    { 0x608D, "pointoncircle" },
    { 0x608F, "poison" },
    { 0x6095, "poolindex" },
    { 0x6096, "pop_first_vo_out_of_queue" },
    { 0x609C, "populate_combat_resource_from_table" },
    { 0x609E, "populateintelchallenges" },
    { 0x60A0, "port_to_escape_spitter_location" },
    { 0x60AB, "pos_info" },
    { 0x60B0, "posclosed" },
    { 0x60B4, "positions" },
    { 0x60B8, "post" },
    { 0x60BA, "post_breach_doors" },
    { 0x60BB, "post_breach_positional_ambience" },
    { 0x60BD, "post_crater_breach_dialogue_2" },
    { 0x60CD, "post_quake_vo" },
    { 0x60DE, "power_down" },
    { 0x60E2, "pre_ambush_scene_org" },
    { 0x60E3, "pre_breach_guys" },
    { 0x60EC, "pre_org" },
    { 0x60F4, "pre_tall_grass_patroller_logic" },
    { 0x60F6, "pre_tall_grass_stealth_settings" },
    { 0x60F8, "pre_tall_grass_went_hot_vo" },
    { 0x60FB, "pre_teleport" },
    { 0x60FC, "pre_teleport_to_start" },
    { 0x6103, "precache_create_fx" },
    { 0x6106, "precache_for_startpoints" },
    { 0x6107, "precache_items" },
    { 0x6108, "precache_please" },
    { 0x610C, "precache_scripts" },
    { 0x610D, "precache_stuff" },
    { 0x610E, "precache_teargas" },
    { 0x6114, "precachesetup" },
    { 0x6117, "precisepositioning" },
    { 0x6119, "precommon" },
    { 0x611A, "precommon_bc" },
    { 0x6120, "predictedspawnpoint" },
    { 0x612A, "prefers_drones" },
    { 0x612B, "pregame_delay" },
    { 0x612D, "prekillcamnotify" },
    { 0x612E, "preload" },
    { 0x612F, "preloop_push_anims" },
    { 0x6131, "prematchperiodend" },
    { 0x6134, "prepare_crates_for_anim" },
    { 0x613C, "prepare_to_regenerate" },
    { 0x613D, "prepareattackplayer" },
    { 0x613F, "prereqs" },
    { 0x6147, "prespawn_decomp_enemies" },
    { 0x614A, "pressurized" },
    { 0x614F, "prestige_getweapondamagescalar" },
    { 0x6151, "prestigedoubleweaponxp" },
    { 0x615D, "prevattack" },
    { 0x616F, "prevleanfracpitch" },
    { 0x6179, "prevstairsstate" },
    { 0x6180, "price_breach_ent_movesto_player" },
    { 0x6182, "price_desired_speed" },
    { 0x6183, "price_match_player_speed" },
    { 0x6184, "priceliner" },
    { 0x6185, "primary_attacker" },
    { 0x6187, "primary_weapon_array" },
    { 0x6189, "primaryprogressbarfontsize" },
    { 0x618C, "primaryprogressbartexty" },
    { 0x61A4, "printboldonteam" },
    { 0x61A9, "printer" },
    { 0x61B1, "printturnrate" },
    { 0x61B5, "prioritize_watch_nodes_toward_enemies" },
    { 0x61BB, "process" },
    { 0x61BD, "process_blend" },
    { 0x61C3, "process_deathflags" },
    { 0x61C7, "process_moving_platform_death" },
    { 0x61CC, "processassist" },
    { 0x61DD, "progress_trigger_callbacks" },
    { 0x61DF, "project" },
    { 0x61E1, "projectileexplode" },
    { 0x61EA, "promo_dof" },
    { 0x61EC, "promote_nearest_friendly_with_classname" },
    { 0x61EE, "promotion_dog" },
    { 0x61F6, "prone_kill_trigger" },
    { 0x61F7, "prone_only" },
    { 0x61FF, "pronelegsstraighttree" },
    { 0x6207, "pronetocrouchwalk" },
    { 0x6208, "pronetopronemove" },
    { 0x6216, "prop_launch" },
    { 0x621B, "props_cleanup" },
    { 0x6221, "proxbartext" },
    { 0x6223, "proximity_spawn" },
    { 0x6228, "proximitykill" },
    { 0x622B, "proxmity_check_stop_loop" },
    { 0x622C, "proxmity_check_stop_relative" },
    { 0x6235, "pt2_force_look_down" },
    { 0x6243, "push_anim" },
    { 0x6246, "push_out_of_doorway" },
    { 0x6264, "quadrantanimweights" },
    { 0x6269, "quake_anims_ref" },
    { 0x626F, "quake_event_disconnect_node" },
    { 0x6273, "quake_event_gas_leak" },
    { 0x6274, "quake_event_hurt" },
    { 0x6275, "quake_event_init" },
    { 0x6279, "quake_event_pole_fall_on_car" },
    { 0x6282, "quake_event_wait" },
    { 0x6285, "quake_hurt_trigger" },
    { 0x6298, "queued_anim_threads" },
    { 0x62AA, "radarviewtime" },
    { 0x62AF, "radial_button_group" },
    { 0x62B1, "radial_button_previous_group" },
    { 0x62B7, "radiationdeath" },
    { 0x62B8, "radiationeffect" },
    { 0x62BA, "radiationsound" },
    { 0x62C5, "radio_dialogue_queue_single" },
    { 0x62C9, "radio_in_use" },
    { 0x62CA, "radio_queue_thread" },
    { 0x62CB, "radio_tower_guy_shot" },
    { 0x62CC, "radioforcedtransmissionqueue" },
    { 0x62D1, "radiousityscale" },
    { 0x62D5, "radius_min_sq" },
    { 0x62D7, "radiusartilleryshellshock" },
    { 0x62D8, "ragdoll_and_delete" },
    { 0x62DD, "ragdoll_immediate" },
    { 0x62E3, "railyard_style" },
    { 0x62E4, "rain" },
    { 0x62E7, "rain_overlay_alpha" },
    { 0x62E9, "raineffectchange" },
    { 0x62F6, "raise_rear_elevator_intro" },
    { 0x62F9, "ramboaccuracymult" },
    { 0x62FF, "ramp_down_push" },
    { 0x6301, "ramp_up_accurracy" },
    { 0x6302, "rampfx" },
    { 0x6303, "ramping_explosions" },
    { 0x631B, "random_mortars_get_target" },
    { 0x631C, "random_mortars_incoming_sound" },
    { 0x6320, "random_pick_r_or_l" },
    { 0x6321, "random_player_wall_push" },
    { 0x6322, "random_player_wall_pushdownup" },
    { 0x632A, "random_weight" },
    { 0x632B, "random_weight_sorted" },
    { 0x632C, "randomaditionaltime" },
    { 0x6330, "randomdeath" },
    { 0x6336, "randominttable" },
    { 0x6338, "randomitem" },
    { 0x633B, "randomizer_create" },
    { 0x633E, "randomly_rotate_and_fire" },
    { 0x6341, "randompathstoptime" },
    { 0x6343, "randomvector" },
    { 0x6344, "randomvectorrange" },
    { 0x6354, "rappel_clear_vertical_limits" },
    { 0x6356, "rappel_combat_two_volume_upstairs" },
    { 0x6357, "rappel_death" },
    { 0x6359, "rappel_disable_fidgit" },
    { 0x635B, "rappel_enemy" },
    { 0x635D, "rappel_enter_death" },
    { 0x635E, "rappel_entry_anim_struct" },
    { 0x6364, "rappel_guy" },
    { 0x6366, "rappel_idle" },
    { 0x6374, "rappel_max_lateral_speed" },
    { 0x6378, "rappel_params" },
    { 0x6380, "rappel_rope_rig" },
    { 0x6381, "rappel_rotate_jump_anim" },
    { 0x6385, "rappel_stealth_checkpoint" },
    { 0x638A, "rappel_stealth_second_floor_combat_vo" },
    { 0x638C, "rappel_type" },
    { 0x638D, "rappel_type_aim" },
    { 0x638E, "rappel_upper_limit" },
    { 0x6397, "raven_player_can_see_ai" },
    { 0x6398, "ravine_jeeps_to_the_right" },
    { 0x639C, "reach_death_notify" },
    { 0x63B2, "react_to_attractor_flare" },
    { 0x63BB, "reaction_sleep_wait_wakeup_dist" },
    { 0x63C2, "reactive_fx_ents" },
    { 0x63C4, "reactive_grass_settings" },
    { 0x63C5, "reactive_grass_settings_pc" },
    { 0x63C6, "reactive_sound_ents" },
    { 0x63CA, "reacttobulletsinterruptcheck" },
    { 0x63CF, "readtriple" },
    { 0x63DC, "real_reload" },
    { 0x63DE, "realorigin" },
    { 0x63E3, "rear_left_anchor_impact" },
    { 0x63E4, "rear_right_anchor_impact" },
    { 0x63E9, "reassign_dir_group" },
    { 0x63EC, "reassign_interrogate" },
    { 0x63EF, "reassign_typer" },
    { 0x63F0, "reassign_vault_guys" },
    { 0x63F7, "rec_blink" },
    { 0x63FA, "recentdamageamount" },
    { 0x63FB, "recentdamages" },
    { 0x63FC, "recentkillcount" },
    { 0x6402, "recipe_getkillstreak" },
    { 0x6403, "recipeclassapplyjuggernaut" },
    { 0x640C, "record_last_player_damage" },
    { 0x640E, "recordfinalkillcam" },
    { 0x6410, "recover_from_careful_disable" },
    { 0x6411, "recover_interval" },
    { 0x6412, "recruit_player_visual" },
    { 0x6416, "red_crosshair" },
    { 0x6418, "redshirt_cq_enc_handles" },
    { 0x641A, "redshirts" },
    { 0x6425, "ref_node" },
    { 0x642A, "refillclip" },
    { 0x642B, "refillexplosiveweapons" },
    { 0x643B, "refresh_reactive_fx_ents" },
    { 0x643C, "refreshspectatorportalfx" },
    { 0x6442, "regenamount" },
    { 0x6444, "regenfastermod" },
    { 0x6448, "regenspeedwatcher" },
    { 0x6449, "register_achievement" },
    { 0x644A, "register_airdrop_sub_items" },
    { 0x644E, "register_default_achievements" },
    { 0x6451, "register_end_game_string_index" },
    { 0x6455, "register_laststand_ammo" },
    { 0x6465, "registeradrenalineinfo" },
    { 0x6467, "registerdamage" },
    { 0x6474, "registerscoreinfo" },
    { 0x647D, "reincrement_count_if_deleted" },
    { 0x6486, "release_use_trigger" },
    { 0x6488, "release_view_count" },
    { 0x648A, "reload_aim_active" },
    { 0x6490, "relocatespeed" },
    { 0x6491, "remaining_kill" },
    { 0x6492, "remote_canreload" },
    { 0x649D, "remote_turret_clear_hud" },
    { 0x64A9, "remote_turret_handle_zoom" },
    { 0x64B7, "remote_turret_monitor_dryfire" },
    { 0x64B8, "remote_turret_next" },
    { 0x64C6, "remote_turret_type" },
    { 0x64D7, "remotefiring" },
    { 0x64DF, "remoteride" },
    { 0x64EE, "remoteuav_explode_on_disconnect" },
    { 0x64F7, "remoteuav_leave" },
    { 0x64FA, "remoteuav_markplayer" },
    { 0x6501, "remoteuav_rumble" },
    { 0x6502, "remoteuav_staticfade" },
    { 0x650C, "remove_all_challenge_cases" },
    { 0x6510, "remove_allies" },
    { 0x6519, "remove_damagefeedback" },
    { 0x6529, "remove_from_bot_damage_targets" },
    { 0x6531, "remove_global_spawn_function" },
    { 0x6543, "remove_nocolor_from_array" },
    { 0x654E, "remove_remote_turret_targets" },
    { 0x6559, "remove_specialized_ammo" },
    { 0x6567, "remove_without_classname" },
    { 0x6572, "removeac130playerongameend" },
    { 0x6578, "removealtituedmesh" },
    { 0x657B, "removebodyarmorondeath" },
    { 0x6581, "removechild" },
    { 0x658B, "removefrombattlebuddywaitlist" },
    { 0x658D, "removefromhelilist" },
    { 0x658F, "removefromimslist" },
    { 0x6595, "removefromtanklist" },
    { 0x659F, "removelockedontarget" },
    { 0x65A8, "removepickup" },
    { 0x65A9, "removeplanefromlist" },
    { 0x65AD, "removesafetyhealth" },
    { 0x65B6, "removetypefromqueue" },
    { 0x65B8, "removeuavmodelondeath" },
    { 0x65C1, "replace_attacker_request" },
    { 0x65C5, "replenishloadout" },
    { 0x65D6, "rescue_think" },
    { 0x65DB, "reserve_turret" },
    { 0x65DC, "reserved" },
    { 0x65E9, "reset_kill_10_with_propane_progress" },
    { 0x65EA, "reset_kill_10_with_traps_progress" },
    { 0x65ED, "reset_laststand" },
    { 0x65EE, "reset_melee_goons_progress" },
    { 0x65F2, "reset_nuke_usage" },
    { 0x65F4, "reset_player_speed" },
    { 0x65FB, "reset_team_hive_performance" },
    { 0x6603, "resetaccuracyandpause" },
    { 0x6607, "resetbrinkofdeathkillstreakshortly" },
    { 0x6612, "resetnow" },
    { 0x6613, "resetoncancel" },
    { 0x6615, "resetplayervariables" },
    { 0x6618, "resetskill" },
    { 0x6619, "resetsniperaim" },
    { 0x661A, "resetstate" },
    { 0x6620, "resetuidvarsondeath" },
    { 0x6628, "respawn_asspectator" },
    { 0x662A, "respawn_friendlies_force_vision_check" },
    { 0x662C, "respawn_max" },
    { 0x662E, "respawn_spawner_org" },
    { 0x6634, "respawn_with_launcher" },
    { 0x6635, "respawneliminatedplayers" },
    { 0x6636, "respawntimerstarttime" },
    { 0x6639, "respondto" },
    { 0x6640, "responsethreatexposed" },
    { 0x6646, "restockammoaura" },
    { 0x6651, "restore_last_weapon" },
    { 0x6663, "restorevision_no_blur" },
    { 0x666D, "retarget_rig" },
    { 0x6677, "retreat_watcher" },
    { 0x667C, "return_collision_model" },
    { 0x667E, "return_false" },
    { 0x6680, "return_on_movement" },
    { 0x6694, "revive_player" },
    { 0x669A, "revivetriggerthink" },
    { 0x669F, "ride_dog_bark" },
    { 0x66A7, "rider_func" },
    { 0x66B0, "rifleshootobjectivenormal" },
    { 0x66B1, "rifleshootobjectivesuppress" },
    { 0x66B4, "rig_fx" },
    { 0x66B5, "rig_model" },
    { 0x66B8, "right_direction_worldspace" },
    { 0x66C3, "rigmodel_anims" },
    { 0x66C4, "rigmodel_pauseend" },
    { 0x66C6, "riing_fx" },
    { 0x66CB, "ring_fx" },
    { 0x66D0, "riotshield_attach" },
    { 0x66D2, "riotshield_detach" },
    { 0x66D3, "riotshield_getmodel" },
    { 0x6702, "rog_add_hud_element_on_target" },
    { 0x6703, "rog_adjust_aoe_target_height" },
    { 0x6705, "rog_airstrike" },
    { 0x670B, "rog_altimeter_game_world_delta" },
    { 0x6711, "rog_ammo_animate_depleted_round" },
    { 0x6716, "rog_aoe_reticle_visibility_logic" },
    { 0x6723, "rog_cleanup_altimeter_hud" },
    { 0x6725, "rog_cleanup_static_hud" },
    { 0x672A, "rog_display_hint_strings" },
    { 0x6736, "rog_firing_fx_at_player" },
    { 0x6744, "rog_hud_logic_reset" },
    { 0x674B, "rog_kill_jet" },
    { 0x6755, "rog_play_countdown_beeps" },
    { 0x6762, "rog_setup_temp_hud" },
    { 0x6764, "rog_shockwave" },
    { 0x6765, "rog_show_target_marker" },
    { 0x6767, "rog_single_velocity_max" },
    { 0x676D, "rog_stop_countdown_beebs" },
    { 0x6773, "rog_target_add_hud_element" },
    { 0x6774, "rog_target_cleaned_up_by_allies" },
    { 0x6776, "rog_target_create_model_logic" },
    { 0x677C, "rog_target_handle_rod_impact" },
    { 0x6784, "rog_target_overkill_death" },
    { 0x678A, "rog_targeting_pip" },
    { 0x678D, "rog_temp_ui" },
    { 0x678F, "rog_uav_camera_logic" },
    { 0x67A1, "rog_zoom_ui_pip_0" },
    { 0x67AB, "roll_speed_max" },
    { 0x67AE, "ronnie_knife_watcher" },
    { 0x67AF, "ronnie_talks" },
    { 0x67B9, "roofcollapse_retreat" },
    { 0x67BF, "rooftop_collapse_end_build_sfx" },
    { 0x67C4, "rooftop_heli" },
    { 0x67E7, "rooftops_encounter_b_ally_vo" },
    { 0x67EA, "rooftops_encounter_b_flank_vo" },
    { 0x67EC, "rooftops_encounter_b_handle_defensive" },
    { 0x67F0, "rooftops_encounter_b_water_vo" },
    { 0x67F1, "rooftops_encounters_ally_logic" },
    { 0x67FE, "rooftops_heli_flyaway" },
    { 0x6802, "rooftops_interior_start_combat_soft_sight_line" },
    { 0x6804, "rooftops_long_jump" },
    { 0x680B, "rooftops_player_start_combat_attack" },
    { 0x681B, "rooftops_water_enter_combat_space_play_effects" },
    { 0x6820, "rooftops_water_heli_movement_logic" },
    { 0x6821, "rooftops_water_intro" },
    { 0x6824, "rooftops_water_intro_flare_setup" },
    { 0x682A, "rooftops_water_player_logic" },
    { 0x6832, "room_has_multiple_doors" },
    { 0x6833, "room_volume" },
    { 0x6834, "root_anim" },
    { 0x6836, "rootsofquadratic" },
    { 0x6839, "rope_cut_death" },
    { 0x683E, "rope_origin" },
    { 0x6840, "rope_prop_anim" },
    { 0x6844, "rope_unwind_anim" },
    { 0x684D, "rorke_glass_cutter" },
    { 0x6852, "rorke_inverted_kill_enemy" },
    { 0x6853, "rorke_inverted_kill_knife" },
    { 0x6854, "rorke_inverted_kill_knife_putaway" },
    { 0x685B, "rorke_react_to_stealth_break" },
    { 0x685E, "rorke_start_shadowkill" },
    { 0x686A, "rotate_helicopter_rotor" },
    { 0x686F, "rotate_rollers_to" },
    { 0x6872, "rotate_turbine" },
    { 0x6873, "rotate_vector" },
    { 0x687E, "rotating_turbines" },
    { 0x6882, "rotation_is_occuring" },
    { 0x6888, "rotunda_kill_gun_sync" },
    { 0x6889, "rotunda_knife" },
    { 0x688E, "round_num" },
    { 0x6892, "roundbegin" },
    { 0x6895, "roundenddelay" },
    { 0x6897, "roundendwait" },
    { 0x689D, "roundup" },
    { 0x689E, "roundwinnerdialog" },
    { 0x68A0, "rpg_ai_attack" },
    { 0x68AC, "rpg_guy_wait_and_fire_at_target" },
    { 0x68AD, "rpg_kibble" },
    { 0x68BB, "rpl_calc_max_yaw_right" },
    { 0x68C2, "rpl_get_max_downward_speed" },
    { 0x68C3, "rpl_get_max_lateral_speed" },
    { 0x68C9, "rpl_get_walk_loop_anim" },
    { 0x68CB, "rpl_get_walk_stop_anim" },
    { 0x68D4, "rpl_legs_get_blend_time" },
    { 0x68D7, "rpl_legs_get_idle_anim" },
    { 0x68D8, "rpl_legs_get_move_directions" },
    { 0x68DC, "rpl_legs_get_stop_move_direction" },
    { 0x68E7, "rpl_legs_should_use_run_loop" },
    { 0x68F3, "rt_fade_in" },
    { 0x68F7, "rt_helo_break_glass" },
    { 0x68F9, "rt_helo_cleanup" },
    { 0x68FC, "rt_helo_crash_train" },
    { 0x68FE, "rt_helo_dam" },
    { 0x6901, "rt_helo_flyout" },
    { 0x6902, "rt_helo_fx_setup" },
    { 0x690F, "rt_helos_kill_ally" },
    { 0x6910, "rt_hero_train_impact" },
    { 0x6914, "rt_run_fic" },
    { 0x6915, "rt_spawn_and_link_helo_glass" },
    { 0x691F, "rubble_flare" },
    { 0x6925, "rumble_ent" },
    { 0x692A, "rumble_ramp_on" },
    { 0x6930, "rumbletrigger" },
    { 0x6936, "run_anim_events" },
    { 0x693B, "run_c17_jet_flyers" },
    { 0x693C, "run_call_after_wait_array" },
    { 0x6946, "run_defend_zodiac" },
    { 0x694D, "run_enemy_destroyer_gun" },
    { 0x695E, "run_nag_vo" },
    { 0x695F, "run_near_enemy" },
    { 0x6960, "run_noself_call_after_wait_array" },
    { 0x696A, "run_overrideanim" },
    { 0x696B, "run_overridebulletreact" },
    { 0x696D, "run_player_out_of_bounds" },
    { 0x6976, "run_spawn_functions" },
    { 0x6977, "run_speed" },
    { 0x6978, "run_speed_state" },
    { 0x697E, "run_to_new_spot_and_setup_gun" },
    { 0x6994, "runnerguys" },
    { 0x6996, "runngun_backward" },
    { 0x69A4, "runportalstatus" },
    { 0x69AB, "runteamintel" },
    { 0x69AC, "runtovehicleoverride" },
    { 0x69AD, "runway_apache_logic" },
    { 0x69C1, "rushtime" },
    { 0x69C7, "saf_large_sign_01_dynamic_func" },
    { 0x69C9, "saf_streetlight_dynamic_func" },
    { 0x69D8, "safe_makerealai" },
    { 0x69E6, "safezone_sound_on" },
    { 0x69F2, "sam_exit" },
    { 0x69F4, "sam_fire_missiles" },
    { 0x6A00, "sam_missile_lockon" },
    { 0x6A04, "sam_reload_missile" },
    { 0x6A06, "sam_start_missile_lockon" },
    { 0x6A08, "sam_update_compass" },
    { 0x6A0D, "sam_use_auto_lock_on" },
    { 0x6A10, "sam_watchleaving" },
    { 0x6A12, "samdamagescale" },
    { 0x6A18, "samproximitydetonate" },
    { 0x6A19, "samtargetent" },
    { 0x6A1A, "samturret" },
    { 0x6A20, "sardines" },
    { 0x6A23, "sardines_balllinkpiece" },
    { 0x6A24, "sardines_ballpanic" },
    { 0x6A28, "sardines_line" },
    { 0x6A36, "sardines_spreadring" },
    { 0x6A39, "sat_base_s" },
    { 0x6A3A, "sat_combat" },
    { 0x6A3C, "sat_crane_arm" },
    { 0x6A4A, "satellite_cleanup" },
    { 0x6A55, "satellite_group_hide" },
    { 0x6A5E, "satellite_script_mover" },
    { 0x6A67, "satellite_view_blink_corners" },
    { 0x6A79, "satellite_view_pip_close_l_corner" },
    { 0x6A7B, "satellite_view_pip_display_name" },
    { 0x6A7F, "satellite_view_pip_set_entity" },
    { 0x6A81, "satellite_view_remove_l_corners" },
    { 0x6A84, "satellite_view_type_anchored_text" },
    { 0x6A86, "satellite_view_type_multiline_text_at_point" },
    { 0x6A8C, "satellite_view_zoom_in_sound" },
    { 0x6A92, "satfarm_global_flags" },
    { 0x6AA0, "save_turret_sharing_info" },
    { 0x6AAB, "saveplayerweaponstatepersistent" },
    { 0x6AAE, "saw_mgturretlink" },
    { 0x6AB0, "say_nag_after_delay" },
    { 0x6AB4, "saylocalsounddelayed" },
    { 0x6AB9, "scale" },
    { 0x6ABB, "scan_blur" },
    { 0x6AC1, "scared_behavior" },
    { 0x6AC2, "scared_danger_dialogue" },
    { 0x6AC8, "scene1emitter" },
    { 0x6ACA, "scene_loop" },
    { 0x6AEB, "scr_anim" },
    { 0x6AF4, "scr_notetrack" },
    { 0x6AF8, "scr_text" },
    { 0x6AFD, "scramble_watchlineofsight" },
    { 0x6B01, "scrambletarget" },
    { 0x6B09, "screen_effect_on_open_bottom" },
    { 0x6B0D, "screen_glitch_org" },
    { 0x6B11, "screen_shake_vehicles" },
    { 0x6B25, "script_airstrike_exp02" },
    { 0x6B28, "script_airstrike_exp05" },
    { 0x6B2B, "script_allowdeath" },
    { 0x6B2D, "script_ammo_alt_extra" },
    { 0x6B44, "script_audio_update_rate" },
    { 0x6B45, "script_audio_zones" },
    { 0x6B68, "script_color_axis" },
    { 0x6B6D, "script_count" },
    { 0x6B6E, "script_count_max" },
    { 0x6B71, "script_crashtypeoverride" },
    { 0x6B77, "script_deathchain" },
    { 0x6B79, "script_deathflag_longdeath" },
    { 0x6B7A, "script_deathroll" },
    { 0x6B7D, "script_decel_fraction" },
    { 0x6B82, "script_delayed_playerseek" },
    { 0x6B84, "script_delete_vehicles" },
    { 0x6B86, "script_destruct_collision" },
    { 0x6B8E, "script_dof_far_start" },
    { 0x6BA2, "script_engage" },
    { 0x6BB3, "script_favoriteenemy" },
    { 0x6BB5, "script_firefx" },
    { 0x6BB9, "script_firelink" },
    { 0x6BBC, "script_fixednode" },
    { 0x6BBD, "script_flag" },
    { 0x6BBE, "script_flag_clear" },
    { 0x6BC2, "script_flag_set" },
    { 0x6BCB, "script_followmax" },
    { 0x6BD3, "script_forcegrenade" },
    { 0x6BE2, "script_gametype_koth" },
    { 0x6BE6, "script_ghettotag" },
    { 0x6BE7, "script_goal_radius" },
    { 0x6BEC, "script_goalvolume" },
    { 0x6BF2, "script_grenadespeed" },
    { 0x6BF5, "script_health" },
    { 0x6BF7, "script_hidden" },
    { 0x6BFA, "script_idleanim" },
    { 0x6BFC, "script_ignore_suppression" },
    { 0x6C0E, "script_light" },
    { 0x6C11, "script_linkto" },
    { 0x6C13, "script_longdeath" },
    { 0x6C17, "script_mapsize_16" },
    { 0x6C18, "script_mapsize_32" },
    { 0x6C24, "script_minspec_level" },
    { 0x6C30, "script_mortargroup_domortar" },
    { 0x6C32, "script_mortargroup_mortarzone" },
    { 0x6C35, "script_move_vehicles" },
    { 0x6C38, "script_mover" },
    { 0x6C39, "script_mover_add_hintstring" },
    { 0x6C3A, "script_mover_add_parameters" },
    { 0x6C40, "script_mover_connect_watch" },
    { 0x6C42, "script_mover_func_on_notify" },
    { 0x6C46, "script_mover_is_script_mover" },
    { 0x6C47, "script_mover_move_to_named_goal" },
    { 0x6C48, "script_mover_move_to_target" },
    { 0x6C4A, "script_mover_parameters" },
    { 0x6C4C, "script_mover_parse_range" },
    { 0x6C4F, "script_mover_save_default_move_parameters" },
    { 0x6C56, "script_multiplier" },
    { 0x6C57, "script_namenumber" },
    { 0x6C5A, "script_node_pausetime" },
    { 0x6C5E, "script_nofriendlywave" },
    { 0x6C62, "script_nosurprise" },
    { 0x6C6B, "script_objective_inactive" },
    { 0x6C6F, "script_oneway" },
    { 0x6C74, "script_painter_treeorient" },
    { 0x6C7C, "script_pet" },
    { 0x6C7D, "script_physics" },
    { 0x6C8E, "script_retreating_allies04_firing" },
    { 0x6C92, "script_rumble" },
    { 0x6C93, "script_savedata" },
    { 0x6C98, "script_shooting_tanks_01" },
    { 0x6C9F, "script_shooting_tanks_08" },
    { 0x6CB1, "script_spawnsubgroup" },
    { 0x6CB2, "script_specialops" },
    { 0x6CC1, "script_stay_drone" },
    { 0x6CD9, "script_transient" },
    { 0x6CDD, "script_triggered_playerseek" },
    { 0x6CE5, "script_turretmg" },
    { 0x6CEF, "script_vehicle_selfremove" },
    { 0x6CFD, "script_vehicletriggergroup" },
    { 0x6D0A, "scriptable_vo_handler" },
    { 0x6D0C, "scriptables" },
    { 0x6D11, "scripted_dialogue" },
    { 0x6D1E, "scubamask_distortion" },
    { 0x6D24, "sd_loadout" },
    { 0x6D25, "sd_loadouts" },
    { 0x6D27, "sd_prisonerobjective" },
    { 0x6D28, "sd_triggers" },
    { 0x6D29, "sdbomb" },
    { 0x6D2C, "sdv_follow_spotted_react" },
    { 0x6D35, "second_bumper_hint" },
    { 0x6D3D, "second_floor_elevator_guy_setup" },
    { 0x6D40, "second_floor_fridge_guy_setup" },
    { 0x6D43, "second_floor_kitchenette_guy_setup" },
    { 0x6D46, "second_floor_poker_table_guys_setup" },
    { 0x6D4B, "second_patroller" },
    { 0x6D4F, "secondarytarget" },
    { 0x6D52, "secondload" },
    { 0x6D57, "section_flag_inits" },
    { 0x6D5E, "security_beeps" },
    { 0x6D63, "seekoutenemytime" },
    { 0x6D64, "seen_attacker" },
    { 0x6D65, "segments" },
    { 0x6D6B, "select_by_substring" },
    { 0x6D6E, "select_last_entity" },
    { 0x6D75, "selected_fx_ents" },
    { 0x6D7E, "selectentrancelocation" },
    { 0x6D7F, "selectidleanimstate" },
    { 0x6D85, "self_cleanup" },
    { 0x6D87, "self_func" },
    { 0x6D95, "send_friends_in" },
    { 0x6D9A, "send_to_node_and_set_flag_if_specified_when_reached" },
    { 0x6D9B, "send_to_volume_and_delete" },
    { 0x6DA0, "sensitive_chopper_in_players_view" },
    { 0x6DA1, "sensor" },
    { 0x6DA2, "sensorscreeneffects" },
    { 0x6DA3, "sentient" },
    { 0x6DAC, "sentry_badplace_delete" },
    { 0x6DAE, "sentry_beep_sounds" },
    { 0x6DB0, "sentry_burst_fire_start" },
    { 0x6DB3, "sentry_burstfirestop" },
    { 0x6DBE, "sentry_enemy_wait" },
    { 0x6DC2, "sentry_gun_weak_settings" },
    { 0x6DCB, "sentry_initsentry" },
    { 0x6DDC, "sentry_place_delay" },
    { 0x6DE9, "sentry_setactive" },
    { 0x6DF3, "sentry_smg_default_settings" },
    { 0x6DF7, "sentry_synch_attack_begin" },
    { 0x6DFD, "sentry_team_show_icon" },
    { 0x6E06, "sentrymode" },
    { 0x6E09, "sentrymodifydamage" },
    { 0x6E16, "set3duseicon" },
    { 0x6E20, "set_ai_bcvoice" },
    { 0x6E29, "set_all_ai_targetnames" },
    { 0x6E2A, "set_all_exceptions" },
    { 0x6E2C, "set_allowdeath" },
    { 0x6E37, "set_animarray_burst_and_semi_fire_crouch" },
    { 0x6E39, "set_animarray_crouching" },
    { 0x6E3A, "set_animarray_crouching_left" },
    { 0x6E4B, "set_audio_mix" },
    { 0x6E4F, "set_bag_visibility" },
    { 0x6E5E, "set_breaching_variable" },
    { 0x6E64, "set_console_status" },
    { 0x6E70, "set_custom_gameskill_func" },
    { 0x6E73, "set_cycle_scalars" },
    { 0x6E7F, "set_default_hud_stuff" },
    { 0x6E87, "set_diff_accuracy" },
    { 0x6E8F, "set_dog_hud_disabled" },
    { 0x6E91, "set_dog_hud_for_guard" },
    { 0x6E9E, "set_dpad_right_level" },
    { 0x6E9F, "set_dpad_up_level" },
    { 0x6EA4, "set_drill_state_run" },
    { 0x6EAB, "set_enter_garage_mall_vf" },
    { 0x6EB7, "set_firing" },
    { 0x6EC0, "set_flag_on_dead" },
    { 0x6EC1, "set_flag_on_dead_or_dying" },
    { 0x6EC5, "set_flag_on_spawned" },
    { 0x6EDA, "set_flavorbursts_team_state" },
    { 0x6EF2, "set_glass_zero_gravity" },
    { 0x6EF7, "set_goal_from_settings" },
    { 0x6EFE, "set_goal_volume" },
    { 0x6F01, "set_grenadeammo" },
    { 0x6F04, "set_high_priority_target_for_bot" },
    { 0x6F07, "set_hud_name_percent_value" },
    { 0x6F09, "set_hud_value" },
    { 0x6F0A, "set_hud_value_internal" },
    { 0x6F0B, "set_hudoutline" },
    { 0x6F12, "set_ignoreall" },
    { 0x6F13, "set_ignoreme" },
    { 0x6F14, "set_ignoresuppression" },
    { 0x6F15, "set_initial_emissive" },
    { 0x6F18, "set_key" },
    { 0x6F22, "set_missile_reaction" },
    { 0x6F24, "set_mission_failed_override" },
    { 0x6F27, "set_mk23_model" },
    { 0x6F28, "set_mortar_on" },
    { 0x6F29, "set_motion_blur" },
    { 0x6F2A, "set_motionblur" },
    { 0x6F30, "set_nerf_scalar" },
    { 0x6F31, "set_new_enemy_volume" },
    { 0x6F39, "set_off_exploders" },
    { 0x6F46, "set_perk_bullet_damage_1" },
    { 0x6F49, "set_perk_bullet_damage_4" },
    { 0x6F4B, "set_perk_health_level_1" },
    { 0x6F56, "set_perk_pistol_m9a1_2" },
    { 0x6F59, "set_perk_pistol_magnum_0" },
    { 0x6F60, "set_perk_pistol_mp443_2" },
    { 0x6F63, "set_perk_pistol_p226_0" },
    { 0x6F69, "set_perk_rigger_1" },
    { 0x6F6B, "set_perk_rigger_3" },
    { 0x6F6C, "set_perk_rigger_4" },
    { 0x6F6F, "set_player_character_model" },
    { 0x6F72, "set_player_hearbeat_rate" },
    { 0x6F79, "set_player_session_rankup" },
    { 0x6F87, "set_room_to_breached" },
    { 0x6F91, "set_silenthawk_override_anims" },
    { 0x6F98, "set_spawners_for_location" },
    { 0x6F99, "set_stage" },
    { 0x6F9C, "set_standing_turns" },
    { 0x6F9D, "set_start_pos" },
    { 0x6F9E, "set_start_positions" },
    { 0x6FA0, "set_stop_ignore_flag" },
    { 0x6FA5, "set_talker_until_msg" },
    { 0x6FA6, "set_team_bcvoice" },
    { 0x6FB5, "set_up_vignette_enemies" },
    { 0x6FB6, "set_up_when_patroller_goes" },
    { 0x6FBF, "set_vision_set" },
    { 0x6FC2, "set_visionset_for_watching_players" },
    { 0x6FC3, "set_vo_currently_playing" },
    { 0x6FC4, "set_vo_system_playing" },
    { 0x6FC7, "set_warelights_off" },
    { 0x6FCF, "set_z" },
    { 0x6FD3, "setac130" },
    { 0x6FD5, "setactivegrenadetimer" },
    { 0x6FD8, "setairdropcratecollision" },
    { 0x6FDA, "setalienloadout" },
    { 0x6FDD, "setanimaimweight" },
    { 0x6FEE, "setbloodrushinternal" },
    { 0x6FF8, "setcarriervisible" },
    { 0x6FF9, "setcarryicon" },
    { 0x6FFD, "setcarryingtank" },
    { 0x6FFE, "setcarryingturret" },
    { 0x7009, "setcomexp" },
    { 0x7018, "setdefaultmodel" },
    { 0x701A, "setdelaymine" },
    { 0x701C, "setdirection" },
    { 0x7033, "setfirstinfected" },
    { 0x7036, "setflashbangimmunity" },
    { 0x703F, "setgamblerinternal" },
    { 0x7041, "setglobalaimsettings" },
    { 0x7046, "setgunsmithinternal" },
    { 0x7048, "sethardshell" },
    { 0x704A, "setheadicon" },
    { 0x7050, "seticonshader" },
    { 0x7059, "setjugg" },
    { 0x7064, "setlightweight" },
    { 0x7067, "setlowermessage" },
    { 0x7068, "setmapcenterfordev" },
    { 0x706B, "setmatchrecordiflower" },
    { 0x706D, "setmeleeattackdist" },
    { 0x7076, "setmovenonforwardanims" },
    { 0x707E, "setnotetrackeffect" },
    { 0x7081, "setobjectivehinttext" },
    { 0x709C, "setplayermodels" },
    { 0x70A6, "setpose" },
    { 0x70A8, "setposemovementfnarray" },
    { 0x70BE, "setscrambled" },
    { 0x70C4, "setshootstyleforsuppression" },
    { 0x70C8, "setsize" },
    { 0x70CC, "setspawnpoint" },
    { 0x70D2, "setstandardspeed" },
    { 0x70D6, "setstreakcounttonext" },
    { 0x70D7, "setstrikermodel" },
    { 0x70F3, "setup_ai" },
    { 0x70F5, "setup_ai_for_end" },
    { 0x7101, "setup_atrium_check" },
    { 0x7105, "setup_blocker_hives" },
    { 0x7107, "setup_bodydrag_startpoint" },
    { 0x710B, "setup_bullet_pinholes" },
    { 0x710D, "setup_camera_roll" },
    { 0x710E, "setup_capture" },
    { 0x711A, "setup_combat" },
    { 0x711B, "setup_common" },
    { 0x7126, "setup_damageable_exploder_rooftanks" },
    { 0x7130, "setup_defend_blowdoors2" },
    { 0x7135, "setup_defense_bunkers" },
    { 0x713A, "setup_dog" },
    { 0x7143, "setup_edge_lean" },
    { 0x7144, "setup_electric_fence" },
    { 0x7146, "setup_enemies_open_gate" },
    { 0x7148, "setup_exfil" },
    { 0x7153, "setup_flags" },
    { 0x7154, "setup_flood_water_anims" },
    { 0x7158, "setup_front_elevator" },
    { 0x7159, "setup_fuel_leak_lighting" },
    { 0x715E, "setup_garden_entry" },
    { 0x7160, "setup_gate_ai" },
    { 0x7163, "setup_hangar_truck_pre_loaded" },
    { 0x7164, "setup_hangar_truck_to_load" },
    { 0x7165, "setup_hill_enemies" },
    { 0x7167, "setup_house" },
    { 0x716C, "setup_individual_exploder" },
    { 0x7176, "setup_inverted_rappel" },
    { 0x7179, "setup_item_data" },
    { 0x7186, "setup_mall" },
    { 0x7194, "setup_neighborhood" },
    { 0x7196, "setup_object_friction_mass" },
    { 0x7197, "setup_objectives" },
    { 0x719C, "setup_office_enemy_vo" },
    { 0x71AF, "setup_retreat_vols" },
    { 0x71B1, "setup_rooftop_dof" },
    { 0x71BC, "setup_scientist" },
    { 0x71C5, "setup_shadow_kill" },
    { 0x71D4, "setup_traverse" },
    { 0x71D9, "setup_tunnel" },
    { 0x71E8, "setup_woods" },
    { 0x71EB, "setupapproachnode" },
    { 0x71F1, "setupcqbpointsofinterest" },
    { 0x71F2, "setupdamageflags" },
    { 0x71F6, "setupdoor" },
    { 0x71FA, "setupforrandomspawn" },
    { 0x7201, "setupmovement" },
    { 0x7202, "setuppaths" },
    { 0x721D, "sfx_airlock_door" },
    { 0x721F, "sfx_ally_ally_grapple" },
    { 0x7220, "sfx_ally_grab_script" },
    { 0x7223, "sfx_ally_plr_grapple_loop" },
    { 0x722F, "sfx_ascend_wind_last" },
    { 0x723F, "sfx_blackice_catwalk_collapse" },
    { 0x7244, "sfx_blackice_derrick_exp6_ss" },
    { 0x7245, "sfx_blackice_door_rollup" },
    { 0x7247, "sfx_blackice_engine_beam_fall" },
    { 0x7248, "sfx_blackice_engine_dist_explo" },
    { 0x7266, "sfx_cmd_console_acknowledge" },
    { 0x7269, "sfx_cmd_seq_end" },
    { 0x7270, "sfx_controlroom_explosions" },
    { 0x7278, "sfx_delete_refinery_alarm_node" },
    { 0x727C, "sfx_derrick_mix_change" },
    { 0x7282, "sfx_distant_oil_rig" },
    { 0x7283, "sfx_drill_destroyed" },
    { 0x7285, "sfx_drill_offline" },
    { 0x7286, "sfx_drill_offline_beep" },
    { 0x7289, "sfx_drill_plant" },
    { 0x728F, "sfx_escape_destruction_fire_puffs" },
    { 0x729F, "sfx_fire_tower_triggers" },
    { 0x72A2, "sfx_flare_stack_arm_expl" },
    { 0x72A5, "sfx_flarestack_door_open" },
    { 0x72B5, "sfx_gas_line_scene_plr_hit" },
    { 0x72B9, "sfx_gen4_fire_puff" },
    { 0x72BB, "sfx_gun_grab_script" },
    { 0x72C4, "sfx_heli_flyaway_cmd_center" },
    { 0x72CC, "sfx_heli_rooftops_engine" },
    { 0x72D3, "sfx_heli_turret_fire_start" },
    { 0x72D6, "sfx_heli_turret_shells_stop" },
    { 0x72E3, "sfx_intro_heli_engine" },
    { 0x72E9, "sfx_jet_passby_01" },
    { 0x72EA, "sfx_jet_passby_02" },
    { 0x72EC, "sfx_jet_passby_04" },
    { 0x72ED, "sfx_jet_passby_05" },
    { 0x72F5, "sfx_kyra_hatch" },
    { 0x7301, "sfx_last_pass" },
    { 0x7305, "sfx_lever_logic" },
    { 0x7307, "sfx_lever_rumbles" },
    { 0x730A, "sfx_loco_breach_02" },
    { 0x730B, "sfx_loco_breach_out" },
    { 0x730E, "sfx_loki_control_room_start" },
    { 0x7310, "sfx_long_pipe_bursts" },
    { 0x731E, "sfx_noticket" },
    { 0x7320, "sfx_odin_decompress" },
    { 0x7321, "sfx_odin_decompress_explode" },
    { 0x7325, "sfx_odin_spin_explosion" },
    { 0x7326, "sfx_odin_spinup" },
    { 0x732B, "sfx_plane_crash_script" },
    { 0x732F, "sfx_play_pipe_sounds" },
    { 0x733D, "sfx_plr_vault" },
    { 0x733E, "sfx_post_infil_door" },
    { 0x734A, "sfx_rescue_heli_flyin" },
    { 0x7353, "sfx_rooftop_collapse" },
    { 0x735B, "sfx_sat_approach_kyra" },
    { 0x735D, "sfx_sat_big_fire_lp" },
    { 0x7362, "sfx_satellite_lat_thruster_loop" },
    { 0x736C, "sfx_set_rog_amb" },
    { 0x736F, "sfx_shaky_camera_moment" },
    { 0x7383, "sfx_stereo_quake" },
    { 0x7385, "sfx_stop_alley_water" },
    { 0x7386, "sfx_stop_ascend_sound_wait" },
    { 0x7397, "sfx_train_derail_sound" },
    { 0x739A, "sfx_truck_sinking" },
    { 0x73B0, "sfxcannonfireloop_1p" },
    { 0x73B8, "sfxmissilefire_3p" },
    { 0x73BB, "shader_fade_out" },
    { 0x73D0, "shadowkill_phone_on" },
    { 0x73D1, "shadowkill_phone_show" },
    { 0x73D5, "shakemag" },
    { 0x73D6, "shaky_moment" },
    { 0x73D8, "shared_turrets" },
    { 0x73DA, "shark_attack_2_fx" },
    { 0x73DF, "shark_collision_model" },
    { 0x73E2, "shark_eat_wandering_player" },
    { 0x73E5, "shark_heartbeat" },
    { 0x73E7, "shark_kill_front" },
    { 0x73EF, "shark_room_faster_hint" },
    { 0x73F0, "shark_think" },
    { 0x740C, "shipartilleryspots" },
    { 0x7423, "shoot_from_ai_to_ai" },
    { 0x742F, "shootanimtime" },
    { 0x7433, "shootenemytarget_bullets" },
    { 0x743A, "shootentvelocity" },
    { 0x743E, "shooter_tracking" },
    { 0x7441, "shootflares" },
    { 0x744C, "shootuntilneedtoturn" },
    { 0x7456, "shot" },
    { 0x7461, "shotgunswitchfinish" },
    { 0x7472, "should_do_immediate_ragdoll" },
    { 0x7477, "should_explode_on_damage" },
    { 0x7480, "should_make_entity_sentient" },
    { 0x7488, "should_put_drill_outline_on" },
    { 0x7489, "should_put_player_outline_on" },
    { 0x748F, "should_stabilize" },
    { 0x7495, "should_toggle_silencer" },
    { 0x74A3, "shouldconserveammotime" },
    { 0x74A6, "shoulddoarrivals" },
    { 0x74AC, "shoulddostopanim" },
    { 0x74B1, "shouldfirewhilechangingpose" },
    { 0x74B4, "shouldhelpadvancingteammate" },
    { 0x74BB, "shouldpreventearlyuse" },
    { 0x74BE, "shouldreturntocover" },
    { 0x74C3, "shouldsniff" },
    { 0x74D6, "show_binoc_scan_hint" },
    { 0x74D7, "show_challenge_outcome" },
    { 0x74DD, "show_ent" },
    { 0x74DE, "show_entity" },
    { 0x74E2, "show_fire_hide_aim_idle" },
    { 0x74E3, "show_garage_debris" },
    { 0x74EB, "show_icon" },
    { 0x74F0, "show_legs" },
    { 0x74F3, "show_missile_launcher_collision" },
    { 0x74FC, "show_train_geo" },
    { 0x74FE, "show_trap_icon_to_team" },
    { 0x7503, "show_weapon_switch_hints" },
    { 0x7505, "showcapturedbaseeffecttoplayer" },
    { 0x7507, "showclosedportalfx" },
    { 0x7508, "showcratesplash" },
    { 0x750C, "showdefendprompt" },
    { 0x7511, "showfx" },
    { 0x7518, "showlastenemysightpos" },
    { 0x7519, "showline" },
    { 0x751B, "showmainmenuforteam" },
    { 0x7521, "showspawnnotifies" },
    { 0x7522, "showspawnpoint" },
    { 0x752C, "shufflekillstreaksdown" },
    { 0x7531, "shut_post_infil_door_flag" },
    { 0x7538, "shuttle_docking" },
    { 0x753C, "shuttle_thrust_fl" },
    { 0x7541, "side_arm_array" },
    { 0x7545, "sideisleftright" },
    { 0x7547, "sidesteprate" },
    { 0x7550, "sign" },
    { 0x7551, "silent_magic_bullet_windshield" },
    { 0x7557, "simulation_speed" },
    { 0x7558, "single_ally_run" },
    { 0x755B, "single_fish_detectdamage" },
    { 0x7562, "single_fish_start" },
    { 0x7563, "single_fish_start_after_frameend" },
    { 0x7568, "sinkingplatform_start" },
    { 0x7574, "sittag_offset" },
    { 0x7576, "sitting_wounded_scene_sound" },
    { 0x757A, "skip_start_transition" },
    { 0x757B, "skip_to_end" },
    { 0x7583, "skiplivesxpscalar" },
    { 0x7584, "skipmodelswapdeath" },
    { 0x758A, "sky_change" },
    { 0x758D, "skybridge_ally_cross" },
    { 0x75A0, "skybridge_player_flinch" },
    { 0x75A1, "skybridge_precursor_emitter" },
    { 0x75A3, "skybridge_rumble_logic" },
    { 0x75AA, "skybridge_to_rooftops_transition" },
    { 0x75B8, "skydomeradiosity_ng" },
    { 0x75C0, "skyway_beach_fade_to_dev_logo_credits" },
    { 0x75C5, "skyway_beach_rorke_again" },
    { 0x75C6, "skyway_checkmate_music" },
    { 0x75C9, "skyway_endshot_sfx" },
    { 0x75CD, "skyway_hud_ammocounterhide" },
    { 0x75D2, "skyway_introscreen" },
    { 0x75DD, "sleeping_enemy_below_alerted_or_killed" },
    { 0x75DF, "sleeping_enemy_below_vo" },
    { 0x75F1, "sliding_jet2" },
    { 0x75F3, "sliding_jet21" },
    { 0x75F7, "slomo_breach_vision_change" },
    { 0x75FA, "slomobasevision" },
    { 0x7606, "slow_intro_pre_load" },
    { 0x7608, "slowconvergence" },
    { 0x760C, "slowmo_breach_damage_trigger_think" },
    { 0x760D, "slowmo_breach_disable_stancemod" },
    { 0x7610, "slowmo_breach_start_delay" },
    { 0x7614, "slowmo_lerp_in" },
    { 0x7623, "small_shakes_on" },
    { 0x762A, "smart_get_nag_line" },
    { 0x762B, "smart_radio_dialogue" },
    { 0x763B, "smoke_thrown" },
    { 0x7647, "snake_cam_enemy_anims" },
    { 0x7649, "snake_cam_hud" },
    { 0x7652, "snake_cam_vision_flicker" },
    { 0x7659, "sniff_mode" },
    { 0x765B, "sniff_system_init" },
    { 0x7660, "sniper_cancel_flag" },
    { 0x7667, "sniper_paces" },
    { 0x766A, "sniper_vision_override" },
    { 0x766B, "sniper_wait_extra_black" },
    { 0x766C, "sniper_zoom_hint_hud" },
    { 0x767C, "snowmobile_decide_shoot" },
    { 0x767D, "snowmobile_decide_shoot_internal" },
    { 0x7684, "snowmobile_loop_driver" },
    { 0x7685, "snowmobile_loop_driver_shooting" },
    { 0x768B, "snowmobile_reload" },
    { 0x768C, "snowmobile_reload_internal" },
    { 0x76A4, "soccer_scene_enemies" },
    { 0x76B1, "softlandingwaiter" },
    { 0x76B3, "solar_array0" },
    { 0x76B5, "solar_panel_handling" },
    { 0x76B9, "solarpanels_damage_think_instant" },
    { 0x76C3, "sonar_boat_audio" },
    { 0x76C8, "sonar_boat_e3" },
    { 0x76D7, "sonar_mines" },
    { 0x76E0, "sonar_ping_light_think" },
    { 0x76E4, "sonar_times_hit" },
    { 0x76E5, "sonar_wreck" },
    { 0x76E6, "sonar_wreck_crash_after" },
    { 0x76F9, "sorter" },
    { 0x76FA, "sortlowermessages" },
    { 0x7704, "sound_effect" },
    { 0x7707, "sound_ents" },
    { 0x770A, "sound_fade_in" },
    { 0x770F, "sound_org_four" },
    { 0x7711, "sound_ping_plr" },
    { 0x7719, "sound_torpedo_ent" },
    { 0x7721, "sounds" },
    { 0x772C, "space_accel" },
    { 0x7732, "space_breach_dialogue" },
    { 0x7733, "space_breach_enemies" },
    { 0x7735, "space_breathing_enabled" },
    { 0x773D, "space_hud_enable" },
    { 0x7750, "sparrow_control" },
    { 0x7752, "sparrow_fire_hint" },
    { 0x775B, "sparrow_run_enemy_logic" },
    { 0x7775, "spawn_ally" },
    { 0x777B, "spawn_and_reinforce" },
    { 0x7799, "spawn_convoy" },
    { 0x779B, "spawn_dam_harrassers" },
    { 0x779D, "spawn_dead_bodies_mudpumps" },
    { 0x77B3, "spawn_event_delayed_wave_spawn" },
    { 0x77B8, "spawn_event_per_alien_activation_increase" },
    { 0x77BD, "spawn_events" },
    { 0x77CC, "spawn_grenade" },
    { 0x77D0, "spawn_group" },
    { 0x77D3, "spawn_gunboat" },
    { 0x77D4, "spawn_guys_until_death_or_no_count" },
    { 0x77D6, "spawn_heroes_checkpoint" },
    { 0x77DF, "spawn_infantry_friends" },
    { 0x77E0, "spawn_infantry_in_blackhawk" },
    { 0x77E1, "spawn_initial_combat_helis" },
    { 0x77EB, "spawn_left" },
    { 0x77FB, "spawn_model_on_me" },
    { 0x77FF, "spawn_node_info" },
    { 0x7807, "spawn_odin_actor_single" },
    { 0x7811, "spawn_random_airdrop_sub_items" },
    { 0x7818, "spawn_rorke_inverted_kill_knife" },
    { 0x781C, "spawn_scaffold_guys" },
    { 0x781D, "spawn_scene_alien" },
    { 0x7820, "spawn_setcharacter" },
    { 0x7825, "spawn_single_stage3_ground_vehicle" },
    { 0x7829, "spawn_space_ais_from_targetname" },
    { 0x782C, "spawn_stage1_bombing_runs" },
    { 0x782E, "spawn_stage2_attack_ground_vehicles" },
    { 0x7832, "spawn_stair_runner" },
    { 0x7849, "spawn_type_vo_monitor" },
    { 0x7850, "spawn_vehicles_from_targetname" },
    { 0x785C, "spawn_zone_exists" },
    { 0x7860, "spawnallypet" },
    { 0x7861, "spawnarmor" },
    { 0x786D, "spawnendofgame" },
    { 0x786E, "spawner_cleanup" },
    { 0x7876, "spawner_random_team" },
    { 0x7888, "spawnfunc_enemies_noticket" },
    { 0x788A, "spawnfunc_enemies_rog" },
    { 0x7891, "spawnfunc_enemy_plan_b" },
    { 0x7892, "spawnfunc_enemy_right" },
    { 0x78AB, "spawnlogicteam" },
    { 0x78B2, "spawnorigin" },
    { 0x78B5, "spawnplayer" },
    { 0x78B6, "spawnpointarray" },
    { 0x78B9, "spawnpoints" },
    { 0x78BB, "spawnpointupdate" },
    { 0x78BC, "spawnpos" },
    { 0x78C0, "spawnsolarpanelsinit" },
    { 0x78C1, "spawnspectator" },
    { 0x78C5, "spawntracelocation" },
    { 0x78C7, "spawnwaypointfriendlies" },
    { 0x78CC, "spec_cg_fireworks_low" },
    { 0x78D3, "special_ammocount" },
    { 0x78DA, "special_ammocount_rank" },
    { 0x78E6, "specialcheck" },
    { 0x78F1, "specialpain" },
    { 0x78F3, "specialreloadanimfunc" },
    { 0x78F7, "species" },
    { 0x78FE, "speechcommands" },
    { 0x790B, "spin_allowance_y" },
    { 0x790D, "spin_ally_logic" },
    { 0x7910, "spin_colliders_go" },
    { 0x7913, "spin_dialogue" },
    { 0x7914, "spin_do_moving_debris" },
    { 0x7915, "spin_do_moving_debris_fx" },
    { 0x791B, "spin_piece_sparks" },
    { 0x7930, "spit_target_location" },
    { 0x7931, "spit_type" },
    { 0x793D, "spitters_against_players_ratio" },
    { 0x794A, "splash_z" },
    { 0x794F, "splashqueue" },
    { 0x7951, "spline" },
    { 0x7955, "spline_min_progress" },
    { 0x7959, "splitarrivalsright" },
    { 0x795C, "splitscreen" },
    { 0x795E, "splshtag11" },
    { 0x796E, "spotlight_loop" },
    { 0x7970, "spotlight_tag_origin_cleanup" },
    { 0x7973, "spotted_an_enemy" },
    { 0x7974, "spotted_enemy" },
    { 0x797A, "sprinkler_fx" },
    { 0x797C, "sprinkler_watch" },
    { 0x7986, "squad" },
    { 0x7987, "squad_kill" },
    { 0x7991, "squadflavorbursttransmissions" },
    { 0x799A, "squadofficerwaiter" },
    { 0x79A0, "squadupdatefuncs" },
    { 0x79B3, "stairwell_crack_flat" },
    { 0x79BB, "stance" },
    { 0x79C1, "stance_handler" },
    { 0x79C5, "stances" },
    { 0x79CA, "standardspeed" },
    { 0x79D1, "standing" },
    { 0x79D5, "standrun_begin" },
    { 0x79D7, "standrun_checkreload" },
    { 0x79DC, "standruntostand" },
    { 0x79DD, "standruntranstime" },
    { 0x79E3, "standwalk_begin" },
    { 0x79F9, "start_ascend_time" },
    { 0x79FA, "start_ascent" },
    { 0x79FB, "start_atrium_combat" },
    { 0x79FD, "start_baker" },
    { 0x79FE, "start_balcony_fall" },
    { 0x7A08, "start_canyon" },
    { 0x7A09, "start_canyon_combat" },
    { 0x7A0A, "start_catwalk_snow" },
    { 0x7A0C, "start_catwalks_end" },
    { 0x7A0E, "start_chaos_a" },
    { 0x7A0F, "start_chaos_b" },
    { 0x7A11, "start_chase_dog" },
    { 0x7A12, "start_chopper_test" },
    { 0x7A14, "start_church_destruction" },
    { 0x7A18, "start_combat_after_seeing_launcher" },
    { 0x7A1A, "start_command_outside" },
    { 0x7A1D, "start_coverheight_water_swap" },
    { 0x7A2D, "start_end_beach" },
    { 0x7A30, "start_end_swim" },
    { 0x7A3B, "start_fuel_explosion_fx" },
    { 0x7A43, "start_hangar" },
    { 0x7A46, "start_hdrsuncolorintensity" },
    { 0x7A47, "start_hesh" },
    { 0x7A58, "start_loco_standoff_nomove" },
    { 0x7A59, "start_mall" },
    { 0x7A5A, "start_mansion" },
    { 0x7A6C, "start_odin_satellite" },
    { 0x7A6D, "start_odin_spin" },
    { 0x7A6E, "start_off_running" },
    { 0x7A76, "start_pickup" },
    { 0x7A77, "start_pipe_deck" },
    { 0x7A79, "start_point" },
    { 0x7A7E, "start_post_crater_dog" },
    { 0x7A7F, "start_post_crater_house" },
    { 0x7A8A, "start_sat" },
    { 0x7A8F, "start_sfx_dam_siren_ext" },
    { 0x7A94, "start_small_wreck" },
    { 0x7A9A, "start_strafe_1" },
    { 0x7A9C, "start_street_sequence" },
    { 0x7A9E, "start_struggle_spin" },
    { 0x7AA0, "start_suncolor" },
    { 0x7AAA, "start_test_1" },
    { 0x7AAE, "start_test_area_a" },
    { 0x7AB2, "start_timed_monitor_fx" },
    { 0x7AB3, "start_to_end_length" },
    { 0x7ABB, "start_vo_system" },
    { 0x7AD2, "startfireandaimidlethread" },
    { 0x7AD8, "starthelipilot" },
    { 0x7ADA, "startingbunkerheli" },
    { 0x7ADD, "startjeep" },
    { 0x7AE0, "startlbsupport" },
    { 0x7AE7, "startnode" },
    { 0x7AE9, "startnpcbombusesound" },
    { 0x7B00, "startwatervisuals" },
    { 0x7B0D, "state" },
    { 0x7B17, "statgetbuffered" },
    { 0x7B1A, "static" },
    { 0x7B1B, "static_damping_factor" },
    { 0x7B32, "stealth_1_encounter" },
    { 0x7B35, "stealth_1_zodiac_sounds" },
    { 0x7B4A, "stealth_ai_reach_idle_and_react_proc" },
    { 0x7B4C, "stealth_alert_level_duration" },
    { 0x7B58, "stealth_behavior_system_init" },
    { 0x7B5A, "stealth_blockers" },
    { 0x7B5C, "stealth_broken_flag" },
    { 0x7B63, "stealth_corpse_default_collect_func" },
    { 0x7B69, "stealth_corpse_forget_time_default" },
    { 0x7B84, "stealth_event_anim_defaults" },
    { 0x7B8A, "stealth_event_mod_all" },
    { 0x7B9A, "stealth_group_corpse_flag_waitopen" },
    { 0x7B9E, "stealth_group_return_groups_with_corpse_flag" },
    { 0x7BA0, "stealth_group_return_groups_with_spotted_flag" },
    { 0x7BA8, "stealth_init" },
    { 0x7BAE, "stealth_kill_01_enemy1" },
    { 0x7BB0, "stealth_kill_02_ally" },
    { 0x7BB5, "stealth_photocopier" },
    { 0x7BB7, "stealth_plugin_accuracy" },
    { 0x7BBD, "stealth_plugin_event_explosion" },
    { 0x7BC4, "stealth_pre_spotted_function_default" },
    { 0x7BC5, "stealth_range_trigger" },
    { 0x7BC6, "stealth_satellite_guys" },
    { 0x7BDC, "stealth_threat_behavior_replace" },
    { 0x7BE3, "stealthbombfx" },
    { 0x7BE4, "stealthed_stream_vo" },
    { 0x7BE5, "stealthnewenemyreactanim" },
    { 0x7BE8, "steering_maxdelta" },
    { 0x7BEF, "stern_corner_dmg" },
    { 0x7BF3, "sticky_grenade_04" },
    { 0x7BF6, "sticky_grenade_07" },
    { 0x7BFB, "stingerfxid" },
    { 0x7C0D, "stop_anim_direction" },
    { 0x7C17, "stop_chopper_bosses" },
    { 0x7C18, "stop_chopper_lean" },
    { 0x7C1C, "stop_drilling_sounds" },
    { 0x7C36, "stop_loopsound" },
    { 0x7C3D, "stop_music_jg" },
    { 0x7C3F, "stop_path_on_damage" },
    { 0x7C40, "stop_patrol_vo" },
    { 0x7C43, "stop_player_scuba" },
    { 0x7C44, "stop_player_space" },
    { 0x7C4B, "stop_sfx_dam_siren_int" },
    { 0x7C4F, "stop_spotlight_killed" },
    { 0x7C50, "stop_spotlight_provoked" },
    { 0x7C57, "stop_vo_on_spotted" },
    { 0x7C63, "stopguy3" },
    { 0x7C65, "stopidlesound" },
    { 0x7C66, "stoplocationselection" },
    { 0x7C6C, "stoponback" },
    { 0x7C7A, "stopusingturretwhennodelost" },
    { 0x7C7E, "store_weapons_status" },
    { 0x7C7F, "storecompletedchallenge" },
    { 0x7C86, "str" },
    { 0x7C8B, "strafecooldown" },
    { 0x7C9F, "stream_enemy_logic" },
    { 0x7CBA, "streets_script_vehicle_cleanup" },
    { 0x7CDF, "string_is_single_digit_integer" },
    { 0x7CE1, "string_starts_with" },
    { 0x7CE2, "string_to_bool" },
    { 0x7CE5, "stringcannotplace" },
    { 0x7CE6, "stringfix" },
    { 0x7CED, "strobe_off_hide_hint" },
    { 0x7CEE, "strobe_on_hide_hint" },
    { 0x7CF0, "strobelight" },
    { 0x7CF4, "struct" },
    { 0x7CF6, "struct_arrayspawn" },
    { 0x7CFC, "structarray_remove" },
    { 0x7CFE, "structarray_remove_undefined" },
    { 0x7D01, "structarray_swaptolast" },
    { 0x7D0F, "stuckenemyentity" },
    { 0x7D12, "stumble_and_quake" },
    { 0x7D21, "stylized_line" },
    { 0x7D36, "sun_angles_intro" },
    { 0x7D3C, "sun_manage" },
    { 0x7D44, "sundir_start" },
    { 0x7D48, "sunflare" },
    { 0x7D55, "sunred_start" },
    { 0x7D57, "sunsamplesizenear_ng" },
    { 0x7D68, "suppressingfiretracking" },
    { 0x7D79, "survive" },
    { 0x7D83, "sus_fd" },
    { 0x7D86, "sus_fr" },
    { 0x7D92, "swap_goal_volumes" },
    { 0x7D93, "swap_head" },
    { 0x7D98, "swap_to_overcast_sky" },
    { 0x7D9E, "sweep_tells_vehicles_to_get_off_path" },
    { 0x7D9F, "swept" },
    { 0x7DAA, "swept_end" },
    { 0x7DB2, "swept_path_anim" },
    { 0x7DB5, "swept_plunge_01" },
    { 0x7DDB, "swim_getanimstartpos" },
    { 0x7DE6, "swim_moveend" },
    { 0x7DEA, "swim_player_surface_anim" },
    { 0x7DFA, "swim_track_set" },
    { 0x7DFD, "swim_truck_surface_anim" },
    { 0x7DFF, "swim_updateleananim" },
    { 0x7E00, "swim_updatestrafeaimanim" },
    { 0x7E01, "swim_updatestrafeanim" },
    { 0x7E04, "swim_waitforapproachpos" },
    { 0x7E13, "switch_node_on_flag" },
    { 0x7E1F, "switchmsg" },
    { 0x7E24, "sync_carry_walk_anims" },
    { 0x7E2D, "synch_directions" },
    { 0x7E30, "syncnotetrackent" },
    { 0x7E34, "system_default_event_distances" },
    { 0x7E38, "system_init_shadows" },
    { 0x7E41, "systems_down_end" },
    { 0x7E51, "table_getweaponbuff" },
    { 0x7E53, "table_getweaponreticle" },
    { 0x7E55, "tablet" },
    { 0x7E58, "tablet_prop" },
    { 0x7E7A, "take_cover_warning" },
    { 0x7E7B, "take_cover_warning_loop" },
    { 0x7E7D, "take_fire_tracking" },
    { 0x7E8C, "takeover_flash" },
    { 0x7E91, "talk_for_time" },
    { 0x7E95, "tall_grass_friendly_exit_logic" },
    { 0x7E9D, "tall_grass_guys_went_hot" },
    { 0x7EAD, "tank_ambient_waits" },
    { 0x7EBE, "tank_damage_allies" },
    { 0x7EC0, "tank_death_allies" },
    { 0x7ECB, "tank_drop_slide_allies" },
    { 0x7ECC, "tank_dropmines" },
    { 0x7ED0, "tank_event" },
    { 0x7ED1, "tank_fire_at_enemies" },
    { 0x7EDA, "tank_handledeath" },
    { 0x7EE0, "tank_hud" },
    { 0x7EE5, "tank_hud_crack_left" },
    { 0x7EF1, "tank_hud_vignette" },
    { 0x7F03, "tank_rumble" },
    { 0x7F04, "tank_save" },
    { 0x7F08, "tank_setinactive" },
    { 0x7F1E, "tank_watchlowhealth" },
    { 0x7F27, "tankfire_spline_jeep" },
    { 0x7F29, "tankflash" },
    { 0x7F39, "tankupdate" },
    { 0x7F3F, "targertname" },
    { 0x7F46, "target_ent_cleanup" },
    { 0x7F47, "target_entity" },
    { 0x7F4C, "target_min_range" },
    { 0x7F55, "target_trace_to_owners_eyes" },
    { 0x7F57, "target_unset_islockedon" },
    { 0x7F63, "targetingoffset" },
    { 0x7F6A, "targetpointtooclose" },
    { 0x7F78, "team1" },
    { 0x7F84, "team_hive_performance" },
    { 0x7F93, "teamflashbangimmunity" },
    { 0x7F94, "teamheadicon" },
    { 0x7F97, "teamkilldelay" },
    { 0x7FA3, "teamprogressbartexty" },
    { 0x7FA4, "teamprogressbarwidth" },
    { 0x7FA6, "teams" },
    { 0x7FB0, "teargas_cough_vo" },
    { 0x7FB1, "teargas_coughing" },
    { 0x7FB9, "teargas_loaded" },
    { 0x7FBF, "teargas_vo" },
    { 0x7FC5, "teleport_add_delta" },
    { 0x7FD8, "teleport_failsafe" },
    { 0x7FDD, "teleport_get_care_packages" },
    { 0x7FE0, "teleport_get_safe_node_near" },
    { 0x7FE4, "teleport_include_killstreaks" },
    { 0x7FE8, "teleport_is_valid_zone" },
    { 0x7FEC, "teleport_notify_death" },
    { 0x7FF6, "teleport_onteleportgrind" },
    { 0x8004, "teleport_pre_onstartgamesd_and_sr" },
    { 0x8007, "teleport_self_add_delta" },
    { 0x8016, "teleport_to_zone_agents" },
    { 0x8017, "teleport_to_zone_character" },
    { 0x801C, "teleport_zone_current" },
    { 0x8027, "tell_player_to_stay" },
    { 0x8043, "test_attack_heli" },
    { 0x8045, "test_connect_paths" },
    { 0x8048, "test_guy_do_goalor_die" },
    { 0x8049, "test_individual_spawner" },
    { 0x8057, "testingapache_animations" },
    { 0x805B, "text1" },
    { 0x8065, "text_titlefadeout" },
    { 0x807E, "thermite2" },
    { 0x808E, "thorough_delete" },
    { 0x8097, "threatcallouttracking" },
    { 0x80AA, "throw_stones_idle" },
    { 0x80B1, "throwingknifeused" },
    { 0x80B2, "thrown_entities" },
    { 0x80B3, "thrown_semtex_grenades" },
    { 0x80C7, "thruster_timer_logic" },
    { 0x80CB, "thunder_big_sound_moving" },
    { 0x80E0, "tilt_exfil_props" },
    { 0x80E1, "tilt_exp_heli_clip" },
    { 0x80E3, "tilt_front_deck_impact" },
    { 0x80E8, "tilt_handler" },
    { 0x80EF, "tilt_osprey_1" },
    { 0x80F4, "tilt_player_stumble" },
    { 0x80F5, "tilt_player_vault" },
    { 0x8109, "tilt_run_forward_monitor" },
    { 0x8117, "timebetweenshots" },
    { 0x8118, "timecheck" },
    { 0x8121, "timelimitthread" },
    { 0x8123, "timeofnextsound" },
    { 0x8124, "timeout" },
    { 0x8126, "timeout_watch" },
    { 0x8128, "timeoutregenfaster" },
    { 0x8129, "timepaused" },
    { 0x812F, "timer_ten_change" },
    { 0x8131, "timername" },
    { 0x8132, "timerpausetime" },
    { 0x8136, "timescale" },
    { 0x8153, "toggle_zoom" },
    { 0x8158, "too_close_to_boat" },
    { 0x815A, "took_damage" },
    { 0x8161, "topattackpasses" },
    { 0x817C, "total_runner_drone_spawn_count" },
    { 0x8186, "totalscore" },
    { 0x8188, "touched" },
    { 0x8193, "tower_a10_flyby" },
    { 0x81A7, "tower_door_pitchfx" },
    { 0x81B0, "tower_entrance_stumblers" },
    { 0x81B5, "tower_hesh_help_wounded" },
    { 0x81BD, "tower_nh90_guys" },
    { 0x81BE, "tower_pickup_b" },
    { 0x81BF, "tower_radio" },
    { 0x81C8, "tower_retreat_mortars_hit" },
    { 0x81CF, "tower_vista_retreat_tank" },
    { 0x81D8, "toy_tvs_flatscreen_sturdy" },
    { 0x81DC, "trace_part_for_efx" },
    { 0x81DE, "trace_to_forward" },
    { 0x81ED, "track_dud" },
    { 0x81F9, "track_odin_down" },
    { 0x8219, "trackingweaponheadshots" },
    { 0x8224, "trackloop_cqbshootpos" },
    { 0x822A, "trackshootentorpos" },
    { 0x8239, "train_nag" },
    { 0x8240, "train_pathing" },
    { 0x824A, "train_tele_trig_proc" },
    { 0x824B, "train_teleport" },
    { 0x8250, "trains_move" },
    { 0x8252, "trajectoryactive" },
    { 0x826E, "transitiontime" },
    { 0x8277, "translate_off_edge" },
    { 0x827E, "trap_cage_init" },
    { 0x827F, "trap_can_player_trigger" },
    { 0x8282, "trap_damage_scalar" },
    { 0x8287, "trap_explosive_init" },
    { 0x8289, "trap_get_launch_dir" },
    { 0x828E, "trap_get_targets" },
    { 0x8290, "trap_get_unset_use_text" },
    { 0x829A, "trap_launcher_init" },
    { 0x829C, "trap_launcher_reset" },
    { 0x82AD, "trap_set_can_use" },
    { 0x82B1, "trap_state_funcs" },
    { 0x82BA, "trap_use_trigger_init" },
    { 0x82C5, "travel_view_fx" },
    { 0x82D5, "traversedeath" },
    { 0x82DA, "traverseinfo" },
    { 0x82E6, "treadfx_override_nh90" },
    { 0x82F5, "trench_beach_allies_ambient" },
    { 0x82F6, "trench_beach_allies_ambient_off" },
    { 0x82FA, "trench_boat_think" },
    { 0x82FC, "trench_chargers_think" },
    { 0x8302, "trench_friendly_orange_guy" },
    { 0x830F, "trench_spawn_functions" },
    { 0x8325, "trig_stairs_setup" },
    { 0x8330, "trigger_battlechatter" },
    { 0x8333, "trigger_createart_transient" },
    { 0x833B, "trigger_disable_on_jump" },
    { 0x833C, "trigger_door" },
    { 0x833F, "trigger_enemy_spawn" },
    { 0x8347, "trigger_flag_set_specialops" },
    { 0x8348, "trigger_flag_set_specialops_clear" },
    { 0x834F, "trigger_functions" },
    { 0x8351, "trigger_glass_break" },
    { 0x8352, "trigger_group" },
    { 0x835D, "trigger_int_building_hits" },
    { 0x836D, "trigger_multiple_depthoffield" },
    { 0x836F, "trigger_multiple_fx_trigger_on_think" },
    { 0x8371, "trigger_multiple_fx_volume_off_target" },
    { 0x8373, "trigger_multiple_sunflare" },
    { 0x8380, "trigger_parse_parameters" },
    { 0x839D, "trigger_spawngroup" },
    { 0x83A5, "trigger_turns_off" },
    { 0x83AD, "trigger_wait" },
    { 0x83BB, "triggers" },
    { 0x83DC, "truck_rumble" },
    { 0x83EC, "try_take_player_currency" },
    { 0x8404, "tryorderto" },
    { 0x8409, "tryuse" },
    { 0x8417, "tryuse_dpad_team_ammo" },
    { 0x841B, "tryuse_dpad_team_explosives" },
    { 0x842E, "tryusedeployableammo" },
    { 0x8445, "tryuseperkstreak" },
    { 0x8449, "tryuserefillammo" },
    { 0x8451, "tryusesam" },
    { 0x845B, "tryuseuplink" },
    { 0x845C, "tryusevanguard" },
    { 0x8464, "tunnel_behavior" },
    { 0x8468, "tunnel_door_clip" },
    { 0x8469, "tunnel_door_scene" },
    { 0x847F, "tunnel_tank_crush" },
    { 0x8480, "tunnel_vehicle_think" },
    { 0x8482, "tunnel_vehix_stop_sounds" },
    { 0x848A, "turn_exit_trigger_on" },
    { 0x8491, "turn_off_gas_fx" },
    { 0x8495, "turn_off_nodes" },
    { 0x84A6, "turn_on_gaztiger_underwater_bubble_fx" },
    { 0x84B5, "turnbackon" },
    { 0x84B9, "turnoff" },
    { 0x84C6, "turret_activate" },
    { 0x84D1, "turret_cleanup_on_unload" },
    { 0x84D4, "turret_coolmonitor" },
    { 0x84DA, "turret_deleteme" },
    { 0x84DF, "turret_function" },
    { 0x84EB, "turret_impact" },
    { 0x84F5, "turret_oncarrierchangedteam" },
    { 0x84F7, "turret_oncarrierdisconnect" },
    { 0x8503, "turret_setcarried" },
    { 0x8512, "turret_target_updater" },
    { 0x8521, "turretattachpoint" },
    { 0x8523, "turretdeathanimroot" },
    { 0x852B, "turretinfo" },
    { 0x853F, "tv_changes_intensity" },
    { 0x8544, "tv_play" },
    { 0x8549, "tweak_player_view_do_rotate" },
    { 0x856C, "uav_rig_controller" },
    { 0x856F, "uav_sort" },
    { 0x857C, "ui_action_slot_force_active_off" },
    { 0x857F, "ui_bomb_planting_defusing" },
    { 0x8581, "ui_dom_securing" },
    { 0x858B, "unblock_curtain" },
    { 0x858F, "underwater" },
    { 0x8592, "underwater_combat_amb" },
    { 0x8595, "underwater_melee_kill_achievement_count" },
    { 0x85BC, "unloadque" },
    { 0x85BF, "unlock_flag" },
    { 0x85C6, "unmatched_death_rig_light_waits_for_lights_off" },
    { 0x85C7, "unresolved_collision_count" },
    { 0x85CA, "unresolved_collision_kill" },
    { 0x85D3, "unset_forcegoal" },
    { 0x85D6, "unset_perk" },
    { 0x85DF, "unset_perk_health_level_3" },
    { 0x85E1, "unset_perk_medic_0" },
    { 0x85ED, "unset_perk_pistol_magnum_2" },
    { 0x85F0, "unset_perk_pistol_mp443_0" },
    { 0x85F3, "unset_perk_pistol_mp443_3" },
    { 0x85FC, "unset_perk_rigger_2" },
    { 0x862B, "unsetlightarmor" },
    { 0x8632, "unsetpainted" },
    { 0x8639, "unsetrefillgrenades" },
    { 0x864E, "unstoppable_friendly_fire_shield" },
    { 0x8658, "update_achievement" },
    { 0x8659, "update_achievement_all_players" },
    { 0x8663, "update_bike_player_avoidance" },
    { 0x867A, "update_goal_vol_from_trigger" },
    { 0x867B, "update_heli_crash_location" },
    { 0x8689, "update_melee_spitter" },
    { 0x868A, "update_motion_tracker_speed" },
    { 0x8690, "update_no_reloads" },
    { 0x86AA, "update_shotguns_only" },
    { 0x86AD, "update_soflam_ammocount" },
    { 0x86B2, "update_steering" },
    { 0x86C1, "update_weaponstats_hits" },
    { 0x86CB, "updateareanodes" },
    { 0x86D7, "updatecheckforceragdoll" },
    { 0x86E6, "updateenemyuse" },
    { 0x86E9, "updatefreeplayedtime" },
    { 0x8709, "updateobjectivetext" },
    { 0x8729, "updatesavedlastweapon" },
    { 0x872E, "updatesentrypositionthread" },
    { 0x8734, "updatesppercent" },
    { 0x8739, "updatestreakcount" },
    { 0x873A, "updatestreakslots" },
    { 0x873F, "updateteamplacement" },
    { 0x8741, "updateteamscores" },
    { 0x8748, "updatetimerpausedness" },
    { 0x8757, "updatewinlossstats" },
    { 0x875A, "updateworldicons" },
    { 0x8760, "uplink_override_moving_platform_death" },
    { 0x8767, "upload_virus_anims" },
    { 0x8779, "use_big_splash" },
    { 0x8780, "use_dpad_glsentry" },
    { 0x878C, "use_dpad_team_randombox" },
    { 0x878E, "use_dpad_war_machine" },
    { 0x8799, "use_switch_toggle_multiple" },
    { 0x87A0, "use_trigger" },
    { 0x87A5, "useballdrone" },
    { 0x87BB, "usehardpoint_regularmp" },
    { 0x87BE, "usehintstring" },
    { 0x87C0, "useholdthinkloop" },
    { 0x87E5, "using_hdr_fog" },
    { 0x87EA, "using_remote_turret_when_died" },
    { 0x87EC, "using_string_tables" },
    { 0x87F1, "usingmg" },
    { 0x87F6, "usingprimary" },
    { 0x87F7, "usingremote" },
    { 0x87F9, "usingriflelikeweapon" },
    { 0x8803, "util_derrick_destroy_quick" },
    { 0x882B, "vanguard_missile_radius" },
    { 0x8833, "vanguard_origin" },
    { 0x8834, "vanguard_pick_node" },
    { 0x883C, "vanguard_showreticletoplayer" },
    { 0x8842, "vanguardfiremisslefunc" },
    { 0x8847, "vantage_nag" },
    { 0x886B, "vault_vol_1" },
    { 0x886F, "vault_water" },
    { 0x8884, "vehicle_ai_turret_gunner_ignore_all_until_unload" },
    { 0x8893, "vehicle_attack_missile" },
    { 0x889A, "vehicle_controlling" },
    { 0x88A0, "vehicle_death_anim" },
    { 0x88A6, "vehicle_death_jolt" },
    { 0x88AB, "vehicle_deathmodel_delay" },
    { 0x88AE, "vehicle_deletegroup" },
    { 0x88B0, "vehicle_detourpaths" },
    { 0x88B9, "vehicle_dynamicpath" },
    { 0x88BE, "vehicle_fire_loop" },
    { 0x88C0, "vehicle_frontarmor" },
    { 0x88C9, "vehicle_getinanim_clear" },
    { 0x88CC, "vehicle_getinstart" },
    { 0x88D1, "vehicle_grenadeshield" },
    { 0x88DA, "vehicle_idle_override" },
    { 0x88DF, "vehicle_isstationary" },
    { 0x88ED, "vehicle_liftoff" },
    { 0x88EF, "vehicle_lights" },
    { 0x88F2, "vehicle_lights_off" },
    { 0x88FC, "vehicle_mainturrets" },
    { 0x8905, "vehicle_play_guy_anim" },
    { 0x8910, "vehicle_resume_named" },
    { 0x8916, "vehicle_ridespawners" },
    { 0x8919, "vehicle_rumble_override" },
    { 0x892C, "vehicle_spawner_adjust_health_and_damage" },
    { 0x8931, "vehicle_startmovegroup" },
    { 0x893B, "vehicle_turn_left" },
    { 0x8945, "vehicle_unloadgroups" },
    { 0x8946, "vehicle_unloadwhenattacked" },
    { 0x894C, "vehicle_zpu_can_target" },
    { 0x8955, "vehiclekilled" },
    { 0x8966, "verifydedicatedconfiguration" },
    { 0x898A, "vfx_hostage1_waterboard_stop" },
    { 0x8999, "vfx_torpedo_wings_out" },
    { 0x899A, "vfxarry" },
    { 0x89A0, "victimteam" },
    { 0x89AB, "view_particle_source_alt" },
    { 0x89B5, "viewfx" },
    { 0x89C3, "vignette_drone_spawn" },
    { 0x89D0, "vignette_register_wait" },
    { 0x89D8, "vignette_unignore_everything" },
    { 0x89D9, "vignette_vehicle_delete" },
    { 0x89E2, "vip_enemy_interrupt" },
    { 0x89F6, "visibleto" },
    { 0x89FD, "vision_mudpumps_end" },
    { 0x8A03, "vision_set_fog_progress" },
    { 0x8A04, "vision_set_refinery_visionsets" },
    { 0x8A0B, "visionset_diff" },
    { 0x8A0C, "visiontest" },
    { 0x8A1B, "vista_tilt_model" },
    { 0x8A1C, "vista_tilt_setup" },
    { 0x8A1D, "vista_vehicles" },
    { 0x8A1F, "visual_origin" },
    { 0x8A33, "vo_flag_timer" },
    { 0x8A37, "vo_system" },
    { 0x8A3F, "voice_is_british_based" },
    { 0x8A40, "voicecanburst" },
    { 0x8A46, "vol" },
    { 0x8A57, "vttype" },
    { 0x8A5F, "vulture_savetostruct" },
    { 0x8A62, "vulture_waitfortriggerstop" },
    { 0x8A64, "w_alpha" },
    { 0x8A6C, "wait_endon" },
    { 0x8A6D, "wait_for_ability_use" },
    { 0x8A91, "wait_for_proximity_check_activation" },
    { 0x8AB7, "wait_till_distance_from_enemy" },
    { 0x8AB9, "wait_till_offscreen_then_delete" },
    { 0x8ABC, "wait_till_shot" },
    { 0x8AC3, "wait_until_anim_finishes" },
    { 0x8AC5, "wait_until_enemies_in_volume" },
    { 0x8ADC, "waitforchangeteams" },
    { 0x8ADD, "waitforclassselect" },
    { 0x8AE8, "waitformovemodechange" },
    { 0x8AF6, "waitforstancechange" },
    { 0x8AFA, "waitforstuck" },
    { 0x8AFC, "waitfortimeornotify" },
    { 0x8B04, "waitingforgate" },
    { 0x8B05, "waitingtodeactivate" },
    { 0x8B0B, "waitloadoutdone" },
    { 0x8B17, "waitskipkillcambutton" },
    { 0x8B18, "waitskipkillcambuttonduringdeathtimer" },
    { 0x8B1E, "waittakekillstreakweapon" },
    { 0x8B2D, "waittill_any_in_array_or_timeout_no_endon_death" },
    { 0x8B2F, "waittill_any_in_array_return_no_endon_death" },
    { 0x8B34, "waittill_any_return_no_endon_death" },
    { 0x8B38, "waittill_any_triggered_return_triggerer" },
    { 0x8B39, "waittill_array_dead" },
    { 0x8B3B, "waittill_breach_enemy_dead" },
    { 0x8B3C, "waittill_clockambush_driver_dead" },
    { 0x8B4D, "waittill_death_respawn" },
    { 0x8B5A, "waittill_entity_in_range" },
    { 0x8B61, "waittill_forever" },
    { 0x8B6B, "waittill_knockdown_moment" },
    { 0x8B75, "waittill_my_death" },
    { 0x8B7C, "waittill_notify_or_timeout" },
    { 0x8B92, "waittill_player_not_looking" },
    { 0x8BA1, "waittill_time_out" },
    { 0x8BA6, "waittill_trigger_ent_targetname" },
    { 0x8BA8, "waittill_triggered_current" },
    { 0x8BAD, "waittill_volume_dead_then_set_flag" },
    { 0x8BAE, "waittill_weapon_placed" },
    { 0x8BCB, "waituntilwaverelease" },
    { 0x8BD1, "wakeup_enemies" },
    { 0x8BD5, "wakeup_to_player_distance" },
    { 0x8BE0, "walkout_do_stop_transition_anim" },
    { 0x8BE6, "walkway_collapse_clip_show" },
    { 0x8BF2, "wall_friction_enabled" },
    { 0x8BF4, "wall_friction_trace_dist" },
    { 0x8BF9, "wall_lights" },
    { 0x8C00, "wants_to_fire" },
    { 0x8C2B, "warhawkcustombotkillstreakfunc" },
    { 0x8C35, "warthog" },
    { 0x8C38, "warthog_player_fadeout_crash" },
    { 0x8C62, "watch_damage" },
    { 0x8C6E, "watch_for_ally_see_convoy" },
    { 0x8C80, "watch_for_round_room_combat" },
    { 0x8C82, "watch_for_shockwave_hit" },
    { 0x8C8E, "watch_if_used" },
    { 0x8C99, "watch_nodes_stop" },
    { 0x8C9D, "watch_placed_sentry" },
    { 0x8CA2, "watch_player_escape" },
    { 0x8CAC, "watch_player_wake_scientists" },
    { 0x8CB4, "watch_smoke" },
    { 0x8CB8, "watch_throwing_knife_land" },
    { 0x8CBB, "watch_tv_for_damage" },
    { 0x8CC5, "watchattackstate" },
    { 0x8CD7, "watchdeployablemarkercancel" },
    { 0x8CDD, "watchdvars" },
    { 0x8D07, "watchoffhandcancel" },
    { 0x8D22, "watchsmokeuse" },
    { 0x8D24, "watchstartweaponchange" },
    { 0x8D28, "watchstoprevenge" },
    { 0x8D29, "watchsuppression" },
    { 0x8D34, "watchweaponpickuphorde" },
    { 0x8D52, "water_surface_think" },
    { 0x8D5F, "waterball_main_side_setup" },
    { 0x8D79, "wave_3_charge_spawn_func" },
    { 0x8D7E, "wave_has_delayed_spawn_type" },
    { 0x8D8D, "weap_has_thermal" },
    { 0x8D91, "weapon_change_timer" },
    { 0x8D9A, "weapon_notify_loop" },
    { 0x8DA1, "weaponattachments" },
    { 0x8DC2, "weaponspeed" },
    { 0x8DC9, "weapontagright" },
    { 0x8DD3, "weld_sound_loop_npc" },
    { 0x8DDA, "weldtool" },
    { 0x8DE4, "whisper" },
    { 0x8DE6, "white_overlay" },
    { 0x8DFA, "window_crash" },
    { 0x8E0C, "wing_tag" },
    { 0x8E11, "winner" },
    { 0x8E12, "winners" },
    { 0x8E15, "wipeout" },
    { 0x8E33, "wolfpack_circle_dialogue" },
    { 0x8E44, "world_vault_clip" },
    { 0x8E51, "wreck_hint_up" },
    { 0x8E52, "wreck_jumper_watcher" },
    { 0x8E5B, "write_log" },
    { 0x8E96, "yoff" },
    { 0x8E9C, "yscale" },
    { 0x8E9E, "z_max" },
    { 0x8EA8, "zipline" },
    { 0x8EA9, "zipline_allies_anims" },
    { 0x8EAA, "zipline_allies_vo" },
    { 0x8EB8, "zipline_launcher_setup_anims" },
    { 0x8EBC, "zipline_rope_baker" },
    { 0x8EC1, "zipline_startpoint" },
    { 0x8EC5, "zipline_trolley_rorke" },
    { 0x8ECA, "zodiac_ally_shoot_targets" },
    { 0x8ECC, "zodiac_b" },
    { 0x8ED7, "zodiac_osprey_2_wave" },
    { 0x8EE0, "zodiac_teleport_logic" },
    { 0x8EE7, "zodiacs_defend_vo" },
    { 0x8EE8, "zoffset" },
    { 0x8EE9, "zone" },
    { 0x8EF2, "zoom_lerp_dof" },
}};

struct __init__
{
    __init__()
    {
        static bool init = false;
        if (init) return;
        init = true;

        opcode_map.reserve(opcode_list.size());
        opcode_map_rev.reserve(opcode_list.size());
        function_map.reserve(function_list.size());
        function_map_rev.reserve(function_list.size());
        method_map.reserve(method_list.size());
        method_map_rev.reserve(method_list.size());
        file_map.reserve(file_list.size());
        file_map_rev.reserve(file_list.size());
        token_map.reserve(token_list.size());
        token_map_rev.reserve(token_list.size());

        for (const auto& entry : opcode_list)
        {
            opcode_map.insert({ entry.first, entry.second });
            opcode_map_rev.insert({ entry.second, entry.first });
        }

        for (const auto& entry : function_list)
        {
            function_map.insert({ entry.first, entry.second });
            function_map_rev.insert({ entry.second, entry.first });
        }

        for (const auto& entry : method_list)
        {
            method_map.insert({ entry.first, entry.second });
            method_map_rev.insert({ entry.second, entry.first });
        }

        for (const auto& entry : file_list)
        {
            file_map.insert({ entry.first, entry.second });
            file_map_rev.insert({ entry.second, entry.first });
        }

        for (const auto& entry : token_list)
        {
            token_map.insert({ entry.first, entry.second });
            token_map_rev.insert({ entry.second, entry.first });
        }
    }
};

__init__ _;

} // namespace xsk::gsc::iw6

#ifdef _MSC_VER
#pragma warning(pop)
#endif
